------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developped by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
-----------------------------------------------------------------------------
------------------------------------------------------------------------------



PACKAGE Framework_Config IS


   PRAGMA Elaborate_Body;



   -- Display all framework configuration variables
   --
   PROCEDURE Put;


   ----------------------------------------------------------------------------------------
   --
   -- Constants defined under this comment can be tuned by users to match user
   -- requirements in terms
   -- of systems that Cheddar can analyse and in terms
   -- of available memory/CPU in the user's machine.
   --
   -- WARNING : DO NOT FORGET TO CHANGE config/config.h
   --
   ----------------------------------------------------------------------------------------


   -- Activate Cheddar's debug mode if true
   -- This debug mode allows the user to have some extra display
   --
   Cheddar_Debug : Boolean := False;


   -- Maximum processors, tasks, messages, buffers and resources
   -- which can be defined in a cheddar XML file
   -- Increasing these constants implies increasing cheddar memory footprint
   --
   Max_Processors      : CONSTANT Natural := 10;
   Max_Tasks           : CONSTANT Natural := 200;
   Max_Messages        : CONSTANT Natural := 5;
   Max_Buffers         : CONSTANT Natural := 30;
   Max_Resources       : CONSTANT Natural := 30;
   Max_Networks        : CONSTANT Natural := 5;
   Max_Event_Analyzers : CONSTANT Natural := 5;
   Max_Address_Space   : CONSTANT Natural := 10;


   -- Define the largest scheduling interval time that the
   -- simulator can compute
   -- Increasing this constant implies increasing cheddar CPU usage
   --
   Max_Scheduling_Period : CONSTANT Natural := 450000;


   -- Definition of priority range. We choose the most current value :
   -- 256 priority levels (from 0 to 255)
   -- Zero is the lowest priority level
   ---
   TYPE Priority_Range IS NEW Natural RANGE 0 .. 255;
   Low_Priority  : CONSTANT Priority_Range := Priority_Range'First;
   High_Priority : CONSTANT Priority_Range := Priority_Range'Last;


   -- Maximum accurrancy of big numbers packages
   -- Increase it if you need to run large XML Chedar project file
   -- Increasing this constant implies increasing cheddar memory footprint
   -- Increasing this constant implies increasing cheddar CPU usage
   --
   Max_Big_Number_Size : CONSTANT Natural := 200;


   -- Define the language used in the graphical editor of Cheddar
   -- Default language is "english"
   --
   TYPE Languages IS
         (Francais,
          English);
   Current_Language : Languages := English;


   -- Constants above define maximum number of tasks that
   -- can use a shared resource or a buffer
   -- increasing this constant implies increasing cheddar memory footprint
   --
   Max_Tasks_For_A_Resource : CONSTANT Natural := 10;
   Max_Tasks_For_A_Buffer   : CONSTANT Natural := 10;


   -- Define maximum of variables that can be defined
   -- in parametric scheduler
   -- Increasing this constant implies increasing cheddar memory footprint
   --
   Max_Variables : CONSTANT Natural := 150;


   -- Maximum number of statements in a parametric
   -- scheduler
   -- Increasing this constant implies increasing cheddar memory footprint
   --
   Max_Statements : CONSTANT Natural := 100;


   -- The langage used to defined a parametric scheduler
   -- is block oriented : the constant below
   -- given the maximum depth level of block
   -- Increasing this constant implies increasing cheddar memory footprint
   --
   Max_Block_Level : CONSTANT Natural := 100;



   -- Maximum number of dependencies of a XML Cheddar project file.
   -- Dependencies can be a message/buffer used by two tasks or a precedencies
   -- constraint.
   -- Increasing this constant implies increasing cheddar memory footprint
   --
   Max_Tasks_Dependencies : CONSTANT Natural := 150;


   -- Define maximum of user's defined task parameter
   --
   Max_User_Defined_Task_Parameters : CONSTANT Natural := 12;


   -- Maximum of parameters during a framework call
   --
   Max_Framework_Parameters : CONSTANT Natural := 20;


   -- Define offset constants : a task can contain at most 3 offsets
   --
   Max_Offsets : CONSTANT Natural := 5;


   -- This constant define the size of "xi" of a discrete
   -- density probability
   --
   Max_Density_Size : CONSTANT Natural := 200;


   -- Maximum size of the framework request table
   -- Increasing this constant implies increasing cheddar memory footprint
   --
   Max_Requests : CONSTANT Natural := 20;



   -- Maximul size of a fixed string
   --
   Max_String_Size : CONSTANT Natural := 200;



   -- This constant is used by the parametric simulation engine
   -- It defines size of tables which store simulation data for
   -- all elements of Cheddar project (except time units)
   --
   Max_Simulation : CONSTANT Natural := Natural'Max (Natural'Max (Max_Processors,
     Max_Tasks), Natural'Max (Natural'Max (Max_Messages, Max_Buffers), Natural'Max
     (Max_Resources, Max_Networks)));


   -- Sum of all objects which can be stored in a XML project
   --
   Sum_Simulation : CONSTANT Natural := Max_Processors + Max_Tasks + Max_Messages
     + Max_Resources + Max_Networks;



END Framework_Config;
