------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------





WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Ada.Exceptions;
USE Ada.Exceptions;
WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Time_Units;
USE  Time_Units;
USE  Time_Units.A_Time_Unit;

WITH Scheduler;
USE Scheduler;

WITH Scheduler.Parametrics;
USE Scheduler.Parametrics;
WITH Scheduler.Fixed_Priority.Dm;
USE Scheduler.Fixed_Priority.Dm;
WITH Scheduler.Fixed_Priority.Rm;
USE Scheduler.Fixed_Priority.Rm;
WITH Scheduler.Fixed_Priority.Hpf;
USE Scheduler.Fixed_Priority.Hpf;
WITH Scheduler.Dynamic_Priority.Edf;
USE Scheduler.Dynamic_Priority.Edf;
WITH Scheduler.Dynamic_Priority.Llf;
USE Scheduler.Dynamic_Priority.Llf;
WITH Scheduler.Dynamic_Priority.Muf.Deadline_Based;
USE Scheduler.Dynamic_Priority.Muf.Deadline_Based;
WITH Scheduler.Dynamic_Priority.Muf.Laxity_Based;
USE Scheduler.Dynamic_Priority.Muf.Laxity_Based;
WITH Scheduler.Dynamic_Priority.D_Over;
USE Scheduler.Dynamic_Priority.D_Over;
WITH Scheduler.Time_Sharing_Based_On_Wait_Time;
USE Scheduler.Time_Sharing_Based_On_Wait_Time;
WITH Scheduler.Time_Sharing_Based_On_Cpu_Usage;
USE Scheduler.Time_Sharing_Based_On_Cpu_Usage;
WITH Scheduler.Round_Robin;
USE Scheduler.Round_Robin;
WITH Scheduler.Nothing;
USE Scheduler.Nothing;

WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;
WITH Tasks;
USE Tasks;

WITH Resource_Set;
USE Resource_Set;
USE Resource_Set.Generic_Resource_Set;
WITH Resources;
USE Resources;
USE Resources.Task_Lists;

WITH Translate;
USE Translate;

WITH Objects;
USE Objects;

WITH Integer_Util;



PACKAGE BODY Address_Space_Set IS




   PROCEDURE Add_Address_Space (
         My_Address_Spaces : IN OUT Address_Spaces_Set;
         Name              : IN     Unbounded_String;
         Cpu_Name          : IN     Unbounded_String;
         Text_Memory_Size  : IN     Integer;
         Stack_Memory_Size : IN     Integer;
         Data_Memory_Size  : IN     Integer;
         Heap_Memory_Size  : IN     Integer;
         Is_Preemptive     : IN     Preemptive_Type    := Preemptive;
         Quantum           : IN     Integer            := 0;
         File_Name         : IN     Unbounded_String   := Empty_String;
         A_Scheduler       : IN     Scheduler_Type     := No_Scheduling_Protocol) IS

      Dummy : Address_Space_Ptr;
   BEGIN
      Add_Address_Space(My_Address_Spaces, Dummy, Name,  Cpu_Name,
         Text_Memory_Size, Stack_Memory_Size, Data_Memory_Size,
         Heap_Memory_Size, Is_Preemptive, Quantum, File_Name, A_Scheduler);
   END Add_Address_Space;



   PROCEDURE Check_Address_Space (
         My_Address_Spaces : IN     Address_Spaces_Set;
         Name              : IN     Unbounded_String;
         Cpu_Name          : IN     Unbounded_String;
         Text_Memory_Size  : IN     Integer;
         Stack_Memory_Size : IN     Integer;
         Data_Memory_Size  : IN     Integer;
         Heap_Memory_Size  : IN     Integer;
         Is_Preemptive     : IN     Preemptive_Type    := Preemptive;
         Quantum           : IN     Integer            := 0;
         File_Name         : IN     Unbounded_String   := Empty_String;
         A_Scheduler       : IN     Scheduler_Type     := No_Scheduling_Protocol) IS

   BEGIN


      IF Cpu_Name=""
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Processor_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;


      IF Name=""
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Address_Space_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;


      IF NOT Is_A_Valid_Identifier(Name)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Address_Space_Name(Current_Language) &
               Lb_Colon & Lb_Invalid_Identifier(Current_Language)));
      END IF;


      IF (Text_Memory_Size<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Text_Memory_Size(Current_Language) &
               Lb_Must_Be(
                  Current_Language) & ">=0"));
      END IF;


      IF (Data_Memory_Size<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Data_Memory_Size(Current_Language) &
               Lb_Must_Be(
                  Current_Language) & ">=0"));
      END IF;



      IF (Heap_Memory_Size<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Heap_Memory_Size(Current_Language) &
               Lb_Must_Be(
                  Current_Language) & ">=0"));
      END IF;



      IF (Stack_Memory_Size<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Stack_Memory_Size(
                  Current_Language) &
               Lb_Must_Be(
                  Current_Language) & ">=0"));
      END IF;

      IF (A_Scheduler=Parametric_Protocol)
            AND (File_Name="")
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_File_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;


      IF (File_Name/="")
            AND (A_Scheduler/=Parametric_Protocol)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_File_Name_Control(Current_Language)));
      END IF;



      IF (Quantum<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & "Quantum" &
               Lb_Must_Be(Current_Language)
               & To_Unbounded_String(" >= 0")));
      END IF;


      IF (Quantum/=0)
            AND  (A_Scheduler/=Posix_1003_Highest_Priority_First_Protocol)
            AND (A_Scheduler/=Round_Robin_Protocol)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Quantum_Control(Current_Language)));
      END IF;


      IF (A_Scheduler=Hierarchical_Cyclic_Protocol)
            OR (A_Scheduler=Hierarchical_Round_Robin_Protocol)
            OR (A_Scheduler=Hierarchical_Fixed_Priority_Protocol)
            OR (A_Scheduler=Hierarchical_Parametric_Protocol)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Invalid_Scheduler(
                  Current_Language)));
      END IF;

   END Check_Address_Space;



   PROCEDURE Add_Address_Space (
         My_Address_Spaces : IN OUT Address_Spaces_Set;
         A_Address_Space   : IN OUT Address_Space_Ptr;
         Name              : IN     Unbounded_String;
         Cpu_Name          : IN     Unbounded_String;
         Text_Memory_Size  : IN     Integer;
         Stack_Memory_Size : IN     Integer;
         Data_Memory_Size  : IN     Integer;
         Heap_Memory_Size  : IN     Integer;
         Is_Preemptive     : IN     Preemptive_Type    := Preemptive;
         Quantum           : IN     Integer            := 0;
         File_Name         : IN     Unbounded_String   := Empty_String;
         A_Scheduler       : IN     Scheduler_Type     := No_Scheduling_Protocol) IS


      My_Iterator : Address_Spaces_Iterator;

   BEGIN

      Check_Address_Space(My_Address_Spaces, Name,
         Cpu_Name,
         Text_Memory_Size,
         Stack_Memory_Size, Data_Memory_Size,
         Heap_Memory_Size, Is_Preemptive, Quantum, File_Name, A_Scheduler);


      IF(Get_Number_Of_Elements(My_Address_Spaces) > 0)
            THEN

         Reset_Iterator(My_Address_Spaces, My_Iterator);

         LOOP
            Current_Element(My_Address_Spaces, A_Address_Space,
               My_Iterator);
            IF (Name = A_Address_Space.Name)
                  THEN
               Raise_Exception(Invalid_Parameter'
                  Identity,
                  To_String(Lb_Address_Space(Current_Language) & " " &
                     Name &
                     " : " & Lb_Address_Space_Name(
                        Current_Language) &
                     Lb_Already_Defined(
                        Current_Language)));
            END IF;

            EXIT WHEN Is_Last_Element(
               My_Address_Spaces, My_Iterator);

            Next_Element(My_Address_Spaces,
               My_Iterator);
         END LOOP;
      END IF;


      A_Address_Space:= NEW Address_Space;

      A_Address_Space.Cpu_Name:=Cpu_Name;
      A_Address_Space.Name:=Name;
      A_Address_Space.Text_Memory_Size:=Text_Memory_Size;
      A_Address_Space.Stack_Memory_Size:=Stack_Memory_Size;
      A_Address_Space.Data_Memory_Size:=Data_Memory_Size;
      A_Address_Space.Heap_Memory_Size:=Heap_Memory_Size;


      CASE A_Scheduler IS
         WHEN Rate_Monotonic_Protocol =>
            A_Address_Space.Scheduler := NEW Rm_Scheduler;
         WHEN Deadline_Monotonic_Protocol =>
            A_Address_Space.Scheduler := NEW Dm_Scheduler;
         WHEN Posix_1003_Highest_Priority_First_Protocol =>
            A_Address_Space.Scheduler := NEW Hpf_Scheduler;
         WHEN Earliest_Deadline_First_Protocol =>
            A_Address_Space.Scheduler := NEW Edf_Scheduler;
         WHEN Least_Laxity_First_Protocol =>
            A_Address_Space.Scheduler := NEW Llf_Scheduler;
         WHEN Round_Robin_Protocol =>
            A_Address_Space.Scheduler := NEW Round_Robin_Scheduler;
         WHEN Maximum_Urgency_First_Based_On_Deadline_Protocol =>
            A_Address_Space.Scheduler := NEW Deadline_Muf_Scheduler;
         WHEN Maximum_Urgency_First_Based_On_Laxity_Protocol =>
            A_Address_Space.Scheduler := NEW Laxity_Muf_Scheduler;
         WHEN Time_Sharing_Based_On_Wait_Time_Protocol =>
            A_Address_Space.Scheduler := NEW
               Time_Sharing_Based_On_Wait_Time_Scheduler;
         WHEN Time_Sharing_Based_On_Cpu_Usage_Protocol =>
            A_Address_Space.Scheduler := NEW
               Time_Sharing_Based_On_Cpu_Usage_Scheduler;
         WHEN D_Over_Protocol =>
            A_Address_Space.Scheduler := NEW D_Over_Scheduler;
         WHEN Parametric_Protocol =>
            A_Address_Space.Scheduler := NEW Parametric_Scheduler;
         WHEN No_Scheduling_Protocol =>
            A_Address_Space.Scheduler := NEW No_Scheduler;
         WHEN OTHERS =>
            Raise_Exception(Invalid_Parameter'Identity,
               To_String(Lb_Address_Space(Current_Language) & " " & Name &
                  " : " & Lb_Invalid_Scheduler(
                     Current_Language)));
      END CASE;

      IF (A_Scheduler=Parametric_Protocol)
            THEN
         Set_Behavior_File_Name (
            Parametric_Scheduler(A_Address_Space.Scheduler.All), File_Name) ;
      END IF;


      Set_Quantum(A_Address_Space.Scheduler.All, Quantum);
      Set_Preemptive(A_Address_Space.Scheduler.All,
         Is_Preemptive);



      Add(My_Address_Spaces,A_Address_Space);


   EXCEPTION
      WHEN Generic_Address_Space_Set.Full_Set =>
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Can_Not_Define_More_Address_Spaces(
                  Current_Language)));

   END Add_Address_Space;





   FUNCTION Search_Address_Space (
         My_Address_Spaces : IN     Address_Spaces_Set;
         Name              : IN     Unbounded_String)
     RETURN Address_Space_Ptr IS


      My_Iterator : Address_Spaces_Iterator;
      A_Addr      : Address_Space_Ptr;
      Result      : Address_Space_Ptr;

      Found : Boolean := False;

   BEGIN

      Reset_Iterator(My_Address_Spaces, My_Iterator);

      LOOP
         Current_Element(My_Address_Spaces,A_Addr,
            My_Iterator);

         IF(A_Addr.Name = Name)
               THEN
            Found:=True;
            Result:=A_Addr;
         END IF;

         EXIT WHEN Is_Last_Element(My_Address_Spaces,
            My_Iterator);

         Next_Element(My_Address_Spaces, My_Iterator);

      END LOOP;

      IF NOT Found
            THEN
         Raise_Exception(Address_Space_Not_Found'Identity,
            To_String(Lb_Address_Space_Name(Current_Language) & "=" & Name));
      END IF;

      RETURN Result;
   END Search_Address_Space;





   PROCEDURE Delete_Processor (
         My_Address_Spaces : IN OUT Address_Spaces_Set;
         A_Processor       : IN     Unbounded_String) IS

      Tmp             : Address_Spaces_Set;
      A_Address_Space : Address_Space_Ptr;
      My_Iterator     : Address_Spaces_Iterator;

   BEGIN

      IF(Get_Number_Of_Elements(My_Address_Spaces) >
            0)
            THEN

         Reset_Iterator(My_Address_Spaces, My_Iterator);
         LOOP
            Current_Element(My_Address_Spaces,
               A_Address_Space,
               My_Iterator);

            IF(A_Address_Space.Cpu_Name = A_Processor)
                  THEN
               Add(Tmp, A_Address_Space);
            END IF;

            EXIT WHEN Is_Last_Element(
               My_Address_Spaces, My_Iterator);
            Next_Element(My_Address_Spaces,
               My_Iterator);
         END LOOP;

         IF NOT Is_Empty(Tmp) THEN
            Reset_Iterator(My_Address_Spaces, My_Iterator);
            LOOP
               Current_Element(Tmp, A_Address_Space,
                  My_Iterator);


               Delete(My_Address_Spaces,
                  A_Address_Space);

               EXIT WHEN Is_Last_Element(Tmp,
                  My_Iterator);
               Next_Element(Tmp, My_Iterator);
            END LOOP;

            Reset(Tmp,False);
         END IF;
      END IF;

   END Delete_Processor;




   --
   -- XML/AADL export sub-programs
   --



   FUNCTION Export_Xml (
         My_Address_Spaces : IN     Address_Spaces_Set)
     RETURN Unbounded_String IS

      My_Iterator : Address_Spaces_Iterator;
      A_Addr      : Address_Space_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN

      IF NOT Is_Empty(My_Address_Spaces) THEN

         Result:= Result & To_Unbounded_String(
            "<address_spaces>") &
            Unbounded_Lf;
         Reset_Iterator(My_Address_Spaces, My_Iterator);

         LOOP
            Current_Element(My_Address_Spaces,A_Addr,
               My_Iterator);

            Result:= Result & To_Unbounded_String(
               Ascii.Ht &
               "<address_space>") & Unbounded_Lf;


            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<name>" & To_String(
                  A_Addr.Name) & "</name>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<cpu_name>" & To_String(
                  A_Addr.Cpu_Name) & "</cpu_name>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<text_memory_size>" &
               A_Addr.
               Text_Memory_Size'
               Img & "</text_memory_size>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<stack_memory_size>" &
               A_Addr.
               Stack_Memory_Size'
               Img & "</stack_memory_size>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<data_memory_size>" &
               A_Addr.
               Data_Memory_Size'
               Img & "</data_memory_size>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<heap_memory_size>" &
               A_Addr.
               Heap_Memory_Size'
               Img & "</heap_memory_size>") &
               Unbounded_Lf;


            Result:= Result & Export_Xml(A_Addr.Scheduler.All);


            Result:= Result & To_Unbounded_String(
               Ascii.Ht &
               "</address_space>") & Unbounded_Lf;

            EXIT WHEN Is_Last_Element(
               My_Address_Spaces, My_Iterator);

            Next_Element(My_Address_Spaces,
               My_Iterator);

         END LOOP;
         Result:= Result & To_Unbounded_String(
            "</address_spaces>") &
            Unbounded_Lf & Unbounded_Lf;
      END IF;

      RETURN Result;

   END Export_Xml;




   FUNCTION Export_Aadl_Implementations (
         My_Address_Spaces : IN     Address_Spaces_Set;
         My_Tasks          : IN     Tasks_Set;
         My_Resources      : IN     Resources_Set)
     RETURN Unbounded_String IS


      My_Iterator1    : Address_Spaces_Iterator;
      A_Address_Space : Address_Space_Ptr;

      My_Iterator2 : Tasks_Iterator;
      A_Task       : Task_Ptr;

      My_Iterator3 : Resources_Iterator;
      A_Resource   : Resource_Ptr;

      Result               : Unbounded_String := Empty_String;
      Thread_Subcomponents : Unbounded_String := Empty_String;
      Data_Subcomponents   : Unbounded_String := Empty_String;
      First_Resource       : Boolean          := True;

   BEGIN


      IF NOT Is_Empty(My_Address_Spaces) THEN


         Reset_Iterator(My_Address_Spaces, My_Iterator1);

         LOOP
            Current_Element(My_Address_Spaces,
               A_Address_Space,
               My_Iterator1);

            Result:= Result & To_Unbounded_String(
               "process " & To_String(
                  A_Address_Space.Name)) &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               "end " & To_String(
                  A_Address_Space.Name) & ";" ) &
               Unbounded_Lf;

            Result:=Result & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               "process implementation " & To_String(
                  A_Address_Space.
                  Name) & ".Impl") & Unbounded_Lf;


            Thread_Subcomponents :=  Export_Aadl_Declarations(My_Tasks,
               A_Address_Space.Name, 2);
            Data_Subcomponents :=  Export_Aadl_Declarations(My_Resources,
               A_Address_Space.Name, 2);
            IF (Thread_Subcomponents /= Empty_String)
                  OR (Data_Subcomponents /= Empty_String)
                  THEN
               Result:= Result & To_Unbounded_String(
                  Ascii.Ht &
                  "subcomponents") & Unbounded_Lf & Thread_Subcomponents &
                  Data_Subcomponents;
            END IF;



            IF Data_Subcomponents/=Empty_String THEN

               Reset_Iterator(My_Tasks, My_Iterator2);

               Result:= Result &
                  To_Unbounded_String( Ascii.Ht &
                  "connections") & Unbounded_Lf;

               LOOP
                  Current_Element(My_Tasks, A_Task,
                     My_Iterator2);

                  IF A_Task.Address_Space_Name =
                        A_Address_Space.Name THEN
                     Result:= Result &
                        Export_Aadl_Connections(
                        My_Resources,
                        A_Task.Name,  2);
                  END IF;

                  EXIT WHEN Is_Last_Element(
                     My_Tasks, My_Iterator2);

                  Next_Element(My_Tasks,
                     My_Iterator2);

               END LOOP;
            END IF;


            Result:= Result & To_Unbounded_String(
               Ascii.Ht &
               "properties") & Unbounded_Lf;

            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "Cheddar_Properties::Source_Global_Text_Size => " &
               A_Address_Space.Text_Memory_Size'
               Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "Cheddar_Properties::Source_Global_Data_Size => " &
               A_Address_Space.Data_Memory_Size'
               Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "Cheddar_Properties::Source_Global_Heap_Size => " &
               A_Address_Space.Heap_Memory_Size'
               Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "Cheddar_Properties::Source_Global_Stack_Size => " &
               A_Address_Space.
               Stack_Memory_Size'Img & ";") &
               Unbounded_Lf;
            IF Get_Name(A_Address_Space.Scheduler.All) /=
                  No_Scheduling_Protocol
                  THEN
               Result:= Result & Export_Aadl_Properties (
                  A_Address_Space.Scheduler.All, 2);
            END IF;

            IF Data_Subcomponents/=Empty_String THEN

               Result:= Result & To_Unbounded_String(  Ascii.Ht &
                  Ascii.Ht &
                  "Cheddar_Properties::Critical_Section => (")  &
                  Unbounded_Lf;

               Reset_Iterator(My_Resources, My_Iterator3);

               LOOP
                  Current_Element(My_Resources,
                     A_Resource,
                     My_Iterator3);

                  IF A_Resource.Address_Space_Name =
                        A_Address_Space.Name THEN

                     IF A_Resource.Task_Tab.Nb_Entries/=0 THEN
                        IF NOT First_Resource THEN
                           Result:= Result & To_Unbounded_String(",") &
                              Unbounded_Lf;
                        END IF;
                        First_Resource:=False;
                        FOR I IN 0..A_Resource.Task_Tab.Nb_Entries-1 LOOP
                           Result:= Result & Ascii.Ht & Ascii.Ht &
                              Ascii.Ht
                              & """" & "instancied_" & To_String(
                              A_Resource.Name) & """" & "," & Unbounded_Lf;
                           Result:= Result &
                              To_Unbounded_String(
                              Ascii.Ht &
                              Ascii.Ht & Ascii.Ht
                              & """" & "instancied_" & To_String(
                                 A_Resource.Task_Tab.Entries(I).
                                 Item) & """,""" & To_String(
                                 Integer_Util.Format(
                                    A_Resource.Task_Tab.Entries(I).
                                    Data.Task_Begin)) &
                              """,""" & To_String(Integer_Util.Format(
                                    A_Resource.
                                    Task_Tab.Entries(I).Data.Task_End)) &
                              """");
                           IF I/=A_Resource.Task_Tab.Nb_Entries-1 THEN
                              Result:= Result & To_Unbounded_String(  ",");
                              Result:=Result & Unbounded_Lf;
                           END IF;
                        END LOOP;
                     END IF;
                  END IF;


                  EXIT WHEN Is_Last_Element(
                     My_Resources, My_Iterator3);

                  Next_Element(My_Resources,
                     My_Iterator3);

               END LOOP;

               Result:= Result & To_Unbounded_String(Ascii.Ht &
                  Ascii.Ht &
                  Ascii.Ht & ");") & Unbounded_Lf;
            END IF;

            Result:= Result & To_Unbounded_String(
               "end " & To_String(
                  A_Address_Space.Name) &
               ".Impl;" ) & Unbounded_Lf;

            Result:=Result & Unbounded_Lf;


            EXIT WHEN Is_Last_Element(
               My_Address_Spaces, My_Iterator1);

            Next_Element(My_Address_Spaces,
               My_Iterator1);

         END LOOP;



      END IF;

      RETURN Result;

   END Export_Aadl_Implementations;



   FUNCTION Export_Aadl_Declarations (
         My_Address_Spaces : IN     Address_Spaces_Set;
         Number_Of_Ht      : IN     Natural)
     RETURN Unbounded_String IS

      My_Iterator     : Address_Spaces_Iterator;
      A_Address_Space : Address_Space_Ptr;

      Result : Unbounded_String := Empty_String;


   BEGIN

      IF NOT Is_Empty(My_Address_Spaces) THEN


         Reset_Iterator(My_Address_Spaces, My_Iterator);

         LOOP
            Current_Element(My_Address_Spaces,
               A_Address_Space,
               My_Iterator);


            FOR I IN 1..Number_Of_Ht LOOP
               Result:= Result & Ascii.Ht;
            END LOOP;

            Result:= Result & To_Unbounded_String(
               "instancied_" &
               To_String(A_Address_Space.
                  Name) &  " : process " & To_String(
                  A_Address_Space.Name) &
               ".Impl;") & Unbounded_Lf;

            EXIT WHEN Is_Last_Element(
               My_Address_Spaces, My_Iterator);

            Next_Element(My_Address_Spaces,
               My_Iterator);

         END LOOP;



      END IF;


      RETURN Result;


   END Export_Aadl_Declarations;



   FUNCTION Export_Aadl_Properties (
         My_Address_Spaces : IN     Address_Spaces_Set;
         Number_Of_Ht      : IN     Natural)
     RETURN Unbounded_String IS


      My_Iterator     : Address_Spaces_Iterator;
      A_Address_Space : Address_Space_Ptr;

      Result : Unbounded_String := Empty_String;


   BEGIN

      IF NOT Is_Empty(My_Address_Spaces) THEN


         Reset_Iterator(My_Address_Spaces, My_Iterator);

         LOOP
            Current_Element(My_Address_Spaces,
               A_Address_Space,
               My_Iterator);

            FOR I IN 1..Number_Of_Ht LOOP
               Result:= Result & Ascii.Ht;
            END LOOP;

            Result:= Result & To_Unbounded_String(
               "Actual_Processor_Binding => reference instancied_" &
               To_String(A_Address_Space.
                  Cpu_Name) &
               " applies to instancied_" & To_String(
                  A_Address_Space.Name) & ";")  &
               Unbounded_Lf;

            EXIT WHEN Is_Last_Element(
               My_Address_Spaces, My_Iterator);

            Next_Element(My_Address_Spaces,
               My_Iterator);

         END LOOP;



      END IF;


      RETURN Result;


   END Export_Aadl_Properties;






END Address_Space_Set;


