------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Ada.Exceptions;
USE Ada.Exceptions;

WITH Translate;
USE Translate;

WITH Buffers;
USE Buffers;
USE Buffers.Buffer_Roles;

WITH Objects;
USE Objects;

WITH Unbounded_Strings;
USE Unbounded_Strings;


PACKAGE BODY Buffer_Set IS






   PROCEDURE Add_Buffer (
         My_Buffers         : IN OUT Buffers_Set;
         Name               : IN     Unbounded_String;
         Size               : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         A_Qs               : IN     Queueing_Systems_Type;
         Roles              : IN     Buffer_Roles_Table) IS

      Dummy : Buffer_Ptr;

   BEGIN
      Add_Buffer(My_Buffers, Dummy, Name, Size, Cpu_Name,
         Address_Space_Name, A_Qs,
         Roles);
   END Add_Buffer;



   PROCEDURE Check_Buffer (
         My_Buffers         : IN     Buffers_Set;
         Name               : IN     Unbounded_String;
         Size               : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String) IS


   BEGIN

      IF (Name = "")
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Buffer_Name(Current_Language)  &
               Lb_Mandatory(Current_Language)));
      END IF;


      IF NOT Is_A_Valid_Identifier(Name)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Buffer(Current_Language) & " " & Name &
               " : " & Lb_Buffer_Name(Current_Language) &
               Lb_Colon & Lb_Invalid_Identifier(Current_Language)));
      END IF;


      IF (Cpu_Name = "")
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Buffer(Current_Language) & " " & Name &
               " : " & Lb_Processor_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;


      IF (Address_Space_Name = "")
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Buffer(Current_Language) & " " & Name &
               " : " & Lb_Address_Space_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;


      IF (Size<=0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Buffer(Current_Language) & " " & Name &
               " : " & Lb_Size(Current_Language) & Lb_Must_Be(
                  Current_Language) & ">0"));
      END IF;

   END Check_Buffer;


   PROCEDURE Add_Buffer (
         My_Buffers         : IN OUT Buffers_Set;
         A_Buffer           : IN OUT Buffer_Ptr;
         Name               : IN     Unbounded_String;
         Size               : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         A_Qs               : IN     Queueing_Systems_Type;
         Roles              : IN     Buffer_Roles_Table) IS

      My_Iterator : Iterator;

   BEGIN

      Check_Buffer(My_Buffers, Name, Size, Cpu_Name, Address_Space_Name);


      IF(Get_Number_Of_Elements(My_Buffers) > 0)
            THEN

         Reset_Iterator(My_Buffers, My_Iterator);

         LOOP
            Current_Element(My_Buffers, A_Buffer, My_Iterator);
            IF (Name = A_Buffer.Name)
                  THEN
               Raise_Exception(Invalid_Parameter'Identity,
                  To_String(Lb_Buffer(Current_Language) & " " & Name &
                     " : " & Lb_Buffer_Name(Current_Language) &
                     Lb_Already_Defined(Current_Language)));
            END IF;

            EXIT WHEN Is_Last_Element(My_Buffers, My_Iterator);

            Next_Element(My_Buffers, My_Iterator);
         END LOOP;
      END IF;


      A_Buffer := NEW Buffer;
      A_Buffer.Name:=Name;
      A_Buffer.Cpu_Name:=Cpu_Name;
      A_Buffer.Address_Space_Name:=Address_Space_Name;
      A_Buffer.Size:=Size;
      A_Buffer.Roles:=Roles;
      A_Buffer.Queueing_System_Type:=A_Qs;

      Add(My_Buffers,A_Buffer);


   END Add_Buffer;



   PROCEDURE Update_Buffer (
         My_Buffers         : IN OUT Buffers_Set;
         Name               : IN     Unbounded_String;
         New_Name           : IN     Unbounded_String;
         Size               : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         Roles              : IN     Buffer_Roles_Table) IS
      The_Buffer : Buffer_Ptr;
   BEGIN
      The_Buffer:=Search_Buffer(My_Buffers,Name);
      The_Buffer.Name:=New_Name;
      The_Buffer.Address_Space_Name:=Address_Space_Name;
      The_Buffer.Cpu_Name:=Cpu_Name;
      The_Buffer.Size:=Size;
      The_Buffer.Roles:=Roles;

   END Update_Buffer;




   FUNCTION Get (
         My_Buffers  : IN     Buffers_Set;
         Buffer_Name : IN     Unbounded_String;
         Param_Name  : IN     Buffer_Parameters)
     RETURN Natural IS

      A_Buffer    : Buffer_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF ( Param_Name /= Size)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Buffers, My_Iterator);

      LOOP
         Current_Element(My_Buffers,A_Buffer,My_Iterator);

         IF (A_Buffer.Name = Buffer_Name)
               THEN
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Buffers, My_Iterator);

         Next_Element(My_Buffers, My_Iterator);

      END LOOP;

      RETURN A_Buffer.Size;

   END Get;



   FUNCTION Get (
         My_Buffers  : IN     Buffers_Set;
         Buffer_Name : IN     Unbounded_String;
         Param_Name  : IN     Buffer_Parameters)
     RETURN Unbounded_String IS

      A_Buffer    : Buffer_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF ( Param_Name /= Cpu_Name)
            AND ( Param_Name /= Address_Space_Name)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Buffers, My_Iterator);

      LOOP
         Current_Element(My_Buffers,A_Buffer,My_Iterator);

         IF (A_Buffer.Name = Buffer_Name)
               THEN
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Buffers, My_Iterator);

         Next_Element(My_Buffers, My_Iterator);

      END LOOP;

      IF ( Param_Name /= Cpu_Name)
            THEN
         RETURN A_Buffer.Cpu_Name;
      ELSE
         RETURN A_Buffer.Address_Space_Name;
      END IF;

   END Get;


   PROCEDURE Set (
         My_Buffers  : IN OUT Buffers_Set;
         Buffer_Name : IN     Unbounded_String;
         Param_Name  : IN     Buffer_Parameters;
         Param_Value : IN     Natural) IS
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF (Param_Name /= Size)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Buffers, My_Iterator);

      LOOP
         Current_Element(My_Buffers,A_Buffer,My_Iterator);

         IF (A_Buffer.Name = Buffer_Name)
               THEN
            A_Buffer.Size:=Param_Value;
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Buffers, My_Iterator);

         Next_Element(My_Buffers, My_Iterator);

      END LOOP;
   END Set;


   PROCEDURE Set (
         My_Buffers  : IN OUT Buffers_Set;
         Buffer_Name : IN     Unbounded_String;
         Param_Name  : IN     Buffer_Parameters;
         Param_Value : IN     Unbounded_String) IS
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF (Param_Name /= Cpu_Name)
            AND (Param_Name /= Address_Space_Name)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Buffers, My_Iterator);

      LOOP
         Current_Element(My_Buffers,A_Buffer,My_Iterator);

         IF (A_Buffer.Name = Buffer_Name)
               THEN
            IF (Param_Name /= Cpu_Name)
                  THEN
               A_Buffer.Cpu_Name:=Param_Value;
            ELSE
               A_Buffer.Address_Space_Name:=Param_Value;
            END IF;
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Buffers, My_Iterator);

         Next_Element(My_Buffers, My_Iterator);

      END LOOP;
   END Set;


   FUNCTION Search_Buffer (
         My_Buffers : IN     Buffers_Set;
         Name       : IN     Unbounded_String)
     RETURN Buffer_Ptr IS

      My_Iterator : Iterator;
      A_Buffer    : Buffer_Ptr;
      Result      : Buffer_Ptr;

      Found : Boolean := False;

   BEGIN

      Reset_Iterator(My_Buffers, My_Iterator);

      LOOP
         Current_Element(My_Buffers,A_Buffer,My_Iterator);

         IF(A_Buffer.Name = Name) THEN
            Found:=True;
            Result:=A_Buffer;
         END IF;

         EXIT WHEN Is_Last_Element(My_Buffers, My_Iterator);

         Next_Element(My_Buffers, My_Iterator);

      END LOOP;

      IF NOT Found THEN
         RAISE Buffer_Not_Found;
      END IF;

      RETURN Result;

   END Search_Buffer;



   PROCEDURE Delete_Task (
         My_Buffers : IN OUT Buffers_Set;
         A_Task     : IN     Unbounded_String) IS

      A_Buffer    : Buffer_Ptr;
      My_Iterator : Buffers_Iterator;
      Nb_Entries  : Buffer_Roles_Range;

   BEGIN

      IF(Get_Number_Of_Elements(My_Buffers) > 0)
            THEN

         Reset_Iterator(My_Buffers, My_Iterator);
         LOOP
            Current_Element(My_Buffers, A_Buffer, My_Iterator);

            -- Looking for a_task in roles
            --
            Nb_Entries:=A_Buffer.Roles.Nb_Entries;
            FOR Index IN 0..Nb_Entries-1 LOOP
               IF (A_Buffer.Roles.Entries(Index).Item = A_Task)
                     THEN
                  A_Buffer.Roles.Entries(Index):=A_Buffer.Roles.Entries(
                     A_Buffer.Roles.Nb_Entries-1);

                  A_Buffer.Roles.Nb_Entries:=
                     A_Buffer.Roles.Nb_Entries-1;
               END IF;
            END LOOP;


            EXIT WHEN Is_Last_Element(My_Buffers, My_Iterator);
            Next_Element(My_Buffers, My_Iterator);
         END LOOP;

      END IF;

   END Delete_Task;



   PROCEDURE Delete_Address_Space (
         My_Buffers : IN OUT Buffers_Set;
         A_Addr     : IN     Unbounded_String) IS

      Tmp         : Buffers_Set;
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Buffers_Iterator;

   BEGIN

      IF(Get_Number_Of_Elements(My_Buffers) > 0)
            THEN

         Reset_Iterator(My_Buffers, My_Iterator);
         LOOP
            Current_Element(My_Buffers, A_Buffer, My_Iterator);

            IF(A_Buffer.Address_Space_Name = A_Addr)
                  THEN
               Add(Tmp, A_Buffer);
            END IF;

            EXIT WHEN Is_Last_Element(My_Buffers, My_Iterator);
            Next_Element(My_Buffers, My_Iterator);
         END LOOP;

         IF NOT Is_Empty(Tmp) THEN
            Reset_Iterator(Tmp, My_Iterator);
            LOOP
               Current_Element(Tmp, A_Buffer, My_Iterator);

               Delete(My_Buffers,A_Buffer);

               EXIT WHEN Is_Last_Element(Tmp, My_Iterator);
               Next_Element(Tmp, My_Iterator);
            END LOOP;
         END IF;


      END IF;
   END Delete_Address_Space;


   FUNCTION Get_Number_Of_Buffer_From_Processor (
         My_Buffers     : IN     Buffers_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Buffers_Range IS

      Number      : Buffers_Range := 0;
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF Is_Empty(My_Buffers)
            THEN
         RETURN 0;
      END IF;

      Reset_Iterator(My_Buffers, My_Iterator);

      LOOP
         Current_Element(My_Buffers, A_Buffer, My_Iterator);

         IF (A_Buffer.Cpu_Name = Processor_Name)
               THEN
            Number:=Number+1;
         END IF;

         EXIT WHEN Is_Last_Element(My_Buffers, My_Iterator);

         Next_Element(My_Buffers, My_Iterator);


      END LOOP;

      RETURN Number;
   END Get_Number_Of_Buffer_From_Processor;




   PROCEDURE Delete_Processor (
         My_Buffers  : IN OUT Buffers_Set;
         A_Processor : IN     Unbounded_String) IS

      Tmp         : Buffers_Set;
      A_Buffer    : Buffer_Ptr;
      My_Iterator : Buffers_Iterator;

   BEGIN

      IF(Get_Number_Of_Elements(My_Buffers) > 0)
            THEN

         Reset_Iterator(My_Buffers, My_Iterator);
         LOOP
            Current_Element(My_Buffers, A_Buffer,
               My_Iterator);

            IF(A_Buffer.Cpu_Name = A_Processor)
                  THEN
               Add(Tmp, A_Buffer);
            END IF;

            EXIT WHEN Is_Last_Element(My_Buffers, My_Iterator);
            Next_Element(My_Buffers, My_Iterator);
         END LOOP;

         IF NOT Is_Empty(Tmp) THEN
            Reset_Iterator(Tmp, My_Iterator);
            LOOP
               Current_Element(Tmp, A_Buffer, My_Iterator);


               Delete(My_Buffers,A_Buffer);

               EXIT WHEN Is_Last_Element(Tmp, My_Iterator);
               Next_Element(Tmp, My_Iterator);
            END LOOP;

            Reset(Tmp,False);
         END IF;
      END IF;

   END Delete_Processor;






   FUNCTION Export_Xml (
         My_Buffers : IN     Buffers_Set)
     RETURN Unbounded_String IS

      My_Iterator : Buffers_Iterator;
      A_Buffer    : Buffer_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN

      IF NOT Is_Empty(My_Buffers) THEN

         Result:= Result & To_Unbounded_String("<buffers>") & Unbounded_Lf;

         Reset_Iterator(My_Buffers, My_Iterator);

         LOOP
            Current_Element(My_Buffers,A_Buffer,My_Iterator);


            Result:= Result & To_Unbounded_String(Ascii.Ht & "<buffer");
            IF A_Buffer.X/=0
                  THEN
               Result:= Result & To_Unbounded_String(" x=""" & A_Buffer.X'
                  Img & """ ");
            END IF;
            IF A_Buffer.Y/=0
                  THEN
               Result:= Result & To_Unbounded_String( "y=""" & A_Buffer.Y'
                  Img & """ ");
            END IF;
            Result:= Result & To_Unbounded_String( ">") & Unbounded_Lf;


            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<cpu_name>" & To_String(
                  A_Buffer.Cpu_Name) & "</cpu_name>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<address_space_name>" &
               To_String(
                  A_Buffer.Address_Space_Name) & "</address_space_name>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<name>" & To_String(
                  A_Buffer.Name) & "</name>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<size>" & A_Buffer.Size'
               Img & "</size>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<qs>" & A_Buffer.Queueing_System_Type'
               Img & "</qs>") & Unbounded_Lf;




            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<buffer_used_by>") & Unbounded_Lf;
            FOR I IN 0..A_Buffer.Roles.Nb_Entries-1 LOOP
               Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
                  Ascii.Ht &
                  "<buffer_user");
               IF  A_Buffer.Roles.Entries(I).Data.The_Role = Consumer
                     THEN
                  Result:= Result & To_Unbounded_String(
                     " buffer_role=""consumer"">");
               ELSE
                  Result:= Result & To_Unbounded_String(">");
               END IF;
               Result:= Result & To_Unbounded_String( " " & To_String(
                     A_Buffer.Roles.Entries(I).Item) &
                  " ");
               Result:= Result & To_Unbounded_String( " " &
                  A_Buffer.Roles.Entries(I).Data.Size'Img &
                  " ");
               Result:= Result & To_Unbounded_String( " " &
                  A_Buffer.Roles.Entries(I).Data.Time'Img &
                  " ");

               Result:= Result & To_Unbounded_String("</buffer_user>") &
                  Unbounded_Lf;
            END LOOP;
            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "</buffer_used_by>") & Unbounded_Lf;


            Result:= Result & To_Unbounded_String(Ascii.Ht & "</buffer>") &
               Unbounded_Lf;


            EXIT WHEN Is_Last_Element(My_Buffers, My_Iterator);

            Next_Element(My_Buffers, My_Iterator);


         END LOOP;
         Result:= Result & To_Unbounded_String("</buffers>") &
            Unbounded_Lf;
      END IF;



      RETURN Result;

   END Export_Xml;



   FUNCTION Export_Aadl_Implementations (
         My_Buffers : IN     Buffers_Set)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;

   BEGIN

      RETURN Result;

   END Export_Aadl_Implementations;




END Buffer_Set;



