------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Text_Io;
USE Text_Io;

WITH Tasks;
USE Tasks;

WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;

WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Time_Units;
USE Time_Units;

WITH Double_Util;
USE Double_Util;


PACKAGE BODY Buffers IS


   PROCEDURE Put (
         B : IN     Buffer_Role) IS
   BEGIN
      Put(B.The_Role'Img & "/" & B.Size'Img & "/" & B.Time'Img);
   END Put;

   PROCEDURE Initialize (
         B :    OUT Buffer_Role) IS
   BEGIN
      B.The_Role:=No_Role;
      B.Size:=0;
      B.Time:=0;
   END Initialize;



   PROCEDURE Initialize (
         A_Buff : IN OUT Buffer) IS
   BEGIN
      A_Buff.Object_Type:=Buffer_Object_Type;
      A_Buff.Cpu_Name:=Empty_String;
      A_Buff.Address_Space_Name:=Empty_String;
      A_Buff.Size:=0;
      Initialize(A_Buff.Roles);
   END Initialize;



   PROCEDURE Put_Name (
         A_Buff : IN     Buffer_Ptr) IS
   BEGIN
      Put(To_String(A_Buff.Name));
   END Put_Name;

   PROCEDURE Put (
         A_Buff : IN     Buffer_Ptr) IS
   BEGIN
      Put(A_Buff.All);
   END Put;

   PROCEDURE Put (
         My_Buff : IN     Buffer) IS
   BEGIN
      Put("Buffer name : " & To_String(My_Buff.Name));
      Put(" ; Cpu : " & To_String(My_Buff.Cpu_Name));
      Put(" ; Address space : " & To_String(My_Buff.Address_Space_Name));
      Put(" ; Size : " & My_Buff.Size'Img);
      Put(" ; x : " & My_Buff.X'Img);
      Put(" ; y : " & My_Buff.Y'Img);
      New_Line;
      Put(My_Buff.Roles);
      New_Line;
   END Put;


   FUNCTION Copy (
         My_Buffer : IN     Buffer_Ptr)
     RETURN Buffer_Ptr IS

      Ptr : Buffer_Ptr;
   BEGIN
      Ptr:=NEW Buffer;
      Ptr.All:=My_Buffer.All;

      RETURN Ptr;
   END Copy;



   PROCEDURE Buffer_Flow_Control (
         My_Buff   : IN     Buffer_Ptr;
         My_Tasks  : IN     Tasks_Set;
         Flow_Cons : IN OUT Double;
         Flow_Prod : IN OUT Double) IS

      A_Task : Task_Ptr;

   BEGIN
      --init
      Flow_Prod := 0.0;
      Flow_Cons := 0.0;

      -- compute data flow
      FOR I IN  0 .. (My_Buff.Roles.Nb_Entries - 1) LOOP

         -- compute producers data flow
         IF (My_Buff.Roles.Entries(I).Data.The_Role = Producer)
               THEN
            A_Task := Search_Task(My_Tasks,My_Buff.Roles.Entries(I).Item);

            CASE A_Task.Task_Type IS
               WHEN Periodic_Type =>
                  Flow_Prod := Flow_Prod +
                     Double(My_Buff.Roles.Entries(I).Data.Size) /
                     Double(Periodic_Task_Ptr(A_Task).Period);

               WHEN Poisson_Type =>
                  Flow_Prod := Flow_Prod +
                     Double(My_Buff.Roles.Entries(I).Data.Size) /
                     Double(Poisson_Task_Ptr(A_Task).Period);
               WHEN OTHERS =>
                  RAISE Flow_Constraint_Not_Respected;
            END CASE;


         END IF;

         -- Compute consumers data flow
         --
         IF (My_Buff.Roles.Entries(I).Data.The_Role = Consumer)
               THEN
            A_Task := Search_Task(My_Tasks,My_Buff.Roles.Entries(I).Item);

            CASE A_Task.Task_Type IS
               WHEN Periodic_Type =>
                  Flow_Cons := Flow_Cons +
                     Double(My_Buff.Roles.Entries(I).Data.Size) /
                     Double(Periodic_Task_Ptr(A_Task).Period);
               WHEN Poisson_Type =>
                  Flow_Cons := Flow_Cons +
                     Double(My_Buff.Roles.Entries(I).Data.Size) /
                     Double(Poisson_Task_Ptr(A_Task).Period);
               WHEN OTHERS =>
                  RAISE Flow_Constraint_Not_Respected;
            END CASE;
         END IF;

      END LOOP;

      IF (NOT Equal(Flow_Prod,Flow_Cons)) AND  (Flow_Prod > Flow_Cons)
            THEN
         RAISE Flow_Constraint_Not_Respected;
      END IF;
   END Buffer_Flow_Control;



   FUNCTION Is_Cons_Prod_Harmonic (
         My_Buff  : IN     Buffer_Ptr;
         My_Tasks : IN     Tasks_Set)
     RETURN Boolean IS

      A_Task1,
      A_Task2 : Task_Ptr;
      Result  : Boolean  := True;
      Tmp_Mod : Natural;

   BEGIN

      FOR I IN  0 .. (My_Buff.Roles.Nb_Entries - 1) LOOP
         A_Task1 := Search_Task(My_Tasks,My_Buff.Roles.Entries(I).Item);
         FOR J IN  I .. (My_Buff.Roles.Nb_Entries - 1) LOOP
            A_Task2 := Search_Task(My_Tasks,My_Buff.Roles.Entries(J).Item);

            IF A_Task1.Task_Type = Periodic_Type THEN

               IF ( Periodic_Task_Ptr(A_Task1).Period >
                     Periodic_Task_Ptr(A_Task2).Period)
                     THEN
                  Tmp_Mod := Periodic_Task_Ptr(A_Task1).Period mod
                     Periodic_Task_Ptr(A_Task2).Period;
               ELSE
                  Tmp_Mod := Periodic_Task_Ptr(A_Task2).Period mod
                     Periodic_Task_Ptr(A_Task1).Period;
               END IF;

            ELSE
               IF A_Task1.Task_Type = Poisson_Type THEN

                  IF ( Poisson_Task_Ptr(A_Task1).Period >
                        Poisson_Task_Ptr(A_Task2).Period)
                        THEN
                     Tmp_Mod := Poisson_Task_Ptr(A_Task1).Period mod
                        Poisson_Task_Ptr(A_Task2).Period;
                  ELSE
                     Tmp_Mod := Poisson_Task_Ptr(A_Task2).Period mod
                        Poisson_Task_Ptr(A_Task1).Period;
                  END IF;
               END IF;
            END IF;

            IF (Tmp_Mod /= 0 )
                  THEN
               Result:=False;
            END IF;
         END LOOP;
      END LOOP;
      RETURN Result;
   END Is_Cons_Prod_Harmonic;








END Buffers;





