------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Framework_Config;
USE Framework_Config;

WITH Xml_Tag;
USE Xml_Tag;
WITH Scheduler;
USE Scheduler;
USE Scheduler.Double_Tasks_Parameters;

WITH Translate;
USE Translate;
WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;

WITH Unbounded_Strings;
USE Unbounded_Strings;


WITH Dependency_Services;
USE Dependency_Services;

WITH Multiprocessor_Services;
USE Multiprocessor_Services;
USE Multiprocessor_Services.Scheduling_Result_Per_Processor;




PACKAGE BODY Call_Dependency_Framework  IS









   PROCEDURE Chetto_Parameter_Modifications (
         Sys                : IN OUT System;
         Result             : IN OUT Unbounded_String;
         A_Processor        : IN     Processor_Ptr;
         Compute_Deadlines  : IN     Boolean;
         Compute_Priorities : IN     Boolean;
         Update_Tasks_Set   : IN     Boolean;
         Output             : IN     Output_Format    := String_Output) IS



      Tmp       : Tasks_Set;
      A_Task    : Task_Ptr;
      Iterator1 : Tasks_Iterator;

   BEGIN


      IF Compute_Deadlines THEN

         Result:=Result & Unbounded_Lf & Lb_Minus & Lb_Chetto_Deadline(
            Current_Language)  &
            To_Unbounded_String(" (") & Lb_See(Current_Language) &
            To_Unbounded_String("[5], [6]).") &
            Unbounded_Lf & Unbounded_Lf;

      ELSE
         Result := Result & Unbounded_Lf & Lb_Minus & Lb_Chetto_Priority(
            Current_Language)  &
            To_Unbounded_String(" (") & Lb_See(Current_Language) &
            To_Unbounded_String("[5], [6]).") &
            Unbounded_Lf & Unbounded_Lf;
      END IF;





      -- If we must update parameters, we do not work on temporary tasks_set
      -- Call logic and display new parameters
      --
      IF Update_Tasks_Set THEN
         Set_Parameters_According_To_Chetto(Sys.Dependencies, Sys.
            Tasks,
            Compute_Deadlines, Compute_Priorities);



         -- Display priorities/deadlines
         --
         Reset_Iterator(Sys.Tasks, Iterator1);
         LOOP

            Current_Element(Sys.Tasks, A_Task, Iterator1);
            Result := Result & To_Unbounded_String("- ") &
               A_Task.
               Name & To_Unbounded_String(" => ");
            IF Compute_Deadlines
                  THEN
               Result:=Result & To_Unbounded_String(A_Task.
                  Deadline'Img) &
                  Unbounded_Lf;
            ELSE
               Result:=Result & To_Unbounded_String(A_Task.
                  Priority'Img) &
                  Unbounded_Lf;
            END IF;

            EXIT WHEN Is_Last_Element(Sys.Tasks, Iterator1);

            Next_Element(Sys.Tasks, Iterator1);
         END LOOP;


         IF Compute_Priorities THEN
            Result:=Result &  Lb_Priorities_Are_Updated(
               Current_Language);
         ELSE
            Result:=Result & Lb_Deadlines_Are_Updated(
               Current_Language);
         END IF;
         Result:=Result & Unbounded_Lf & Unbounded_Lf &
            Unbounded_Lf;


      ELSE
         Duplicate(Sys.Tasks, Tmp);
         Set_Parameters_According_To_Chetto(Sys.Dependencies, Tmp,
            Compute_Deadlines, Compute_Priorities);


         -- Display priorities/deadlines
         --
         Reset_Iterator(Tmp, Iterator1);
         LOOP

            Current_Element(Tmp, A_Task, Iterator1);
            Result := Result & To_Unbounded_String("- ") &
               A_Task.
               Name & To_Unbounded_String(" => ");
            IF Compute_Deadlines
                  THEN
               Result := Result & To_Unbounded_String(A_Task.
                  Deadline'Img) &
                  Unbounded_Lf;
            ELSE
               Result:=Result &  To_Unbounded_String(A_Task.
                  Priority'Img) &
                  Unbounded_Lf;
            END IF;

            EXIT WHEN Is_Last_Element(Tmp, Iterator1);

            Next_Element(Tmp, Iterator1);
         END LOOP;

         Free(Tmp);

      END IF;

      Result:=Result & Unbounded_Lf ;




   EXCEPTION
      WHEN Dependency_Services.Dependencies_Period_Error =>
         Result := Result & Lb_Chetto_Error2(Current_Language) &
            Unbounded_Lf;
      WHEN Dependency_Services.Cyclic_Graph_Error =>
         Result := Result & Lb_Chetto_Error1(Current_Language) &
            Unbounded_Lf;
      WHEN Dependency_Services.Task_Model_Error =>
         Result := Result & Lb_Chetto_Error3(Current_Language) &
            Unbounded_Lf;




   END Chetto_Parameter_Modifications;




   PROCEDURE Compute_End_To_End_Response_Time (
         Sys              : IN OUT System;
         Result           : IN OUT Unbounded_String;
         Update_Tasks_Set : IN     Boolean;
         One_Step         : IN     Boolean;
         Output           : IN     Output_Format    := String_Output) IS


      Rt          : Response_Time_Table;
      Referencies : Unbounded_String    := Empty_String;
      Message     : Unbounded_String    := Empty_String;

   BEGIN

      Initialize(Rt);
      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;

      BEGIN

         Compute_End_To_End_Response_Time(Sys.Dependencies, Sys.
            Processors, Sys.Tasks, One_Step, Update_Tasks_Set,
            Referencies,
            Rt);



         IF One_Step THEN
            Message:=Message & Start_Line & Lb_Minus &
               Lb_Jitter_From_Response_Time_One_Step(Current_Language)&
               Start_Ref;
         ELSE
            Message:= Message & Start_Line & Lb_Minus &
               Lb_Jitter_From_Response_Time_All_Steps(
               Current_Language)& Start_Ref;
         END IF;


         Message:=Message & Referencies & End_Ref  & End_Line &
            Unbounded_Lf;



         IF Update_Tasks_Set THEN
            Message:=Message & Start_Line & Lb_Jitters_Are_Updated(
               Current_Language) & End_Line &
               Unbounded_Lf;
         END IF;



         Message:=Message & Start_Line & Lb_Task_Response_Time(
            Current_Language) &
            Unbounded_Lf;



         FOR I IN 0..Rt.Nb_Entries-1 LOOP
            IF (Rt.Entries(I).Item /= NULL)
                  THEN
               Message:=Message & Start_Task &  To_Unbounded_String(
                  "    ") & Rt.
                  Entries(
                  I).
                  Item.Name & To_Unbounded_String(" => ")
                  & Integer'Image(Integer(Rt.Entries(I).Data))
                  & End_Task  & Lb_Comma & Start_Jit & Lb_Jitter(
                  Current_Language) &
                  To_Unbounded_String("=")  & Periodic_Task_Ptr(Rt.
                  Entries(I).Item).Jitter'Img &
                  To_Unbounded_String(" ") & End_Jit;
               IF (Natural(Rt.Entries(I).Data) > Rt.Entries(I).Item.
                     Deadline)
                     THEN
                  Message := Message & Lb_Comma & Lb_Check_Deadline(
                     Current_Language) & Rt.Entries(I).Item.Deadline'
                     Img;
                  Message:=Message  &
                     To_Unbounded_String(")");
               END IF;

               Message:=Message & Unbounded_Lf;
            END IF;
         END LOOP;
         Message:=Message & End_Line & Unbounded_Lf;
         Message:=Message & End_End_To_End & Unbounded_Lf;



      EXCEPTION
         WHEN Task_Set.Task_Must_Have_Period_Equal_To_Deadline=>
            Message:=Message & Lb_Compute_Scheduling_Error_4(
               Current_Language) &
               Unbounded_Lf;
         WHEN Task_Set.Task_Must_Be_Periodic=>
            Message:=Message & Lb_Compute_Scheduling_Error_5(
               Current_Language) &
               Unbounded_Lf;
         WHEN Task_Set.Task_Model_Error =>
            Message:=Message & Lb_Compute_Scheduling_Error_6(
               Current_Language) &
               Unbounded_Lf;
         WHEN Scheduler.Processor_Utilization_Exceeded =>
            Message:=Message & Lb_Compute_Scheduling_Error_7(
               Current_Language) &
               Unbounded_Lf;
         WHEN Task_Set.Start_Time_Error =>
            Message:=Message & Lb_Compute_Scheduling_Error_12(
               Current_Language) &
               Unbounded_Lf;
         WHEN Task_Set.Offset_Error =>
            Message:=Message & Lb_Compute_Scheduling_Error_13(
               Current_Language) &
               Unbounded_Lf;


      END;



      Result:=Result & Message;





   END Compute_End_To_End_Response_Time;




END Call_Dependency_Framework;




