------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Text_Io;
USE  Text_Io;

with events;
use events;

WITH Xml_Tag;
USE Xml_Tag;
WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;
WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;

WITH Ada.Exceptions;
USE Ada.Exceptions;

WITH Processors;
USE Processors;
WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;

WITH Call_Scheduling_Framework;
USE Call_Scheduling_Framework;
WITH Call_Memory_Framework;
USE Call_Memory_Framework;
USE Call_Memory_Framework.Buffer_Result_Package;
WITH Call_Resource_Framework;
USE Call_Resource_Framework;
WITH Call_Dependency_Framework;
USE Call_Dependency_Framework;
WITH Call_Random_Framework;
USE Call_Random_Framework;

WITH Scheduler;
USE Scheduler;
WITH Time_Units;
USE Time_Units;

WITH Multiprocessor_Services;
USE Multiprocessor_Services;
USE Multiprocessor_Services.Scheduling_Result_Per_Processor;


WITH Translate;
USE Translate;


PACKAGE BODY Call_Framework  IS



   TYPE Sub_Framework_Request IS
      RECORD
         Statement : Framework_Statement_Type;
         Param     : Framework_Parameters_Table;
         Target    : Processor_Ptr;
      END RECORD;





   PROCEDURE Sequential_Framework_Request (
         Sys      : IN OUT System;
         Request  : IN     Framework_Request_Table;
         Response : IN OUT Framework_Response_Table) IS
   BEGIN
      Sequential_Framework_Request(Sys, Request, Response, Total_Order,
         String_Output);
   END   Sequential_Framework_Request;


   PROCEDURE Put (
         F : IN     Framework_Request) IS
   BEGIN
      Put(F.Statement'Img & "/" & To_String(F.Target));
   END Put;


   PROCEDURE Initialize (
         F : IN OUT Framework_Request) IS
   BEGIN
      F.Target:=Empty_String;
      Initialize(F.Param);
   END Initialize;


   PROCEDURE Initialize (
         F : IN OUT Framework_Response) IS
   BEGIN
      F.Title:=Empty_String;
      F.Text:=Empty_String;
   END Initialize;


   PROCEDURE Put (
         F : IN     Framework_Response) IS
   BEGIN
      Put_Line(To_String(F.Title));
      Put_Line(To_String(F.Text));
   END Put;




   PROCEDURE Call_A_Request (
         Sys        : IN OUT System;
         Request    : IN     Sub_Framework_Request;
         A_Response : IN OUT Unbounded_String;
         Output     : IN     Output_Format         := String_Output) IS


      -- Parameters to be used for calling the "computing time line" service
      --
      Event_To_Generate          : Time_Unit_Event_Type_Boolean_Table;
      Period                     : Natural                            :=
        0;
      Schedule_With_Offsets      : Boolean                            :=
        True;
      Schedule_With_Precedencies : Boolean                            :=
        True;
      Schedule_With_Resources    : Boolean                            :=
        True;
      With_Task_Specific_Seed    : Boolean                            :=
        True;
      Global_Seed_Value          : Integer                            :=
        0;
      Predictable_Global_Seed    : Boolean                            :=
        True;


   BEGIN



      CASE Request.Statement IS



         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Simulation_Time_Line =>


            Initialize (Event_To_Generate, True);


            -- Check parameters
            --
            FOR I IN 0..Request.Param.Nb_Entries-1 LOOP
               IF (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "period")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "seed_value")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "schedule_with_offsets")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "schedule_with_precedencies")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "schedule_with_resources")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "predictable")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "start_of_task_capacity")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "end_of_task_capacity")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "write_to_buffer")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "read_from_buffer")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "running_task")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "task_activation")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "send_message")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "receive_message")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "allocate_resource")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "release_resource")) AND
                     (Request.Param.Entries(I).Name/=To_Unbounded_String(
                        "wait_for_resource")) THEN
                  Raise_Exception(
                     Unexpected_Parameter'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;



               IF  (Request.Param.Entries(I).Ptype/=Integer_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "period")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Integer_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "seed_value")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "schedule_with_offsets")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "schedule_with_precedencies")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "schedule_with_resources")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "predictable")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "start_of_task_capacity")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "end_of_task_capacity")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "write_to_buffer")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "read_from_buffer")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "running_task")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "task_activation")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "send_message")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "receive_message")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "allocate_resource")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "release_resource")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;
               IF  (Request.Param.Entries(I).Ptype/=Boolean_Parameter) AND
                     (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "wait_for_resource")) THEN
                  Raise_Exception(Parameter_Type_Error'Identity,
                     To_String(Request.Param.Entries(I).Name));
               END IF;



               IF  (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "period")) THEN
                  Period:=Request.Param.Entries(I).Integer_Value;
               END IF;
               IF  (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "seed_value")) THEN
                  Global_Seed_Value:=Request.Param.Entries(I).
                     Integer_Value;
                  With_Task_Specific_Seed:=False;
               END IF;
               IF       (Request.Param.Entries(I).Name=
                     To_Unbounded_String(
                        "schedule_with_offsets")) THEN
                  Schedule_With_Offsets:=Request.Param.Entries(I).
                     Boolean_Value;
               END IF;
               IF      (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "schedule_with_precedencies")) THEN
                  Schedule_With_Precedencies:=Request.Param.Entries(I).
                     Boolean_Value;
               END IF;
               IF      (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "schedule_with_resources")) THEN
                  Schedule_With_Resources:=Request.Param.Entries(I).
                     Boolean_Value;
               END IF;
               IF       (Request.Param.Entries(I).Name=
                     To_Unbounded_String(
                        "predictable")) THEN
                  Predictable_Global_Seed:=Request.Param.Entries(I).
                     Boolean_Value;
               END IF;
               IF       (Request.Param.Entries(I).Name=
                     To_Unbounded_String(
                        "start_of_task_capacity")) THEN
                  Event_To_Generate(Start_Of_Task_Capacity):=Request.
                     Param.Entries(I).Boolean_Value;
               END IF;
               IF       (Request.Param.Entries(I).Name=
                     To_Unbounded_String(
                        "end_of_task_capacity")) THEN
                  Event_To_Generate(End_Of_Task_Capacity):=Request.Param.
                     Entries(I).Boolean_Value;
               END IF;
               IF       (Request.Param.Entries(I).Name=
                     To_Unbounded_String(
                        "write_to_buffer")) THEN
                  Event_To_Generate(Write_To_Buffer):=Request.Param.
                     Entries(I).Boolean_Value;
               END IF;
               IF      (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "read_from_buffer")) THEN
                  Event_To_Generate(Read_From_Buffer):=Request.Param.
                     Entries(I).Boolean_Value;
               END IF;
               IF      (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "running_task")) THEN
                  Event_To_Generate(Running_Task):=Request.Param.Entries(
                     I).Boolean_Value;
               END IF;
               IF      (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "task_activation")) THEN
                  Event_To_Generate(Task_Activation):=Request.Param.
                     Entries(I).Boolean_Value;
               END IF;
               IF      (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "send_message")) THEN
                  Event_To_Generate(Send_Message):=Request.Param.Entries(
                     I).Boolean_Value;
               END IF;
               IF      (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "receive_message")) THEN
                  Event_To_Generate(Receive_Message):=Request.Param.
                     Entries(I).Boolean_Value;
               END IF;
               IF      (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "allocate_resource")) THEN
                  Event_To_Generate(Receive_Message):=Request.Param.
                     Entries(I).Boolean_Value;
               END IF;
               IF      (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "release_resource")) THEN
                  Event_To_Generate(Release_Resource):=Request.Param.
                     Entries(I).Boolean_Value;
               END IF;
               IF      (Request.Param.Entries(I).Name=To_Unbounded_String(
                        "wait_for_resource")) THEN
                  Event_To_Generate(Wait_For_Resource):=Request.Param.
                     Entries(I).Boolean_Value;
               END IF;

            END LOOP;


            -- Call framework
            --
            Compute_Simulation_Time_Line (
               Sys,  A_Response,

               -- With period
               --
               Period,

               -- With offsets
               --
               Schedule_With_Offsets,

               -- With Task precedencies
               --
               Schedule_With_Precedencies,

               -- With Shared resources
               --
               Schedule_With_Resources,

               -- With task specific seed
               --
               With_Task_Specific_Seed,

               -- With a global seed value
               --
               Global_Seed_Value,

               -- Predictable seed of not ?
               --
               Predictable_Global_Seed,

               -- Which event we should generate
               --
               Event_To_Generate,


               Output);



         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Simulation_Basics =>

            Compute_Simulation_Basics (
               Sys, A_Response, Request.Target,Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Simulation_Run_Event_Handler =>

            Run_User_Defined_Event_Handlers (Sys, A_Response, Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Simulation_Priority_Inversion =>

            Priority_Inversion_Detection  (
               Sys, A_Response, Request.Target, Output);


         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Feasibility_General_Task=>


            Partionning_Tasks(
               Sys, A_Response, General_Task, Output);


         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Feasibility_First_Fit =>


            Partionning_Tasks(
               Sys, A_Response, First_Fit, Output);


         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN  Scheduling_Feasibility_Next_Fit =>


            Partionning_Tasks(
               Sys, A_Response, Next_Fit, Output);


         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN  Scheduling_Feasibility_Best_Fit =>


            Partionning_Tasks(
               Sys, A_Response, Best_Fit, Output);


         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN  Scheduling_Feasibility_Small_Task =>

            Partionning_Tasks(
               Sys, A_Response, Small_Task, Output);

         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Simulation_Deadlock =>

            Deadlock_Detection  (
               Sys, A_Response, Request.Target,Output);







         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Simulation_Context_Switch_Number =>

            Compute_Simulation_Number_Of_Context_Switch (
               Sys, A_Response, Request.Target,Output);









         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN  Scheduling_Simulation_Preemption_Number =>

            Compute_Simulation_Number_Of_Preemption (
               Sys, A_Response, Request.Target, Output);








         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Feasibility_Response_Time   =>


            IF (Get_Number_Of_Task_From_Processor(Sys.Tasks, Request.
                     Target.Name)/=0)
                  THEN


               Compute_Feasibility_Response_Time(
                  Sys, A_Response, Request.Target, Output);

            END IF;

         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Feasibility_Cpu_Utilization =>


            IF (Get_Number_Of_Task_From_Processor(Sys.Tasks, Request.
                     Target.Name)/=0)
                  THEN

               Compute_Feasibility_Processor_Utilization_Factor(
                  Sys, A_Response, Request.Target, Output);

            END IF;


         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Simulation_Response_Time =>


            -- Check parameters
            --
            IF (Request.Param.Nb_Entries/=3)
                  THEN
               Raise_Exception(Invalid_Number_Of_Parameter'Identity,
                  "3 expected parameters");
            END IF;

            IF (Request.Param.Entries(0).Name/=To_Unbounded_String(
                     "worst_case")) THEN
               Raise_Exception(Unexpected_Parameter'Identity,
                  "worst_case");
            END IF;

            IF (Request.Param.Entries(1).Name/=To_Unbounded_String(
                     "best_case")) THEN
               Raise_Exception(Unexpected_Parameter'Identity,
                  "best_case");
            END IF;

            IF (Request.Param.Entries(2).Name/=To_Unbounded_String(
                     "average_case")) THEN
               Raise_Exception(Unexpected_Parameter'Identity,
                  "average_case");
            END IF;

            IF (Request.Param.Entries(0).Ptype/=Boolean_Parameter) THEN
               Raise_Exception(Parameter_Type_Error'Identity,
                  "worst_case");
            END IF;
            IF (Request.Param.Entries(1).Ptype/=Boolean_Parameter) THEN
               Raise_Exception(Parameter_Type_Error'Identity,
                  "best_case");
            END IF;
            IF (Request.Param.Entries(2).Ptype/=Boolean_Parameter) THEN
               Raise_Exception(Parameter_Type_Error'Identity,
                  "average_case");
            END IF;

            Compute_Simulation_Response_Time (
               Sys, A_Response, Request.Target,
               Request.Param.Entries(0).Boolean_Value,
               Request.Param.Entries(1).Boolean_Value,
               Request.Param.Entries(2).Boolean_Value,
               Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Simulation_Blocking_Time =>


            -- Check parameters
            --
            IF (Request.Param.Nb_Entries/=3)
                  THEN
               Raise_Exception(Invalid_Number_Of_Parameter'Identity,
                  "3 expected parameters");
            END IF;

            IF (Request.Param.Entries(0).Name/=To_Unbounded_String(
                     "worst_case")) THEN
               Raise_Exception(Unexpected_Parameter'Identity,
                  "worst_case");
            END IF;

            IF (Request.Param.Entries(1).Name/=To_Unbounded_String(
                     "best_case")) THEN
               Raise_Exception(Unexpected_Parameter'Identity,
                  "best_case");
            END IF;

            IF (Request.Param.Entries(2).Name/=To_Unbounded_String(
                     "average_case")) THEN
               Raise_Exception(Unexpected_Parameter'Identity,
                  "average_case");
            END IF;

            IF (Request.Param.Entries(0).Ptype/=Boolean_Parameter) THEN
               Raise_Exception(Parameter_Type_Error'Identity,
                  "worst_case");
            END IF;
            IF (Request.Param.Entries(1).Ptype/=Boolean_Parameter) THEN
               Raise_Exception(Parameter_Type_Error'Identity,
                  "best_case");
            END IF;
            IF (Request.Param.Entries(2).Ptype/=Boolean_Parameter) THEN
               Raise_Exception(Parameter_Type_Error'Identity,
                  "average_case");
            END IF;

            Compute_Simulation_Blocking_Time (Sys, A_Response,
               Request.Target,
               Request.Param.Entries(0).Boolean_Value,
               Request.Param.Entries(1).Boolean_Value,
               Request.Param.Entries(2).Boolean_Value,
               Output);



         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Simulation_All_Response_Time =>

            NULL;




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN  Scheduling_Feasibility_Compute_And_Set_Blocking_Time =>

            Compute_Bound_On_Blocking_Time (Sys, A_Response, Request.
               Target, True,
               Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN   Scheduling_Feasibility_Compute_Blocking_Time =>

            Compute_Bound_On_Blocking_Time (Sys, A_Response, Request.
               Target, False,
               Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Feasibility_Basics =>

            IF (Get_Number_Of_Task_From_Processor(Sys.Tasks, Request.
                     Target.Name)/=0)
                  THEN

               Compute_Feasibility_Basics (
                  Sys, A_Response, Request.Target,Output);

            END IF;




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN
               Scheduling_Set_Priorities_According_To_Deadline_Monotonic =>

            Set_Priorities (
               Sys, A_Response, Request.Target,
               Deadline_Monotonic_Protocol,
               Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Scheduling_Set_Priorities_According_To_Rate_Monotonic =>

            Set_Priorities (
               Sys, A_Response, Request.Target, Rate_Monotonic_Protocol,
               Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Buffer_Feasibility_Tests =>

            IF (Get_Number_Of_Buffer_From_Processor (Sys.Buffers,
                     Request.
                     Target.
                     Name) /= 0 )
                  THEN

               Compute_Buffer_Feasibility_Tests (Sys, A_Response, Request.
                  Target,
                  Output);
            END IF;


         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Global_Memory_Compute_Requirement_Analysis  =>
            Compute_Memory_Requirement_Analysis (Sys, A_Response,  False,
               Output);

         WHEN Global_Memory_Set_Requirement_Analysis  =>
            Compute_Memory_Requirement_Analysis (Sys, A_Response, True,
               Output);

         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Buffer_Scheduling_Simulation   =>
            IF (Get_Number_Of_Buffer_From_Processor (Sys.Buffers,
                     Request.
                     Target.
                     Name) /= 0 )                  THEN
               Compute_Buffer_Scheduling_Simulation(Sys, A_Response,
                  Request.
                  Target,
                  Output);
            END IF;


         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Random_Response_Time_Density    =>

            Compute_Response_Time_Density (
               Sys, A_Response, Request.Target, Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Dependency_Compute_Chetto_Blazewicz_Priority  =>

            Chetto_Parameter_Modifications(
               Sys, A_Response, Request.Target, False, True, False,
               Output);



         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Dependency_Compute_Chetto_Blazewicz_Deadline   =>

            Chetto_Parameter_Modifications(
               Sys, A_Response, Request.Target, True, False, False,
               Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Dependency_Set_Chetto_Blazewicz_Deadline =>

            Chetto_Parameter_Modifications(
               Sys, A_Response, Request.Target, True, False, True,
               Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Dependency_Set_Chetto_Blazewicz_Priority  =>

            Chetto_Parameter_Modifications(
               Sys, A_Response, Request.Target, False, True, True,
               Output);





         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN  Dependency_Compute_End_To_End_Response_Time_One_Step =>

            Compute_End_To_End_Response_Time (
               Sys, A_Response, False, True, Output);



         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Dependency_Set_End_To_End_Response_Time_One_Step   =>

            Compute_End_To_End_Response_Time (
               Sys, A_Response, True, True, Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN  Dependency_Compute_End_To_End_Response_Time_All_Steps  =>

            Compute_End_To_End_Response_Time (
               Sys, A_Response, False, False, Output);




         -----------------------------------------------------------
         -----------------------------------------------------------
         WHEN Dependency_Set_End_To_End_Response_Time_All_Steps   =>

            Compute_End_To_End_Response_Time (
               Sys, A_Response, True, False, Output);




      END CASE;
   END Call_A_Request;








   PROCEDURE Initialize_Before_Request (
         Request : IN     Framework_Request) IS
   BEGIN
      CASE Request.Statement IS
         WHEN Scheduling_Simulation_Time_Line =>
            Initialize(Sched.All);
         WHEN Scheduling_Simulation_Basics =>
            NULL;
         WHEN  Scheduling_Simulation_Preemption_Number =>
            NULL;
         WHEN  Scheduling_Simulation_Context_Switch_Number =>
            NULL;
         WHEN Scheduling_Simulation_Response_Time =>
            NULL;
         WHEN Scheduling_Simulation_All_Response_Time =>
            NULL;
         WHEN Scheduling_Simulation_Blocking_Time =>
            NULL;
         WHEN Scheduling_Simulation_Run_Event_Handler =>
            NULL;
         WHEN Scheduling_Feasibility_First_Fit =>
            NULL;
         WHEN Scheduling_Feasibility_Next_Fit =>
            NULL;
         WHEN  Scheduling_Feasibility_Best_Fit =>
            NULL;
         WHEN  Scheduling_Feasibility_Small_Task =>
            NULL;
         WHEN Scheduling_Feasibility_General_Task =>
            NULL;
         WHEN Scheduling_Feasibility_Response_Time   =>
            NULL;
         WHEN Scheduling_Feasibility_Cpu_Utilization =>
            NULL;
         WHEN Scheduling_Feasibility_Basics =>
            NULL;
         WHEN
               Scheduling_Set_Priorities_According_To_Deadline_Monotonic =>
            NULL;
         WHEN
               Scheduling_Set_Priorities_According_To_Rate_Monotonic =>
            NULL;
         WHEN
               Scheduling_Feasibility_Compute_And_Set_Blocking_Time =>
            NULL;
         WHEN  Scheduling_Feasibility_Compute_Blocking_Time =>
            NULL;
         WHEN Scheduling_Simulation_Priority_Inversion =>
            NULL;
         WHEN Scheduling_Simulation_Deadlock =>
            NULL;


         WHEN  Global_Memory_Set_Requirement_Analysis =>
            NULL;
         WHEN  Global_Memory_Compute_Requirement_Analysis =>
            NULL;
         WHEN Buffer_Feasibility_Tests =>
            NULL;
         WHEN Buffer_Scheduling_Simulation  =>
            Initialize(Buff);


         WHEN Random_Response_Time_Density    =>
            NULL;


         WHEN Dependency_Set_Chetto_Blazewicz_Deadline =>
            NULL;
         WHEN Dependency_Compute_End_To_End_Response_Time_One_Step =>
            NULL;
         WHEN Dependency_Set_End_To_End_Response_Time_One_Step   =>
            NULL;
         WHEN Dependency_Compute_End_To_End_Response_Time_All_Steps  =>
            NULL;
         WHEN Dependency_Set_End_To_End_Response_Time_All_Steps   =>
            NULL;
         WHEN Dependency_Compute_Chetto_Blazewicz_Priority  =>
            NULL;
         WHEN Dependency_Compute_Chetto_Blazewicz_Deadline   =>
            NULL;
         WHEN Dependency_Set_Chetto_Blazewicz_Priority  =>
            NULL;
      END CASE;
   END Initialize_Before_Request;



   FUNCTION Build_A_Title (
         Request : IN     Framework_Request)
     RETURN Unbounded_String IS
   BEGIN

      CASE Request.Statement IS

         WHEN Scheduling_Simulation_Time_Line =>
            RETURN Empty_String;

         WHEN Scheduling_Simulation_Basics =>
            RETURN Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Simulation(
               Current_Language);

         WHEN  Scheduling_Simulation_Preemption_Number =>
            RETURN Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Simulation(
               Current_Language);

         WHEN  Scheduling_Simulation_Context_Switch_Number =>
            RETURN Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Simulation(
               Current_Language);

         WHEN Scheduling_Simulation_All_Response_Time =>
            RETURN Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Simulation(
               Current_Language);

         WHEN Scheduling_Simulation_Priority_Inversion =>
            RETURN Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Simulation(
               Current_Language);

         WHEN Scheduling_Simulation_Deadlock =>
            RETURN Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Simulation(
               Current_Language);

         WHEN Scheduling_Simulation_Response_Time =>
            RETURN Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Simulation(
               Current_Language);

         WHEN Scheduling_Simulation_Blocking_Time =>
            RETURN Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Simulation(
               Current_Language);

         WHEN Scheduling_Simulation_Run_Event_Handler =>
            RETURN Start_Sched_Sim & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Simulation(
               Current_Language) &
               Lb_Comma &
               Lb_Event_Analyzer(Current_Language) ;

         WHEN Scheduling_Feasibility_First_Fit =>
            RETURN Start_Block & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Feasibility(Current_Language) &
               Lb_Comma &
               Lb_Partition(Current_Language) & Lb_Partition_First_Fit(
               Current_Language);

         WHEN Scheduling_Feasibility_General_Task =>
            RETURN Start_Block & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Feasibility(Current_Language) &
               Lb_Comma &
               Lb_Partition(Current_Language) &
               Lb_Partition_General_Task(
               Current_Language);

         WHEN  Scheduling_Feasibility_Best_Fit =>
            RETURN Start_Block & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Feasibility(Current_Language) &
               Lb_Comma &
               Lb_Partition(Current_Language) & Lb_Partition_Best_Fit(
               Current_Language);

         WHEN Scheduling_Feasibility_Next_Fit =>
            RETURN Start_Block & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Feasibility(Current_Language) &
               Lb_Comma &
               Lb_Partition(Current_Language) & Lb_Partition_Next_Fit(
               Current_Language);

         WHEN  Scheduling_Feasibility_Small_Task =>
            RETURN Start_Block & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Feasibility(Current_Language) &
               Lb_Comma &
               Lb_Partition(Current_Language) & Lb_Partition_Small_Task(
               Current_Language);

         WHEN Scheduling_Feasibility_Compute_And_Set_Blocking_Time =>
            RETURN Start_Block & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Feasibility(Current_Language);

         WHEN  Scheduling_Feasibility_Compute_Blocking_Time =>
            RETURN Start_Block & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Feasibility(Current_Language);

         WHEN Scheduling_Feasibility_Response_Time   =>
            RETURN Start_Sched_Feas & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Feasibility(
               Current_Language);

         WHEN Scheduling_Feasibility_Basics =>
            RETURN Start_Sched_Feas & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Feasibility(Current_Language);

         WHEN Scheduling_Feasibility_Cpu_Utilization =>
            RETURN Start_Sched_Feas & Unbounded_Lf & Start_Title &
               Lb_Scheduling_Feasibility(Current_Language);

         WHEN
               Scheduling_Set_Priorities_According_To_Deadline_Monotonic =>
            RETURN Start_Set_Prio & Unbounded_Lf & Start_Title &
               Unbounded_Lf & Lb_Set_Priorities(Current_Language);

         WHEN Scheduling_Set_Priorities_According_To_Rate_Monotonic =>
            RETURN Start_Set_Prio & Unbounded_Lf & Start_Title &
               Unbounded_Lf & Lb_Set_Priorities(Current_Language);


         WHEN  Global_Memory_Set_Requirement_Analysis =>
            RETURN Start_Buf_Bound & Unbounded_Lf & Start_Title &
               Lb_Memory_Requirement_Analysis(
               Current_Language);
         WHEN  Global_Memory_Compute_Requirement_Analysis =>
            RETURN Start_Buf_Bound & Unbounded_Lf & Start_Title &
               Lb_Memory_Requirement_Analysis(
               Current_Language);
         WHEN Buffer_Feasibility_Tests =>
            RETURN Start_Buf_Bound & Unbounded_Lf & Start_Title &
               Lb_Compute_Buffer(Current_Language);
         WHEN Buffer_Scheduling_Simulation   =>
            RETURN Unbounded_Lf & Lb_Draw_Buffer(Current_Language);



         WHEN Random_Response_Time_Density    =>
            RETURN Start_Response_Time_Density & Start_Title &
               Unbounded_Lf & Lb_Response_Time_Density(
               Current_Language);


         WHEN Dependency_Compute_End_To_End_Response_Time_One_Step =>
            RETURN  Unbounded_Lf & Lb_Jitter_From_Response_Time(
               Current_Language);
         WHEN Dependency_Set_End_To_End_Response_Time_One_Step   =>
            RETURN  Unbounded_Lf & Lb_Jitter_From_Response_Time(
               Current_Language);
         WHEN Dependency_Compute_End_To_End_Response_Time_All_Steps  =>
            RETURN Start_End_To_End & Unbounded_Lf & Start_Title &
               Lb_Jitter_From_Response_Time(
               Current_Language);
         WHEN Dependency_Set_End_To_End_Response_Time_All_Steps   =>
            RETURN Start_End_To_End & Unbounded_Lf & Start_Title &
               Lb_Jitter_From_Response_Time(
               Current_Language);

         WHEN Dependency_Compute_Chetto_Blazewicz_Priority  =>
            RETURN Lb_Chetto(Current_Language);
         WHEN Dependency_Compute_Chetto_Blazewicz_Deadline   =>
            RETURN Lb_Chetto(Current_Language);
         WHEN Dependency_Set_Chetto_Blazewicz_Priority  =>
            RETURN Lb_Chetto(Current_Language);
         WHEN Dependency_Set_Chetto_Blazewicz_Deadline =>
            RETURN Lb_Chetto(Current_Language);

      END CASE;

   END Build_A_Title;


   PROCEDURE Sequential_Framework_Request (
         Sys      : IN OUT System_Ptr;
         Request  : IN     Framework_Request_Table;
         Response : IN OUT Framework_Response_Table;
         Order    : IN     Perform_Order;
         Output   : IN     Output_Format) IS
   BEGIN
      Sequential_Framework_Request(Sys.All, Request, Response,
         Order,
         Output);
   END  Sequential_Framework_Request;


   PROCEDURE Sequential_Framework_Request (
         Sys      : IN OUT System;
         Request  : IN     Framework_Request_Table;
         Response : IN OUT Framework_Response_Table;
         Order    : IN     Perform_Order;
         Output   : IN     Output_Format) IS

      A_Response  : Framework_Response;
      A_Processor : Processor_Ptr;
      My_Iterator : Processors_Iterator;
      A_Request   : Sub_Framework_Request;

   BEGIN

      FOR I IN 0..Request.Nb_Entries-1 LOOP


         Initialize_Before_Request(Request.Entries(I));



         -- Here, we should take care of two cases :
         -- case 1 : the called service should be run for
         --          each processor (that's the most of services)
         -- case 2 : the called service is global to all processor and
         --          should be called one time only
         --
         --
         IF (Request.Entries(I).Statement=
               Scheduling_Simulation_Run_Event_Handler)
               OR (Request.Entries(I).Statement=
               Global_Memory_Set_Requirement_Analysis)
               OR (Request.Entries(I).Statement=
               Global_Memory_Compute_Requirement_Analysis)
               OR (Request.Entries(I).Statement=
               Scheduling_Simulation_Time_Line)
               OR (Request.Entries(I).Statement=
               Scheduling_Feasibility_Small_Task)
               OR (Request.Entries(I).Statement=
               Scheduling_Feasibility_Best_Fit)
               OR (Request.Entries(I).Statement=
               Scheduling_Feasibility_Next_Fit)
               OR (Request.Entries(I).Statement=
               Scheduling_Feasibility_First_Fit)
               OR (Request.Entries(I).Statement=
               Scheduling_Feasibility_General_Task)
               OR (Request.Entries(I).Statement=
               Dependency_Compute_End_To_End_Response_Time_One_Step)
               OR (Request.Entries(I).Statement=
               Dependency_Set_End_To_End_Response_Time_One_Step)
               OR (Request.Entries(I).Statement=
               Dependency_Compute_End_To_End_Response_Time_All_Steps)
               OR (Request.Entries(I).Statement=
               Dependency_Set_End_To_End_Response_Time_All_Steps)
               OR (Request.Entries(I).Statement=
               Dependency_Set_Chetto_Blazewicz_Deadline)
               OR (Request.Entries(I).Statement=
               Dependency_Compute_Chetto_Blazewicz_Priority)
               OR (Request.Entries(I).Statement=
               Dependency_Compute_Chetto_Blazewicz_Deadline)
               OR (Request.Entries(I).Statement=
               Dependency_Set_Chetto_Blazewicz_Priority)

               THEN

            -- Only one call for all processors
            --
            Initialize(A_Response);
            A_Request.Param:=Request.Entries(I).Param;
            A_Request.Statement:=Request.Entries(I).Statement;

            Call_A_Request(Sys, A_Request,
               A_Response.Text, Output);

            IF A_Response.Text/=Empty_String THEN

               A_Response.Title:=Build_A_Title (Request.Entries(
                     I));
               IF  A_Response.Title/=Empty_String THEN
                  A_Response.Title:= Unbounded_Lf &
                     Unbounded_Lf &
                     A_Response.Title &
                     Lb_Colon & Unbounded_Lf & End_Title;
               END IF;

               Add(Response, A_Response);
            END IF;

         ELSE

            -- A call for each processor
            --
            Reset_Iterator(Sys.Processors, My_Iterator);


            LOOP
               Current_Element(Sys.Processors, A_Processor,
                  My_Iterator);

               -- Do a request
               --
               IF (A_Processor.Name=Request.Entries(I).Target) OR
                     (Request.Entries(I).Target=Empty_String) THEN

                  Initialize(A_Response);
                  A_Request.Target:=A_Processor;
                  A_Request.Param:=Request.Entries(I).Param;
                  A_Request.Statement:=Request.Entries(I).
                     Statement;

                  Call_A_Request(Sys, A_Request,
                     A_Response.Text, Output);

                  IF A_Response.Text/=Empty_String THEN

                     A_Response.Title:=Build_A_Title (Request.
                        Entries(
                           I));
                     IF  A_Response.Title/=Empty_String THEN
                        A_Response.Title:=
                           Unbounded_Lf &
                           A_Response.Title & Lb_Comma &
                           Lb_Processor(Current_Language) &
                           Start_Proc & A_Processor.Name &
                           End_Proc &
                           Lb_Colon & Unbounded_Lf & End_Title;
                     END IF;

                     Add(Response, A_Response);
                  END IF;

               END IF;

               EXIT WHEN Is_Last_Element(Sys.Processors,
                  My_Iterator);

               Next_Element(Sys.Processors, My_Iterator);
            END LOOP;
         END IF;


      END LOOP;
   END Sequential_Framework_Request;



   PROCEDURE Parallel_Framework_Request (
         Sys      : IN OUT System;
         Request  : IN     Framework_Request_Table;
         Response : IN OUT Framework_Response_Table;
         Order    : IN     Perform_Order            := Total_Order;
         Output   : IN     Output_Format            := String_Output) IS
   BEGIN
      NULL;
   END Parallel_Framework_Request;



   PROCEDURE Distributed_Framework_Request (
         Sys      : IN OUT System;
         Request  : IN     Framework_Request_Table;
         Response : IN OUT Framework_Response_Table;
         Order    : IN     Perform_Order            := Total_Order;
         Output   : IN     Output_Format            := String_Output) IS
   BEGIN
      NULL;
   END Distributed_Framework_Request ;






END Call_Framework;


