------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Translate;
USE Translate;

WITH Xml_Tag;
USE Xml_Tag;
WITH Task_Set;
USE Task_Set;

WITH Buffers;
USE Buffers;
USE Buffers.Buffer_Roles;
WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;

WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Call_Scheduling_Framework;
USE Call_Scheduling_Framework;

WITH Multiprocessor_Services;
USE Multiprocessor_Services;
USE Multiprocessor_Services.Scheduling_Result_Per_Processor;

WITH Scheduling_Sequences;
USE Scheduling_Sequences;
USE Scheduling_Sequences.Buffer_Size_Package;
USE Scheduling_Sequences.Waiting_Time_Package;

WITH Tasks;
USE Tasks;

WITH Queueing_System;
USE Queueing_System;
WITH Queueing_System.Theoritical;
USE Queueing_System.Theoritical;
WITH Queueing_System.Theoritical.Mp1;
USE Queueing_System.Theoritical.Mp1;
WITH Queueing_System.Theoritical.Pp1;
USE Queueing_System.Theoritical.Pp1;
WITH Queueing_System.Theoritical.Mm1;
USE Queueing_System.Theoritical.Mm1;
WITH Queueing_System.Theoritical.Md1;
USE Queueing_System.Theoritical.Md1;
WITH Queueing_System.Theoritical.Mg1;
USE Queueing_System.Theoritical.Mg1;

WITH Double_Util;
USE Double_Util;


WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;

WITH Text_Io;
USE Text_Io;


PACKAGE BODY Call_Memory_Framework IS




   PROCEDURE Compute_Buffer_Scheduling_Simulation (
         Sys         : IN     System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         Output      : IN     Output_Format    := String_Output) IS


      A_Buffer        : Buffer_Ptr;
      My_Iterator     : Buffers_Iterator;
      Index           : Scheduling_Table_Range := 0;
      A_Buffer_Result : Buffer_Size_Table_Ptr;
      A_Task          : Task_Ptr;

      Maximum_Size         : Natural;
      Maximum_Waiting_Time : Natural;
      Average_Size         : Double;
      Average_Waiting_Time : Double;

      Average_Consumer_Period : Double := 0.0;
      Sum_Period              : Double := 0.0;
      Nb_Consumer             : Double := 0.0;

   BEGIN


      -- find the processor on which we want to compute buffer utilization
      --
      FOR I IN 0..Call_Scheduling_Framework.Sched.Nb_Entries-1 LOOP

         IF Sched.Entries(I).Item.Name=A_Processor.Name THEN
            IF Sched.Entries(I).Data.Error_Msg=Empty_String THEN
               Index:=I;
               EXIT;
            END IF;
         END IF;
      END LOOP;


      -- Compute buffer utilization on all buffer of the
      -- selected processor
      --



      A_Buffer_Result := NEW Buffer_Size_Table;


      LOOP
         Current_Element(Sys.Buffers, A_Buffer, My_Iterator);


         IF A_Processor.Name = A_Buffer.Cpu_Name
               THEN

            Initialize(A_Buffer_Result.All);

            Compute_Buffer_Size_From_Simulation(Sched.Entries(Index).
               Data.Result,
               A_Buffer, A_Buffer_Result);

            Add(Buff, A_Buffer, A_Buffer_Result.All);


            -- Compute average consumer period
            --
            Average_Consumer_Period :=0.0;
            Sum_Period  :=0.0;
            Nb_Consumer  :=0.0;
            FOR J IN  0 .. A_Buffer.Roles.Nb_Entries - 1 LOOP
               IF (A_Buffer.Roles.Entries(J).Data.The_Role =
                     Consumer)
                     THEN

                  A_Task:=Search_Task(Sys.Tasks, A_Buffer.Roles.Entries(J).
                     Item);
                  IF (A_Task.Task_Type = Periodic_Type)
                        THEN
                     Sum_Period:=+Double(Periodic_Task_Ptr(A_Task).Period);
                     Nb_Consumer:= Nb_Consumer+1.0;
                  END IF;

               END IF;
            END LOOP;

            Average_Consumer_Period:=Sum_Period/Nb_Consumer;


            -- Compute results
            --
            Maximum_Size:=Compute_Maximum_Buffer_Size_From_Simulation (
               A_Buffer_Result);
            Average_Size:=Compute_Average_Buffer_Size_From_Simulation (
               A_Buffer_Result);
            Maximum_Waiting_Time:=
               Compute_Maximum_Waiting_Time_From_Simulation (
               A_Buffer_Result, Average_Consumer_Period);
            Average_Waiting_Time:=
               Compute_Average_Waiting_Time_From_Simulation (
               A_Buffer_Result, Average_Consumer_Period);


            -- Display results
            --
            Result:= Result & Unbounded_Lf & Unbounded_Lf &
               Start_Task & Lb_Buffer(
               Current_Language) & " " &
               A_Buffer.Name &
               To_Unbounded_String(" => ") & Unbounded_Lf;


            Result:= Result & To_Unbounded_String("    ") &
               Lb_Maximum_Buffer_Size(
               Current_Language) &
               Maximum_Size'Img & Unbounded_Lf  &
               To_Unbounded_String("    ") &
               Lb_Maximum_Waiting_Time(
               Current_Language) &
               Maximum_Waiting_Time'Img & Unbounded_Lf  ;

            Result:= Result & To_Unbounded_String("    ") &
               Lb_Average_Buffer_Size(
               Current_Language) &
               Format(Average_Size,8) & Unbounded_Lf  &
               To_Unbounded_String("    ") &
               Lb_Average_Waiting_Time(
               Current_Language) &
               Format(Average_Waiting_Time,8) & Unbounded_Lf  ;


         END IF;

         EXIT WHEN Is_Last_Element(Sys.Buffers, My_Iterator);
         Next_Element(Sys.Buffers, My_Iterator);

      END LOOP;

      Free(A_Buffer_Result);

   END Compute_Buffer_Scheduling_Simulation;







   PROCEDURE Periodic_Buffer_Control (
         My_Tasks : IN     Tasks_Set;
         A_Buffer : IN     Buffer_Ptr) IS


      Iterator1 : Tasks_Iterator;
      Task1     : Task_Ptr;

   BEGIN

      FOR I IN 0..A_Buffer.Roles.Nb_Entries-1  LOOP


         Reset_Iterator(My_Tasks, Iterator1);

         LOOP
            Current_Element(My_Tasks, Task1, Iterator1);

            IF (Task1.Name = A_Buffer.Roles.Entries(I).Item) THEN
               IF  (Task1.Task_Type /= Periodic_Type)
                     THEN
                  RAISE Task_Must_Be_Periodic;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Tasks, Iterator1);
            Next_Element(My_Tasks, Iterator1);

         END LOOP;

      END LOOP;

   END Periodic_Buffer_Control;





   PROCEDURE Compute_Buffer_Feasibility_Tests (
         Sys         : IN     System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         Output      : IN     Output_Format    := String_Output) IS



      A_Buffer : Buffer_Ptr;

      Waiting_Time : Double := 0.0;
      Occupation   : Double := 0.0;

      Reference,
      Ref1,
      Ref2        : Unbounded_String;
      My_Iterator : Iterator;

      Nb_Arrival : Double := 0.0;
      Nb_Server  : Double := 0.0;

      My_Consumer_Task : Task_Ptr;

      Service_Rate : Double := 0.0;
      Arrival_Rate : Double := 0.0;

      A_Mm1_Queueing_System : Mm1_Queueing_System_Theoritical;
      A_Md1_Queueing_System : Md1_Queueing_System_Theoritical;
      A_Mg1_Queueing_System : Mg1_Queueing_System_Theoritical;
      A_Mp1_Queueing_System : Mp1_Queueing_System_Theoritical;
      A_Pp1_Queueing_System : Pp1_Queueing_System_Theoritical;

      No_Performance_Criteria : Boolean := False;


   BEGIN

      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;


      Result :=  Unbounded_Lf & Start_Line & Unbounded_Lf;



      Reset_Iterator(Sys.Buffers, My_Iterator);
      LOOP
         Current_Element(Sys.Buffers, A_Buffer, My_Iterator);
         IF (A_Buffer.Cpu_Name = A_Processor.Name) THEN
            IF (Natural(A_Buffer.Roles.Nb_Entries) >= 1)
                  THEN
               BEGIN

                  Result:= Result & Unbounded_Lf & Unbounded_Lf &
                     Start_Task & Lb_Buffer(
                     Current_Language) & " " &
                     A_Buffer.Name &
                     To_Unbounded_String(" => (") &
                     Qs_To_Display_String(
                     A_Buffer.Queueing_System_Type)
                     & To_Unbounded_String(")") & Unbounded_Lf;

                  Ref1:= Empty_String;
                  Ref2:= Empty_String;
                  Reference:= Empty_String;
                  Nb_Arrival:=0.0;
                  Nb_Server:=0.0;
                  No_Performance_Criteria:=False;


                  -- Number of producer (arrival) and consumer (server)
                  --
                  FOR J IN  0 .. A_Buffer.Roles.Nb_Entries - 1 LOOP
                     IF (A_Buffer.Roles.Entries(J).Data.The_Role =
                           Producer)
                           THEN
                        Nb_Arrival := Nb_Arrival + 1.0;
                     END IF;
                     IF (A_Buffer.Roles.Entries(J).Data.The_Role =
                           Consumer)
                           THEN
                        My_Consumer_Task := Search_Task(Sys.Tasks,
                           A_Buffer.Roles.
                           Entries(J).Item);

                        Nb_Server := Nb_Server + 1.0;
                     END IF;
                  END LOOP;

                  -- Control before buffer computation
                  --
                  Periodic_Buffer_Control (Sys.Tasks, A_Buffer);
                  Buffer_Flow_Control (A_Buffer, Sys.Tasks,
                     Service_Rate,
                     Arrival_Rate);


                  CASE A_Buffer.Queueing_System_Type IS


                     WHEN Qs_Mm1 =>

                        -- initalization
                        --
                        Reset(A_Mm1_Queueing_System);
                        Set_Qs_Arrival_Rate(A_Mm1_Queueing_System,
                           Arrival_Rate);
                        Set_Qs_Service_Rate(A_Mm1_Queueing_System,
                           Service_Rate);
                        Set_Qs_Nb_Arrival(A_Mm1_Queueing_System,
                           Nb_Arrival);
                        Set_Qs_Nb_Server(A_Mm1_Queueing_System,
                           Nb_Server);
                        Set_Qs_Harmonic(A_Mm1_Queueing_System,
                           Is_Cons_Prod_Harmonic(A_Buffer,Sys.Tasks));

                        -- consumer response time
                        --
                        Set_Constant_Consumer_Response_Time (
                           A_Mm1_Queueing_System,
                           Double(My_Consumer_Task.Deadline));


                        IF (Nb_Server <= 1.0) THEN

                           -- occupation computation
                           --
                           Queueing_System.Theoritical.Mm1.Qs_Average_Number_Customer (
                              A_Mm1_Queueing_System,
                              Occupation);

                           -- waiting time computation
                           --
                           Queueing_System.Theoritical.Mm1.Qs_Average_Waiting_Time (
                              A_Mm1_Queueing_System,
                              Waiting_Time);


                           -- Update message to be displayed
                           --
                           Ref1:=To_Unbounded_String(" (") & Lb_See(
                              Current_Language) &
                              To_Unbounded_String("[12], page 96, ") &
                              Lb_Theorem(
                              Current_Language) &
                              To_Unbounded_String("3.24). ");


                           Ref2:=To_Unbounded_String(" (") & Lb_See(
                              Current_Language) &
                              To_Unbounded_String("[12], page 98, ") &
                              Lb_Theorem(
                              Current_Language) &
                              To_Unbounded_String("3.26). ");


                           Reference:= To_Unbounded_String("    ") &
                              Lb_Average_Buffer_Size(
                              Current_Language) &
                              Format(Occupation,8) & Unbounded_Lf  &
                              To_Unbounded_String("    ") & Start_Ref &
                              Ref1 &
                              End_Ref  & Unbounded_Lf  &
                              To_Unbounded_String("    ") &
                              Lb_Average_Waiting_Time(
                              Current_Language) &
                              Format(Waiting_Time,8) & Unbounded_Lf  &
                              To_Unbounded_String("    ") & Start_Ref &
                              Ref2 &
                              End_Ref & End_Task &
                              Unbounded_Lf & End_Line &
                              End_Buf_Bound & Unbounded_Lf;

                        ELSE
                           No_Performance_Criteria:=True;
                        END IF;


                     WHEN Qs_Md1 =>

                        -- initalization
                        --
                        Reset(A_Md1_Queueing_System);
                        Set_Qs_Arrival_Rate(A_Md1_Queueing_System,
                           Arrival_Rate);
                        Set_Qs_Service_Rate(A_Md1_Queueing_System,
                           Service_Rate);
                        Set_Qs_Nb_Arrival(A_Md1_Queueing_System,
                           Nb_Arrival);
                        Set_Qs_Nb_Server(A_Md1_Queueing_System,
                           Nb_Server);
                        Set_Qs_Harmonic(A_Md1_Queueing_System,
                           Is_Cons_Prod_Harmonic(A_Buffer,Sys.Tasks));

                        -- consumer response time
                        --
                        Set_Constant_Consumer_Response_Time (
                           A_Md1_Queueing_System,
                           Double(My_Consumer_Task.Deadline));



                        IF (Nb_Server <= 1.0) THEN

                           -- occupation computation
                           --
                           Queueing_System.Theoritical.Md1.Qs_Average_Number_Customer (
                              A_Md1_Queueing_System,
                              Occupation);

                           -- waiting time computation
                           --
                           Queueing_System.Theoritical.Md1.Qs_Average_Waiting_Time (
                              A_Md1_Queueing_System,
                              Waiting_Time);


                           -- Update message to be displayed
                           --
                           Ref1:=To_Unbounded_String("to be filled");

                           Ref2:=To_Unbounded_String("to be filled");



                           Reference:= To_Unbounded_String("    ") &
                              Lb_Average_Buffer_Size(
                              Current_Language) &
                              Format(Occupation,8) & Unbounded_Lf  &
                              To_Unbounded_String("    ") & Start_Ref &
                              Ref1 &
                              End_Ref  & Unbounded_Lf  &
                              To_Unbounded_String("    ") &
                              Lb_Average_Waiting_Time(
                              Current_Language) &
                              Format(Waiting_Time,8) & Unbounded_Lf  &
                              To_Unbounded_String("    ") & Start_Ref &
                              Ref2 &
                              End_Ref & End_Task &
                              Unbounded_Lf & End_Line &
                              End_Buf_Bound & Unbounded_Lf;

                        ELSE
                           No_Performance_Criteria:=True;
                        END IF;


                     WHEN Qs_Mg1 =>

                        -- initalization
                        --
                        Reset(A_Mg1_Queueing_System);
                        Set_Qs_Arrival_Rate(A_Mg1_Queueing_System,
                           Arrival_Rate);
                        Set_Qs_Service_Rate(A_Mg1_Queueing_System,
                           Service_Rate);
                        Set_Qs_Nb_Arrival(A_Mg1_Queueing_System,
                           Nb_Arrival);
                        Set_Qs_Nb_Server(A_Mg1_Queueing_System,
                           Nb_Server);
                        Set_Qs_Harmonic(A_Mg1_Queueing_System,
                           Is_Cons_Prod_Harmonic(A_Buffer,Sys.Tasks));

                        -- consumer response time
                        --
                        Set_Constant_Consumer_Response_Time (
                           A_Mg1_Queueing_System,
                           Double(My_Consumer_Task.Deadline));



                        IF (Nb_Server <= 1.0) THEN

                           -- occupation computation
                           --
                           Queueing_System.Theoritical.Mg1.Qs_Average_Number_Customer (
                              A_Mg1_Queueing_System,
                              Occupation);

                           -- waiting time computation
                           --
                           Queueing_System.Theoritical.Mg1.Qs_Average_Waiting_Time (
                              A_Mg1_Queueing_System,
                              Waiting_Time);


                           -- Update message to be displayed
                           --
                           Ref1:=To_Unbounded_String("to be filled");

                           Ref2:=To_Unbounded_String("to be filled");


                           Reference:= To_Unbounded_String("    ") &
                              Lb_Average_Buffer_Size(
                              Current_Language) &
                              Format(Occupation,8) & Unbounded_Lf  &
                              To_Unbounded_String("    ") & Start_Ref &
                              Ref1 &
                              End_Ref  & Unbounded_Lf  &
                              To_Unbounded_String("    ") &
                              Lb_Average_Waiting_Time(
                              Current_Language) &
                              Format(Waiting_Time,8) & Unbounded_Lf  &
                              To_Unbounded_String("    ") & Start_Ref &
                              Ref2 &
                              End_Ref & End_Task &
                              Unbounded_Lf & End_Line &
                              End_Buf_Bound & Unbounded_Lf;

                        ELSE
                           No_Performance_Criteria:=True;
                        END IF;


                     WHEN Qs_Mp1 =>

                        -- initalization
                        --
                        Reset(A_Mp1_Queueing_System);
                        Set_Qs_Arrival_Rate(A_Mp1_Queueing_System,
                           Arrival_Rate);
                        Set_Qs_Service_Rate(A_Mp1_Queueing_System,
                           Service_Rate);
                        Set_Qs_Nb_Arrival(A_Mp1_Queueing_System,
                           Nb_Arrival);
                        Set_Qs_Nb_Server(A_Mp1_Queueing_System,
                           Nb_Server);
                        Set_Qs_Harmonic(A_Mp1_Queueing_System,
                           Is_Cons_Prod_Harmonic(A_Buffer,Sys.Tasks));

                        -- consumer response time
                        --
                        Set_Constant_Consumer_Response_Time (
                           A_Mp1_Queueing_System,
                           Double(My_Consumer_Task.Deadline));



                        IF (Nb_Server <= 1.0) THEN

                           -- occupation computation
                           --
                           Queueing_System.Theoritical.Mp1.Qs_Average_Number_Customer (
                              A_Mp1_Queueing_System,
                              Occupation);

                           -- waiting time computation
                           --
                           Queueing_System.Theoritical.Mp1.Qs_Average_Waiting_Time (
                              A_Mp1_Queueing_System,
                              Waiting_Time);


                           -- Update message to be displayed
                           --
                           Ref1:=To_Unbounded_String(" (") & Lb_See(
                              Current_Language) &
                              To_Unbounded_String("[10], ") &
                              Lb_Theorem(
                              Current_Language) &
                              To_Unbounded_String("6). ");

                           Ref2:=To_Unbounded_String(" (") & Lb_See(
                              Current_Language) &
                              To_Unbounded_String("[10], ") &
                              Lb_Theorem(
                              Current_Language) &
                              To_Unbounded_String("6). ");

                           Reference:= To_Unbounded_String("    ") &
                              Lb_Average_Buffer_Size(
                              Current_Language) &
                              Format(Occupation,8) & Unbounded_Lf  &
                              To_Unbounded_String("    ") & Start_Ref &
                              Ref1 &
                              End_Ref  & Unbounded_Lf  &
                              To_Unbounded_String("    ") &
                              Lb_Average_Waiting_Time(
                              Current_Language) &
                              Format(Waiting_Time,8) & Unbounded_Lf  &
                              To_Unbounded_String("    ") & Start_Ref &
                              Ref2 &
                              End_Ref & End_Task &
                              Unbounded_Lf & End_Line &
                              End_Buf_Bound & Unbounded_Lf;


                        ELSE
                           No_Performance_Criteria:=True;
                        END IF;


                     WHEN Qs_Pp1 =>

                        -- initalization
                        --
                        Reset(A_Pp1_Queueing_System);
                        Set_Qs_Arrival_Rate(A_Pp1_Queueing_System,
                           Arrival_Rate);
                        Set_Qs_Service_Rate(A_Pp1_Queueing_System,
                           Service_Rate);
                        Set_Qs_Nb_Arrival(A_Pp1_Queueing_System,
                           Nb_Arrival);
                        Set_Qs_Nb_Server(A_Pp1_Queueing_System,
                           Nb_Server);
                        Set_Qs_Harmonic(A_Pp1_Queueing_System,
                           Is_Cons_Prod_Harmonic(A_Buffer,Sys.Tasks));

                        -- consumer response time
                        --
                        Set_Constant_Consumer_Response_Time (
                           A_Mp1_Queueing_System,
                           Double(My_Consumer_Task.Deadline));


                        IF (Nb_Server <= 1.0) THEN

                           -- occupation computation
                           --
                           Queueing_System.Theoritical.Pp1.Qs_Maximum_Number_Customer (
                              A_Pp1_Queueing_System,
                              Occupation);

                           -- waiting time computation
                           --
                           Queueing_System.Theoritical.Pp1.Qs_Maximum_Waiting_Time (
                              A_Pp1_Queueing_System,
                              Waiting_Time);


                           -- Update message to be displayed
                           --
                           Ref1:=To_Unbounded_String(" (") & Lb_See(
                              Current_Language) &
                              To_Unbounded_String("[4,10], ") &
                              Lb_Theorem(
                              Current_Language) &
                              To_Unbounded_String("1 or 8). ");

                           Ref2:=To_Unbounded_String(" (") & Lb_See(
                              Current_Language) &
                              To_Unbounded_String("[10], ") &
                              Lb_Theorem(
                              Current_Language) &
                              To_Unbounded_String("8). ");

                           Reference:= To_Unbounded_String("    ") &
                              Lb_Maximum_Buffer_Size(
                              Current_Language) &
                              Format(Occupation,8) & Unbounded_Lf  &
                              To_Unbounded_String("    ") & Start_Ref &
                              Ref1 &
                              End_Ref  & Unbounded_Lf  &
                              To_Unbounded_String("    ") &
                              Lb_Maximum_Waiting_Time(
                              Current_Language) &
                              Format(Waiting_Time,8) & Unbounded_Lf  &
                              To_Unbounded_String("    ") & Start_Ref &
                              Ref2 &
                              End_Ref & End_Task &
                              Unbounded_Lf & End_Line &
                              End_Buf_Bound & Unbounded_Lf;

                        ELSE
                           No_Performance_Criteria:=True;
                        END IF;

                     WHEN OTHERS =>
                        No_Performance_Criteria:=True;

                  END CASE;


                  IF No_Performance_Criteria
                        THEN
                     Result:= Result & To_Unbounded_String("    ") &
                        Lb_Compute_Buffer_Error_4(
                        Current_Language) & Unbounded_Lf;
                  ELSE
                     Result:=Result & Reference;
                  END IF;

               EXCEPTION
                  WHEN Not_Implemented =>
                     Result:= Result & To_Unbounded_String("    ") &
                        Lb_Compute_Buffer_Error_4(
                        Current_Language) &
                        Unbounded_Lf;
                  WHEN Task_Set.Task_Model_Error=>
                     Result:= Result & To_Unbounded_String("    ") &
                        Lb_Compute_Buffer_Error_3(
                        Current_Language) &
                        Unbounded_Lf;
                  WHEN Task_Must_Be_Periodic=>
                     Result:=Result & To_Unbounded_String("    ") &
                        Lb_Compute_Buffer_Error_2(
                        Current_Language) &
                        Unbounded_Lf;
                  WHEN Buffers.Flow_Constraint_Not_Respected =>
                     Result:=Result & To_Unbounded_String("    ") &
                        Lb_Compute_Buffer_Error_1(
                        Current_Language) &
                        Unbounded_Lf;
               END;
            ELSE
               Result:= Result &  To_Unbounded_String("    ") &
                  A_Buffer.
                  Name
                  & To_Unbounded_String(" => NA");
            END IF;
         END IF;
         EXIT WHEN Is_Last_Element(Sys.Buffers, My_Iterator);
         Next_Element(Sys.Buffers, My_Iterator);
      END LOOP;

   END Compute_Buffer_Feasibility_Tests;


   PROCEDURE Compute_Memory_Requirement_Analysis (
         Sys                       : IN     System;
         Result                    : IN OUT Unbounded_String;
         Update_Address_Spaces_Set : IN     Boolean;
         Output                    : IN     Output_Format    := String_Output) IS
   BEGIN
      Result:= To_Unbounded_String("to be written");
   END  Compute_Memory_Requirement_Analysis;



END Call_Memory_Framework;




