------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Text_Io;
USE Text_Io;

WITH Xml_Tag;
USE Xml_Tag;

WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;

WITH Resource_Set;
USE Resource_Set;
USE Resource_Set.Generic_Resource_Set;

WITH Processors;
USE Processors;
WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;

WITH Translate;
USE Translate;


WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Framework_Config;
USE Framework_Config;


WITH Scheduler;
USE Scheduler;
WITH Scheduler.Fixed_Priority;
USE Scheduler.Fixed_Priority;

WITH Scheduling_Sequences;
USE Scheduling_Sequences;
USE Scheduling_Sequences.Deadlock_List_Package;
USE Scheduling_Sequences.Priority_Inversion_List_Package;

WITH Multiprocessor_Services;
USE Multiprocessor_Services;
USE Multiprocessor_Services.Scheduling_Result_Per_Processor;

WITH Call_Scheduling_Framework;
USE Call_Scheduling_Framework;

WITH Double_Util;
USE Double_Util;
WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;


PACKAGE BODY Call_Resource_Framework  IS




   PROCEDURE Compute_Bound_On_Blocking_Time (
         Sys                  : IN OUT System;
         Result               : IN OUT Unbounded_String;
         A_Processor          : IN     Processor_Ptr;
         Inject_Blocking_Time : IN     Boolean;
         Output               : IN     Output_Format    := String_Output) IS

      A_Task               : Task_Ptr;
      Has_Raised_Exception : Boolean                       := False;
      Msg                  : Unbounded_String;
      Blocking_Time        : Scheduler.Blocking_Time_Table;
      A_Scheduler          : Scheduler_Ptr;

   BEGIN


      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;


      -- And now, compute blocking time
      --
      Result:= Result & Start_Line & Lb_Minus & Lb_Bound_Blocking_Time(
         Current_Language) & Start_Ref & To_Unbounded_String(
         " (") &
         Lb_See(Current_Language) & To_Unbounded_String("[3]) ") &
         End_Ref & Lb_Colon & Unbounded_Lf;
      Msg:=Empty_String;
      A_Scheduler:=Get_Scheduler(A_Processor.All);
      Compute_Blocking_Time (A_Scheduler.All,
         Sys.Tasks,
         Sys.Resources,
         A_Processor.Name,
         Msg,
         Blocking_Time);
      IF Msg/=Empty_String
            THEN
         Has_Raised_Exception:=True;
         Result:=Result & Msg;
      ELSE
         Result:=Result & Unbounded_Lf;
      END IF;

      -- Display all task blocking time
      --

      FOR I IN 0..Scheduler.Blocking_Time_Range(
            Get_Number_Of_Task_From_Processor(
               Sys.
               Tasks,A_Processor.Name)-1) LOOP
         IF (Blocking_Time.Entries(I).Item /= NULL)
               THEN
            Result:=Result & Start_Task & To_Unbounded_String("    ") &
               Blocking_Time.
               Entries(
               I).
               Item.Name & To_Unbounded_String(" =>") &
               Integer'Image(Integer(Blocking_Time.Entries(I).Data)) &
               End_Task &  Unbounded_Lf;
         END IF;
      END LOOP;

      Result:= Result & Unbounded_Lf & End_Line & Unbounded_Lf &
         Unbounded_Lf;


      -- Update blocking time
      --
      IF Inject_Blocking_Time
            THEN

         IF Has_Raised_Exception
               THEN

            Result := Result   & Unbounded_Lf &
               Start_Line & Lb_Blocking_Time_Inject_Failed(
               Current_Language) & End_Line &
               Unbounded_Lf & End_Block & Unbounded_Lf;

         ELSE

            FOR I IN 0..Scheduler.Blocking_Time_Range(
                  Get_Number_Of_Task_From_Processor(
                     Sys.
                     Tasks,A_Processor.Name)-1) LOOP
               IF (Blocking_Time.Entries(I).Item /= NULL)
                     THEN

                  A_Task:=Search_Task(Sys.Tasks, Blocking_Time.Entries(I).
                     Item.Name);
                  A_Task.Blocking_Time :=Blocking_Time.Entries(I).Data;
               END IF;
            END LOOP;

            Result := Result   & Unbounded_Lf &
               Start_Line & Lb_Blocking_Time_Inject_Success(
               Current_Language) & End_Line &
               Unbounded_Lf & End_Block & Unbounded_Lf;
         END IF;
      END IF;


   EXCEPTION
      WHEN Invalid_Scheduler =>
         Result := Result &  Unbounded_Lf &
            Lb_Compute_Blocking_Error2(
            Current_Language) &
            Unbounded_Lf;


   END Compute_Bound_On_Blocking_Time;



   PROCEDURE Compute_Simulation_Blocking_Time (
         Sys          : IN OUT System;
         Result       : IN OUT Unbounded_String;
         A_Processor  : IN     Processor_Ptr;
         Worst_Case   : IN     Boolean;
         Best_Case    : IN     Boolean;
         Average_Case : IN     Boolean;
         Output       : IN     Output_Format    := String_Output) IS



      A_Task      : Task_Ptr;
      My_Iterator : Tasks_Iterator;
      Min         : Natural        := 0;
      Max         : Natural        := 0;
      Average     : Double         := 0.0;

   BEGIN


      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;


      FOR J IN 0.. Sched.Nb_Entries-1 LOOP
         IF Sched.Entries(J).Item.Name=A_Processor.Name THEN

            Result:= Result & Start_Line & Lb_Minus &
               Lb_Simulation_Blocking_Time(
               Current_Language) & Lb_Colon & Unbounded_Lf;


            Reset_Iterator (Sys.Tasks, My_Iterator);

            LOOP
               Current_Element(Sys.Tasks, A_Task, My_Iterator);


               IF (A_Processor.Name = A_Task.Cpu_Name) THEN

                  Blocking_Time_From_Simulation (
                     A_Task, Sys.Resources, Sched.Entries(J).Data.Result,
                     Max, Min,
                     Average);

                  Result:=Result & Start_Task & To_Unbounded_String(
                     "    ") & A_Task.Name  & To_Unbounded_String(" =>");
                  IF Worst_Case THEN
                     Result := Result & Max'Img & To_Unbounded_String(
                        "/worst ");
                  END IF;
                  IF Best_Case THEN
                     Result := Result & Min'Img & To_Unbounded_String(
                        "/best ");
                  END IF;
                  IF Average_Case THEN
                     Result := Result & Format(Average) &
                        To_Unbounded_String(
                        "/average ");
                  END IF;
                  Result:=Result & End_Task & Unbounded_Lf;

               END IF;

               EXIT WHEN Is_Last_Element(Sys.Tasks, My_Iterator);

               Next_Element(Sys.Tasks, My_Iterator);

            END LOOP;

         END IF;
      END LOOP;

   END Compute_Simulation_Blocking_Time;



   PROCEDURE Priority_Inversion_Detection (
         Sys         : IN OUT System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         Output      : IN     Output_Format    := String_Output) IS


      Priority_Inversion :          Priority_Inversion_List;
      Item               :          Priority_Inversion_Item_Ptr;
      Tmp_Tasks          :          Tasks_Set;
      My_Scheduler       : CONSTANT Scheduler_Ptr               := Get_Scheduler
        (A_Processor.
      ALL);


   BEGIN


      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;


      Current_Processor_Name:=A_Processor.Name;
      Select_And_Copy(Sys.Tasks, Tmp_Tasks, Select_Cpu'ACCESS );


      -- Set priority according to the scheduler
      --
      IF (Get_Name(My_Scheduler.All) = Deadline_Monotonic_Protocol)
            THEN
         Set_Priority_According_To_Dm(Fixed_Priority_Scheduler(
               My_Scheduler.All), Tmp_Tasks);
      ELSE
         IF (Get_Name(My_Scheduler.All) = Rate_Monotonic_Protocol)
               THEN
            Set_Priority_According_To_Rm(Fixed_Priority_Scheduler(
                  My_Scheduler.All), Tmp_Tasks);
         END IF;
      END IF;


      FOR J IN 0.. Sched.Nb_Entries-1 LOOP
         IF Sched.Entries(J).Item.Name=A_Processor.Name THEN



            Priority_Inversion:=Priority_Inversion_From_Simulation (
               Tmp_Tasks, Sys.Resources, A_Processor.Name,
               Sched.Entries(J).Data.Result);

            IF Is_Empty(Priority_Inversion)
                  THEN
               Result:= Result &
                  Lb_No_Priority_Inversion_Found(
                  Current_Language)  & Unbounded_Lf;
            ELSE

               Result:= Result & Start_Line & Lb_Minus &
                  Lb_Simulation_Priority_Inversion(
                  Current_Language) & Lb_Colon & Unbounded_Lf;


               WHILE NOT Is_Empty(Priority_Inversion)
                     LOOP
                  Item:=Get_Head(Priority_Inversion);

                  Result:= Result
                     & To_Unbounded_String("   ")
                     & Item.Task_Name
                     & Lb_Has_Priority_Inversion (Current_Language)
                     & Item.Resource_Name
                     & Lb_From_The_Time (Current_Language)
                     & To_Unbounded_String (Item.
                     Start_Time'Img)
                     & Lb_To_The_Time (Current_Language)
                     & To_Unbounded_String (Item.
                     End_Time'
                     Img)
                     & Lb_Dot & Unbounded_Lf;

                  Delete(Priority_Inversion, Item);

               END LOOP;
            END IF;
         END IF;
      END LOOP;

      Free(Tmp_Tasks);
      Result:=Result & Unbounded_Lf & Unbounded_Lf;

   END Priority_Inversion_Detection;




   PROCEDURE Deadlock_Detection (
         Sys         : IN OUT System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         Output      : IN     Output_Format    := String_Output) IS


      Deadlock : Deadlock_List;
      Item     : Deadlock_Item_Ptr;

   BEGIN

      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;


      FOR J IN 0.. Sched.Nb_Entries-1 LOOP
         IF Sched.Entries(J).Item.Name=A_Processor.Name THEN

            Deadlock:=Deadlock_From_Simulation (
               Sys.Tasks, Sys.Resources, A_Processor.Name,
               Sched.Entries(J).Data.Result);


            IF Is_Empty(Deadlock)
                  THEN
               Result:= Result &
                  Lb_No_Deadlock_Found(
                  Current_Language)  & Unbounded_Lf;

            ELSE

               Result:= Result & Start_Line & Lb_Minus &
                  Lb_Simulation_Deadlock(
                  Current_Language) & Lb_Colon & Unbounded_Lf;

               WHILE NOT Is_Empty(Deadlock)
                     LOOP
                  Item:=Get_Head(Deadlock);

                  Result:= Result
                     & To_Unbounded_String("   ")
                     & Lb_Deadlock_At_Time (Current_Language)
                     & To_Unbounded_String(Item.Time'
                     Img)
                     & Lb_Colon;
                  Result:=Result & " " & Item.
                     Task_Name;
                  Result:=Result & " " & Item.
                     Resource_Name & Lb_Dot & Unbounded_Lf;

                  Delete(Deadlock, Item);

               END LOOP;
            END IF;
         END IF;
      END LOOP;

   END Deadlock_Detection;



END Call_Resource_Framework;









