------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Text_Io;
USE Text_Io;

WITH Xml_Tag;
USE Xml_Tag;
WITH Scheduler;
USE Scheduler;
USE Scheduler.Double_Tasks_Parameters;
WITH Scheduler.Fixed_Priority;
USE  Scheduler.Fixed_Priority;
WITH Scheduler.Fixed_Priority.Dm;
USE  Scheduler.Fixed_Priority.Dm;
WITH Scheduler.Fixed_Priority.Rm;
USE  Scheduler.Fixed_Priority.Rm;

WITH Double_Util;
USE Double_Util;
WITH Translate;
USE Translate;
WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;
WITH Event_Analyzers;
USE Event_Analyzers;
WITH Event_Analyzer_Set;
USE Event_Analyzer_Set;
USE Event_Analyzer_Set.Generic_Event_Analyzer_Set;

WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;
WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Scheduling_Sequences;
USE Scheduling_Sequences;

WITH Multiprocessor_Services;
USE Multiprocessor_Services;
USE Multiprocessor_Services.Scheduling_Result_Per_Processor;

WITH  Gnat.Current_Exception;
USE  Gnat.Current_Exception;

WITH Expressions;
USE Expressions;

WITH Framework_Config;
USE Framework_Config;

WITH Time_Units;
USE Time_Units;


PACKAGE BODY Call_Scheduling_Framework  IS







   PROCEDURE Compute_Simulation_Time_Line (
         Sys                        : IN OUT System;
         Result                     : IN OUT Unbounded_String;
         Period                     : IN     Natural;
         Schedule_With_Offsets      : IN     Boolean;
         Schedule_With_Precedencies : IN     Boolean;
         Schedule_With_Resources    : IN     Boolean;
         With_Task_Specific_Seed    : IN     Boolean                            := True;
         Global_Seed_Value          : IN     Integer                            := 0;
         Predictable_Global_Seed    : IN     Boolean                            := True;
         Event_To_Generate          : IN     Time_Unit_Event_Type_Boolean_Table;
         Output                     : IN     Output_Format                      := String_Output) IS

   BEGIN

      Sort(Sys.Tasks, Increasing_Name'ACCESS);
      Scheduling_Base_Period:=Period;

      Initialize(Sched.All);

      Build_Multiprocessor_Scheduling(
         Sys.Processors, Sys.Address_Spaces, Sys.Tasks, Sys.Resources,
         Sys.Buffers, Sys.Messages, Sys.Dependencies,
         Sched, Event_To_Generate, Period,
         Schedule_With_Offsets, Schedule_With_Precedencies,
         Schedule_With_Resources,
         With_Task_Specific_Seed, Global_Seed_Value,
         Predictable_Global_Seed);


   EXCEPTION
      WHEN Parametric_File_Error =>
         Result:= Result & Lb_Minus & Lb_Simulation_Error(
            Current_Language) &
            Lb_Parametric_File_Error(Current_Language)  & Lb_Comma &
            Exception_Message & Unbounded_Lf;
      WHEN Type_Error =>
         Result:= Result & Lb_Minus & Lb_Simulation_Error(
            Current_Language) &
            Lb_Type_Error(Current_Language)  & Lb_Comma &
            Exception_Message & Unbounded_Lf;
      WHEN Statement_Error =>
         Result:= Result & Lb_Minus & Lb_Simulation_Error(
            Current_Language) &
            Lb_Statement_Error(Current_Language) & Lb_Comma &
            Exception_Message & Unbounded_Lf;
      WHEN Variable_Error =>
         Result:= Result & Lb_Minus & Lb_Simulation_Error(
            Current_Language) &
            Lb_Variable_Error(Current_Language)  & Lb_Comma &
            Exception_Message & Unbounded_Lf;
      WHEN Syntax_Error =>
         Result:= Result & Lb_Minus & Lb_Simulation_Error(
            Current_Language) &
            Lb_Syntax_Error(Current_Language) & Lb_Comma &
            Exception_Message & Unbounded_Lf;
      WHEN  A_Time_Unit.Indexed_Table_Full =>
         Result:= Result &
            Lb_Compute_Scheduling_Error_22(Current_Language) &
            Unbounded_Lf;
         Initialize(Sched.All);



   END Compute_Simulation_Time_Line;



   PROCEDURE Compute_Simulation_Basics (
         Sys         : IN     System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         Output      : IN     Output_Format    := String_Output) IS

      Check : Unbounded_String;

   BEGIN

      Check := Empty_String;
      Compute_Simulation_Number_Of_Context_Switch (
         Sys, Check, A_Processor, Output);
      Result:=Result  & Check;


      Check := Empty_String;
      Compute_Simulation_Number_Of_Preemption (
         Sys, Check, A_Processor, Output);
      Result:=Result  & Check;


      Check := Empty_String;
      Compute_Simulation_Response_Time (
         Sys, Check, A_Processor, True, False, False,
         Output);
      Result:=Result  & Unbounded_Lf & Check  & Unbounded_Lf;

   END Compute_Simulation_Basics;




   PROCEDURE Compute_Simulation_Number_Of_Context_Switch (
         Sys         : IN     System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         Output      : IN     Output_Format    := String_Output) IS

      Nb : Natural := 0;

   BEGIN

      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;

      FOR J IN 0.. Sched.Nb_Entries-1 LOOP

         IF Sched.Entries(J).Item.Name=A_Processor.Name THEN

            IF Sched.Entries(J).Data.Error_Msg/=Empty_String THEN
               Result:= Result &  Sched.Entries(J).Data.Error_Msg &
                  Unbounded_Lf;

            ELSE

               Nb:=Number_Of_Switch_Context_From_Simulation(Sched.Entries(
                     J).
                  Data.Result);
               Result := Result &  Start_Line &
                  Lb_Minus & Lb_Number_Of_Context_Switch(
                  Current_Language) & Lb_Colon &
                  Start_Num_Of_Context_Switch &
                  Nb'Img &
                  End_Num_Of_Context_Switch &
                  End_Line  &
                  Unbounded_Lf;

            END IF;
         END IF;
      END LOOP;
   END Compute_Simulation_Number_Of_Context_Switch;




   PROCEDURE Compute_Simulation_Number_Of_Preemption (
         Sys         : IN     System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         Output      : IN     Output_Format    := String_Output) IS

      Nb : Natural := 0;

   BEGIN


      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;


      FOR J IN 0.. Sched.Nb_Entries-1 LOOP

         IF Sched.Entries(J).Item.Name=A_Processor.Name THEN


            IF Sched.Entries(J).Data.Error_Msg/=Empty_String THEN
               Result:= Result &  Sched.Entries(J).Data.Error_Msg &
                  Unbounded_Lf;

            ELSE
               Nb:=Number_Of_Preemption_From_Simulation(Sched.
                  Entries(J).Data.Result, Sys.Tasks);

               Result:= Result &  Sched.Entries(J).Data.Scheduling_Msg;
               Result := Result &  Start_Line &  Lb_Minus &
                  Lb_Number_Of_Preemption(
                  Current_Language) & Lb_Colon & Start_Num_Of_Preemp &
                  Nb'Img & End_Num_Of_Preemp &
                  End_Line  &
                  Unbounded_Lf;
            END IF;

         END IF;
      END LOOP;


   END Compute_Simulation_Number_Of_Preemption;



   PROCEDURE Compute_Simulation_Response_Time (
         Sys          : IN     System;
         Result       : IN OUT Unbounded_String;
         A_Processor  : IN     Processor_Ptr;
         Worst_Case   :        Boolean;
         Best_Case    :        Boolean;
         Average_Case :        Boolean;
         Output       : IN     Output_Format    := String_Output) IS

      Check       : Unbounded_String := Empty_String;
      A_Task      : Task_Ptr;
      My_Iterator : Tasks_Iterator;
      Min         : Natural          := 0;
      Max         : Natural          := 0;
      Average     : Double           := 0.0;
      Missed      : Boolean          := False;

   BEGIN

      FOR J IN 0.. Sched.Nb_Entries-1 LOOP
         IF Sched.Entries(J).Item.Name=A_Processor.Name THEN


            IF Get_Number_Of_Task_From_Processor (Sys.Tasks, A_Processor.
                  Name) > 0 THEN

               Result := Lb_Minus & Lb_Task_Response_Time_From_Simulation (
                  Current_Language)
                  & Lb_Colon & Unbounded_Lf;

               Reset_Iterator (Sys.Tasks, My_Iterator);

               LOOP
                  Current_Element(Sys.Tasks, A_Task, My_Iterator);

                  Check:=Empty_String;
                  IF (A_Processor.Name = A_Task.Cpu_Name) THEN
                     Response_Time_By_Simulation(A_Task, Sched.Entries(J).
                        Data.Result, Check, Max, Min, Average);

                     Result:=Result & Start_Task & To_Unbounded_String(
                        "    ") & A_Task.Name  & To_Unbounded_String(
                        " =>");
                     IF Worst_Case THEN
                        Result := Result & Max'Img & To_Unbounded_String(
                           "/worst ");
                     END IF;
                     IF Best_Case THEN
                        Result := Result & Min'Img & To_Unbounded_String(
                           "/best ");
                     END IF;
                     IF Average_Case THEN
                        Result := Result & Format(Average) &
                           To_Unbounded_String(
                           "/average ");
                     END IF;
                     Result:=Result & Check & Unbounded_Lf;

                     IF Check/=Empty_String
                           THEN
                        Missed:=True;
                     END IF;
                  END IF;

                  EXIT WHEN Is_Last_Element(Sys.Tasks, My_Iterator);

                  Next_Element(Sys.Tasks, My_Iterator);

               END LOOP;
            END IF;
         END IF;
      END LOOP;

      IF Missed=False THEN
         Result := Result &
            Lb_No_Deadline_Missed_In_The_Computed_Scheduling(
            Current_Language) & Unbounded_Lf;
      ELSE
         Result := Result &
            Lb_Deadline_Will_Be_Missed_Task_Are_Not_Schedulable(
            Current_Language) & Unbounded_Lf;

      END IF;

   END Compute_Simulation_Response_Time;





   PROCEDURE Compute_Feasibility_Period (
         Sys         : IN     System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         Output      : IN     Output_Format    := String_Output) IS


      Period_String  : Unbounded_String;
      Free_Ut_String : Unbounded_String;

      Period  : Double := 0.0;
      Util    : Double := 0.0;
      Free_Ut : Double := 0.0;

   BEGIN

      IF Output = Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;


      BEGIN
         Period:=Scheduling_Period(Sys.Tasks,
            A_Processor.Name);

      EXCEPTION
         -- A large task set with many different period values can lead to
         -- a CONSTRAINT_ERROR exception due to type overflow
         -- This "Scheduling Period" is computed with the help of a "big
         -- number" package. If you want to increase the accurrancy of this
         -- computation, increase Max_Big_Number_Size in config/config.ads
         --
         WHEN OTHERS =>
            NULL;
      END;



      IF Period > Double(Natural'Last)
            THEN
         Period_String:=To_Unbounded_String(Double'Image(Period));
      ELSE
         Period_String:=To_Unbounded_String(Natural'Image(Natural(Period)));
      END IF;


      Result:= Result & Start_Line & Lb_Scheduling_Period(
         Current_Language) & Start_Bas_Per &
         Period_String
         & End_Bas_Per  & Start_Ref  &
         To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[18], page 5). ") & End_Ref &  End_Line &
         Unbounded_Lf;

      Util:=Processor_Utilization_Over_Period(Sys.Tasks,
         A_Processor.Name);

      Free_Ut:=Period*(1.0-Util);
      Free_Ut:=Double'Max(0.0,Free_Ut);

      IF Free_Ut > Double(Natural'Last)
            THEN
         Free_Ut_String:=To_Unbounded_String(Double'Image(Free_Ut));
      ELSE
         Free_Ut_String:=To_Unbounded_String(Natural'Image(Natural(
                  Free_Ut)));
      END IF;

      Result:= Result & Start_Line & To_Unbounded_String("-")
         & Start_Unit & Free_Ut_String & End_Unit &
         Lb_Free_Unit_Time(Current_Language) & End_Line  & Unbounded_Lf;

   EXCEPTION
      WHEN Task_Set.Task_Must_Be_Periodic=>
         Result:= Result & Lb_Compute_Scheduling_Error_2(Current_Language) &
            Unbounded_Lf;
      WHEN Task_Set.Task_Model_Error =>
         Result:= Result & Lb_Compute_Scheduling_Error_3(Current_Language) &
            Unbounded_Lf;


   END Compute_Feasibility_Period;



   PROCEDURE Compute_Feasibility_Processor_Utilization_Factor (
         Sys         : IN     System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         Output      : IN     Output_Format    := String_Output) IS

      Util   : Double;
      Utildi : Double;

   BEGIN


      IF Output = Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;

      Result := Result & Lb_Feasibility_Number1(Current_Language) &
         Unbounded_Lf & Unbounded_Lf;


      Compute_Feasibility_Period (
         Sys, Result, A_Processor, Output);


      BEGIN
         Utildi:=Processor_Utilization_Over_Deadline(Sys.Tasks,
            A_Processor.Name);
         Result:= Result & Start_Line &  Lb_Utilization_With_Deadline(
            Current_Language) &
            Start_Deadl_Fact  & Format(Utildi) & End_Deadl_Fact &
            Start_Ref &
            To_Unbounded_String(" (") & Lb_See(
            Current_Language) &
            To_Unbounded_String("[1], page 6). ") & End_Ref & End_Line &
            Unbounded_Lf;
      EXCEPTION
         WHEN Task_Set.Task_Must_Be_Periodic=>
            Result:= Result & Lb_Compute_Scheduling_Error_9(
               Current_Language) &
               Unbounded_Lf;
         WHEN Task_Set.Task_Model_Error =>
            Result:= Result & Lb_Compute_Scheduling_Error_10(
               Current_Language) &
               Unbounded_Lf;
         WHEN Deadline_Error =>
            Result:= Result & Lb_Compute_Scheduling_Error_21(
               Current_Language) &
               Unbounded_Lf;


      END;


      BEGIN
         Util:=Processor_Utilization_Over_Period(Sys.Tasks,
            A_Processor.Name);
         Result:= Result & Start_Line & Lb_Utilization_With_Period(
            Current_Language) &
            Start_Per_Fact & Format(Util) & End_Per_Fact &
            Start_Ref & To_Unbounded_String(" (") & Lb_See(
            Current_Language) &
            To_Unbounded_String("[1], page 6). ") & End_Ref &
            Unbounded_Lf;
      EXCEPTION
         WHEN Task_Set.Task_Must_Be_Periodic=>
            Result:= Result & Lb_Compute_Scheduling_Error_9(
               Current_Language) &
               Unbounded_Lf;
         WHEN Task_Set.Task_Model_Error =>
            Result:= Result & Lb_Compute_Scheduling_Error_10(
               Current_Language) &
               Unbounded_Lf;
      END;

      BEGIN

         Result := Result & Start_Line;
         Utilization_Factor_Feasibility_Test(A_Processor.Scheduler.All,
            Sys.Tasks, A_Processor.Name, Result);

      EXCEPTION
         WHEN Task_Set.Task_Must_Be_Periodic=>
            Result:= Result & Lb_Compute_Scheduling_Error_9(
               Current_Language) &
               Unbounded_Lf;
         WHEN Task_Set.Task_Model_Error =>
            Result:= Result & Lb_Compute_Scheduling_Error_10(
               Current_Language) &
               Unbounded_Lf;
         WHEN Deadline_Error =>
            Result:= Result & Lb_Compute_Scheduling_Error_21(
               Current_Language) &
               Unbounded_Lf;


      END;


   END Compute_Feasibility_Processor_Utilization_Factor;




   PROCEDURE Compute_Feasibility_Response_Time (
         Sys         : IN OUT System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         Output      : IN     Output_Format    := String_Output) IS

      Rt     : Response_Time_Table;
      Msg    : Unbounded_String    := Empty_String;
      Missed : Boolean             := False;

   BEGIN

      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;


      Result := Result & Unbounded_Lf & Lb_Feasibility_Number2(
         Current_Language) & Unbounded_Lf & Unbounded_Lf;


      Compute_Response_Time(A_Processor.Scheduler.All, Sys.Tasks,
         A_Processor.Name, Msg, Rt);

      Result:=Result & Start_Line & Lb_Minus &
         Lb_Bound_On_Task_Response_Time(
         Current_Language) & Lb_Colon &
         Start_Ref & Msg & End_Ref &
         Unbounded_Lf;



      FOR I IN 0..Response_Time_Range(Get_Number_Of_Task_From_Processor(
               Sys.
               Tasks,A_Processor.Name)-1) LOOP
         IF (Rt.Entries(I).Item /= NULL)
               THEN
            Result:=Result & Start_Task & To_Unbounded_String("    ") &
               Rt.
               Entries(
               I).
               Item.Name & To_Unbounded_String(" =>") &
               Integer'Image(Integer(Rt.Entries(I).Data)) & End_Task;

            IF (Natural(Rt.Entries(I).Data) > Rt.Entries(I).Item.
                  Deadline)
                  THEN
               Result := Result & Lb_Comma & Lb_Check_Deadline(
                  Current_Language) & Rt.Entries(I).Item.Deadline'Img &
                  To_Unbounded_String(")");
               Missed:=True;
            END IF;


            Result := Result  &  Unbounded_Lf;

         END IF;
      END LOOP;

      IF Missed=False
            THEN
         Result := Result &
            Lb_No_Deadline_Will_Be_Missed_Task_Are_Schedulable(
            Current_Language);
      ELSE
         Result := Result &
            Lb_Deadline_Will_Be_Missed_Task_Are_Not_Schedulable(
            Current_Language);
      END IF;
      Result:=Result & End_Line  & Unbounded_Lf & End_Sched_Feas  &
         Unbounded_Lf;

   EXCEPTION
      WHEN Task_Set.Task_Must_Have_Period_Equal_To_Deadline=>
         Result := Result & Lb_Compute_Scheduling_Error_4(
            Current_Language) &
            Unbounded_Lf;
      WHEN Task_Set.Task_Must_Be_Periodic=>
         Result := Result & Lb_Compute_Scheduling_Error_5(
            Current_Language) &
            Unbounded_Lf;
      WHEN Task_Set.Task_Model_Error =>
         Result := Result & Lb_Compute_Scheduling_Error_6(
            Current_Language) &
            Unbounded_Lf;
      WHEN Scheduler.Processor_Utilization_Exceeded =>
         Result:= Result & Lb_Compute_Scheduling_Error_7(Current_Language) &
            Unbounded_Lf;
      WHEN Task_Set.Start_Time_Error =>
         Result:= Result & Lb_Compute_Scheduling_Error_12(
            Current_Language) &
            Unbounded_Lf;
      WHEN Task_Set.Offset_Error =>
         Result:= Result & Lb_Compute_Scheduling_Error_13(
            Current_Language) &
            Unbounded_Lf;
      WHEN Invalid_Scheduler =>
         Result:= Result & Lb_Compute_Scheduling_Error_18(
            Current_Language) &
            Unbounded_Lf;
   END Compute_Feasibility_Response_Time;




   PROCEDURE Set_Priorities (
         Sys         : IN OUT System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         A_Scheduler : IN     Scheduler_Type;
         Output      : IN     Output_Format    := String_Output) IS


      A_Rm_Scheduler   : Rm_Scheduler;
      A_Dm_Scheduler   : Dm_Scheduler;
      Ok               : Boolean        := True;
      A_Task           : Task_Ptr;
      My_Task_Iterator : Tasks_Iterator;

   BEGIN

      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;

      BEGIN
         IF(A_Scheduler=Rate_Monotonic_Protocol) THEN
            Set_Priority_According_To_Rm(A_Rm_Scheduler, Sys.Tasks,
               A_Processor.Name) ;
         ELSE
            Set_Priority_According_To_Dm(A_Dm_Scheduler, Sys.Tasks,
               A_Processor.Name) ;
         END IF;


      EXCEPTION
         WHEN Task_Set.Task_Must_Be_Periodic=>
            Ok:=False;

      END;

      IF Ok
            THEN
         IF(A_Scheduler=Rate_Monotonic_Protocol)
               THEN
            Result := Result & Start_Line &  Lb_Set_Priorities_Rm(
               Current_Language)& End_Line;
         ELSE
            Result := Result & Start_Line &  Lb_Set_Priorities_Dm(
               Current_Language)& End_Line;
         END IF;


         Result :=Result & Unbounded_Lf & Start_Line & Lb_Priorities(
            Current_Language);

         Reset_Iterator(Sys.Tasks, My_Task_Iterator);
         LOOP
            Current_Element(Sys.Tasks, A_Task, My_Task_Iterator);

            IF (A_Task.Cpu_Name = A_Processor.Name)
                  THEN
               Result :=Result & Unbounded_Lf & Start_Task &
                  To_Unbounded_String("    ") &
                  A_Task.Name & To_Unbounded_String(" =>") &
                  To_Unbounded_String(A_Task.Priority'Img)& End_Task;
            END IF;

            EXIT WHEN Is_Last_Element(Sys.Tasks, My_Task_Iterator);
            Next_Element(Sys.Tasks, My_Task_Iterator);
         END LOOP;
         Result:=Result & Unbounded_Lf & End_Line &Unbounded_Lf&
            End_Set_Prio;

      ELSE
         Result:= Result &  Lb_Priorities_Error(Current_Language);
      END IF;


   END Set_Priorities;



   PROCEDURE Run_User_Defined_Event_Handlers (
         Sys    : IN     System;
         Result : IN OUT Unbounded_String;
         Output : IN     Output_Format    := String_Output) IS


      Msg               : Unbounded_String;
      My_Iterator       : Event_Analyzers_Iterator;
      An_Event_Analyzer : Event_Analyzer_Ptr;

   BEGIN

      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;


      IF(Get_Number_Of_Elements(Sys.Event_Analyzers) > 0)
            THEN


         Reset_Iterator(Sys.Event_Analyzers, My_Iterator);

         LOOP
            Current_Element(Sys.Event_Analyzers, An_Event_Analyzer,
               My_Iterator);

            -- Run an event analyzer
            ---
            BEGIN
               Msg := Empty_String;


               Result := Result & Unbounded_Lf & Unbounded_Lf;

               Result:= Result & Lb_Minus & Lb_Event_Analyzer_Name(
                  Current_Language)
                  & Lb_Colon & An_Event_Analyzer.Behavior_File_Name &
                  Unbounded_Lf & Unbounded_Lf;

               Run_An_Event_Analyzer (
                  Sys.Tasks, Sys.Processors, Sys.Address_Spaces,
                  Sys.Resources, Sys.Buffers,
                  Sys.Messages,
                  Sched,  An_Event_Analyzer, Msg);

               Result:= Result & Msg;

               Result := Result   & Unbounded_Lf & Unbounded_Lf;

            EXCEPTION
               WHEN Parametric_File_Error =>
                  Result:= Result & Lb_Minus & Lb_Simulation_Error(
                     Current_Language) &
                     Lb_Parametric_File_Error(Current_Language)  &
                     Lb_Comma &
                     Exception_Message & Unbounded_Lf;
               WHEN Type_Error =>
                  Result:= Result & Lb_Minus & Lb_Simulation_Error(
                     Current_Language) &
                     Lb_Type_Error(Current_Language)  & Lb_Comma &
                     Exception_Message & Unbounded_Lf;
               WHEN Statement_Error =>
                  Result:= Result & Lb_Minus & Lb_Simulation_Error(
                     Current_Language) &
                     Lb_Statement_Error(Current_Language) & Lb_Comma &
                     Exception_Message & Unbounded_Lf;
               WHEN Variable_Error =>
                  Result:= Result & Lb_Minus & Lb_Simulation_Error(
                     Current_Language) &
                     Lb_Variable_Error(Current_Language)  & Lb_Comma &
                     Exception_Message & Unbounded_Lf;
               WHEN Syntax_Error =>
                  Result:= Result & Lb_Minus & Lb_Simulation_Error(
                     Current_Language) &
                     Lb_Syntax_Error(Current_Language) & Lb_Comma &
                     Exception_Message & Unbounded_Lf;
            END;


            EXIT WHEN Is_Last_Element(Sys.Event_Analyzers, My_Iterator);

            Next_Element(Sys.Event_Analyzers, My_Iterator);


         END LOOP;

      END IF;


   END Run_User_Defined_Event_Handlers;









   PROCEDURE Partionning_Tasks (
         Sys             : IN OUT System;
         Result          : IN OUT Unbounded_String;
         A_Parition_Rule : IN     Partioning_Type;
         Output          : IN     Output_Format    := String_Output) IS


      Partitionned_Task_Set : Tasks_Set;
      A_Task                : Task_Ptr;
      My_Task_Iterator      : Tasks_Iterator;
      Msg                   : Unbounded_String := Empty_String;

   BEGIN

      IF Output=Xml_Output THEN
         Set_Tag;
      ELSE
         Set_Empty;
      END IF;


      CASE A_Parition_Rule IS

         WHEN General_Task =>
            Partition_General_Task(Sys.Processors, Sys.Tasks, Msg,
               Partitionned_Task_Set);
         WHEN Best_Fit =>
            Partition_Best_Fit(Sys.Processors, Sys.Tasks, Msg,
               Partitionned_Task_Set);
         WHEN Next_Fit =>
            Partition_Next_Fit(Sys.Processors, Sys.Tasks, Msg,
               Partitionned_Task_Set);
         WHEN First_Fit =>
            Partition_First_Fit(Sys.Processors, Sys.Tasks, Msg,
               Partitionned_Task_Set);
         WHEN Small_Task =>
            Partition_Small_Task(Sys.Processors, Sys.Tasks, Msg,
               Partitionned_Task_Set);
      END CASE;


      Result:=Result & Start_Line & Lb_Minus & Lb_Partitioning_Result(
         Current_Language) &
         Start_Ref & Msg & End_Ref & Lb_Colon &
         Unbounded_Lf;


      Sys.Tasks:=Partitionned_Task_Set;


      Reset_Iterator(Sys.Tasks, My_Task_Iterator);
      LOOP
         Current_Element(Sys.Tasks, A_Task, My_Task_Iterator);

         Result :=Result & Unbounded_Lf &
            To_Unbounded_String("    ") &
            A_Task.Name & To_Unbounded_String(" =>") &
            A_Task.Cpu_Name;

         EXIT WHEN Is_Last_Element(Sys.Tasks, My_Task_Iterator);
         Next_Element(Sys.Tasks, My_Task_Iterator);
      END LOOP;
      Result:=Result & Unbounded_Lf & Unbounded_Lf;


   EXCEPTION
      WHEN Processors_Should_Have_The_Same_Scheduler =>
         Result:= Result &  Lb_Partition_Error1(Current_Language);
      WHEN No_Such_Processors =>
         Result:= Result &  Lb_Partition_Error2(Current_Language);
      WHEN Invalid_Scheduler_Multiprocessors =>
         Result:= Result &  Lb_Partition_Error3(Current_Language);

   END Partionning_Tasks;





   PROCEDURE Compute_Feasibility_Basics (
         Sys         : IN OUT System;
         Result      : IN OUT Unbounded_String;
         A_Processor : IN     Processor_Ptr;
         Output      : IN     Output_Format    := String_Output) IS
   BEGIN

      Compute_Feasibility_Processor_Utilization_Factor (
         Sys,  Result, A_Processor, Output);

      Compute_Feasibility_Response_Time (
         Sys, Result, A_Processor, Output);


   END Compute_Feasibility_Basics ;





BEGIN


   Sched:=NEW Scheduling_Table;
   Initialize(Sched.All);




END Call_Scheduling_Framework;




