------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;


WITH Text_Io;
USE Text_Io;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;


WITH Unbounded_Strings;
USE  Unbounded_Strings;


WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;

WITH Framework_Config;
USE Framework_Config;

WITH Task_Dependencies;
USE Task_Dependencies;

WITH Processors;
USE Processors;
WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;



WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;


WITH Scheduler;
USE Scheduler;
USE Scheduler.Double_Tasks_Parameters;





PACKAGE BODY Dependency_Services IS




   PROCEDURE Set_Parameters_According_To_Chetto (
         My_Dependencies    : IN     Tasks_Dependencies_Ptr;
         My_Tasks           : IN OUT Tasks_Set;
         Compute_Deadlines  : IN     Boolean;
         Compute_Priorities : IN     Boolean) IS


      Leaf,
      Step       : Tasks_Set;
      Previous   : Tasks_Set;
      Real_Left,
      Real_Right,
      Left,
      Right      : Task_Ptr;
      Ite1,
      Ite2       : Tasks_Iterator;

   BEGIN

      IF Is_Cyclic(My_Dependencies)
            THEN
         RAISE Cyclic_Graph_Error;
      END IF;


      Dependencies_Periods_Control (
         My_Tasks, My_Dependencies, Empty_String);



      -- Assume the graph is acyclic.
      -- Extract leaves of the graph and update deadline
      -- of previous node
      --
      Step:=Get_Leaf_Tasks(My_Dependencies);

      WHILE NOT Is_Empty(Step)    LOOP

         Reset(Leaf);
         Duplicate(Step, Leaf);
         Reset(Step);

         Reset_Iterator(Leaf, Ite1);
         FOR I IN 0 .. Get_Number_Of_Elements(Leaf)-1 LOOP

            -- The task to be updated
            --
            Current_Element(Leaf, Right, Ite1);


            -- Take previous tasks
            --
            IF Has_Predecessor(My_Dependencies, Right) THEN

               Reset(Previous);
               Previous:=Get_Predecessors_List(My_Dependencies, Right);

               Reset_Iterator(Previous, Ite2);

               FOR J IN 0 .. Get_Number_Of_Elements(Previous)-1 LOOP

                  Current_Element(Previous, Left, Ite2);
                  Add(Step, Left);

                  Real_Left:=Search_Task (My_Tasks, Left.Name);
                  Real_Right:=Search_Task (My_Tasks, Right.Name);

                  IF Compute_Priorities THEN
                     IF Right.Priority>=1
                           THEN
                        Real_Left.Priority:= Priority_Range'Max(
                           Real_Left.Priority,
                           Real_Right.Priority+1);
                     END IF;
                  END IF;

                  IF Compute_Deadlines THEN
                     Real_Left.Deadline:= Natural'Min(
                        Real_Left.Deadline,  Real_Right.Deadline
                        - Real_Right.Capacity);
                  END IF;



                  Next_Element(Previous, Ite2);
               END LOOP;
            END IF;


            Next_Element(Leaf, Ite1);
         END LOOP;

      END LOOP;


   END Set_Parameters_According_To_Chetto;



   FUNCTION Get_Response_Time (
         Rtn    : IN     Scheduler.Response_Time_Table;
         A_Task :        Task_Ptr)
     RETURN Double IS
   BEGIN
      FOR I IN Scheduler.Response_Time_Range LOOP
         IF (Rtn.Entries(I).Item.Name = A_Task.Name)
               THEN
            RETURN Rtn.Entries(I).Data;
         END IF;
      END LOOP;

      RAISE Task_Set.Task_Not_Found;
   END Get_Response_Time;



   PROCEDURE Inject_Response_Time_Into_Jitter (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Tasks        : IN OUT Tasks_Set;
         Rtn             : IN     Scheduler.Response_Time_Table) IS


      Max_Jitter : Double         := 0.0;
      Ite1,
      Ite2       : Tasks_Iterator;
      Previous,
      A_Task     : Task_Ptr;
      Pred       : Tasks_Set;

   BEGIN

      -- Scan each task. If a task has a predecessor, set its jitter
      -- to the response time of its predecessors
      --
      Reset_Iterator(My_Tasks, Ite1);
      LOOP
         Current_Element(My_Tasks, A_Task, Ite1);


         -- Find maximum response time of predecessors
         --
         IF Has_Predecessor(My_Dependencies, A_Task) THEN
            Reset(Pred);
            Pred:=Get_Predecessors_List(My_Dependencies, A_Task);

            IF NOT Is_Empty(Pred) THEN
               Reset_Iterator(Pred, Ite2);
               Max_Jitter := 0.0;
               FOR I IN 0 .. Get_Number_Of_Elements(Pred)-1 LOOP

                  Current_Element(Pred, Previous, Ite2);

                  Max_Jitter:=Double'Max(Max_Jitter,
                     Get_Response_Time(Rtn, Previous) );

                  Next_Element(Pred, Ite2);
               END LOOP;

               -- set jitter with max_jitter
               --
               CASE A_Task.Task_Type IS
                  WHEN Periodic_Type =>
                     Periodic_Task_Ptr(A_Task).Jitter:=
                        Natural'Max(Periodic_Task_Ptr(A_Task).Jitter,
                        Natural(Max_Jitter) );
                  WHEN Poisson_Type =>
                     Poisson_Task_Ptr(A_Task).Jitter:=
                        Natural'Max(Poisson_Task_Ptr(A_Task).Jitter,
                        Natural(Max_Jitter) );
                  WHEN OTHERS =>
                     RAISE Constraint_Error;
               END CASE;
            END IF;
         END IF;


         EXIT WHEN Is_Last_Element(My_Tasks, Ite1);
         Next_Element(My_Tasks, Ite1);
      END LOOP;

   END Inject_Response_Time_Into_Jitter;




   PROCEDURE Compute_All_Response_Times (
         My_Processors : IN     Processors_Set;
         My_Tasks      : IN OUT Tasks_Set;
         Msg           : IN OUT Unbounded_String;
         Rtn           : IN OUT Scheduler.Response_Time_Table) IS


      I           : Natural;
      K,
      J           : Scheduler.Response_Time_Range;
      A_Processor : Processor_Ptr;
      My_Iterator : Processors_Iterator;

      TYPE Rt IS ARRAY (0 .. Max_Processors - 1) OF Scheduler.Response_Time_Table;
      Rt_By_Processor : Rt;

   BEGIN
      -- Compute tasks
      --
      I:=0;
      Reset_Iterator(My_Processors, My_Iterator);

      LOOP
         Current_Element(My_Processors, A_Processor, My_Iterator);


         Msg:=Empty_String;
         IF Get_Number_Of_Task_From_Processor (
               My_Tasks, A_Processor.Name) > 0 THEN
            Compute_Response_Time(A_Processor.Scheduler.All, My_Tasks,
               A_Processor.Name, Msg, Rt_By_Processor(I));
         END IF;

         EXIT WHEN Is_Last_Element(My_Processors, My_Iterator);

         Next_Element(My_Processors, My_Iterator);
         I:=I+1;
      END LOOP;

      I:=0;
      K:=0;

      Reset_Iterator(My_Processors, My_Iterator);

      LOOP
         J:=0;
         Current_Element(My_Processors, A_Processor, My_Iterator);

         FOR L IN 0 .. (Get_Number_Of_Task_From_Processor(My_Tasks,
                  A_Processor.Name)-1) LOOP
            Rtn.Entries(K):=Rt_By_Processor(I).Entries(J);
            Rtn.Nb_Entries:=Rtn.Nb_Entries+1;
            K:=K+1;
            J:=J+1;
         END LOOP;

         EXIT WHEN Is_Last_Element(My_Processors, My_Iterator);

         Next_Element(My_Processors, My_Iterator);
         I:=I+1;


      END LOOP;

   END Compute_All_Response_Times;




   PROCEDURE Compute_End_To_End_Response_Time (
         My_Dependencies  : IN     Tasks_Dependencies_Ptr;
         My_Processors    : IN     Processors_Set;
         My_Tasks         : IN OUT Tasks_Set;
         One_Step         : IN     Boolean;
         Update_Tasks_Set : IN     Boolean;
         Msg              : IN OUT Unbounded_String;
         Rt               : IN OUT Response_Time_Table) IS



      Rtn,
      Rtn1      : Response_Time_Table;
      Tmp_Tasks : Tasks_Set;


   BEGIN

      Initialize(Rtn);
      Initialize(Rtn1);

      IF NOT Update_Tasks_Set THEN
         Duplicate(My_Tasks, Tmp_Tasks);
      END IF;

      Compute_All_Response_Times(My_Processors, My_Tasks, Msg, Rtn);

      LOOP

         -- Stop when end to end response time do not change
         -- from an iteration to another
         --
         EXIT WHEN Is_Equal(Rtn,Rtn1);



         -- Update jitter ....
         --
         IF NOT Update_Tasks_Set THEN
            Inject_Response_Time_Into_Jitter(My_Dependencies, Tmp_Tasks,
               Rtn);
         ELSE
            Inject_Response_Time_Into_Jitter(My_Dependencies, My_Tasks,
               Rtn);
         END IF;

         IF One_Step THEN
            EXIT;
         END IF;


         Rtn1:=Rtn;
         Initialize(Rtn);

         IF NOT Update_Tasks_Set THEN
            Compute_All_Response_Times(My_Processors, Tmp_Tasks, Msg, Rtn);
         ELSE
            Compute_All_Response_Times(My_Processors, My_Tasks, Msg, Rtn);
         END IF;


         -- Stop when all deadlines are missed
         --
         EXIT WHEN All_Deadlines_Missed(Rtn);


      END LOOP;


      Rt:=Rtn;


   END Compute_End_To_End_Response_Time;







   FUNCTION All_Deadlines_Missed (
         Rtn : IN     Response_Time_Table)
     RETURN Boolean IS
   BEGIN

      FOR I IN 0.. Rtn.Nb_Entries -1 LOOP
         IF Rtn.Entries(I).Data < Double(Rtn.Entries(I).Item.Deadline)
               THEN
            RETURN False;
         END IF;
      END LOOP;


      RETURN True;
   END All_Deadlines_Missed ;




   FUNCTION Is_Equal (
         Rtn  : IN     Response_Time_Table;
         Rtn1 : IN     Response_Time_Table)
     RETURN Boolean IS
   BEGIN

      IF Rtn.Nb_Entries /= Rtn1.Nb_Entries
            THEN
         RETURN False;
      END IF;

      FOR I IN 0.. Rtn.Nb_Entries -1 LOOP
         IF Rtn.Entries(I).Data /= Rtn1.Entries(I).Data
               THEN
            RETURN False;
         END IF;
      END LOOP;


      RETURN True;
   END Is_Equal;



   -- To schedule with precedencies
   -- all task of a dag must have the same period
   --
   PROCEDURE Dependencies_Periods_Control (
         My_Tasks       : IN     Tasks_Set;
         My_Deps        : IN     Tasks_Dependencies_Ptr;
         Processor_Name : IN     Unbounded_String) IS


      Next       : Tasks_Set;
      Real_Task1,
      Real_Right,
      Right      : Task_Ptr;
      Ite1,
      Ite2       : Tasks_Iterator;
      Task1      : Task_Ptr;

   BEGIN

      Reset_Iterator(My_Tasks, Ite1);

      LOOP
         Current_Element(My_Tasks, Task1, Ite1);


         IF (Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String)
               THEN


            -- Check period of successors
            --
            IF Has_Successor(My_Deps, Task1) THEN

               Reset(Next);
               Next:=Get_Successors_List(My_Deps, Task1);

               Reset_Iterator(Next, Ite2);
               FOR J IN 0 .. Get_Number_Of_Elements(Next)-1 LOOP

                  Current_Element(Next, Right, Ite2);

                  Real_Task1:=Search_Task (My_Tasks, Task1.Name);
                  Real_Right:=Search_Task (My_Tasks, Right.Name);

                  IF (Real_Right.Task_Type /= Real_Task1.Task_Type)
                        THEN
                     RAISE Dependencies_Period_Error;
                  END IF;

                  IF (Real_Right.Task_Type = Parametric_Type)
                        THEN
                     RAISE Dependencies_Period_Error;
                  END IF;

                  IF (Real_Right.Task_Type = Periodic_Type)
                        THEN
                     IF (Periodic_Task_Ptr(Real_Right).Period /=
                           Periodic_Task_Ptr(Real_Task1).Period)
                           THEN
                        RAISE Dependencies_Period_Error;
                     END IF;
                  END IF;

                  IF (Real_Right.Task_Type = Poisson_Type)
                        THEN
                     IF (Poisson_Task_Ptr(Real_Right).Period /=
                           Poisson_Task_Ptr(Real_Task1).Period)
                           THEN
                        RAISE Dependencies_Period_Error;
                     END IF;
                  END IF;

                  Next_Element(Next, Ite2);
               END LOOP;
            END IF;

            -- Check period of predecessors
            --
            IF Has_Predecessor(My_Deps, Task1) THEN

               Reset(Next);
               Next:=Get_Predecessors_List(My_Deps, Task1);

               Reset_Iterator(Next, Ite2);
               FOR J IN 0 .. Get_Number_Of_Elements(Next)-1 LOOP

                  Current_Element(Next, Right, Ite2);

                  Real_Task1:=Search_Task (My_Tasks, Task1.Name);
                  Real_Right:=Search_Task (My_Tasks, Right.Name);


                  IF (Real_Right.Task_Type /= Real_Task1.Task_Type)
                        THEN
                     RAISE Dependencies_Period_Error;
                  END IF;


                  IF (Real_Right.Task_Type = Parametric_Type)
                        THEN
                     RAISE Dependencies_Period_Error;
                  END IF;


                  IF (Real_Right.Task_Type = Periodic_Type)
                        THEN
                     IF (Periodic_Task_Ptr(Real_Right).Period /=
                           Periodic_Task_Ptr(Real_Task1).Period)
                           THEN
                        RAISE Dependencies_Period_Error;
                     END IF;
                  END IF;

                  IF (Real_Right.Task_Type = Poisson_Type)
                        THEN
                     IF (Poisson_Task_Ptr(Real_Right).Period /=
                           Poisson_Task_Ptr(Real_Task1).Period)
                           THEN
                        RAISE Dependencies_Period_Error;
                     END IF;
                  END IF;


                  Next_Element(Next, Ite2);
               END LOOP;
            END IF;



         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, Ite1);
         Next_Element(My_Tasks, Ite1);

      END LOOP;
   END Dependencies_Periods_Control;






END Dependency_Services;







