------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------





WITH Ada.Exceptions;
USE Ada.Exceptions;

WITH Objects;
USE Objects;

WITH Translate;
USE Translate;

WITH Unbounded_Strings;
USE Unbounded_Strings;



PACKAGE BODY Event_Analyzer_Set IS





   PROCEDURE Check_Event_Analyzer (
         My_Event_Analyzers : IN     Event_Analyzers_Set;
         File_Name          : IN     Unbounded_String) IS


   BEGIN


      IF File_Name=""
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Event_Analyzer_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;

      IF NOT Is_A_Valid_Identifier(File_Name)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Event_Analyzer_Name(Current_Language) &
               Lb_Colon & Lb_Invalid_Identifier(Current_Language)));
      END IF;


   END Check_Event_Analyzer;




   PROCEDURE Add_Event_Analyzer (
         My_Event_Analyzers : IN OUT Event_Analyzers_Set;
         File_Name          : IN     Unbounded_String) IS


      An_Event_Analyzer : Event_Analyzer_Ptr;
      My_Iterator       : Iterator;


   BEGIN


      Check_Event_Analyzer(My_Event_Analyzers ,File_Name);


      IF(Get_Number_Of_Elements(My_Event_Analyzers) > 0)
            THEN

         Reset_Iterator(My_Event_Analyzers, My_Iterator);

         LOOP
            Current_Element(My_Event_Analyzers, An_Event_Analyzer,
               My_Iterator);
            IF (File_Name = An_Event_Analyzer.Behavior_File_Name)
                  THEN
               Raise_Exception(Invalid_Parameter'
                  Identity,
                  To_String(Lb_Event_Analyzer(Current_Language) & " " &
                     File_Name &
                     " : " & Lb_Event_Analyzer_Name(
                        Current_Language) &
                     Lb_Already_Defined(
                        Current_Language)));
            END IF;

            EXIT WHEN Is_Last_Element(My_Event_Analyzers, My_Iterator);

            Next_Element(My_Event_Analyzers, My_Iterator);


         END LOOP;

      END IF;


      An_Event_Analyzer:= NEW Event_Analyzer;
      An_Event_Analyzer.Behavior_File_Name:= File_Name;

      Add(My_Event_Analyzers, An_Event_Analyzer);


   EXCEPTION
      WHEN Full_Set =>
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Can_Not_Define_More_Event_Analyzers(
                  Current_Language)));

   END Add_Event_Analyzer;



   FUNCTION Search_Event_Analyzer (
         My_Event_Analyzers : IN     Event_Analyzers_Set;
         Name               : IN     Unbounded_String)
     RETURN Event_Analyzer_Ptr IS

      My_Iterator       : Iterator;
      An_Event_Analyzer : Event_Analyzer_Ptr;
      Result            : Event_Analyzer_Ptr;

      Found : Boolean := False;

   BEGIN

      Reset_Iterator(My_Event_Analyzers, My_Iterator);

      LOOP
         Current_Element(My_Event_Analyzers,An_Event_Analyzer,My_Iterator);

         IF(An_Event_Analyzer.Behavior_File_Name = Name)
               THEN
            Found:=True;
            Result:=An_Event_Analyzer;
         END IF;

         EXIT WHEN Is_Last_Element(My_Event_Analyzers, My_Iterator);

         Next_Element(My_Event_Analyzers, My_Iterator);

      END LOOP;

      IF NOT Found
            THEN
         RAISE Event_Analyzer_Not_Found;
      END IF;

      RETURN Result;
   END Search_Event_Analyzer;








   FUNCTION Export_Xml (
         My_Event_Analyzers : IN     Event_Analyzers_Set)
     RETURN Unbounded_String IS


      A_Event_Analyzer : Event_Analyzer_Ptr;
      My_Iterator      : Event_Analyzers_Iterator;

      Result : Unbounded_String := Empty_String;

   BEGIN

      IF NOT Is_Empty(My_Event_Analyzers) THEN

         Result:= Result & To_Unbounded_String( "<event_analyzers>") &
            Unbounded_Lf;
         Reset_Iterator(My_Event_Analyzers, My_Iterator);

         LOOP
            Current_Element(My_Event_Analyzers, A_Event_Analyzer,
               My_Iterator);

            Result:= Result & To_Unbounded_String(Ascii.Ht &
               "<event_analyzer");
            Result:= Result & To_Unbounded_String(
               " parametric_file_name=""" & To_String(
                  A_Event_Analyzer.Behavior_File_Name) &
               """" & "> </event_analyzer>") & Unbounded_Lf;

            EXIT WHEN Is_Last_Element(My_Event_Analyzers, My_Iterator);

            Next_Element(My_Event_Analyzers, My_Iterator);

         END LOOP;
         Result:= Result & To_Unbounded_String( "</event_analyzers>") &
            Unbounded_Lf;
      END IF;


      RETURN Result;

   END Export_Xml;


   FUNCTION Export_Aadl_Properties (
         My_Event_Analyzers : IN     Event_Analyzers_Set;
         Number_Of_Ht       : IN     Natural)
     RETURN Unbounded_String IS

      An_Event_Analyzer : Event_Analyzer_Ptr;
      My_Iterator       : Event_Analyzers_Iterator;

      Result : Unbounded_String := Empty_String;

   BEGIN

      IF NOT Is_Empty(My_Event_Analyzers) THEN


         Reset_Iterator(My_Event_Analyzers, My_Iterator);

         FOR I IN 1..Number_Of_Ht LOOP
            Result:= Result & Ascii.Ht;
         END LOOP;
         Result:= Result & To_Unbounded_String(
            "Source_Text => ") & """";

         LOOP
            Current_Element(My_Event_Analyzers,
               An_Event_Analyzer,
               My_Iterator);

            Result:=Result &
               To_String(An_Event_Analyzer.
               Behavior_File_Name);

            EXIT WHEN Is_Last_Element(
               My_Event_Analyzers, My_Iterator);

            Next_Element(My_Event_Analyzers,
               My_Iterator);

            Result:= Result & ",";

         END LOOP;

         Result:= Result & """" & ";" & Unbounded_Lf;

      END IF;


      RETURN Result;

   END Export_Aadl_Properties;




END Event_Analyzer_Set;
