------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Parameters;
USE Parameters;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;


WITH Time_Units;
USE  Time_Units;
USE  Time_Units.A_Time_Unit;

WITH Double_Util;
USE Double_Util;

WITH Unbounded_Strings;
USE Unbounded_Strings;


PACKAGE BODY Message_Set IS



   PROCEDURE Add_Message (
         My_Messages        : IN OUT Messages_Set;
         Name               : IN     Unbounded_String;
         Size               : IN     Natural;
         Period             : IN     Natural;
         Deadline           : IN     Natural;
         Jitter             : IN     Natural;
         Param              : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Response_Time      : IN     Natural;
         Communication_Time : IN     Natural) IS

      Dummy : Message_Ptr;
   BEGIN
      Add_Message(My_Messages, Dummy, Name, Size,
         Period, Deadline, Jitter, Param, Response_Time,
         Communication_Time);
   END Add_Message;


   PROCEDURE Check_Message (
         My_Messages        : IN     Messages_Set;
         Name               : IN     Unbounded_String;
         Size               : IN     Natural;
         Period             : IN     Natural;
         Deadline           : IN     Natural;
         Jitter             : IN     Natural;
         Param              : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Response_Time      : IN     Natural;
         Communication_Time : IN     Natural) IS

      My_Iterator : Iterator;
      A_Message   : Message_Ptr;

   BEGIN

      IF (Name = "")
            THEN
         RAISE Invalid_Parameter;
      END IF;

      IF (Size = 0)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      IF (Jitter > Deadline)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      IF(Get_Number_Of_Elements(My_Messages) > 0)
            THEN

         Reset_Iterator(My_Messages, My_Iterator);

         LOOP
            Current_Element(My_Messages, A_Message, My_Iterator);
            IF (Name = A_Message.Name)
                  THEN
               RAISE Message_Already_Defined;
            END IF;

            EXIT WHEN Is_Last_Element(My_Messages, My_Iterator);

            Next_Element(My_Messages, My_Iterator);
         END LOOP;
      END IF;

   END Check_Message;


   PROCEDURE Add_Message (
         My_Messages        : IN OUT Messages_Set;
         A_Message          : IN OUT Message_Ptr;
         Name               : IN     Unbounded_String;
         Size               : IN     Natural;
         Period             : IN     Natural;
         Deadline           : IN     Natural;
         Jitter             : IN     Natural;
         Param              : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Response_Time      : IN     Natural;
         Communication_Time : IN     Natural) IS


      New_Periodic_Message  : Periodic_Message_Ptr;
      New_Aperiodic_Message : Aperiodic_Message_Ptr;

   BEGIN

      Check_Message(My_Messages, Name, Size, Period, Deadline, Jitter,
         Param, Response_Time, Communication_Time);

      IF (Period /= 0)
            THEN
         New_Periodic_Message := NEW Periodic_Message;
         New_Periodic_Message.Jitter:=Jitter;
         New_Periodic_Message.Period :=Period;
         A_Message:=Message_Ptr(New_Periodic_Message);

      ELSE
         New_Aperiodic_Message:= NEW Aperiodic_Message;
         A_Message:=Message_Ptr(New_Aperiodic_Message);
      END IF;


      A_Message.Name:=Name;
      A_Message.Size:=Size;
      A_Message.Response_Time := Response_Time;
      A_Message.Communication_Time := Communication_Time;
      A_Message.Deadline:=Deadline;
      A_Message.Param:=Param;


      Add(My_Messages,A_Message);

   END Add_Message;


   PROCEDURE Update_Message (
         My_Messages        : IN OUT Messages_Set;
         Name               : IN     Unbounded_String;
         New_Name           : IN     Unbounded_String;
         Size               : IN     Natural;
         Period             : IN     Natural;
         Deadline           : IN     Natural;
         Jitter             : IN     Natural;
         Param              : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Response_Time      : IN     Natural;
         Communication_Time : IN     Natural) IS

      A_Message          : Message_Ptr;
      A_Periodic_Message : Periodic_Message_Ptr;
   BEGIN
      A_Message:=Search_Message(My_Messages,Name);
      A_Message.Name:=New_Name;
      A_Message.Size:=Size;
      A_Message.Deadline:=Deadline;
      A_Message.Param:=Param;
      A_Message.Response_Time:=Response_Time;
      A_Message.Communication_Time := Communication_Time;
      IF(A_Message.Message_Type=Periodic_Type) THEN
         A_Periodic_Message := Periodic_Message_Ptr(A_Message);
         A_Periodic_Message.Period:=Period;
         A_Periodic_Message.Jitter:=Jitter;
      END IF;

   END Update_Message;


   FUNCTION Have_Deadlines_Equal_Than_Periods (
         My_Messages : IN     Messages_Set)
     RETURN Boolean IS

      Equal_Than  : Boolean     := True;
      A_Message   : Message_Ptr;
      My_Iterator : Iterator;

   BEGIN

      Reset_Iterator(My_Messages, My_Iterator);

      LOOP
         Current_Element(My_Messages,A_Message,My_Iterator);


         IF (A_Message.Message_Type /= Periodic_Type)
               THEN
            RETURN False;
         END IF;

         IF  (Periodic_Message_Ptr(A_Message).Deadline =
               Periodic_Message_Ptr(A_Message).Period)
               THEN
            Equal_Than := False;
         END IF;


         EXIT WHEN Is_Last_Element(My_Messages, My_Iterator);

         Next_Element(My_Messages, My_Iterator);

      END LOOP;

      RETURN Equal_Than;

   END Have_Deadlines_Equal_Than_Periods ;



   PROCEDURE Get_Global_Message_Type (
         My_Messages     : IN     Messages_Set;
         Has_Global_Type :    OUT Boolean;
         Global_Type     :    OUT Messages_Type) IS

      A_Message   : Message_Ptr;
      My_Iterator : Iterator;

   BEGIN

      Has_Global_Type:=True;
      Reset_Iterator(My_Messages, My_Iterator);


      -- Find the first message of the processor
      --
      Current_Element(My_Messages,A_Message,My_Iterator);

      Global_Type := A_Message.Message_Type;
      IF Is_Last_Element(My_Messages, My_Iterator)
            THEN
         RETURN;
      ELSE
         Next_Element(My_Messages, My_Iterator);
      END IF;

      LOOP
         Current_Element(My_Messages,A_Message,My_Iterator);

         IF (Global_Type /= A_Message.Message_Type)
               THEN
            Has_Global_Type:=False;
            RETURN;
         END IF;

         EXIT WHEN Is_Last_Element(My_Messages, My_Iterator);

         Next_Element(My_Messages, My_Iterator);

      END LOOP;

   END Get_Global_Message_Type;




   FUNCTION Search_Message (
         My_Messages : IN     Messages_Set;
         Name        : IN     Unbounded_String)
     RETURN Message_Ptr IS

      My_Iterator : Iterator;
      A_Message   : Message_Ptr;
      Result      : Message_Ptr;

      Found : Boolean := False;

   BEGIN

      Reset_Iterator(My_Messages, My_Iterator);

      LOOP
         Current_Element(My_Messages,A_Message,My_Iterator);

         IF(A_Message.Name = Name)
               THEN
            Found:=True;
            Result:=A_Message;
         END IF;

         EXIT WHEN Is_Last_Element(My_Messages, My_Iterator);

         Next_Element(My_Messages, My_Iterator);

      END LOOP;

      IF NOT Found
            THEN
         RAISE Message_Not_Found;
      END IF;

      RETURN Result;
   END Search_Message;


   FUNCTION Increasing_Period (
         Op1 : IN     Message_Ptr;
         Op2 : IN     Message_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Periodic_Message_Ptr(Op1).Period <= Periodic_Message_Ptr(
            Op2).Period);
   END Increasing_Period;


   FUNCTION Decreasing_Period (
         Op1 : IN     Message_Ptr;
         Op2 : IN     Message_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Periodic_Message_Ptr(Op1).Period >= Periodic_Message_Ptr(
            Op2).Period);
   END Decreasing_Period;


   FUNCTION Increasing_Deadline (
         Op1 : IN     Message_Ptr;
         Op2 : IN     Message_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Periodic_Message_Ptr(Op1).Deadline <=
         Periodic_Message_Ptr(
            Op2).Deadline);
   END Increasing_Deadline;


   FUNCTION Decreasing_Deadline (
         Op1 : IN     Message_Ptr;
         Op2 : IN     Message_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Periodic_Message_Ptr(Op1).Deadline >=
         Periodic_Message_Ptr(
            Op2).Deadline);
   END Decreasing_Deadline;


   FUNCTION Get (
         My_Messages  : IN     Messages_Set;
         Message_Name : IN     Unbounded_String;
         Param_Name   : IN     Message_Parameters)
     RETURN Natural IS

      A_Message   : Message_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF ( ( Param_Name /= Size) AND
            (Param_Name /= Period) AND
            (Param_Name /= Deadline) AND
            (Param_Name /= Response_Time) AND
            (Param_Name /= Jitter) )
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Messages, My_Iterator);

      LOOP
         Current_Element(My_Messages,A_Message,My_Iterator);

         IF (A_Message.Name = Message_Name)
               THEN
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Messages, My_Iterator);

         Next_Element(My_Messages, My_Iterator);

      END LOOP;

      IF (Param_Name = Size)
            THEN
         RETURN A_Message.Size;
      ELSE
         IF (Param_Name = Deadline)
               THEN
            RETURN Periodic_Message_Ptr(A_Message).Deadline;
         ELSE
            IF (Param_Name = Response_Time)
                  THEN
               RETURN A_Message.Response_Time;
            ELSE
               IF (Param_Name = Communication_Time)
                     THEN
                  RETURN A_Message.Communication_Time;
               ELSE
                  IF (Param_Name = Period)
                        THEN
                     RETURN Periodic_Message_Ptr(A_Message).Period;
                  ELSE
                     RETURN Periodic_Message_Ptr(A_Message).Jitter;
                  END IF;
               END IF;
            END IF;
         END IF;
      END IF;
   END Get;


   PROCEDURE Set (
         My_Messages  : IN OUT Messages_Set;
         Message_Name : IN     Unbounded_String;
         Param_Name   : IN     Message_Parameters;
         Param_Value  : IN     Natural) IS
      A_Message   : Message_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF ( (Param_Name /= Size) AND
            (Param_Name /= Period) AND
            (Param_Name /= Deadline) AND
            (Param_Name /= Jitter) ) AND
            (Param_Name /= Response_Time) AND
            (Param_Name /= Communication_Time)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Messages, My_Iterator);

      LOOP
         Current_Element(My_Messages,A_Message,My_Iterator);

         IF (A_Message.Name = Message_Name)
               THEN
            IF (Param_Name = Size)
                  THEN
               A_Message.Size:=Param_Value;
            ELSE
               IF (Param_Name = Response_Time)
                     THEN
                  A_Message.Response_Time := Param_Value;
               ELSE
                  IF (Param_Name = Communication_Time)
                        THEN
                     A_Message.Communication_Time := Param_Value;
                     EXIT;
                  END IF;
               END IF;
            END IF;
         END IF;
         IF (Param_Name = Deadline)
               THEN
            Periodic_Message_Ptr(A_Message).Deadline:=Param_Value;
         END IF;
         IF (Param_Name = Period)
               THEN
            Periodic_Message_Ptr(A_Message).Period:=Param_Value;
            EXIT;
         END IF;
         IF (Param_Name = Jitter)
               THEN
            Periodic_Message_Ptr(A_Message).Jitter:=Param_Value;
            EXIT;
         END IF;
         EXIT WHEN Is_Last_Element(My_Messages, My_Iterator);

         Next_Element(My_Messages, My_Iterator);

      END LOOP;
   END Set;




   FUNCTION Export_Xml (
         My_Messages : IN     Messages_Set)
     RETURN Unbounded_String IS

      My_Iterator : Messages_Iterator;
      A_Message   : Message_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN

      IF NOT Is_Empty(My_Messages) THEN

         Result:= Result & To_Unbounded_String( "<messages>") &
            Unbounded_Lf;
         Reset_Iterator(My_Messages, My_Iterator);

         LOOP
            Current_Element(My_Messages,A_Message,My_Iterator);

            Result:= Result & To_Unbounded_String(Ascii.Ht & "<message");
            IF A_Message.X/=0
                  THEN
               Result:= Result & To_Unbounded_String( " x=""" &
                  A_Message.X'Img & """ ");
            END IF;
            IF A_Message.Y/=0
                  THEN
               Result:= Result & To_Unbounded_String( "y=""" & A_Message.Y'
                  Img & """ ");
            END IF;
            Result:= Result & To_Unbounded_String( ">") & Unbounded_Lf;




            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<name>" & To_String(
                  A_Message.Name) & "</name>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<size>" & A_Message.
               Size'
               Img & "</size>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<deadline>" &
               A_Message.
               Deadline'Img & "</deadline>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<response_time>" &
               A_Message.
               Response_Time'Img & "</response_time>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<communication_time>" &
               A_Message.
               Communication_Time'Img & "</communication_time>") &
               Unbounded_Lf;

            IF (A_Message.Message_Type = Periodic_Type)
                  THEN
               Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
                  "<period>" &
                  Periodic_Message_Ptr(A_Message).Period'Img &
                  "</period>") & Unbounded_Lf;
               Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
                  "<jitter>" &
                  Periodic_Message_Ptr(A_Message).Jitter'Img &
                  "</jitter>") & Unbounded_Lf;
            END IF;


            IF A_Message.Param.Nb_Entries>0 THEN
               Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
                  "<parameters>") & Unbounded_Lf;
               FOR I IN 0..A_Message.Param.Nb_Entries-1 LOOP

                  Result:= Result & To_Unbounded_String(Ascii.Ht &
                     Ascii.Ht & Ascii.Ht &
                     "<parameter");
                  IF A_Message.Param.Entries(I).Ptype =
                        Boolean_Parameter THEN
                     Result:= Result & To_Unbounded_String(
                        " parameter_type=""boolean""> ");
                     Result:= Result & To_Unbounded_String( " " &
                        A_Message.Param.Entries(I).
                        Boolean_Value'
                        Img & " " );
                  ELSE
                     IF A_Message.Param.Entries(I).Ptype =
                           Integer_Parameter THEN
                        Result:= Result & To_Unbounded_String(
                           " parameter_type=""integer""> ");
                        Result:= Result & To_Unbounded_String( " " &
                           A_Message.Param.Entries(I).
                           Integer_Value'Img & " " );
                     ELSE
                        IF A_Message.Param.Entries(I).Ptype =
                              Double_Parameter THEN
                           Result:= Result & To_Unbounded_String(
                              " parameter_type=""double""> ");
                           Result:= Result & To_Unbounded_String( " " &
                              To_String(Format(A_Message.
                                    Param.
                                    Entries(I).
                                    Double_Value)) & " " );
                        ELSE
                           Result:= Result & To_Unbounded_String(
                              " parameter_type=""string""> ");
                           Result:= Result & To_Unbounded_String( " " &
                              To_String(A_Message.Param.
                                 Entries(
                                    I).
                                 String_Value) & " " );
                        END IF;
                     END IF;
                  END IF;
                  Result:= Result & To_Unbounded_String( To_String(" " &
                        A_Message.Param. Entries(I).
                        Name) &
                     " ");
                  Result:= Result & To_Unbounded_String("</parameter>") &
                     Unbounded_Lf;
               END LOOP;
               Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
                  "</parameters>") & Unbounded_Lf;
            END IF;

            Result:= Result & To_Unbounded_String(Ascii.Ht & "</message>") &
               Unbounded_Lf;

            EXIT WHEN Is_Last_Element(My_Messages, My_Iterator);

            Next_Element(My_Messages, My_Iterator);

         END LOOP;
         Result:= Result & To_Unbounded_String( "</messages>") &
            Unbounded_Lf;
      END IF;

      RETURN Result;

   END Export_Xml;







   FUNCTION Export_Aadl_User_Defined_Properties (
         My_Messages : IN     Messages_Set)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;

   BEGIN
      RETURN Result;
   END Export_Aadl_User_Defined_Properties;



   FUNCTION Export_Aadl_Implementations (
         My_Messages : IN     Messages_Set)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;

   BEGIN
      RETURN Result;
   END Export_Aadl_Implementations;





END Message_Set;












