------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Ada.Numerics;
USE Ada.Numerics;
WITH Ada.Numerics.Elementary_Functions;
USE Ada.Numerics.Elementary_Functions;


WITH  Scheduler_Io;
USE  Scheduler_Io;

WITH Statements;
USE Statements;
USE Statements.Sets_Type_Package;
WITH Expressions;
USE Expressions;
USE Expressions.Variables_Type_Package;

WITH Processors;
USE Processors;
WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;

WITH Address_Space_Set;
USE Address_Space_Set;
USE Address_Space_Set.Generic_Address_Space_Set;

WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;

WITH Resources;
USE Resources;
USE Resources.Task_Lists;
WITH Resource_Set;
USE Resource_Set;
USE Resource_Set.Generic_Resource_Set;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Ada.Strings.Unbounded.Text_Io;
USE Ada.Strings.Unbounded.Text_Io;

WITH Scheduling_Sequences;
USE Scheduling_Sequences;


WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Scheduler;
USE Scheduler;

WITH Dependency_Services;
USE Dependency_Services;

WITH Translate;
USE Translate;

WITH Io_Tools;
USE Io_Tools;
WITH Ada.Io_Exceptions;
USE Ada.Io_Exceptions;
WITH Ada.Exceptions;
USE Ada.Exceptions;

WITH Xml_Parsers;
USE Xml_Parsers;
WITH Input_Sources.File;
USE Input_Sources.File;
WITH Sax.Readers;
USE Sax.Readers;

WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;

WITH Text_Io;
USE Text_Io;


WITH Parser;
USE Parser;




PACKAGE BODY Multiprocessor_Services IS



   PROCEDURE Put (
         R : IN     Scheduling_Result) IS
   BEGIN
      Put(To_Unbounded_String("Error Message : ") & R.Error_Msg);
      New_Line;
      Put(To_Unbounded_String("Scheduling Message : ") & R.Scheduling_Msg);
      New_Line;

      Put(R.Result.All);
   END Put;


   PROCEDURE Initialize (
         R :    OUT Scheduling_Result) IS
   BEGIN
      R.Scheduling_Msg:=Empty_String;
      R.Error_Msg:=Empty_String;
      R.Result := NEW Scheduling_Sequence;
      Initialize(R.Result.All);
   END Initialize;




   PROCEDURE Build_Multiprocessor_Scheduling (
         My_Processors           : IN     Processors_Set;
         My_Address_Spaces       : IN     Address_Spaces_Set;
         My_Tasks                : IN     Tasks_Set;
         My_Resources            : IN     Resources_Set;
         My_Buffers              : IN     Buffers_Set;
         My_Messages             : IN     Messages_Set;
         My_Dependencies         : IN     Tasks_Dependencies_Ptr;
         Result                  : IN OUT Scheduling_Table_Ptr;
         Event_To_Generate       : IN     Time_Unit_Event_Type_Boolean_Table;
         Last_Time               : IN     Natural;
         With_Offsets            : IN     Boolean                            := True;
         With_Precedencies       : IN     Boolean                            := True;
         With_Resources          : IN     Boolean                            := True;
         With_Task_Specific_Seed : IN     Boolean                            := True;
         Global_Seed_Value       : IN     Integer                            := 0;
         Predictable_Global_Seed : IN     Boolean                            := True) IS

      Tmp_Buffers   : ARRAY (Scheduling_Table_Range) OF Buffers_Set;
      Tmp_Tasks     : ARRAY (Scheduling_Table_Range) OF Tasks_Set;
      Tmp_Resources : ARRAY (Scheduling_Table_Range) OF Resources_Set;

      Si : Scheduling_Information;

      Iterator3 : Processors_Iterator;

      A_Processor : Processor_Ptr;

      Last_Current_Time : Natural := 0;

      Current_Time : ARRAY (Scheduling_Table_Range) OF Natural := (OTHERS
        => 0);
      Elected      : Tasks_Range                               := Tasks_Range'First;

      No_Task : Boolean := True;

   BEGIN

      Si.Number_Of_Tasks:=0;
      Si.Number_Of_Resources:=0;
      Si.Number_Of_Processors:=Integer(Get_Number_Of_Elements(
            My_Processors));
      Si.Number_Of_Address_Spaces:=Integer(Get_Number_Of_Elements(
            My_Address_Spaces));
      Si.Simulation_Length:=Last_Time;


      -- Before scheduling, check if the tasks can be
      -- schedule with the selected scheduler
      --
      Reset_Iterator(My_Processors, Iterator3);

      LOOP
         Current_Element(My_Processors, A_Processor, Iterator3);

         BEGIN

            Result.Entries(Result.Nb_Entries).Item:=A_Processor;
            Result.Entries(Result.Nb_Entries).Data.Scheduling_Msg:=
               Empty_String;
            Result.Entries(Result.Nb_Entries).Data.Error_Msg:=
               Empty_String;

            IF (Get_Number_Of_Task_From_Processor(My_Tasks,
                     A_Processor.
                     Name)/=0)
                  THEN


               Check_Before_Scheduling(A_Processor.Scheduler.All,
                  My_Tasks,
                  A_Processor.Name);


               IF With_Precedencies THEN
                  Dependencies_Periods_Control(My_Tasks,
                     My_Dependencies,
                     A_Processor.Name );
               END IF;


            ELSE
               Result.Entries(Result.Nb_Entries).Data.Error_Msg:=
                  Lb_Minus &
                  Lb_Define_Tasks_Before(Current_Language);

            END IF;


         EXCEPTION
            WHEN Dependency_Services.Dependencies_Period_Error=>
               Result.Entries(Result.Nb_Entries).Data.Error_Msg:=
                  Lb_Precedencies_Period_Error(Current_Language) &
                  Unbounded_Lf;
            WHEN Task_Set.Task_Must_Be_Periodic=>
               Result.Entries(Result.Nb_Entries).Data.Error_Msg:=
                  Lb_Compute_Scheduling_Error_15(Current_Language) &
                  Unbounded_Lf;
            WHEN Task_Set.Task_Must_Have_Period_Equal_To_Deadline=>
               Result.Entries(Result.Nb_Entries).Data.Error_Msg:=
                  Lb_Compute_Scheduling_Error_14(Current_Language) &
                  Unbounded_Lf;
            WHEN Task_Set.Task_Model_Error =>
               Result.Entries(Result.Nb_Entries).Data.Error_Msg:=
                  Lb_Compute_Scheduling_Error_1(Current_Language) &
                  Unbounded_Lf;
            WHEN Task_Set.Priority_Error =>
               Result.Entries(Result.Nb_Entries).Data.Error_Msg:=
                  Lb_Compute_Scheduling_Error_11(Current_Language) &
                  Unbounded_Lf;

         END;

         Result.Nb_Entries:=Result.Nb_Entries+1;

         EXIT WHEN Is_Last_Element(My_Processors, Iterator3);
         Next_Element(My_Processors, Iterator3);

      END LOOP;



      -- First : gather information
      --
      FOR J IN 0..Result.Nb_Entries-1 LOOP



         -- Duplicate data before scheduling (if case of data are changed)
         --
         IF Result.Entries(J).Data.Error_Msg=Empty_String THEN

            Current_Processor_Name:=Result.Entries(J).Item.Name;

            Select_And_Copy(
               My_Tasks, Tmp_Tasks(J), Select_Cpu'ACCESS);
            Select_And_Copy(
               My_Resources, Tmp_Resources(J), Select_Cpu'ACCESS);
            Select_And_Copy(
               My_Buffers, Tmp_Buffers(J), Select_Cpu'ACCESS);



            Scheduler_Initialization (
               Result.Entries(J).Item.Scheduler.All,  Si,
               Result.Entries(J).Item.Name, Tmp_Tasks(J),
               Tmp_Resources(J), Tmp_Buffers(J), Result.Entries(J).
               Data.Result,
               With_Offsets, With_Precedencies, With_Resources,
               With_Task_Specific_Seed, Global_Seed_Value,
               Predictable_Global_Seed,
               Last_Time, Event_To_Generate);


            Specific_Scheduler_Initialization(
               Result.Entries(J).Item.Scheduler.All, Si,
               Result.Entries(J).Item.Name,
               Tmp_Tasks(J), Tmp_Resources(J), Tmp_Buffers(J),
               My_Messages, Result.Entries(J).Data.
               Scheduling_Msg);


         END IF;

      END LOOP;





      -- Second : compute scheduling
      --

      WHILE(Last_Current_Time < Last_Time) LOOP


         Last_Current_Time:=Natural'Last;

         FOR J IN 0..Result.Nb_Entries -1 LOOP


            IF (Result.Entries(J).Data.Error_Msg = Empty_String) THEN


               IF (Current_Time(J) < Last_Time) THEN

                  Do_Election(Result.Entries(J).
                     Item.Scheduler.All,
                     Si, Result.Entries(J).Data.Result,
                     Result.Entries(J).
                     Data.Scheduling_Msg,
                     Current_Time(J),
                     Result.Entries(J).Item.Name,
                     My_Dependencies, With_Offsets,
                     With_Precedencies, With_Resources,
                     Event_To_Generate,
                     Elected,
                     No_Task);


                  -- No task to schedule
                  --
                  IF No_Task
                        THEN
                     Current_Time(J):=Current_Time(J)+1;

                  ELSE

                     Run_Task(Result.Entries(J).
                        Item.Scheduler.All,
                        Result.Entries(J).Item.Name,
                        Si, My_Dependencies, Elected, Result.Entries(
                           J).
                        Data.Result,
                        Current_Time(J), Last_Time,
                        With_Offsets, With_Precedencies,
                        With_Resources,
                        Event_To_Generate);

                  END IF;

                  Last_Current_Time:=Natural'Min(Last_Current_Time,
                     Current_Time(J));

               END IF;
            END IF;

         END LOOP;

      END LOOP;


   END Build_Multiprocessor_Scheduling;





   PROCEDURE Put (
         Sched : IN     Scheduling_Table_Ptr) IS

      Global_Sched : ARRAY (Scheduling_Table_Range) OF Scheduling_Sequence_Ptr;

      Global_Sched_Index : ARRAY (Scheduling_Table_Range) OF Time_Unit_Range :=
        (OTHERS => 0);

      Global_Sched_Number        : Scheduling_Table_Range := 0;
      Min_Time                   : Natural                := 0;
      Current_Min_Time_Processor : Scheduling_Table_Range;



   BEGIN

      New_Line;
      New_Line;
      Put_Line("Event Table after scheduling : ");
      New_Line;
      Put_Line(
         "Event type/Task, buffer, message or resource name/Date");
      New_Line;
      New_Line;

      -- how many processors scheduling should we run analysis ?
      --
      FOR I IN   0.. Sched.Nb_Entries-1 LOOP
         IF (Sched.Entries(I).Data.Error_Msg = Empty_String)
               THEN
            Global_Sched(Global_Sched_Number):=Sched.Entries(I).
               Data.Result;
            Global_Sched_Number:=Global_Sched_Number+1;
         END IF;
      END LOOP;


      -- Now, scan each time unit
      --
      WHILE   Min_Time/=Natural'Last  LOOP


         -- find the smallest time unit
         --
         Min_Time:=Natural'Last;
         FOR I IN   0.. Global_Sched_Number-1 LOOP
            IF Global_Sched_Index(I) < Global_Sched(I).
                  Nb_Entries THEN
               IF Global_Sched(I).Entries(Global_Sched_Index(I)).
                     Data < Min_Time
                     THEN
                  Min_Time:=Global_Sched(I).Entries(
                     Global_Sched_Index(I)).Data;
                  Current_Min_Time_Processor:=I;
               END IF;
            END IF;
         END LOOP;


         IF Min_Time /= Natural'Last THEN

            Put (Global_Sched(Current_Min_Time_Processor).
               Entries(Global_Sched_Index(
                     Current_Min_Time_Processor))  );


            Global_Sched_Index(Current_Min_Time_Processor):=
               Global_Sched_Index(Current_Min_Time_Processor)+1;

         END IF;


      END LOOP;


      New_Line;
      New_Line;

   END Put;




   PROCEDURE Run_An_Event_Analyzer (
         My_Tasks          : IN     Tasks_Set;
         My_Processors     : IN     Processors_Set;
         My_Address_Spaces : IN     Address_Spaces_Set;
         My_Resources      : IN     Resources_Set;
         My_Buffers        : IN     Buffers_Set;
         My_Messages       : IN     Messages_Set;
         Sched             : IN     Scheduling_Table_Ptr;
         An_Event_Analyzer : IN     Event_Analyzer_Ptr;
         Result            : IN OUT Unbounded_String) IS


      Current : Statement_Ptr;

      Var,
      Var_Time,
      Var_Type,
      Var_Processor,
      Var_Message,
      Var_Resource,
      Var_Task,
      Var_Buffer    : Variables_Range;

      Global_Sched : ARRAY (Scheduling_Table_Range) OF Scheduling_Sequence_Ptr;

      Global_Sched_Index    : ARRAY (Scheduling_Table_Range) OF Time_Unit_Range  :=
        (OTHERS => 0);
      Global_Processor_Name : ARRAY (Scheduling_Table_Range) OF Unbounded_String;

      Global_Sched_Number        : Scheduling_Table_Range := 0;
      Min_Time                   : Natural                := 0;
      Current_Min_Time_Processor : Scheduling_Table_Range;



   BEGIN

      BEGIN
         An_Event_Analyzer.String_Behavior:=Read_Sequential_File(
            An_Event_Analyzer.Behavior_File_Name);


      EXCEPTION
         WHEN Ada.Io_Exceptions.Name_Error =>
            Raise_Exception(Parametric_File_Error'Identity,
               To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  An_Event_Analyzer.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(Current_Language))  &
               ", Name_Error");
         WHEN Ada.Io_Exceptions.Status_Error =>
            Raise_Exception(Parametric_File_Error'Identity,
               To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  An_Event_Analyzer.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Status_Error");
         WHEN Ada.Io_Exceptions.Mode_Error   =>
            Raise_Exception(Parametric_File_Error'Identity,
               To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  An_Event_Analyzer.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Mode_Error");
         WHEN Ada.Io_Exceptions.Use_Error     =>
            Raise_Exception(Parametric_File_Error'Identity,
               To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  An_Event_Analyzer.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(Current_Language)) &
               ", Use_Error");
         WHEN Ada.Io_Exceptions.Device_Error  =>
            Raise_Exception(Parametric_File_Error'Identity,
               To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  An_Event_Analyzer.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(Current_Language)) &
               ", Device_Error");
         WHEN Ada.Io_Exceptions.End_Error    =>
            Raise_Exception(Parametric_File_Error'Identity,
               To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  An_Event_Analyzer.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", End_Error");
         WHEN Ada.Io_Exceptions.Data_Error   =>
            Raise_Exception(Parametric_File_Error'Identity,
               To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  An_Event_Analyzer.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Data_Error");
         WHEN Ada.Io_Exceptions.Layout_Error  =>
            Raise_Exception(Parametric_File_Error'Identity,
               To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  An_Event_Analyzer.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Layout_Error");
         WHEN OTHERS =>
            Raise_Exception(Parametric_File_Error'Identity,
               To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  An_Event_Analyzer.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) );
      END;


      -- We should check the scheduler behavior syntax ...
      --
      Open_Input(An_Event_Analyzer.String_Behavior);
      Parser.Init_Parser;
      Create_Parametric_Variables(Parser.Variables_Table, My_Tasks);
      Parser.Yyparse;

      -- Store syntax tree and compiling information
      --
      An_Event_Analyzer.Root_Statement_Pointer:=
         Parser.Root_Statement_Pointer;

      An_Event_Analyzer.Variables_Table:=
         Parser.Variables_Table;

      An_Event_Analyzer.Sets_Table:=
         Parser.Sets_Table;



      -- Initialize parametric variables which are constant
      --
      Initialize_Parametric_Variables(
         An_Event_Analyzer.Variables_Table,
         My_Messages,
         My_Buffers,
         My_Resources,
         My_Tasks);



      -- "nb_processors"
      --
      Var:=Find_Variable(An_Event_Analyzer.Variables_Table,
         To_Unbounded_String("nb_processors"));
      An_Event_Analyzer.Variables_Table.Entries(Var).Simulation.
         Integer_Value:=Integer(Get_Number_Of_Elements(My_Processors));

      -- "nb_address_spaces"
      --
      Var:=Find_Variable(An_Event_Analyzer.Variables_Table,
         To_Unbounded_String("nb_address_spaces"));
      An_Event_Analyzer.Variables_Table.Entries(Var).Simulation.
         Integer_Value:=Integer(Get_Number_Of_Elements(
            My_Address_Spaces));





      -- Now : run "start" statements
      --
      Current := An_Event_Analyzer.Root_Statement_Pointer (
         Statements.Start);
      Dispatch(Current, Start, An_Event_Analyzer.Variables_Table,
         Result);

      -- Now : run "event analyzer" statements
      --
      Current := An_Event_Analyzer.Root_Statement_Pointer (
         Statements.Gather_Event_Analyzer);


      -- Find variables before running the simulation
      --
      Var_Time:=Find_Variable(An_Event_Analyzer.Variables_Table,
         To_Unbounded_String("events.time"));
      Var_Type:=Find_Variable(An_Event_Analyzer.Variables_Table,
         To_Unbounded_String("events.type"));
      Var_Processor:=Find_Variable(
         An_Event_Analyzer.Variables_Table,
         To_Unbounded_String("events.processor_name"));
      Var_Message:=Find_Variable(An_Event_Analyzer.Variables_Table,
         To_Unbounded_String("events.message_name"));
      Var_Resource:=Find_Variable(An_Event_Analyzer.Variables_Table,
         To_Unbounded_String("events.resource_name"));
      Var_Task:=Find_Variable(An_Event_Analyzer.Variables_Table,
         To_Unbounded_String("events.task_name"));
      Var_Buffer:=Find_Variable(An_Event_Analyzer.Variables_Table,
         To_Unbounded_String("events.buffer_name"));



      -- how many processors scheduling should we run analysis ?
      --
      FOR I IN   0.. Sched.Nb_Entries-1 LOOP
         IF (Sched.Entries(I).Data.Error_Msg = Empty_String)
               THEN
            Global_Sched(Global_Sched_Number):=Sched.Entries(I).
               Data.Result;
            Global_Processor_Name(Global_Sched_Number):=
               Sched.Entries(I).
               Item.Name;
            Global_Sched_Number:=Global_Sched_Number+1;
         END IF;
      END LOOP;

      -- Now, scan each time unit
      --
      WHILE   Min_Time/=Natural'Last  LOOP


         -- find the smallest time unit
         --
         Min_Time:=Natural'Last;
         FOR I IN   0.. Global_Sched_Number-1 LOOP
            IF Global_Sched_Index(I) < Global_Sched(I).
                  Nb_Entries THEN
               IF Global_Sched(I).Entries(Global_Sched_Index(I)).
                     Data < Min_Time
                     THEN
                  Min_Time:=Global_Sched(I).Entries(
                     Global_Sched_Index(I)).Data;
                  Current_Min_Time_Processor:=I;
               END IF;
            END IF;
         END LOOP;


         IF Min_Time /= Natural'Last THEN

            -- "events.time"
            --
            An_Event_Analyzer.Variables_Table.Entries(Var_Time).
               Simulation.
               Integer_Value:=
               Integer(Global_Sched(Current_Min_Time_Processor).
               Entries(Global_Sched_Index(
                     Current_Min_Time_Processor)).
               Data);

            -- "events.type"
            --
            An_Event_Analyzer.Variables_Table.Entries(Var_Type).
               Simulation.
               String_Value:=To_Lower(Global_Sched(
                  Current_Min_Time_Processor).
               Entries(Global_Sched_Index(
                     Current_Min_Time_Processor)).
               Item.Event_Type'Img);

            -- "events.processor_name"
            --
            An_Event_Analyzer.Variables_Table.Entries(Var_Processor).
               Simulation.
               String_Value:=Global_Processor_Name(
               Current_Min_Time_Processor);

            -- "events.buffer_name"
            --
            An_Event_Analyzer.Variables_Table.Entries(Var_Buffer).
               Simulation.
               String_Value:=Empty_String;

            -- "events.resource_name"
            --
            An_Event_Analyzer.Variables_Table.Entries(Var_Resource).
               Simulation.
               String_Value:=Empty_String;

            -- "events.message_name"
            --
            An_Event_Analyzer.Variables_Table.Entries(Var_Message).
               Simulation.
               String_Value:=Empty_String;


            CASE Global_Sched(Current_Min_Time_Processor).
                     Entries(
                     Global_Sched_Index(Current_Min_Time_Processor)).
                     Item.Event_Type IS

               WHEN Start_Of_Task_Capacity =>
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Task).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Start_Task.Name;

               WHEN End_Of_Task_Capacity =>
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Task).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.End_Task.Name;

               WHEN Write_To_Buffer =>
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Task).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Write_Task.Name;
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Buffer).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Write_Buffer.Name;

               WHEN Read_From_Buffer =>
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Task).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Read_Task.Name;
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Buffer).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Read_Buffer.Name;

               WHEN Running_Task =>
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Task).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Running_Task.Name;

               WHEN Task_Activation =>
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Task).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Activation_Task.Name;

               WHEN Allocate_Resource =>
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Task).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Allocate_Task.Name;
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Resource).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Allocate_Resource.Name;

               WHEN Release_Resource =>
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Task).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Release_Task.Name;
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Resource).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Release_Resource.Name;

               WHEN Wait_For_Resource =>
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Task).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Wait_Task.Name;
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Resource).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Wait_Resource.Name;

               WHEN Send_Message =>
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Task).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Send_Task.Name;
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Message).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Send_Message.Name;

               WHEN Receive_Message =>
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Task).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Receive_Task.Name;
                  An_Event_Analyzer.Variables_Table.Entries(
                     Var_Message).
                     Simulation.
                     String_Value:=Global_Sched(
                     Current_Min_Time_Processor).
                     Entries(Global_Sched_Index(
                        Current_Min_Time_Processor)).
                     Item.Receive_Message.Name;

            END CASE;




            -- Do analysis on the current time : run "event_analyzer"  section/statements
            --
            Dispatch(Current, Statements.Gather_Event_Analyzer,
               An_Event_Analyzer.
               Variables_Table, Result);


            Global_Sched_Index(Current_Min_Time_Processor):=
               Global_Sched_Index(Current_Min_Time_Processor)+1;


         END IF;


      END LOOP;


      -- The end of the line : run "display_event" section/statements
      --
      Current := An_Event_Analyzer.Root_Statement_Pointer (
         Statements.Display_Event_Analyzer);
      Dispatch(Current, Display_Event_Analyzer, An_Event_Analyzer.
         Variables_Table, Result);


   END Run_An_Event_Analyzer;



   PROCEDURE Free (
         Sched : IN OUT Scheduling_Table_Ptr) IS

      PROCEDURE Free_Pointer IS
      NEW Unchecked_Deallocation
         (Scheduling_Table, Scheduling_Table_Ptr);

   BEGIN
      FOR I IN Scheduling_Table_Range LOOP
         IF Sched.Entries(I).Data.Result/=NULL
               THEN
            Free(Sched.Entries(I).Data.Result);
         END IF;
      END LOOP;
      Free_Pointer(Sched);
   END Free;


   PROCEDURE Read_From_Xml_File (
         Sched     : IN OUT Scheduling_Table_Ptr;
         Sys       : IN     System;
         File_Name : IN     String) IS
   BEGIN
      Read_From_Xml_File(Sched, Sys, To_Unbounded_String(File_Name));
   END Read_From_Xml_File;


   PROCEDURE Read_From_Xml_File (
         Sched     : IN OUT Scheduling_Table_Ptr;
         Sys       : IN     System;
         File_Name : IN     Unbounded_String) IS

      Read       :          File_Input;
      My_Reader  :          Xml_Event_Table_Parser;
      Name_Start :          Natural;
      S          : CONSTANT String                 := To_String (File_Name);


   BEGIN

      --  Base file name should be used as the public Id
      --
      Name_Start := S'Last;
      WHILE Name_Start >= S'First  AND THEN S (Name_Start) /= '/' LOOP
         Name_Start := Name_Start - 1;
      END LOOP;
      Set_Public_Id (Read, S (Name_Start + 1 .. S'Last));
      Set_System_Id (Read, S);
      Open (S, Read);

      --  xmlns:* attributes will be reported in Start_Element
      --
      Set_Feature (My_Reader, Namespace_Prefixes_Feature, False);
      Set_Feature (My_Reader, Validation_Feature, True);

      Set_Scheduled_System(My_Reader, Sys);
      Parse (My_Reader, Read);
      Free(Sched);
      Sched:=Get_Parsed_Event_Table(My_Reader);
      Close (Read);


   EXCEPTION
      WHEN Xml_Fatal_Error =>
         Close(Read);
         RAISE Xml_Read_Error;

   END  Read_From_Xml_File;


   PROCEDURE Write_To_Xml_File (
         Sched     : IN     Scheduling_Table_Ptr;
         Sys       : IN     System;
         File_Name : IN     String) IS
   BEGIN
      Write_To_Xml_File(Sched, Sys, To_Unbounded_String(File_Name));
   END Write_To_Xml_File;



   PROCEDURE Write_To_Xml_File (
         Sched     : IN     Scheduling_Table_Ptr;
         Sys       : IN     System;
         File_Name : IN     Unbounded_String) IS


      Into         : File_Type;
      An_Event     : Time_Unit_Event_Ptr;
      My_Scheduler : Scheduler_Ptr;
      A_Processor  : Processor_Ptr;

   BEGIN

      -- Open file and Write DTD/XML Header
      --
      Create(Into,
         Mode => Out_File,
         Name => To_String(File_Name));

      Put_Line(Into, "<?xml version=""1.0"" standalone=""yes""?>  ");
      Put_Line(Into,
         "<?xml-stylesheet type=""text/xsl"" href=""event_table.xsl""?>");
      Put_Line(Into, "<!DOCTYPE event_table [  ");
      FOR E IN Time_Unit_Event_Type'RANGE LOOP
         Put_Line(Into, "<!ELEMENT " & To_Lower(E'Img) &
            " (#PCDATA) >  ");
      END LOOP;
      Put_Line(Into, "<!ELEMENT name (#PCDATA) >  ");
      Put_Line(Into, "<!ELEMENT processor (#PCDATA) >  ");

      Put(Into,
         "<!ELEMENT processor (name");
      FOR E IN Time_Unit_Event_Type'RANGE LOOP
         Put(Into, "|(" & To_Lower(E'Img) & ")?");
      END LOOP;
      Put(Into, " ) >  ");

      Put_Line(Into, "] >  ");
      New_Line(Into, 2);


      Put_Line(Into, "<event_table>");



      -- Export event table for each processor
      --
      FOR I IN   0.. Sched.Nb_Entries-1 LOOP
         IF (Sched.Entries(I).Data.Error_Msg = Empty_String)
               THEN

            Put_Line(Into, "   <processor> " );
            Put_Line(Into, "     <name>"  & To_String(Sched.Entries(
                     I).
                  Item.Name) & " </name>");

            A_Processor:=Search_Processor(Sys.Processors,
               Sched.Entries(I).
               Item.Name);
            My_Scheduler:=Get_Scheduler(A_Processor.All);

            FOR J IN 0..Sched.Entries(I).Data.Result.Nb_Entries-
                  1 LOOP

               An_Event:= Sched.Entries(I).Data.Result.Entries(
                  J).Item;
               Put(Into, "      <" & To_Lower(An_Event.Event_Type'
                     Img) &
                  "> ");
               Put(Into, " " & Sched.Entries(I).Data.Result.
                  Entries(
                     J).Data'Img & " ");

               CASE An_Event.Event_Type IS
                  WHEN Start_Of_Task_Capacity =>
                     Put(Into, To_String(
                           Export_Xml_Event_Start_Of_Task_Capacity (
                              My_Scheduler.All, An_Event)));

                  WHEN     End_Of_Task_Capacity =>
                     Put(Into, To_String(
                           Export_Xml_Event_End_Of_Task_Capacity (
                              My_Scheduler.All, An_Event)));

                  WHEN Write_To_Buffer =>
                     Put(Into, To_String(
                           Export_Xml_Event_Write_To_Buffer (
                              My_Scheduler.All, An_Event)));

                  WHEN          Read_From_Buffer =>
                     Put(Into, To_String(
                           Export_Xml_Event_Read_From_Buffer (
                              My_Scheduler.All, An_Event)));

                  WHEN          Running_Task =>
                     Put(Into, To_String(
                           Export_Xml_Event_Running_Task (
                              My_Scheduler.All, An_Event)));

                  WHEN          Task_Activation =>
                     Put(Into, To_String(
                           Export_Xml_Event_Task_Activation (
                              My_Scheduler.All, An_Event)));

                  WHEN          Send_Message =>
                     Put(Into, To_String(
                           Export_Xml_Event_Send_Message (
                              My_Scheduler.All, An_Event)));

                  WHEN         Receive_Message =>
                     Put(Into, To_String(
                           Export_Xml_Event_Receive_Message (
                              My_Scheduler.All, An_Event)));

                  WHEN          Allocate_Resource =>
                     Put(Into, To_String(
                           Export_Xml_Event_Allocate_Resource (
                              My_Scheduler.All,
                              An_Event)));

                  WHEN          Release_Resource =>
                     Put(Into, To_String(
                           Export_Xml_Event_Release_Resource (
                              My_Scheduler.All,
                              An_Event)));

                  WHEN         Wait_For_Resource =>
                     Put(Into, To_String(
                           Export_Xml_Event_Wait_For_Resource (
                              My_Scheduler.All,
                              An_Event)));

               END CASE;

               Put(Into, "</" & To_Lower(
                     An_Event.Event_Type'
                     Img) & "> ");
               New_Line(Into);

            END LOOP;
            Put_Line(Into,
               "   </processor>");
            New_Line(Into);

         END IF;
      END LOOP;


      Put_Line(Into, "</event_table>");
      New_Line(Into);
      New_Line(Into);

      Close(Into);

   END  Write_To_Xml_File;






   PROCEDURE Partition_Small_Task (
         Src_Processors   : IN     Processors_Set;
         Src_Tasks        : IN     Tasks_Set;
         Msg              : IN OUT Unbounded_String;
         Result_Tasks     : IN OUT Tasks_Set;
         Processor_Result :    OUT Processors_Iterator) IS

      A_Task_Result           : Periodic_Task_Ptr;
      My_Task_Iterator_Result : Tasks_Iterator;
      A_Processor             : Processor_Ptr;
      My_Processor_Iterator   : Processors_Iterator;

      Processor_Util,
      Beta           : Float   := 0.0;
      S_Cur          : Float   := 1.0;
      First_Run      : Boolean := True;

   BEGIN

      Validate_Multiprocessor_Tasks(
         Src_Processors, Src_Tasks);

      Duplicate(Src_Tasks, Result_Tasks);
      Sort(Result_Tasks, Increasing_Si'ACCESS);

      Reset_Iterator(Result_Tasks,
         My_Task_Iterator_Result);
      Current_Element(Result_Tasks,
         Task_Ptr(A_Task_Result),
         My_Task_Iterator_Result);
      Reset_Iterator(Src_Processors,
         My_Processor_Iterator);
      Current_Element(Src_Processors,
         A_Processor,
         My_Processor_Iterator);

      LOOP
         IF First_Run = False THEN
            IF Is_Last_Element(
                  Src_Processors,
                  My_Processor_Iterator) THEN
               Duplicate(Src_Tasks,
                  Result_Tasks);
               RAISE No_Such_Processors;
            END IF;
            Next_Element(Src_Processors,
               My_Processor_Iterator);
            Current_Element(
               Src_Processors,
               A_Processor,
               My_Processor_Iterator);
         END IF;
         First_Run := False;
         IF Float(A_Task_Result.Capacity) /
               Float(A_Task_Result.Period) >
               1.0 THEN
            Duplicate(Src_Tasks,
               Result_Tasks);
            RAISE No_Such_Processors;
         END IF;
         A_Task_Result.Cpu_Name :=
            A_Processor.Name;
         Processor_Util := Float(
            A_Task_Result.Capacity) /
            Float(
            A_Task_Result.Period);
         S_Cur := Float(Log(Float(
                  A_Task_Result.Period),
               2.0)) - Float'
            Floor(Float(Log(Float(
                     A_Task_Result.Period),
                  2.0)));
         Beta := 0.0;
         EXIT WHEN Is_Last_Element(
            Result_Tasks,
            My_Task_Iterator_Result);

         LOOP
            IF Is_Last_Element(
                  Result_Tasks,
                  My_Task_Iterator_Result) THEN
               First_Run := True;
               EXIT;
            END IF;
            Next_Element(Result_Tasks,
               My_Task_Iterator_Result);
            Current_Element(Result_Tasks,
               Task_Ptr(A_Task_Result),
               My_Task_Iterator_Result);
            Beta := (Float(Log(Float(
                        A_Task_Result.Period),
                     2.0)) -
               Float'Floor(Float(Log(
                        Float(
                           A_Task_Result.Period),
                        2.0)))) -
               S_Cur;
            Processor_Util :=
               Processor_Util + Float(
               A_Task_Result.Capacity) /
               Float(A_Task_Result.Period);
            IF (Processor_Util <= Float'
                  Max(Float(Log(2.0, E)),
                     1.0 - Beta *
                     Float(Log(2.0, E)))) THEN
               A_Task_Result.Cpu_Name :=
                  A_Processor.Name;
            ELSE
               EXIT;
            END IF;
         END LOOP;

      END LOOP;
      Processor_Result :=
         My_Processor_Iterator;
      Sort(Result_Tasks, Increasing_Name'ACCESS);

   END Partition_Small_Task;

   PROCEDURE Partition_Small_Task (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set;
         Msg            : IN OUT Unbounded_String;
         Result_Tasks   : IN OUT Tasks_Set) IS
      Tmp : Processors_Iterator;
   BEGIN


      Msg:= To_Unbounded_String(" (") &
         Lb_See(Current_Language) &
         To_Unbounded_String(
         "[9], [10]) ");


      Partition_Small_Task(
         Src_Processors, Src_Tasks, Msg,
         Result_Tasks,
         Tmp);
   END Partition_Small_Task;

   PROCEDURE Partition_Small_Task (
         Src_Processors   : IN     Processors_Set;
         Result_Tasks     : IN OUT Tasks_Set;
         Processor_Result :    OUT Processors_Iterator) IS
      Src_Tasks : Tasks_Set;
      Msg       : Unbounded_String;
   BEGIN
      Duplicate(Result_Tasks, Src_Tasks);
      Partition_Small_Task(
         Src_Processors, Src_Tasks, Msg,
         Result_Tasks,
         Processor_Result);
   END Partition_Small_Task;


   PROCEDURE Rm_General_Tasks (
         Src_Processors : IN     Processors_Set;
         Result_Tasks   : IN OUT Tasks_Set) IS

      A_Task_Result           : Periodic_Task_Ptr;
      A_Task_Tmp              : Periodic_Task_Ptr;
      A_Task_Iterator         : Tasks_Iterator;
      My_Task_Iterator_Result : Tasks_Iterator;
      Tmp_Iterator            : Tasks_Iterator;
      A_Processor             : Processor_Ptr;
      My_Processor_Iterator   : Processors_Iterator;
      Src_Tasks               : Tasks_Set;

      Num_Of_Tasks      : Integer := 0;
      Task_Utilization,
      Total_Utilization : Float   := 0.0;
      Goto_First_Loop   : Boolean := False;

   BEGIN
      Duplicate(Result_Tasks, Src_Tasks);
      Reset_Iterator(Result_Tasks,
         My_Task_Iterator_Result);
      Current_Element(Result_Tasks,
         Task_Ptr(A_Task_Result),
         My_Task_Iterator_Result);
      Reset_Iterator(Src_Processors,
         My_Processor_Iterator);
      Current_Element(Src_Processors,
         A_Processor,
         My_Processor_Iterator);

      Sort(Result_Tasks,
         Increasing_Period'ACCESS);

      -- Zero processor locations
      --
      LOOP
         A_Task_Result.Cpu_Name :=
            To_Unbounded_String("");
         EXIT WHEN Is_Last_Element(
            Result_Tasks,
            My_Task_Iterator_Result);
         Next_Element(Result_Tasks,
            My_Task_Iterator_Result);
         Current_Element(Result_Tasks,
            Task_Ptr(A_Task_Result),
            My_Task_Iterator_Result);
      END LOOP;

      Reset_Iterator(Result_Tasks,
         My_Task_Iterator_Result);
      Current_Element(Result_Tasks,
         Task_Ptr(A_Task_Result),
         My_Task_Iterator_Result);
      LOOP

         LOOP
            IF Goto_First_Loop THEN
               Goto_First_Loop := False;
               EXIT;
            END IF;
            -- Fetch information for condition-IP
            --
            Tmp_Iterator :=
               My_Task_Iterator_Result;
            Reset_Iterator(Result_Tasks,
               My_Task_Iterator_Result);
            Current_Element(Result_Tasks,
               Task_Ptr(A_Task_Result),
               My_Task_Iterator_Result);
            Num_Of_Tasks := 0;
            Total_Utilization := 0.0;
            LOOP
               IF A_Task_Result.Cpu_Name =
                     A_Processor.Name THEN
                  Num_Of_Tasks :=
                     Num_Of_Tasks + 1;
                  Total_Utilization :=
                     Total_Utilization +
                     Float(
                     A_Task_Result.Capacity) /
                     Float(A_Task_Result.
                     Period);
               END IF;
               EXIT WHEN Is_Last_Element(
                  Result_Tasks,
                  My_Task_Iterator_Result);
               Next_Element(Result_Tasks,
                  My_Task_Iterator_Result);
               Current_Element(
                  Result_Tasks, Task_Ptr(
                     A_Task_Result),
                  My_Task_Iterator_Result);
            END LOOP;
            My_Task_Iterator_Result :=
               Tmp_Iterator;
            Current_Element(Result_Tasks,
               Task_Ptr(A_Task_Result),
               My_Task_Iterator_Result);
            Task_Utilization := Float(
               A_Task_Result.Capacity) /
               Float(
               A_Task_Result.Period);

            IF Task_Utilization <= 2.0 * (
                  1.0 / ((1.0 +
                        Total_Utilization /
                        Float(
                           Num_Of_Tasks))) **
                  Num_Of_Tasks) - 1.0 THEN
               IF Total_Utilization = 0.0 THEN
                  A_Task_Result.Cpu_Name :=
                     A_Processor.Name;
                  EXIT;
               END IF;
               Reset_Iterator(
                  Result_Tasks,
                  A_Task_Iterator);
               Current_Element(
                  Result_Tasks, Task_Ptr(
                     A_Task_Tmp),
                  A_Task_Iterator);

               LOOP
                  IF A_Task_Tmp.Cpu_Name =
                        A_Processor.Name AND
                        A_Task_Tmp /=
                        A_Task_Result THEN

                     IF
                           A_Task_Result.Period <
                           A_Task_Tmp.Period AND
                           (Integer(Float'
                              Floor(
                                 Float(
                                    A_Task_Tmp.Period) /
                                 Float(
                                    A_Task_Result.Period))) * (
                              A_Task_Tmp.Period -
                              A_Task_Result.
                              Capacity) >=
                           A_Task_Tmp.Capacity OR
                           A_Task_Tmp.Period >=
                           Integer(Float'
                              Ceiling(
                                 Float(
                                    A_Task_Tmp.
                                    Period) /
                                 Float(
                                    A_Task_Result.Period))) *
                           A_Task_Result.Capacity +
                           A_Task_Tmp.Capacity) THEN
                        A_Task_Result.Cpu_Name :=
                           A_Processor.Name;
                        Reset_Iterator(
                           Src_Processors,
                           My_Processor_Iterator);
                        Goto_First_Loop :=
                           True;
                        EXIT;
                     ELSE
                        IF
                              A_Task_Result.Period >=
                              A_Task_Tmp.Period AND (((
                                    Integer(
                                       Float'
                                       Floor(
                                          Float(
                                             A_Task_Result.Period) /
                                          Float(
                                             A_Task_Tmp.Period))) * (
                                       A_Task_Result.Period -
                                       A_Task_Tmp.Capacity)) >=
                                 A_Task_Result.Capacity) OR
                              (
                                 A_Task_Result.Period >= (
                                    Integer(
                                       Float'
                                       Ceiling(
                                          Float(
                                             A_Task_Result.
                                             Period) /
                                          Float(
                                             A_Task_Tmp.Period))) *
                                    A_Task_Tmp.Capacity +
                                    A_Task_Result.Capacity))) THEN
                           A_Task_Result.Cpu_Name :=
                              A_Processor.Name;
                           Reset_Iterator(
                              Src_Processors,
                              My_Processor_Iterator);
                           Goto_First_Loop :=
                              True;
                           EXIT;
                        ELSE
                           IF
                                 Is_Last_Element(
                                 Src_Processors,
                                 My_Processor_Iterator) THEN
                              Duplicate(
                                 Src_Tasks,
                                 Result_Tasks);
                              RAISE
                                 No_Such_Processors;
                           END IF;
                           Next_Element(
                              Src_Processors,
                              My_Processor_Iterator);
                           Current_Element(
                              Src_Processors,
                              A_Processor,
                              My_Processor_Iterator);
                           EXIT;
                        END IF;
                     END IF;
                  END IF;
                  EXIT WHEN
                     Is_Last_Element(
                     Result_Tasks,
                     A_Task_Iterator);
                  Next_Element(
                     Result_Tasks,
                     A_Task_Iterator);
                  Current_Element(
                     Result_Tasks,
                     Task_Ptr(A_Task_Tmp),
                     A_Task_Iterator);
               END LOOP;

            ELSE
               IF Is_Last_Element(
                     Src_Processors,
                     My_Processor_Iterator) THEN
                  Duplicate(Src_Tasks,
                     Result_Tasks);
                  RAISE
                     No_Such_Processors;
               END IF;
               Next_Element(
                  Src_Processors,
                  My_Processor_Iterator);
               Current_Element(
                  Src_Processors,
                  A_Processor,
                  My_Processor_Iterator);
            END IF;

         END LOOP;

         EXIT WHEN Is_Last_Element(
            Result_Tasks,
            My_Task_Iterator_Result);
         Next_Element(Result_Tasks,
            My_Task_Iterator_Result);
      END LOOP;

   END Rm_General_Tasks;



   PROCEDURE Partition_General_Task (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set;
         Msg            : IN OUT Unbounded_String;
         Result_Tasks   : IN OUT Tasks_Set) IS

      A_Task_Result           : Periodic_Task_Ptr;
      My_Task_Iterator_Result : Tasks_Iterator;
      My_Processor_Iterator   : Processors_Iterator;
      A_Processor             : Processor_Ptr;

      Rmst_Tasks : Tasks_Set;
      Rmgt_Tasks : Tasks_Set;
      Tmp_Cpus   : Processors_Set;

      Rmst_Index,
      Rmgt,
      Num_Of_Tasks : Integer := 0;

   BEGIN


      Validate_Multiprocessor_Tasks(
         Src_Processors, Src_Tasks);

      Msg:= To_Unbounded_String(" (") &
         Lb_See(Current_Language) &
         To_Unbounded_String(
         "[9], [10]) ");

      Duplicate(Src_Tasks, Result_Tasks);
      Sort(Result_Tasks,
         Increasing_Utilization'ACCESS);

      -- Scan for max Alfa to make division between RMST and RMGT tasks
      --
      Reset_Iterator(Result_Tasks,
         My_Task_Iterator_Result);
      Current_Element(Result_Tasks,
         Task_Ptr(A_Task_Result),
         My_Task_Iterator_Result);
      LOOP
         IF Float(A_Task_Result.Capacity) /
               Float(A_Task_Result.
               Period) >
               1.0/3.0 THEN
            EXIT;
         ELSE
            Rmgt := Rmgt + 1;
         END IF;
         EXIT WHEN Is_Last_Element(
            Result_Tasks,
            My_Task_Iterator_Result);
         Next_Element(Result_Tasks,
            My_Task_Iterator_Result);
         Current_Element(Result_Tasks,
            Task_Ptr(A_Task_Result),
            My_Task_Iterator_Result);
      END LOOP;

      -- Make bin-packing for RMST and RMGT
      --
      Reset_Iterator(Result_Tasks,
         My_Task_Iterator_Result);
      Current_Element(Result_Tasks,
         Task_Ptr(A_Task_Result),
         My_Task_Iterator_Result);
      LOOP
         EXIT WHEN Rmst_Index = Rmgt;
         Add(Rmst_Tasks, Task_Ptr(
               A_Task_Result));
         Rmst_Index := Rmst_Index + 1;
         Next_Element(Result_Tasks,
            My_Task_Iterator_Result);
         Current_Element(Result_Tasks,
            Task_Ptr(A_Task_Result),
            My_Task_Iterator_Result);
      END LOOP;

      Num_Of_Tasks := Integer(
         Get_Number_Of_Elements(
            Result_Tasks));
      IF Rmgt /= Num_Of_Tasks THEN
         LOOP
            Current_Element(Result_Tasks,
               Task_Ptr(A_Task_Result),
               My_Task_Iterator_Result);
            Add(Rmgt_Tasks, Task_Ptr(
                  A_Task_Result));
            EXIT WHEN Is_Last_Element(
               Result_Tasks,
               My_Task_Iterator_Result);
            Next_Element(Result_Tasks,
               My_Task_Iterator_Result);
         END LOOP;
      END IF;

      -- Call RMST and RMGT functions respectively
      --
      IF(Is_Empty(Rmst_Tasks) = False) THEN
         Partition_Small_Task(
            Src_Processors, Rmst_Tasks,
            My_Processor_Iterator);
      END IF;

      IF False = Is_Last_Element(
            Src_Processors,
            My_Processor_Iterator) AND
            Is_Empty(Rmgt_Tasks) = False THEN
         -- Make processor set for RMGT
         --
         Next_Element(Src_Processors,
            My_Processor_Iterator);
         Current_Element(Src_Processors,
            A_Processor,
            My_Processor_Iterator);
         LOOP
            Add(Tmp_Cpus, A_Processor);
            EXIT WHEN Is_Last_Element(
               Src_Processors,
               My_Processor_Iterator);
            Next_Element(Src_Processors,
               My_Processor_Iterator);
            Current_Element(
               Src_Processors,
               A_Processor,
               My_Processor_Iterator);
         END LOOP;
         Rm_General_Tasks(Tmp_Cpus,
            Rmgt_Tasks);

      ELSE
         IF Is_Last_Element(
               Src_Processors,
               My_Processor_Iterator) AND
               Is_Empty(Rmgt_Tasks) = False THEN
            Duplicate(Src_Tasks,
               Result_Tasks);
            RAISE No_Such_Processors;
         END IF;
      END IF;

      -- Put results of two functions to returning task set
      --
      IF NOT Is_Empty(Rmst_Tasks) THEN
         Duplicate(Rmst_Tasks,
            Result_Tasks);
      END IF;

      IF NOT Is_Empty(Rmgt_Tasks) THEN
         Reset_Iterator(Rmgt_Tasks,
            My_Task_Iterator_Result);
         Current_Element(Rmgt_Tasks,
            Task_Ptr(A_Task_Result),
            My_Task_Iterator_Result);
         LOOP
            Add(Result_Tasks, Task_Ptr(
                  A_Task_Result));
            EXIT WHEN Is_Last_Element(
               Rmgt_Tasks,
               My_Task_Iterator_Result);
            Next_Element(Rmgt_Tasks,
               My_Task_Iterator_Result);
            Current_Element(Rmgt_Tasks,
               Task_Ptr(A_Task_Result),
               My_Task_Iterator_Result);
         END LOOP;
      END IF;
      Sort(Result_Tasks, Increasing_Name'ACCESS);

   END Partition_General_Task;

   PROCEDURE Partition_Next_Fit (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set;
         Msg            : IN OUT Unbounded_String;
         Result_Tasks   : IN OUT Tasks_Set) IS

      A_Task_Result           : Periodic_Task_Ptr;
      My_Task_Iterator_Result : Tasks_Iterator;
      A_Processor             : Processor_Ptr;
      My_Processor_Iterator   : Processors_Iterator;

      M                 : Integer := 2;
      Total_Utilization : Float   := 0.0;
      Condition_Ip      : Boolean := False;

   BEGIN

      Validate_Multiprocessor_Tasks(
         Src_Processors, Src_Tasks);

      Msg:= To_Unbounded_String(" (") &
         Lb_See(Current_Language) &
         To_Unbounded_String(
         "[9], [10]) ");


      Duplicate(Src_Tasks, Result_Tasks);
      Sort(Result_Tasks,
         Increasing_Period'ACCESS);

      Reset_Iterator(Result_Tasks,
         My_Task_Iterator_Result);
      Current_Element(Result_Tasks,
         Task_Ptr(A_Task_Result),
         My_Task_Iterator_Result);
      Reset_Iterator(Src_Processors,
         My_Processor_Iterator);
      Current_Element(Src_Processors,
         A_Processor,
         My_Processor_Iterator);

      LOOP
         IF Float(A_Task_Result.Capacity) /
               Float(A_Task_Result.
               Period) <=
               1.0 THEN
            A_Task_Result.Cpu_Name :=
               A_Processor.Name;
         ELSE
            Duplicate(Src_Tasks,
               Result_Tasks);
            RAISE No_Such_Processors;
         END IF;
         EXIT WHEN Is_Last_Element(
            Result_Tasks,
            My_Task_Iterator_Result);
         LOOP
            EXIT WHEN Is_Last_Element(
               Result_Tasks,
               My_Task_Iterator_Result);
            Total_Utilization :=
               Total_Utilization + Float(
               A_Task_Result.Capacity) /
               Float(A_Task_Result.
               Period);
            Next_Element(Result_Tasks,
               My_Task_Iterator_Result);
            Current_Element(Result_Tasks,
               Task_Ptr(A_Task_Result),
               My_Task_Iterator_Result);

            IF Float(
                  A_Task_Result.Capacity) /
                  Float(
                  A_Task_Result.Period) <=
                  2.0 * (1.0 / (1.0 +
                     Total_Utilization /
                     Float((M-
                           1))) ** (
                     M-1)) - 1.0 AND
                  Total_Utilization <=
                  Float(M-1) * (2.0 ** (1.0 /
                     Float(M-1)) - 1.0) THEN
               Condition_Ip := True;
            ELSE
               Condition_Ip := False;
            END IF;

            IF Condition_Ip THEN
               A_Task_Result.Cpu_Name :=
                  A_Processor.Name;
               M := M + 1;
            ELSE
               IF Is_Last_Element(
                     Src_Processors,
                     My_Processor_Iterator) THEN
                  Duplicate(Src_Tasks,
                     Result_Tasks);
                  RAISE
                     No_Such_Processors;
               END IF;
               Next_Element(
                  Src_Processors,
                  My_Processor_Iterator);
               Current_Element(
                  Src_Processors,
                  A_Processor,
                  My_Processor_Iterator);
               Total_Utilization := 0.0;
               M := 2;
               EXIT;
            END IF;
         END LOOP;

      END LOOP;

      Sort(Result_Tasks, Increasing_Name'ACCESS);

   END Partition_Next_Fit;


   PROCEDURE Partition_First_Fit (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set;
         Msg            : IN OUT Unbounded_String;
         Result_Tasks   : IN OUT Tasks_Set) IS

      A_Task_Result           : Periodic_Task_Ptr;
      My_Task_Iterator_Result : Tasks_Iterator;
      Temp_Task_Iterator      : Tasks_Iterator;
      A_Processor             : Processor_Ptr;
      My_Processor_Iterator   : Processors_Iterator;

      Num_Of_Tasks      : Integer := 0;
      Task_Utilization,
      Total_Utilization : Float   := 0.0;
   BEGIN
      Validate_Multiprocessor_Tasks(
         Src_Processors, Src_Tasks);


      Msg:= To_Unbounded_String(" (") &
         Lb_See(Current_Language) &
         To_Unbounded_String("[9], [10]) ");


      Duplicate(Src_Tasks, Result_Tasks);
      Sort(Result_Tasks,
         Increasing_Period'ACCESS);

      Reset_Iterator(Result_Tasks,
         My_Task_Iterator_Result);
      Temp_Task_Iterator :=
         My_Task_Iterator_Result;
      Current_Element(Result_Tasks,
         Task_Ptr(A_Task_Result),
         My_Task_Iterator_Result);
      Reset_Iterator(Src_Processors,
         My_Processor_Iterator);
      Current_Element(Src_Processors,
         A_Processor,
         My_Processor_Iterator);

      -- Zero processor locations
      --
      LOOP
         A_Task_Result.Cpu_Name :=
            To_Unbounded_String("");
         EXIT WHEN Is_Last_Element(
            Result_Tasks,
            My_Task_Iterator_Result);
         Next_Element(Result_Tasks,
            My_Task_Iterator_Result);
         Current_Element(Result_Tasks,
            Task_Ptr(A_Task_Result),
            My_Task_Iterator_Result);
      END LOOP;

      Reset_Iterator(Result_Tasks,
         My_Task_Iterator_Result);
      Current_Element(Result_Tasks,
         Task_Ptr(A_Task_Result),
         My_Task_Iterator_Result);
      LOOP
         LOOP
            -- Fetch information for condition-IP
            --
            Reset_Iterator(Result_Tasks,
               My_Task_Iterator_Result);
            Current_Element(Result_Tasks,
               Task_Ptr(
                  A_Task_Result),
               My_Task_Iterator_Result);
            Num_Of_Tasks := 0;
            Total_Utilization := 0.0;
            LOOP
               IF A_Task_Result.Cpu_Name =
                     A_Processor.Name THEN
                  Num_Of_Tasks :=
                     Num_Of_Tasks + 1;
                  Total_Utilization :=
                     Total_Utilization +
                     Float(
                     A_Task_Result.Capacity) /
                     Float(
                     A_Task_Result.Period);
               END IF;
               EXIT WHEN Is_Last_Element(
                  Result_Tasks,
                  My_Task_Iterator_Result);
               Next_Element(Result_Tasks,
                  My_Task_Iterator_Result);
               Current_Element(
                  Result_Tasks, Task_Ptr(
                     A_Task_Result),
                  My_Task_Iterator_Result);
            END LOOP;
            My_Task_Iterator_Result :=
               Temp_Task_Iterator;
            Current_Element(Result_Tasks,
               Task_Ptr(
                  A_Task_Result),
               My_Task_Iterator_Result);

            Task_Utilization := Float(
               A_Task_Result.Capacity) /
               Float(
               A_Task_Result.Period);
            IF (Task_Utilization <= 2.0 * (
                     1.0 / (1.0 +
                        Total_Utilization /
                        Float(
                           Num_Of_Tasks)) **
                     Num_Of_Tasks) -
                  1.0) THEN
               A_Task_Result.Cpu_Name :=
                  A_Processor.Name;
               Reset_Iterator(
                  Src_Processors,
                  My_Processor_Iterator);
               Current_Element(
                  Src_Processors,
                  A_Processor,
                  My_Processor_Iterator);
               EXIT;
            ELSE
               IF Is_Last_Element(
                     Src_Processors,
                     My_Processor_Iterator) THEN
                  Duplicate(Src_Tasks,
                     Result_Tasks);
                  RAISE
                     No_Such_Processors;
               END IF;
               Next_Element(
                  Src_Processors,
                  My_Processor_Iterator);
               Current_Element(
                  Src_Processors,
                  A_Processor,
                  My_Processor_Iterator);
            END IF;
         END LOOP;

         EXIT WHEN Is_Last_Element(
            Result_Tasks,
            My_Task_Iterator_Result);
         Next_Element(Result_Tasks,
            My_Task_Iterator_Result);
         Temp_Task_Iterator :=
            My_Task_Iterator_Result;
      END LOOP;

      Sort(Result_Tasks, Increasing_Name'ACCESS);

   END Partition_First_Fit;


   PROCEDURE Partition_Best_Fit (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set;
         Msg            : IN OUT Unbounded_String;
         Result_Tasks   : IN OUT Tasks_Set) IS

      A_Task_Result           : Periodic_Task_Ptr;
      My_Task_Iterator_Result : Tasks_Iterator;
      A_Processor             : Processor_Ptr;
      My_Processor_Iterator   : Processors_Iterator;
      Temp_Task_Iterator      : Tasks_Iterator;
      Best_Cpu                : Processors_Iterator;

      M                  : Integer := 1;
      Best_Fit_Value     : Float   := 2.0;
      Best_Fit_Tmp_Value,
      Tot_Util           : Float;
      First_Task         : Boolean := True;
      Assignable         : Boolean;

   BEGIN

      Validate_Multiprocessor_Tasks(
         Src_Processors, Src_Tasks);

      Msg:= To_Unbounded_String(" (") &
         Lb_See(Current_Language) &
         To_Unbounded_String("[9], [10]) ");

      Duplicate(Src_Tasks, Result_Tasks);
      Sort(Result_Tasks,
         Increasing_Period'ACCESS);

      Reset_Iterator(Result_Tasks,
         My_Task_Iterator_Result);
      Current_Element(Result_Tasks,
         Task_Ptr(A_Task_Result),
         My_Task_Iterator_Result);

      -- Zero processor locations
      --
      LOOP
         A_Task_Result.Cpu_Name :=
            To_Unbounded_String("");
         EXIT WHEN Is_Last_Element(
            Result_Tasks,
            My_Task_Iterator_Result);
         Next_Element(Result_Tasks,
            My_Task_Iterator_Result);
         Current_Element(Result_Tasks,
            Task_Ptr(
               A_Task_Result),
            My_Task_Iterator_Result);
      END LOOP;
      Reset_Iterator(Result_Tasks,
         My_Task_Iterator_Result);
      Current_Element(Result_Tasks,
         Task_Ptr(A_Task_Result),
         My_Task_Iterator_Result);


      LOOP
         Reset_Iterator(Src_Processors,
            My_Processor_Iterator);
         Current_Element(Src_Processors,
            A_Processor,
            My_Processor_Iterator);

         -- Loop scans best processor for task to be placed
         --
         LOOP
            Temp_Task_Iterator :=
               My_Task_Iterator_Result;

            -- get CPU data
            --
            Reset_Iterator(Result_Tasks,
               My_Task_Iterator_Result);
            Current_Element(Result_Tasks,
               Task_Ptr(
                  A_Task_Result),
               My_Task_Iterator_Result);
            M := 1;
            Tot_Util := 0.0;
            LOOP
               IF A_Task_Result.Cpu_Name =
                     A_Processor.
                     Name THEN
                  M := M + 1;
                  Tot_Util := Tot_Util +
                     Float(
                     A_Task_Result.Capacity) /
                     Float(
                     A_Task_Result.Period);
               END IF;
               EXIT WHEN Is_Last_Element(
                  Result_Tasks,
                  My_Task_Iterator_Result);
               Next_Element(Result_Tasks,
                  My_Task_Iterator_Result);
               Current_Element(
                  Result_Tasks, Task_Ptr(
                     A_Task_Result),
                  My_Task_Iterator_Result);
            END LOOP;
            My_Task_Iterator_Result :=
               Temp_Task_Iterator;
            Current_Element(Result_Tasks,
               Task_Ptr(
                  A_Task_Result),
               My_Task_Iterator_Result);

            -- test for condition IP
            --
            IF First_Task THEN
               IF Float(
                     A_Task_Result.Capacity) /
                     Float(
                     A_Task_Result.Period) <=
                     1.0 THEN
                  Reset_Iterator(
                     Src_Processors,
                     Best_Cpu);
                  My_Processor_Iterator :=
                     Best_Cpu;
                  First_Task := False;
                  Assignable := True;
                  EXIT;
               ELSE
                  Assignable := False;
                  EXIT;
               END IF;
            ELSE
               IF Float(
                     A_Task_Result.Capacity) /
                     Float(
                     A_Task_Result.Period) <=
                     2.0 * (1.0 / (1.0 +
                        Tot_Util / Float((
                              M-
                              1))) ** (
                        M-1)) - 1.0 THEN
                  -- Tricky Condition IP testing because mathematics function
                  -- doesn't seem to understand handle
                  -- zero number very well in following check (M := 0
                  -- happens when processor is empty)
                  --
                  IF M > 1 THEN
                     IF Tot_Util <= Float(
                           M-1) * (2.0 **
                           Float(1.0 /
                              Float(
                                 M-1)) -
                           1.0) THEN
                        Best_Fit_Tmp_Value :=
                           2.0 * (1.0/((
                                 1.0 +
                                 Tot_Util/
                                 Float(M-
                                    1))**(
                                 M-1))) -
                           1.0;
                        IF
                              Best_Fit_Tmp_Value <
                              Best_Fit_Value THEN
                           Best_Fit_Value :=
                              Best_Fit_Tmp_Value;
                           Best_Cpu :=
                              My_Processor_Iterator;
                           Assignable :=
                              True;
                        END IF;
                     END IF;
                  ELSE
                     Best_Fit_Tmp_Value :=
                        2.0 * (1.0/(1.0 +
                           Tot_Util/
                           Float(M-1)**(
                              M-1))) -
                        1.0;
                     IF
                           Best_Fit_Tmp_Value <
                           Best_Fit_Value THEN
                        Best_Fit_Value :=
                           Best_Fit_Tmp_Value;
                        Best_Cpu :=
                           My_Processor_Iterator;
                        Assignable :=
                           True;
                     END IF;
                  END IF;
                  -- End of tricky testing
                  --
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(
               Src_Processors,
               My_Processor_Iterator);
            Next_Element(Src_Processors,
               My_Processor_Iterator);
            Current_Element(
               Src_Processors,
               A_Processor,
               My_Processor_Iterator);
         END LOOP;

         IF Assignable THEN
            -- save task to best CPU
            --
            Current_Element(
               Src_Processors,
               A_Processor,
               Best_Cpu);
            A_Task_Result.Cpu_Name :=
               A_Processor.Name;
            Assignable := False;
            Best_Fit_Value := 2.0;
         ELSE
            Duplicate(Src_Tasks,
               Result_Tasks);
            RAISE No_Such_Processors;
         END IF;

         EXIT WHEN Is_Last_Element(
            Result_Tasks,
            My_Task_Iterator_Result);
         Next_Element(Result_Tasks,
            My_Task_Iterator_Result);
         Current_Element(Result_Tasks,
            Task_Ptr(
               A_Task_Result),
            My_Task_Iterator_Result);
      END LOOP;

      Sort(Result_Tasks, Increasing_Name'ACCESS);
   END Partition_Best_Fit;


   PROCEDURE Validate_Multiprocessor_Tasks (
         Src_Processors : IN     Processors_Set;
         Src_Tasks      : IN     Tasks_Set) IS

      A_Processor           : Processor_Ptr;
      My_Processor_Iterator : Processors_Iterator;


   BEGIN
      -- check if tasks are periodic
      --
      Reset_Iterator(Src_Processors,
         My_Processor_Iterator);
      LOOP
         Current_Element(Src_Processors,
            A_Processor,
            My_Processor_Iterator);
         Periodic_Control(Src_Tasks,
            A_Processor.Name);
         EXIT WHEN True = Is_Last_Element(
            Src_Processors,
            My_Processor_Iterator);
         Next_Element(Src_Processors,
            My_Processor_Iterator);
      END LOOP;

      -- check if processors are rate-monotonic and non pre-emptive
      --
      Reset_Iterator(Src_Processors,
         My_Processor_Iterator);
      LOOP
         Current_Element(Src_Processors,
            A_Processor,
            My_Processor_Iterator);
         IF Get_Name(
               A_Processor.Scheduler.All) /=
               Rate_Monotonic_Protocol THEN
            RAISE
               Invalid_Scheduler_Multiprocessors;
         END IF;
         IF Get_Preemptive(
               A_Processor.Scheduler.All) =
               Not_Preemptive THEN
            RAISE
               Invalid_Scheduler_Multiprocessors;
         END IF;
         EXIT WHEN Is_Last_Element(
            Src_Processors,
            My_Processor_Iterator);
         Next_Element(Src_Processors,
            My_Processor_Iterator);
      END LOOP;

   END Validate_Multiprocessor_Tasks;



END Multiprocessor_Services;



