------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Unbounded_Strings;
USE Unbounded_Strings;



PACKAGE BODY Network_Set IS




   PROCEDURE Check_Network (
         My_Networks  : IN     Networks_Set;
         Name         : IN     Unbounded_String;
         Network_Type : IN     Unbounded_String) IS

      My_Iterator : Iterator;
      A_Network   : Network_Ptr;

   BEGIN

      IF (Name = "") THEN
         RAISE Invalid_Parameter;
      END IF;

      IF(Get_Number_Of_Elements(My_Networks) > 0) THEN
         Reset_Iterator(My_Networks, My_Iterator);

         LOOP
            Current_Element(My_Networks, A_Network, My_Iterator);

            IF (Name = A_Network.Name) THEN
               RAISE Network_Already_Defined;
            END IF;

            EXIT WHEN Is_Last_Element(My_Networks, My_Iterator);

            Next_Element(My_Networks, My_Iterator);

         END LOOP;
      END IF;


      IF ( Network_Type /= "BOUNDED_DELAY" )  AND
            ( Network_Type /= "PARAMETRIC_DELAY" ) THEN
         RAISE Invalid_Parameter;
      END IF;

   END Check_Network;



   PROCEDURE Add_Network (
         My_Networks  : IN OUT Networks_Set;
         Name         : IN     Unbounded_String;
         Network_Type : IN     Unbounded_String) IS

      A_Network : Network_Ptr;

   BEGIN

      Check_Network(My_Networks, Name, Network_Type);


      A_Network := NEW Network;

      IF ( Network_Type = "BOUNDED_DELAY" )THEN
         A_Network.Network_Type := Bounded_Delay;
      ELSE
         A_Network.Network_Type := Parametric_Delay;
      END IF;

      A_Network.Name := Name;

      Add(My_Networks,A_Network);

   END Add_Network;


   PROCEDURE Update_Network (
         My_Networks  : IN OUT Networks_Set;
         Name         : IN     Unbounded_String;
         Network_Type : IN     Unbounded_String) IS

      A_Network : Network_Ptr;

   BEGIN

      A_Network := Search_Network(My_Networks,Name);
      Delete(My_Networks,A_Network);
      Add_Network(My_Networks,Name,Network_Type);

   END Update_Network;

   FUNCTION Search_Network (
         My_Networks : IN     Networks_Set;
         Name        : IN     Unbounded_String)
     RETURN Network_Ptr IS

      My_Iterator : Iterator;
      A_Network   : Network_Ptr;
      Result      : Network_Ptr;

      Found : Boolean := False;

   BEGIN

      Reset_Iterator(My_Networks, My_Iterator);

      LOOP
         Current_Element(My_Networks, A_Network, My_Iterator);

         IF(A_Network.Name = Name) THEN
            Found:=True;
            Result:=A_Network;
         END IF;

         EXIT WHEN Is_Last_Element(My_Networks, My_Iterator);

         Next_Element(My_Networks, My_Iterator);

      END LOOP;

      IF NOT Found THEN
         RAISE Network_Not_Found;
      END IF;

      RETURN Result;

   END Search_Network;




   FUNCTION Export_Xml (
         My_Networks : IN     Networks_Set)
     RETURN Unbounded_String IS

      My_Iterator : Network_Iterator;
      A_Network   : Network_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN

      IF NOT Is_Empty(My_Networks) THEN

         Result:= Result & To_Unbounded_String( "<networks>") &
            Unbounded_Lf;

         Reset_Iterator(My_Networks, My_Iterator);

         LOOP
            Current_Element(My_Networks, A_Network, My_Iterator);

            Result:= Result & To_Unbounded_String(Ascii.Ht & "<network>") &
               Unbounded_Lf;

            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<name>" & To_String(
                  A_Network.Name) & "</name>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht & Ascii.Ht &
               "<type>" & A_Network.
               Network_Type'Img & "</type>") & Unbounded_Lf;

            Result:= Result & To_Unbounded_String(Ascii.Ht & "</network>") &
               Unbounded_Lf;

            EXIT WHEN Is_Last_Element(My_Networks, My_Iterator);

            Next_Element(My_Networks, My_Iterator);

         END LOOP;
         Result:= Result & To_Unbounded_String( "</networks>") &
            Unbounded_Lf;
      END IF;


      RETURN Result;

   END Export_Xml;


   FUNCTION Export_Aadl_Implementations (
         My_Networks : IN     Networks_Set)
     RETURN Unbounded_String IS

      My_Iterator : Iterator;
      A_Network   : Network_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN

      IF NOT Is_Empty(My_Networks) THEN

         Reset_Iterator(My_Networks, My_Iterator);

         LOOP
            Current_Element(My_Networks, A_Network, My_Iterator);
            EXIT WHEN Is_Last_Element(My_Networks, My_Iterator);

            Result:= Result & To_Unbounded_String( "bus " & To_String(
                  A_Network.Name));
            Result:= Result & To_Unbounded_String( "end " & To_String(
                  A_Network.Name) & ";" );

            Result:= Result & To_Unbounded_String( "bus implementation " &
               To_String(A_Network.
                  Name) & ".Impl" );
            Result:= Result & To_Unbounded_String( "end " & To_String(
                  A_Network.Name) & ".Impl" );


            Next_Element(My_Networks, My_Iterator);

         END LOOP;
      END IF;


      RETURN Result;

   END Export_Aadl_Implementations;






   FUNCTION Export_Aadl_Declarations (
         My_Networks  : IN     Networks_Set;
         Number_Of_Ht : IN     Natural)
     RETURN Unbounded_String IS


      My_Iterator : Iterator;
      A_Network   : Network_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN

      IF NOT Is_Empty(My_Networks) THEN

         Reset_Iterator(My_Networks, My_Iterator);

         LOOP
            Current_Element(My_Networks, A_Network, My_Iterator);
            EXIT WHEN Is_Last_Element(My_Networks, My_Iterator);


            FOR I IN 1..Number_Of_Ht LOOP
               Result:= Result & Ascii.Ht;
            END LOOP;
            Result:= Result & To_Unbounded_String(  "instancied_" &
               To_String(A_Network.Name) &
               " : bus " & To_String(A_Network.Name) & ".Impl;");

            Next_Element(My_Networks, My_Iterator);

         END LOOP;
      END IF;


      RETURN Result;

   END Export_Aadl_Declarations;


END Network_Set;







