------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Text_Io;
USE Text_Io;

WITH Ada.Strings.Unbounded.Text_Io;
USE Ada.Strings.Unbounded.Text_Io;


WITH Ada.Exceptions;
USE Ada.Exceptions;


WITH Translate;
USE Translate;

WITH Time_Units;
USE Time_Units;

WITH Qs_Tools;
USE Qs_Tools;

WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;
WITH Ada.Numerics.Float_Random;
USE Ada.Numerics.Float_Random;

WITH Messages;
USE Messages;
WITH Tasks;
USE Tasks;
WITH Buffers;
USE Buffers;
WITH Resources;
USE Resources;


WITH Parameters;
USE Parameters;
USE Parameters.User_Defined_Parameters_Table_Package;


WITH Integer_Util;



PACKAGE BODY Expressions IS





   PROCEDURE Put (
         V : IN     Variable_Record_Ptr) IS
   BEGIN
      Put(V.Variable);
      Put(V.Simulation);
   END Put;





   FUNCTION Find_Variable (
         Var_Table : IN     Variables_Table_Type;
         S         : IN     Unbounded_String;
         Line      : IN     Natural              := 0)
     RETURN Variables_Range IS

   BEGIN
      FOR I IN 0..Var_Table.Nb_Entries-1 LOOP
         IF Var_Table.Entries(I).Variable.Name=S
               THEN
            RETURN I;
         END IF;
      END LOOP;

      IF (Line=0)
            THEN
         Raise_Exception(Expressions.Syntax_Error'Identity, To_String(
               S & Lb_Comma & Lb_Undeclared_Identifier(Current_Language) ));
      ELSE
         Raise_Exception(Expressions.Syntax_Error'Identity, " Line " &
            Line'Img & To_String(Lb_Comma &
               S & Lb_Comma & Lb_Undeclared_Identifier(Current_Language) ));
      END IF;

   END Find_Variable;


   FUNCTION Find_Variable (
         Var_Table : IN     Variables_Table_Type;
         S         : IN     String;
         Line      : IN     Natural              := 0)
     RETURN Variables_Range IS
   BEGIN
      RETURN Find_Variable(Var_Table, To_Unbounded_String(S),Line);
   END Find_Variable;


   PROCEDURE Check_Variable_Declaration (
         Var_Table : IN     Variables_Table_Type;
         S         : IN     Unbounded_String;
         Line      : IN     Natural              := 0) IS

      Dummy : Variables_Range;

   BEGIN
      Dummy:=Find_Variable(Var_Table,S,Line);
   END Check_Variable_Declaration;



   PROCEDURE Initialize_Parametric_Variables (
         Var_Table      : IN OUT Variables_Table_Type;
         My_Messages    : IN     Messages_Set;
         My_Buffers     : IN     Buffers_Set;
         My_Resources   : IN     Resources_Set;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String     := Empty_String) IS


      Var   : Variables_Range;
      Index : Simulations_Range;

      A_Message          : Message_Ptr;
      A_Message_Iterator : Messages_Iterator;

      A_Buffer          : Buffer_Ptr;
      A_Buffer_Iterator : Buffers_Iterator;

      A_Task          : Task_Ptr;
      A_Task_Iterator : Tasks_Iterator;

      A_Resource          : Resource_Ptr;
      A_Resource_Iterator : Resources_Iterator;

   BEGIN

      ---------------------------------------------------------------
      --   Set message variables
      ---------------------------------------------------------------

      -- "nb_messages"
      --
      Var:=Find_Variable(Var_Table,
         To_Unbounded_String("nb_messages"));
      Var_Table.Entries(Var).Simulation.
         Integer_Value:=Integer(Get_Number_Of_Elements(My_Messages));


      IF NOT Is_Empty(My_Messages)
            THEN
         Reset_Iterator(My_Messages, A_Message_Iterator);
         Index:=0;

         LOOP
            Current_Element(My_Messages, A_Message, A_Message_Iterator);


            -- "messages.name"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("messages.name"));
            Var_Table.Entries(Var).Simulation.
               String_Table_Value(Index):=
               A_Message.Name;


            -- "messages.deadline"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("messages.deadline"));
            Var_Table.Entries(Var).Simulation.
               Integer_Table_Value(Index):=
               Integer(A_Message.Deadline);


            -- "messages.delay"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("messages.delay"));
            Var_Table.Entries(Var).Simulation.
               Integer_Table_Value(Index):=
               Integer(A_Message.Response_Time);


            -- "messages.size"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("messages.size"));
            Var_Table.Entries(Var).Simulation.
               Integer_Table_Value(Index):=
               Integer(A_Message.Size);


            -- "messages.period" and "messages.jitter"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("messages.period"));
            CASE A_Message.Message_Type IS
               WHEN Aperiodic_Type =>
                  Var_Table.Entries(Var).Simulation.
                     Integer_Table_Value(Index):=0;
               WHEN OTHERS =>
                  Var_Table.Entries(Var).Simulation.
                     Integer_Table_Value(Index):=
                     Integer(Periodic_Message_Ptr(A_Message).Period);
            END CASE;
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("messages.jitter"));
            CASE A_Message.Message_Type IS
               WHEN Aperiodic_Type =>
                  Var_Table.Entries(Var).Simulation.
                     Integer_Table_Value(Index):=0;
               WHEN OTHERS =>
                  Var_Table.Entries(Var).Simulation.
                     Integer_Table_Value(Index):=
                     Integer(Periodic_Message_Ptr(A_Message).Jitter);
            END CASE;


            -- "messages.ready"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("messages.ready"));
            Var_Table.Entries(Var).Simulation.
               Boolean_Table_Value(Index):=False;


            EXIT WHEN Is_Last_Element(My_Messages, A_Message_Iterator);
            Next_Element(My_Messages, A_Message_Iterator);
            Index:=Index+1;

         END LOOP;
      END IF;


      ---------------------------------------------------------------
      --   Set buffer variables
      ---------------------------------------------------------------


      -- "nb_buffers"
      --
      Var:=Find_Variable(Var_Table,
         To_Unbounded_String("nb_buffers"));
      IF Processor_Name = Empty_String
            THEN
         Var_Table.Entries(Var).Simulation.
            Integer_Value:=Integer(Get_Number_Of_Elements(My_Buffers));
      ELSE
         Var_Table.Entries(Var).Simulation.
            Integer_Value:=Integer(Get_Number_Of_Buffer_From_Processor(
               My_Buffers, Processor_Name));
      END IF;


      IF NOT Is_Empty(My_Buffers)
            THEN
         Reset_Iterator(My_Buffers, A_Buffer_Iterator);
         Index:=0;

         LOOP
            Current_Element(My_Buffers, A_Buffer, A_Buffer_Iterator);


            -- "buffers.processor_name"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("buffers.processor_name"));
            Var_Table.Entries(Var).Simulation.
               String_Table_Value(Index):=
               A_Buffer.Cpu_Name;



            -- "buffers.name"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("buffers.name"));
            Var_Table.Entries(Var).Simulation.
               String_Table_Value(Index):=
               A_Buffer.Name;



            -- "buffers.max_size"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("buffers.max_size"));
            Var_Table.Entries(Var).Simulation.
               Integer_Table_Value(Index):=
               Integer(A_Buffer.Size);



            EXIT WHEN Is_Last_Element(My_Buffers, A_Buffer_Iterator);
            Next_Element(My_Buffers, A_Buffer_Iterator);
            Index:=Index+1;

         END LOOP;
      END IF;


      ---------------------------------------------------------------
      --   Set resource variables
      ---------------------------------------------------------------


      -- "nb_resources"
      --
      Var:=Find_Variable(Var_Table,
         To_Unbounded_String("nb_resources"));
      IF Processor_Name = Empty_String
            THEN
         Var_Table.Entries(Var).Simulation.
            Integer_Value:=Integer(Get_Number_Of_Elements(My_Resources));
      ELSE
         Var_Table.Entries(Var).Simulation.
            Integer_Value:=Integer(Get_Number_Of_Resource_From_Processor(
               My_Resources, Processor_Name));
      END IF;


      IF NOT Is_Empty(My_Resources)
            THEN
         Reset_Iterator(My_Resources, A_Resource_Iterator);
         Index:=0;

         LOOP
            Current_Element(My_Resources, A_Resource, A_Resource_Iterator);


            -- "resources.name"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("resources.name"));
            Var_Table.Entries(Var).Simulation.
               String_Table_Value(Index):=
               A_Resource.Name;


            -- "resources.initial_state"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("resources.initial_state"));
            Var_Table.Entries(Var).Simulation.
               Integer_Table_Value(Index):=
               A_Resource.State;


            -- "resources.protocol"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("resources.protocol"));
            Var_Table.Entries(Var).Simulation.
               String_Table_Value(Index):=
               To_Unbounded_String(A_Resource.Protocol'Img);


            -- "resources.processor_name"
            --
            Var:=Find_Variable(Var_Table,
               To_Unbounded_String("resources.processor_name"));
            Var_Table.Entries(Var).Simulation.
               String_Table_Value(Index):=
               A_Resource.Cpu_Name;



            EXIT WHEN Is_Last_Element(My_Resources, A_Resource_Iterator);
            Next_Element(My_Resources, A_Resource_Iterator);
            Index:=Index+1;

         END LOOP;
      END IF;



      ---------------------------------------------------------------
      --   Set task variables
      ---------------------------------------------------------------


      -- "nb_tasks"
      --
      Var:=Find_Variable(Var_Table,
         To_Unbounded_String("nb_tasks"));
      IF Processor_Name = Empty_String
            THEN
         Var_Table.Entries(Var).Simulation.
            Integer_Value:=Integer(Get_Number_Of_Elements(My_Tasks));
      ELSE
         Var_Table.Entries(Var).Simulation.
            Integer_Value:=Integer(Get_Number_Of_Task_From_Processor(
               My_Tasks, Processor_Name));
      END IF;


      Reset_Iterator(My_Tasks, A_Task_Iterator);
      Index:=0;

      LOOP
         Current_Element(My_Tasks, A_Task, A_Task_Iterator);


         -- "tasks.priority"
         --
         Var:=Find_Variable(Var_Table,
            To_Unbounded_String("tasks.priority"));
         Var_Table.Entries(Var).Simulation.
            Integer_Table_Value(Index):=
            Integer(A_Task.Priority);


         -- "tasks.deadline"
         --
         Var:=Find_Variable(Var_Table,
            To_Unbounded_String("tasks.deadline"));
         Var_Table.Entries(Var).Simulation.
            Integer_Table_Value(Index):=
            Integer(A_Task.Deadline);


         -- "tasks.start_time"
         --
         Var:=Find_Variable(Var_Table,
            To_Unbounded_String("tasks.start_time"));
         Var_Table.Entries(Var).Simulation.
            Integer_Table_Value(Index):=
            Integer(A_Task.Start_Time);


         -- "tasks.processor_name"
         --
         Var:=Find_Variable(Var_Table,
            To_Unbounded_String("tasks.processor_name"));
         Var_Table.Entries(Var).Simulation.
            String_Table_Value(Index):=
            A_Task.Cpu_Name;


         -- "tasks.name"
         --
         Var:=Find_Variable(Var_Table,
            To_Unbounded_String("tasks.name"));
         Var_Table.Entries(Var).Simulation.
            String_Table_Value(Index):=
            A_Task.Name;


         -- "tasks.capacity"
         --
         Var:=Find_Variable(Var_Table,
            To_Unbounded_String("tasks.capacity"));
         Var_Table.Entries(Var).Simulation.
            Integer_Table_Value(Index):=
            Integer(A_Task.Capacity);


         -- "tasks.blocking_time"
         --
         Var:=Find_Variable(Var_Table,
            To_Unbounded_String("tasks.blocking_time"));
         Var_Table.Entries(Var).Simulation.
            Integer_Table_Value(Index):=
            Integer(A_Task.Blocking_Time);


         -- "tasks.period"
         --
         Var:=Find_Variable(Var_Table,
            To_Unbounded_String("tasks.period"));
         CASE A_Task.Task_Type IS
            WHEN Aperiodic_Type =>
               Var_Table.Entries(Var).Simulation.
                  Integer_Table_Value(Index):=0;
            WHEN OTHERS =>
               Var_Table.Entries(Var).Simulation.
                  Integer_Table_Value(Index):=
                  Integer(Periodic_Task_Ptr(A_Task).Period);
         END CASE;


         -- "tasks.jitter"
         --
         Var:=Find_Variable(Var_Table,
            To_Unbounded_String("tasks.jitter"));
         CASE A_Task.Task_Type IS
            WHEN Aperiodic_Type =>
               Var_Table.Entries(Var).Simulation.
                  Integer_Table_Value(Index):=0;
            WHEN OTHERS =>
               Var_Table.Entries(Var).Simulation.
                  Integer_Table_Value(Index):=
                  Integer(Periodic_Task_Ptr(A_Task).Jitter);
         END CASE;



         -- user defined task's parameter :
         --

         FOR I IN 0..A_Task.Param.Nb_Entries-1 LOOP

            -- Find the user defined variable
            -- and set its value
            --
            FOR J IN 0..Var_Table.Nb_Entries-1 LOOP
               IF Var_Table.Entries(J).Variable.Name=
                     To_Unbounded_String("tasks.") & A_Task.Param.Entries(
                     I).Name
                     THEN

                  -- Now, set the value
                  --
                  IF A_Task.Param.Entries(I).Ptype=
                        Boolean_Parameter THEN
                     Var_Table.Entries(J).
                        Simulation.
                        Boolean_Table_Value(Index):=
                        Boolean(A_Task.
                        Param.Entries(I).Boolean_Value);
                  END IF;
                  IF A_Task.Param.Entries(I).Ptype=
                        Integer_Parameter THEN
                     Var_Table.Entries(J).
                        Simulation.
                        Integer_Table_Value(Index):=
                        Integer(A_Task.
                        Param.Entries(I).Integer_Value);
                  END IF;
                  IF A_Task.Param.Entries(I).Ptype=
                        Double_Parameter THEN
                     Var_Table.Entries(J).
                        Simulation.
                        Double_Table_Value(Index):=Double(
                        A_Task.
                        Param.Entries(I).Double_Value);
                  END IF;
                  IF A_Task.Param.Entries(I).Ptype=
                        String_Parameter THEN
                     Var_Table.Entries(J).
                        Simulation.
                        String_Table_Value(Index):=
                        A_Task.
                        Param.Entries(I).String_Value;
                  END IF;
               END IF;
            END LOOP;
         END LOOP;


         EXIT WHEN Is_Last_Element(My_Tasks, A_Task_Iterator);
         Next_Element(My_Tasks, A_Task_Iterator);
         Index:=Index+1;

      END LOOP;



   END Initialize_Parametric_Variables;




   PROCEDURE Create_Parametric_Variables (
         Variables_Table : IN OUT Variables_Table_Type;
         My_Tasks        : IN     Tasks_Set) IS


      New_Var           : Variable_Record_Ptr;
      New_Expr_Array    : Array_Variable_Expr_Ptr;
      New_Expr_Variable : Variable_Expr_Ptr;
      Do_Insert         : Boolean;
      A_Task            : Task_Ptr;
      Iterator          : Tasks_Iterator;

   BEGIN


      Initialize(Variables_Table);


      ------------------------------------------------------
      --              common variables
      ------------------------------------------------------


      -- "previously_elected"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("previously_elected");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "nb_tasks"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("nb_tasks");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "nb_processors"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("nb_processors");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "nb_messages"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("nb_messages");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "nb_buffers"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("nb_buffers");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "nb_resources"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("nb_resources");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);



      -- "nb_address_spaces"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("nb_address_spaces");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "nb_time_units"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("nb_time_units");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "simulation_time"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("simulation_time");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);



      -- "simulation_length"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("simulation_length");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      ------------------------------------------------------
      --              Buffer variables
      ------------------------------------------------------


      -- "buffers.name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_String;
      New_Expr_Variable.Name:=To_Unbounded_String("buffers.name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "buffers.max_size"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("buffers.max_size");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "buffers.processor_name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_String;
      New_Expr_Variable.Name:=To_Unbounded_String(
         "buffers.processor_name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "buffers.users.time"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("buffers.users.time");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "buffers.users.size"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("buffers.users.size");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "buffers.users.task_name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_String;
      New_Expr_Variable.Name:=To_Unbounded_String(
         "buffers.users.task_name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      ------------------------------------------------------
      --              Message variables
      ------------------------------------------------------


      -- "messages.jitter"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("messages.jitter");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);



      -- "messages.delay"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("messages.delay");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "messages.name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_String;
      New_Expr_Variable.Name:=To_Unbounded_String("messages.name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "messages.deadline"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("messages.deadline");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "messages.period"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("messages.period");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "messages.size"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("messages.size");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "messages.users.time"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("messages.users.time");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "messages.users.type"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_String;
      New_Expr_Variable.Name:=To_Unbounded_String("messages.users.type");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "messages.users.task_name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_String;
      New_Expr_Variable.Name:=To_Unbounded_String(
         "messages.users.task_name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);



      ------------------------------------------------------
      --              Event variables
      ------------------------------------------------------


      -- "events.time"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String("events.time");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "events.processor_name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_String;
      New_Expr_Variable.Name:=To_Unbounded_String("events.processor_name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "events.type"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_String;
      New_Expr_Variable.Name:=To_Unbounded_String("events.type");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "events.task_name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_String;
      New_Expr_Variable.Name:=To_Unbounded_String("events.task_name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "events.buffer_name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_String;
      New_Expr_Variable.Name:=To_Unbounded_String("events.buffer_name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "events.resource_name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_String;
      New_Expr_Variable.Name:=To_Unbounded_String("events.resource_name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "events.message_name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_String;
      New_Expr_Variable.Name:=To_Unbounded_String("events.message_name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      ------------------------------------------------------
      --              Resource variables
      ------------------------------------------------------



      -- "resources.initial_state"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String(
         "resources.initial_state");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "resources.processor_name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_String;
      New_Expr_Variable.Name:=To_Unbounded_String(
         "resources.processor_name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "resources.protocol"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_String;
      New_Expr_Variable.Name:=To_Unbounded_String("resources.protocol");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "resources.name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_String;
      New_Expr_Variable.Name:=To_Unbounded_String("resources.name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "resources.current_state"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String(
         "resources.current_state");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "resources.users.task_name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_String;
      New_Expr_Variable.Name:=To_Unbounded_String(
         "resources.users.task_name");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "resources.users.start_time"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String(
         "resources.users.start_time");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);


      -- "resources.users.end_time"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Variable := NEW Array_Variable_Expr;
      New_Expr_Variable.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Variable.Name:=To_Unbounded_String(
         "resources.users.end_time");
      New_Var.Variable:=New_Expr_Variable;
      Add(Variables_Table, New_Var);



      ------------------------------------------------------
      --              Task variables
      ------------------------------------------------------


      -- "tasks.priority"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.priority");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.deadline"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.deadline");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.start_time"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.start_time");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.capacity"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.capacity");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.rest_of_capacity"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.rest_of_capacity");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.activation_number"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.activation_number");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.used_cpu"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.used_cpu");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.period"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.period");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.ready"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_Boolean;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.ready");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.processor_name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_String;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.processor_name");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.blocking_time"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.blocking_time");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.name"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_String;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.name");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- "tasks.jitter"
      --
      New_Var:= NEW Variable_Record;
      New_Expr_Array := NEW Array_Variable_Expr;
      New_Expr_Array.Variable_Type:=Simulation_Array_Integer;
      New_Expr_Array.Name:=To_Unbounded_String("tasks.jitter");
      New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
      Add(Variables_Table, New_Var);


      -- Task's user defined parameters
      --
      -- Find user defined parameters : scan all task
      -- Variables and add them
      --
      IF NOT Is_Empty(My_Tasks)
            THEN

         Reset_Iterator(My_Tasks, Iterator);
         LOOP
            Current_Element(My_Tasks, A_Task, Iterator);

            FOR I IN 0..A_Task.Param.Nb_Entries-1 LOOP
               -- Is the Variable already exists ?
               --
               Do_Insert:=True;
               FOR J IN 0..Variables_Table.Nb_Entries-1 LOOP
                  IF Variables_Table.Entries(J).Variable.Name=
                        To_Unbounded_String("tasks.") &
                        A_Task.Param.Entries(I).Name
                        THEN
                     Do_Insert:=False;
                  END IF;
               END LOOP;

               IF Do_Insert
                     THEN
                  New_Var:= NEW Variable_Record;
                  New_Expr_Array := NEW
                     Array_Variable_Expr;
                  IF A_Task.Param.Entries(I).Ptype=Integer_Parameter
                        THEN
                     New_Expr_Array.Variable_Type:=
                        Simulation_Array_Integer;
                  END IF;
                  IF A_Task.Param.Entries(I).Ptype=Double_Parameter
                        THEN
                     New_Expr_Array.Variable_Type:=
                        Simulation_Array_Double;
                  END IF;
                  IF A_Task.Param.Entries(I).Ptype=Boolean_Parameter
                        THEN
                     New_Expr_Array.Variable_Type:=
                        Simulation_Array_Boolean;
                  END IF;
                  IF A_Task.Param.Entries(I).Ptype=String_Parameter
                        THEN
                     New_Expr_Array.Variable_Type:=
                        Simulation_Array_String;
                  END IF;

                  New_Expr_Array.Name:=To_Unbounded_String("tasks.") &
                     A_Task.Param.Entries(I).Name;
                  New_Var.Variable:=Variable_Expr_Ptr(New_Expr_Array);
                  Add(Variables_Table, New_Var);
               END IF;

            END LOOP;

            EXIT WHEN Is_Last_Element(My_Tasks, Iterator);
            Next_Element(My_Tasks, Iterator);

         END LOOP;
      END IF;




      ---------------------------------------------------------------
      -- Initialize simulation data : provide initial values for
      -- all variables
      ---------------------------------------------------------------


      -- Allocate simulation data
      --  	
      FOR I IN 0..Variables_Table.Nb_Entries-1 LOOP

         Variables_Table.Entries(I).Simulation:=NEW
            Simulation_Value(
            Variable_Expr_Ptr(Variables_Table.
               Entries(
                  I).
               Variable).Variable_Type);

      END LOOP;


   END Create_Parametric_Variables;





   --------------------------------------------------


   PROCEDURE Put (
         S : IN     Constant_Expr) IS
   BEGIN
      Put(To_Unbounded_String(S));
   END Put;

   PROCEDURE Put (
         S : IN     Constant_Expr_Ptr) IS
   BEGIN
      Put(To_Unbounded_String(S.All));
   END Put;



   PROCEDURE Put (
         S : IN     Variable_Expr) IS
   BEGIN
      Put(S.Name);
      Put("/");
      Put(S.Variable_Type'Img);
   END Put;

   PROCEDURE Put (
         S : IN     Variable_Expr_Ptr) IS
   BEGIN
      Put(S.All);
   END Put;



   PROCEDURE Put (
         S : IN     Array_Variable_Expr) IS
   BEGIN
      Put(S.Name);
      Put("/");
      Put(S.Variable_Type'Img);
      IF (S.Array_Index/=NULL)  THEN
         Put(S.Array_Index.All);
      END IF;
   END Put;

   PROCEDURE Put (
         S : IN     Array_Variable_Expr_Ptr) IS
   BEGIN
      Put(S.All);
   END Put;


   PROCEDURE Put (
         S : IN     Binary_Expr) IS
   BEGIN
      IF (S.Lvalue/=NULL)  THEN
         Put(S.Lvalue.All);
      END IF;
      Put(S.Ope);
      IF (S.Rvalue/=NULL)  THEN
         Put(S.Rvalue.All);
      END IF;
   END Put;

   PROCEDURE Put (
         S : IN     Binary_Expr_Ptr) IS
   BEGIN
      Put(S.All);
   END Put;




   PROCEDURE Put (
         S : IN     Unary_Expr) IS
   BEGIN
      Put(To_Unbounded_String(S));
   END Put;

   PROCEDURE Put (
         S : IN     Unary_Expr_Ptr) IS
   BEGIN
      Put(To_Unbounded_String(S.All));
   END Put;


   PROCEDURE Put (
         S : IN     Operator) IS
   BEGIN
      CASE S IS
         WHEN Plus=>
            Put(" + ");
         WHEN Minus =>
            Put(" - ");
         WHEN Div  =>
            Put(" / ");
         WHEN Mult =>
            Put(" * ");
         WHEN Exp =>
            Put(" ** ");
         WHEN Modulo =>
            Put(" mod ");
         WHEN Equal =>
            Put(" = ");
         WHEN Not_Equal =>
            Put(" /= ");
         WHEN Equal_Less =>
            Put(" <= ");
         WHEN Equal_Greater =>
            Put(" >= ");
         WHEN Inf =>
            Put(" < ");
         WHEN Sup =>
            Put(" > ");
         WHEN Logic_And =>
            Put(" and ");
         WHEN Logic_Or =>
            Put(" or ");
         WHEN Logic_Not =>
            Put(" not ");
         WHEN Lcm =>
            Put(" lcm ");
         WHEN Min_Operator =>
            Put(" min ");
         WHEN Max_Operator =>
            Put(" max ");
         WHEN Min_To_Index =>
            Put(" min_to_index ");
         WHEN Max_To_Index =>
            Put(" max_to_index ");
         WHEN Abs_Operator =>
            Put(" abs ");
         WHEN Get_Buffer_Index =>
            Put(" get_buffer_index ");
         WHEN Get_Message_Index =>
            Put(" get_message_index ");
         WHEN Get_Resource_Index =>
            Put(" get_resource_index ");
         WHEN Get_Task_Index =>
            Put(" get_task_index ");
         WHEN Concat =>
            Put(" & ");
      END CASE;
   END Put;




   ------------------------------------------------------


   PROCEDURE Initialize (
         S : IN OUT Constant_Expr) IS
   BEGIN
      S.Expression_Type:=Constant_Type;
   END Initialize;


   PROCEDURE Initialize (
         S : IN OUT Variable_Expr) IS
   BEGIN
      S.Expression_Type:=Variable_Type;
   END Initialize;


   PROCEDURE Initialize (
         S : IN OUT Array_Variable_Expr) IS
   BEGIN
      S.Expression_Type:=Array_Variable_Type;
   END Initialize;


   PROCEDURE Initialize (
         S : IN OUT Binary_Expr) IS
   BEGIN
      S.Expression_Type:=Binary_Type;
   END Initialize;


   PROCEDURE Initialize (
         S : IN OUT Unary_Expr) IS
   BEGIN
      S.Expression_Type:=Unary_Type;
   END Initialize;



   ---------------------------------------------------------------



   FUNCTION To_Unbounded_String (
         C : Constant_Expr)
     RETURN Unbounded_String IS
   BEGIN
      RETURN     To_Unbounded_String(C.Value);
   END To_Unbounded_String;


   FUNCTION To_Unbounded_String (
         C : Variable_Expr)
     RETURN Unbounded_String IS
   BEGIN
      RETURN C.Name;
   END To_Unbounded_String;


   FUNCTION To_Unbounded_String (
         C : Array_Variable_Expr)
     RETURN Unbounded_String IS
   BEGIN
      RETURN C.Name;
   END To_Unbounded_String;


   FUNCTION To_Unbounded_String (
         C : Binary_Expr)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;

   BEGIN
      CASE C.Ope IS
         WHEN Minus =>
            Result:=Result & "-(";
         WHEN Plus =>
            Result:=Result & "+(";
         WHEN Mult =>
            Result:=Result & "*(";
         WHEN Div =>
            Result:=Result & "/(";
         WHEN Equal =>
            Result:=Result & "=(";
         WHEN Not_Equal =>
            Result:=Result & "/=(";
         WHEN Equal_Less =>
            Result:=Result & "<=(";
         WHEN Equal_Greater =>
            Result:=Result & ">=(";
         WHEN Inf =>
            Result:=Result & "<(";
         WHEN Sup =>
            Result:=Result & ">(";
         WHEN Logic_And =>
            Result:=Result & "and(";
         WHEN Logic_Or =>
            Result:=Result & "or(";
         WHEN Max_Operator =>
            Result:=Result & "max(";
         WHEN Min_Operator =>
            Result:=Result & "min(";
         WHEN Lcm =>
            Result:=Result & "lcm(";
         WHEN Concat =>
            Result:=Result & "&(";
         WHEN Exp =>
            Result:=Result & "**(";
         WHEN Modulo =>
            NULL;
         WHEN Logic_Not =>
            NULL;
         WHEN Min_To_Index =>
            NULL;
         WHEN Max_To_Index =>
            NULL;
         WHEN Abs_Operator =>
            NULL;
         WHEN Get_Resource_Index =>
            NULL;
         WHEN Get_Task_Index =>
            NULL;
         WHEN Get_Message_Index =>
            NULL;
         WHEN Get_Buffer_Index =>
            NULL;
      END CASE;
      Result:=Result & To_Unbounded_String(C.Lvalue.All);
      Result:=Result & ",";
      Result:=Result & To_Unbounded_String(C.Rvalue.All);
      Result:=Result & ")";
      RETURN Result;
   END To_Unbounded_String;




   FUNCTION To_Unbounded_String (
         C : Unary_Expr)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;

   BEGIN
      CASE C.Ope IS
         WHEN Minus =>
            NULL;
         WHEN Plus =>
            NULL;
         WHEN Mult =>
            NULL;
         WHEN Div =>
            NULL;
         WHEN Equal =>
            NULL;
         WHEN Not_Equal =>
            NULL;
         WHEN Equal_Less =>
            NULL;
         WHEN Equal_Greater =>
            NULL;
         WHEN Inf =>
            NULL;
         WHEN Sup =>
            NULL;
         WHEN Logic_And =>
            NULL;
         WHEN Logic_Or =>
            NULL;
         WHEN Max_Operator =>
            NULL;
         WHEN Min_Operator =>
            NULL;
         WHEN Concat =>
            NULL;
         WHEN Exp =>
            NULL;
         WHEN Lcm =>
            NULL;
         WHEN Modulo =>
            Result:=Result & "mod(";
         WHEN Logic_Not =>
            Result:=Result & "not(";
         WHEN Min_To_Index =>
            Result:=Result & "min_to_index(";
         WHEN Max_To_Index =>
            Result:=Result & "max_to_index(";
         WHEN Abs_Operator =>
            Result:=Result & "abs(";
         WHEN Get_Resource_Index =>
            Result:=Result & "get_resource_index(";
         WHEN Get_Task_Index =>
            Result:=Result & "get_task_index(";
         WHEN Get_Message_Index =>
            Result:=Result & "get_message_index(";
         WHEN Get_Buffer_Index =>
            Result:=Result & "get_buffer_index(";
      END CASE;
      Result:=Result & To_Unbounded_String(C.Value.All);
      Result:=Result & ")";
      RETURN Result;
   END To_Unbounded_String;





   ---------------------------------------------------------------



   FUNCTION Get_Type (
         C : Constant_Expr)
     RETURN Simulation_Type IS
   BEGIN
      RETURN C.Value.Ptype;
   END Get_Type;


   FUNCTION Get_Type (
         C : Variable_Expr)
     RETURN Simulation_Type IS
   BEGIN
      RETURN C.Variable_Type;
   END Get_Type;



   FUNCTION Get_Type (
         C : Binary_Expr)
     RETURN Simulation_Type IS
   BEGIN
      RETURN Get_Type(C.Rvalue.All);
   END Get_Type;


   FUNCTION Get_Type (
         C : Unary_Expr)
     RETURN Simulation_Type IS
   BEGIN
      RETURN Get_Type(C.Value.All);
   END Get_Type;





   -------------------------------------------------------------------------






   FUNCTION Apply_Random_Operator (
         Seed       : IN     Generator;
         Parameters : IN     Random_Law_Parameters_Type)
     RETURN Integer;

   FUNCTION Apply_Operator (
         Lvalue : Integer;
         Rvalue : Integer;
         Ope    : Operator)
     RETURN Integer;

   FUNCTION Apply_Operator (
         Lvalue : Double;
         Rvalue : Double;
         Ope    : Operator)
     RETURN Double;

   FUNCTION Apply_Operator (
         Lvalue : Double;
         Rvalue : Double;
         Ope    : Operator)
     RETURN Boolean;

   FUNCTION Apply_Operator (
         Lvalue : Unbounded_String;
         Rvalue : Unbounded_String;
         Ope    : Operator)
     RETURN Unbounded_String;

   FUNCTION Apply_Operator (
         Lvalue : Unbounded_String;
         Rvalue : Unbounded_String;
         Ope    : Operator)
     RETURN Boolean;

   FUNCTION Apply_Operator (
         Lvalue : Integer;
         Rvalue : Integer;
         Ope    : Operator)
     RETURN Boolean;

   FUNCTION Apply_Operator (
         Lvalue : Boolean;
         Rvalue : Boolean;
         Ope    : Operator)
     RETURN Boolean;

   FUNCTION Compute_Binary_Value (
         Lvalue : Simulation_Value_Ptr;
         Rvalue : Simulation_Value_Ptr;
         Ope    : Operator)
     RETURN Simulation_Value_Ptr;

   FUNCTION Apply_Operator (
         Value : Boolean;
         Ope   : Operator)
     RETURN Boolean;

   FUNCTION Apply_Operator (
         Value : Integer;
         Ope   : Operator)
     RETURN Integer;

   FUNCTION Apply_Operator (
         Value : Double;
         Ope   : Operator)
     RETURN Double;

   FUNCTION Apply_Operator (
         Value : Unbounded_String;
         V     : Variables_Table_Type;
         Ope   : Operator)
     RETURN Integer;




   FUNCTION Apply_Random_Operator (
         Seed       : IN     Generator;
         Parameters : IN     Random_Law_Parameters_Type)
     RETURN Integer IS

      Diff  : Double;
      Value : Integer;

   BEGIN

      CASE Parameters.Law IS

         WHEN Exponential_Law =>
            Value:=Integer(Get_Exponential_Time(Double(Parameters.
                     Parameter1), Seed));

         WHEN Uniform_Law =>
            Diff:=Double(Parameters.Parameter2-Parameters.Parameter1);
            Value:=Integer(Double(Parameters.Parameter1)+Double(Random(
                     Seed))*Diff);
      END CASE;

      RETURN Value;

   END Apply_Random_Operator;




   FUNCTION Apply_Operator (
         Lvalue : Integer;
         Rvalue : Integer;
         Ope    : Operator)
     RETURN Integer IS

   BEGIN

      IF Ope=Min_Operator THEN
         IF Lvalue<Rvalue THEN
            RETURN Lvalue;
         ELSE
            RETURN Rvalue;
         END IF;
      END IF;
      IF Ope=Max_Operator THEN
         IF Lvalue>Rvalue THEN
            RETURN Lvalue;
         ELSE
            RETURN Rvalue;
         END IF;
      END IF;
      IF Ope=Lcm THEN
         RETURN Integer_Util.Lcm(Lvalue,Rvalue);
      END IF;
      IF Ope=Plus THEN
         RETURN Lvalue+Rvalue;
      END IF;
      IF Ope=Minus THEN
         RETURN Lvalue-Rvalue;
      END IF;
      IF Ope=Mult THEN
         RETURN Lvalue*Rvalue;
      END IF;
      IF Ope=Div THEN
         RETURN Lvalue/Rvalue;
      END IF;
      IF Ope=Modulo THEN
         RETURN Lvalue mod Rvalue;
      END IF;


      Raise_Exception(Type_Error'Identity,
         To_String(Lb_Uncompatible_Type_Error(Current_Language) &
            Lb_Colon & To_Unbounded_String(Lvalue'Img) &
            Lb_Comma  & To_Unbounded_String(Rvalue'Img) &
            Lb_Comma &
            To_Unbounded_String(Ope'Img)  ) );


      -- Arithmetic exception : return 0 in any cases
      -- to avoid stopping the application
      --
   EXCEPTION
      WHEN OTHERS =>
         RETURN 0;

   END Apply_Operator;


   FUNCTION Apply_Operator (
         Lvalue : Double;
         Rvalue : Double;
         Ope    : Operator)
     RETURN Double IS
   BEGIN

      IF Ope=Min_Operator THEN
         IF Lvalue<Rvalue THEN
            RETURN Lvalue;
         ELSE
            RETURN Rvalue;
         END IF;
      END IF;
      IF Ope=Max_Operator THEN
         IF Lvalue>Rvalue THEN
            RETURN Lvalue;
         ELSE
            RETURN Rvalue;
         END IF;
      END IF;
      IF Ope=Lcm THEN
         RETURN Double(Integer_Util.Lcm(Integer(Lvalue),Integer(Rvalue)));
      END IF;
      IF Ope=Plus THEN
         RETURN Lvalue+Rvalue;
      END IF;
      IF Ope=Minus THEN
         RETURN Lvalue-Rvalue;
      END IF;
      IF Ope=Mult THEN
         RETURN Lvalue*Rvalue;
      END IF;

      IF Ope=Div THEN
         RETURN Lvalue/Rvalue;
      END IF;

      Raise_Exception(Type_Error'Identity,
         To_String(Lb_Uncompatible_Type_Error(Current_Language) &
            Lb_Colon & To_Unbounded_String(Lvalue'Img) &
            Lb_Comma  & To_Unbounded_String(Rvalue'Img) &
            Lb_Comma &
            To_Unbounded_String(Ope'Img)  ) );


      -- Arithmetic exception : return 0 in any cases
      -- to avoid stopping the application
      --
   EXCEPTION
      WHEN OTHERS =>
         RETURN 0.0;
   END Apply_Operator;


   FUNCTION Apply_Operator (
         Lvalue : Double;
         Rvalue : Double;
         Ope    : Operator)
     RETURN Boolean IS
   BEGIN
      IF Ope=Equal THEN
         RETURN Lvalue=Rvalue;
      END IF;
      IF Ope=Not_Equal THEN
         RETURN Lvalue/=Rvalue;
      END IF;
      IF Ope=Equal_Less THEN
         RETURN Lvalue<=Rvalue;
      END IF;
      IF Ope=Equal_Greater THEN
         RETURN Lvalue>=Rvalue;
      END IF;
      IF Ope=Inf THEN
         RETURN Lvalue<Rvalue;
      END IF;
      IF Ope=Sup THEN
         RETURN Lvalue>Rvalue;
      END IF;

      Raise_Exception(Type_Error'Identity,
         To_String(Lb_Uncompatible_Type_Error(Current_Language) &
            Lb_Colon & To_Unbounded_String(Lvalue'Img) &
            Lb_Comma  & To_Unbounded_String(Rvalue'Img) &
            Lb_Comma &
            To_Unbounded_String(Ope'Img)  ) );


   END Apply_Operator;


   FUNCTION Apply_Operator (
         Lvalue : Unbounded_String;
         Rvalue : Unbounded_String;
         Ope    : Operator)
     RETURN Unbounded_String IS
   BEGIN
      IF Ope=Concat
            THEN
         RETURN Lvalue & Rvalue;
      END IF;


      Raise_Exception(Type_Error'Identity,
         To_String(Lb_Uncompatible_Type_Error(Current_Language) &
            Lb_Colon & Lvalue &
            Lb_Comma  & Rvalue &
            Lb_Comma &
            To_Unbounded_String(Ope'Img)) );

   END Apply_Operator;



   FUNCTION Apply_Operator (
         Lvalue : Unbounded_String;
         Rvalue : Unbounded_String;
         Ope    : Operator)
     RETURN Boolean IS
   BEGIN
      IF Ope=Equal THEN
         RETURN Lvalue=Rvalue;
      END IF;
      IF Ope=Not_Equal THEN
         RETURN Lvalue/=Rvalue;
      END IF;
      IF Ope=Equal_Less THEN
         RETURN Lvalue<=Rvalue;
      END IF;
      IF Ope=Equal_Greater THEN
         RETURN Lvalue>=Rvalue;
      END IF;
      IF Ope=Inf THEN
         RETURN Lvalue<Rvalue;
      END IF;
      IF Ope=Sup THEN
         RETURN Lvalue>Rvalue;
      END IF;

      Raise_Exception(Type_Error'Identity,
         To_String(Lb_Uncompatible_Type_Error(Current_Language) &
            Lb_Colon & Lvalue &
            Lb_Comma  & Rvalue &
            Lb_Comma &
            To_Unbounded_String(Ope'Img)  ) );

   END Apply_Operator;


   FUNCTION Apply_Operator (
         Lvalue : Integer;
         Rvalue : Integer;
         Ope    : Operator)
     RETURN Boolean IS
   BEGIN
      IF Ope=Equal THEN
         RETURN Lvalue=Rvalue;
      END IF;
      IF Ope=Not_Equal THEN
         RETURN Lvalue/=Rvalue;
      END IF;
      IF Ope=Equal_Less THEN
         RETURN Lvalue<=Rvalue;
      END IF;
      IF Ope=Equal_Greater THEN
         RETURN Lvalue>=Rvalue;
      END IF;
      IF Ope=Inf THEN
         RETURN Lvalue<Rvalue;
      END IF;

      IF Ope=Sup THEN
         RETURN Lvalue>Rvalue;
      END IF;


      Raise_Exception(Type_Error'Identity,
         To_String(Lb_Uncompatible_Type_Error(Current_Language) &
            Lb_Colon & To_Unbounded_String(Lvalue'Img) &
            Lb_Comma  & To_Unbounded_String(Rvalue'Img) &
            Lb_Comma &
            To_Unbounded_String(Ope'Img)  ) );

   END Apply_Operator;


   FUNCTION Apply_Operator (
         Lvalue : Boolean;
         Rvalue : Boolean;
         Ope    : Operator)
     RETURN Boolean IS
   BEGIN

      IF Ope=Equal THEN
         RETURN Lvalue=Rvalue;
      END IF;
      IF Ope=Not_Equal THEN
         RETURN Lvalue/=Rvalue;
      END IF;
      IF Ope=Equal_Less THEN
         RETURN Lvalue<=Rvalue;
      END IF;
      IF Ope=Equal_Greater THEN
         RETURN Lvalue>=Rvalue;
      END IF;
      IF Ope=Inf THEN
         RETURN Lvalue<Rvalue;
      END IF;
      IF Ope=Sup THEN
         RETURN Lvalue>Rvalue;
      END IF;
      IF Ope=Logic_And THEN
         RETURN Lvalue AND Rvalue;
      END IF;
      IF Ope=Logic_Or THEN
         RETURN Lvalue OR Rvalue;
      END IF;


      Raise_Exception(Type_Error'Identity,
         To_String(Lb_Uncompatible_Type_Error(Current_Language) &
            Lb_Colon & To_Unbounded_String(Lvalue'Img) &
            Lb_Comma  & To_Unbounded_String(Rvalue'Img) &
            Lb_Comma &
            To_Unbounded_String(Ope'Img)  ) );


   END Apply_Operator;





   FUNCTION Compute_Binary_Value (
         Lvalue : Simulation_Value_Ptr;
         Rvalue : Simulation_Value_Ptr;
         Ope    : Operator)
     RETURN Simulation_Value_Ptr IS

      Res : Simulation_Value_Ptr;

   BEGIN

      CASE Lvalue.Ptype  IS



         -- Nothing to compute in the cases of random type
         -- because value_of functions already do the job
         -- by returning integer values
         --
         WHEN Simulation_Random =>
            NULL;

         WHEN Simulation_Array_Random =>
            NULL;

         WHEN Simulation_Time_Unit_Array_Random =>
            NULL;




         WHEN Simulation_Integer =>

            IF Rvalue.Ptype=Simulation_Integer THEN

               IF (Ope=Equal) OR
                     (Ope=Not_Equal) OR
                     (Ope=Equal_Less) OR
                     (Ope=Equal_Greater) OR
                     (Ope=Inf) OR
                     (Ope=Sup) THEN
                  Res:=NEW Simulation_Value(Simulation_Boolean);
                  Res.Boolean_Value:=Apply_Operator(Lvalue.Integer_Value,
                     Rvalue.Integer_Value, Ope);
               ELSE
                  Res:=NEW Simulation_Value(Simulation_Integer);
                  Res.Integer_Value:=Apply_Operator(Lvalue.Integer_Value,
                     Rvalue.Integer_Value, Ope);
               END IF;

            ELSE
               IF Rvalue.Ptype=Simulation_Array_Integer THEN
                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res:=NEW Simulation_Value(
                        Simulation_Array_Boolean);
                  ELSE
                     Res:=NEW Simulation_Value(
                        Simulation_Array_Integer);
                  END IF;

                  FOR I IN Simulations_Range LOOP
                     IF (Ope=Equal) OR
                           (Ope=Not_Equal) OR
                           (Ope=Equal_Less) OR
                           (Ope=Equal_Greater) OR
                           (Ope=Inf) OR
                           (Ope=Sup) THEN
                        Res.Boolean_Table_Value(I):=Apply_Operator(
                           Lvalue.Boolean_Value,
                           Rvalue.Boolean_Table_Value(I),Ope);
                     ELSE
                        Res.Integer_Table_Value(I):=Apply_Operator(
                           Lvalue.Integer_Value,
                           Rvalue.Integer_Table_Value(I),Ope);
                     END IF;
                  END LOOP;

               ELSE

                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res:=NEW Simulation_Value(
                        Simulation_Time_Unit_Array_Boolean);
                  ELSE
                     Res:=NEW Simulation_Value(
                        Simulation_Time_Unit_Array_Integer);
                  END IF;

                  FOR I IN Time_Unit_Range LOOP
                     IF (Ope=Equal) OR
                           (Ope=Not_Equal) OR
                           (Ope=Equal_Less) OR
                           (Ope=Equal_Greater) OR
                           (Ope=Inf) OR
                           (Ope=Sup) THEN
                        Res.Boolean_Time_Unit_Table_Value(I):=
                           Apply_Operator(
                           Lvalue.Boolean_Value,
                           Rvalue.Boolean_Time_Unit_Table_Value(I),Ope);
                     ELSE
                        Res.Integer_Time_Unit_Table_Value(I):=
                           Apply_Operator(
                           Lvalue.Integer_Value,
                           Rvalue.Integer_Time_Unit_Table_Value(I),Ope);
                     END IF;
                  END LOOP;
               END IF;
            END IF;


         WHEN Simulation_String =>

            IF Rvalue.Ptype=Simulation_String THEN

               IF (Ope=Equal) OR
                     (Ope=Not_Equal) OR
                     (Ope=Equal_Less) OR
                     (Ope=Equal_Greater) OR
                     (Ope=Inf) OR
                     (Ope=Sup) THEN
                  Res:=NEW Simulation_Value(Simulation_Boolean);
                  Res.Boolean_Value:=Apply_Operator(Lvalue.String_Value,
                     Rvalue.String_Value, Ope);
               ELSE
                  Res:=NEW Simulation_Value(Simulation_String);
                  Res.String_Value:=Apply_Operator(Lvalue.String_Value,
                     Rvalue.String_Value, Ope);
               END IF;

            ELSE
               IF Rvalue.Ptype=Simulation_Array_String THEN
                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res:=NEW Simulation_Value(
                        Simulation_Array_Boolean);
                  ELSE
                     Res:=NEW Simulation_Value(
                        Simulation_Array_String);
                  END IF;

                  FOR I IN Simulations_Range LOOP
                     IF (Ope=Equal) OR
                           (Ope=Not_Equal) OR
                           (Ope=Equal_Less) OR
                           (Ope=Equal_Greater) OR
                           (Ope=Inf) OR
                           (Ope=Sup) THEN
                        Res.Boolean_Table_Value(I):=Apply_Operator(
                           Lvalue.Boolean_Value,
                           Rvalue.Boolean_Table_Value(I),Ope);
                     ELSE
                        Res.String_Table_Value(I):=Apply_Operator(
                           Lvalue.String_Value,
                           Rvalue.String_Table_Value(I),Ope);
                     END IF;
                  END LOOP;

               ELSE

                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res:=NEW Simulation_Value(
                        Simulation_Time_Unit_Array_Boolean);
                  ELSE
                     Res:=NEW Simulation_Value(
                        Simulation_Time_Unit_Array_String);
                  END IF;

                  FOR I IN Time_Unit_Range LOOP
                     IF (Ope=Equal) OR
                           (Ope=Not_Equal) OR
                           (Ope=Equal_Less) OR
                           (Ope=Equal_Greater) OR
                           (Ope=Inf) OR
                           (Ope=Sup) THEN
                        Res.Boolean_Time_Unit_Table_Value(I):=
                           Apply_Operator(
                           Lvalue.Boolean_Value,
                           Rvalue.Boolean_Time_Unit_Table_Value(I),Ope);
                     ELSE
                        Res.String_Time_Unit_Table_Value(I):=
                           Apply_Operator(
                           Lvalue.String_Value,
                           Rvalue.String_Time_Unit_Table_Value(I),Ope);
                     END IF;
                  END LOOP;
               END IF;
            END IF;





         WHEN Simulation_Double =>


            IF Rvalue.Ptype=Simulation_Double THEN

               IF (Ope=Equal) OR
                     (Ope=Not_Equal) OR
                     (Ope=Equal_Less) OR
                     (Ope=Equal_Greater) OR
                     (Ope=Inf) OR
                     (Ope=Sup) THEN
                  Res:=NEW Simulation_Value(Simulation_Boolean);
                  Res.Boolean_Value:=Apply_Operator(Lvalue.Double_Value,
                     Rvalue.Double_Value, Ope);
               ELSE
                  Res:=NEW Simulation_Value(Simulation_Double);
                  Res.Double_Value:=Apply_Operator(Lvalue.Double_Value,
                     Rvalue. Double_Value, Ope);
               END IF;

            ELSE
               IF Rvalue.Ptype=Simulation_Array_Double THEN
                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res:=NEW Simulation_Value(
                        Simulation_Array_Boolean);
                  ELSE
                     Res:=NEW Simulation_Value(
                        Simulation_Array_Double);
                  END IF;

                  FOR I IN Simulations_Range LOOP
                     IF (Ope=Equal) OR
                           (Ope=Not_Equal) OR
                           (Ope=Equal_Less) OR
                           (Ope=Equal_Greater) OR
                           (Ope=Inf) OR
                           (Ope=Sup) THEN
                        Res.Boolean_Table_Value(I):=Apply_Operator(
                           Lvalue.Boolean_Value,
                           Rvalue.Boolean_Table_Value(I),Ope);
                     ELSE
                        Res.Double_Table_Value(I):=Apply_Operator(
                           Lvalue.Double_Value,
                           Rvalue.Double_Table_Value(I),Ope);
                     END IF;
                  END LOOP;

               ELSE

                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res:=NEW Simulation_Value(
                        Simulation_Time_Unit_Array_Boolean);
                  ELSE
                     Res:=NEW Simulation_Value(
                        Simulation_Time_Unit_Array_Double);
                  END IF;

                  FOR I IN Time_Unit_Range LOOP
                     IF (Ope=Equal) OR
                           (Ope=Not_Equal) OR
                           (Ope=Equal_Less) OR
                           (Ope=Equal_Greater) OR
                           (Ope=Inf) OR
                           (Ope=Sup) THEN
                        Res.Boolean_Time_Unit_Table_Value(I):=
                           Apply_Operator(
                           Lvalue.Boolean_Value,
                           Rvalue.Boolean_Time_Unit_Table_Value(I),Ope);
                     ELSE
                        Res.Double_Time_Unit_Table_Value(I):=
                           Apply_Operator(
                           Lvalue.Double_Value,
                           Rvalue.Double_Time_Unit_Table_Value(I),Ope);
                     END IF;
                  END LOOP;
               END IF;
            END IF;






         WHEN Simulation_Boolean =>


            IF Rvalue.Ptype=Simulation_Boolean THEN

               Res:=NEW Simulation_Value(Simulation_Boolean);
               Res.Boolean_Value:=Apply_Operator(Lvalue.Boolean_Value,
                  Rvalue.Boolean_Value, Ope);
            ELSE
               IF Rvalue.Ptype=Simulation_Array_Boolean THEN

                  Res:=NEW Simulation_Value(
                     Simulation_Array_Boolean);

                  FOR I IN Simulations_Range LOOP
                     Res.Boolean_Table_Value(I):=Apply_Operator(
                        Lvalue.Boolean_Value,
                        Rvalue.Boolean_Table_Value(I),Ope);
                  END LOOP;

               ELSE

                  Res:=NEW Simulation_Value(
                     Simulation_Time_Unit_Array_Boolean);

                  FOR I IN Time_Unit_Range LOOP
                     Res.Boolean_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.Boolean_Value,
                        Rvalue.Boolean_Time_Unit_Table_Value(I),Ope);
                  END LOOP;

               END IF;
            END IF;





         WHEN Simulation_Array_Integer =>


            IF (Ope=Equal) OR
                  (Ope=Not_Equal) OR
                  (Ope=Equal_Less) OR
                  (Ope=Equal_Greater) OR
                  (Ope=Inf) OR
                  (Ope=Sup) THEN
               Res:=NEW Simulation_Value(
                  Simulation_Array_Boolean);
            ELSE
               Res:=NEW Simulation_Value(
                  Simulation_Array_Integer);
            END IF;


            FOR I IN Simulations_Range LOOP

               IF Rvalue.Ptype=Simulation_Array_Integer THEN

                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Table_Value(I):=Apply_Operator(Lvalue.
                        Boolean_Table_Value(I), Rvalue.
                        Boolean_Table_Value(I), Ope);
                  ELSE
                     Res.Integer_Table_Value(I):=Apply_Operator(Lvalue.
                        Integer_Table_Value(I), Rvalue.
                        Integer_Table_Value(I), Ope);
                  END IF;

               ELSE

                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Table_Value(I):=Apply_Operator(Lvalue.
                        Boolean_Table_Value(I), Rvalue.Boolean_Value, Ope);
                  ELSE
                     Res.Integer_Table_Value(I):=Apply_Operator(Lvalue.
                        Integer_Table_Value(I), Rvalue.Integer_Value, Ope);
                  END IF;

               END IF;

            END LOOP;



         WHEN Simulation_Array_Double =>


            IF (Ope=Equal) OR
                  (Ope=Not_Equal) OR
                  (Ope=Equal_Less) OR
                  (Ope=Equal_Greater) OR
                  (Ope=Inf) OR
                  (Ope=Sup) THEN
               Res:=NEW Simulation_Value(
                  Simulation_Array_Boolean);
            ELSE
               Res:=NEW Simulation_Value(
                  Simulation_Array_Double);
            END IF;

            FOR I IN Simulations_Range LOOP

               IF Rvalue.Ptype=Simulation_Array_Double THEN
                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Table_Value(I):=Apply_Operator(Lvalue.
                        Boolean_Table_Value(I), Rvalue.
                        Boolean_Table_Value(I),Ope);
                  ELSE
                     Res.Double_Table_Value(I):=Apply_Operator(Lvalue.
                        Double_Table_Value(I), Rvalue.Double_Table_Value(
                           I),Ope);
                  END IF;
               ELSE
                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Table_Value(I):=Apply_Operator(Lvalue.
                        Boolean_Table_Value(I), Rvalue.Boolean_Value, Ope);
                  ELSE
                     Res.Double_Table_Value(I):=Apply_Operator(Lvalue.
                        Double_Table_Value(I), Rvalue.Double_Value, Ope);
                  END IF;
               END IF;

            END LOOP;




         WHEN Simulation_Array_String =>


            IF (Ope=Equal) OR
                  (Ope=Not_Equal) OR
                  (Ope=Equal_Less) OR
                  (Ope=Equal_Greater) OR
                  (Ope=Inf) OR
                  (Ope=Sup) THEN
               Res:=NEW Simulation_Value(
                  Simulation_Array_Boolean);
            ELSE
               Res:=NEW Simulation_Value(
                  Simulation_Array_String);
            END IF;

            FOR I IN Simulations_Range LOOP

               IF Rvalue.Ptype=Simulation_Array_String THEN
                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Table_Value(I):=Apply_Operator(Lvalue.
                        Boolean_Table_Value(I), Rvalue.
                        Boolean_Table_Value(I),Ope);
                  ELSE
                     Res.String_Table_Value(I):=Apply_Operator(Lvalue.
                        String_Table_Value(I), Rvalue.String_Table_Value(
                           I),Ope);
                  END IF;
               ELSE
                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Table_Value(I):=Apply_Operator(Lvalue.
                        Boolean_Table_Value(I), Rvalue.Boolean_Value, Ope);
                  ELSE
                     Res.String_Table_Value(I):=Apply_Operator(Lvalue.
                        String_Table_Value(I), Rvalue.String_Value, Ope);
                  END IF;
               END IF;

            END LOOP;




         WHEN Simulation_Array_Boolean =>


            Res:=NEW Simulation_Value(
               Simulation_Array_Boolean);

            FOR I IN Simulations_Range LOOP

               IF Rvalue.Ptype=Simulation_Array_Boolean THEN
                  Res.Boolean_Table_Value(I):=Apply_Operator(
                     Lvalue.Boolean_Table_Value(I),
                     Rvalue.Boolean_Table_Value(I),Ope);
               ELSE
                  Res.Boolean_Table_Value(I):=Apply_Operator(
                     Lvalue.
                     Boolean_Table_Value(I),
                     Rvalue.Boolean_Value, Ope);
               END IF;

            END LOOP;




         WHEN Simulation_Time_Unit_Array_Integer =>


            IF (Ope=Equal) OR
                  (Ope=Not_Equal) OR
                  (Ope=Equal_Less) OR
                  (Ope=Equal_Greater) OR
                  (Ope=Inf) OR
                  (Ope=Sup) THEN
               Res:=NEW Simulation_Value(
                  Simulation_Time_Unit_Array_Boolean);
            ELSE
               Res:=NEW Simulation_Value(
                  Simulation_Time_Unit_Array_Integer);
            END IF;


            FOR I IN Time_Unit_Range LOOP

               IF Rvalue.Ptype=Simulation_Time_Unit_Array_Integer THEN

                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        Boolean_Time_Unit_Table_Value(I), Rvalue.
                        Boolean_Time_Unit_Table_Value(I), Ope);
                  ELSE
                     Res.Integer_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        Integer_Time_Unit_Table_Value(I), Rvalue.
                        Integer_Time_Unit_Table_Value(I), Ope);
                  END IF;

               ELSE

                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        Boolean_Time_Unit_Table_Value(I), Rvalue.
                        Boolean_Value, Ope);
                  ELSE
                     Res.Integer_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        Integer_Time_Unit_Table_Value(I), Rvalue.
                        Integer_Value, Ope);
                  END IF;

               END IF;

            END LOOP;





         WHEN Simulation_Time_Unit_Array_Double =>


            IF (Ope=Equal) OR
                  (Ope=Not_Equal) OR
                  (Ope=Equal_Less) OR
                  (Ope=Equal_Greater) OR
                  (Ope=Inf) OR
                  (Ope=Sup) THEN
               Res:=NEW Simulation_Value(
                  Simulation_Time_Unit_Array_Boolean);
            ELSE
               Res:=NEW Simulation_Value(
                  Simulation_Time_Unit_Array_Double);
            END IF;

            FOR I IN Time_Unit_Range LOOP

               IF Rvalue.Ptype=Simulation_Time_Unit_Array_Double THEN
                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        Boolean_Time_Unit_Table_Value(I), Rvalue.
                        Boolean_Time_Unit_Table_Value(I),Ope);
                  ELSE
                     Res.Double_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        Double_Time_Unit_Table_Value(I), Rvalue.
                        Double_Time_Unit_Table_Value(
                           I),Ope);
                  END IF;
               ELSE
                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        Boolean_Time_Unit_Table_Value(I), Rvalue.
                        Boolean_Value, Ope);
                  ELSE
                     Res.Double_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        Double_Time_Unit_Table_Value(I), Rvalue.
                        Double_Value, Ope);
                  END IF;
               END IF;

            END LOOP;





         WHEN Simulation_Time_Unit_Array_Boolean =>


            Res:=NEW Simulation_Value(
               Simulation_Time_Unit_Array_Boolean);

            FOR I IN Time_Unit_Range LOOP

               IF Rvalue.Ptype=Simulation_Time_Unit_Array_Boolean THEN
                  Res.Boolean_Time_Unit_Table_Value(I):=Apply_Operator(
                     Lvalue.Boolean_Time_Unit_Table_Value(I),
                     Rvalue.Boolean_Time_Unit_Table_Value(I),Ope);
               ELSE
                  Res.Boolean_Time_Unit_Table_Value(I):=Apply_Operator(
                     Lvalue.
                     Boolean_Time_Unit_Table_Value(I),
                     Rvalue.Boolean_Value, Ope);
               END IF;

            END LOOP;





         WHEN Simulation_Time_Unit_Array_String =>


            IF (Ope=Equal) OR
                  (Ope=Not_Equal) OR
                  (Ope=Equal_Less) OR
                  (Ope=Equal_Greater) OR
                  (Ope=Inf) OR
                  (Ope=Sup) THEN
               Res:=NEW Simulation_Value(
                  Simulation_Time_Unit_Array_Boolean);
            ELSE
               Res:=NEW Simulation_Value(
                  Simulation_Time_Unit_Array_String);
            END IF;

            FOR I IN Time_Unit_Range LOOP

               IF Rvalue.Ptype=Simulation_Time_Unit_Array_String THEN
                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        Boolean_Time_Unit_Table_Value(I), Rvalue.
                        Boolean_Time_Unit_Table_Value(I),Ope);
                  ELSE
                     Res.String_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        String_Time_Unit_Table_Value(I), Rvalue.
                        String_Time_Unit_Table_Value(
                           I),Ope);
                  END IF;
               ELSE
                  IF (Ope=Equal) OR
                        (Ope=Not_Equal) OR
                        (Ope=Equal_Less) OR
                        (Ope=Equal_Greater) OR
                        (Ope=Inf) OR
                        (Ope=Sup) THEN
                     Res.Boolean_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        Boolean_Time_Unit_Table_Value(I), Rvalue.
                        Boolean_Value, Ope);
                  ELSE
                     Res.String_Time_Unit_Table_Value(I):=Apply_Operator(
                        Lvalue.
                        String_Time_Unit_Table_Value(I), Rvalue.
                        String_Value, Ope);
                  END IF;
               END IF;

            END LOOP;





      END CASE;

      RETURN Res;

   END Compute_Binary_Value;





   FUNCTION Apply_Operator (
         Value : Boolean;
         Ope   : Operator)
     RETURN Boolean IS
   BEGIN

      IF Ope=Logic_Not THEN
         RETURN NOT Value;
      END IF;

      Raise_Exception(Type_Error'Identity,
         To_String(Lb_Uncompatible_Type_Error(Current_Language) &
            Lb_Colon & To_Unbounded_String(Value'Img) &
            Lb_Comma &
            To_Unbounded_String(Ope'Img)) );
   END Apply_Operator;


   FUNCTION Apply_Operator (
         Value : Integer;
         Ope   : Operator)
     RETURN Integer IS
   BEGIN

      IF Ope=Abs_Operator THEN
         RETURN ABS Value;
      END IF;

      Raise_Exception(Type_Error'Identity,
         To_String(Lb_Uncompatible_Type_Error(Current_Language) &
            Lb_Colon & To_Unbounded_String(Value'Img) &
            Lb_Comma &
            To_Unbounded_String(Ope'Img)) );
   END Apply_Operator;


   FUNCTION Apply_Operator (
         Value : Double;
         Ope   : Operator)
     RETURN Double IS
   BEGIN

      IF Ope=Abs_Operator THEN
         RETURN ABS Value;
      END IF;

      Raise_Exception(Type_Error'Identity,
         To_String(Lb_Uncompatible_Type_Error(Current_Language) &
            Lb_Colon & To_Unbounded_String(Value'Img) &
            Lb_Comma &
            To_Unbounded_String(Ope'Img)) );
   END Apply_Operator;


   FUNCTION Apply_Operator (
         Value : Unbounded_String;
         V     : Variables_Table_Type;
         Ope   : Operator)
     RETURN Integer IS

      Var : Variables_Range;


   BEGIN

      IF Ope=Get_Task_Index THEN
         Var:=Find_Variable(V,
            To_Unbounded_String("tasks.name"));
         FOR I IN String_Table'RANGE LOOP
            IF V.Entries(Var).Simulation.
                  String_Table_Value(I) = Value
                  THEN
               RETURN Integer(I);
            END IF;
         END LOOP;
         Raise_Exception(Statement_Error'Identity,
            To_String(Lb_Task_Not_Found(Current_Language) &
               Lb_Colon & Value &
               Lb_Comma &
               To_Unbounded_String(Ope'Img)) );
      END IF;

      IF Ope=Get_Buffer_Index THEN
         Var:=Find_Variable(V,
            To_Unbounded_String("buffers.name"));
         FOR I IN String_Table'RANGE LOOP
            IF V.Entries(Var).Simulation.
                  String_Table_Value(I) = Value
                  THEN
               RETURN Integer(I);
            END IF;
         END LOOP;
         Raise_Exception(Statement_Error'Identity,
            To_String(Lb_Buffer_Not_Found(Current_Language) &
               Lb_Colon & Value &
               Lb_Comma &
               To_Unbounded_String(Ope'Img)) );
      END IF;


      IF Ope=Get_Message_Index THEN
         Var:=Find_Variable(V,
            To_Unbounded_String("messages.name"));
         FOR I IN String_Table'RANGE LOOP
            IF V.Entries(Var).Simulation.
                  String_Table_Value(I) = Value
                  THEN
               RETURN Integer(I);
            END IF;
         END LOOP;
         Raise_Exception(Statement_Error'Identity,
            To_String(Lb_Message_Not_Found(Current_Language) &
               Lb_Colon & Value &
               Lb_Comma &
               To_Unbounded_String(Ope'Img)) );
      END IF;

      IF Ope=Get_Resource_Index THEN
         Var:=Find_Variable(V,
            To_Unbounded_String("resources.name"));
         FOR I IN String_Table'RANGE LOOP
            IF V.Entries(Var).Simulation.
                  String_Table_Value(I) = Value
                  THEN
               RETURN Integer(I);
            END IF;
         END LOOP;
         Raise_Exception(Statement_Error'Identity,
            To_String(Lb_Resource_Not_Found(Current_Language) &
               Lb_Colon & Value &
               Lb_Comma &
               To_Unbounded_String(Ope'Img)) );
      END IF;

      Raise_Exception(Type_Error'Identity,
         To_String(Lb_Uncompatible_Type_Error(Current_Language) &
            Lb_Colon & Value &
            Lb_Comma &
            To_Unbounded_String(Ope'Img)) );
   END Apply_Operator;



   ---------------------------------------------------------









   FUNCTION Value_Of (
         C : Binary_Expr;
         V : Variables_Table_Type)
     RETURN Simulation_Value_Ptr IS

      Res,
      Lvalue,
      Rvalue : Simulation_Value_Ptr;


   BEGIN

      Lvalue:=Value_Of(C.Lvalue.All, V);
      Rvalue:=Value_Of(C.Rvalue.All, V);


      -- First case : it's a logical expression : the two values
      -- have to have the same type and the same size
      --
      -- Second case : the two values has the same type (boolean, integer, double) and the
      -- same size (scalar or vector)
      --
      -- Third case :  the two values has the same type but a data is scalar and
      -- the other is a vector
      --
      -- Last case : types are differents : it's an error !
      --

      IF (Rvalue.Ptype/=Lvalue.Ptype)  AND
            (
            (C.Ope = Equal)
            OR (C.Ope = Not_Equal)
            OR (C.Ope = Equal_Less)
            OR (C.Ope = Equal_Greater)
            OR (C.Ope = Logic_And)
            OR (C.Ope = Logic_Or)
            OR (C.Ope = Logic_Not)
            OR (C.Ope = Inf)
            OR (C.Ope = Sup)  )
            THEN
         Raise_Exception(Type_Error'Identity,
            To_String(Lb_Uncompatible_Type_Error(Current_Language) &
               Lb_Colon & To_Unbounded_String(C.Lvalue.All) &
               Lb_Comma &
               To_Unbounded_String(C.Rvalue.All)));
      END IF;


      IF (
            (Rvalue.Ptype=Lvalue.Ptype)  OR



            ( (Lvalue.Ptype=Simulation_Array_Integer) AND  (
                  Rvalue.Ptype=Simulation_Integer) ) OR

            ( (Lvalue.Ptype=Simulation_Array_Double) AND  (
                  Rvalue.Ptype=Simulation_Double) ) OR

            ( (Lvalue.Ptype=Simulation_Array_Boolean) AND  (
                  Rvalue.Ptype=Simulation_Boolean) ) OR

            ( (Lvalue.Ptype=Simulation_Array_String) AND  (
                  Rvalue.Ptype=Simulation_String) ) OR

            ( (Lvalue.Ptype=Simulation_Integer) AND  (Rvalue.
                  Ptype=Simulation_Array_Integer) )  OR

            ( (Lvalue.Ptype=Simulation_Double) AND  (Rvalue.
                  Ptype=Simulation_Array_Double) ) OR

            ( (Lvalue.Ptype=Simulation_String) AND  (Rvalue.
                  Ptype=Simulation_Array_String) ) OR

            ( (Lvalue.Ptype=Simulation_Boolean) AND  (Rvalue.
                  Ptype=Simulation_Array_Boolean)) OR




            ( (Lvalue.Ptype=Simulation_Time_Unit_Array_Integer) AND  (
                  Rvalue.Ptype=Simulation_Integer) ) OR

            ( (Lvalue.Ptype=Simulation_Time_Unit_Array_Double) AND  (
                  Rvalue.Ptype=Simulation_Double) ) OR

            ( (Lvalue.Ptype=Simulation_Time_Unit_Array_Boolean) AND  (
                  Rvalue.Ptype=Simulation_Boolean) ) OR

            ( (Lvalue.Ptype=Simulation_Time_Unit_Array_String) AND  (
                  Rvalue.Ptype=Simulation_String) ) OR

            ( (Lvalue.Ptype=Simulation_Integer) AND  (Rvalue.
                  Ptype=Simulation_Time_Unit_Array_Integer) )  OR

            ( (Lvalue.Ptype=Simulation_Double) AND  (Rvalue.
                  Ptype=Simulation_Time_Unit_Array_Double) ) OR

            ( (Lvalue.Ptype=Simulation_String) AND  (Rvalue.
                  Ptype=Simulation_Time_Unit_Array_String) ) OR

            ( (Lvalue.Ptype=Simulation_Boolean) AND  (Rvalue.
                  Ptype=Simulation_Time_Unit_Array_Boolean) )
            )

            THEN
         Res:=Compute_Binary_Value(Lvalue, Rvalue, C.Ope);

      ELSE
         Raise_Exception(Type_Error'Identity,
            To_String(Lb_Uncompatible_Type_Error(Current_Language) &
               Lb_Colon & To_Unbounded_String(C.Lvalue.All) &
               Lb_Comma &
               To_Unbounded_String(C.Rvalue.All)));
      END IF;

      Free(Lvalue);
      Free(Rvalue);

      RETURN Res;

   END Value_Of;



   FUNCTION Value_Of (
         C : Unary_Expr;
         V : Variables_Table_Type)
     RETURN Simulation_Value_Ptr IS

      Res,
      Value : Simulation_Value_Ptr;


   BEGIN


      Value:=Value_Of(C.Value.All,V);


      IF (Value.Ptype=Simulation_Array_String)
            OR (Value.Ptype=Simulation_Time_Unit_Array_String)
            THEN
         Raise_Exception(Type_Error'Identity,
            To_String(Lb_Uncompatible_Type_Error(Current_Language) &
               Lb_Colon & Value.Ptype'Img &
               Lb_Comma &
               To_Unbounded_String(C.Ope'Img)) );
      END IF;


      IF (C.Ope=Max_To_Index) OR (C.Ope=Min_To_Index)
            THEN
         RETURN Value;
      END IF;

      -- Apply operator from an unary expression
      --
      CASE Value.Ptype IS

         WHEN Simulation_String=>
            Res:=NEW Simulation_Value(
               Simulation_Integer);
            Res.Integer_Value:=Apply_Operator(Value.String_Value, V, C.
               Ope);

         WHEN Simulation_Array_String =>
            NULL;
         WHEN Simulation_Time_Unit_Array_String =>
            NULL;

         -- Nothing to compute in the cases of random type
         -- because value_of functions already do the job
         -- by returing integer values
         --
         WHEN Simulation_Random =>
            NULL;

         WHEN Simulation_Array_Random =>
            NULL;

         WHEN Simulation_Time_Unit_Array_Random =>
            NULL;



         WHEN Simulation_Time_Unit_Array_Integer =>
            Res:=NEW Simulation_Value(
               Simulation_Time_Unit_Array_Integer);
            FOR I IN Time_Unit_Range LOOP
               Res.Integer_Time_Unit_Table_Value(I):=
                  Apply_Operator(Value.Integer_Time_Unit_Table_Value(I),
                  C.
                  Ope);
            END LOOP;


         WHEN Simulation_Array_Integer =>
            Res:=NEW Simulation_Value(
               Simulation_Array_Integer);
            FOR I IN Simulations_Range LOOP
               Res.Integer_Table_Value(I):=
                  Apply_Operator(Value.Integer_Table_Value(I), C.Ope);
            END LOOP;

         WHEN Simulation_Integer =>
            Res:=NEW Simulation_Value(
               Simulation_Integer);
            Res.Integer_Value:=Apply_Operator(Value.Integer_Value, C.Ope);

         WHEN Simulation_Time_Unit_Array_Double =>
            Res:=NEW Simulation_Value(
               Simulation_Time_Unit_Array_Double);
            FOR I IN Time_Unit_Range LOOP
               Res.Double_Time_Unit_Table_Value(I):=
                  Apply_Operator(Value.Double_Time_Unit_Table_Value(I), C.
                  Ope);
            END LOOP;

         WHEN Simulation_Array_Double =>
            Res:=NEW Simulation_Value(
               Simulation_Array_Double);
            FOR I IN Simulations_Range LOOP
               Res.Double_Table_Value(I):=
                  Apply_Operator(Value.Double_Table_Value(I), C.Ope);
            END LOOP;

         WHEN Simulation_Double =>
            Res:=NEW Simulation_Value(
               Simulation_Double);
            Res.Double_Value:=Apply_Operator(Value.Double_Value, C.Ope);

         WHEN Simulation_Time_Unit_Array_Boolean =>
            Res:=NEW Simulation_Value(
               Simulation_Time_Unit_Array_Boolean);
            FOR I IN Time_Unit_Range LOOP
               Res.Boolean_Time_Unit_Table_Value(I):=
                  Apply_Operator(Value.Boolean_Time_Unit_Table_Value(I),
                  C.
                  Ope);
            END LOOP;

         WHEN Simulation_Array_Boolean =>
            Res:=NEW Simulation_Value(
               Simulation_Array_Boolean);
            FOR I IN Simulations_Range LOOP
               Res.Boolean_Table_Value(I):=
                  Apply_Operator(Value.Boolean_Table_Value(I), C.Ope);
            END LOOP;

         WHEN Simulation_Boolean =>
            Res:=NEW Simulation_Value(
               Simulation_Boolean);
            Res.Boolean_Value:=Apply_Operator(Value.Boolean_Value, C.Ope);
      END CASE;

      RETURN Res;

   END Value_Of;

   FUNCTION Value_Of (
         C : Constant_Expr;
         V : Variables_Table_Type)
     RETURN Simulation_Value_Ptr IS

   BEGIN
      RETURN Copy(C.Value);
   END Value_Of;


   FUNCTION Value_Of (
         C : Variable_Expr;
         V : Variables_Table_Type)
     RETURN Simulation_Value_Ptr IS

      New_Val : Simulation_Value_Ptr;
      Seed    : Generator;

   BEGIN
      FOR I IN 0..V.Nb_Entries-1 LOOP
         IF V.Entries(I).Variable.Name=C.Name
               THEN
            IF   (V.Entries(I).Simulation.Ptype = Simulation_Random)
                  THEN
               Reset(Seed, V.Entries(I).Simulation.Random_Value);
               New_Val:=NEW Simulation_Value(Simulation_Integer);
               New_Val.Integer_Value:=
                  Apply_Random_Operator(Seed, V.Entries(I).Simulation.
                  Random_Law_Parameters);
               Save(Seed, V.Entries(I).Simulation.Random_Value);
               RETURN New_Val;
            ELSE
               RETURN Copy(V.Entries(I).Simulation);
            END IF;
         END IF;
      END LOOP;

      Raise_Exception(Variable_Error'Identity,
         To_String(Lb_Undeclared_Identifier(Current_Language) & Lb_Colon &
            C.Name));
   END Value_Of;



   FUNCTION Value_Of (
         C : Array_Variable_Expr;
         V : Variables_Table_Type)
     RETURN Simulation_Value_Ptr IS

      New_Val : Simulation_Value_Ptr;
      Index   : Simulation_Value_Ptr;
      Seed    : Generator;

   BEGIN
      FOR I IN 0..V.Nb_Entries-1 LOOP
         IF V.Entries(I).Variable.Name=C.Name
               THEN

            -- Two cases : return the all task array table or
            --          if an index is provided, return
            --          a scalar data
            --

            IF (C.Array_Index/=NULL)


                  -- indexed expression : return scalar data
                  --
                  THEN

               Index:=Value_Of(C.Array_Index.All,V);
               IF Index.Ptype/=Simulation_Integer
                     THEN
                  Raise_Exception(Type_Error'Identity, To_String(
                        Lb_Index_Error(Current_Language) & Lb_Colon & C.
                        Name));
               END IF;

               IF C.Variable_Type=Simulation_Array_Integer THEN
                  New_Val:=NEW Simulation_Value(Simulation_Integer);
                  New_Val.Integer_Value:=V.Entries(I).Simulation.
                     Integer_Table_Value(Simulations_Range(Index.
                        Integer_Value));
               END IF;
               IF C.Variable_Type=Simulation_Array_Random THEN
                  Reset(Seed, V.Entries(I).Simulation.Random_Table_Value(
                        Simulations_Range(Index.Integer_Value)));
                  New_Val:=NEW Simulation_Value(Simulation_Integer);
                  New_Val.Integer_Value:=
                     Apply_Random_Operator(Seed, V.Entries(I).Simulation.
                     Random_Table_Law_Parameters);
                  Save(Seed, V.Entries(I).Simulation.Random_Table_Value(
                        Simulations_Range(Index.Integer_Value)));
               END IF;
               IF C.Variable_Type=Simulation_Array_Double THEN
                  New_Val:=NEW Simulation_Value(Simulation_Double);
                  New_Val.Double_Value:=V.Entries(I).Simulation.
                     Double_Table_Value(Simulations_Range(Index.
                        Integer_Value));
               END IF;
               IF C.Variable_Type=Simulation_Array_Boolean THEN
                  New_Val:=NEW Simulation_Value(Simulation_Boolean);
                  New_Val.Boolean_Value:=V.Entries(I).Simulation.
                     Boolean_Table_Value(Simulations_Range(Index.
                        Integer_Value));
               END IF;
               IF C.Variable_Type=Simulation_Array_String THEN
                  New_Val:=NEW Simulation_Value(Simulation_String);
                  New_Val.String_Value:=V.Entries(I).Simulation.
                     String_Table_Value(Simulations_Range(Index.
                        Integer_Value));
               END IF;
               IF C.Variable_Type=Simulation_Time_Unit_Array_Integer THEN
                  New_Val:=NEW Simulation_Value(Simulation_Integer);
                  New_Val.Integer_Value:=V.Entries(I).Simulation.
                     Integer_Time_Unit_Table_Value(Time_Unit_Range(Index.
                        Integer_Value));
               END IF;
               IF C.Variable_Type=Simulation_Time_Unit_Array_Random THEN
                  Reset(Seed, V.Entries(I).
                     Simulation.Random_Time_Unit_Table_Value(
                        Time_Unit_Range(Index.Integer_Value)));
                  New_Val:=NEW Simulation_Value(Simulation_Integer);
                  New_Val.Integer_Value:=
                     Apply_Random_Operator(Seed, V.Entries(I).Simulation.
                     Random_Time_Unit_Table_Law_Parameters);
                  Save(Seed, V.Entries(I).Simulation.
                     Random_Time_Unit_Table_Value(
                        Time_Unit_Range(Index.Integer_Value)));
               END IF;
               IF C.Variable_Type=Simulation_Time_Unit_Array_Double THEN
                  New_Val:=NEW Simulation_Value(Simulation_Double);
                  New_Val.Double_Value:=V.Entries(I).Simulation.
                     Double_Time_Unit_Table_Value(Time_Unit_Range(Index.
                        Integer_Value));
               END IF;
               IF C.Variable_Type=Simulation_Time_Unit_Array_Boolean THEN
                  New_Val:=NEW Simulation_Value(Simulation_Boolean);
                  New_Val.Boolean_Value:=V.Entries(I).Simulation.
                     Boolean_Time_Unit_Table_Value(Time_Unit_Range(Index.
                        Integer_Value));
               END IF;
               IF C.Variable_Type=Simulation_Time_Unit_Array_String THEN
                  New_Val:=NEW Simulation_Value(Simulation_String);
                  New_Val.String_Value:=V.Entries(I).Simulation.
                     String_Time_Unit_Table_Value(Time_Unit_Range(Index.
                        Integer_Value));
               END IF;

               Free(Index);

               RETURN New_Val;


            ELSE
               -- Array tables.
               --

               IF C.Variable_Type=Simulation_Array_Random THEN
                  New_Val:=NEW Simulation_Value(Simulation_Array_Integer);
                  FOR J IN Simulations_Range LOOP
                     Reset(Seed, V.Entries(I).
                        Simulation.Random_Table_Value(J));
                     New_Val.Integer_Table_Value(J):=
                        Apply_Random_Operator(Seed, V.Entries(I).
                        Simulation.Random_Table_Law_Parameters);
                     Save(Seed, V.Entries(I).
                        Simulation.Random_Table_Value(J));
                  END LOOP;
               END IF;

               IF C.Variable_Type=Simulation_Array_Integer THEN
                  New_Val:=NEW Simulation_Value(
                     Simulation_Array_Integer);
                  New_Val.Integer_Table_Value:=V.Entries(I).Simulation.
                     Integer_Table_Value;
               END IF;

               IF C.Variable_Type=Simulation_Array_Double THEN
                  New_Val:=NEW Simulation_Value(
                     Simulation_Array_Double);
                  New_Val.Double_Table_Value:=V.Entries(I).Simulation.
                     Double_Table_Value;
               END IF;

               IF C.Variable_Type=Simulation_Array_Boolean THEN
                  New_Val:=NEW Simulation_Value(
                     Simulation_Array_Boolean);
                  New_Val.Boolean_Table_Value:=V.Entries(I).Simulation.
                     Boolean_Table_Value;
               END IF;

               IF C.Variable_Type=Simulation_Array_String THEN
                  New_Val:=NEW Simulation_Value(
                     Simulation_Array_String);
                  New_Val.String_Table_Value:=V.Entries(I).Simulation.
                     String_Table_Value;
               END IF;

               IF C.Variable_Type=Simulation_Time_Unit_Array_Random THEN
                  New_Val:=NEW Simulation_Value(
                     Simulation_Time_Unit_Array_Integer);
                  FOR J IN Time_Unit_Range LOOP
                     Reset(Seed, V.Entries(I).
                        Simulation.Random_Time_Unit_Table_Value(J));
                     New_Val.Integer_Time_Unit_Table_Value(J):=
                        Apply_Random_Operator(Seed, V.Entries(I).
                        Simulation.Random_Time_Unit_Table_Law_Parameters);
                     Save(Seed, V.Entries(I).
                        Simulation.Random_Time_Unit_Table_Value(J));
                  END LOOP;
               END IF;

               IF C.Variable_Type=Simulation_Time_Unit_Array_Integer THEN
                  New_Val:=NEW Simulation_Value(
                     Simulation_Time_Unit_Array_Integer);
                  New_Val.Integer_Time_Unit_Table_Value:=V.Entries(I).
                     Simulation.
                     Integer_Time_Unit_Table_Value;
               END IF;

               IF C.Variable_Type=Simulation_Time_Unit_Array_Double THEN
                  New_Val:=NEW Simulation_Value(
                     Simulation_Time_Unit_Array_Double);
                  New_Val.Double_Time_Unit_Table_Value:=V.Entries(I).
                     Simulation.
                     Double_Time_Unit_Table_Value;
               END IF;

               IF C.Variable_Type=Simulation_Time_Unit_Array_Boolean THEN
                  New_Val:=NEW Simulation_Value(
                     Simulation_Time_Unit_Array_Boolean);
                  New_Val.Boolean_Time_Unit_Table_Value:=V.Entries(I).
                     Simulation.
                     Boolean_Time_Unit_Table_Value;
               END IF;

               IF C.Variable_Type=Simulation_Time_Unit_Array_String THEN
                  New_Val:=NEW Simulation_Value(
                     Simulation_Time_Unit_Array_String);
                  New_Val.String_Time_Unit_Table_Value:=V.Entries(I).
                     Simulation.
                     String_Time_Unit_Table_Value;
               END IF;


               RETURN New_Val;
            END IF;


         END IF;

      END LOOP;


      Raise_Exception(Variable_Error'Identity,
         To_String(Lb_Undeclared_Identifier(Current_Language) & Lb_Colon &
            C.Name));
   END Value_Of;







END Expressions;






