------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Message_Set;
USE Message_Set;
USE Message_Set.Generic_Message_Set;
WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;
WITH Resource_Set;
USE Resource_Set;
USE Resource_Set.Generic_Resource_Set;


WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Framework_Config;
USE Framework_Config;

WITH Simulations;
USE Simulations;

WITH Tables;
WITH Ada.Finalization;





PACKAGE Expressions IS




   -- The data type does not match the expected type
   --
   Type_Error : EXCEPTION;

   -- Statement exception (sush as unexpected statement)
   --
   Statement_Error : EXCEPTION;

   -- Exception raised when an error occurs
   -- on variable  operations
   --
   Variable_Error : EXCEPTION;

   -- Syntax error detected
   --
   Syntax_Error : EXCEPTION;

   -- This exception is raised when a parametric file can not be read
   --
   Parametric_File_Error : EXCEPTION;





   ---------------------------------------------------------
   -- Valid operator used to build expressions
   --
   TYPE Operator IS
         (Plus,
          Minus,
          Div,
          Mult,
          Exp,
          Modulo,
          Equal,
          Not_Equal,
          Equal_Less,
          Equal_Greater,
          Inf,
          Sup,
          Logic_And,
          Logic_Or,
          Logic_Not,
          Min_Operator,
          Max_Operator,
          Min_To_Index,
          Max_To_Index,
          Lcm,
          Abs_Operator,
          Get_Resource_Index,
          Get_Task_Index,
          Get_Message_Index,
          Get_Buffer_Index,
          Concat);

   PROCEDURE Put (
         S : IN     Operator);


   -----------------------------------------------------------


   -- A predefinition of the variable type
   --
   TYPE Variable_Expr;
   TYPE Variable_Expr_Ptr IS ACCESS ALL Variable_Expr'Class;

   ---------------------------------------------------------
   -- Table to store the list of variables of
   -- a parametric scheduler. For each variable,
   -- the current simulation values are also stored
   --

   TYPE Variable_Record IS
      RECORD
         Variable   : Variable_Expr_Ptr;
         Simulation : Simulation_Value_Ptr;
      END RECORD;
   TYPE Variable_Record_Ptr IS ACCESS Variable_Record;


   PROCEDURE Put (
         V : IN     Variable_Record_Ptr);



   PACKAGE Variables_Type_Package IS NEW Tables(Variable_Record_Ptr,
      Max_Variables,Put);
   USE Variables_Type_Package;

   SUBTYPE Variables_Range IS Variables_Type_Package.Table_Range;
   SUBTYPE Variables_Table_Type IS Variables_Type_Package.Table;



   -- Looking for a variable in the variable table
   --
   FUNCTION Find_Variable (
         Var_Table : IN     Variables_Table_Type;
         S         : IN     Unbounded_String;
         Line      : IN     Natural              := 0)
     RETURN Variables_Range;

   FUNCTION Find_Variable (
         Var_Table : IN     Variables_Table_Type;
         S         : IN     String;
         Line      : IN     Natural              := 0)
     RETURN Variables_Range;


   -- Check is a given variable is already defined in a
   -- variable table
   --
   PROCEDURE Check_Variable_Declaration (
         Var_Table : IN     Variables_Table_Type;
         S         : IN     Unbounded_String;
         Line      : IN     Natural              := 0);



   -- Load into the parametric simulator
   -- data from task, processor, message and buffer elements
   --
   -- The number of object (tasks, resources, buffers) is initialized
   -- accordig to the correspondig processor (parameter processor_name)
   -- if processor_name is an empty string, all tasks/buffers/resources
   -- of all procesors are considered
   --
   PROCEDURE Initialize_Parametric_Variables (
         Var_Table      : IN OUT Variables_Table_Type;
         My_Messages    : IN     Messages_Set;
         My_Buffers     : IN     Buffers_Set;
         My_Resources   : IN     Resources_Set;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String     := Empty_String);


   -- Add the parametric variable in a variable table
   --
   PROCEDURE Create_Parametric_Variables (
         Variables_Table : IN OUT Variables_Table_Type;
         My_Tasks        : IN     Tasks_Set);




   ---------------------------------------------------------

   TYPE Expressions_Type IS
         (Constant_Type,
          Variable_Type,
          Array_Variable_Type,
          Binary_Type,
          Unary_Type,
          Random_Type);


   TYPE Generic_Expression IS ABSTRACT NEW Ada.Finalization.Controlled WITH
      RECORD
         Expression_Type : Expressions_Type;
      END RECORD;
   TYPE Expression_Ptr IS ACCESS ALL Generic_Expression'Class;


   PROCEDURE Put (
         S : IN     Generic_Expression) IS
   ABSTRACT;

   PROCEDURE Put (
         S : IN     Expression_Ptr) IS
   ABSTRACT;

   FUNCTION To_Unbounded_String (
         C : Generic_Expression)
     RETURN Unbounded_String IS
   ABSTRACT;

   FUNCTION Get_Type (
         C : Generic_Expression)
     RETURN Simulation_Type IS
   ABSTRACT;

   FUNCTION Value_Of (
         C : Generic_Expression;
         V : Variables_Table_Type)
     RETURN Simulation_Value_Ptr IS
   ABSTRACT;



   ------------------------------------------

   TYPE Constant_Expr IS NEW Generic_Expression WITH
      RECORD
         Value : Simulation_Value_Ptr;
      END RECORD;

   TYPE Constant_Expr_Ptr IS ACCESS ALL Constant_Expr'Class;

   PROCEDURE Put (
         S : IN     Constant_Expr);
   PROCEDURE Put (
         S : IN     Constant_Expr_Ptr);


   FUNCTION To_Unbounded_String (
         C : Constant_Expr)
     RETURN Unbounded_String;

   FUNCTION Get_Type (
         C : Constant_Expr)
     RETURN Simulation_Type;

   PROCEDURE Initialize (
         S : IN OUT Constant_Expr);

   FUNCTION Value_Of (
         C : Constant_Expr;
         V : Variables_Table_Type)
     RETURN Simulation_Value_Ptr;



   ------------------------------------------


   TYPE Variable_Expr IS NEW Generic_Expression WITH
      RECORD

         -- Name of the variable
         --
         Name : Unbounded_String;

         -- Type of the variable (integer, boolean, ...)
         --
         Variable_Type : Simulation_Type;
      END RECORD;


   PROCEDURE Put (
         S : IN     Variable_Expr);
   PROCEDURE Put (
         S : IN     Variable_Expr_Ptr);

   FUNCTION To_Unbounded_String (
         C : Variable_Expr)
     RETURN Unbounded_String;

   FUNCTION Get_Type (
         C : Variable_Expr)
     RETURN Simulation_Type;

   PROCEDURE Initialize (
         S : IN OUT Variable_Expr);

   FUNCTION Value_Of (
         C : Variable_Expr;
         V : Variables_Table_Type)
     RETURN Simulation_Value_Ptr;



   ------------------------------------------

   TYPE Array_Variable_Expr IS NEW Variable_Expr WITH
      RECORD
         Array_Index : Expression_Ptr;
      END RECORD;

   TYPE Array_Variable_Expr_Ptr IS ACCESS ALL Array_Variable_Expr'Class;

   PROCEDURE Put (
         S : IN     Array_Variable_Expr);
   PROCEDURE Put (
         S : IN     Array_Variable_Expr_Ptr);

   FUNCTION To_Unbounded_String (
         C : Array_Variable_Expr)
     RETURN Unbounded_String;

   PROCEDURE Initialize (
         S : IN OUT Array_Variable_Expr);

   FUNCTION Value_Of (
         C : Array_Variable_Expr;
         V : Variables_Table_Type)
     RETURN Simulation_Value_Ptr;



   ------------------------------------------

   TYPE Binary_Expr IS NEW Generic_Expression WITH
      RECORD
         Ope    : Operator;
         Rvalue : Expression_Ptr;
         Lvalue : Expression_Ptr;
      END RECORD;
   TYPE Binary_Expr_Ptr IS ACCESS ALL Binary_Expr'Class;

   PROCEDURE Put (
         S : IN     Binary_Expr);
   PROCEDURE Put (
         S : IN     Binary_Expr_Ptr);

   FUNCTION To_Unbounded_String (
         C : Binary_Expr)
     RETURN Unbounded_String;

   FUNCTION Get_Type (
         C : Binary_Expr)
     RETURN Simulation_Type;

   PROCEDURE Initialize (
         S : IN OUT Binary_Expr);

   FUNCTION Value_Of (
         C : Binary_Expr;
         V : Variables_Table_Type)
     RETURN Simulation_Value_Ptr;


   ------------------------------------------

   TYPE Unary_Expr IS NEW Generic_Expression WITH
      RECORD
         Ope   : Operator;
         Value : Expression_Ptr;
      END RECORD;
   TYPE Unary_Expr_Ptr IS ACCESS ALL Unary_Expr'Class;

   PROCEDURE Put (
         S : IN     Unary_Expr);
   PROCEDURE Put (
         S : IN     Unary_Expr_Ptr);

   FUNCTION To_Unbounded_String (
         C : Unary_Expr)
     RETURN Unbounded_String;

   FUNCTION Get_Type (
         C : Unary_Expr)
     RETURN Simulation_Type;

   PROCEDURE Initialize (
         S : IN OUT Unary_Expr);

   FUNCTION Value_Of (
         C : Unary_Expr;
         V : Variables_Table_Type)
     RETURN Simulation_Value_Ptr;






END Expressions;


