------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------




WITH Framework_Config;
USE Framework_Config;

WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;
WITH Ada.Numerics.Float_Random;
USE Ada.Numerics.Float_Random;


WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;

WITH Time_Units;
USE Time_Units;

WITH Unchecked_Deallocation;



PACKAGE Simulations IS




   -------------------------------------------------------------
   -- Random Law used to built random expressions
   --
   TYPE Law_Type IS
         (Uniform_Law,
          Exponential_Law);

   PROCEDURE Put (
         S : IN     Law_Type);




   ------------------------------------------------------------------
   -- Type of data stored during a scheduling simulation
   ------------------------------------------------------------------

   TYPE Simulations_Range IS NEW Integer RANGE 0 .. Max_Simulation;
   TYPE Sum_Simulations_Range IS NEW Integer RANGE 0 .. Sum_Simulation;


   TYPE Random_Law_Parameters_Type IS
      RECORD
         Law        : Law_Type;
         Parameter1 : Integer;
         Parameter2 : Integer;
      END RECORD;


   TYPE Boolean_Table IS ARRAY (Simulations_Range) OF Boolean;
   TYPE Integer_Table IS ARRAY (Simulations_Range) OF Integer;
   TYPE Double_Table IS ARRAY (Simulations_Range) OF Double;
   TYPE String_Table IS ARRAY (Simulations_Range) OF Unbounded_String;
   TYPE Random_Table IS ARRAY (Simulations_Range) OF State;

   TYPE Boolean_Time_Unit_Table IS ARRAY (Time_Unit_Range) OF Boolean;
   TYPE Integer_Time_Unit_Table IS ARRAY (Time_Unit_Range) OF Integer;
   TYPE Double_Time_Unit_Table IS ARRAY (Time_Unit_Range) OF Double;
   TYPE String_Time_Unit_Table IS ARRAY (Time_Unit_Range) OF Unbounded_String;
   TYPE Random_Time_Unit_Table IS ARRAY (Time_Unit_Range) OF State;


   TYPE Simulation_Type IS
         (Simulation_Boolean,
          Simulation_Integer,
          Simulation_Double,
          Simulation_Random,
          Simulation_String,
          Simulation_Array_Boolean,
          Simulation_Array_Integer,
          Simulation_Array_Double,
          Simulation_Array_Random,
          Simulation_Array_String,
          Simulation_Time_Unit_Array_Boolean,
          Simulation_Time_Unit_Array_Integer,
          Simulation_Time_Unit_Array_Double,
          Simulation_Time_Unit_Array_Random,
          Simulation_Time_Unit_Array_String);



   TYPE Simulation_Value
      (Ptype : Simulation_Type) IS
   RECORD
      CASE Ptype IS

         WHEN Simulation_Boolean =>
            Boolean_Value : Boolean := True;
         WHEN Simulation_Integer=>
            Integer_Value : Integer := 0;
         WHEN Simulation_Double=>
            Double_Value : Double := 0.0;
         WHEN Simulation_Random=>
            Random_Value          : State;
            Random_Law_Parameters : Random_Law_Parameters_Type;
         WHEN Simulation_String=>
            String_Value : Unbounded_String;


         WHEN Simulation_Array_Boolean=>
            Boolean_Table_Value : Boolean_Table := (OTHERS => True);
         WHEN Simulation_Array_Integer=>
            Integer_Table_Value : Integer_Table := (OTHERS => 0);
         WHEN Simulation_Array_Double=>
            Double_Table_Value : Double_Table := (OTHERS => 0.0);
         WHEN Simulation_Array_Random=>
            Random_Table_Value          : Random_Table;
            Random_Table_Law_Parameters : Random_Law_Parameters_Type;
         WHEN Simulation_Array_String=>
            String_Table_Value : String_Table;


         WHEN Simulation_Time_Unit_Array_Boolean=>
            Boolean_Time_Unit_Table_Value : Boolean_Time_Unit_Table := (OTHERS =>
               True);
         WHEN Simulation_Time_Unit_Array_Integer=>
            Integer_Time_Unit_Table_Value : Integer_Time_Unit_Table := (OTHERS =>
               0);
         WHEN Simulation_Time_Unit_Array_Double=>
            Double_Time_Unit_Table_Value : Double_Time_Unit_Table := (OTHERS =>
               0.0);
         WHEN Simulation_Time_Unit_Array_Random=>
            Random_Time_Unit_Table_Value          : Random_Time_Unit_Table;
            Random_Time_Unit_Table_Law_Parameters :
               Random_Law_Parameters_Type;
         WHEN Simulation_Time_Unit_Array_String=>
            String_Time_Unit_Table_Value : String_Time_Unit_Table;


      END CASE;
   END RECORD;

   TYPE Simulation_Value_Ptr IS ACCESS Simulation_Value;



   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Simulation_Value, Simulation_Value_Ptr);




   PROCEDURE Put (
         S : Simulation_Value_Ptr);


   FUNCTION To_Unbounded_String (
         S    : IN     Simulation_Value_Ptr;
         From,
         To   : IN     Natural)
     RETURN Unbounded_String;


   FUNCTION To_Unbounded_String (
         S : IN     Simulation_Value_Ptr)
     RETURN Unbounded_String;


   FUNCTION Copy (
         My_Simulation : IN     Simulation_Value_Ptr)
     RETURN Simulation_Value_Ptr;




END Simulations;



