------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;

WITH Expressions;
USE Expressions;

WITH Ada.Finalization;
WITH Tables;


WITH Framework_Config;
USE Framework_Config;

WITH Simulations;
USE Simulations;




PACKAGE Statements IS



   ------------------------------------------------------------------------
   -- Definition of all kind of statement we can find in parametric code
   ------------------------------------------------------------------------

   TYPE Table_Types IS
         (Task_Table_Type,
          Message_Table_Type,
          Processor_Table_Type,
          Buffer_Table_Type,
          Time_Unit_Table_Type,
          Resource_Table_Type);



   TYPE Generic_Statement;
   TYPE Statement_Ptr IS ACCESS ALL Generic_Statement'Class;


   TYPE Statements_Type IS
         (Put_Statement_Type,
          Nop_Statement_Type,
          If_Statement_Type,
          Affectation_Statement_Type,
          For_Statement_Type,
          Return_Statement_Type,
          While_Statement_Type,
          Set_Statement_Type,
          Subprogram_Statement_Type,
          Subprogram_Call_Statement_Type,
          Random_Initialize_Statement_Type);



   TYPE Generic_Statement IS ABSTRACT NEW Ada.Finalization.Controlled WITH
      RECORD
         Statement_Type : Statements_Type;
         Line_Number    : Natural         := 1;
         Next_Statement : Statement_Ptr   := NULL;
      END RECORD;


   PROCEDURE Put (
         S : IN     Generic_Statement);
   PROCEDURE Recursive_Put (
         S : IN     Statement_Ptr);


   FUNCTION Copy (
         S : IN     Statement_Ptr)
     RETURN Statement_Ptr;


   ------------------------------------------

   TYPE Nop_Statement IS NEW Generic_Statement WITH NULL RECORD;

   TYPE Nop_Statement_Ptr IS ACCESS ALL Nop_Statement'Class;

   PROCEDURE Initialize (
         S : IN OUT Nop_Statement);



   PROCEDURE Put (
         S : IN     Nop_Statement);



   ------------------------------------------

   TYPE Put_Statement IS NEW Generic_Statement WITH
      RECORD
         Expression_To_Be_Displayed : Expression_Ptr;
         Put_From                   : Expression_Ptr;
         Put_To                     : Expression_Ptr;
      END RECORD;

   TYPE Put_Statement_Ptr IS ACCESS ALL Put_Statement'Class;

   PROCEDURE Initialize (
         S : IN OUT Put_Statement);


   PROCEDURE Put (
         S : IN     Put_Statement);


   ------------------------------------------

   TYPE If_Statement IS NEW Generic_Statement WITH
      RECORD
         Bool_Expression : Expression_Ptr;
         Else_Statement  : Statement_Ptr;
         Then_Statement  : Statement_Ptr;
      END RECORD;

   TYPE If_Statement_Ptr IS ACCESS ALL If_Statement'Class;

   PROCEDURE Initialize (
         S : IN OUT If_Statement);


   PROCEDURE Put (
         S : IN     If_Statement);

   ------------------------------------------

   TYPE Affectation_Statement IS NEW Generic_Statement WITH
      RECORD
         Lvalue : Expression_Ptr;
         Rvalue : Expression_Ptr;
      END RECORD;

   TYPE Affectation_Statement_Ptr IS ACCESS ALL Affectation_Statement'Class;

   PROCEDURE Initialize (
         S : IN OUT Affectation_Statement);


   PROCEDURE Put (
         S : IN     Affectation_Statement);

   ------------------------------------------


   TYPE For_Statement IS NEW Generic_Statement WITH
      RECORD
         For_Type           : Table_Types;
         Included_Statement : Statement_Ptr;
         For_Index          : Variable_Expr_Ptr;
      END RECORD;

   TYPE For_Statement_Ptr IS ACCESS ALL For_Statement'Class;

   PROCEDURE Initialize (
         S : IN OUT For_Statement);


   PROCEDURE Put (
         S : IN     For_Statement);

   ------------------------------------------

   TYPE Return_Statement IS NEW Generic_Statement WITH
      RECORD
         Return_Value : Expression_Ptr;
      END RECORD;

   TYPE Return_Statement_Ptr IS ACCESS ALL Return_Statement'Class;

   PROCEDURE Initialize (
         S : IN OUT Return_Statement);

   PROCEDURE Put (
         S : IN     Return_Statement);


   ------------------------------------------

   TYPE While_Statement IS NEW Generic_Statement WITH
      RECORD
         Included_Statement : Statement_Ptr;
         Condition          : Expression_Ptr;
      END RECORD;

   TYPE While_Statement_Ptr IS ACCESS ALL While_Statement'Class;

   PROCEDURE Initialize (
         S : IN OUT While_Statement);

   PROCEDURE Put (
         S : IN     While_Statement);


   ------------------------------------------

   TYPE Random_Initialize_Statement IS NEW Generic_Statement WITH
      RECORD
         Lvalue     : Unbounded_String;
         Law        : Law_Type;
         Parameter1 : Expression_Ptr;
         Parameter2 : Expression_Ptr;
      END RECORD;

   TYPE Random_Initialize_Statement_Ptr IS ACCESS ALL Random_Initialize_Statement'Class;

   PROCEDURE Initialize (
         S : IN OUT Random_Initialize_Statement);

   PROCEDURE Put (
         S : IN     Random_Initialize_Statement);

   ------------------------------------------

   TYPE Set_Statement IS NEW Generic_Statement WITH
      RECORD
         Set_Value : Expression_Ptr;
         Set_Id    : Unbounded_String;
      END RECORD;

   TYPE Set_Statement_Ptr IS ACCESS ALL Set_Statement'Class;

   PROCEDURE Initialize (
         S : IN OUT Set_Statement);

   PROCEDURE Put (
         S : IN     Set_Statement);

   PROCEDURE Put (
         S : IN     Set_Statement_Ptr);


   ------------------------------------------

   TYPE Subprogram_Statement IS NEW Generic_Statement WITH
      RECORD
         Included_Statement : Statement_Ptr;
         Is_A_Function      : Boolean;
         Subprogram_Name    : Unbounded_String;
      END RECORD;

   TYPE Subprogram_Statement_Ptr IS ACCESS ALL Subprogram_Statement'Class;

   PROCEDURE Initialize (
         S : IN OUT Subprogram_Statement);

   PROCEDURE Put (
         S : IN     Subprogram_Statement);


   ------------------------------------------

   TYPE Subprogram_Call_Statement IS NEW Generic_Statement WITH
      RECORD
         Is_A_Function     : Boolean;
         Called_Subprogram : Statement_Ptr;
         Return_Value      : Expression_Ptr;
      END RECORD;

   TYPE Subprogram_Call_Statement_Ptr IS ACCESS ALL Subprogram_Call_Statement'Class;

   PROCEDURE Initialize (
         S : IN OUT Subprogram_Call_Statement);

   PROCEDURE Put (
         S : IN     Subprogram_Call_Statement);


   ---------------------------------------------------------
   -- Table to store the list of sets statements
   ---------------------------------------------------------

   PACKAGE Sets_Type_Package IS NEW Tables(Set_Statement_Ptr,
      Max_Statements,Put);
   USE Sets_Type_Package;

   SUBTYPE Sets_Range IS Sets_Type_Package.Table_Range;
   SUBTYPE Sets_Table_Type IS Sets_Type_Package.Table;



   -- Looking for a set statement in the table
   --
   FUNCTION Find_Set (
         Var : IN     Sets_Table_Type;
         S   : IN     Unbounded_String)
     RETURN Sets_Range;



   PROCEDURE Check_Set_Declaration (
         Var : IN     Sets_Table_Type;
         S   : IN     Unbounded_String);



   ------------------------------------------------------------------------
   -- Parametric codes are sorted in different sections. Each section is associated
   -- to a parametric functionnality
   ------------------------------------------------------------------------

   TYPE Sections IS
         (Start,
          Priority,
          Election,
          Activation,
          Check_Resource,
          Allocate_Resource,
          Release_Resource,
          Gather_Event_Analyzer,
          Display_Event_Analyzer,
          Subprogram);

   TYPE Section_Pointers IS ARRAY (Sections'RANGE) OF Statement_Ptr;



   ---------------------------------------------------------------------
   --- Functions and procedure in this part of the package run
   --  parametric scheduler statements
   ----------------------------------------------------------------------


   PROCEDURE Dispatch (
         Current         : IN     Statement_Ptr;
         Current_Section : IN     Sections;
         Variables_Table : IN OUT Variables_Table_Type;
         Msg             : IN OUT Unbounded_String);


   PROCEDURE If_Dispatch (
         Current         : IN     If_Statement_Ptr;
         Current_Section : IN     Sections;
         Variables_Table : IN OUT Variables_Table_Type;
         Msg             : IN OUT Unbounded_String;
         Next            : IN OUT Statement_Ptr);


   PROCEDURE Affectation_Dispatch (
         Current         : IN     Affectation_Statement_Ptr;
         Variables_Table : IN OUT Variables_Table_Type;
         Msg             : IN OUT Unbounded_String;
         Next            : IN OUT Statement_Ptr);


   PROCEDURE Random_Initialize_Dispatch (
         Current         : IN     Random_Initialize_Statement_Ptr;
         Variables_Table : IN OUT Variables_Table_Type;
         Msg             : IN OUT Unbounded_String;
         Next            : IN OUT Statement_Ptr);


   PROCEDURE Put_Dispatch (
         Current         : IN     Put_Statement_Ptr;
         Variables_Table : IN OUT Variables_Table_Type;
         Msg             : IN OUT Unbounded_String;
         Next            : IN OUT Statement_Ptr);


   PROCEDURE While_Dispatch (
         Current         : IN     While_Statement_Ptr;
         Current_Section : IN     Sections;
         Variables_Table : IN OUT Variables_Table_Type;
         Msg             : IN OUT Unbounded_String;
         Next            : IN OUT Statement_Ptr);


   PROCEDURE For_Dispatch (
         Current         : IN     For_Statement_Ptr;
         Current_Section : IN     Sections;
         Variables_Table : IN OUT Variables_Table_Type;
         Msg             : IN OUT Unbounded_String;
         Next            : IN OUT Statement_Ptr);




END Statements;


