------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Ada.Exceptions;
USE Ada.Exceptions;


WITH Scheduler.Parametrics;
USE Scheduler.Parametrics;
WITH Scheduler.Fixed_Priority.Dm;
USE Scheduler.Fixed_Priority.Dm;
WITH Scheduler.Fixed_Priority.Rm;
USE Scheduler.Fixed_Priority.Rm;
WITH Scheduler.Fixed_Priority.Hpf;
USE Scheduler.Fixed_Priority.Hpf;
WITH Scheduler.Dynamic_Priority.Edf;
USE Scheduler.Dynamic_Priority.Edf;
WITH Scheduler.Dynamic_Priority.Llf;
USE Scheduler.Dynamic_Priority.Llf;
WITH Scheduler.Dynamic_Priority.Muf.Deadline_Based;
USE Scheduler.Dynamic_Priority.Muf.Deadline_Based;
WITH Scheduler.Dynamic_Priority.Muf.Laxity_Based;
USE Scheduler.Dynamic_Priority.Muf.Laxity_Based;
WITH Scheduler.Dynamic_Priority.D_Over;
USE Scheduler.Dynamic_Priority.D_Over;
WITH Scheduler.Time_Sharing_Based_On_Wait_Time;
USE Scheduler.Time_Sharing_Based_On_Wait_Time;
WITH Scheduler.Time_Sharing_Based_On_Cpu_Usage;
USE Scheduler.Time_Sharing_Based_On_Cpu_Usage;
WITH Scheduler.Round_Robin;
USE Scheduler.Round_Robin;
WITH Scheduler.Hierarchical;
USE Scheduler.Hierarchical;
WITH Scheduler.Hierarchical.Parametric;
USE Scheduler.Hierarchical.Parametric;
WITH Scheduler.Hierarchical.Round_Robin;
USE Scheduler.Hierarchical.Round_Robin;



WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Translate;
USE Translate;

WITH Objects;
USE Objects;



PACKAGE BODY Processor_Set IS





   PROCEDURE Check_Processor (
         My_Processors : IN     Processors_Set;
         Name          : IN     Unbounded_String;
         Is_Preemptive : IN     Preemptive_Type;
         Quantum       : IN     Integer;
         File_Name     : IN     Unbounded_String;
         A_Scheduler   : IN     Scheduler_Type;
         A_Network     : IN     Unbounded_String) IS


   BEGIN

      IF (Name = "")
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Processor_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;


      IF NOT Is_A_Valid_Identifier(Name)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Processor(Current_Language) & Name &
               " : " & Lb_Processor_Name(Current_Language) &
               Lb_Colon & Lb_Invalid_Identifier(Current_Language)));
      END IF;


      IF ( (A_Scheduler=Parametric_Protocol) OR (A_Scheduler=
               Hierarchical_Parametric_Protocol))
            AND (File_Name="")
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Processor(Current_Language) & Name &
               " : " & Lb_File_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;


      IF (File_Name/="")
            AND (A_Scheduler/=Parametric_Protocol)
            AND  (A_Scheduler=Hierarchical_Parametric_Protocol)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Processor(Current_Language) & Name &
               " : " & Lb_File_Name_Control(Current_Language)));
      END IF;


      IF (Quantum/=0)
            AND  (A_Scheduler/=Posix_1003_Highest_Priority_First_Protocol)
            AND (A_Scheduler/=Round_Robin_Protocol)
            AND (A_Scheduler/=Hierarchical_Round_Robin_Protocol)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Processor(Current_Language) & Name &
               " : " & Lb_Quantum_Control(Current_Language)));
      END IF;


      IF (Quantum<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Processor(Current_Language) & Name &
               " : " & "Quantum" &
               Lb_Must_Be(Current_Language)
               & To_Unbounded_String(" >= 0")));
      END IF;

      IF(A_Scheduler=No_Scheduling_Protocol)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Processor(Current_Language) & " " & Name &
               " : " & Lb_Invalid_Scheduler(
                  Current_Language)));
      END IF;

   END Check_Processor;




   PROCEDURE Add_Processor (
         My_Processors : IN OUT Processors_Set;
         A_Processor   : IN OUT Processor_Ptr;
         Name          : IN     Unbounded_String;
         Is_Preemptive : IN     Preemptive_Type;
         Quantum       : IN     Integer;
         File_Name     : IN     Unbounded_String;
         A_Scheduler   : IN     Scheduler_Type;
         A_Network     : IN     Unbounded_String) IS


      My_Iterator : Iterator;

   BEGIN

      Check_Processor(My_Processors , Name, Is_Preemptive, Quantum,
         File_Name, A_Scheduler, A_Network);


      IF(Get_Number_Of_Elements(My_Processors) > 0)
            THEN

         Reset_Iterator(My_Processors, My_Iterator);

         LOOP
            Current_Element(My_Processors, A_Processor, My_Iterator);
            IF (Name = A_Processor.Name)
                  THEN
               Raise_Exception(Invalid_Parameter'
                  Identity,
                  To_String(Lb_Processor(Current_Language) & " " & Name &
                     " : " & Lb_Processor_Name(
                        Current_Language) &
                     Lb_Already_Defined(
                        Current_Language)));
            END IF;

            EXIT WHEN Is_Last_Element(My_Processors, My_Iterator);

            Next_Element(My_Processors, My_Iterator);
         END LOOP;
      END IF;

      A_Processor:= NEW Processor;
      A_Processor.Name := To_Unbounded_String(To_String(Name));
      CASE A_Scheduler IS

         WHEN Rate_Monotonic_Protocol =>
            A_Processor.Scheduler := NEW Rm_Scheduler;
         WHEN Deadline_Monotonic_Protocol =>
            A_Processor.Scheduler := NEW Dm_Scheduler;
         WHEN Posix_1003_Highest_Priority_First_Protocol =>
            A_Processor.Scheduler := NEW Hpf_Scheduler;
         WHEN Earliest_Deadline_First_Protocol =>
            A_Processor.Scheduler := NEW Edf_Scheduler;
         WHEN Least_Laxity_First_Protocol =>
            A_Processor.Scheduler := NEW Llf_Scheduler;
         WHEN Round_Robin_Protocol =>
            A_Processor.Scheduler := NEW Round_Robin_Scheduler;
         WHEN Maximum_Urgency_First_Based_On_Deadline_Protocol =>
            A_Processor.Scheduler := NEW Deadline_Muf_Scheduler;
         WHEN Maximum_Urgency_First_Based_On_Laxity_Protocol =>
            A_Processor.Scheduler := NEW Laxity_Muf_Scheduler;
         WHEN Time_Sharing_Based_On_Wait_Time_Protocol =>
            A_Processor.Scheduler := NEW
               Time_Sharing_Based_On_Wait_Time_Scheduler;
         WHEN Time_Sharing_Based_On_Cpu_Usage_Protocol =>
            A_Processor.Scheduler := NEW
               Time_Sharing_Based_On_Cpu_Usage_Scheduler;
         WHEN D_Over_Protocol =>
            A_Processor.Scheduler := NEW D_Over_Scheduler;
         WHEN Parametric_Protocol =>
            A_Processor.Scheduler := NEW Parametric_Scheduler;

         WHEN Hierarchical_Cyclic_Protocol =>
            A_Processor.Scheduler := NEW Rm_Scheduler;
         WHEN Hierarchical_Round_Robin_Protocol =>
            A_Processor.Scheduler := NEW
               Hierarchical_Round_Robin_Scheduler;
         WHEN Hierarchical_Fixed_Priority_Protocol =>
            A_Processor.Scheduler := NEW Rm_Scheduler;
         WHEN Hierarchical_Parametric_Protocol  =>
            A_Processor.Scheduler := NEW Hierarchical_Parametric_Scheduler;
         WHEN No_Scheduling_Protocol =>
            Raise_Exception(Invalid_Parameter'Identity,
               To_String(Lb_Processor(Current_Language) & " " & Name &
                  " : " & Lb_Invalid_Scheduler(
                     Current_Language)));
      END CASE;


      A_Processor.Network :=A_Network;


      IF (A_Scheduler=Parametric_Protocol) OR (A_Scheduler=
            Hierarchical_Parametric_Protocol)
            THEN
         Set_Behavior_File_Name (
            Parametric_Scheduler(A_Processor.Scheduler.All), File_Name) ;
      END IF;


      Set_Quantum(A_Processor.Scheduler.All, Quantum);
      Set_Preemptive(A_Processor.Scheduler.All,
         Is_Preemptive);


      Add(My_Processors, A_Processor);


   EXCEPTION
      WHEN Full_Set =>
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Can_Not_Define_More_Processors(
                  Current_Language)));

   END Add_Processor;



   PROCEDURE Add_Processor (
         My_Processors : IN OUT Processors_Set;
         Name          : IN     Unbounded_String;
         Is_Preemptive : IN     Preemptive_Type;
         Quantum       : IN     Integer;
         File_Name     : IN     Unbounded_String;
         A_Scheduler   : IN     Scheduler_Type;
         A_Network     : IN     Unbounded_String) IS

      Dummy : Processor_Ptr;

   BEGIN
      Add_Processor(My_Processors, Dummy,
         Name, Is_Preemptive, Quantum, File_Name, A_Scheduler,
         A_Network);
   END Add_Processor;






   FUNCTION Get (
         My_Processors  : IN     Processors_Set;
         Processor_Name : IN     Unbounded_String;
         Param_Name     : IN     Processor_Parameters)
     RETURN Scheduler_Ptr IS

      A_Processor : Processor_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF (Param_Name /= Scheduler)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Processors, My_Iterator);

      LOOP
         Current_Element(My_Processors,A_Processor,My_Iterator);

         IF (A_Processor.Name = Processor_Name)
               THEN
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Processors, My_Iterator);

         Next_Element(My_Processors, My_Iterator);

      END LOOP;

      RETURN A_Processor.Scheduler;

   END Get;



   PROCEDURE Set (
         My_Processors  : IN OUT Processors_Set;
         Processor_Name : IN     Unbounded_String;
         Param_Name     : IN     Processor_Parameters;
         Param_Value    : IN     Scheduler_Ptr) IS

      A_Processor : Processor_Ptr;
      My_Iterator : Iterator;
   BEGIN

      IF (Param_Name /= Scheduler)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Processors, My_Iterator);

      LOOP
         Current_Element(My_Processors,A_Processor,My_Iterator);

         IF (A_Processor.Name = Processor_Name)
               THEN
            A_Processor.Scheduler:=Param_Value;
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Processors, My_Iterator);

         Next_Element(My_Processors, My_Iterator);

      END LOOP;
   END Set;


   FUNCTION Search_Processor (
         My_Processors : IN     Processors_Set;
         Name          : IN     Unbounded_String)
     RETURN Processor_Ptr IS

      My_Iterator : Iterator;
      A_Processor : Processor_Ptr;
      Result      : Processor_Ptr;

      Found : Boolean := False;

   BEGIN

      Reset_Iterator(My_Processors, My_Iterator);

      LOOP
         Current_Element(My_Processors, A_Processor, My_Iterator);

         IF(A_Processor.Name = Name)
               THEN
            Found:=True;
            Result:=A_Processor;
         END IF;

         EXIT WHEN Is_Last_Element(My_Processors, My_Iterator);

         Next_Element(My_Processors, My_Iterator);

      END LOOP;

      IF NOT Found
            THEN
         Raise_Exception(Processor_Not_Found'Identity,
            To_String(Lb_Processor_Name(Current_Language) & "=" & Name));
      END IF;

      RETURN Result;

   END Search_Processor;







   FUNCTION Export_Xml (
         My_Processors : IN     Processors_Set)
     RETURN Unbounded_String IS


      A_Processor : Processor_Ptr;
      My_Iterator : Processors_Iterator;

      Result : Unbounded_String := Empty_String;


   BEGIN

      IF NOT Is_Empty(My_Processors) THEN

         Result:= Result & To_Unbounded_String( "<processors>") &
            Unbounded_Lf;
         Reset_Iterator(My_Processors, My_Iterator);

         LOOP
            Current_Element(My_Processors, A_Processor, My_Iterator);

            Result:= Result & To_Unbounded_String( Ascii.Ht &
               "<processor>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<name>" & To_String(
                  A_Processor.Name) & "</name>") & Unbounded_Lf;


            Result:=Result & Export_Xml(A_Processor.Scheduler.All);


            IF A_Processor.Network /= Empty_String
                  THEN
               Result:= Result & To_Unbounded_String( Ascii.Ht &
                  Ascii.Ht &
                  "<network_link> " &
                  To_String(
                     A_Processor.Network) & " </network_link>") &
                  Unbounded_Lf;
            END IF;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               "</processor>") & Unbounded_Lf;


            EXIT WHEN Is_Last_Element(My_Processors, My_Iterator);

            Next_Element(My_Processors, My_Iterator);

         END LOOP;
         Result:= Result & To_Unbounded_String(  "</processors>") &
            Unbounded_Lf;
      END IF;

      RETURN Result;

   END Export_Xml;





   FUNCTION Export_Aadl_Implementations (
         My_Processors : IN     Processors_Set)
     RETURN Unbounded_String IS

      My_Iterator : Iterator;
      A_Processor : Processor_Ptr;

      Result : Unbounded_String := Empty_String;


   BEGIN

      IF NOT Is_Empty(My_Processors) THEN

         Reset_Iterator(My_Processors, My_Iterator);

         LOOP
            Current_Element(My_Processors, A_Processor, My_Iterator);


            Result:= Result & To_Unbounded_String(  "processor " &
               To_String(A_Processor.Name)) & Unbounded_Lf;

            IF A_Processor.Network /= To_Unbounded_String(
                  "No_Network")
                  THEN
               Result:= Result & To_Unbounded_String( Ascii.Ht &
                  "features") &
                  Unbounded_Lf;
               Result:= Result & To_Unbounded_String(  Ascii.Ht &
                  Ascii.Ht &
                  "bus_connected : requires bus access " & To_String(
                     A_Processor.Network) &  ";") &
                  Unbounded_Lf;
            END IF;
            Result:= Result & To_Unbounded_String(  "end " &
               To_String(
                  A_Processor.Name) & ";" ) & Unbounded_Lf &
               Unbounded_Lf;

            Result:= Result & To_Unbounded_String(
               "processor implementation " & To_String(
                  A_Processor.Name) & ".Impl" ) & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(  Ascii.Ht &
               "properties ") & Unbounded_Lf;

            Result:= Result & Export_Aadl_Properties (
               A_Processor.Scheduler.All, 2);

            Result:= Result & To_Unbounded_String(  "end " &
               To_String(
                  A_Processor.Name) &
               ".Impl;" ) & Unbounded_Lf & Unbounded_Lf;



            EXIT WHEN Is_Last_Element(My_Processors, My_Iterator);

            Next_Element(My_Processors, My_Iterator);

         END LOOP;


      END IF;

      RETURN Result;

   END Export_Aadl_Implementations;






   FUNCTION Export_Aadl_Declarations (
         My_Processors : IN     Processors_Set;
         Number_Of_Ht  : IN     Natural)
     RETURN Unbounded_String IS

      My_Iterator : Iterator;
      A_Processor : Processor_Ptr;

      Result : Unbounded_String := Empty_String;


   BEGIN

      IF NOT Is_Empty(My_Processors) THEN


         Reset_Iterator(My_Processors, My_Iterator);

         LOOP
            Current_Element(My_Processors, A_Processor, My_Iterator);

            FOR I IN 1..Number_Of_Ht LOOP
               Result:= Result & Ascii.Ht;
            END LOOP;
            Result:= Result & To_Unbounded_String(  "instancied_" &
               To_String(A_Processor.Name) &
               " : processor " & To_String(A_Processor.Name) &
               ".Impl;") &
               Unbounded_Lf;

            EXIT WHEN Is_Last_Element(My_Processors, My_Iterator);

            Next_Element(My_Processors, My_Iterator);

         END LOOP;

      END IF;

      RETURN Result;

   END Export_Aadl_Declarations;



END Processor_Set;
