------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automaticly check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developped by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;

WITH Indexed_Tables;

WITH Natural_Util;

WITH Double_Util;
USE Double_Util;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;

WITH Convert_Strings;
WITH Convert_Unbounded_Strings;
WITH Ada.Finalization;


PACKAGE Queueing_System IS

   Not_Implemented : EXCEPTION;

   -- different type of queueing sytem
   -- (there are not all implemented)
   --
   TYPE Queueing_Systems_Type IS
         (Qs_Pp1,
          Qs_Mm1,
          Qs_Md1,
          Qs_Mp1,
          Qs_Mg1,
          Qs_Mms,
          Qs_Mds,
          Qs_Mps,
          Qs_Mgs,
          Qs_Mm1n,
          Qs_Md1n,
          Qs_Mp1n,
          Qs_Mg1n,
          Qs_Mmsn,
          Qs_Mdsn,
          Qs_Mpsn,
          Qs_Mgsn);



   PROCEDURE To_Qs IS
   NEW Convert_Strings(Queueing_Systems_Type, Qs_Mm1);
   PROCEDURE To_Qs IS
   NEW Convert_Unbounded_Strings(Queueing_Systems_Type, Qs_Mm1);


   FUNCTION Qs_To_Display_String (
         A_Qs : IN     Queueing_Systems_Type)
     RETURN Unbounded_String;
   PROCEDURE Display_String_To_Qs (
         From : IN     Unbounded_String;
         To   :    OUT Queueing_Systems_Type;
         Ok   :    OUT Boolean);


   Arrival_Rate_Table_Size       : CONSTANT Natural := 100;
   Service_Rate_Table_Size       : CONSTANT Natural := 100;
   Resp_Time_Table_Size          : CONSTANT Natural := 1000;
   Resp_Time_Consumer_Table_Size : CONSTANT Natural := 1000;

   PACKAGE A_Arrival_Rate IS NEW Indexed_Tables(Double,
      Natural,
      Arrival_Rate_Table_Size,
      0,
      Double_Util.Put,
      Double_Util.Initialize,
      Natural_Util.Put);
   USE A_Arrival_Rate;

   PACKAGE A_Service_Rate IS NEW Indexed_Tables(Double,
      Natural,
      Service_Rate_Table_Size,
      0,
      Double_Util.Put,
      Double_Util.Initialize,
      Natural_Util.Put);
   USE A_Service_Rate;

   PACKAGE A_Resp_Time IS NEW Indexed_Tables(Double,
      Natural,
      Resp_Time_Table_Size,
      0,
      Double_Util.Put,
      Double_Util.Initialize,
      Natural_Util.Put);
   USE A_Resp_Time;

   PACKAGE A_Resp_Time_Consumer IS NEW Indexed_Tables(Double,
      Natural,
      Resp_Time_Consumer_Table_Size,
      0,
      Double_Util.Put,
      Double_Util.Initialize,
      Natural_Util.Put);
   USE A_Resp_Time_Consumer;


   SUBTYPE Arrival_Rate_Table IS A_Arrival_Rate.Indexed_Table;
   SUBTYPE Arrival_Rate_Range IS A_Arrival_Rate.Indexed_Table_Range;

   SUBTYPE Service_Rate_Table IS A_Service_Rate.Indexed_Table;
   SUBTYPE Service_Rate_Range IS A_Service_Rate.Indexed_Table_Range;

   SUBTYPE Resp_Time_Consumer_Table IS A_Resp_Time_Consumer.Indexed_Table;
   SUBTYPE Resp_Time_Consumer_Table_Ptr IS A_Resp_Time_Consumer.
      Indexed_Table_Ptr;
   SUBTYPE Resp_Time_Consumer_Range IS A_Resp_Time_Consumer.
      Indexed_Table_Range;

   SUBTYPE Resp_Time_Table IS A_Resp_Time.Indexed_Table;
   SUBTYPE Resp_Time_Table_Ptr IS A_Resp_Time.Indexed_Table_Ptr;
   SUBTYPE Resp_Time_Range IS A_Resp_Time.Indexed_Table_Range;

   TYPE Generic_Queueing_System IS ABSTRACT NEW Ada.Finalization.Controlled WITH
      RECORD
         Queueing_System_Type : Queueing_Systems_Type;
         Arrival_Rate         : Arrival_Rate_Table;
         Service_Rate         : Service_Rate_Table;
         Nb_Of_Messages       : Double;

         Var_Service : Double; -- variance service time for mg1
         Avg_Service : Double; -- average service time for mg1

         Waiting_Time : Double;
         Utilisation  : Double;
         Max_State    : Natural;

         Nb_Arrival : Double;  -- nb of producer or arrival flow
         Nb_Server  : Double;  -- nb of server or consumer
         Harmonic   : Boolean; -- for PP1 qs

         -- mp1 criteria computation
         --
         Consumer_Resp_Time     : Resp_Time_Consumer_Table_Ptr;
         Consumer_Resp_Time_Cpt : Resp_Time_Consumer_Range;
      END RECORD;

   TYPE Queueing_System_Ptr IS ACCESS ALL Generic_Queueing_System'Class;

   PROCEDURE Initialize (
         A_Queueing_System : IN OUT Generic_Queueing_System);


   -- Initialize a queueing system with a constant consumer response time
   --
   PROCEDURE Set_Constant_Consumer_Response_Time (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Response_Time     : IN     Double);


   -- reset queueing system values
   --
   PROCEDURE Reset (
         A_Queueing_System : IN OUT Generic_Queueing_System);


   -- display all queueing system caracteristics
   --
   PROCEDURE Put (
         A_Queueing_System : IN     Generic_Queueing_System);

   -- get or set queueing system caracteristics
   --
   PROCEDURE Set_Qs_Arrival_Rate (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Value             : IN     Double);

   PROCEDURE Set_Qs_Arrival_Rate (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Place             : IN     Arrival_Rate_Range;
         Value             : IN     Double);

   FUNCTION Get_Qs_Arrival_Rate (
         A_Queueing_System : IN     Generic_Queueing_System'Class;
         Place             : IN     Arrival_Rate_Range)
     RETURN Double;

   PROCEDURE Set_Qs_Service_Rate (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Value             : IN     Double);

   PROCEDURE Set_Qs_Service_Rate (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Place             : IN     Service_Rate_Range;
         Value             : IN     Double);

   FUNCTION Get_Qs_Service_Rate (
         A_Queueing_System : IN     Generic_Queueing_System'Class;
         Place             : IN     Service_Rate_Range)
     RETURN Double;

   PROCEDURE Set_Qs_Utilisation (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Value             :        Double);

   FUNCTION Get_Qs_Utilisation (
         A_Queueing_System : IN     Generic_Queueing_System'Class)
     RETURN Double;

   PROCEDURE Set_Qs_Type (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Qs_Type           : IN     Queueing_Systems_Type);

   FUNCTION Get_Qs_Type (
         A_Queueing_System : IN     Generic_Queueing_System'Class)
     RETURN Queueing_Systems_Type;

   PROCEDURE Set_Qs_Nb_Of_Messages (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Value             :        Double);

   FUNCTION Get_Qs_Nb_Of_Messages (
         A_Queueing_System : IN     Generic_Queueing_System'Class)
     RETURN Double;

   PROCEDURE Set_Qs_Waiting_Time (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Value             :        Double);

   FUNCTION Get_Qs_Waiting_Time (
         A_Queueing_System : IN     Generic_Queueing_System'Class)
     RETURN Double;

   PROCEDURE Set_Qs_Max_State (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Value             :        Natural);

   FUNCTION Get_Qs_Max_State (
         A_Queueing_System : IN     Generic_Queueing_System'Class)
     RETURN Natural;

   PROCEDURE Set_Qs_Nb_Arrival (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Value             :        Double);

   FUNCTION Get_Qs_Nb_Arrival (
         A_Queueing_System : IN     Generic_Queueing_System'Class)
     RETURN Double;

   PROCEDURE Set_Qs_Nb_Server (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Value             :        Double);

   FUNCTION Get_Qs_Nb_Server (
         A_Queueing_System : IN     Generic_Queueing_System'Class)
     RETURN Double;

   PROCEDURE Set_Qs_Harmonic (
         A_Queueing_System : IN OUT Generic_Queueing_System'Class;
         Value             :        Boolean);

   FUNCTION Get_Qs_Harmonic (
         A_Queueing_System : IN     Generic_Queueing_System'Class)
     RETURN Boolean;

   -- return the name of the queueing system
   --
   FUNCTION Get_Name (
         A_Queueing_System : IN     Generic_Queueing_System'Class)
     RETURN Unbounded_String;

   -- set queueing system caracteristics
   -- needed, for instance, if you want theoritical results after a simulation
   --
   PROCEDURE Set_Qs_Basic_Parameters (
         A_Queueing_System        : IN OUT Generic_Queueing_System'Class;
         A_Source_Queueing_System : IN     Generic_Queueing_System);


END Queueing_System;




