------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developped by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;

WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;

WITH Ada.Numerics.Float_Random;
USE Ada.Numerics.Float_Random;


WITH Queueing_System;
USE Queueing_System;


WITH Tasks;
USE Tasks;

WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;

WITH Text_Io;
USE Text_Io;

WITH Scheduler;
USE Scheduler;

WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Task_Offsets;
USE Task_Offsets;
USE Task_Offsets.Offsets_Table_Package;

WITH Text_Io;
USE Text_Io;

WITH Random_Tools;
USE Random_Tools;

WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;

WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;

WITH Natural_Util;
USE Natural_Util;

WITH Double_Util;
USE Double_Util;





PACKAGE BODY Random_Project IS

   PROCEDURE Random_Project_1_N (
         Nb_Prod      : IN     Buffer_Roles_Range;
         Per_Max_Prod : IN     Natural;
         Per_Cons     : IN     Natural;
         Dea_Max      : IN     Natural;
         Max_Base_Per : IN     Natural;
         Sys          :    OUT System) IS


   BEGIN

      Random_Project (Nb_Prod, 1, Per_Max_Prod, Per_Cons, Per_Cons,
         Dea_Max,Max_Base_Per ,Sys);

   END Random_Project_1_N;


   PROCEDURE Random_Project (
         Nb_Prod      : IN     Buffer_Roles_Range;
         Nb_Cons      : IN     Buffer_Roles_Range;
         Per_Max_Prod : IN     Natural;
         Per_Max_Cons : IN     Natural;
         Per_Min_Cons : IN     Natural;
         Dea_Max      : IN     Natural;
         Max_Base_Per : IN     Natural;
         Sys          :    OUT System) IS

      -- variables
      Offset    : Offsets_Table;
      Roles     : Buffer_Roles_Table;
      Tptr      : Task_Ptr;
      Cpu       : Unbounded_String   := To_Unbounded_String ("cpu");
      Ea        : Unbounded_String   := To_Unbounded_String ("ea");
      Per_Prod  : Natural;
      Per_Cons  : Natural;
      Seed      : Generator;
      Aux_Prod,
      Flow_Aux,
      Flow_Prod : Double;
      Min,
      Max       : Natural;
      Deadline  : Natural;
      Lcm       : Natural;

   BEGIN

      Lcm := Max_Base_Per + 1;
      WHILE (Lcm > Max_Base_Per) LOOP

         -- initialization
         Initialize(Roles);
         Initialize(Offset);
         Reset(Seed);
         Initialize(Sys);
         Aux_Prod := 0.0;
         Flow_Prod := 0.0;

         -- consumers generation
         FOR I IN 1 .. Nb_Cons LOOP
            Per_Cons :=Get_Rand_Parameter(Per_Min_Cons,Per_Max_Cons,Seed);

            Deadline := Get_Rand_Parameter(Per_Cons,Dea_Max*Per_Cons,Seed);
            Add_Task(Sys.Tasks,
               Tptr,
               To_Unbounded_String("Cons"),
               Cpu, Ea,
               Periodic_Type,
               0,
               1,
               Per_Cons,
               Deadline,
               0,
               0,
               1,
               0,
               Sched_Fifo,
               Offset);

            Roles.Entries(0).Item:=Tptr.Name;
            Roles.Entries(0).Data.Size:=1;
            Roles.Entries(0).Data.The_Role:=Consumer;

            Roles.Nb_Entries := Nb_Prod + 1;

            Aux_Prod := Aux_Prod + 1.0/Double(Per_Cons);

         END LOOP;

         -- Per_Prod := Per_Cons;

         Max := Per_Max_Prod;
         Flow_Aux := Aux_Prod - Double(Nb_Prod - 1)/Double(Max);

         -- test of the maximum
         IF (Flow_Aux < 0.0)
               THEN

            -- new maximum calculation
            Max := Natural(Nb_Prod) * Per_Cons;
            Flow_Aux := 1.0/Double(Per_Cons) - Double(Nb_Prod - 1)/Double(
               Max);
            Put_Line ("No solution with period maximum :" &
               Per_Max_Prod'Img);
            Put_Line ("new  maximum :" & Max'Img);

         END IF;

         -- Producers generation
         FOR J IN 1 .. Nb_Prod LOOP
            -- period of producers is based on the respect of the conservative data flow : the data produce is egal to the data consum
            IF (J = Nb_Prod) -- last producer
                  THEN
               Per_Prod := Natural(Double'Ceiling(1.0/Aux_Prod));
               --Put_Line ("Period = " & Per_Prod'Img);
            ELSE
               Min := Natural(Double'Ceiling(1.0/Flow_Aux));
               Per_Prod := Get_Rand_Parameter(Min,Max,Seed);
               --Put_Line ("Period = " & Per_Prod'Img);

            END IF;

            -- last producer period calculation
            Aux_Prod := Aux_Prod - 1.0/Double(Per_Prod);

            -- bandswith for next producers
            Flow_Aux := Flow_Aux - 1.0 / Double(Per_Prod) + 1.0/Double(
               Max);

            -- intermediary calculation of total data flow of the producers
            Flow_Prod := Flow_Prod + 1.0 / Double(Per_Prod);

            -- add the tasks to the system
            Deadline := Get_Rand_Parameter(Per_Prod,Dea_Max*Per_Prod,Seed);
            Add_Task(Sys.Tasks,Tptr, (To_Unbounded_String("Prod") &
                  Format(Natural(J))),
               Cpu , Ea, Periodic_Type,0, 1, Per_Prod, Deadline , 0, 0, 1,
               0,
               Sched_Fifo,
               Offset);

            -- array of maximum generation
            Roles.Entries(J).Item:=Tptr.Name;
            Roles.Entries(J).Data.Size:=1;
            Roles.Entries(J).Data.The_Role:=Producer;

         END LOOP;

         -- add the processor
         Add_Processor(Sys.Processors,Cpu,Preemptive, 0, Empty_String,
            Posix_1003_Highest_Priority_First_Protocol, Empty_String);


         -- add the buffer
         Add_Buffer(Sys.Buffers, To_Unbounded_String("Buffer"), 1, Cpu, Ea,
            Qs_PP1, Roles);

         -- display the total data of producers : checking
         -- Put_Line ("Producers data flow = " & Flow_Prod'Img);

         Lcm := Scheduling_Period(Sys.Tasks, Cpu);
         IF (Max_Base_Per = 0)
               THEN
            Lcm := 0;
         END IF;

         --Put_Line("LCM = " & LCM'img);

      END LOOP;


   END Random_Project;


   PROCEDURE Random_Project_1_N_Harmonic (
         Nb_Prod       : IN     Buffer_Roles_Range;
         Coef_Max_Prod : IN     Natural;
         Per_Cons      : IN     Natural;
         Dea_Max       : IN     Natural;
         Sys           :    OUT System) IS

      -- variables
      Offset      : Offsets_Table;
      Roles       : Buffer_Roles_Table;
      Tptr        : Task_Ptr;
      Cpu         : Unbounded_String   := To_Unbounded_String ("cpu");
      Ea          : Unbounded_String   := To_Unbounded_String ("ea");
      Per_Prod    : Natural;
      Seed        : Generator;
      Flow_Aux    : Double;
      Flow_Prod   : Double;
      Min,
      Max         : Natural;
      Last_Period : Natural;
      Deadline    : Natural;


   BEGIN

      -- initialization
      Initialize(Roles);
      Initialize(Offset);
      Reset(Seed);
      Initialize(Sys);

      -- add the consumer
      Deadline := Get_Rand_Parameter(Per_Cons,Dea_Max*Per_Cons,Seed);
      Add_Task(Sys.Tasks,Tptr, To_Unbounded_String("Cons"),
         Cpu, Ea, Periodic_Type,0, 1, Per_Cons, Deadline, 0, 0, 1, 0,
         Sched_Fifo,
         Offset);
      Roles.Entries(0).Item:=Tptr.Name;
      Roles.Entries(0).Data.Size:=1;
      Roles.Entries(0).Data.The_Role:=Consumer;
      Roles.Nb_Entries := Nb_Prod + 1;

      -- variables initialization
      Last_Period := Per_Cons;
      Flow_Aux := 1.0/ Double(Per_Cons);
      Min := 2;
      Max := Coef_Max_Prod;

      -- Put_Line("consumer flow = " & Flow_Aux'img);

      -- first producer
      Per_Prod := Last_Period * Get_Rand_Parameter(Min,Max,Seed);
      Add_Task(Sys.Tasks,Tptr, To_Unbounded_String("Prod1"),
         Cpu , Ea, Periodic_Type,0, 1, Per_Prod, Per_Prod , 0, 0, 1, 0,
         Sched_Fifo,
         Offset);
      Roles.Entries(1).Item:=Tptr.Name;
      Roles.Entries(1).Data.Size:=1;
      Roles.Entries(1).Data.The_Role:=Producer;

      Last_Period := Per_Prod;
      Flow_Aux := Flow_Aux - 1.0 / Double(Per_Prod);
      Flow_Prod := 1.0 / Double(Per_Prod);
      -- Put_Line("P 1" & " = " & Per_Prod'img);

      -- Producers generation
      FOR J IN 2 .. Nb_Prod LOOP
         IF (J /= Nb_Prod) -- not the last producer
               THEN
            IF(Equal(Flow_Aux,1.0 / Double(Last_Period)))
                  THEN
               Min := 2;
            ELSE
               Min := 1;
            END IF;

            -- new producer period
            Per_Prod := Last_Period * Get_Rand_Parameter(Min,Max,Seed);
            Last_Period := Per_Prod;

            -- data flow still available
            Flow_Aux := Flow_Aux - 1.0 / Double(Per_Prod);
         ELSE
            Per_Prod := Last_Period;
         END IF;

         Flow_Prod := Flow_Prod + 1.0 / Double(Per_Prod);

         -- Put_Line("P" & J'Img & " = " & Per_Prod'Img);

         -- add the tasks to the system
         Deadline := Get_Rand_Parameter(Per_Prod,Dea_Max*Per_Prod,Seed);
         Add_Task(Sys.Tasks,Tptr, (To_Unbounded_String("Prod") &
               Format(Natural(J))),
            Cpu , Ea, Periodic_Type, 0, 1, Per_Prod, Deadline , 0, 0, 1, 0,
            Sched_Fifo,
            Offset);

         Roles.Entries(J).Item:=Tptr.Name;
         Roles.Entries(J).Data.Size:=1;
         Roles.Entries(J).Data.The_Role:=Producer;

      END LOOP;

      -- Put_Line("producer flow = " & Flow_prod'Img);


      -- add the processor
      Add_Processor(Sys.Processors,Cpu,Preemptive, 0, Empty_String,
         Posix_1003_Highest_Priority_First_Protocol, Empty_String);

      -- add the buffer
      Add_Buffer(Sys.Buffers, To_Unbounded_String("Buffer"), 1, Cpu, Ea,
         Qs_PP1, Roles);

   END Random_Project_1_N_Harmonic;



END Random_Project;
