------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developped by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Tasks;
USE Tasks;

WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;

WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;

WITH Scheduler;
USE Scheduler;
USE Scheduler.Double_Tasks_Parameters;

WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;

WITH Processors;
USE Processors;

WITH Random_Tools;
USE Random_Tools;

WITH Ada.Numerics.Float_Random;
USE Ada.Numerics.Float_Random;

WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;

WITH Text_Io;
USE Text_Io;

WITH Buffers;
USE Buffers;
USE Buffers.Buffer_Roles;

WITH Processors;
USE Processors;

WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;

WITH Fixed_Strings;
USE Fixed_Strings;

WITH Natural_Util;
USE Natural_Util;

WITH Double_Util;
USE Double_Util;


WITH Ada.Sequential_Io;

PACKAGE BODY Random_Scheduling IS

   -- display an array of respone time table
   PROCEDURE Put (
         Res_Time_Set    : IN     Response_Table_Set;
         Nb_Res_Time_Set : IN     Natural) IS

   BEGIN
      FOR J IN 1 .. Nb_Res_Time_Set LOOP
         FOR I IN 0 .. Res_Time_Set(J).Nb_Entries-1 LOOP
            Put_Line("Task " &
               To_String(Res_Time_Set(J).Entries(I).Item.Name) &
               ", r = " &
               Res_Time_Set(J).Entries(I).Data'Img);
         END LOOP;
      END LOOP;
   END Put;

   PROCEDURE Random_Buff_Size_Plot (
         Nb_Rand_Scheduling : Natural;
         Nb_Base_Period     : Natural;
         Save_Response      : Boolean;
         Save_Buff_Response : Natural;
         Save_File_Name     : Unbounded_String;
         Sys                : System) IS

      TYPE Save_Descriptor IS
         RECORD
            Task_Name : Fixed_String;
            Eol1      : Character    := Ascii.Lf;
            Resp_Time : Fixed_String;
            Deadline  : Fixed_String;
            Buff_Size : Fixed_String;
         END RECORD;

      PACKAGE Save_File IS NEW Ada.Sequential_Io( Save_Descriptor );
      USE Save_File;

      -- types
      TYPE Natural_Tab IS ARRAY (Tasks_Range RANGE <>) OF Natural;
      TYPE Natural_Tab_Ptr IS ACCESS Natural_Tab;

      -- variables
      A_Buffer         : Buffer_Ptr;
      A_Task           : Task_Ptr;
      A_Processor      : Processor_Ptr;
      My_Iterator_Task : Tasks_Iterator;
      My_Iterator_Buf  : Buffers_Iterator;
      My_Iterator_Proc : Processors_Iterator;
      Res_Time         : Response_Time_Table;
      -- array of random response time
      Buff_Size       : Natural;             -- size of the buffer
      Seed            : Generator;
      I_Min_Resp_Time : Response_Time_Range;
      Cnt_Final       : Natural_Tab_Ptr;
      -- array of the maximum number of response time generation
      Continu       :          Boolean;
      Max_Buff_Size :          Natural;
      Buff_Bound    : CONSTANT Natural := 0;
      -- maximum size of the buffer
      Lcm               : Natural;             -- study period
      Min_Resp          : Natural;
      Max_Resp          : Natural;
      Nb_Max_Activation : Natural;
      Nb_Tasks          : Tasks_Range;
      L                 : Tasks_Range;
      K                 : Response_Time_Range;
      F_Desc            : Save_File.File_Type;
      Desc              : Save_Descriptor;
      File_Name         : Unbounded_String;
      Cnt_File          : Natural;


      -- return the index of the minimum response time
      FUNCTION Min_Res_Time (
            Min : IN     Response_Time_Table)
        RETURN Response_Time_Range IS

         Minimum : Response_Time_Range;
      BEGIN
         Minimum := 0;
         FOR I IN 1 .. (Min.Nb_Entries - 1) LOOP
            IF (Min.Entries(I).Data < Min.Entries(Minimum).Data)
                  THEN
               Minimum := I;
            END IF;
         END LOOP;

         RETURN Minimum;
      END Min_Res_Time;

   BEGIN

      -- !!! INIT !!!

      -- control : one processor, one buffer and only producer and consumer

      Reset_Iterator(Sys.Processors, My_Iterator_Proc);
      Current_Element(Sys.Processors, A_Processor, My_Iterator_Proc);

      Reset_Iterator(Sys.Buffers, My_Iterator_Buf);
      Current_Element(Sys.Buffers, A_Buffer, My_Iterator_Buf);

      Lcm := Scheduling_Period(Sys. -- only one processor
         Tasks, A_Processor.Name);

      K := 0;
      L := 0;
      Res_Time.Nb_Entries := 0;
      Nb_Tasks := Get_Number_Of_Task_From_Processor(Sys.Tasks,
         A_Processor.
         Name);
      Cnt_Final := NEW Natural_Tab(0 .. Nb_Tasks);
      Reset_Iterator(Sys.Tasks, My_Iterator_Task);
      LOOP
         -- take a task
         Current_Element(Sys.Tasks, A_Task, My_Iterator_Task);

         -- number of response time generation for each task
         Cnt_Final(L) := Lcm / Periodic_Task_Ptr(A_Task).Period *
            Nb_Base_Period*Nb_Rand_Scheduling;

         -- intialization of the response time table
         Res_Time.Entries(K).Data := Double'Last;
         Res_Time.Entries(K).Item:= A_Task ;
         Res_Time.Nb_Entries := Res_Time.Nb_Entries + 1;

         -- display the information
         Put_Line("#" & To_String(A_Task.Name) &
            " P = " & Periodic_Task_Ptr(A_Task).Period'Img &
            ", D = " & A_Task.Deadline'Img );

         -- next task
         L := L + 1;
         K := K + 1;
         EXIT WHEN Is_Last_Element(Sys.Tasks, My_Iterator_Task);
         Next_Element(Sys.Tasks, My_Iterator_Task);
      END LOOP;
      Reset(Seed);
      Cnt_File :=0;



      Put_Line("# Buffer bound = " & Buff_Bound'Img);


      -- plot the buffer size
      FOR T IN 0 .. (Nb_Rand_Scheduling - 1) LOOP

         -- init
         Buff_Size := 0;
         Max_Buff_Size := 0;

         IF (Save_Response)
               THEN
            File_Name := Save_File_Name & To_Unbounded_String("_") &
               Format(Cnt_File) & To_Unbounded_String(".sav");
            -- Put_Line(File_name);
            Create(F_Desc,
               Mode => Out_File,
               Name => To_String(File_Name));
         END IF;

         -- start random response time generation
         Continu := True;
         WHILE (Continu) LOOP

            Continu := False;
            -- if the buffer size doesn't change, stop the program
            K := 0;
            L := 0;

            Reset_Iterator(Sys.Tasks, My_Iterator_Task);
            LOOP
               -- take a task
               Current_Element(Sys.Tasks, A_Task, My_Iterator_Task);

               -- number maximum of activation during base period
               Nb_Max_Activation :=
                  Lcm / Periodic_Task_Ptr(A_Task).Period * Nb_Base_Period;

               -- generate a response time if needed
               IF (Res_Time.Entries(K).Data = Double'Last AND
                     Cnt_Final(L) > (Nb_Rand_Scheduling - T-1)*
                     Nb_Max_Activation)
                     THEN
                  Min_Resp := (Nb_Max_Activation * (Nb_Rand_Scheduling - T)  -
                     Cnt_Final(L))
                     *Periodic_Task_Ptr(A_Task).Period;
                  Max_Resp := Min_Resp + A_Task.Deadline-1;

                  Res_Time.Entries(K).Data :=
                     Double(Get_Rand_Parameter(Min_Resp,
                        Max_Resp,
                        Seed));
                  Cnt_Final(L) := Cnt_Final(L)-1;

                  -- save data
                  IF (Save_Response)
                        THEN
                     Desc.Task_Name := To_Bounded_String(To_String(
                           A_Task.
                           Name));
                     Desc.Resp_Time :=
                        To_Bounded_String(To_String(Format(Res_Time.
                              Entries(K).Data)) );
                     Desc.Deadline :=
                        To_Bounded_String(To_String(To_Unbounded_String(
                              Max_Resp'Img)));
                     Desc.Buff_Size := To_Bounded_String(To_String(
                           To_Unbounded_String(Buff_Size'Img)));

                     -- save the tasks response time in a file for further studies
                     Write(F_Desc,Desc);
                  END IF;

               END IF;

               -- next task
               L := L + 1;
               K := K + 1;
               EXIT WHEN Is_Last_Element(Sys.Tasks, My_Iterator_Task);
               Next_Element(Sys.Tasks, My_Iterator_Task);
            END LOOP;

            -- find the minimum response time
            I_Min_Resp_Time := Min_Res_Time(Res_Time);

            -- the task
            A_Task := Res_Time.Entries(I_Min_Resp_Time).Item;

            -- Put_Line("min " & A_Task.Name & ", r=" & Res_Time.Entries(I_Min_Resp_Time).Data'img);
            -- search for the type of the task which has the minimum response time
            -- (if equalitie, worst case if consumption before production)
            FOR J IN 0 .. Buffer_Roles_Range(Nb_Tasks-1) LOOP

               IF (Res_Time.Entries(I_Min_Resp_Time).Data < Double'Last )
                     THEN

                  IF (A_Buffer.Roles.Entries(J).Item = A_Task.Name)
                        THEN
                     IF (A_Buffer.Roles.Entries(J).Data.The_Role =
                           Consumer)
                           THEN
                        IF (Buff_Size > 0)
                              THEN
                           Buff_Size := Buff_Size - 1;
                        END IF;
                        Continu := True;
                        -- new response time generation
                        Res_Time.Entries(I_Min_Resp_Time).Data := Double'
                           Last;

                        -- Put_Line(To_Unbounded_String("Buffer size = ") & Buff_Size'img);
                     ELSE
                        IF (A_Buffer.Roles.Entries(J).Data.The_Role =
                              Producer)
                              THEN
                           Buff_Size := Buff_Size + 1;

                           IF (Buff_Size > Max_Buff_Size)
                                 THEN
                              Max_Buff_Size := Buff_Size;
                           END IF;
                           Continu := True;
                           -- new response time generation
                           Res_Time.Entries(I_Min_Resp_Time).Data :=
                              Double'Last;
                           -- Put_Line(To_Unbounded_String("Buffer size = ") & Buff_Size'img);
                        END IF;
                     END IF;
                  END IF;
               END IF;
            END LOOP; -- end for

            -- Put_Line(To_Unbounded_String("Buffer size = ") & Buff_Size'img);

         END LOOP; -- end while
         -- Put_Line("Maximum buffer size = " & Max_Buff_Size'img);

         IF (Max_Buff_Size >= Save_Buff_Response)
               THEN
            Cnt_File := Cnt_File + 1;

         END IF;
         IF (Max_Buff_Size > Buff_Bound)
               THEN
            Put_Line("# WARNING !!! max occupation = " & Max_Buff_Size'
               Img & "Bound = " & Buff_Bound'Img);
         END IF;
         IF (Save_Response)
               THEN
            Close(F_Desc); -- close the old file
         END IF;

         -- use with gnuplot
         Put_Line(T'Img & " " & Max_Buff_Size'Img);

      END LOOP; -- end for nb tasks set generation

   END Random_Buff_Size_Plot;

END Random_Scheduling;
