------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------




WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;

WITH Resources;
USE Resources;

WITH Time_Units;
USE  Time_Units;
USE  Time_Units.A_Time_Unit;

WITH Unbounded_Strings;
USE Unbounded_Strings;


WITH Ada.Exceptions;
USE Ada.Exceptions;


WITH Translate;
USE Translate;
WITH Objects;
USE Objects;



PACKAGE BODY Resource_Set IS


   PROCEDURE Task_Begin_End (
         Task_Name      : IN     Unbounded_String;
         Resource_Name  : IN     Unbounded_String;
         A_Resource_Set : IN     Resources_Set;
         Starts_Ends    : IN OUT Start_End_Table;
         Elements       : IN OUT Task_List_Range) IS

      My_Resource_Iterator : Resources_Iterator;
      A_Resource           : Resource_Ptr;
      Index                : Task_List_Range    := 0;

      -- looking for a resource and a task the end and the begin of this task on this resource
      -- Element is the number of element, must start to 0
      --
   BEGIN
      Reset_Iterator(A_Resource_Set, My_Resource_Iterator);
      LOOP
         Current_Element(A_Resource_Set, A_Resource, My_Resource_Iterator);
         IF A_Resource.Name = Resource_Name THEN

            FOR I IN 0..A_Resource.Task_Tab.Nb_Entries-1 LOOP
               IF (A_Resource.Task_Tab.Entries(Index).Item= Task_Name) THEN
                  Starts_Ends(Elements+1).Task_Begin:= A_Resource.
                     Task_Tab.Entries(Index).Data.Task_Begin;
                  Starts_Ends(Elements+1).Task_End:= A_Resource.Task_Tab.
                     Entries(Index).Data.Task_End;
                  Elements := Elements + 1;
               END IF;
               Index := Index + 1;
            END LOOP;
            EXIT;
         END IF;
         EXIT WHEN Is_Last_Element(A_Resource_Set, My_Resource_Iterator);
         Next_Element(A_Resource_Set, My_Resource_Iterator);
      END LOOP;

   END Task_Begin_End;




   PROCEDURE Update_Resource (
         My_Resources       : IN OUT Resources_Set;
         Name               : IN     Unbounded_String;
         State              : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         Protocol           : IN     Resources_Type;
         Affected_Tasks     : IN     Task_List) IS

      The_Resource : Resource_Ptr;

   BEGIN

      The_Resource := Search_Resource(My_Resources, Name);

      Check_Resource(My_Resources, Name, State, Cpu_Name,
         Address_Space_Name,
         Protocol);
      Delete(My_Resources, The_Resource);
      Add_Resource(My_Resources, Name, State, Cpu_Name, Address_Space_Name,
         Protocol, Affected_Tasks);

   END Update_Resource;


   PROCEDURE Add_Resource (
         My_Resources       : IN OUT Resources_Set;
         Name               : IN     Unbounded_String;
         State              : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         Protocol           : IN     Resources_Type;
         Affected_Tasks     : IN     Task_List) IS


      A_Resource_Ptr : Resource_Ptr;

   BEGIN

      Add_Resource(My_Resources, A_Resource_Ptr, Name, State, Cpu_Name,
         Address_Space_Name, Protocol,
         Affected_Tasks);

   END Add_Resource;



   PROCEDURE Check_Resource (
         My_Resources       : IN     Resources_Set;
         Name               : IN     Unbounded_String;
         State              : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         Protocol           : IN     Resources_Type) IS

   BEGIN

      IF Name=""
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Resource_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;

      IF NOT Is_A_Valid_Identifier(Name)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Resource(Current_Language) & " " & Name &
               " : " & Lb_Resource_Name(Current_Language) &
               Lb_Colon & Lb_Invalid_Identifier(Current_Language)));
      END IF;

      IF Cpu_Name=""
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Resource(Current_Language) & " " & Name &
               " : " & Lb_Processor_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;


      IF Address_Space_Name=""
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Resource(Current_Language) & " " & Name &
               " : " & Lb_Address_Space_Name(
                  Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;


      IF (Protocol /= No_Protocol)
            AND (Protocol /= Priority_Ceiling_Protocol)
            AND (Protocol /= Priority_Inheritance_Protocol)
            AND (Protocol /= Immediate_Priority_Ceiling_Protocol)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Resource(Current_Language) & " " & Name &
               " : " & Lb_Invalid_Protocol(Current_Language)));
      END IF;


   END Check_Resource;



   PROCEDURE Add_Resource (
         My_Resources       : IN OUT Resources_Set;
         A_Resource         : IN OUT Resource_Ptr;
         Name               : IN     Unbounded_String;
         State              : IN     Integer;
         Cpu_Name           : IN     Unbounded_String;
         Address_Space_Name : IN     Unbounded_String;
         Protocol           : IN     Resources_Type;
         Affected_Tasks     : IN     Task_List) IS


      New_Protocol      : Resources_Type;
      New_Pcp_Resource  : Pcp_Resource_Ptr;
      New_Pip_Resource  : Pip_Resource_Ptr;
      New_Np_Resource   : Np_Resource_Ptr;
      New_IPcp_Resource : IPcp_Resource_Ptr;

      My_Iterator : Iterator;

   BEGIN


      Check_Resource(My_Resources, Name, State, Cpu_Name,
         Address_Space_Name, Protocol);



      -- Resource already exist ?
      --
      IF(Get_Number_Of_Elements(My_Resources) > 0) THEN

         Reset_Iterator(My_Resources, My_Iterator);

         LOOP
            Current_Element(My_Resources, A_Resource, My_Iterator);

            IF (Name = A_Resource.Name) THEN
               Raise_Exception(Invalid_Parameter'
                  Identity,
                  To_String(Lb_Resource(Current_Language) & " " & Name &
                     " : " & Lb_Resource_Name(
                        Current_Language) &
                     Lb_Already_Defined(
                        Current_Language)));
            END IF;

            EXIT WHEN Is_Last_Element(My_Resources, My_Iterator);

            Next_Element(My_Resources, My_Iterator);

         END LOOP;

      END IF;


      IF (Protocol = Priority_Ceiling_Protocol)
            THEN
         New_Pcp_Resource := NEW Pcp_Resource;
         New_Protocol := Priority_Ceiling_Protocol;
         A_Resource:=Resource_Ptr(New_Pcp_Resource);

      ELSE
         IF (Protocol = Priority_Inheritance_Protocol)
               THEN
            New_Pip_Resource := NEW Pip_Resource;
            New_Protocol := Priority_Inheritance_Protocol;
            A_Resource:=Resource_Ptr(New_Pip_Resource);
         ELSE
            IF (Protocol = No_Protocol)
                  THEN
               New_Np_Resource := NEW Np_Resource;
               New_Protocol := No_Protocol;
               A_Resource:=Resource_Ptr(New_Np_Resource);
            ELSE
               New_IPcp_Resource := NEW IPcp_Resource;
               New_Protocol := Immediate_Priority_Ceiling_Protocol;
               A_Resource:=Resource_Ptr(New_IPcp_Resource);
            END IF;
         END IF;
      END IF;

      A_Resource.Name:=Name;
      A_Resource.State:=State;
      A_Resource.Cpu_Name:=Cpu_Name;
      A_Resource.Address_Space_Name:=Address_Space_Name;
      A_Resource.Protocol := New_Protocol;

      A_Resource.Task_Tab:=Affected_Tasks;


      Add(My_Resources,A_Resource);


   EXCEPTION
      WHEN Full_Set =>
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Can_Not_Define_More_Resources(
                  Current_Language)));

   END Add_Resource;




   FUNCTION Search_Resource (
         My_Resources : IN     Resources_Set;
         Name         : IN     Unbounded_String)
     RETURN Resource_Ptr IS

      My_Iterator : Iterator;
      A_Resource  : Resource_Ptr;
      Result      : Resource_Ptr;

      Found : Boolean := False;

   BEGIN

      Reset_Iterator(My_Resources, My_Iterator);

      LOOP
         Current_Element(My_Resources,A_Resource,My_Iterator);

         IF(A_Resource.Name = Name) THEN
            Found:=True;
            Result:=A_Resource;
         END IF;

         EXIT WHEN Is_Last_Element(My_Resources, My_Iterator);

         Next_Element(My_Resources, My_Iterator);

      END LOOP;

      IF NOT Found THEN
         Raise_Exception(Resource_Not_Found'Identity,
            To_String(Lb_Resource_Name(Current_Language) & "=" & Name));
      END IF;

      RETURN Result;

   END Search_Resource;

   PROCEDURE Delete_Task (
         My_Resources : IN OUT Resources_Set;
         A_Task       : IN     Unbounded_String) IS

      A_Resource  : Resource_Ptr;
      My_Iterator : Resources_Iterator;
      Index       : Task_List_Range    := 1;

   BEGIN

      IF(Get_Number_Of_Elements(My_Resources) > 0) THEN

         Reset_Iterator(My_Resources, My_Iterator);
         LOOP
            Current_Element(My_Resources, A_Resource, My_Iterator);

            -- Looking for a_task in the task_list and delete it
            --
            Index:=0;
            WHILE (Index<A_Resource.Task_Tab.Nb_Entries)
                  LOOP
               IF (A_Resource.Task_Tab.Entries(Index).Item =
                     A_Task)
                     THEN
                  A_Resource.Task_Tab.Entries(Index):=
                     A_Resource.Task_Tab.
                     Entries(A_Resource.Task_Tab.Nb_Entries-1);

                  A_Resource.Task_Tab.Nb_Entries:=
                     A_Resource.Task_Tab.Nb_Entries-1;
               END IF;
               Index:=Index+1;
            END LOOP;


            EXIT WHEN Is_Last_Element(My_Resources, My_Iterator);
            Next_Element(My_Resources, My_Iterator);
         END LOOP;
      END IF;

   END Delete_Task;



   PROCEDURE Delete_Address_Space (
         My_Resources : IN OUT Resources_Set;
         A_Addr       : IN     Unbounded_String) IS

      Tmp         : Resources_Set;
      A_Resource  : Resource_Ptr;
      My_Iterator : Resources_Iterator;

   BEGIN
      IF(Get_Number_Of_Elements(My_Resources) > 0) THEN

         Reset_Iterator(My_Resources, My_Iterator);
         LOOP
            Current_Element(My_Resources, A_Resource, My_Iterator);

            IF(A_Resource.Address_Space_Name = A_Addr)
                  THEN
               Add(Tmp, A_Resource);
            END IF;

            EXIT WHEN Is_Last_Element(My_Resources, My_Iterator);
            Next_Element(My_Resources, My_Iterator);
         END LOOP;

         IF NOT Is_Empty(Tmp) THEN
            Reset_Iterator(Tmp, My_Iterator);
            LOOP
               Current_Element(Tmp, A_Resource, My_Iterator);

               Delete(My_Resources,A_Resource);

               EXIT WHEN Is_Last_Element(Tmp, My_Iterator);
               Next_Element(Tmp, My_Iterator);
            END LOOP;
         END IF;

      END IF;
   END Delete_Address_Space;






   PROCEDURE Delete_Processor (
         My_Resources : IN OUT Resources_Set;
         A_Processor  : IN     Unbounded_String) IS

      Tmp         : Resources_Set;
      A_Resource  : Resource_Ptr;
      My_Iterator : Resources_Iterator;

   BEGIN

      IF(Get_Number_Of_Elements(My_Resources) > 0)
            THEN

         Reset_Iterator(My_Resources, My_Iterator);
         LOOP
            Current_Element(My_Resources, A_Resource,
               My_Iterator);

            IF(A_Resource.Cpu_Name = A_Processor)
                  THEN
               Add(Tmp, A_Resource);
            END IF;

            EXIT WHEN Is_Last_Element(My_Resources, My_Iterator);
            Next_Element(My_Resources, My_Iterator);
         END LOOP;

         IF NOT Is_Empty(Tmp) THEN
            Reset_Iterator(Tmp, My_Iterator);
            LOOP
               Current_Element(Tmp, A_Resource, My_Iterator);


               Delete(My_Resources,A_Resource);

               EXIT WHEN Is_Last_Element(Tmp, My_Iterator);
               Next_Element(Tmp, My_Iterator);
            END LOOP;

            Reset(Tmp,False);
         END IF;
      END IF;

   END Delete_Processor;




   PROCEDURE Same_Protocol_Control (
         My_Resources : IN     Resources_Set;
         A_Task       : IN     Unbounded_String) IS

      A_Protocol : Resources_Type;
      Is_First   : Boolean            := True;
      Iterator1  : Resources_Iterator;
      Resource1  : Resource_Ptr;

   BEGIN

      Reset_Iterator(My_Resources, Iterator1);

      LOOP
         Current_Element(My_Resources, Resource1, Iterator1);

         -- Looking for task user in the task_list index_table
         --
         FOR I IN 0..Resource1.Task_Tab.Nb_Entries-1 LOOP
            IF A_Task = Resource1.Task_Tab.Entries(I).Item
                  THEN
               IF  Is_First
                     THEN
                  Is_First:=False;
                  A_Protocol:=Resource1.Protocol;
               ELSE
                  IF A_Protocol /= Resource1.Protocol
                        THEN
                     RAISE Can_Not_Used_Different_Protocol;
                  END IF;
               END IF;
            END IF;
         END LOOP;

         EXIT WHEN Is_Last_Element(My_Resources, Iterator1);
         Next_Element(My_Resources, Iterator1);

      END LOOP;

   END Same_Protocol_Control;



   FUNCTION Export_Xml (
         My_Resources : IN     Resources_Set)
     RETURN Unbounded_String IS

      My_Iterator : Resources_Iterator;
      A_Resource  : Resource_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN

      IF NOT Is_Empty(My_Resources) THEN

         Result:= Result & To_Unbounded_String( "<resources>") &
            Unbounded_Lf;

         Reset_Iterator(My_Resources, My_Iterator);

         LOOP
            Current_Element(My_Resources,A_Resource,My_Iterator);

            Result:= Result & To_Unbounded_String( Ascii.Ht &
               "<resource>") &
               Unbounded_Lf;

            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<cpu_name>" & To_String(
                  A_Resource.Cpu_Name) & "</cpu_name>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<address_space_name>" &
               To_String(
                  A_Resource.Address_Space_Name) &
               "</address_space_name>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<name>" & To_String(
                  A_Resource.Name) & "</name>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<state>" & A_Resource.
               State'Img & "</state>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<protocol>" & A_Resource.
               Protocol'Img & "</protocol>") & Unbounded_Lf;


            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<resource_used_by>") & Unbounded_Lf;
            FOR I IN 0..A_Resource.Task_Tab.Nb_Entries-1 LOOP
               Result:= Result & To_Unbounded_String( Ascii.Ht &
                  Ascii.Ht &
                  Ascii.Ht & "<resource_user>");
               Result:= Result & To_Unbounded_String( " " &
                  To_String(
                     A_Resource.Task_Tab.Entries(I).
                     Item) & " ");
               Result:= Result & To_Unbounded_String(
                  A_Resource.Task_Tab.Entries(I).Data.Task_Begin'
                  Img & " " );
               Result:= Result & To_Unbounded_String(
                  A_Resource.Task_Tab.Entries(I).Data.Task_End'
                  Img &
                  " " );
               Result:= Result & To_Unbounded_String(
                  "</resource_user>") &
                  Unbounded_Lf;
            END LOOP;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "</resource_used_by>") & Unbounded_Lf;


            Result:= Result & To_Unbounded_String( Ascii.Ht &
               "</resource>") & Unbounded_Lf;


            EXIT WHEN Is_Last_Element(My_Resources, My_Iterator);

            Next_Element(My_Resources, My_Iterator);

         END LOOP;
         Result:= Result & To_Unbounded_String( "</resources>") &
            Unbounded_Lf;
      END IF;

      RETURN Result;

   END Export_Xml;



   FUNCTION Export_Aadl_Implementations (
         My_Resources : IN     Resources_Set)
     RETURN Unbounded_String IS

      My_Iterator : Resources_Iterator;
      A_Resource  : Resource_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN

      IF NOT Is_Empty(My_Resources) THEN

         Reset_Iterator(My_Resources, My_Iterator);

         LOOP
            Current_Element(My_Resources,A_Resource,My_Iterator);

            Result:= Result & To_Unbounded_String( "data " &
               To_String(
                  A_Resource.Name)) & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( "end " &
               To_String(
                  A_Resource.Name) & ";" ) & Unbounded_Lf &
               Unbounded_Lf;


            Result:= Result & To_Unbounded_String(
               "data implementation " &
               To_String(
                  A_Resource.
                  Name) & ".Impl" ) & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               "properties ") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "Cheddar_Properties::Data_Concurrency_State => " &
               A_Resource.State'Img  & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(  Ascii.Ht &
               Ascii.Ht &
               "Concurrency_Control_Protocol => " &
               A_Resource.Protocol'
               Img & ";") & Unbounded_Lf;

            Result:= Result & To_Unbounded_String(  "end " &
               To_String(
                  A_Resource.Name) & ".Impl;" ) & Unbounded_Lf &
               Unbounded_Lf;


            EXIT WHEN Is_Last_Element(My_Resources, My_Iterator);

            Next_Element(My_Resources, My_Iterator);

         END LOOP;

      END IF;

      RETURN Result;

   END Export_Aadl_Implementations;



   FUNCTION Export_Aadl_Declarations (
         My_Resources       : IN     Resources_Set;
         Address_Space_Name :        Unbounded_String;
         Number_Of_Ht       : IN     Natural)
     RETURN Unbounded_String IS


      My_Iterator : Resources_Iterator;
      A_Resource  : Resource_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN


      IF NOT Is_Empty(My_Resources) THEN


         Reset_Iterator(My_Resources, My_Iterator);

         LOOP
            Current_Element(My_Resources,A_Resource,My_Iterator);

            IF A_Resource.Address_Space_Name = Address_Space_Name THEN
               FOR I IN 1..Number_Of_Ht LOOP
                  Result:= Result & Ascii.Ht;
               END LOOP;
               Result:= Result & To_Unbounded_String(
                  "instancied_" &
                  To_String(A_Resource.Name) &
                  " : data " & To_String(A_Resource.Name) &
                  ".Impl;") &
                  Unbounded_Lf;
            END IF;


            EXIT WHEN Is_Last_Element(My_Resources, My_Iterator);

            Next_Element(My_Resources, My_Iterator);

         END LOOP;
      END IF;

      RETURN Result;

   END Export_Aadl_Declarations;


   FUNCTION Export_Aadl_Connections (
         My_Resources : IN     Resources_Set;
         A_Task_Name  : IN     Unbounded_String;
         Number_Of_Ht : IN     Natural)
     RETURN Unbounded_String IS

      My_Iterator : Resources_Iterator;
      A_Resource  : Resource_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN
      IF NOT Is_Empty(My_Resources) THEN


         Reset_Iterator(My_Resources, My_Iterator);

         LOOP
            Current_Element(My_Resources, A_Resource, My_Iterator);

            IF A_Resource.Task_Tab.Nb_Entries/=0 THEN
               FOR I IN 0..A_Resource.Task_Tab.Nb_Entries-1 LOOP
                  IF  A_Resource.Task_Tab.Entries(I).Item =
                        A_Task_Name THEN
                     FOR J IN 1..Number_Of_Ht LOOP
                        Result:= Result & Ascii.Ht;
                     END LOOP;
                     Result:= Result & To_Unbounded_String(
                        "data access " &
                        To_String("instancied_" & A_Task_Name &
                           "." & A_Resource.Name & "_features" ) &
                        " -> instancied_" & To_String(
                           A_Resource.Name) &
                        ";") &
                        Unbounded_Lf;
                  END IF;
               END LOOP;
            END IF;

            EXIT WHEN Is_Last_Element(My_Resources, My_Iterator);

            Next_Element(My_Resources, My_Iterator);

         END LOOP;
      END IF;

      RETURN Result;

   END   Export_Aadl_Connections;



   FUNCTION Get_Number_Of_Resource_From_Processor (
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Resources_Range IS

      Number      : Resources_Range := 0;
      A_Resource  : Resource_Ptr;
      My_Iterator : Iterator;

   BEGIN


      IF Is_Empty(My_Resources)
            THEN
         RETURN 0;
      END IF;


      Reset_Iterator(My_Resources, My_Iterator);

      LOOP
         Current_Element(My_Resources, A_Resource, My_Iterator);

         IF (A_Resource.Cpu_Name = Processor_Name)
               THEN
            Number:=Number+1;
         END IF;

         EXIT WHEN Is_Last_Element(My_Resources, My_Iterator);

         Next_Element(My_Resources, My_Iterator);

      END LOOP;

      RETURN Number;
   END Get_Number_Of_Resource_From_Processor;



END Resource_Set;








