------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Framework_Config;
USE Framework_Config;
WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;


WITH Text_Io;
USE Text_Io;


WITH Scheduler;
USE Scheduler;


WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;







PACKAGE BODY Scheduler.Dynamic_Priority.Muf.Laxity_Based IS


   PROCEDURE Initialize (
         A_Scheduler : IN OUT Laxity_Muf_Scheduler) IS
   BEGIN
      Reset(A_Scheduler);
      A_Scheduler.Name:= Maximum_Urgency_First_Based_On_Laxity_Protocol;
   END Initialize;



   FUNCTION Copy (
         A_Scheduler : IN     Laxity_Muf_Scheduler)
     RETURN Scheduler_Ptr IS

      Ptr : Muf_Scheduler_Ptr;

   BEGIN

      Ptr:= NEW Laxity_Muf_Scheduler;

      Ptr.Name:=A_Scheduler.Name;
      Ptr.Is_Preemptive:=A_Scheduler.Is_Preemptive;
      Ptr.Quantum:=A_Scheduler.Quantum;
      Ptr.Previously_Elected:=A_Scheduler.Previously_Elected;

      RETURN Scheduler_Ptr(Ptr);

   END Copy;


   PROCEDURE Do_Election (
         My_Scheduler      : IN OUT Laxity_Muf_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Msg               : IN OUT Unbounded_String;
         Current_Time      : IN     Natural;
         Processor_Name    : IN     Unbounded_String;
         My_Dependencies   : IN     Tasks_Dependencies_Ptr;
         With_Offsets      : IN     Boolean;
         With_Precedencies : IN     Boolean;
         With_Resources    : IN     Boolean;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table;
         Elected           : IN OUT Tasks_Range;
         No_Task           : IN OUT Boolean) IS



      Smallest_Laxity : Natural     := Natural'Last;
      Laxity          : Natural     := 0;
      I               : Tasks_Range := 0;
      Almost_One_Task : Boolean     := False;

   BEGIN

      -- Find the first ready task and assume we'll run this one
      --
      LOOP
         IF (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) THEN

            -- Compute current laxity
            --
            Laxity:=Dynamic_Priority_Tcb_Ptr(Si.Tcbs(I)).Dynamic_Deadline -
               Si.Tcbs(I).
               Rest_Of_Capacity;

            -- Is the task ready to run ?
            --
            IF (Si.Tcbs(I).Wake_Up_Time <= Current_Time) AND
                  (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                  THEN
               IF (With_Offsets=False) OR
                     Check_Offset(Si.Tcbs(I), Current_Time)
                     THEN
                  IF (With_Precedencies=False) OR
                        Check_Precedencies(Si, My_Dependencies,
                        Current_Time,
                        Si.Tcbs(I))
                        THEN

                     Almost_One_Task :=True;
                     Smallest_Laxity:=Laxity;
                     Elected:=I;

                  END IF;
               END IF;
            END IF;
         END IF;

         I:=I+1;
         EXIT WHEN Si.Tcbs(I)=NULL;
      END LOOP;



      IF NOT Almost_One_Task
            THEN
         No_Task:=True;
      ELSE
         No_Task:=False;


         -- Let see now if some more urgent/critical task exists
         --
         I:=0;
         LOOP
            IF (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) THEN

               -- Compute current laxity
               --
               Laxity:=Dynamic_Priority_Tcb_Ptr(Si.Tcbs(I)).
                  Dynamic_Deadline -
                  Si.Tcbs(I).
                  Rest_Of_Capacity;

               -- Is the task ready to run ?
               --
               IF (Si.Tcbs(I).Wake_Up_Time <= Current_Time) AND
                     (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                     THEN
                  IF (With_Offsets=False) OR
                        Check_Offset(Si.Tcbs(I), Current_Time)
                        THEN
                     IF (With_Precedencies=False) OR
                           Check_Precedencies(Si, My_Dependencies,
                           Current_Time,
                           Si.Tcbs(I))
                           THEN


                        -- Select according to the laxity, criticality and priority
                        --

                        -- First case : we find a more critical task
                        --
                        IF (Si.Tcbs(I).Tsk.Criticality > Si.Tcbs(Elected).
                              Tsk.Criticality)
                              THEN
                           Smallest_Laxity:=Laxity;
                           Elected:=I;
                        ELSE

                           -- Second case : we find a more urgent task
                           --
                           IF (Si.Tcbs(I).Tsk.Criticality = Si.Tcbs(
                                    Elected).Tsk.Criticality)
                                 AND (Laxity < Smallest_Laxity)
                                 THEN
                              Smallest_Laxity:=Laxity;
                              Elected:=I;
                           ELSE

                              -- Last case : we find a task with a highiest priority
                              --
                              IF (Si.Tcbs(I).Tsk.Criticality = Si.Tcbs(
                                       Elected).Tsk.Criticality)
                                    AND (Laxity = Smallest_Laxity)
                                    AND (Si.Tcbs(I).Tsk.Priority > Si.Tcbs(
                                       Elected).Tsk.Priority)
                                    THEN
                                 Smallest_Laxity:=Laxity;
                                 Elected:=I;
                              END IF;
                           END IF;
                        END IF;



                     END IF;
                  END IF;
               END IF;
            END IF;

            I:=I+1;
            EXIT WHEN Si.Tcbs(I)=NULL;
         END LOOP;


         -- last time unit : compute dynamic deadline for periodic task
         --
         IF (Si.Tcbs(Elected).Rest_Of_Capacity = 1)
               AND (Si.Tcbs(Elected).Tsk.Task_Type = Periodic_Type)
               THEN
            Dynamic_Priority_Tcb_Ptr(Si.Tcbs(Elected)).Dynamic_Deadline:=
               Dynamic_Priority_Tcb_Ptr(Si.Tcbs(Elected)).
               Dynamic_Deadline
               +Periodic_Task_Ptr(Si.Tcbs(Elected).Tsk).Period;
         END IF;

      END IF;


   END Do_Election;






END Scheduler.Dynamic_Priority.Muf.Laxity_Based ;


