------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------




WITH Text_Io;
USE Text_Io;


WITH Translate;
USE Translate;

WITH Tasks;
USE Tasks;

WITH Scheduler;
USE Scheduler;
USE Scheduler.Double_Tasks_Parameters;

WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;


WITH Unbounded_Strings;
USE Unbounded_Strings;





PACKAGE BODY Scheduler.Dynamic_Priority IS


   FUNCTION Build_Tcb (
         My_Scheduler : IN     Dynamic_Priority_Scheduler;
         A_Task       :        Task_Ptr)
     RETURN Tcb_Ptr IS
      A_Tcb : Dynamic_Priority_Tcb_Ptr;
   BEGIN
      A_Tcb:=NEW Dynamic_Priority_Tcb;
      Initialize(Tcb(A_Tcb.All), A_Task);
      Initialize(A_Tcb.All);
      RETURN Tcb_Ptr(A_Tcb);
   END Build_Tcb;


   PROCEDURE Initialize (
         A_Tcb : IN OUT Dynamic_Priority_Tcb) IS
   BEGIN
      A_Tcb.Dynamic_Deadline:=
         A_Tcb.Tsk.Deadline+A_Tcb.Wake_Up_Time;
   END Initialize;




   PROCEDURE Check_Before_Scheduling (
         My_Scheduler   : IN     Dynamic_Priority_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

   BEGIN
      NULL;
   END Check_Before_Scheduling;



   PROCEDURE Compute_Blocking_Time (
         My_Scheduler   : IN     Dynamic_Priority_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Blocking_Time  :    OUT Blocking_Time_Table) IS
   BEGIN
      RAISE Invalid_Scheduler;
   END Compute_Blocking_Time;



   PROCEDURE Bound_On_Processor_Utilization (
         My_Scheduler   : IN     Dynamic_Priority_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Result         : IN OUT Double;
         Msg            : IN OUT Unbounded_String) IS
   BEGIN
      Result:=1.0;
      Msg:=To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[1], page 8, ") & Lb_Theorem(
         Current_Language) &
         To_Unbounded_String("2). ") & Unbounded_Lf;

   END Bound_On_Processor_Utilization;



   FUNCTION Compute_Wi_A_T (
         My_Tasks : IN     Tasks_Set;
         Taski    : IN     Task_Ptr;
         A        : IN     Integer;
         T        : IN     Double)
     RETURN Double IS

      Result   : Double         := 0.0;
      Min      : Double         := 0.0;
      Iterator : Tasks_Iterator;
      Taskj    : Task_Ptr;
   BEGIN

      Reset_Iterator(My_Tasks, Iterator);
      LOOP
         Current_Element(My_Tasks, Taskj, Iterator);

         IF ((Taskj.Name /= Taski.Name) AND (Taskj.Deadline <= A + Taski.
                  Deadline) )
               THEN
            Min := 1.0 + Double'Floor(
               (Double(A) + Double(Taski.Deadline) - Double(Taskj.
                     Deadline))
               / Double(Periodic_Task_Ptr(Taskj).Period)
               ) ;

            IF ( Double'Ceiling(T/Double(Periodic_Task_Ptr(Taskj).Period)) <
                  Min )
                  THEN
               Min := Double'Ceiling(T/Double(Periodic_Task_Ptr(Taskj).
                     Period));
            END IF ;
            Result := Result + Min * Double(Taskj.Capacity) ;
         END IF;

         -- exit loop when there is no more task
         --
         EXIT WHEN Is_Last_Element(My_Tasks, Iterator);
         Next_Element(My_Tasks, Iterator);
      END LOOP ;

      RETURN Result ;
   END Compute_Wi_A_T ;

   FUNCTION Compute_Li_A (
         My_Scheduler : IN     Dynamic_Priority_Scheduler;
         My_Tasks     : IN     Tasks_Set;
         Taski        : IN     Task_Ptr;
         A            : IN     Integer)
     RETURN Double IS

      Li_A     : Double         := 0.1;
      Li_A_1   : Double         := 0.0;
      Iterator : Tasks_Iterator;
      Taskj    : Task_Ptr;
      Max_Cj   : Natural        := 0;
      Tmp      : Double         := 0.0;
      Min      : Double         := 0.0;
      Min1     : Double         := 0.0;
      Min2     : Double         := 0.0;
   BEGIN

      IF ( My_Scheduler.Is_Preemptive = Preemptive )
            THEN
         -- preemptive case
         --
         WHILE Li_A_1 /= Li_A LOOP

            Li_A := Li_A_1 ;

            Li_A_1 := Compute_Wi_A_T(My_Tasks, Taski, A, Li_A) ;

            CASE Taski.Task_Type IS
               WHEN Periodic_Type =>
                  Li_A_1 := Li_A_1 + ( 1.0 + Double'Floor( Double(A) /
                        Double(Periodic_Task_Ptr(Taski).Period) ) )
                     * Double(Taski.Capacity) ;
               WHEN OTHERS =>
                  RAISE Constraint_Error;
            END CASE;
         END LOOP ;

      ELSE
         -- non preemptive case
         --
         WHILE Li_A_1 /= Li_A LOOP

            Li_A := Li_A_1 ;
            Tmp := 0.0 ;
            Reset_Iterator(My_Tasks, Iterator);
            LOOP
               Current_Element(My_Tasks, Taskj, Iterator);

               IF ( (A + Taski.Deadline) < Taskj.Deadline )
                     THEN
                  IF ( Taskj.Capacity > Max_Cj)
                        THEN
                     Max_Cj := Taskj.Capacity ;
                  END IF ;
               END IF ;

               IF ((Taskj.Name /= Taski.Name) AND (Taskj.Deadline <= A +
                        Taski.Deadline) )
                     THEN
                  Min1 := 1.0
                     + Double'Floor( Li_A / Double(Periodic_Task_Ptr(
                           Taskj).Period) ) ;

                  Min2 := 1.0
                     + Double'Floor(Double(A) + Double(Taski.Deadline) -
                     Double(Taskj.Deadline))
                     / Double(Periodic_Task_Ptr(Taskj).Period) ;

                  Min := Min1 ;
                  IF(Min2 < Min) THEN
                     Min := Min2 ;
                  END IF ;

                  Tmp := Tmp + Min * Double(Taskj.Capacity)
                     +  Double'Floor(Double(A) / Double(Periodic_Task_Ptr(
                           Taskj).Period)) * Double(Taski.Capacity) ;
               END IF ;

               -- exit loop when there is no more task
               --
               EXIT WHEN Is_Last_Element(My_Tasks, Iterator);
               Next_Element(My_Tasks, Iterator);
            END LOOP ;

            Li_A_1 := Double(Max_Cj) + Tmp ;
         END LOOP ;

      END IF ;

      RETURN Li_A ;
   END Compute_Li_A ;


   -- This function may present an infinite loop for non-preemptive case
   -- It may come from the upper bound for a: (exit when a >= Li(a))
   --
   PROCEDURE Compute_Response_Time (
         My_Scheduler   : IN     Dynamic_Priority_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Response_Time  :    OUT Response_Time_Table) IS

      Tmp       : Tasks_Set;
      Iterator1 : Tasks_Iterator;
      Taski     : Task_Ptr;
      Iterator2 : Tasks_Iterator;
      Taskj     : Task_Ptr;
      I         : Response_Time_Range := 0;

      A : Integer := 0;

      B_Sup : Double  := 0.0;
      Li_A  : Double  := 0.0;
      K     : Natural := 0;

   BEGIN
      Double_Tasks_Parameters.Initialize(Response_Time);

      -- check if tasks are periodics
      --
      Periodic_Control(My_Tasks, Processor_Name);

      -- check start time
      --
      Start_Time_Control(My_Tasks, Processor_Name);

      -- check offset
      --
      Offset_Control(My_Tasks, Processor_Name);


      -- Check processor utilization first
      --
      IF (Processor_Utilization_Over_Period(My_Tasks,
               Processor_Name) > 1.0)
            THEN
         RAISE Processor_Utilization_Exceeded;
      END IF;


      -- Select task to proceed
      --
      Current_Processor_Name:=Processor_Name;
      Select_And_Copy( My_Tasks, Tmp, Select_Cpu'ACCESS );
      Reset_Iterator(My_Tasks, Iterator1);


      LOOP
         -- selection of the current task
         --
         Current_Element(Tmp, Taski, Iterator1);

         -- initialize response time for current task
         --
         Response_Time.Entries(I).Data := 0.0;
         Response_Time.Entries(I).Item := Taski;
         Response_Time.Nb_Entries := Response_Time.Nb_Entries + 1;

         IF ( My_Scheduler.Is_Preemptive = Preemptive )
               THEN
            Response_Time.Entries(I).Data := Double(Taski.Capacity);
         END IF ;

         -- compute S "on the fly"
         --
         Reset_Iterator(My_Tasks, Iterator2);

         LOOP
            Current_Element(Tmp, Taskj, Iterator2);

            K := 0 ;
            LOOP
               CASE Taskj.Task_Type IS
                  WHEN Periodic_Type =>

                     A:= (K * Periodic_Task_Ptr(Taskj).Period)
                        + Taskj.Deadline
                        - Taski.Deadline ;

                  WHEN OTHERS =>
                     RAISE Constraint_Error;
               END CASE;

               Li_A := Compute_Li_A(My_Scheduler, Tmp, Taski, A) ;

               IF ( (Li_A - Double(A) ) > Response_Time.Entries(I).Data )
                     THEN
                  Response_Time.Entries(I).Data := Li_A - Double(A) ;

                  -- PROBLEM
                  -- with non preemptive case response_time has no upper bound in some cases
                  -- Put(Natural(Response_Time.Entries(i).Data)'Img) ;
               END IF ;

               IF ( My_Scheduler.Is_Preemptive = Preemptive )
                     THEN
                  B_Sup := Li_A ;
               ELSE
                  B_Sup := Li_A ;
               END IF ;

               EXIT WHEN (Double(A) >= B_Sup) ;


               -- PROBLEM : the lines below try yo fix the problem described 10 lines
               -- before (see morin2003.doc)
               -- The "exit when K" statement is not a part of the EDF response time
               -- computation : it's just a way to be sure that the response
               -- time computation never loop for ever .... This Exit statement
               -- HAVE TO BE REMOVED BEFORE EDF response time computation will be
               -- strongly tested by users
               --
               --exit when Response_Time.Entries(I).Data > Double(
               --  Taski.Deadline);
               EXIT WHEN K > 10000;

               K := K + 1 ;
            END LOOP ;

            EXIT WHEN Is_Last_Element(Tmp, Iterator2);
            -- go to the next task
            --
            Next_Element(Tmp, Iterator2);
         END LOOP ;

         EXIT WHEN Is_Last_Element(Tmp, Iterator1);

         -- go to the next task
         --
         Next_Element(Tmp, Iterator1);

         I := I + 1 ;
      END LOOP ;


   END Compute_Response_Time;



   PROCEDURE Specific_Scheduler_Initialization (
         My_Scheduler   : IN OUT Dynamic_Priority_Scheduler;
         Si             : IN OUT Scheduling_Information;
         Processor_Name : IN     Unbounded_String;
         My_Tasks       : IN OUT Tasks_Set;
         My_Resources   : IN OUT Resources_Set;
         My_Buffers     : IN OUT Buffers_Set;
         My_Messages    : IN     Messages_Set;
         Msg            : IN OUT Unbounded_String) IS

   BEGIN
      NULL;
   END Specific_Scheduler_Initialization;



END Scheduler.Dynamic_Priority ;


