------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Framework_Config;
USE Framework_Config;
WITH Xml_Tag;
USE Xml_Tag;
WITH Double_Util;
USE Double_Util;
WITH Translate;
USE Translate;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;
WITH Unbounded_Strings;
USE Unbounded_Strings;


WITH Scheduler;
USE Scheduler;
USE Scheduler.Double_Tasks_Parameters;


WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;


WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;


PACKAGE BODY Scheduler.Fixed_Priority.Dm IS


   PROCEDURE Initialize (
         A_Scheduler : IN OUT Dm_Scheduler) IS
   BEGIN
      Reset(A_Scheduler);
      A_Scheduler.Name:= Deadline_Monotonic_Protocol;
   END Initialize;


   FUNCTION Copy (
         A_Scheduler : IN     Dm_Scheduler)
     RETURN Scheduler_Ptr IS

      Ptr : Dm_Scheduler_Ptr;

   BEGIN

      Ptr:= NEW Dm_Scheduler;

      Ptr.Name:=A_Scheduler.Name;
      Ptr.Is_Preemptive:=A_Scheduler.Is_Preemptive;
      Ptr.Quantum:=A_Scheduler.Quantum;
      Ptr.Previously_Elected:=A_Scheduler.Previously_Elected;
      Ptr.Used_Resource:=A_Scheduler.Used_Resource;

      RETURN Scheduler_Ptr(Ptr);

   END Copy;


   PROCEDURE Check_Before_Scheduling (
         My_Scheduler   : IN     Dm_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

   BEGIN

      Periodic_Control(My_Tasks, Processor_Name);

   END Check_Before_Scheduling;




   PROCEDURE Compute_Response_Time (
         My_Scheduler   : IN     Dm_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Response_Time  :    OUT Response_Time_Table) IS

   BEGIN


      -- check if tasks are periodics
      Periodic_Control(My_Tasks, Processor_Name);


      -- Check processor utilization first
      IF (Processor_Utilization_Over_Period(My_Tasks,
               Processor_Name) > 1.0)
            THEN
         RAISE Processor_Utilization_Exceeded;
      END IF;

      -- check start time
      Start_Time_Control(My_Tasks, Processor_Name);

      -- check offset
      Offset_Control(My_Tasks, Processor_Name);


      -- Compute response time
      --
      Scheduler.Fixed_Priority.Compute_Response_Time(
         Fixed_Priority_Scheduler(My_Scheduler),
         My_Tasks, Processor_Name, Msg, Response_Time);


   END Compute_Response_Time;


   PROCEDURE Utilization_Factor_Feasibility_Test (
         My_Scheduler   : IN     Dm_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Result         : IN OUT Unbounded_String) IS


      Bound  : Double           := 0.0;
      Utildi : Double           := 0.0;
      Msg    : Unbounded_String := Empty_String;

   BEGIN

      Utildi:=Processor_Utilization_Over_Deadline(My_Tasks,
         Processor_Name);
      Bound_On_Processor_Utilization(My_Scheduler,
         My_Tasks,
         Processor_Name, Bound, Msg);

      CASE Get_Preemptive(My_Scheduler) IS

         WHEN Preemptive =>

            Result := Result & Lb_Utilization_Bound1(Current_Language);
            Result := Result & Lb_Dm(Current_Language);

            IF  Utildi <= Bound THEN
               Result:= Result & Lb_Sched_Explanation1(Current_Language);
               Result := Result & Start_Fact_Bound & Format(Utildi) &
                  End_Fact_Bound ;
               Result := Result &  Lb_Sched_Explanation12(
                  Current_Language);
               Result := Result & Start_Fact_Bound & Format(Bound) &
                  End_Fact_Bound ;
            ELSE
               Result:= Result & Lb_Sched_Explanation2(Current_Language);
               Result := Result & Start_Fact_Bound & Format(Utildi) &
                  End_Fact_Bound ;
               Result := Result & Lb_Sched_Explanation22(Current_Language);
               Result := Result & Start_Fact_Bound & Format(Bound) &
                  End_Fact_Bound ;
            END IF;
            Result := Result & Start_Ref & To_Unbounded_String(" (") &
               Lb_See(
               Current_Language) &
               To_Unbounded_String("[7]).") & End_Ref &
               Unbounded_Lf;


         WHEN Not_Preemptive =>

            Result:=Result & Lb_Pb_Sched_Unknown(Current_Language) &
               Unbounded_Lf;
      END CASE;

   END Utilization_Factor_Feasibility_Test;








END Scheduler.Fixed_Priority.Dm ;


