------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Translate;
USE Translate;
WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;


WITH Text_Io;
USE Text_Io;

WITH Scheduler;
USE Scheduler;
USE Scheduler.Double_Tasks_Parameters;

WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;


PACKAGE BODY Scheduler.Fixed_Priority.Hpf IS


   FUNCTION Build_Tcb (
         My_Scheduler : IN     Hpf_Scheduler;
         A_Task       :        Task_Ptr)
     RETURN Tcb_Ptr IS
      A_Tcb : Hpf_Tcb_Ptr;
   BEGIN
      A_Tcb:=NEW Hpf_Tcb;
      Initialize(Tcb(A_Tcb.All), A_Task);
      Initialize(Fixed_Priority_Tcb(A_Tcb.All));
      Initialize(A_Tcb.All);
      RETURN Tcb_Ptr(A_Tcb);
   END Build_Tcb;


   PROCEDURE Initialize (
         A_Tcb : IN OUT Hpf_Tcb) IS
   BEGIN
      A_Tcb.Task_Quantum:= 0;
   END Initialize;


   PROCEDURE Initialize (
         A_Scheduler : IN OUT Hpf_Scheduler) IS
   BEGIN
      Reset(A_Scheduler);
      A_Scheduler.Name:= Posix_1003_Highest_Priority_First_Protocol;
   END Initialize;



   FUNCTION Copy (
         A_Scheduler : IN     Hpf_Scheduler)
     RETURN Scheduler_Ptr IS

      Ptr : Hpf_Scheduler_Ptr;

   BEGIN

      Ptr:= NEW Hpf_Scheduler;

      Ptr.Name:=A_Scheduler.Name;
      Ptr.Is_Preemptive:=A_Scheduler.Is_Preemptive;
      Ptr.Quantum:=A_Scheduler.Quantum;
      Ptr.Previously_Elected:=A_Scheduler.Previously_Elected;
      Ptr.Used_Resource:=A_Scheduler.Used_Resource;
      Ptr.Priority_Fifos:=A_Scheduler.Priority_Fifos;

      RETURN Scheduler_Ptr(Ptr);

   END Copy;




   PROCEDURE Check_Before_Scheduling (
         My_Scheduler   : IN     Hpf_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

   BEGIN

      NULL;

   END Check_Before_Scheduling;




   PROCEDURE Compute_Response_Time (
         My_Scheduler   : IN     Hpf_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Response_Time  :    OUT Response_Time_Table) IS


   BEGIN

      -- Check if tasks are periodics
      --      Periodic_Control(My_Tasks, Processor_Name);

      -- Check processor utilization first
      --
      IF (Processor_Utilization_Over_Period(My_Tasks,
               Processor_Name) > 1.0)
            THEN
         RAISE Processor_Utilization_Exceeded;
      END IF;

      -- Check start time
      --
      Start_Time_Control(My_Tasks, Processor_Name);

      -- Check offset
      --
      Offset_Control(My_Tasks, Processor_Name);


      -- Compute response time
      --
      Scheduler.Fixed_Priority.Compute_Response_Time (
         Fixed_Priority_Scheduler(My_Scheduler), My_Tasks, Processor_Name,
         Msg, Response_Time);


   END Compute_Response_Time;





   PROCEDURE Put_Tcb (
         E : Hpf_Tcb_Ptr) IS
   BEGIN
      NULL;
   END Put_Tcb;




   PROCEDURE Specific_Scheduler_Initialization (
         My_Scheduler   : IN OUT Hpf_Scheduler;
         Si             : IN OUT Scheduling_Information;
         Processor_Name : IN     Unbounded_String;
         My_Tasks       : IN OUT Tasks_Set;
         My_Resources   : IN OUT Resources_Set;
         My_Buffers     : IN OUT Buffers_Set;
         My_Messages    : IN     Messages_Set;
         Msg            : IN OUT Unbounded_String) IS


      Index : Tasks_Range := 0;

   BEGIN
      FOR I IN Priority_Range LOOP
         Reset(My_Scheduler.Priority_Fifos(I));
      END LOOP;



      FOR I IN 0 .. Si.Number_Of_Tasks-1 LOOP
         IF Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name THEN
            Fixed_Priority_Tcb_Ptr(Si.Tcbs(I)).Current_Priority:=
               Si.Tcbs(I).Tsk.Priority;
            Hpf_Tcb_Ptr(Si.Tcbs(I)).Task_Quantum:=My_Scheduler.
               Quantum;
         END IF;
      END LOOP;



      -- Insert task info fifos according to
      -- their priority level
      --
      LOOP
         IF Si.Tcbs(Index).Tsk.Cpu_Name = Processor_Name THEN
            Insert(My_Scheduler.Priority_Fifos(Hpf_Tcb_Ptr(Si.Tcbs(Index)).
                  Current_Priority), Hpf_Tcb_Ptr(Si.Tcbs(Index)));
         END IF;

         Index:=Index+1;
         EXIT WHEN Si.Tcbs(Index)=NULL;
      END LOOP;


      -- Set priority ceiling of resources (only for PCP resource
      --
      Compute_Ceiling_Of_Resources (
         My_Scheduler  ,
         Si            ,
         Processor_Name ,
         My_Tasks      ,
         My_Resources              );



   END Specific_Scheduler_Initialization;




   PROCEDURE Do_Election (
         My_Scheduler      : IN OUT Hpf_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Msg               : IN OUT Unbounded_String;
         Current_Time      : IN     Natural;
         Processor_Name    : IN     Unbounded_String;
         My_Dependencies   : IN     Tasks_Dependencies_Ptr;
         With_Offsets      : IN     Boolean;
         With_Precedencies : IN     Boolean;
         With_Resources    : IN     Boolean;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table;
         Elected           : IN OUT Tasks_Range;
         No_Task           : IN OUT Boolean) IS


      Ready,
      Found             : Boolean;
      Selected_Priority : Priority_Range;
      Index             : Tasks_Range    := 0;
      Current           : Hpf_Tcb_Ptr;
      Elected_Task      : Hpf_Tcb_Ptr    := NULL;
      Max_Cpu_Usage     : Natural        := Natural'Last;


   BEGIN


      -- For each task, call "check_ressources" to
      -- take care of priority modification (PCP and PIP)
      --

      FOR I IN 0..Si.Number_Of_Tasks-1 LOOP

         IF (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) THEN
            IF (Si.Tcbs(I).Wake_Up_Time <= Current_Time)
                  AND (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                  THEN
               IF (With_Offsets=False) OR
                     Check_Offset(Si.Tcbs(I), Current_Time)
                     THEN
                  IF (With_Precedencies=False) OR
                        Check_Precedencies(Si, My_Dependencies,
                        Current_Time,
                        Si.Tcbs(I))
                        THEN
                     IF With_Resources THEN
                        Check_Resource(My_Scheduler, Si, Result,
                           Current_Time, Si.Tcbs(I),
                           Fixed_Priority_Tcb_Ptr(Si.Tcbs(I)).Is_Ready,
                           Event_To_Generate);
                     END IF;
                  END IF;
               END IF;
            END IF;
         END IF;
      END LOOP;


      -- Scan fifos to find the highest priority level
      -- with a ready task
      -- Warning : WE DO NOT CHANGE TASK POSITION IN THE
      --        FIFOS
      --
      Selected_Priority:=Priority_Range'Last;
      Found:=False;
      LOOP

         IF NOT Is_Empty(My_Scheduler.Priority_Fifos(Selected_Priority))

               THEN

            FOR I IN 1..Get_Size(My_Scheduler.Priority_Fifos(
                     Selected_Priority)) LOOP

               Current:=Consult(My_Scheduler.Priority_Fifos(
                     Selected_Priority));

               -- is the task ready ?
               --
               Ready:=False;
               IF (Current.Wake_Up_Time <= Current_Time) AND
                     (Current.Rest_Of_Capacity /= 0)
                     THEN
                  IF (With_Offsets=False) OR
                        Check_Offset(Tcb_Ptr(Current), Current_Time)
                        THEN
                     IF (With_Precedencies=False) OR
                           Check_Precedencies(Si, My_Dependencies,
                           Current_Time,
                           Tcb_Ptr(Current))
                           THEN
                        IF (With_Resources=False) OR
                              Fixed_Priority_Tcb_Ptr(
                              Current).Is_Ready
                              THEN

                           -- the priority level is selected
                           --
                           Ready:=True;
                           Found:=True;
                           EXIT;
                        END IF;
                     END IF;
                  END IF;
               END IF;

               -- The current task is not ready ...
               -- put it the queue of the fifo
               -- and look for the Next element
               -- of the fifo
               --
               IF (Ready=False)
                     THEN
                  Extract(My_Scheduler.Priority_Fifos(Selected_Priority),
                     Current);
                  Insert(My_Scheduler.Priority_Fifos(Selected_Priority),
                     Current);
               END IF;


            END LOOP;

         END IF;



         -- A fifo is already selected ... get out !!
         -- and never come back .........
         --
         IF Found
               THEN
            No_Task:=False;
            EXIT;
         END IF;


         -- No task found = no task to schedule
         --
         IF (Selected_Priority = Priority_Range'First)
               THEN
            No_Task:=True;
            RETURN;
         END IF;


         Selected_Priority:=
            Selected_Priority-1;

      END LOOP;






      -- Priority level is selected : find the task
      -- in the head of the fifo : this task is the elected
      -- task
      --

      -- Task sharing policies
      --
      IF (Selected_Priority = 0)
            THEN
         -- Loop and looking for the less consuming
         -- task
         --
         Max_Cpu_Usage:=Natural'Last;
         FOR I IN 1..Get_Size(My_Scheduler.Priority_Fifos(
                  Selected_Priority)) LOOP

            Extract(My_Scheduler.Priority_Fifos(Selected_Priority),
               Current);
            Insert(My_Scheduler.Priority_Fifos(Selected_Priority), Current);
            IF (Current.Wake_Up_Time <= Current_Time) AND
                  (Current.Rest_Of_Capacity /= 0)
                  THEN
               IF (With_Offsets=False) OR
                     Check_Offset(Tcb_Ptr(Current), Current_Time)
                     THEN
                  IF (With_Resources=False) OR Fixed_Priority_Tcb_Ptr(
                        Current).Is_Ready
                        THEN

                     IF (Current.Used_Cpu <= Max_Cpu_Usage)
                           THEN
                        Max_Cpu_Usage:=Current.Used_Cpu;
                        Elected_Task:=Current;
                     END IF;
                  END IF;
               END IF;
            END IF;
         END LOOP;



      ELSE

         -- SCHED_FIFO or SCHED_RR task
         --
         Elected_Task:=Consult(My_Scheduler.Priority_Fifos(
               Selected_Priority));

         -- is the task quantum exausted ?
         -- if so and if SCHED_RR is set
         -- the task have to be moved to
         -- the queue of the fifo
         --

         IF (Elected_Task.Tsk.Policy = Sched_Rr) THEN
            IF (My_Scheduler.Quantum>0) THEN


               Elected_Task.Task_Quantum:=Elected_Task.Task_Quantum-1;

               IF (Elected_Task.Task_Quantum=0)
                     THEN
                  Elected_Task.Task_Quantum:=My_Scheduler.Quantum;
                  Extract(My_Scheduler.Priority_Fifos(Selected_Priority),
                     Elected_Task );
                  Insert(My_Scheduler.Priority_Fifos(Selected_Priority),
                     Elected_Task);
               END IF;
            END IF;
         END IF;
      END IF;



      IF (Elected_Task /= NULL)
            THEN

         Index:=0;
         LOOP
            IF (Si.Tcbs(Index).Tsk.Name
                  = Elected_Task.Tsk.Name)
                  THEN
               Elected:=Index;
               EXIT;
            END IF;

            Index:=Index+1;
            EXIT WHEN Si.Tcbs(Index)=NULL;
         END LOOP;

      END IF;


   END Do_Election;



   PROCEDURE Dispatched_Change_Current_Priority (
         My_Scheduler : IN OUT Hpf_Scheduler;
         A_Tcb        : IN     Fixed_Priority_Tcb_Ptr;
         New_Priority :        Priority_Range) IS


   BEGIN
      Extract_Any_Where(My_Scheduler.Priority_Fifos(
            A_Tcb.Current_Priority),Hpf_Tcb_Ptr(A_Tcb));

      A_Tcb.Current_Priority:=New_Priority;
      Insert(My_Scheduler.Priority_Fifos(A_Tcb.Current_Priority),
         Hpf_Tcb_Ptr(A_Tcb));

   END Dispatched_Change_Current_Priority;


   PROCEDURE Utilization_Factor_Feasibility_Test (
         My_Scheduler   : IN     Hpf_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Result         : IN OUT Unbounded_String) IS

   BEGIN
      Result:=Result & Lb_Pb_Sched_Unknown(Current_Language) &
         Unbounded_Lf;
   END Utilization_Factor_Feasibility_Test;



END Scheduler.Fixed_Priority.Hpf ;


