------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Framework_Config;
USE Framework_Config;
WITH Xml_Tag;
USE Xml_Tag;
WITH Double_Util;
USE Double_Util;
WITH Translate;
USE Translate;
WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;
WITH Unbounded_Strings;
USE Unbounded_Strings;





WITH Text_Io;
USE  Text_Io;


WITH Scheduler;
USE Scheduler;
USE Scheduler.Double_Tasks_Parameters;


WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;




PACKAGE BODY Scheduler.Fixed_Priority.Rm IS


   PROCEDURE Initialize (
         A_Scheduler : IN OUT Rm_Scheduler) IS
   BEGIN
      Reset(A_Scheduler);
      A_Scheduler.Name:= Rate_Monotonic_Protocol;
   END Initialize;


   FUNCTION Copy (
         A_Scheduler : IN     Rm_Scheduler)
     RETURN Scheduler_Ptr IS

      Ptr : Rm_Scheduler_Ptr;

   BEGIN

      Ptr:= NEW Rm_Scheduler;

      Ptr.Name:=A_Scheduler.Name;
      Ptr.Is_Preemptive:=A_Scheduler.Is_Preemptive;
      Ptr.Quantum:=A_Scheduler.Quantum;
      Ptr.Previously_Elected:=A_Scheduler.Previously_Elected;
      Ptr.Used_Resource:=A_Scheduler.Used_Resource;

      RETURN Scheduler_Ptr(Ptr);

   END Copy;



   PROCEDURE Check_Before_Scheduling (
         My_Scheduler   : IN     Rm_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

   BEGIN

      Periodic_Control(My_Tasks, Processor_Name);
      Have_Deadlines_Equal_Than_Periods_Control
         (My_Tasks, Processor_Name);

   END Check_Before_Scheduling;







   PROCEDURE Compute_Response_Time (
         My_Scheduler   : IN     Rm_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Response_Time  :    OUT Response_Time_Table) IS

   BEGIN


      -- check if tasks are periodic
      --
      Periodic_Control(My_Tasks, Processor_Name);

      -- check start time
      --
      Start_Time_Control(My_Tasks, Processor_Name);

      -- check offset
      --
      Offset_Control(My_Tasks, Processor_Name);


      -- check if deadlines are equal  than period
      --
      Have_Deadlines_Equal_Than_Periods_Control
         (My_Tasks, Processor_Name);


      -- Check processor utilization first
      --
      IF (Processor_Utilization_Over_Period(My_Tasks,
               Processor_Name) > 1.0)
            THEN
         RAISE Processor_Utilization_Exceeded;
      END IF;


      -- Compute response time
      --

      Scheduler.Fixed_Priority.Compute_Response_Time(
         Fixed_Priority_Scheduler(My_Scheduler),
         My_Tasks, Processor_Name, Msg, Response_Time);
   END Compute_Response_Time;



   PROCEDURE Utilization_Factor_Feasibility_Test (
         My_Scheduler   : IN     Rm_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Result         : IN OUT Unbounded_String) IS

      Tmp               : Tasks_Set;
      Bound             : Double           := 0.0;
      Util              : Double           := 0.0;
      Msg,
      Msg2              : Unbounded_String := Empty_String;
      Verification      : Boolean          := True;
      A_Task            : Task_Ptr;
      My_Task_Iterator  : Tasks_Iterator;
      A_Task2           : Task_Ptr;
      My_Task_Iterator2 : Tasks_Iterator;
      Compute           : Double           := 0.0;
      Max_Bi            : Double           := 0.0;
      I_Bound           : Double           := 0.0;
      Task_Number       : Natural          := 0;

   BEGIN

      Util:=Processor_Utilization_Over_Period(My_Tasks,
         Processor_Name);

      Bound_On_Processor_Utilization(My_Scheduler,
         My_Tasks,
         Processor_Name, Bound, Msg);

      CASE Get_Preemptive(My_Scheduler) IS

         WHEN Preemptive =>


            Result := Result & Lb_Utilization_Bound1(Current_Language)
               &  Lb_Rm(Current_Language);

            IF  Util <= Bound THEN
               Result:= Result & Lb_Sched_Explanation1(Current_Language);
               Result := Result & Start_Fact_Bound & Format(Util) &
                  End_Fact_Bound ;
               Result := Result & Lb_Sched_Explanation12(Current_Language);
               Result := Result & Start_Fact_Bound & Format(Bound) &
                  End_Fact_Bound ;
            ELSE
               Result:= Result & Lb_Sched_Explanation3(Current_Language);
               Result := Result & Start_Fact_Bound & Format(Util) &
                  End_Fact_Bound ;
               Result := Result & Lb_Sched_Explanation22(Current_Language);
               Result := Result & Start_Fact_Bound & Format(Bound) &
                  End_Fact_Bound ;
            END IF;
            Result := Result & Start_Ref & Msg & End_Ref &
               Unbounded_Lf;

         WHEN Not_Preemptive =>

            Result := Result & Lb_Utilization_Bound11(Current_Language)
               &  Lb_Rm(Current_Language);

            Current_Processor_Name:=Processor_Name;
            Select_And_Copy( My_Tasks, Tmp, Select_Cpu'ACCESS );

            Sort(Tmp, Increasing_Period'ACCESS);


            Msg2:= Empty_String;

            -- Main loop of all tasks
            --
            Verification:=True;
            Reset_Iterator(Tmp, My_Task_Iterator);
            LOOP
               Current_Element(Tmp, A_Task, My_Task_Iterator);
               IF (A_Task.Cpu_Name = Processor_Name) THEN
                  Compute:=0.0;

                  -- Loop until the current task
                  --
                  Reset_Iterator(Tmp, My_Task_Iterator2);
                  Task_Number:=0;
                  LOOP
                     Current_Element(Tmp, A_Task2, My_Task_Iterator2);
                     IF (A_Task2.Cpu_Name = Processor_Name) THEN
                        Task_Number:=Task_Number+1;
                        Compute:=Compute + Double(A_Task2.Capacity)/
                           Double(
                           Periodic_Task_Ptr(A_Task2).Period);
                     END IF;
                     EXIT WHEN A_Task.Name=A_Task2.Name;
                     Next_Element(Tmp, My_Task_Iterator2);
                  END LOOP;

                  -- Looking for the max Bi
                  --
                  Max_Bi:=0.0;
                  LOOP
                     EXIT WHEN Is_Last_Element(Tmp, My_Task_Iterator2);
                     Next_Element(Tmp, My_Task_Iterator2);
                     Current_Element(Tmp, A_Task2, My_Task_Iterator2);
                     IF (A_Task2.Cpu_Name = Processor_Name) THEN

                        IF Double(A_Task2.Capacity)>Max_Bi THEN
                           Max_Bi:=Double(A_Task2.Capacity);
                        END IF;
                     END IF;
                  END LOOP;
                  Compute:=Compute + Max_Bi/Double(Periodic_Task_Ptr(
                        A_Task).
                     Period);
                  I_Bound:= (Double (Task_Number))* (Pow (2.0,(1.0/(
                              Double (
                                 Task_Number)))) - 1.0);

                  IF Compute > I_Bound THEN
                     Verification:=False;
                     Msg2:=Msg2 & A_Task.Name & " ";
                  END IF;
               END IF;
               EXIT WHEN Is_Last_Element(Tmp, My_Task_Iterator);
               Next_Element(My_Tasks, My_Task_Iterator);
            END LOOP;

            -- Checks the result
            --
            IF Verification THEN
               Result:=Result & Lb_Sched_Explanation5(Current_Language);
               Result:=Result & Lb_Sched_Explanation52(Current_Language);
               Result:=Result & Lb_See(Current_Language) & "[3]."
                  & Unbounded_Lf;
            ELSE
               Result:=Result & Lb_Sched_Explanation8(Current_Language);
               Result:=Result & Lb_Sched_Explanation81(Current_Language) &
                  To_Unbounded_String(" (") & Lb_See(Current_Language) &
                  "[3])."
                  & Unbounded_Lf & Lb_Sched_Explanation82(
                  Current_Language) &
                  Msg2 & Unbounded_Lf ;
            END IF;

      END CASE;

      Free(Tmp);

   END Utilization_Factor_Feasibility_Test;








END Scheduler.Fixed_Priority.Rm ;


