------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Text_Io;
USE Text_Io;


WITH Translate;
USE Translate;

WITH Tasks;
USE Tasks;

WITH Resources;
USE Resources;

WITH Scheduler;
USE Scheduler;
USE Scheduler.Double_Tasks_Parameters;


WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;


WITH Unbounded_Strings;
USE  Unbounded_Strings;



PACKAGE BODY Scheduler.Fixed_Priority  IS


   FUNCTION Build_Tcb (
         My_Scheduler : IN     Fixed_Priority_Scheduler;
         A_Task       :        Task_Ptr)
     RETURN Tcb_Ptr IS
      A_Tcb : Fixed_Priority_Tcb_Ptr;
   BEGIN
      A_Tcb:=NEW Fixed_Priority_Tcb;
      Initialize(Tcb(A_Tcb.All), A_Task);
      Initialize(A_Tcb.All);
      RETURN Tcb_Ptr(A_Tcb);
   END Build_Tcb;


   PROCEDURE Initialize (
         A_Tcb : IN OUT Fixed_Priority_Tcb) IS
   BEGIN
      A_Tcb.Current_Priority:=A_Tcb.Tsk.Priority;
   END Initialize;


   PROCEDURE Do_Election (
         My_Scheduler      : IN OUT Fixed_Priority_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Msg               : IN OUT Unbounded_String;
         Current_Time      : IN     Natural;
         Processor_Name    : IN     Unbounded_String;
         My_Dependencies   : IN     Tasks_Dependencies_Ptr;
         With_Offsets      : IN     Boolean;
         With_Precedencies : IN     Boolean;
         With_Resources    : IN     Boolean;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table;
         Elected           : IN OUT Tasks_Range;
         No_Task           : IN OUT Boolean) IS


      Highiest_Priority : Natural     := Natural'First;
      I                 : Tasks_Range := 0;
   BEGIN


      -- For each task, call "check_resources" to
      -- take care of priority modification (PCP and PIP)
      --
      LOOP

         IF (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) THEN
            IF (Si.Tcbs(I).Wake_Up_Time <= Current_Time)
                  AND (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                  THEN
               IF (With_Offsets=False) OR
                     Check_Offset(Si.Tcbs(I), Current_Time)
                     THEN
                  IF (With_Precedencies=False) OR
                        Check_Precedencies(Si, My_Dependencies,
                        Current_Time,
                        Si.Tcbs(I))
                        THEN
                     IF With_Resources THEN
                        Check_Resource(My_Scheduler, Si, Result,
                           Current_Time, Si.Tcbs(I),
                           Fixed_Priority_Tcb_Ptr(Si.Tcbs(I)).Is_Ready,
                           Event_To_Generate);
                     END IF;
                  END IF;
               END IF;
            END IF;
         END IF;


         I:=I+1;
         EXIT WHEN Si.Tcbs(I)=NULL;
      END LOOP;


      -- We can compute scheduling since priority are ok
      --
      I:=0;
      LOOP

         IF (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) THEN
            IF (Si.Tcbs(I).Wake_Up_Time <= Current_Time) AND
                  (Natural(Fixed_Priority_Tcb_Ptr(Si.Tcbs(I)).
                     Current_Priority) > Highiest_Priority) AND
                  (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                  THEN
               IF (With_Offsets=False) OR
                     Check_Offset(Si.Tcbs(I), Current_Time)
                     THEN
                  IF (With_Precedencies=False) OR
                        Check_Precedencies(Si, My_Dependencies,
                        Current_Time,
                        Si.Tcbs(I))
                        THEN
                     IF (With_Resources=False) OR Fixed_Priority_Tcb_Ptr(
                           Si.Tcbs(I)).Is_Ready
                           THEN

                        Highiest_Priority:=Natural(Fixed_Priority_Tcb_Ptr(
                              Si.Tcbs(I)).Current_Priority);
                        Elected:=I;
                     END IF;
                  END IF;
               END IF;
            END IF;
         END IF;


         I:=I+1;
         EXIT WHEN Si.Tcbs(I)=NULL;
      END LOOP;

      IF Highiest_Priority = Natural'First
            THEN
         No_Task:=True;
      ELSE
         No_Task:=False;
      END IF;

   END Do_Election;



   PROCEDURE Bound_On_Processor_Utilization (
         My_Scheduler   : IN     Fixed_Priority_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Result         : IN OUT Double;
         Msg            : IN OUT Unbounded_String) IS

      Nb : CONSTANT Double := Double (Get_Number_Of_Task_From_Processor (My_Tasks,
        Processor_Name));
   BEGIN

      IF NOT Is_Harmonic(My_Tasks, Processor_Name)
            THEN
         Msg:=To_Unbounded_String(" (") & Lb_See(Current_Language) &
            To_Unbounded_String("[1], page 16, ") & Lb_Theorem(
            Current_Language) &
            To_Unbounded_String("8). ") & Unbounded_Lf;
         Result:=Nb*( Pow(2.0,(1.0/Nb)) - 1.0);
      ELSE
         Msg:=To_Unbounded_String(" (") & Lb_See(Current_Language) &
            To_Unbounded_String("[19], page 13).")  & Unbounded_Lf;
         Result:=1.0;
      END IF;


   END Bound_On_Processor_Utilization;




   FUNCTION Compute_Workload (
         My_Tasks : IN     Tasks_Set;
         Taski    : IN     Task_Ptr;
         Q        : IN     Integer)
     RETURN Double IS

      Iterator2 : Tasks_Iterator;
      Taskj     : Task_Ptr;
      Wiq_K,
      Wiq_K1    : Double;

   BEGIN

      Wiq_K  := -0.1 ;
      Wiq_K1 :=  0.0 ;

      -- approximations for Wiq
      -- exit loop when approximations converge
      --
      WHILE Wiq_K /= Wiq_K1  LOOP

         Reset_Iterator(My_Tasks, Iterator2);

         Wiq_K  := Wiq_K1 ;
         Wiq_K1 := Double((Q+1)*Taski.Capacity) + Double(Taski.
            Blocking_Time) ;

         -- Iterate on all tasks
         --
         LOOP
            Current_Element(My_Tasks, Taskj, Iterator2);

            -- If Taskj is in hp(Taski)
            --
            IF(Taskj.Priority> Taski.Priority)
                  THEN
               Wiq_K1 :=   Wiq_K1
                  + Double'Ceiling((Wiq_K + Double(Periodic_Task_Ptr(
                           Taskj).Jitter))
                  / Double(Periodic_Task_Ptr(Taskj).Period))
                  * Double(Taskj.Capacity) ;
            END IF ;

            -- exit loop when there is no more task
            --
            EXIT WHEN Is_Last_Element(My_Tasks, Iterator2);

            Next_Element(My_Tasks, Iterator2);

         END LOOP ;

      END LOOP ;

      RETURN Wiq_K1;

   END Compute_Workload;


   PROCEDURE Compute_Response_Time (
         My_Scheduler   : IN     Fixed_Priority_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Response_Time  :    OUT Response_Time_Table) IS

      Tmp       : Tasks_Set;
      Iterator1 : Tasks_Iterator;
      Taski     : Task_Ptr;
      I         : Response_Time_Range := 0;
      Wiq,
      Ri        : Double              := 0.0;
      Q         : Integer;

   BEGIN



      Double_Tasks_Parameters.Initialize(Response_Time);

      Current_Processor_Name:=Processor_Name;
      Select_And_Copy( My_Tasks, Tmp, Select_Cpu'ACCESS );


      -- Set priority according to the scheduler
      --
      IF (My_Scheduler.Name = Deadline_Monotonic_Protocol)
            THEN
         Set_Priority_According_To_Dm(My_Scheduler, Tmp);
      ELSE
         IF (My_Scheduler.Name = Rate_Monotonic_Protocol)
               THEN
            Set_Priority_According_To_Rm(My_Scheduler, Tmp);
         END IF;
      END IF;


      Sort(Tmp, Increasing_Priority'ACCESS);


      -- Bibliographical references
      --
      Msg:=To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[2], page 3, ") & Lb_Equation(
         Current_Language);

      IF ( My_Scheduler.Is_Preemptive = Not_Preemptive )
            THEN
         Msg:= Msg & To_Unbounded_String("4 ") & Lb_And(Current_Language) &
            To_Unbounded_String("[2], page 29, ") & Lb_Theorem(
            Current_Language)
            & To_Unbounded_String("15). ");
      ELSE
         Msg:=Msg & To_Unbounded_String("4). ");
      END IF ;



      Reset_Iterator(Tmp, Iterator1);

      LOOP

         -- Selection of the current task
         --
         Current_Element(Tmp, Taski, Iterator1);

         -- Initialize response time for current task
         --
         Response_Time.Entries(I).Data := 0.0;
         Response_Time.Entries(I).Item := Taski;
         Response_Time.Nb_Entries := Response_Time.Nb_Entries+1;

         -- Find the maximum for Ri in [0..Q]
         --
         Q := 0 ;
         LOOP

            Wiq := Compute_Workload(Tmp, Taski, Q);

            -- Ri = Wiq + Ji - qTi
            --
            CASE Taski.Task_Type IS
               WHEN Periodic_Type =>
                  Ri :=   Wiq
                     + Double(Periodic_Task_Ptr(Taski).Jitter)
                     - Double(Q) * Double(Periodic_Task_Ptr(Taski).Period) ;
               WHEN OTHERS =>
                  RAISE Constraint_Error;
            END CASE;


            -- Is Ri greater than current maximum for Ri ?
            --
            IF ( Ri > Response_Time.Entries(I).Data )
                  THEN
               Response_Time.Entries(I).Data := Ri ;
            END IF ;

            -- Exit loop when Q is found ( WiQ<=(Q+1)Ti )
            --
            CASE Taski.Task_Type IS
               WHEN Periodic_Type =>
                  EXIT WHEN Wiq <= Double(Q+1) *
                     Double(Periodic_Task_Ptr(Taski).Period);
               WHEN OTHERS =>
                  RAISE Constraint_Error;
            END CASE;

            Q := Q + 1 ;

         END LOOP ;

         -- If the scheduler is non premptive add Ci to Ri
         --
         IF ( My_Scheduler.Is_Preemptive = Not_Preemptive )
               THEN
            Response_Time.Entries(I).Data := Response_Time.Entries(I).
               Data + Double(Periodic_Task_Ptr(Taski).Capacity) ;
         END IF ;

         EXIT WHEN Is_Last_Element(Tmp, Iterator1);

         -- Go to the next task
         --
         Next_Element(Tmp, Iterator1);
         I:=I+1 ;

      END LOOP ;

   END Compute_Response_Time;



   PROCEDURE Set_Priority_According_To_Dm (
         My_Scheduler   : IN     Fixed_Priority_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String         := Empty_String) IS

      Iterator1    : Tasks_Iterator;
      Task1        : Task_Ptr;
      Iterator2    : Tasks_Iterator;
      Task2        : Task_Ptr;
      Current_Prio : Priority_Range := 1;
      Tmp          : Tasks_Set;

   BEGIN

      IF Processor_Name = Empty_String THEN
         Duplicate(My_Tasks, Tmp);
      ELSE
         Current_Processor_Name:=Processor_Name;
         Select_And_Copy(My_Tasks, Tmp, Select_Cpu'ACCESS );
      END IF;


      Periodic_Control(Tmp, Processor_Name);

      Sort(Tmp, Decreasing_Deadline'ACCESS);



      -- Assign priorities
      --
      Reset_Iterator(Tmp, Iterator1);

      LOOP
         Current_Element(Tmp, Task1, Iterator1);

         IF (Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String)
               THEN

            Task1.Priority:=Current_Prio;
            Current_Prio:=Current_Prio+1;
         END IF;

         EXIT WHEN Is_Last_Element(Tmp, Iterator1);
         Next_Element(Tmp, Iterator1);

      END LOOP;

      -- Copy resulting task objects in My_Tasks
      --
      Reset_Iterator(Tmp, Iterator1);
      LOOP
         Current_Element(Tmp, Task1, Iterator1);

         Reset_Iterator(My_Tasks, Iterator2);
         LOOP
            Current_Element(My_Tasks, Task2, Iterator2);

            IF (Task2.Name = Task1.Name) THEN
               Task2.Priority:=Task1.Priority;
            END IF;

            EXIT WHEN Task2.Name=Task1.Name;
            EXIT WHEN Is_Last_Element(My_Tasks, Iterator2);
            Next_Element(My_Tasks, Iterator2);
         END LOOP;

         EXIT WHEN Is_Last_Element(Tmp, Iterator1);
         Next_Element(Tmp, Iterator1);
      END LOOP;

      Free(Tmp);

   END Set_Priority_According_To_Dm;



   PROCEDURE Set_Priority_According_To_Rm (
         My_Scheduler   : IN     Fixed_Priority_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String         := Empty_String) IS

      Iterator1    : Tasks_Iterator;
      Task1        : Task_Ptr;
      Iterator2    : Tasks_Iterator;
      Task2        : Task_Ptr;
      Current_Prio : Priority_Range := 1;
      Tmp          : Tasks_Set;

   BEGIN

      IF Processor_Name = Empty_String THEN
         Duplicate(My_Tasks, Tmp);
      ELSE
         Current_Processor_Name:=Processor_Name;
         Select_And_Copy(My_Tasks, Tmp, Select_Cpu'ACCESS );
      END IF;

      Periodic_Control(Tmp, Processor_Name);


      Sort(Tmp, Decreasing_Period'ACCESS);


      -- Assign priorities
      --
      Reset_Iterator(Tmp, Iterator1);
      LOOP
         Current_Element(Tmp, Task1, Iterator1);

         IF (Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String)
               THEN

            Task1.Priority:=Current_Prio;
            Current_Prio:=Current_Prio+1;
         END IF;

         EXIT WHEN Is_Last_Element(Tmp, Iterator1);
         Next_Element(Tmp, Iterator1);

      END LOOP;

      -- Copy resulting task objects in My_Tasks
      --
      Reset_Iterator(Tmp, Iterator1);
      LOOP
         Current_Element(Tmp, Task1, Iterator1);

         Reset_Iterator(My_Tasks, Iterator2);
         LOOP
            Current_Element(My_Tasks, Task2, Iterator2);

            IF (Task2.Name = Task1.Name) THEN
               Task2.Priority:=Task1.Priority;
            END IF;

            EXIT WHEN Task2.Name=Task1.Name;
            EXIT WHEN Is_Last_Element(My_Tasks, Iterator2);
            Next_Element(My_Tasks, Iterator2);
         END LOOP;

         EXIT WHEN Is_Last_Element(Tmp, Iterator1);
         Next_Element(Tmp, Iterator1);
      END LOOP;

      Free(Tmp);

   END Set_Priority_According_To_Rm;



   PROCEDURE Compute_Ceiling_Of_Resources (
         My_Scheduler   : IN OUT Fixed_Priority_Scheduler;
         Si             : IN OUT Scheduling_Information;
         Processor_Name : IN     Unbounded_String;
         My_Tasks       : IN OUT Tasks_Set;
         My_Resources   : IN OUT Resources_Set) IS


   BEGIN
      -- Private scheduler data
      --
      My_Scheduler.Used_Resource := False;


      -- Set priority ceiling of resources (only for PCP and IPCP resources)
      --
      FOR K IN 0 .. Si.Number_Of_Resources-1 LOOP

         IF (Si.Shared_Resources(K).Shared.Protocol =
               Priority_Ceiling_Protocol) OR (Si.Shared_Resources(K).
               Shared.Protocol = Immediate_Priority_Ceiling_Protocol) THEN
            IF Si.Shared_Resources(K).Shared.Cpu_Name = Processor_Name THEN

               Fixed_Priority_Resource_Ptr(Si.Shared_Resources(K)).
                  Priority_Ceiling:=Priority_Range'First;

               FOR I IN 0..Si.Shared_Resources(K).Shared.Task_Tab.
                     Nb_Entries-1
                     LOOP
                  --    Looking for the task
                  --
                  FOR J IN 0..Si.Number_Of_Tasks-1 LOOP

                     IF Si.Tcbs(J).Tsk.Cpu_Name = Processor_Name THEN

                        IF Si.Shared_Resources(K).Shared.Task_Tab.Entries(
                              I).
                              Item = Si.Tcbs(J).
                              Tsk.Name THEN
                           Fixed_Priority_Resource_Ptr(Si.
                              Shared_Resources(K)).
                              Priority_Ceiling:=Priority_Range'Max(
                              Fixed_Priority_Tcb_Ptr(Si.Tcbs(J)).
                              Current_Priority,
                              Fixed_Priority_Resource_Ptr(Si.
                                 Shared_Resources(K)).Priority_Ceiling);
                        END IF;
                     END IF;

                  END LOOP;

               END LOOP;
            END IF;
         END IF;

      END LOOP;


   END Compute_Ceiling_Of_Resources;


   PROCEDURE Specific_Scheduler_Initialization (
         My_Scheduler   : IN OUT Fixed_Priority_Scheduler;
         Si             : IN OUT Scheduling_Information;
         Processor_Name : IN     Unbounded_String;
         My_Tasks       : IN OUT Tasks_Set;
         My_Resources   : IN OUT Resources_Set;
         My_Buffers     : IN OUT Buffers_Set;
         My_Messages    : IN     Messages_Set;
         Msg            : IN OUT Unbounded_String) IS


   BEGIN



      -- Set priority according to the scheduler
      --
      IF (My_Scheduler.Name = Deadline_Monotonic_Protocol)
            THEN
         Set_Priority_According_To_Dm(My_Scheduler, My_Tasks);
      ELSE
         IF (My_Scheduler.Name = Rate_Monotonic_Protocol)
               THEN
            Set_Priority_According_To_Rm(My_Scheduler, My_Tasks);
         END IF;
      END IF;

      FOR I IN 0 .. Si.Number_Of_Tasks-1 LOOP
         IF Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name THEN
            Fixed_Priority_Tcb_Ptr(Si.Tcbs(I)).Current_Priority:=
               Si.Tcbs(I).Tsk.Priority;
         END IF;
      END LOOP;


      -- Set priority ceiling of resources (only to PCP resource
      --
      Compute_Ceiling_Of_Resources (
         My_Scheduler  ,
         Si            ,
         Processor_Name ,
         My_Tasks      ,
         My_Resources              );

   END Specific_Scheduler_Initialization;


   PROCEDURE Release_Resource (
         My_Scheduler      : IN OUT Fixed_Priority_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table) IS

      Current_Capacity : Natural             := 0;
      A_Item           : Time_Unit_Event_Ptr;
      B                : Boolean;

   BEGIN


      IF (Si.Number_Of_Resources>0)
            THEN
         Current_Capacity:=A_Tcb.Tsk.Capacity-A_Tcb.Rest_Of_Capacity+1;


         FOR Index1 IN 0.. Si.Number_Of_Resources-1 LOOP


            -- For each resource, check if the task hold resources
            -- and then release them
            --
            FOR Index2 IN 0..Si.Shared_Resources(Index1).Shared.
                  Task_Tab.Nb_Entries-1
                  LOOP


               IF (Current_Capacity=Si.Shared_Resources(Index1).
                     Shared.Task_Tab.Entries(Index2).Data.Task_End)
                     AND (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.Name)
                     THEN
                  Si.Shared_Resources(Index1).Shared.State:=
                     Si.Shared_Resources(Index1).Shared.State+1;
                  FOR I IN 0 .. Si.Shared_Resources(Index1).
                        Nb_Allocated-1 LOOP
                     IF Si.Shared_Resources(Index1).
                           Allocated_By(I)=A_Tcb.Tsk.Name
                           THEN
                        Si.Shared_Resources(Index1).
                           Allocated_By(I):= Si.
                           Shared_Resources(Index1).Allocated_By(
                           Si.Shared_Resources(Index1).
                           Nb_Allocated-1);
                        Si.Shared_Resources(Index1).
                           Nb_Allocated:=Si.Shared_Resources(
                           Index1).Nb_Allocated-1;

                        IF   Event_To_Generate(Release_Resource) THEN
                           A_Item:= NEW Time_Unit_Event(Release_Resource);
                           A_Item.Release_Resource:=Si.Shared_Resources(
                              Index1).
                              Shared;
                           A_Item.Release_Task:=A_Tcb.Tsk;
                           Add(Result.All, A_Item, Current_Time);
                        END IF;

                        EXIT;
                     END IF;
                  END LOOP;

                  -- If PIP or PCP is used, releasing a shared resource
                  -- should restore initial priority level of the
                  -- task
                  --
                  IF (Si.Shared_Resources(Index1).Shared.
                        Protocol = Priority_Inheritance_Protocol) OR
                        (Si.Shared_Resources(Index1).Shared.Protocol
                        = Priority_Ceiling_Protocol)
                        THEN

                     Change_Current_Priority( My_Scheduler,
                        Fixed_Priority_Tcb_Ptr(A_Tcb),
                        A_Tcb.Tsk.Priority);

                  END IF;

                  -- If IPCP is used, once all IPCP resources have been
                  -- released by the task, initial priority level of the
                  -- task should be restored
                  --
                  IF  (Si.Shared_Resources(Index1).Shared.
                        Protocol = Immediate_Priority_Ceiling_Protocol)
                        THEN
                     B:=True;
                     FOR I1 IN 0.. Si.Number_Of_Resources-1 LOOP
                        FOR I2 IN 0..Si.Shared_Resources(I1).Shared.
                              Task_Tab.Nb_Entries-1
                              LOOP
                           IF (Si.Shared_Resources(I1).Shared.
                                 Task_Tab.Entries(I2).Item=A_Tcb.Tsk.Name)
                                 AND (Current_Capacity<
                                 Si.Shared_Resources(I1).
                                 Shared.Task_Tab.Entries(Index2).
                                 Data.Task_End)
                                 AND (Current_Capacity>=
                                 Si.Shared_Resources(I1).
                                 Shared.Task_Tab.Entries(Index2).
                                 Data.Task_Begin)
                                 AND (Si.Shared_Resources(I1).
                                 Shared.Protocol
                                 = Immediate_Priority_Ceiling_Protocol)
                                 THEN
                              B:=False;
                           END IF;
                        END LOOP;
                     END LOOP;
                     IF B  AND (A_Tcb.Tsk.Priority /=
                           Fixed_Priority_Tcb_Ptr(A_Tcb).
                           Current_Priority)
                           THEN
                        Change_Current_Priority( My_Scheduler,
                           Fixed_Priority_Tcb_Ptr(A_Tcb),
                           A_Tcb.Tsk.Priority);
                     END IF;
                  END IF;
               END IF;

            END LOOP;


         END LOOP;
      END IF;

   END Release_Resource;





   PROCEDURE Allocate_Resource (
         My_Scheduler      : IN OUT Fixed_Priority_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table) IS


      Current_Capacity : Natural             := 0;
      A_Item           : Time_Unit_Event_Ptr;

   BEGIN


      IF (Si.Number_Of_Resources>0)
            THEN

         Current_Capacity:=A_Tcb.Tsk.Capacity-A_Tcb.Rest_Of_Capacity+1;

         FOR Index1 IN 0 .. Si.Number_Of_Resources -1 LOOP


            -- For each resource, check if the task request the resource
            -- and if the resource is free
            -- otherwise, block the task
            --
            FOR Index2 IN 0..Si.Shared_Resources(Index1).Shared.
                  Task_Tab.Nb_Entries-1
                  LOOP


               IF (Current_Capacity=Si.Shared_Resources(Index1).
                     Shared.Task_Tab.Entries(Index2).Data.Task_Begin)
                     AND (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.Name)
                     THEN


                  Si.Shared_Resources(Index1).Shared.State:=
                     Si.Shared_Resources(Index1).Shared.State-1;

                  Si.Shared_Resources(Index1).Allocated_By(
                     Si.Shared_Resources(Index1).Nb_Allocated):=
                     A_Tcb.Tsk.Name;
                  Si.Shared_Resources(Index1).Nb_Allocated:=
                     Si.Shared_Resources(Index1).Nb_Allocated+1;

                  IF   Event_To_Generate(Allocate_Resource) THEN
                     A_Item:= NEW Time_Unit_Event(Allocate_Resource);
                     A_Item.Allocate_Resource:=Si.Shared_Resources(Index1).
                        Shared;
                     A_Item.Allocate_Task:=A_Tcb.Tsk;
                     Add(Result.All, A_Item, Current_Time);
                  END IF;

               END IF;

            END LOOP;


         END LOOP;
      END IF;


   END Allocate_Resource;





   PROCEDURE Check_Resource (
         My_Scheduler      : IN OUT Fixed_Priority_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             : IN     Tcb_Ptr;
         Is_Ready          :    OUT Boolean;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table) IS


      Current_Capacity : Natural             := 0;
      A_Item           : Time_Unit_Event_Ptr;


   BEGIN
      Is_Ready:=True;

      IF (Si.Number_Of_Resources>0)
            THEN

         Current_Capacity:=A_Tcb.Tsk.Capacity-A_Tcb.Rest_Of_Capacity+1;

         FOR Index1 IN 0 .. Si.Number_Of_Resources -1 LOOP


            -- For each resource, check if the task requests the resource
            -- and if the resource is free
            --
            FOR Index2 IN 0..Si.Shared_Resources(Index1).Shared.
                  Task_Tab.Nb_Entries-1
                  LOOP


               IF (Current_Capacity=Si.Shared_Resources(Index1).
                     Shared.Task_Tab.Entries(Index2).Data.Task_Begin)
                     AND (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.Name)
                     THEN


                  -- PCP blocking : current ceiling priority should be less or equal than
                  -- the priority of the required resource
                  --
                  IF (Si.Shared_Resources(Index1).Shared.
                        Protocol = Priority_Ceiling_Protocol)
                        THEN


                     -- Scan already allocated resources to check if the requiring task
                     -- has a highier priority than the resource
                     -- ceiling value : block the task in the other case
                     --
                     FOR Index3 IN 0 .. Si.Number_Of_Resources -1
                           LOOP
                        IF Si.Shared_Resources(Index3).Nb_Allocated>0 THEN
                           FOR Index4 IN 0 .. Si.Shared_Resources(Index3).
                                 Nb_Allocated-1 LOOP
                              IF
                                    (Si.Shared_Resources(Index3).
                                    Allocated_By(Index4) /= A_Tcb.Tsk.
                                    Name) AND
                                    (Fixed_Priority_Resource_Ptr(Si.
                                       Shared_Resources(
                                          Index3)).Priority_Ceiling >=
                                    Fixed_Priority_Tcb_Ptr(A_Tcb).
                                    Current_Priority)

                                    THEN
                                 Is_Ready:=False;
                              END IF;
                           END LOOP;
                        END IF;
                     END LOOP;


                  END IF;


                  IF (Si.Shared_Resources(Index1).Shared.State<=
                        0)
                        THEN
                     Is_Ready:=False;

                     -- Find if we must change a priority level of a task
                     -- that takes the resource (PIP or PCP protocols)
                     --
                     IF (Si.Shared_Resources(Index1).Shared.
                           Protocol = Priority_Inheritance_Protocol) OR
                           (Si.Shared_Resources(Index1).Shared.
                           Protocol = Priority_Ceiling_Protocol)
                           THEN
                        FOR I IN 0 .. Si.Shared_Resources(
                              Index1).Nb_Allocated-1 LOOP
                           FOR J IN 0..Si.Number_Of_Tasks-1  LOOP

                              IF Si.Shared_Resources(Index1).
                                    Allocated_By(I)=Si.Tcbs(J).
                                    Tsk.Name
                                    THEN

                                 Change_Current_Priority( My_Scheduler,
                                    Fixed_Priority_Tcb_Ptr(Si.Tcbs(
                                          J)),
                                    Priority_Range'Max(
                                       Fixed_Priority_Tcb_Ptr(A_Tcb).
                                       Current_Priority,
                                       Fixed_Priority_Tcb_Ptr(Si.
                                          Tcbs(J)).Current_Priority) );

                              END IF;
                           END LOOP;
                        END LOOP;
                     END IF;
                  END IF;


                  IF  Is_Ready=False THEN
                     IF   Event_To_Generate(Wait_For_Resource) THEN
                        A_Item:= NEW Time_Unit_Event(Wait_For_Resource);
                        A_Item.Wait_Resource:=Si.Shared_Resources(Index1).
                           Shared;
                        A_Item.Wait_Task:=A_Tcb.Tsk;
                        Add(Result.All, A_Item, Current_Time);
                     END IF;
                  END IF;

                  -- when allocating an IPCP resource, task priority should be
                  -- raised to resource priority ceiling
                  --
                  IF Is_Ready AND(Si.Shared_Resources(Index1).Shared.
                        Protocol = Immediate_Priority_Ceiling_Protocol)
                        AND ( Fixed_Priority_Resource_Ptr(Si.
                           Shared_Resources(
                              Index1)).Priority_Ceiling >
                        Fixed_Priority_Tcb_Ptr(A_Tcb).
                        Current_Priority)
                        THEN
                     Change_Current_Priority( My_Scheduler,
                        Fixed_Priority_Tcb_Ptr(A_Tcb),
                        Fixed_Priority_Resource_Ptr(Si.
                           Shared_Resources(
                              Index1)).Priority_Ceiling);
                  END IF;

               END IF;


            END LOOP;


         END LOOP;
      END IF;


   END Check_Resource;





   FUNCTION Build_Resource (
         My_Scheduler : IN     Fixed_Priority_Scheduler;
         A_Resource   :        Resource_Ptr)
     RETURN Shared_Resource_Ptr IS

      New_A_Resource : Fixed_Priority_Resource_Ptr;

   BEGIN

      New_A_Resource:= NEW Fixed_Priority_Resource;
      New_A_Resource.Shared:=A_Resource;

      -- Set priority ceiling of the resource
      --
      New_A_Resource.Priority_Ceiling:=Low_Priority;

      RETURN Shared_Resource_Ptr(New_A_Resource);

   END Build_Resource;




   PROCEDURE Change_Current_Priority (
         My_Scheduler : IN OUT Fixed_Priority_Scheduler'Class;
         A_Tcb        : IN     Fixed_Priority_Tcb_Ptr;
         New_Priority :        Priority_Range) IS
   BEGIN
      Dispatched_Change_Current_Priority(My_Scheduler, A_Tcb, New_Priority);
   END Change_Current_Priority;



   PROCEDURE Dispatched_Change_Current_Priority (
         My_Scheduler : IN OUT Fixed_Priority_Scheduler;
         A_Tcb        : IN     Fixed_Priority_Tcb_Ptr;
         New_Priority :        Priority_Range) IS


   BEGIN
      A_Tcb.Current_Priority:=New_Priority;
   END Dispatched_Change_Current_Priority;




   FUNCTION Compute_Pcp_Blocking_Time (
         My_Tasks     : IN     Tasks_Set;
         My_Resources : IN     Resources_Set;
         Task_Name    : IN     Unbounded_String)
     RETURN Natural IS

      Blocking_Time : Natural := 0;

      My_Resource_Iterator : Resources_Iterator;
      A_Resource           : Resource_Ptr;

   BEGIN
      Reset_Iterator(My_Resources, My_Resource_Iterator);
      LOOP
         Current_Element(My_Resources, A_Resource, My_Resource_Iterator);


         IF A_Resource.Protocol /= Priority_Ceiling_Protocol
               THEN
            RAISE Other_Resources_Protocol_Found;
         END IF;


         FOR I IN 0..A_Resource.Task_Tab.Nb_Entries-1 LOOP

            -- We are on a resource used by Task_Name
            -- Let find now the largest private section
            --
            IF (A_Resource.Task_Tab.Entries(I).Item=Task_Name) THEN

               FOR J IN 0..A_Resource.Task_Tab.Nb_Entries-1 LOOP

                  IF (A_Resource.Task_Tab.Entries(J).Item /= Task_Name) THEN
                     Blocking_Time:= Natural'Max(Blocking_Time,
                        A_Resource.
                        Task_Tab.Entries(
                           J).Data.
                        Task_End -
                        A_Resource.Task_Tab.Entries(J).Data.Task_Begin + 1);
                  END IF;

               END LOOP;
            END IF;
         END LOOP;

         EXIT WHEN Is_Last_Element(My_Resources, My_Resource_Iterator);
         Next_Element(My_Resources, My_Resource_Iterator);
      END LOOP;

      RETURN Blocking_Time;

   END Compute_Pcp_Blocking_Time;



   FUNCTION Compute_Pip_Blocking_Time (
         My_Tasks     : IN     Tasks_Set;
         My_Resources : IN     Resources_Set;
         Task_Name    : IN     Unbounded_String)
     RETURN Natural IS


      Blocking_Time : Natural := 0;

      My_Resource_Iterator : Resources_Iterator;
      A_Resource           : Resource_Ptr;
      A_Task1_Ptr          : Task_Ptr;
      A_Task2_Ptr          : Task_Ptr;

   BEGIN
      Reset_Iterator(My_Resources, My_Resource_Iterator);
      LOOP
         Current_Element(My_Resources, A_Resource, My_Resource_Iterator);


         IF A_Resource.Protocol /= Priority_Inheritance_Protocol
               THEN
            RAISE Other_Resources_Protocol_Found;
         END IF;


         FOR I IN 0..A_Resource.Task_Tab.Nb_Entries-1 LOOP

            -- We are on a resource used by Task_Name
            -- Let see now all lowest priority tasks whose access to
            -- to this resource
            --
            IF (A_Resource.Task_Tab.Entries(I).Item=Task_Name) THEN

               FOR J IN 0..A_Resource.Task_Tab.Nb_Entries-1 LOOP

                  IF (A_Resource.Task_Tab.Entries(J).Item /= Task_Name) THEN


                     -- Is the task a lower priority one ?
                     --
                     A_Task1_Ptr:=Search_Task(My_Tasks, Task_Name);
                     A_Task2_Ptr:=Search_Task(My_Tasks, A_Resource.
                        Task_Tab.Entries(J).Item);
                     IF A_Task2_Ptr.Priority < A_Task1_Ptr.Priority THEN
                        Blocking_Time:= Blocking_Time + A_Resource.
                           Task_Tab.Entries(J).Data.
                           Task_End -
                           A_Resource.Task_Tab.Entries(J).Data.Task_Begin +
                           1;
                     END IF;
                  END IF;

               END LOOP;
            END IF;
         END LOOP;

         EXIT WHEN Is_Last_Element(My_Resources, My_Resource_Iterator);
         Next_Element(My_Resources, My_Resource_Iterator);
      END LOOP;
      RETURN Blocking_Time;


   END Compute_Pip_Blocking_Time;



   FUNCTION Compute_Bound_On_Task_Blocking_Time (
         My_Scheduler : IN     Fixed_Priority_Scheduler;
         My_Tasks     : IN     Tasks_Set;
         My_Resources : IN     Resources_Set;
         A_Task       : IN     Unbounded_String)
     RETURN Natural IS


      Bound_On_Blocking_Time : Natural            := 0;
      Iterator1              : Resources_Iterator;
      Resource1              : Resource_Ptr;

   BEGIN

      -- First check that each resource used by the task have the same protocol
      --
      Same_Protocol_Control(My_Resources, A_Task);


      -- Dispatch blocking computation according to the protocol of resources
      --
      Reset_Iterator(My_Resources, Iterator1);

      LOOP
         Current_Element(My_Resources, Resource1, Iterator1);

         -- Looking for task user in the task_list index_table
         --
         FOR I IN 0..Resource1.Task_Tab.Nb_Entries-1 LOOP
            IF A_Task = Resource1.Task_Tab.Entries(I).Item
                  THEN

               CASE Resource1.Protocol  IS
                  WHEN Priority_Ceiling_Protocol =>
                     Bound_On_Blocking_Time:=Compute_Pcp_Blocking_Time(
                        My_Tasks, My_Resources, A_Task);
                  WHEN Priority_Inheritance_Protocol =>
                     Bound_On_Blocking_Time:=Compute_Pip_Blocking_Time(
                        My_Tasks, My_Resources, A_Task);
                  WHEN  OTHERS =>
                     RAISE Invalid_Protocol;
               END CASE;

               RETURN Bound_On_Blocking_Time;


            END IF;
         END LOOP;

         EXIT WHEN Is_Last_Element(My_Resources, Iterator1);
         Next_Element(My_Resources, Iterator1);

      END LOOP;

      RETURN Bound_On_Blocking_Time;


   END Compute_Bound_On_Task_Blocking_Time;





   PROCEDURE Compute_Blocking_Time (
         My_Scheduler   : IN     Fixed_Priority_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Blocking_Time  :    OUT Blocking_Time_Table) IS


      My_Task_Iterator1 : Tasks_Iterator;
      Tmp               : Tasks_Set;
      A_Task1           : Task_Ptr;
      A_Task2           : Task_Ptr;

      Has_Common_Resource   : Boolean            := False;
      My_Resource_Iterator1 : Resources_Iterator;
      A_Resource1           : Resource_Ptr;


   BEGIN

      Current_Processor_Name:=Processor_Name;
      Select_And_Copy(My_Tasks, Tmp, Select_Cpu'ACCESS );


      -- Set priority according to the scheduler
      --
      IF (My_Scheduler.Name = Deadline_Monotonic_Protocol)
            THEN
         Set_Priority_According_To_Dm(My_Scheduler, Tmp);
      ELSE
         IF (My_Scheduler.Name = Rate_Monotonic_Protocol)
               THEN
            Set_Priority_According_To_Rm(My_Scheduler, Tmp);
         END IF;
      END IF;


      Initialize(Blocking_Time);


      -- Compute blocking time for each task
      --
      Reset_Iterator(Tmp, My_Task_Iterator1);
      LOOP

         Current_Element(Tmp, A_Task1, My_Task_Iterator1);

         IF A_Task1.Cpu_Name = Processor_Name THEN

            BEGIN
               Blocking_Time.Entries(Blocking_Time.Nb_Entries).Data :=
                  Compute_Bound_On_Task_Blocking_Time (My_Scheduler,
                  My_Tasks, My_Resources, A_Task1.Name);
               Blocking_Time.Entries(Blocking_Time.Nb_Entries).Item :=
                  A_Task1;
               Blocking_Time.Nb_Entries := Blocking_Time.Nb_Entries+1;
            EXCEPTION
               WHEN  Can_Not_Used_Different_Protocol =>
                  Msg := Msg &  Unbounded_Lf &
                     To_Unbounded_String("    ") &  A_Task1.Name &
                     " => " &
                     Lb_Compute_Blocking_Error1(
                     Current_Language);
               WHEN  Invalid_Protocol =>
                  Msg := Msg &  Unbounded_Lf &
                     To_Unbounded_String("    ") & A_Task1.Name & " => " &
                     Lb_Compute_Blocking_Error3(
                     Current_Language);
            END;

         END IF;

         EXIT WHEN Is_Last_Element(Tmp, My_Task_Iterator1);

         Next_Element(Tmp, My_Task_Iterator1);

      END LOOP;

      IF Msg/=Empty_String
            THEN
         Msg:= Msg & Unbounded_Lf;
      END IF;


      --       Sort by priority and apply inheritance blocking time :
      --       a high priority task which has a non null blocking time implyes
      --       this blocking to the lowest task
      --
      Sort(Tmp, Increasing_Priority'ACCESS);

      FOR Ti IN 0..Blocking_Time.Nb_Entries-1 LOOP
         A_Task1:=Blocking_Time.Entries(TI).Item;

         -- Look for lower priority tasks on the same processor.
         -- If these lower priority tasks do not access to the same resources
         -- than A_Task1, then perform blocking time inheritance
         --
         FOR Tj IN 0..Blocking_Time.Nb_Entries-1 LOOP
            A_Task2:=Blocking_Time.Entries(TJ).Item;

            -- Check that A_Task1 and A_Task2 have no common shared resources
            --
            Has_Common_Resource:=False;
            Reset_Iterator(My_Resources, My_Resource_Iterator1);
            LOOP
               Current_Element(My_Resources, A_Resource1,
                  My_Resource_Iterator1);

               FOR RI IN 0..A_Resource1.Task_Tab.Nb_Entries-1 LOOP
                  IF A_Task1.Name = A_Resource1.Task_Tab.Entries(RI).
                        Item
                        THEN
                     FOR RJ IN 0..A_Resource1.Task_Tab.Nb_Entries-1 LOOP
                        IF A_Task2.Name =
                              A_Resource1.Task_Tab.Entries(RJ).Item
                              THEN
                           Has_Common_Resource:=True;
                        END IF;
                     END LOOP;
                  END IF;
               END LOOP;

               EXIT WHEN Is_Last_Element(My_Resources,
                  My_Resource_Iterator1);
               Next_Element(My_Resources, My_Resource_Iterator1);

            END LOOP;

            -- Compute inherited blocking time
            --
            IF NOT Has_Common_Resource THEN
               Blocking_Time.Entries(Tj).Data:=Blocking_Time.Entries(Ti).
                  Data+
                  Blocking_Time.Entries(Tj).Data;
            END IF;

         END LOOP;
      END LOOP;

      Free(Tmp);

   END Compute_Blocking_Time;



   FUNCTION Produce_Running_Task_Event (
         My_Scheduler : IN     Fixed_Priority_Scheduler;
         A_Task       : IN     Tcb_Ptr)
     RETURN Time_Unit_Event_Ptr IS

      A_Item : Time_Unit_Event_Ptr;

   BEGIN
      A_Item:= NEW Time_Unit_Event(Running_Task);
      A_Item.Running_Task:=A_Task.Tsk;
      A_Item.Current_Priority:=Fixed_Priority_Tcb_Ptr(A_Task).
         Current_Priority;
      RETURN A_Item;
   END Produce_Running_Task_Event;


   FUNCTION Export_Xml_Event_Running_Task (
         My_Scheduler : IN     Fixed_Priority_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result:= " " & An_Event.Running_Task.Name &
         " " & Priority_Range'Image(An_Event.Current_Priority);
      RETURN Result;
   END Export_Xml_Event_Running_Task;



END Scheduler.Fixed_Priority;


