------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------




WITH Text_Io;
USE Text_Io;

WITH Framework_Config;
USE Framework_Config;

WITH Translate;
USE Translate;

WITH Tasks;
USE Tasks;

WITH Scheduler;
USE Scheduler;
USE Scheduler.Double_Tasks_Parameters;

WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;


WITH Unbounded_Strings;
USE Unbounded_Strings;





PACKAGE BODY Scheduler.Hierarchical.Parametric  IS


   FUNCTION Export_Aadl_Properties (
         My_Scheduler : IN     Hierarchical_Parametric_Scheduler;
         Number_Of_Ht : IN     Natural)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Export_Aadl_Properties (Generic_Scheduler
        (My_Scheduler), Number_Of_Ht);

   BEGIN

      FOR I IN 1..Number_Of_Ht LOOP
         Result:= Result & Ascii.Ht;
      END LOOP;

      Result:= Result & To_Unbounded_String(
         "Source_text => """ &
         To_String(Get_Behavior_File_Name(My_Scheduler)) & """;");

      RETURN Result;

   END Export_Aadl_Properties;

   FUNCTION Export_Xml (
         My_Scheduler : IN     Hierarchical_Parametric_Scheduler)
     RETURN Unbounded_String IS

      Result        :          Unbounded_String := Empty_String;
      File          :          Unbounded_String := Empty_String;
      Quantum       : CONSTANT Natural          := Get_Quantum (My_Scheduler);
      Is_Preemptive : CONSTANT Preemptive_Type  := Get_Preemptive (My_Scheduler);
      Name          :          Unbounded_String := Get_Name (My_Scheduler);

   BEGIN

      Result:= Result & To_Unbounded_String( Ascii.Ht &
         Ascii.Ht &
         "<scheduler" );


      IF Quantum/=0
            THEN
         Result:= Result & To_Unbounded_String( " quantum=""" &
            Quantum'Img & """");
      END IF;


      IF Is_Preemptive /= Preemptive
            THEN
         Result:= Result & To_Unbounded_String(
            " is_preemptive=""" &
            Is_Preemptive'Img & """");
      END IF;


      File:=Get_Behavior_File_Name(My_Scheduler);

      Result:= Result & To_Unbounded_String(
         " parametric_file_name=""" & To_String(File) &
         """");

      Result:= Result & To_Unbounded_String( ">" );
      Result:= Result & To_Unbounded_String(" ") & Name &
         To_Unbounded_String(" ");
      Result:= Result & To_Unbounded_String(  "</scheduler>" ) &
         Unbounded_Lf;


      RETURN Result;


   END Export_Xml;




   PROCEDURE Initialize (
         A_Scheduler : IN OUT Hierarchical_Parametric_Scheduler) IS
   BEGIN
      Reset(A_Scheduler);
      A_Scheduler.Name:= Hierarchical_Parametric_Protocol;
      A_Scheduler.String_Behavior:= NEW Unbounded_String;
   END Initialize;


   FUNCTION Copy (
         A_Scheduler : IN     Hierarchical_Parametric_Scheduler)
     RETURN Scheduler_Ptr IS

      Ptr : Hierarchical_Parametric_Scheduler_Ptr;

   BEGIN

      Ptr:= NEW Hierarchical_Parametric_Scheduler;

      Ptr.Name:=A_Scheduler.Name;
      Ptr.Is_Preemptive:=A_Scheduler.Is_Preemptive;
      Ptr.Quantum:=A_Scheduler.Quantum;
      Ptr.Previously_Elected:=A_Scheduler.Previously_Elected;

      Ptr.String_Behavior:= NEW Unbounded_String;
      Ptr.String_Behavior.All:=A_Scheduler.String_Behavior.All;
      --  Ptr.Root_Statement_Pointer:=A_Scheduler.Root_Statement_Pointer;
      --  Ptr.Variables_Table:=A_Scheduler.Variables_Table;

      RETURN Scheduler_Ptr(Ptr);

   END Copy;





   FUNCTION Get_String_Behavior (
         My_Scheduler : IN     Hierarchical_Parametric_Scheduler)
     RETURN Unbounded_String IS

   BEGIN
      RETURN My_Scheduler.String_Behavior.All;
   END Get_String_Behavior;


   FUNCTION Get_String_Behavior (
         My_Scheduler : IN     Hierarchical_Parametric_Scheduler)
     RETURN Unbounded_String_Ptr IS

   BEGIN
      RETURN My_Scheduler.String_Behavior;
   END Get_String_Behavior;


   PROCEDURE Put (
         My_Scheduler : IN     Hierarchical_Parametric_Scheduler) IS
   BEGIN
      Put(Generic_Scheduler(My_Scheduler));
      Put_Line("Behavior file  name : " & To_String(
            My_Scheduler.Behavior_File_Name));
      Put_Line("Behavior file  : " & To_String(
            My_Scheduler.String_Behavior.All));
      --    Put(My_Scheduler.Root_Statement_Pointer);
      --  Put(My_Scheduler.Variables_Table);
   END Put;


   PROCEDURE Set_Behavior_File_Name (
         My_Scheduler : IN OUT Hierarchical_Parametric_Scheduler;
         To_Set       : IN     Unbounded_String) IS
   BEGIN
      My_Scheduler.Behavior_File_Name:=To_Set;
   END Set_Behavior_File_Name;


   FUNCTION Get_Behavior_File_Name (
         My_Scheduler : IN     Hierarchical_Parametric_Scheduler)
     RETURN Unbounded_String IS
   BEGIN
      RETURN My_Scheduler.Behavior_File_Name;
   END Get_Behavior_File_Name;


   PROCEDURE Set_String_Behavior (
         My_Scheduler : IN OUT Hierarchical_Parametric_Scheduler;
         To_Set       : IN     Unbounded_String) IS
   BEGIN
      My_Scheduler.String_Behavior:= NEW Unbounded_String;
      My_Scheduler.String_Behavior.All:=To_Set;
   END Set_String_Behavior;


   PROCEDURE Check_Before_Scheduling (
         My_Scheduler   : IN     Hierarchical_Parametric_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

   BEGIN
      NULL;
   END Check_Before_Scheduling;







END Scheduler.Hierarchical.Parametric ;


