------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Io_Tools;
USE Io_Tools;
WITH Ada.Io_Exceptions;
USE Ada.Io_Exceptions;

WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Ada.Strings.Unbounded.Text_Io;
USE Ada.Strings.Unbounded.Text_Io;
WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Ada.Exceptions;
USE Ada.Exceptions;

WITH Translate;
USE Translate;

WITH  Scheduler_Io;
USE  Scheduler_Io;
WITH Text_Io;
USE Text_Io;

WITH Statements;
USE Statements;
USE Statements.Sets_Type_Package;
WITH Simulations;
USE Simulations;

WITH Parameters;
USE Parameters;
USE Parameters.User_Defined_Parameters_Table_Package;

WITH Parser;
USE Parser;


WITH Task_Set;
USE Task_Set;

WITH Debug;
USE Debug;



PACKAGE BODY Scheduler.Parametrics IS


   FUNCTION Export_Aadl_Properties (
         My_Scheduler : IN     Parametric_Scheduler;
         Number_Of_Ht : IN     Natural)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Export_Aadl_Properties (Generic_Scheduler
        (My_Scheduler), Number_Of_Ht);

   BEGIN

      FOR I IN 1..Number_Of_Ht LOOP
         Result:= Result & Ascii.Ht;
      END LOOP;

      Result:= Result & To_Unbounded_String(
         "Source_Text => """ &
         To_String(Get_Behavior_File_Name(My_Scheduler)) & """;") &
         Unbounded_Lf;

      RETURN Result;

   END Export_Aadl_Properties;


   FUNCTION Export_Xml (
         My_Scheduler : IN     Parametric_Scheduler)
     RETURN Unbounded_String IS

      Result        :          Unbounded_String := Empty_String;
      File          :          Unbounded_String := Empty_String;
      Quantum       : CONSTANT Natural          := Get_Quantum (My_Scheduler);
      Is_Preemptive : CONSTANT Preemptive_Type  := Get_Preemptive (My_Scheduler);
      Name          :          Unbounded_String := Get_Name (My_Scheduler);

   BEGIN

      Result:= Result & To_Unbounded_String( Ascii.Ht &
         Ascii.Ht &
         "<scheduler" );


      IF Quantum/=0
            THEN
         Result:= Result & To_Unbounded_String( " quantum=""" &
            Quantum'Img & """");
      END IF;


      IF Is_Preemptive /= Preemptive
            THEN
         Result:= Result & To_Unbounded_String(
            " is_preemptive=""" &
            Is_Preemptive'Img & """");
      END IF;


      File:=Get_Behavior_File_Name(My_Scheduler);

      Result:= Result & To_Unbounded_String(
         " parametric_file_name=""" & To_String(File) &
         """");

      Result:= Result & To_Unbounded_String( ">" );
      Result:= Result & To_Unbounded_String(" ") & Name &
         To_Unbounded_String(" ");
      Result:= Result & To_Unbounded_String(  "</scheduler>" ) &
         Unbounded_Lf;


      RETURN Result;


   END Export_Xml;



   PROCEDURE Initialize (
         A_Scheduler : IN OUT Parametric_Scheduler) IS
   BEGIN
      Reset(A_Scheduler);
      A_Scheduler.Name:= Parametric_Protocol;
      A_Scheduler.String_Behavior:= NEW Unbounded_String;
   END Initialize;


   FUNCTION Copy (
         A_Scheduler : IN     Parametric_Scheduler)
     RETURN Scheduler_Ptr IS

      Ptr : Parametric_Scheduler_Ptr;

   BEGIN

      Ptr:= NEW Parametric_Scheduler;

      Ptr.Name:=A_Scheduler.Name;
      Ptr.Is_Preemptive:=A_Scheduler.Is_Preemptive;
      Ptr.Quantum:=A_Scheduler.Quantum;
      Ptr.Previously_Elected:=A_Scheduler.Previously_Elected;

      Ptr.Behavior_File_Name:= A_Scheduler.Behavior_File_Name;
      Ptr.String_Behavior:= NEW Unbounded_String;
      Ptr.String_Behavior.All:=A_Scheduler.String_Behavior.All;
      Ptr.Root_Statement_Pointer:=A_Scheduler.Root_Statement_Pointer;
      Ptr.Variables_Table:=A_Scheduler.Variables_Table;

      RETURN Scheduler_Ptr(Ptr);

   END Copy;


   ---------------------------------------------------------------------
   --  Standard procedures and functions of scheduler class
   ----------------------------------------------------------------------

   PROCEDURE Compute_Blocking_Time (
         My_Scheduler   : IN     Parametric_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Blocking_Time  :    OUT Blocking_Time_Table) IS
   BEGIN
      RAISE Invalid_Scheduler;
   END Compute_Blocking_Time;


   PROCEDURE Bound_On_Processor_Utilization (
         My_Scheduler   : IN     Parametric_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Result         : IN OUT Double;
         Msg            : IN OUT Unbounded_String) IS

   BEGIN
      RAISE Invalid_Scheduler;
   END Bound_On_Processor_Utilization;


   PROCEDURE Compute_Response_Time (
         My_Scheduler   : IN     Parametric_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Response_Time  :    OUT Response_Time_Table) IS

   BEGIN
      RAISE Invalid_Scheduler;
   END Compute_Response_Time;


   FUNCTION Get_String_Behavior (
         My_Scheduler : IN     Parametric_Scheduler)
     RETURN Unbounded_String IS

   BEGIN
      RETURN My_Scheduler.String_Behavior.All;
   END Get_String_Behavior;


   FUNCTION Get_String_Behavior (
         My_Scheduler : IN     Parametric_Scheduler)
     RETURN Unbounded_String_Ptr IS

   BEGIN
      RETURN My_Scheduler.String_Behavior;
   END Get_String_Behavior;


   PROCEDURE Put (
         My_Scheduler : IN     Parametric_Scheduler) IS
   BEGIN
      Put(Generic_Scheduler(My_Scheduler));
      Put_Line("Behavior file  name : " & My_Scheduler.Behavior_File_Name);
      Put_Line("Behavior file  : " & My_Scheduler.String_Behavior.All);
      Put(My_Scheduler.Root_Statement_Pointer);
      Put(My_Scheduler.Variables_Table);
   END Put;


   PROCEDURE Set_Behavior_File_Name (
         My_Scheduler : IN OUT Parametric_Scheduler;
         To_Set       : IN     Unbounded_String) IS
   BEGIN
      My_Scheduler.Behavior_File_Name:=To_Set;
   END Set_Behavior_File_Name;


   FUNCTION Get_Behavior_File_Name (
         My_Scheduler : IN     Parametric_Scheduler)
     RETURN Unbounded_String IS
   BEGIN
      RETURN My_Scheduler.Behavior_File_Name;
   END Get_Behavior_File_Name;


   PROCEDURE Set_String_Behavior (
         My_Scheduler : IN OUT Parametric_Scheduler;
         To_Set       : IN     Unbounded_String) IS
   BEGIN
      My_Scheduler.String_Behavior:= NEW Unbounded_String;
      My_Scheduler.String_Behavior.All:=To_Set;
   END Set_String_Behavior;


   PROCEDURE Check_Before_Scheduling (
         My_Scheduler   : IN     Parametric_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

   BEGIN
      NULL;
   END Check_Before_Scheduling;



   PROCEDURE Do_Election (
         My_Scheduler      : IN OUT Parametric_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Msg               : IN OUT Unbounded_String;
         Current_Time      : IN     Natural;
         Processor_Name    : IN     Unbounded_String;
         My_Dependencies   : IN     Tasks_Dependencies_Ptr;
         With_Offsets      : IN     Boolean;
         With_Precedencies : IN     Boolean;
         With_Resources    : IN     Boolean;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table;
         Elected           : IN OUT Tasks_Range;
         No_Task           : IN OUT Boolean) IS


      Return_Val             : Expression_Ptr;
      Current                : Statement_Ptr;
      Election_Value         : Simulation_Value_Ptr;
      Current_Selected_Value : Simulation_Value_Ptr;
      I                      : Tasks_Range          := 0;
      Var                    : Variables_Range;
      Var_Task_Name          : Variables_Range;

   BEGIN


      -- By default, no task are elected : it's a way
      -- to safer schedule a parametric scheduler without
      -- a correctly designed code (election part of it)
      --
      No_Task:=True;




      -- Look for the variable which stores task names
      --
      Var_Task_Name:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("tasks.name"));



      -- We set the variables changed by the simulation each unit of time
      --
      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("tasks.activation_number"));
      FOR I IN 0..Si.Number_Of_Tasks -1 LOOP
         IF Processor_Name = Si.Tcbs(I).Tsk.Cpu_Name
               THEN
            FOR J IN 0..Si.Number_Of_Tasks -1  LOOP
               IF My_Scheduler.Variables_Table.Entries(Var_Task_Name).
                     Simulation.
                     String_Table_Value(Simulations_Range(J)) = Si.Tcbs(I).
                     Tsk.Name
                     THEN
                  My_Scheduler.Variables_Table.Entries(Var).Simulation.
                     Integer_Table_Value(Simulations_Range(J)):=
                     Integer(Si.Tcbs(I).Activation);
               END IF;
            END LOOP;
         END IF;
      END LOOP;



      FOR I IN 0..Si.Number_Of_Tasks -1 LOOP
         IF My_Scheduler.Variables_Table.Entries(Var_Task_Name).
               Simulation.
               String_Table_Value(Simulations_Range(I)) = Si.Tcbs(
               My_Scheduler.Previously_Elected).Tsk.Name
               THEN
            Var:=Find_Variable(My_Scheduler.Variables_Table,
               To_Unbounded_String("tasks.rest_of_capacity"));
            My_Scheduler.Variables_Table.Entries(Var).Simulation.
               Integer_Table_Value(Simulations_Range(I)):=
               Integer(Si.Tcbs(My_Scheduler.Previously_Elected).
               Rest_Of_Capacity);

            Var:=Find_Variable(My_Scheduler.Variables_Table,
               To_Unbounded_String("tasks.used_cpu"));
            My_Scheduler.Variables_Table.Entries(Var).Simulation.
               Integer_Table_Value(Simulations_Range(I)):=
               Integer(Si.Tcbs(My_Scheduler.Previously_Elected).
               Used_Cpu);
         END IF;
      END LOOP;


      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("previously_elected"));
      My_Scheduler.Variables_Table.Entries(Var).Simulation.
         Integer_Value:=Integer(My_Scheduler.Previously_Elected);


      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("simulation_time"));
      My_Scheduler.Variables_Table.Entries(Var).Simulation.
         Integer_Value:=Integer(Current_Time);


      -- Check if tasks are ready or not
      --
      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("tasks.ready"));

      FOR I IN 0..Si.Number_Of_Tasks -1 LOOP
         IF (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) THEN
            FOR J IN 0..Si.Number_Of_Tasks -1 LOOP
               IF My_Scheduler.Variables_Table.Entries(
                     Var_Task_Name).Simulation.
                     String_Table_Value(Simulations_Range(J)) = Si.Tcbs(I).
                     Tsk.Name
                     THEN
                  My_Scheduler.Variables_Table.Entries(Var).
                     Simulation.
                     Boolean_Table_Value(Simulations_Range(J)):=
                     False;
               END IF;
            END LOOP;
         END IF;
      END LOOP;


      FOR I IN 0..Si.Number_Of_Tasks -1 LOOP
         IF (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) THEN
            IF (Si.Tcbs(I).Wake_Up_Time <= Current_Time) AND
                  (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                  THEN
               IF (With_Offsets=False) OR
                     Check_Offset(Si.Tcbs(I), Current_Time)
                     THEN
                  IF (With_Precedencies=False) OR
                        Check_Precedencies(Si, My_Dependencies,
                        Current_Time,
                        Si.Tcbs(I))
                        THEN
                     FOR J IN 0..Si.Number_Of_Tasks -1 LOOP
                        IF My_Scheduler.Variables_Table.Entries(
                              Var_Task_Name).Simulation.
                              String_Table_Value(Simulations_Range(J)) =
                              Si.Tcbs(I).
                              Tsk.Name
                              THEN
                           My_Scheduler.Variables_Table.Entries(Var).
                              Simulation.
                              Boolean_Table_Value(Simulations_Range(J)):=
                              True;
                        END IF;
                     END LOOP;
                  END IF;
               END IF;
            END IF;
         END IF;
      END LOOP;



      -- Now, do "priority" section
      --
      Current := My_Scheduler.Root_Statement_Pointer(Priority);
      Dispatch(Current, Priority, My_Scheduler.Variables_Table, Msg);


      -- After priority section, we
      -- save some task run time information
      --
      Save_Read_Write_Parametric_Variables(My_Scheduler, Si,
         Processor_Name);



      -- Do "election" section
      --

      -- A election is done in 3 steps :
      --        1) we check and retrieve return information
      --        2) tasks queues should be updated
      --        3) we find the running task
      --
      Current := My_Scheduler.Root_Statement_Pointer(Election);


      -- We should have, at least a election_section !
      --
      IF Current=NULL THEN
         Raise_Exception (Statement_Error'Identity,
            "An election section is mandatory");
      END IF;


      -- Only one type of statement is allowed on "election" section :
      -- a return statement
      --
      IF Current.All NOT IN Return_Statement'Class
            THEN
         Raise_Exception (Statement_Error'Identity,
            "Election section, line " & Current.Line_Number'Img &
            ", return statement expected");
      END IF;


      -- Do return statement
      --

      -- The return statement should be :
      --      - either a min/max_to_index expression composed of
      --        a array variable only. Boolean/double/string array variable are forbidden
      --      - or an integer constant
      --      - or an integer variable
      --
      Return_Val:=Return_Statement_Ptr(Current).Return_Value;




      ------------------------------------------------------------
      -- The return statement is a constant or a variable ?
      ------------------------------------------------------------


      IF (Return_Val.Expression_Type=Variable_Type) OR (Return_Val.
            Expression_Type=Constant_Type) THEN


         IF (Return_Val.Expression_Type=Constant_Type)
               THEN
            Put_Debug("Do a constant election");
            IF Constant_Expr_Ptr(Return_Val).Value.Ptype/=
                  Simulation_Integer
                  THEN
               Raise_Exception (Statement_Error'Identity,
                  "Election section, line " & Current.Line_Number'
                  Img &
                  ", invalid constant expression type referenced in a return statement");
            END IF;
            IF (Constant_Expr_Ptr(Return_Val).
                  Value.Integer_Value<
                  Integer(Tasks_Range'First)) OR
                  (Constant_Expr_Ptr(Return_Val).Value.
                  Integer_Value>
                  Integer(Tasks_Range'
                     Last)) THEN
               Raise_Exception (Statement_Error'Identity,
                  "Election section, line " & Current.Line_Number'
                  Img &
                  ", invalid range constant type referenced in a return statement");
            END IF;

            I:=Tasks_Range(Constant_Expr_Ptr(Return_Val).Value.
               Integer_Value);
         END IF;

         IF   (Return_Val.Expression_Type=Variable_Type)
               THEN
            Put_Debug("Do a variable election");
            Var:=Find_Variable(My_Scheduler.Variables_Table,
               Variable_Expr_Ptr(Return_Val).Name);
            IF My_Scheduler.Variables_Table.Entries(Var).
                  Simulation.
                  Ptype/=Simulation_Integer
                  THEN
               Raise_Exception (Statement_Error'Identity,
                  "Election section, line " & Current.Line_Number'
                  Img &
                  ", invalid variable expression type referenced in a return statement");
            END IF;
            IF (My_Scheduler.Variables_Table.Entries(Var).
                  Simulation.Integer_Value<Integer(
                     Simulations_Range'First)) OR
                  (My_Scheduler.Variables_Table.Entries(Var).
                  Simulation.Integer_Value>Integer(
                     Tasks_Range'
                     Last)) THEN
               Raise_Exception (Statement_Error'Identity,
                  "Election section, line " & Current.Line_Number'
                  Img &
                  ", invalid range variable type referenced in a return statement");
            END IF;
            I:=Tasks_Range(My_Scheduler.Variables_Table.Entries(
                  Var).
               Simulation.Integer_Value);
         END IF;


         -- Convert parametric variable index towards tcb index
         --
         FOR J IN 0..Si.Number_Of_Tasks -1 LOOP
            IF My_Scheduler.Variables_Table.Entries(Var_Task_Name).
                  Simulation.
                  String_Table_Value(Simulations_Range(I)) = Si.Tcbs(
                  J).Tsk.Name
                  THEN

               -- Check that the task is ready ... and elect it !
               --
               IF (Si.Tcbs(J).Tsk.Cpu_Name = Processor_Name)
                     THEN
                  IF (Si.Tcbs(J).Wake_Up_Time <= Current_Time)
                        THEN
                     IF   (Si.Tcbs(J).Rest_Of_Capacity /= 0)
                           THEN
                        IF (With_Offsets=False) OR
                              Check_Offset(Si.Tcbs(J), Current_Time)
                              THEN
                           IF (With_Precedencies=False) OR
                                 Check_Precedencies(Si, My_Dependencies,
                                 Current_Time,
                                 Si.Tcbs(J))
                                 THEN
                              Elected:=J;
                              No_Task:=False;
                           END IF;
                        END IF;
                     END IF;
                  END IF;
               END IF;

            END IF;
         END LOOP;


         ------------------------------------------------------------------------------
         -- Now, let see the case when the return expression is a max/min_to_index
         ------------------------------------------------------------------------------


      ELSE


         Put_Debug("Do a min/max_to_index election");

         IF (Return_Val.Expression_Type=Unary_Type) THEN
            IF (Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index) OR (
                  Unary_Expr_Ptr(Return_Val).Ope=Max_To_Index) THEN



               IF (Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index) THEN
                  IF (Unary_Expr_Ptr(Return_Val).Value.
                        Expression_Type/=
                        Array_Variable_Type)
                        THEN
                     Raise_Exception (Statement_Error'Identity,
                        "Election section, line " & Current.
                        Line_Number'
                        Img &
                        ", invalid min_to_index expression in return statement");
                  END IF;

                  IF (Array_Variable_Expr_Ptr(Unary_Expr_Ptr(
                              Return_Val).
                           Value).
                        Variable_Type
                        /=Simulation_Array_Integer)
                        AND (Array_Variable_Expr_Ptr(
                           Unary_Expr_Ptr(
                              Return_Val).
                           Value).
                        Variable_Type
                        /=Simulation_Array_Double)

                        THEN
                     Raise_Exception (Statement_Error'Identity,
                        "Election section, line " & Current.
                        Line_Number'
                        Img &
                        ", min_to_index expression in return statement should be an integer array variable");
                  END IF;
               END IF;
            END IF;

            IF (Return_Val.Expression_Type=Unary_Type) THEN
               IF (Unary_Expr_Ptr(Return_Val).Ope=Max_To_Index) THEN
                  IF (Unary_Expr_Ptr(Return_Val).Value.
                        Expression_Type/=
                        Array_Variable_Type)
                        THEN
                     Raise_Exception (Statement_Error'Identity,
                        "Election section, line " & Current.
                        Line_Number'
                        Img &
                        ", invalid max_to_index expression in return statement");
                  END IF;


                  IF (Array_Variable_Expr_Ptr(Unary_Expr_Ptr(
                              Return_Val).
                           Value).
                        Variable_Type
                        /=Simulation_Array_Integer)
                        THEN
                     Raise_Exception (Statement_Error'Identity,
                        "Election section, line " & Current.
                        Line_Number'
                        Img &
                        ", max_to_index expression in return statement should be an integer array variable");
                  END IF;
               END IF;
            END IF;

            -----------------------------------------------------
            -- All the code is now done : we do the last
            -- operation : election !!!
            -----------------------------------------------------

            Election_Value:=Value_Of(Array_Variable_Expr(
                  Unary_Expr_Ptr(
                     Return_Val).Value.All), My_Scheduler.
               Variables_Table);


            IF (Election_Value.Ptype=Simulation_Array_Double) THEN
               Current_Selected_Value:= NEW Simulation_Value (
                  Simulation_Double);
               IF Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index THEN
                  Current_Selected_Value.Double_Value:=Double'Last;
               ELSE
                  Current_Selected_Value.Double_Value:=Double'First;
               END IF;
            ELSE
               Current_Selected_Value:= NEW Simulation_Value (
                  Simulation_Integer);
               IF Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index THEN
                  Current_Selected_Value.Integer_Value:=Integer'
                     Last;
               ELSE
                  Current_Selected_Value.Integer_Value:=Integer'
                     First;
               END IF;
            END IF;


            -- Now, loop on the array variable to find the appropriate task to be
            -- elected
            --

            I:=0;
            LOOP
               IF (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) THEN
                  IF (Si.Tcbs(I).Wake_Up_Time <= Current_Time) AND
                        (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                        THEN
                     IF (With_Offsets=False) OR
                           Check_Offset(Si.Tcbs(I), Current_Time)
                           THEN
                        IF (With_Precedencies=False) OR
                              Check_Precedencies(Si, My_Dependencies,
                              Current_Time,
                              Si.Tcbs(I))
                              THEN

                           IF (Election_Value.Ptype=
                                 Simulation_Array_Integer) THEN

                              IF ((Unary_Expr_Ptr(Return_Val).Ope=
                                       Min_To_Index) AND
                                    (Election_Value.
                                       Integer_Table_Value(
                                          Simulations_Range(I)) <
                                       Current_Selected_Value.
                                       Integer_Value))
                                    OR
                                    ((Unary_Expr_Ptr(Return_Val).
                                       Ope=
                                       Max_To_Index) AND
                                    (Election_Value.
                                       Integer_Table_Value(
                                          Simulations_Range(I)) >
                                       Current_Selected_Value.
                                       Integer_Value))

                                    THEN
                                 Elected:=I;
                                 Current_Selected_Value.Integer_Value:=
                                    Election_Value.
                                    Integer_Table_Value(
                                    Simulations_Range(I));
                              END IF;
                           ELSE
                              IF ((Unary_Expr_Ptr(Return_Val).Ope=
                                       Min_To_Index) AND
                                    (
                                       Election_Value.Double_Table_Value(
                                          Simulations_Range(I)) <
                                       Current_Selected_Value.
                                       Double_Value))
                                    OR
                                    ((Unary_Expr_Ptr(Return_Val).
                                       Ope=
                                       Max_To_Index) AND
                                    (
                                       Election_Value.Double_Table_Value(
                                          Simulations_Range(I)) >
                                       Current_Selected_Value.
                                       Double_Value))

                                    THEN
                                 Elected:=I;
                                 Current_Selected_Value.Double_Value:=
                                    Election_Value.
                                    Double_Table_Value(
                                    Simulations_Range(I));
                              END IF;


                           END IF;
                        END IF;
                     END IF;
                  END IF;
               END IF;


               I:=I+1;
               EXIT WHEN Si.Tcbs(I)=NULL;
            END LOOP;



            IF (Election_Value.Ptype=
                  Simulation_Array_Integer) THEN

               IF ( (Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index) AND (
                        Current_Selected_Value.Integer_Value/=
                        Integer'
                        Last)  )
                     OR ( (Unary_Expr_Ptr(Return_Val).Ope=
                        Max_To_Index) AND (
                        Current_Selected_Value.Integer_Value/=
                        Integer'
                        First)  )
                     THEN
                  No_Task:=False;
               END IF;
            ELSE
               IF ( (Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index) AND (
                        Current_Selected_Value.Double_Value/=Double'
                        Last)  )
                     OR ( (Unary_Expr_Ptr(Return_Val).Ope=
                        Max_To_Index) AND (
                        Current_Selected_Value.Double_Value/=Double'
                        First)  )
                     THEN
                  No_Task:=False;
               END IF;
            END IF;

            Free(Election_Value);


            -- End of return statement
            --

            ------------------------------------------------------------
            -- Others expression in a return statement is a mistake
            ------------------------------------------------------------
         ELSE

            Raise_Exception (Statement_Error'Identity,
               "Election section, line " & Current.Line_Number'Img &
               ", invalid expression referenced in a return statement");


         END IF;
      END IF;

   END Do_Election;



   -- Do "start" section statements
   --
   PROCEDURE Specific_Scheduler_Initialization (
         My_Scheduler   : IN OUT Parametric_Scheduler;
         Si             : IN OUT Scheduling_Information;
         Processor_Name : IN     Unbounded_String;
         My_Tasks       : IN OUT Tasks_Set;
         My_Resources   : IN OUT Resources_Set;
         My_Buffers     : IN OUT Buffers_Set;
         My_Messages    : IN     Messages_Set;
         Msg            : IN OUT Unbounded_String) IS


      Current : Statement_Ptr;
      Var     : Variables_Range;


   BEGIN


      -- Read all parametric files before simulation
      --
      BEGIN
         My_Scheduler.String_Behavior:= NEW Unbounded_String;
         My_Scheduler.String_Behavior.All:=Read_Sequential_File(
            My_Scheduler.Behavior_File_Name);


      EXCEPTION
         WHEN Ada.Io_Exceptions.Name_Error =>
            Raise_Exception(Parametric_File_Error'Identity, To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(Current_Language))  &
               ", Name_Error");
         WHEN Ada.Io_Exceptions.Status_Error =>
            Raise_Exception(Parametric_File_Error'Identity, To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Status_Error");
         WHEN Ada.Io_Exceptions.Mode_Error   =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Mode_Error");
         WHEN Ada.Io_Exceptions.Use_Error     =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(Current_Language)) &
               ", Use_Error");
         WHEN Ada.Io_Exceptions.Device_Error  =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(Current_Language)) &
               ", Device_Error");
         WHEN Ada.Io_Exceptions.End_Error    =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", End_Error");
         WHEN Ada.Io_Exceptions.Data_Error   =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Data_Error");
         WHEN Ada.Io_Exceptions.Layout_Error  =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Layout_Error");
         WHEN OTHERS =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) );
      END;



      -- We should check the scheduler behavior syntax ...
      --
      Open_Input(My_Scheduler.String_Behavior.All);
      Parser.Init_Parser;
      Create_Parametric_Variables(Parser.Variables_Table, My_Tasks);
      Parser.Yyparse;



      -- Store syntax tree and compiling information
      --
      My_Scheduler.Root_Statement_Pointer:=
         Parser.Root_Statement_Pointer;

      My_Scheduler.Variables_Table:=
         Parser.Variables_Table;

      My_Scheduler.Sets_Table:=
         Parser.Sets_Table;

      Current := My_Scheduler.Root_Statement_Pointer (Start);



      -- Initialize parametric variables which are constant
      --
      Initialize_Parametric_Variables(My_Scheduler.Variables_Table,
         My_Messages,
         My_Buffers,
         My_Resources,
         My_Tasks,
         Processor_Name);



      -- Initialize scheduler-parametrics.adb specific parametric variables
      --

      -- "nb_processors"
      --
      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("nb_processors"));
      My_Scheduler.Variables_Table.Entries(Var).Simulation.
         Integer_Value:=Si.Number_Of_Processors;

      -- "nb_address_spaces"
      --
      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("nb_address_spaces"));
      My_Scheduler.Variables_Table.Entries(Var).Simulation.
         Integer_Value:=Si.Number_Of_Address_Spaces;

      -- "simulation_length"
      --
      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("simulation_length"));
      My_Scheduler.Variables_Table.Entries(Var).Simulation.
         Integer_Value:=Si.Simulation_Length;


      -- Load parametric variable which can be changed by the programmer
      -- before executing sections
      --
      Load_Read_Write_Parametric_Variables(My_Scheduler, Si,
         Processor_Name);


      -- Now : run "start" statements
      --
      Dispatch(Current, Start, My_Scheduler.Variables_Table, Msg);


      -- After start section, we
      -- save some task run time information
      --
      Save_Read_Write_Parametric_Variables(My_Scheduler, Si,
         Processor_Name);


   END Specific_Scheduler_Initialization;


   PROCEDURE Compute_Activation_Time (
         My_Scheduler : IN     Parametric_Scheduler;
         Si           : IN OUT Scheduling_Information;
         Elected      : IN     Tasks_Range;
         Value        : IN OUT Natural) IS


      Set_Id           : Sets_Range;
      Parametric_Delay : Simulation_Value_Ptr;

   BEGIN

      Set_Id:=Find_Set(My_Scheduler.Sets_Table,
         Parametric_Task_Ptr(Si.Tcbs(Elected).Tsk).Activation_Rule);

      Parametric_Delay := Value_Of (My_Scheduler.
         Sets_Table.Entries(
            Set_Id).Set_Value.All,
         My_Scheduler.Variables_Table);


      IF (Parametric_Delay.Ptype /= Simulation_Integer)
            AND  (Parametric_Delay.Ptype /=
            Simulation_Array_Integer) THEN
         Raise_Exception(Type_Error'Identity,
            "Task activation rule, activation rule " &
            To_String(Parametric_Task_Ptr(Si.Tcbs(Elected).Tsk).
               Activation_Rule)  &
            To_String(Lb_Comma & Lb_Uncompatible_Type_Error(
                  Current_Language) ));
      END IF;



      -- If it's a vector, return the corresponding index
      -- else, just return the computed integer
      --
      IF Parametric_Delay.Ptype=Simulation_Integer
            THEN
         Value:=Parametric_Delay.Integer_Value;
      ELSE
         Value:=Parametric_Delay.Integer_Table_Value(Simulations_Range(
               Elected));
      END IF;


   END Compute_Activation_Time;



   PROCEDURE Utilization_Factor_Feasibility_Test (
         My_Scheduler   : IN     Parametric_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Result         : IN OUT Unbounded_String) IS

   BEGIN
      Result := Result &  Lb_Compute_Scheduling_Error_19(
         Current_Language) &
         Unbounded_Lf;
   END Utilization_Factor_Feasibility_Test;





   PROCEDURE Load_Read_Write_Parametric_Variables (
         My_Scheduler   : IN OUT Parametric_Scheduler;
         Si             : IN     Scheduling_Information;
         Processor_Name : IN     Unbounded_String) IS

      Var           : Variables_Range;
      Var_Task_Name : Variables_Range;

   BEGIN

      Var_Task_Name:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("tasks.name"));


      FOR J IN 0..Si.Number_Of_Tasks-1 LOOP

         IF Processor_Name = Si.Tcbs(J).Tsk.Cpu_Name
               THEN


            FOR Z IN 0..Si.Number_Of_Tasks -1 LOOP
               IF My_Scheduler.Variables_Table.Entries(
                     Var_Task_Name).Simulation.
                     String_Table_Value(Simulations_Range(Z)) = Si.Tcbs(J).
                     Tsk.Name
                     THEN


                  -- "tasks.priority"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.priority"));
                  My_Scheduler.Variables_Table.Entries(Var).Simulation.
                     Integer_Table_Value(Simulations_Range(Z)):=
                     Integer(Si.Tcbs(J).Tsk.Priority);


                  -- "tasks.deadline"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.deadline"));
                  My_Scheduler.Variables_Table.Entries(Var).Simulation.
                     Integer_Table_Value(Simulations_Range(Z)):=
                     Integer(Si.Tcbs(J).Tsk.Deadline);


                  -- "tasks.start_time"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.start_time"));
                  My_Scheduler.Variables_Table.Entries(Var).Simulation.
                     Integer_Table_Value(Simulations_Range(Z)):=
                     Integer(Si.Tcbs(J).Tsk.Start_Time);


                  -- "tasks.processor_name"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.processor_name"));
                  My_Scheduler.Variables_Table.Entries(Var).Simulation.
                     String_Table_Value(Simulations_Range(Z)):=
                     Si.Tcbs(J).Tsk.Cpu_Name;


                  -- "tasks.name"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.name"));
                  My_Scheduler.Variables_Table.Entries(Var).Simulation.
                     String_Table_Value(Simulations_Range(Z)):=
                     Si.Tcbs(J).Tsk.Name;


                  -- "tasks.capacity"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.capacity"));
                  My_Scheduler.Variables_Table.Entries(Var).Simulation.
                     Integer_Table_Value(Simulations_Range(Z)):=
                     Integer(Si.Tcbs(J).Tsk.Capacity);


                  -- "tasks.rest_of_capacity"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.rest_of_capacity"));
                  My_Scheduler.Variables_Table.Entries(Var).Simulation.
                     Integer_Table_Value(Simulations_Range(Z)):=
                     Integer(Si.Tcbs(J).Rest_Of_Capacity);


                  -- "tasks.blocking_time"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.blocking_time"));
                  My_Scheduler.Variables_Table.Entries(Var).Simulation.
                     Integer_Table_Value(Simulations_Range(Z)):=
                     Integer(Si.Tcbs(J).Tsk.Blocking_Time);


                  -- "tasks.period"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.period"));
                  CASE Si.Tcbs(J).Tsk.Task_Type IS
                     WHEN Aperiodic_Type =>
                        My_Scheduler.Variables_Table.Entries(Var).
                           Simulation.
                           Integer_Table_Value(Simulations_Range(Z)):=0;
                     WHEN OTHERS =>
                        My_Scheduler.Variables_Table.Entries(Var).
                           Simulation.
                           Integer_Table_Value(Simulations_Range(Z)):=
                           Integer(Periodic_Task_Ptr(Si.Tcbs(J).Tsk).
                           Period);
                  END CASE;


                  -- "tasks.jitter"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.jitter"));
                  CASE Si.Tcbs(J).Tsk.Task_Type IS
                     WHEN Aperiodic_Type =>
                        My_Scheduler.Variables_Table.Entries(Var).
                           Simulation.
                           Integer_Table_Value(Simulations_Range(Z)):=0;
                     WHEN OTHERS =>
                        My_Scheduler.Variables_Table.Entries(Var).
                           Simulation.
                           Integer_Table_Value(Simulations_Range(Z)):=
                           Integer(Periodic_Task_Ptr(Si.Tcbs(J).Tsk).
                           Jitter);
                  END CASE;



                  -- user defined task's parameter :
                  --

                  FOR I IN 0..Si.Tcbs(J).Tsk.Param.Nb_Entries-1 LOOP

                     -- Find the user defined variable
                     -- and set its value
                     --
                     FOR K IN 0..Variables_Table.Nb_Entries-1 LOOP
                        IF My_Scheduler.Variables_Table.Entries(K).
                              Variable.Name=
                              To_Unbounded_String("tasks.") & Si.Tcbs(J).
                              Tsk.Param.
                              Entries(
                              I).Name
                              THEN

                           -- Now, set the value
                           --
                           IF Si.Tcbs(J).Tsk.Param.Entries(I).Ptype=
                                 Boolean_Parameter THEN
                              My_Scheduler.Variables_Table.Entries(K).
                                 Simulation.
                                 Boolean_Table_Value(Simulations_Range(Z)):=
                                 Boolean(Si.Tcbs(J).Tsk.
                                 Param.Entries(I).Boolean_Value);
                           END IF;
                           IF Si.Tcbs(J).Tsk.Param.Entries(I).Ptype=
                                 Integer_Parameter THEN
                              My_Scheduler.Variables_Table.Entries(K).
                                 Simulation.
                                 Integer_Table_Value(Simulations_Range(Z)):=
                                 Integer(Si.Tcbs(J).Tsk.
                                 Param.Entries(I).Integer_Value);
                           END IF;
                           IF Si.Tcbs(J).Tsk.Param.Entries(I).Ptype=
                                 Double_Parameter THEN
                              My_Scheduler.Variables_Table.Entries(K).
                                 Simulation.
                                 Double_Table_Value(Simulations_Range(Z)):=
                                 Double(
                                 Si.Tcbs(J).Tsk.
                                 Param.Entries(I).Double_Value);
                           END IF;
                           IF Si.Tcbs(J).Tsk.Param.Entries(I).Ptype=
                                 String_Parameter THEN
                              My_Scheduler.Variables_Table.Entries(K).
                                 Simulation.
                                 String_Table_Value(Simulations_Range(Z)):=
                                 Si.Tcbs(J).Tsk.
                                 Param.Entries(I).String_Value;
                           END IF;
                        END IF;
                     END LOOP;
                  END LOOP;


               END IF;
            END LOOP;


         END IF;
      END LOOP;


   END Load_Read_Write_Parametric_Variables;





   PROCEDURE Save_Read_Write_Parametric_Variables (
         My_Scheduler   : IN     Parametric_Scheduler;
         Si             : IN OUT Scheduling_Information;
         Processor_Name : IN     Unbounded_String) IS

      Var           : Variables_Range;
      Var_Task_Name : Variables_Range;

   BEGIN

      Var_Task_Name:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("tasks.name"));


      FOR J IN 0..Si.Number_Of_Tasks-1 LOOP

         IF Processor_Name = Si.Tcbs(J).Tsk.Cpu_Name
               THEN



            FOR Z IN 0..Si.Number_Of_Tasks -1 LOOP
               IF My_Scheduler.Variables_Table.Entries(
                     Var_Task_Name).Simulation.
                     String_Table_Value(Simulations_Range(Z)) = Si.Tcbs(J).
                     Tsk.Name
                     THEN


                  -- "tasks.priority"
                  --
                  BEGIN
                     Var:=Find_Variable(Variables_Table,
                        To_Unbounded_String("tasks.priority"));
                     Si.Tcbs(J).Tsk.Priority:=Priority_Range(
                        My_Scheduler.
                        Variables_Table.Entries(
                           Var).Simulation.
                        Integer_Table_Value(Simulations_Range(Z)));
                  EXCEPTION
                     WHEN Constraint_Error=>
                        Raise_Exception(Type_Error'Identity,
                           "Task priority range check error  " &
                           To_String(Parametric_Task_Ptr(Si.Tcbs(J).Tsk).
                              Name) & " " & Integer'Image(My_Scheduler.
                              Variables_Table.Entries(
                                 Var).Simulation.
                              Integer_Table_Value(Simulations_Range(Z))));
                  END;



                  -- "tasks.deadline"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.deadline"));
                  Si.Tcbs(J).Tsk.Deadline:=
                     My_Scheduler.Variables_Table.Entries(
                     Var).Simulation.
                     Integer_Table_Value(Simulations_Range(Z));


                  -- "tasks.start_time"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.start_time"));
                  Si.Tcbs(J).Tsk.Start_Time:=
                     My_Scheduler.Variables_Table.Entries(
                     Var).Simulation.
                     Integer_Table_Value(Simulations_Range(Z));


                  -- "tasks.capacity"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.capacity"));
                  Si.Tcbs(J).Tsk.Capacity:=
                     My_Scheduler.Variables_Table.Entries(
                     Var).
                     Simulation.
                     Integer_Table_Value(Simulations_Range(Z));


                  -- "tasks.rest_of_capacity"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.rest_of_capacity"));
                  Si.Tcbs(J).Rest_Of_Capacity:=
                     My_Scheduler.Variables_Table.Entries(
                     Var).
                     Simulation.
                     Integer_Table_Value(Simulations_Range(Z));


                  -- "tasks.blocking_time"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String(
                        "tasks.blocking_time"));
                  Si.Tcbs(J).Tsk.Blocking_Time:=
                     My_Scheduler.Variables_Table.
                     Entries(Var).
                     Simulation.
                     Integer_Table_Value(Simulations_Range(Z));


                  -- "tasks.period"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.period"));
                  CASE Si.Tcbs(J).Tsk.Task_Type IS
                     WHEN Aperiodic_Type =>
                        NULL;
                     WHEN OTHERS =>
                        Periodic_Task_Ptr(Si.Tcbs(J).Tsk).Period:=
                           My_Scheduler.
                           Variables_Table.
                           Entries(Var).Simulation.
                           Integer_Table_Value(Simulations_Range(Z));
                  END CASE;


                  -- "tasks.jitter"
                  --
                  Var:=Find_Variable(Variables_Table,
                     To_Unbounded_String("tasks.jitter"));
                  CASE Si.Tcbs(J).Tsk.Task_Type IS
                     WHEN Aperiodic_Type =>
                        NULL;
                     WHEN OTHERS =>
                        Periodic_Task_Ptr(Si.Tcbs(J).Tsk).Jitter:=
                           My_Scheduler.
                           Variables_Table.
                           Entries(Var).Simulation.
                           Integer_Table_Value(Simulations_Range(Z));
                  END CASE;


                  -- user defined task's parameter :
                  --
                  FOR I IN 0..Si.Tcbs(J).Tsk.Param.
                        Nb_Entries-1 LOOP

                     -- Find the user defined variable
                     -- and set its value
                     --
                     FOR K IN 0..Variables_Table.
                           Nb_Entries-1 LOOP
                        IF My_Scheduler.
                              Variables_Table.
                              Entries(K).Variable.
                              Name=
                              To_Unbounded_String(
                              "tasks.") & Si.Tcbs(J).
                              Tsk.Param.
                              Entries(
                              I).Name
                              THEN

                           -- Now, set the value
                           --
                           IF Si.Tcbs(J).Tsk.Param.
                                 Entries(I).Ptype=
                                 Boolean_Parameter THEN
                              Si.Tcbs(J).
                                 Tsk.
                                 Param.Entries(I).
                                 Boolean_Value:=My_Scheduler.
                                 Variables_Table.
                                 Entries(K).
                                 Simulation.
                                 Boolean_Table_Value(
                                 Simulations_Range(Z));
                           END IF;
                           IF Si.Tcbs(J).Tsk.Param.
                                 Entries(I).Ptype=
                                 Integer_Parameter THEN
                              Si.Tcbs(J).
                                 Tsk.
                                 Param.Entries(I).
                                 Integer_Value:=My_Scheduler.
                                 Variables_Table.
                                 Entries(K).
                                 Simulation.
                                 Integer_Table_Value(
                                 Simulations_Range(Z));
                           END IF;
                           IF Si.Tcbs(J).Tsk.Param.
                                 Entries(I).Ptype=
                                 Double_Parameter THEN
                              Si.Tcbs(J).Tsk.
                                 Param.Entries(I).
                                 Double_Value:=My_Scheduler.
                                 Variables_Table.
                                 Entries(K).
                                 Simulation.
                                 Double_Table_Value(
                                 Simulations_Range(Z));
                           END IF;
                           IF Si.Tcbs(J).Tsk.Param.
                                 Entries(I).Ptype=
                                 String_Parameter THEN
                              Si.Tcbs(J).Tsk.
                                 Param.Entries(I).
                                 String_Value:=My_Scheduler.
                                 Variables_Table.
                                 Entries(K).
                                 Simulation.
                                 String_Table_Value(
                                 Simulations_Range(Z));
                           END IF;
                        END IF;
                     END LOOP;
                  END LOOP;

               END IF;
            END LOOP;

         END IF;
      END LOOP;

   END Save_Read_Write_Parametric_Variables;




END Scheduler.Parametrics ;


