------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Translate;
USE Translate;

WITH Text_Io;
USE Text_Io;

WITH Task_Set;
USE Task_Set;




PACKAGE BODY Scheduler.Round_Robin IS






   PROCEDURE Initialize (
         A_Tcb : IN OUT Rr_Tcb) IS
   BEGIN
      A_Tcb.Task_Quantum:= 0;
   END Initialize;


   FUNCTION Build_Tcb (
         My_Scheduler : IN     Round_Robin_Scheduler;
         A_Task       :        Task_Ptr)
     RETURN Tcb_Ptr IS
      A_Tcb : Rr_Tcb_Ptr;
   BEGIN
      A_Tcb:=NEW Rr_Tcb;
      Initialize(Tcb(A_Tcb.All), A_Task);
      Initialize(A_Tcb.All);
      RETURN Tcb_Ptr(A_Tcb);
   END Build_Tcb;


   PROCEDURE Initialize (
         A_Scheduler : IN OUT Round_Robin_Scheduler) IS
   BEGIN
      Reset(A_Scheduler);
      A_Scheduler.Name:= Round_Robin_Protocol;
   END Initialize;




   FUNCTION Copy (
         A_Scheduler : IN     Round_Robin_Scheduler)
     RETURN Scheduler_Ptr IS

      Ptr : Round_Robin_Scheduler_Ptr;

   BEGIN

      Ptr:= NEW Round_Robin_Scheduler;

      Ptr.Name:=A_Scheduler.Name;
      Ptr.Is_Preemptive:=A_Scheduler.Is_Preemptive;
      Ptr.Quantum:=A_Scheduler.Quantum;
      Ptr.Previously_Elected:=A_Scheduler.Previously_Elected;

      RETURN Scheduler_Ptr(Ptr);

   END Copy;


   PROCEDURE Bound_On_Processor_Utilization (
         My_Scheduler   : IN     Round_Robin_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Result         : IN OUT Double;
         Msg            : IN OUT Unbounded_String) IS

   BEGIN
      RAISE Invalid_Scheduler;
   END Bound_On_Processor_Utilization;


   PROCEDURE Compute_Response_Time (
         My_Scheduler   : IN     Round_Robin_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Response_Time  :    OUT Response_Time_Table) IS

   BEGIN
      RAISE Invalid_Scheduler;
   END Compute_Response_Time;





   PROCEDURE Put (
         My_Scheduler : IN     Round_Robin_Scheduler) IS
   BEGIN
      Put(Generic_Scheduler(My_Scheduler));
   END Put;




   PROCEDURE Check_Before_Scheduling (
         My_Scheduler   : IN     Round_Robin_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

   BEGIN
      NULL;
   END Check_Before_Scheduling;




   PROCEDURE Do_Election (
         My_Scheduler      : IN OUT Round_Robin_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Msg               : IN OUT Unbounded_String;
         Current_Time      : IN     Natural;
         Processor_Name    : IN     Unbounded_String;
         My_Dependencies   : IN     Tasks_Dependencies_Ptr;
         With_Offsets      : IN     Boolean;
         With_Precedencies : IN     Boolean;
         With_Resources    : IN     Boolean;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table;
         Elected           : IN OUT Tasks_Range;
         No_Task           : IN OUT Boolean) IS


      -- find the ready task which is directly after the task "Elected"
      --
      I                : Tasks_Range := 0;
      Has_A_Ready_Task : Boolean     := False;
      First_Task       : Tasks_Range := 0;

   BEGIN


      -- Before switching to the next task, first check that
      -- the task quantum is not exceeded. If the current task
      -- can be run .... chose it
      --
      IF (My_Scheduler.Quantum>0)
            AND  (Rr_Tcb_Ptr(Si.Tcbs(My_Scheduler.Previously_Elected)).
            Task_Quantum/=0)
            AND (Si.Tcbs(My_Scheduler.Previously_Elected).Wake_Up_Time <=
            Current_Time)
            AND (Si.Tcbs(My_Scheduler.Previously_Elected).
            Rest_Of_Capacity /= 0)
            AND
            ( (With_Offsets=False) OR
            Check_Offset(Si.Tcbs(I), Current_Time) )
            AND
            ( (With_Precedencies=False) OR
            Check_Precedencies(Si, My_Dependencies,
               Current_Time,
               Si.Tcbs(My_Scheduler.Previously_Elected)) )
            THEN

         No_Task:=False;
         Elected:=My_Scheduler.Previously_Elected;
         Rr_Tcb_Ptr(Si.Tcbs(My_Scheduler.Previously_Elected)).
            Task_Quantum:=
            Rr_Tcb_Ptr(Si.Tcbs(My_Scheduler.Previously_Elected)).
            Task_Quantum-1;

      ELSE

         -- Look for the next task
         --
         IF Si.Tcbs(My_Scheduler.Previously_Elected+1)=NULL
               THEN
            First_Task:=0;
         ELSE
            First_Task:=My_Scheduler.Previously_Elected+1;
         END IF;
         I:=First_Task;

         LOOP
            IF (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) AND
                  (Si.Tcbs(I).Wake_Up_Time <= Current_Time) AND
                  (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                  THEN
               IF (With_Offsets=False) OR
                     Check_Offset(Si.Tcbs(I), Current_Time)
                     THEN
                  IF (With_Precedencies=False) OR
                        Check_Precedencies(Si, My_Dependencies,
                        Current_Time,
                        Si.Tcbs(I))
                        THEN

                     Elected:=I;
                     Has_A_Ready_Task:=True;
                     IF (My_Scheduler.Quantum>0)
                           THEN
                        Rr_Tcb_Ptr(Si.Tcbs(I)).Task_Quantum:=
                           My_Scheduler.Quantum-1;
                     END IF;
                     EXIT;
                  END IF;
               END IF;
            END IF;

            I:=I+1;
            IF Si.Tcbs(I)=NULL
                  THEN
               I:=0;
            END IF;
            EXIT WHEN I=First_Task;
         END LOOP;

         IF Has_A_Ready_Task
               THEN
            No_Task:=False;
         ELSE
            No_Task:=True;
         END IF;

      END IF;

   END Do_Election;



   PROCEDURE Specific_Scheduler_Initialization (
         My_Scheduler   : IN OUT Round_Robin_Scheduler;
         Si             : IN OUT Scheduling_Information;
         Processor_Name : IN     Unbounded_String;
         My_Tasks       : IN OUT Tasks_Set;
         My_Resources   : IN OUT Resources_Set;
         My_Buffers     : IN OUT Buffers_Set;
         My_Messages    : IN     Messages_Set;
         Msg            : IN OUT Unbounded_String) IS

   BEGIN
      NULL;
   END Specific_Scheduler_Initialization;


   PROCEDURE Compute_Activation_Time (
         My_Scheduler : IN     Round_Robin_Scheduler;
         Si           : IN OUT Scheduling_Information;
         Elected      : IN     Tasks_Range;
         Value        : IN OUT Natural) IS


   BEGIN
      RAISE Invalid_Scheduler;
   END Compute_Activation_Time;



   PROCEDURE Compute_Blocking_Time (
         My_Scheduler   : IN     Round_Robin_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Blocking_Time  :    OUT Blocking_Time_Table) IS
   BEGIN
      RAISE Invalid_Scheduler;
   END Compute_Blocking_Time;



   PROCEDURE Utilization_Factor_Feasibility_Test (
         My_Scheduler   : IN     Round_Robin_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Result         : IN OUT Unbounded_String) IS

   BEGIN
      Result := Result &  Lb_Compute_Scheduling_Error_19(Current_Language) &
         Unbounded_Lf;
   END Utilization_Factor_Feasibility_Test;



END Scheduler.Round_Robin;


