------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Text_Io;
USE Text_Io;

WITH Tasks;
USE Tasks;
WITH Resources;
USE Resources;


WITH Task_Offsets;
USE Task_Offsets;
USE Task_Offsets.Offsets_Table_Package;

WITH Qs_Tools;
USE  Qs_Tools;

WITH Ada.Exceptions;
USE Ada.Exceptions;

WITH Expressions;
USE Expressions;

WITH Dependencies;
USE Dependencies;

WITH Messages;
USE Messages;
WITH Buffers;
USE Buffers;
USE Buffers.Buffer_Roles;

WITH Task_Dependencies;
USE Task_Dependencies;
USE Task_Dependencies.Half_Dep_Set;



WITH Unbounded_Strings;
USE Unbounded_Strings;



PACKAGE BODY Scheduler IS




   FUNCTION Export_Aadl_Properties (
         My_Scheduler : IN     Generic_Scheduler;
         Number_Of_Ht : IN     Natural)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;

   BEGIN


      FOR I IN 1..Number_Of_Ht LOOP
         Result:= Result & Ascii.Ht;
      END LOOP;

      Result:= Result & To_Unbounded_String(
         "Scheduling_Protocol => " & To_String(Get_Name(
               My_Scheduler)) & ";" ) & Unbounded_Lf;


      FOR I IN 1..Number_Of_Ht LOOP
         Result:= Result & Ascii.Ht;
      END LOOP;

      Result:= Result & To_Unbounded_String(
         "Cheddar_Properties::Scheduler_Quantum => " &
         Get_Quantum(My_Scheduler)  & ";") &
         Unbounded_Lf;


      FOR I IN 1..Number_Of_Ht LOOP
         Result:= Result & Ascii.Ht;
      END LOOP;

      IF Get_Preemptive(My_Scheduler) =
            Preemptive THEN
         Result:= Result & To_Unbounded_String(
            "Cheddar_Properties::Preemptive_Scheduler => True;") &
            Unbounded_Lf;
      ELSE
         Result:= Result & To_Unbounded_String(
            "Cheddar_Properties::Preemptive_Scheduler => False;") &
            Unbounded_Lf;
      END IF;


      RETURN Result;

   END Export_Aadl_Properties;



   FUNCTION Export_Xml (
         My_Scheduler : IN     Generic_Scheduler)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;

      Quantum       : CONSTANT Natural          := Get_Quantum (My_Scheduler);
      Is_Preemptive : CONSTANT Preemptive_Type  := Get_Preemptive (My_Scheduler);
      Name          :          Unbounded_String := Get_Name (My_Scheduler);

   BEGIN

      Result:= Result & To_Unbounded_String( Ascii.Ht &
         Ascii.Ht &
         "<scheduler" );


      IF Quantum/=0
            THEN
         Result:= Result & To_Unbounded_String( " quantum=""" &
            Quantum'Img & """");
      END IF;


      IF Is_Preemptive /= Preemptive
            THEN
         Result:= Result & To_Unbounded_String(
            " is_preemptive=""" &
            Is_Preemptive'Img & """");
      END IF;

      Result:= Result & To_Unbounded_String( ">" );
      Result:= Result & To_Unbounded_String(" ") & Name &
         To_Unbounded_String(" ");
      Result:= Result & To_Unbounded_String(  "</scheduler>" ) &
         Unbounded_Lf;



      RETURN Result;


   END Export_Xml;




   PROCEDURE Set_Preemptive (
         My_Scheduler : IN OUT Generic_Scheduler'Class;
         Preempt      : IN     Preemptive_Type) IS
   BEGIN
      My_Scheduler.Is_Preemptive:=Preempt;
   END Set_Preemptive;


   FUNCTION Get_Preemptive (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN Preemptive_Type IS
   BEGIN
      RETURN My_Scheduler.Is_Preemptive;
   END Get_Preemptive;



   FUNCTION Get_Preemptive (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN String IS
   BEGIN
      RETURN My_Scheduler.Is_Preemptive'Img;
   END Get_Preemptive;



   PROCEDURE Put (
         My_Scheduler : IN     Generic_Scheduler) IS
   BEGIN
      Put_Line("Scheduler name : " & My_Scheduler.Name'Img);
      Put_Line("Is preemptive  : " & My_Scheduler.Is_Preemptive'Img);
      Put_Line("Quantum        : " & My_Scheduler.Quantum'Img);
   END Put;


   PROCEDURE Reset (
         A_Scheduler : IN OUT Generic_Scheduler'Class) IS
   BEGIN
      A_Scheduler.Is_Preemptive:=Preemptive;
      A_Scheduler.Quantum:=0;
   END Reset;



   FUNCTION Get_Name (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN Unbounded_String IS
   BEGIN
      RETURN To_Unbounded_String(My_Scheduler.Name'Img);
   END Get_Name;

   FUNCTION Get_Name (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN Scheduler_Type IS
   BEGIN
      RETURN My_Scheduler.Name;
   END Get_Name;


   FUNCTION Build_Tcb (
         My_Scheduler : IN     Generic_Scheduler;
         A_Task       :        Task_Ptr)
     RETURN Tcb_Ptr IS

      A_Tcb : Tcb_Ptr;
   BEGIN
      A_Tcb:=NEW Tcb;
      Initialize(A_Tcb.All, A_Task);
      RETURN A_Tcb;
   END Build_Tcb;


   PROCEDURE Initialize (
         A_Tcb  : IN OUT Tcb;
         A_Task :        Task_Ptr) IS

      Seed : Generator;

   BEGIN
      A_Tcb.Tsk:=A_Task;

      A_Tcb.Activation:=1;
      A_Tcb.End_Time:=0;
      A_Tcb.Used_Cpu:=0;
      A_Tcb.Wake_Up_Time:=A_Tcb.Tsk.Start_Time;
      A_Tcb.Rest_Of_Capacity:=A_Tcb.Tsk.Capacity;

      IF (A_Tcb.Tsk.Task_Type = Poisson_Type)
            OR (A_Tcb.Tsk.Task_Type = Parametric_Type)
            THEN
         IF (Poisson_Task_Ptr(A_Tcb.Tsk).Predictable)
               THEN
            Reset(Seed, Poisson_Task_Ptr(A_Tcb.Tsk).Seed);
         ELSE
            Reset(Seed);
         END IF;

         Save(Seed, A_Tcb.Task_Seed);
      END IF;
   END Initialize;


   PROCEDURE Scheduler_Initialization (
         My_Scheduler            : IN OUT Generic_Scheduler'Class;
         Si                      : IN OUT Scheduling_Information;
         Processor_Name          : IN     Unbounded_String;
         My_Tasks                : IN OUT Tasks_Set;
         My_Resources            : IN OUT Resources_Set;
         My_Buffers              : IN OUT Buffers_Set;
         Result                  : IN OUT Scheduling_Sequence_Ptr;
         With_Offsets            :        Boolean;
         With_Precedencies       :        Boolean;
         With_Resources          :        Boolean;
         With_Task_Specific_Seed :        Boolean                            := True;
         Global_Seed_Value       :        Integer                            := 0;
         Predictable_Global_Seed :        Boolean                            := True;
         Given_Last_Time         : IN     Natural;
         Event_To_Generate       : IN     Time_Unit_Event_Type_Boolean_Table) IS


      Iterator1     : Tasks_Iterator;
      Iterator2     : Resources_Iterator;
      A_Buffer_Size : Buffer_Scheduling_Information_Ptr;
      Iterator3     : Buffers_Iterator;
      A_Buffer      : Buffer_Ptr;
      A_Task        : Task_Ptr;
      A_Resource    : Resource_Ptr;
      Seed          : Generator;
      A_Item        : Time_Unit_Event_Ptr;


   BEGIN

      My_Scheduler.Previously_Elected := Tasks_Range'First;

      IF  (Predictable_Global_Seed)
            THEN
         Reset(Seed,Global_Seed_Value);
      ELSE
         Reset(Seed);
      END IF;

      Si.With_Specific_Task_Seed :=With_Task_Specific_Seed;
      Save(Seed,Si.Global_Seed);


      -- Build Task Control Blocks
      --
      Reset_Iterator(My_Tasks, Iterator1);

      LOOP
         Current_Element(My_Tasks, A_Task, Iterator1);

         Si.Tcbs(Si.Number_Of_Tasks):=Build_Tcb(
            My_Scheduler, A_Task);

         Si.Number_Of_Tasks:=Si.Number_Of_Tasks+1;

         EXIT WHEN Is_Last_Element(My_Tasks, Iterator1);
         Next_Element(My_Tasks, Iterator1);
      END LOOP;

      -- Build of buffer list with their current size
      -- (in order to generate Write/Read buffer events
      --
      IF (With_Precedencies)
            THEN
         IF (Get_Number_Of_Elements(My_Buffers)>0)
               THEN

            Reset_Iterator(My_Buffers, Iterator3);
            LOOP

               Current_Element(My_Buffers, A_Buffer, Iterator3);

               A_Buffer_Size:= NEW Buffer_Scheduling_Information;
               A_Buffer_Size.Written_Buffer:=A_Buffer;
               A_Buffer_Size.Current_Size:=0;
               Add(Si.Written_Buffers, A_Buffer_Size);

               EXIT WHEN Is_Last_Element(My_Buffers, Iterator3);
               Next_Element(My_Buffers, Iterator3);

            END LOOP;
         END IF;
      END IF;

      IF (With_Resources)
            THEN
         IF (Get_Number_Of_Elements(My_Resources)>0)
               THEN

            Reset_Iterator(My_Resources, Iterator2);
            LOOP

               Current_Element(My_Resources, A_Resource, Iterator2);

               Si.Shared_Resources(Si.
                  Number_Of_Resources):=Build_Resource(My_Scheduler,
                  A_Resource);

               Si.Number_Of_Resources:=Si.
                  Number_Of_Resources+1;

               EXIT WHEN Is_Last_Element(My_Resources, Iterator2);
               Next_Element(My_Resources, Iterator2);


            END LOOP;
         END IF;
      END IF;

      Initialize(Result.All);

      -- First activation for all tasks
      --
      FOR I IN 0 .. Si.Number_Of_Tasks-1 LOOP

         IF  Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name
               THEN

            IF   Event_To_Generate(Task_Activation) THEN
               A_Item:= NEW Time_Unit_Event(Task_Activation);
               A_Item.Activation_Task:=Si.Tcbs(I).Tsk;

               Add(Result.All, A_Item, Si.Tcbs(I).Tsk.Start_Time);
            END IF;


         END IF;

      END LOOP;

   END Scheduler_Initialization;



   PROCEDURE Build_Scheduling_Sequence (
         My_Scheduler            : IN OUT Generic_Scheduler'Class;
         My_Tasks                : IN     Tasks_Set;
         My_Resources            : IN     Resources_Set;
         My_Buffers              : IN     Buffers_Set;
         My_Messages             : IN     Messages_Set;
         My_Dependencies         : IN     Tasks_Dependencies_Ptr;
         Processor_Name          : IN     Unbounded_String;
         Result                  :    OUT Scheduling_Sequence_Ptr;
         Msg                     : IN OUT Unbounded_String;
         Event_To_Generate       : IN     Time_Unit_Event_Type_Boolean_Table;
         Given_Last_Time         :        Natural                            := 0;
         With_Offsets            :        Boolean                            := True;
         With_Precedencies       :        Boolean                            := True;
         With_Resources          :        Boolean                            := True;
         With_Task_Specific_Seed :        Boolean                            := True;
         Global_Seed_Value       :        Integer                            := 0;
         Predictable_Global_Seed :        Boolean                            := True) IS


      Tmp_Tasks     : Tasks_Set;
      Tmp_Resources : Resources_Set;
      Tmp_Buffers   : Buffers_Set;
      Tmp_Messages  : Messages_Set;

      Last_Time    : Natural;
      Current_Time : Natural := 0;


      -- index for scheduling table
      --
      Elected : Tasks_Range := Tasks_Range'First;
      No_Task : Boolean     := True;

      Si : Scheduling_Information;


   BEGIN


      Si.Number_Of_Tasks:=0;
      Si.Number_Of_Resources:=0;





      -- Before scheduling, check if the tasks can be
      -- scheduled with the selected scheduler
      --

      Check_Before_Scheduling(My_Scheduler, My_Tasks, Processor_Name);



      -- First : gather information
      --
      Current_Processor_Name:=Processor_Name;

      Select_And_Copy(
         My_Tasks, Tmp_Tasks, Select_Cpu'ACCESS);
      Select_And_Copy(
         My_Resources, Tmp_Resources, Select_Cpu'ACCESS);
      Select_And_Copy(
         My_Buffers, Tmp_Buffers, Select_Cpu'ACCESS);
      Duplicate(My_Messages, Tmp_Messages);


      Scheduler_Initialization (
         My_Scheduler, Si, Processor_Name, Tmp_Tasks, Tmp_Resources,
         Tmp_Buffers,
         Result, With_Offsets, With_Precedencies, With_Resources,
         With_Task_Specific_Seed, Global_Seed_Value,
         Predictable_Global_Seed,
         Given_Last_Time, Event_To_Generate);


      Specific_Scheduler_Initialization(
         My_Scheduler, Si, Processor_Name, Tmp_Tasks, Tmp_Resources,
         Tmp_Buffers, Tmp_Messages, Msg);




      -- Compute last scheduled time unit
      --

      IF(Given_Last_Time = 0)
            THEN
         Last_Time:=Natural'Min (Max_Scheduling_Period,
            Scheduling_Period (Tmp_Tasks, Processor_Name));
      ELSE
         Last_Time:=Given_Last_Time;
      END IF;
      Si.Simulation_Length:=Last_Time;


      -- Second : compute scheduling
      --
      WHILE(Current_Time < Last_Time) LOOP

         Do_Election(My_Scheduler, Si,  Result, Msg,
            Current_Time, Processor_Name,
            My_Dependencies,
            With_Offsets, With_Precedencies, With_Resources,
            Event_To_Generate,
            Elected, No_Task);


         -- No task to schedule
         --
         IF No_Task
               THEN
            Current_Time:=Current_Time+1;

         ELSE

            Run_Task(My_Scheduler, Processor_Name, Si, My_Dependencies,
               Elected, Result,
               Current_Time,  Last_Time,
               With_Offsets, With_Precedencies, With_Resources,
               Event_To_Generate);

         END IF;

      END LOOP;


   END Build_Scheduling_Sequence;



   PROCEDURE Run_Task (
         My_Scheduler      : IN OUT Generic_Scheduler'Class;
         Processor_Name    : IN     Unbounded_String;
         Si                : IN OUT Scheduling_Information;
         My_Dependencies   : IN     Tasks_Dependencies_Ptr;
         Elected           : IN     Tasks_Range;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN OUT Natural;
         Last_Time         : IN     Natural;
         With_Offsets      :        Boolean                            := True;
         With_Precedencies :        Boolean                            := True;
         With_Resources    :        Boolean                            := True;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table) IS


      -- For Poisson Process task
      --
      Seed : Generator;
      Temp : Natural;

      -- For Parametric task
      --
      Parametric_Delay : Natural := 0;

      A_Item : Time_Unit_Event_Ptr;



   BEGIN

      ------------------------------------------------------
      -- Allocate/release shared resources
      ------------------------------------------------------

      IF With_Resources
            THEN
         Allocate_Resource(My_Scheduler, Si, Result,
            Current_Time,
            Si.Tcbs(Elected), Event_To_Generate);

         Release_Resource(My_Scheduler, Si, Result,
            Current_Time,
            Si.Tcbs(Elected), Event_To_Generate);
      END IF;


      ------------------------------------------------------
      -- Send/Receive Messages
      -- write/read on buffers
      ------------------------------------------------------

      IF  With_Precedencies
            THEN

         Send_Message(My_Scheduler, Si, Result,
            Current_Time,
            Si.Tcbs(Elected), My_Dependencies, Event_To_Generate);

         Receive_Message(My_Scheduler, Si, Result,
            Current_Time,
            Si.Tcbs(Elected), My_Dependencies, Event_To_Generate);

         Buffer_Write(My_Scheduler, Si, Result,
            Current_Time,
            Si.Tcbs(Elected), My_Dependencies, Event_To_Generate);

         Buffer_Read(My_Scheduler, Si, Result,
            Current_Time,
            Si.Tcbs(Elected), My_Dependencies, Event_To_Generate);

      END IF;


      ------------------------------------------------------
      -- Computed scheduling
      ------------------------------------------------------

      -- Preemptive case
      --


      IF (My_Scheduler.Is_Preemptive = Preemptive)
            THEN

         IF Si.Tcbs(Elected).Rest_Of_Capacity=Si.Tcbs(Elected).Tsk.
               Capacity THEN
            IF   Event_To_Generate(Start_Of_Task_Capacity) THEN
               A_Item:= NEW Time_Unit_Event(Start_Of_Task_Capacity);
               A_Item.Start_Task:=Si.Tcbs(Elected).Tsk;
               Add(Result.All, A_Item, Current_Time);
            END IF;
         END IF;

         IF   Event_To_Generate(Running_Task) THEN
            A_Item:=Produce_Running_Task_Event (My_Scheduler, Si.Tcbs(
                  Elected));
            Add(Result.All, A_Item, Current_Time);
         END IF;


         Si.Tcbs(Elected).Rest_Of_Capacity:=Si.
            Tcbs(Elected).Rest_Of_Capacity - 1;
         Si.Tcbs(Elected).Used_Cpu:=Si.Tcbs(Elected).Used_Cpu + 1;


         Current_Time:=Current_Time + 1;

         IF Si.Tcbs(Elected).Rest_Of_Capacity=0 THEN
            IF   Event_To_Generate(End_Of_Task_Capacity) THEN
               A_Item:= NEW Time_Unit_Event(End_Of_Task_Capacity);
               A_Item.End_Task:=Si.Tcbs(Elected).Tsk;
               Add(Result.All, A_Item, Current_Time);
            END IF;
         END IF;


      ELSE


         -- Non preemptive case
         --

         IF   Event_To_Generate(Start_Of_Task_Capacity) THEN
            A_Item:= NEW Time_Unit_Event(Start_Of_Task_Capacity);
            A_Item.Start_Task:=Si.Tcbs(Elected).Tsk;
            Add(Result.All, A_Item, Current_Time);
         END IF;

         FOR I IN 0 .. (Si.Tcbs(Elected).Tsk.Capacity - 1)
               LOOP
            IF   Event_To_Generate(Running_Task) THEN
               A_Item:=Produce_Running_Task_Event (My_Scheduler, Si.Tcbs(
                     Elected));
               Add(Result.All, A_Item, Current_Time);
            END IF;
            Current_Time:=Current_Time + 1;
            Si.Tcbs(Elected).Used_Cpu:=Si.Tcbs(Elected).Used_Cpu + 1;
         END LOOP;

         IF   Event_To_Generate(End_Of_Task_Capacity) THEN
            A_Item:= NEW Time_Unit_Event(End_Of_Task_Capacity);
            A_Item.End_Task:=Si.Tcbs(Elected).Tsk;
            Add(Result.All, A_Item, Current_Time);
         END IF;

         Si.Tcbs(Elected).Rest_Of_Capacity:=0;

      END IF;



      ------------------------------------------------------
      -- Compute the next task activations
      ------------------------------------------------------


      IF (Si.Tcbs(Elected).Rest_Of_Capacity = 0) THEN

         Si.Tcbs(Elected).End_Time:=Current_Time;
         Si.Tcbs(Elected).Activation:=
            Si.Tcbs(Elected).Activation+1;

         CASE Si.Tcbs(Elected).Tsk.Task_Type IS

            WHEN Periodic_Type =>
               Si.Tcbs(Elected).Rest_Of_Capacity:=
                  Si.Tcbs(Elected).Tsk.Capacity;
               Si.Tcbs(Elected).Wake_Up_Time:=
                  Si.Tcbs(Elected).Wake_Up_Time +
                  Periodic_Task_Ptr(Si.Tcbs(Elected).Tsk).Period;


            WHEN Poisson_Type =>
               Si.Tcbs(Elected).Rest_Of_Capacity:=
                  Si.Tcbs(Elected).Tsk.Capacity;

               IF Si.With_Specific_Task_Seed THEN
                  Reset(Seed,Si.Tcbs(Elected).Task_Seed);
               ELSE
                  Reset(Seed,Si.Global_Seed);
               END IF;

               Temp := Natural(Get_Exponential_Time(Double
                     (Poisson_Task_Ptr(Si.Tcbs(Elected).Tsk).Period),Seed));

               Si.Tcbs(Elected).Wake_Up_Time:=Si.
                  Tcbs(Elected).Wake_Up_Time + Temp;

               IF Si.With_Specific_Task_Seed THEN
                  Save(Seed,Si.Tcbs(Elected).Task_Seed);
               ELSE
                  Save(Seed,Si.Global_Seed);
               END IF;

            WHEN Sporadic_Type =>
               Si.Tcbs(Elected).Rest_Of_Capacity:=
                  Si.Tcbs(Elected).Tsk.Capacity;

               IF Si.With_Specific_Task_Seed THEN
                  Reset(Seed,Si.Tcbs(Elected).Task_Seed);
               ELSE
                  Reset(Seed,Si.Global_Seed);
               END IF;

               Temp := Natural(Get_Exponential_Time(Double
                     (Poisson_Task_Ptr(Si.Tcbs(Elected).Tsk).Period),Seed));

               Si.Tcbs(Elected).Wake_Up_Time:=Si.
                  Tcbs(Elected).Wake_Up_Time + Natural'Max(Temp,
                  Periodic_Task_Ptr(Si.Tcbs(Elected).Tsk).Period);

               IF Si.With_Specific_Task_Seed THEN
                  Save(Seed,Si.Tcbs(Elected).Task_Seed);
               ELSE
                  Save(Seed,Si.Global_Seed);
               END IF;



            WHEN Aperiodic_Type =>
               NULL;


            WHEN Parametric_Type =>
               Si.Tcbs(Elected).Rest_Of_Capacity:=
                  Si.Tcbs(Elected).Tsk.Capacity;


               -- The scheduler has to be a parametric one !!!
               --
               IF (My_Scheduler.Name/=Parametric_Protocol) THEN
                  Raise_Exception(Statement_Error'Identity,
                     "Parametric task only permitted with parametric scheduler" );
               END IF;

               Compute_Activation_Time(My_Scheduler,
                  Si,
                  Elected, Parametric_Delay);

               Si.Tcbs(Elected).Wake_Up_Time:=Si.
                  Tcbs(Elected).Wake_Up_Time + Parametric_Delay;

         END CASE;


         -- Add to the scheduling table the next activation time
         --
         IF Si.Tcbs(Elected).Tsk.Task_Type /= Aperiodic_Type AND
               Si.Tcbs(Elected).Wake_Up_Time <= Last_Time THEN

            IF   Event_To_Generate(Task_Activation) THEN
               A_Item:= NEW Time_Unit_Event(Task_Activation);
               A_Item.Activation_Task:=Si.Tcbs(Elected).Tsk;
               Add(Result.All, A_Item, Si.Tcbs(Elected).
                  Wake_Up_Time);
            END IF;

         END IF;
      END IF;


      My_Scheduler.Previously_Elected:=Elected;



   END Run_Task;


   -- Return true is the task "task_id" can be schedule now
   -- according its offsets
   --
   FUNCTION Check_Offset (
         A_Tcb   : Tcb_Ptr;
         At_Time : Natural)
     RETURN Boolean IS
      Target_Time : Integer := 0;
   BEGIN

      FOR K IN 0..A_Tcb.Tsk.Offset.Nb_Entries-1 LOOP
         IF A_Tcb.Tsk.Offset.Entries(K).Activation=A_Tcb.Activation
               THEN
            Target_Time:=A_Tcb.Wake_Up_Time+A_Tcb.Tsk.Offset.Entries(
               K).Value;
            IF At_Time<Target_Time
                  THEN
               RETURN False;
            END IF;
         END IF;
      END LOOP;

      RETURN True;
   END Check_Offset;





   FUNCTION Select_Cpu (
         Op1 : IN     Task_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Op1.Cpu_Name = Current_Processor_Name);
   END Select_Cpu;


   FUNCTION Select_Cpu (
         Op1 : IN     Resource_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Op1.Cpu_Name = Current_Processor_Name);
   END Select_Cpu;


   FUNCTION Select_Cpu (
         Op1 : IN     Buffer_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Op1.Cpu_Name = Current_Processor_Name);
   END Select_Cpu;




   PROCEDURE Send_Message (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Deps              :        Tasks_Dependencies_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table) IS

      A_Item                           : Time_Unit_Event_Ptr;
      My_Iterator                      : Tasks_Dependencies_Iterator;
      A_Half_Dep_Ptr                   : Dependency_Ptr;
      A_Message                        : Message_Ptr;
      A_Message_Scheduling_Information : Message_Scheduling_Information_Ptr;





   BEGIN

      -- is the time to send message ?
      --
      IF (A_Tcb.Rest_Of_Capacity = 1) THEN


         -- Check if the task has a message to send
         --
         IF NOT Is_Empty(Deps.All) THEN
            Reset_Iterator(Deps.All, My_Iterator);
            LOOP
               Current_Element(Deps.All, A_Half_Dep_Ptr, My_Iterator);


               -- Is the task concerned ?
               -- Is it a communication dependency ?
               -- Is it a sending operation ?
               --
               IF ( A_Half_Dep_Ptr.Depend_Type=
                     Communication_Dependency ) THEN
                  IF ( A_Half_Dep_Ptr.Communication_Dependent_Task.Name =
                        A_Tcb.Tsk.Name )
                        AND  ( A_Half_Dep_Ptr.Communication_Orientation  =
                        From_Task_To_Object) THEN

                     A_Message := A_Half_Dep_Ptr.
                        Communication_Dependency_Object;

                     A_Message_Scheduling_Information:= NEW
                        Message_Scheduling_Information;
                     A_Message_Scheduling_Information.Send_Time :=
                        Current_Time+1;
                     A_Message_Scheduling_Information.Sended_Message :=
                        A_Message;
                     Add(Si.Sended_Messages,
                        A_Message_Scheduling_Information);


                     IF   Event_To_Generate(Send_Message) THEN
                        A_Item:= NEW Time_Unit_Event(Send_Message);
                        A_Item.Send_Message:=A_Message;
                        A_Item.Send_Task:=A_Tcb.Tsk;
                        Add(Result.All, A_Item,
                           A_Message_Scheduling_Information.Send_Time);
                     END IF;

                  END IF;
               END IF;

               EXIT WHEN Is_Last_Element(Deps.All, My_Iterator);

               Next_Element(Deps.All, My_Iterator);

            END LOOP;

         END IF;
      END IF;
   END Send_Message;


   PROCEDURE Receive_Message (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Deps              :        Tasks_Dependencies_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table) IS

      A_Item         : Time_Unit_Event_Ptr;
      My_Iterator    : Tasks_Dependencies_Iterator;
      A_Half_Dep_Ptr : Dependency_Ptr;
      A_Message      : Message_Ptr;
      List_Ite       : Message_Scheduling_Information_Iterator;
      Msg            : Message_Scheduling_Information_Ptr;

   BEGIN

      -- Is The time to receive a message ?
      --
      IF (A_Tcb.Rest_Of_Capacity = A_Tcb.Tsk.Capacity) THEN


         -- Check the sended_messages to see if this
         -- messages was already sent
         --
         IF NOT Is_Empty(Si.Sended_Messages) THEN

            Reset_Head_Iterator(Si.Sended_Messages, List_Ite);
            LOOP
               Current_Element (Si.Sended_Messages, Msg,
                  List_Ite);


               Reset_Iterator(Deps.All, My_Iterator);
               LOOP
                  Current_Element(Deps.All, A_Half_Dep_Ptr, My_Iterator);

                  -- Is the task concerned ?
                  -- Is it a communication dependency ?
                  -- Is it a receiving operation ?
                  --
                  IF ( A_Half_Dep_Ptr.Depend_Type=
                        Communication_Dependency ) THEN
                     IF ( A_Half_Dep_Ptr.Communication_Dependent_Task.
                           Name =
                           A_Tcb.Tsk.Name )
                           AND  ( A_Half_Dep_Ptr.
                           Communication_Orientation  =
                           From_Object_To_Task) THEN


                        A_Message := A_Half_Dep_Ptr.
                           Communication_Dependency_Object;

                        -- Check the sended_messages to see if this
                        -- messages was already sent
                        --
                        -- Is it the right message ?
                        -- Is the message arrived at the receiving processor ?
                        --
                        IF (Msg.Sended_Message.Name=A_Message.
                              Name) THEN

                           IF  ( Current_Time >= (Msg.Send_Time+
                                    A_Message.
                                    Response_Time-1)) THEN

                              IF   Event_To_Generate(Receive_Message) THEN
                                 A_Item:= NEW Time_Unit_Event(
                                    Receive_Message);
                                 A_Item.Receive_Message:=A_Message;
                                 A_Item.Receive_Task:=A_Tcb.Tsk;
                                 Add(Result.All, A_Item,
                                    Current_Time);
                              END IF;

                              -- Delete received message
                              --
                              Delete(Si.Sended_Messages, Msg);

                              -- Currently, a task reads only one message per activation
                              --
                              RETURN;

                           END IF;
                        END IF;

                     END IF;
                  END IF;


                  EXIT WHEN Is_Last_Element(Deps.All, My_Iterator);

                  Next_Element(Deps.All, My_Iterator);

               END LOOP;

               IF  Is_Tail_Element (Si.Sended_Messages,
                     List_Ite) THEN
                  EXIT;
               END IF;
               Next_Element (Si.Sended_Messages, List_Ite);
            END LOOP;

         END IF;
      END IF;



   END Receive_Message;


   PROCEDURE Buffer_Read (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Deps              :        Tasks_Dependencies_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table) IS

      A_Item         : Time_Unit_Event_Ptr;
      My_Iterator    : Tasks_Dependencies_Iterator;
      A_Half_Dep_Ptr : Dependency_Ptr;
      A_Buffer       : Buffer_Ptr;
      A_Buffer_Size  : Buffer_Scheduling_Information_Ptr;
      List_Ite       : Buffer_Scheduling_Information_Iterator;


   BEGIN

      -- Check if the task has a buffer to write
      --
      IF ( Has_Buffer_To_Read (Deps, A_Tcb.Tsk )  ) THEN

         -- for all tasks which have a precedency with the current task
         --
         LOOP
            Current_Element(Deps.All, A_Half_Dep_Ptr, My_Iterator);

            IF ( A_Half_Dep_Ptr.Depend_Type=
                  Buffer_Dependency ) THEN
               IF (A_Half_Dep_Ptr.Buffer_Dependent_Task.Name =
                     A_Tcb.Tsk.Name)
                     AND (A_Half_Dep_Ptr.Buffer_Orientation=
                     From_Object_To_Task)
                     THEN

                  A_Buffer := A_Half_Dep_Ptr.
                     Buffer_Dependency_Object;

                  -- Must we generate Read event ?
                  --
                  IF  Event_To_Generate(Read_From_Buffer) THEN
                     FOR I IN 0..A_Buffer.Roles.Nb_Entries-1 LOOP

                        -- Is the right task ?
                        --
                        IF (A_Buffer.Roles.Entries(I).Item = A_Tcb.
                              Tsk.Name)
                              THEN

                           -- Is it time to read ?
                           --
                           IF (A_Buffer.Roles.Entries(I).Data.Time =
                                 (A_Tcb.Tsk.Capacity +1
                                    -A_Tcb.Rest_Of_Capacity) )
                                 THEN

                              -- Does the task is consumer ?
                              --
                              IF  (A_Buffer.Roles.Entries(I).Data.
                                    The_Role =
                                    Consumer)
                                    THEN

                                 -- Can we find some information to read ?
                                 --
                                 IF NOT Is_Empty(Si.Written_Buffers) THEN

                                    Reset_Head_Iterator(
                                       Si.Written_Buffers, List_Ite);
                                    LOOP
                                       Current_Element (
                                          Si.Written_Buffers,
                                          A_Buffer_Size,
                                          List_Ite);

                                       IF
                                             (
                                             A_Buffer_Size.Written_Buffer.Name =
                                             A_Buffer.Name)
                                             AND
                                             (A_Buffer_Size.Current_Size>=
                                             A_Buffer.Roles.Entries(I).
                                             Data.Size)
                                             THEN
                                          A_Item:= NEW Time_Unit_Event(
                                             Read_From_Buffer);

                                          A_Item.Read_Buffer:=A_Buffer;
                                          A_Item.Read_Task:=A_Tcb.Tsk;
                                          A_Item.Read_Size:=
                                             A_Buffer.Roles.
                                             Entries(I).
                                             Data.
                                             Size;
                                          Add(Result.All, A_Item,
                                             Current_Time);
                                          EXIT;
                                       END IF;

                                       IF  Is_Tail_Element (
                                             Si.Written_Buffers,
                                             List_Ite) THEN
                                          EXIT;
                                       END IF;
                                       Next_Element (Si.Written_Buffers,
                                          List_Ite);
                                    END LOOP;
                                 END IF;


                              END IF;
                           END IF;
                        END IF;
                     END LOOP;
                  END IF;


               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(Deps.All, My_Iterator);

            Next_Element(Deps.All, My_Iterator);

         END LOOP;

      END IF;

   END Buffer_Read;







   PROCEDURE Buffer_Write (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Deps              :        Tasks_Dependencies_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table) IS


      A_Item         : Time_Unit_Event_Ptr;
      My_Iterator    : Tasks_Dependencies_Iterator;
      A_Half_Dep_Ptr : Dependency_Ptr;
      A_Buffer       : Buffer_Ptr;
      A_Buffer_Size  : Buffer_Scheduling_Information_Ptr;
      List_Ite       : Buffer_Scheduling_Information_Iterator;


   BEGIN

      -- Check if the task has a buffer to write
      --
      IF ( Has_Buffer_To_Write (Deps, A_Tcb.Tsk ) ) THEN

         -- for all tasks which have a precedency with the current task
         --
         LOOP
            Current_Element(Deps.All, A_Half_Dep_Ptr, My_Iterator);


            IF ( A_Half_Dep_Ptr.Depend_Type=
                  Buffer_Dependency ) THEN

               IF (A_Half_Dep_Ptr.Buffer_Dependent_Task .Name =
                     A_Tcb.Tsk.Name)
                     AND (A_Half_Dep_Ptr.Buffer_Orientation=
                     From_Task_To_Object)
                     THEN

                  A_Buffer := A_Half_Dep_Ptr.
                     Buffer_Dependency_Object;

                  -- Must we generate Write event ?
                  --
                  IF   Event_To_Generate(Write_To_Buffer)
                        THEN
                     FOR I IN 0..A_Buffer.Roles.Nb_Entries-1 LOOP

                        -- Is the right task ?
                        --
                        IF (A_Buffer.Roles.Entries(I).Item = A_Tcb.
                              Tsk.Name)
                              THEN
                           -- Does the task is producer ?
                           --
                           IF  (A_Buffer.Roles.Entries(I).Data.
                                 The_Role =
                                 Producer)
                                 THEN
                              -- Is it time to write ?
                              --
                              IF (A_Buffer.Roles.Entries(I).Data.Time =
                                    (A_Tcb.Tsk.Capacity +1
                                       -A_Tcb.Rest_Of_Capacity) )
                                    THEN
                                 A_Item:= NEW Time_Unit_Event(
                                    Write_To_Buffer);

                                 A_Item.Write_Buffer:=A_Buffer;
                                 A_Item.Write_Task:=A_Tcb.Tsk;
                                 A_Item.Write_Size:=A_Buffer.Roles.
                                    Entries(I).
                                    Data.
                                    Size;
                                 Add(Result.All, A_Item,
                                    Current_Time);

                                 -- Add the size write event to the buffer list
                                 --
                                 IF NOT Is_Empty(Si.Written_Buffers) THEN

                                    Reset_Head_Iterator(
                                       Si.Written_Buffers, List_Ite);
                                    LOOP
                                       Current_Element (
                                          Si.Written_Buffers,
                                          A_Buffer_Size,
                                          List_Ite);

                                       IF
                                             A_Buffer_Size.Written_Buffer.Name =
                                             A_Buffer.Name
                                             THEN
                                          A_Buffer_Size.Current_Size:=
                                             A_Buffer_Size.Current_Size+
                                             A_Item.Write_Size;
                                          EXIT;
                                       END IF;

                                       IF  Is_Tail_Element (
                                             Si.Written_Buffers,
                                             List_Ite) THEN
                                          EXIT;
                                       END IF;
                                       Next_Element (Si.Written_Buffers,
                                          List_Ite);
                                    END LOOP;
                                 END IF;





                              END IF;
                           END IF;
                        END IF;
                     END LOOP;
                  END IF;


               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(Deps.All, My_Iterator);

            Next_Element(Deps.All, My_Iterator);

         END LOOP;

      END IF;

   END Buffer_Write;



   PROCEDURE Release_Resource (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table) IS


      Current_Capacity : Natural             := 0;
      A_Item           : Time_Unit_Event_Ptr;

   BEGIN

      IF (Si.Number_Of_Resources>0)
            THEN
         Current_Capacity:=A_Tcb.Tsk.Capacity-A_Tcb.
            Rest_Of_Capacity+1;

         FOR Index1 IN 0.. Si.Number_Of_Resources -1 LOOP


            -- For each resource, check if the task hold resources
            -- and then release them
            --
            FOR Index2 IN 0 .. Si.Shared_Resources(Index1).
                  Shared.Task_Tab.Nb_Entries-1
                  LOOP


               IF (Current_Capacity=Si.Shared_Resources(Index1).
                     Shared.Task_Tab.Entries(Index2).Data.Task_End)
                     AND (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.Name)
                     THEN

                  Si.Shared_Resources(Index1).Shared.State:=
                     Si.Shared_Resources(Index1).Shared.State+1;
                  FOR I IN 0 .. Si.Shared_Resources(Index1).
                        Nb_Allocated-1 LOOP
                     IF Si.Shared_Resources(Index1).
                           Allocated_By(I)=A_Tcb.Tsk.Name
                           THEN

                        Si.Shared_Resources(Index1).
                           Allocated_By(I):= Si.
                           Shared_Resources(Index1).Allocated_By(
                           Si.Shared_Resources(Index1).
                           Nb_Allocated-1);
                        Si.Shared_Resources(Index1).
                           Nb_Allocated:=Si.Shared_Resources(
                           Index1).Nb_Allocated-1;


                        IF   Event_To_Generate(Release_Resource) THEN
                           A_Item:= NEW Time_Unit_Event(
                              Release_Resource);
                           A_Item.Release_Resource:=Si.
                              Shared_Resources(
                              Index1).
                              Shared;
                           A_Item.Release_Task:=A_Tcb.Tsk;
                           Add(Result.All, A_Item, Current_Time);
                        END IF;


                        EXIT;
                     END IF;
                  END LOOP;

               END IF;

            END LOOP;

         END LOOP;
      END IF;
   END Release_Resource;



   PROCEDURE Allocate_Resource (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table) IS


      Current_Capacity : Natural             := 0;
      A_Item           : Time_Unit_Event_Ptr;


   BEGIN

      IF (Si.Number_Of_Resources>0)
            THEN
         Current_Capacity:=A_Tcb.Tsk.Capacity-A_Tcb.
            Rest_Of_Capacity+1;

         FOR Index1 IN 0 .. Si.Number_Of_Resources -1 LOOP


            -- For each resource, check if the task request the resource
            -- and if the resource is free
            -- otherwise, block the task
            --
            FOR Index2 IN 0..Si.Shared_Resources(Index1).Shared.
                  Task_Tab.Nb_Entries-1
                  LOOP


               IF (Current_Capacity=Si.Shared_Resources(Index1).
                     Shared.Task_Tab.Entries(Index2).Data.
                     Task_Begin)
                     AND (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.
                     Name)
                     THEN

                  Si.Shared_Resources(Index1).Shared.State:=
                     Si.Shared_Resources(Index1).Shared.State-1;

                  Si.Shared_Resources(Index1).Allocated_By(
                     Si.Shared_Resources(Index1).Nb_Allocated):=
                     A_Tcb.Tsk.Name;
                  Si.Shared_Resources(Index1).Nb_Allocated:=
                     Si.Shared_Resources(Index1).Nb_Allocated+1;

                  IF   Event_To_Generate(Allocate_Resource) THEN
                     A_Item:= NEW Time_Unit_Event(Allocate_Resource);
                     A_Item.Allocate_Resource:=Si.Shared_Resources(
                        Index1).
                        Shared;
                     A_Item.Allocate_Task:=A_Tcb.Tsk;
                     Add(Result.All, A_Item, Current_Time);
                  END IF;

               END IF;

            END LOOP;


         END LOOP;
      END IF;
   END Allocate_Resource;



   PROCEDURE Check_Resource (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             : IN     Tcb_Ptr;
         Is_Ready          :    OUT Boolean;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table) IS

      Current_Capacity : Natural             := 0;
      A_Item           : Time_Unit_Event_Ptr;

   BEGIN

      Is_Ready:=True;

      IF (Si.Number_Of_Resources>0)
            THEN

         Current_Capacity:=A_Tcb.Tsk.Capacity-A_Tcb.
            Rest_Of_Capacity+1;

         FOR Index1 IN 0 .. Si.Number_Of_Resources -1 LOOP


            -- For each resource, check if the task request the resource
            -- and if the resource is free
            --
            FOR Index2 IN 0 ..Si.Shared_Resources(Index1).
                  Shared.Task_Tab.Nb_Entries-1
                  LOOP


               IF (Current_Capacity=Si.Shared_Resources(
                        Index1).
                     Shared.Task_Tab.Entries(Index2).Data.
                     Task_Begin)
                     AND (Si.Shared_Resources(Index1).Shared.
                     Task_Tab.Entries(Index2).Item=A_Tcb.Tsk.
                     Name)
                     THEN


                  IF (Si.Shared_Resources(Index1).Shared.
                        State<=
                        0)
                        THEN
                     Is_Ready:=False;
                  END IF;


                  IF  NOT Is_Ready THEN

                     IF   Event_To_Generate(Wait_For_Resource) THEN
                        A_Item:= NEW Time_Unit_Event(
                           Wait_For_Resource);
                        A_Item.Wait_Resource:=Si.
                           Shared_Resources(
                           Index1).
                           Shared;
                        A_Item.Wait_Task:=A_Tcb.Tsk;
                        Add(Result.All, A_Item,
                           Current_Time);
                     END IF;

                  END IF;


               END IF;


            END LOOP;


         END LOOP;
      END IF;
   END Check_Resource;




   PROCEDURE Set_Quantum (
         My_Scheduler : IN OUT Generic_Scheduler'Class;
         Q            : IN     Natural) IS
   BEGIN
      My_Scheduler.Quantum:=Q;
   END Set_Quantum;


   FUNCTION Get_Quantum (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN Natural IS
   BEGIN
      RETURN My_Scheduler.Quantum;
   END Get_Quantum;


   FUNCTION Get_Quantum (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN Unbounded_String IS
   BEGIN
      RETURN To_Unbounded_String(My_Scheduler.Quantum'Img);
   END Get_Quantum;


   FUNCTION Get_Quantum (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN String IS
   BEGIN
      RETURN My_Scheduler.Quantum'Img;
   END Get_Quantum;



   FUNCTION Build_Resource (
         My_Scheduler : IN     Generic_Scheduler;
         A_Resource   :        Resource_Ptr)
     RETURN Shared_Resource_Ptr IS

      New_A_Resource : Shared_Resource_Ptr;

   BEGIN

      New_A_Resource:= NEW Shared_Resource;
      New_A_Resource.Shared:=A_Resource;

      RETURN New_A_Resource;

   END Build_Resource;




   PROCEDURE Initialize (
         S : IN OUT Scheduling_Information) IS
   BEGIN
      S.Number_Of_Tasks  := 0;
      S.Number_Of_Resources := 0;
   END Initialize;


   -- If this function return "True", it means that
   -- the task "A_Tcb" can be selected because of all its
   -- predecessors have ended their activation
   --
   FUNCTION Check_Precedencies (
         Si           : IN     Scheduling_Information;
         Deps         :        Tasks_Dependencies_Ptr;
         Current_Time : IN     Natural;
         A_Tcb        :        Tcb_Ptr)
     RETURN Boolean IS


      Has_Message_To_Receive : Boolean := False;
      Wait_Message           : Boolean := False;

      Previous : Tasks_Set;
      Left     : Task_Ptr;
      Ite1     : Tasks_Iterator;

      A_Half_Dep_Ptr : Dependency_Ptr;
      A_Message      : Message_Ptr;

      My_Iterator : Tasks_Dependencies_Iterator;
      List_Ite    : Message_Scheduling_Information_Iterator;
      Msg         : Message_Scheduling_Information_Ptr;


   BEGIN


      -------------------------------------------------
      -- Check if the task has a task precedency and
      -- if they are met
      -------------------------------------------------


      IF Has_Predecessor(Deps, A_Tcb.Tsk) THEN

         Reset(Previous);
         Previous:=Get_Predecessors_List(Deps, A_Tcb.Tsk);

         Reset_Iterator(Previous, Ite1);

         FOR J IN 0 .. Get_Number_Of_Elements(Previous)-1 LOOP

            Current_Element(Previous, Left, Ite1);

            FOR I IN 0..Si.Number_Of_Tasks-1 LOOP
               IF Si.Tcbs(I).Tsk.Name = Left.Name
                     THEN

                  IF (Si.Tcbs(I).Activation<=
                        A_Tcb.Activation)
                        THEN
                     RETURN False;
                  END IF;

                  -- Same activation ... be sure the previous task
                  -- is ended
                  --
                  IF (Si.Tcbs(I).Activation=A_Tcb.Activation+1)
                        AND (Si.Tcbs(I).End_Time>Current_Time)
                        THEN
                     RETURN False;
                  END IF;

               END IF;
            END LOOP;

            Next_Element(Previous, Ite1);
         END LOOP;
      END IF;



      -------------------------------------------------
      -- Check communication dependencies first
      -- A "receiver" has to read at most one message
      -- to be activated
      -------------------------------------------------

      IF NOT Is_Empty(Deps.All) THEN


         Reset_Iterator(Deps.All, My_Iterator);
         LOOP
            Current_Element(Deps.All, A_Half_Dep_Ptr,
               My_Iterator);

            -- Is the task concerned ?
            -- Is it a communication dependency ?
            -- Is it a receiving operation ?
            --
            IF ( A_Half_Dep_Ptr.Depend_Type=
                  Communication_Dependency ) THEN
               IF ( A_Half_Dep_Ptr.
                     Communication_Dependent_Task.
                     Name =
                     A_Tcb.Tsk.Name )
                     AND  ( A_Half_Dep_Ptr.
                     Communication_Orientation  =
                     From_Object_To_Task) THEN


                  A_Message := A_Half_Dep_Ptr.
                     Communication_Dependency_Object;


                  -- Is The task waiting for a message ?
                  --
                  IF (A_Tcb.Rest_Of_Capacity = A_Tcb.Tsk.Capacity)
                        THEN
                     Wait_Message:=True;
                  END IF;

                  -- Check the sended_messages to see if this
                  -- messages is just arrived
                  --
                  IF NOT Is_Empty(Si.Sended_Messages) THEN
                     Reset_Head_Iterator(Si.Sended_Messages,
                        List_Ite);
                     LOOP
                        Current_Element (Si.Sended_Messages,
                           Msg,
                           List_Ite);

                        -- Is it the right message ?
                        -- Is the message arrived at the receiving processor ?
                        --
                        IF Wait_Message THEN
                           IF (Msg.Sended_Message.Name=A_Message.
                                 Name) THEN
                              IF  ( Current_Time >= (Msg.
                                       Send_Time+
                                       A_Message.
                                       Response_Time)) THEN
                                 Has_Message_To_Receive:=True;
                              END IF;
                           END IF;
                        END IF;

                        IF  Is_Tail_Element (Si.
                              Sended_Messages,
                              List_Ite) THEN
                           EXIT;
                        END IF;

                        Next_Element (Si.Sended_Messages,
                           List_Ite);
                     END LOOP;
                  END IF;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(Deps.All, My_Iterator);

            Next_Element(Deps.All, My_Iterator);

         END LOOP;
      END IF;


      -- If the task wait for a message which is not
      -- arrived, do not wake up it
      --
      IF Wait_Message
            AND (NOT Has_Message_To_Receive) THEN
         RETURN False;
      END IF;


      -- All dependency constraints are met
      --
      RETURN True;


   END Check_Precedencies;


   PROCEDURE Compute_Activation_Time (
         My_Scheduler : IN     Generic_Scheduler;
         Si           : IN OUT Scheduling_Information;
         Elected      : IN     Tasks_Range;
         Value        : IN OUT Natural) IS

   BEGIN
      Value:=0;
   END Compute_Activation_Time;




   PROCEDURE Put (
         M : IN     Message_Scheduling_Information_Ptr) IS
   BEGIN
      Put("Message " & To_String(M.Sended_Message.Name));
      Put_Line(" ; send_time = " & M.Send_Time'Img);
   END Put;



   PROCEDURE Put (
         M : IN     Buffer_Scheduling_Information_Ptr) IS
   BEGIN
      Put("Buffer " & To_String(M.Written_Buffer.Name));
      Put_Line(" ; current_size = " & M.Current_Size'Img);
   END Put;


   FUNCTION Produce_Running_Task_Event (
         My_Scheduler : IN     Generic_Scheduler;
         A_Task       : IN     Tcb_Ptr)
     RETURN Time_Unit_Event_Ptr IS

      A_Item : Time_Unit_Event_Ptr;

   BEGIN
      A_Item:= NEW Time_Unit_Event(Running_Task);
      A_Item.Running_Task:=A_Task.Tsk;
      RETURN A_Item;
   END Produce_Running_Task_Event;


   FUNCTION Export_Xml_Event_Write_To_Buffer (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result:= " " & An_Event.Write_Buffer.Name &
         " " & An_Event.Write_Task.Name &
         " " & An_Event.Write_Size'Img;
      RETURN Result;
   END Export_Xml_Event_Write_To_Buffer;


   FUNCTION Export_Xml_Event_Read_From_Buffer (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result:= " " & An_Event.Read_Buffer.Name &
         " " & An_Event.Read_Task.Name &
         " " & An_Event.Read_Size'Img;
      RETURN Result;
   END Export_Xml_Event_Read_From_Buffer;


   FUNCTION Export_Xml_Event_Running_Task (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result:= " " & An_Event.Running_Task.Name;
      RETURN Result;
   END Export_Xml_Event_Running_Task;


   FUNCTION Export_Xml_Event_Task_Activation (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result:= " " & An_Event.Activation_Task.Name;
      RETURN Result;
   END Export_Xml_Event_Task_Activation;


   FUNCTION Export_Xml_Event_Start_Of_Task_Capacity (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result:= " " & An_Event.Start_Task.Name;
      RETURN Result;
   END Export_Xml_Event_Start_Of_Task_Capacity;


   FUNCTION Export_Xml_Event_End_Of_Task_Capacity (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result:= " " & An_Event.End_Task.Name;
      RETURN Result;
   END Export_Xml_Event_End_Of_Task_Capacity;


   FUNCTION Export_Xml_Event_Send_Message (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result:= " " &  An_Event.Send_Message.Name &
         " " & An_Event.Send_Task.Name;
      RETURN Result;
   END Export_Xml_Event_Send_Message;


   FUNCTION Export_Xml_Event_Receive_Message (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result:= " " & An_Event.Receive_Message.Name &
         " " & An_Event.Receive_Task.Name;
      RETURN Result;
   END Export_Xml_Event_Receive_Message;


   FUNCTION Export_Xml_Event_Allocate_Resource (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result := " " & An_Event.Allocate_Resource.Name &
         " " & An_Event.Allocate_Task.Name;
      RETURN Result;
   END Export_Xml_Event_Allocate_Resource;


   FUNCTION Export_Xml_Event_Release_Resource (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result:= " " & An_Event.Release_Resource.Name &
         " " & An_Event.Release_Task.Name;
      RETURN Result;
   END Export_Xml_Event_Release_Resource;


   FUNCTION Export_Xml_Event_Wait_For_Resource (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String IS
      Result : Unbounded_String := Empty_String;
   BEGIN
      Result:= " " & An_Event.Wait_Resource.Name &
         " " & An_Event.Wait_Task.Name;
      RETURN Result;
   END Export_Xml_Event_Wait_For_Resource;

END Scheduler;


