------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Events;
USE Events;

WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;

WITH Resources;
USE Resources;
USE Resources.Task_Lists;
WITH Resource_Set;
USE Resource_Set;
USE Resource_Set.Generic_Resource_Set;

WITH Buffers;
USE Buffers;
WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;

WITH Messages;
USE Messages;
WITH Message_Set;
USE Message_Set;
USE Message_Set.Generic_Message_Set;


WITH Task_Dependencies;
USE Task_Dependencies;


WITH Ada.Finalization;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;

WITH Framework_Config;
USE Framework_Config;

WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;
WITH Ada.Numerics.Float_Random;
USE Ada.Numerics.Float_Random;

WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;

WITH Scheduling_Sequences;
USE Scheduling_Sequences;




WITH Indexed_Tables;
WITH Convert_Strings;
WITH Convert_Unbounded_Strings;
WITH Double_Util;
WITH Natural_Util;
WITH Lists;
WITH Unchecked_Deallocation;



PACKAGE Scheduler IS



   -------------------------------------------------------------
   -------------------------------------------------------------
   --   Basic type definitions
   -------------------------------------------------------------
   -------------------------------------------------------------


   -- All kind of scheduler currently supported by cheddar
   --

   TYPE Scheduler_Type IS
         (Parametric_Protocol,
          Earliest_Deadline_First_Protocol,
          Least_Laxity_First_Protocol,
          Rate_Monotonic_Protocol,
          Deadline_Monotonic_Protocol,
          Round_Robin_Protocol,
          Time_Sharing_Based_On_Wait_Time_Protocol,
          Posix_1003_Highest_Priority_First_Protocol,
          D_Over_Protocol,
          Maximum_Urgency_First_Based_On_Laxity_Protocol,
          Maximum_Urgency_First_Based_On_Deadline_Protocol,
          Time_Sharing_Based_On_Cpu_Usage_Protocol,
          No_Scheduling_Protocol,
          Hierarchical_Cyclic_Protocol,
          Hierarchical_Round_Robin_Protocol,
          Hierarchical_Fixed_Priority_Protocol,
          Hierarchical_Parametric_Protocol);


   PROCEDURE To_Scheduler IS
   NEW Convert_Strings(Scheduler_Type, No_Scheduling_Protocol);
   PROCEDURE To_Scheduler IS
   NEW Convert_Unbounded_Strings(Scheduler_Type, No_Scheduling_Protocol);



   -- A scheduler is preemptive of not ?
   --


   TYPE Preemptive_Type IS
         (Preemptive,
          Not_Preemptive);


   PROCEDURE To_Preemptive IS
   NEW Convert_Strings(Preemptive_Type, Preemptive);
   PROCEDURE To_Preemptive IS
   NEW Convert_Unbounded_Strings(Preemptive_Type, Preemptive);




   -- Constant to be used when no resource or no dependency are
   -- present in a project
   --
   No_Resource_Set       : Resources_Set;
   No_Tasks_Dependencies : Tasks_Dependencies;




   -- Table of real values : used to compute criterion for each task
   -- of a given processor (eg. task response time, task shared reousrces blocking time, ...)
   --
   PACKAGE Double_Tasks_Parameters IS NEW Indexed_Tables(Double, Task_Ptr,
      Framework_Config.Max_Tasks, 0, Double_Util.Put,
      Double_Util.Initialize,
      Put_Name);
   USE Double_Tasks_Parameters;

   SUBTYPE Response_Time_Table IS Double_Tasks_Parameters.Indexed_Table;
   SUBTYPE Response_Time_Range IS
      Double_Tasks_Parameters.Indexed_Table_Range;


   PACKAGE Natural_Tasks_Parameters IS NEW Indexed_Tables(Natural,
      Task_Ptr,
      Framework_Config.Max_Tasks, 0, Natural_Util.Put,
      Natural_Util.Initialize,
      Put_Name);
   USE Natural_Tasks_Parameters;

   SUBTYPE Blocking_Time_Table IS Natural_Tasks_Parameters.Indexed_Table;
   SUBTYPE Blocking_Time_Range IS
      Natural_Tasks_Parameters.Indexed_Table_Range;



   -------------------------------------------------------------
   -------------------------------------------------------------
   -- The type definitions below are used by schedulers to store
   -- required information on task/processor/buffers/... in order to
   -- compute the scheduling
   -------------------------------------------------------------
   -------------------------------------------------------------




   -- Information related to the tasks
   --
   TYPE Tcb IS TAGGED
      RECORD
         Tsk              : Task_Ptr;
         Wake_Up_Time     : Natural;
         Rest_Of_Capacity : Natural;
         Activation       : Natural;
         End_Time         : Natural;
         Used_Cpu         : Natural;
         Task_Seed        : State;
      END RECORD;

   TYPE Tcb_Ptr IS ACCESS ALL Tcb'Class;
   TYPE Tcb_Table IS ARRAY (Tasks_Range) OF Tcb_Ptr;


   PROCEDURE Initialize (
         A_Tcb  : IN OUT Tcb;
         A_Task :        Task_Ptr);



   -- Information related to the access on shared resources
   --
   TYPE Allocation_Table IS ARRAY (Natural RANGE 0 .. Max_Tasks_For_A_Resource) OF Unbounded_String;

   TYPE Shared_Resource IS TAGGED
      RECORD
         Shared       : Resource_Ptr;
         Allocated_By : Allocation_Table;
         Nb_Allocated : Natural          := 0;
      END RECORD;

   TYPE Shared_Resource_Ptr IS ACCESS ALL Shared_Resource'Class;
   TYPE Shared_Resource_Table IS ARRAY (Resources_Range) OF Shared_Resource_Ptr;


   -- Information related to messages exchanges
   -- Messages are data shared between tasks running on different processors
   --

   TYPE Message_Scheduling_Information IS
      RECORD
         Sended_Message : Message_Ptr;
         Send_Time      : Natural     := 0;
      END RECORD;

   TYPE Message_Scheduling_Information_Ptr IS ACCESS ALL Message_Scheduling_Information;


   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Message_Scheduling_Information, Message_Scheduling_Information_Ptr);

   PROCEDURE Put (
         M : IN     Message_Scheduling_Information_Ptr);


   PACKAGE Message_Scheduling_Information_List_Package IS NEW Lists(
      Message_Scheduling_Information_Ptr, Put, Free);
   USE Message_Scheduling_Information_List_Package;
   SUBTYPE Message_Scheduling_Information_List IS
      Message_Scheduling_Information_List_Package.List;
   SUBTYPE Message_Scheduling_Information_Iterator IS
      Message_Scheduling_Information_List_Package.Iterator;


   -- Information related to buffer read/write
   -- Messages are data shared between tasks running on the same processor
   --

   TYPE Buffer_Scheduling_Information IS
      RECORD
         Written_Buffer : Buffer_Ptr;
         Current_Size   : Natural    := 0;
      END RECORD;

   TYPE Buffer_Scheduling_Information_Ptr IS ACCESS ALL Buffer_Scheduling_Information;


   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Buffer_Scheduling_Information, Buffer_Scheduling_Information_Ptr);

   PROCEDURE Put (
         M : IN     Buffer_Scheduling_Information_Ptr);


   PACKAGE Buffer_Scheduling_Information_List_Package IS NEW Lists(
      Buffer_Scheduling_Information_Ptr, Put, Free);
   USE Buffer_Scheduling_Information_List_Package;
   SUBTYPE Buffer_Scheduling_Information_List IS
      Buffer_Scheduling_Information_List_Package.List;
   SUBTYPE Buffer_Scheduling_Information_Iterator IS
      Buffer_Scheduling_Information_List_Package.Iterator;


   -- Finally, the type which merge all the information declared
   -- Above (task, Resources, Buffers, Messages)
   --

   TYPE Scheduling_Information IS
      RECORD
         Simulation_Length        : Natural := 0;
         Number_Of_Processors     : Natural := 0;
         Number_Of_Address_Spaces : Natural := 0;

         Tcbs            : Tcb_Table;
         Number_Of_Tasks : Tasks_Range := 0;

         Shared_Resources    : Shared_Resource_Table;
         Number_Of_Resources : Resources_Range       := 0;

         Sended_Messages : Message_Scheduling_Information_List;
         Written_Buffers : Buffer_Scheduling_Information_List;

         With_Specific_Task_Seed : Boolean;
         Global_Seed             : State;
      END RECORD;


   PROCEDURE Initialize (
         S : IN OUT Scheduling_Information);


   -------------------------------------------------------------
   -------------------------------------------------------------
   -- Abstract scheduler declaration : main entry point
   -- for scheduling source code
   -------------------------------------------------------------
   -------------------------------------------------------------


   TYPE Generic_Scheduler IS ABSTRACT TAGGED PRIVATE;
   TYPE Scheduler_Ptr IS ACCESS ALL Generic_Scheduler'Class;


   PROCEDURE Reset (
         A_Scheduler : IN OUT Generic_Scheduler'Class);


   PROCEDURE Put (
         My_Scheduler : IN     Generic_Scheduler);


   -------------------------------------------------------------
   -------------------------------------------------------------
   -- sub-program which can be used in order to access
   -- scheduler data
   -------------------------------------------------------------
   -------------------------------------------------------------


   PROCEDURE Set_Preemptive (
         My_Scheduler : IN OUT Generic_Scheduler'Class;
         Preempt      : IN     Preemptive_Type);

   FUNCTION Get_Preemptive (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN Preemptive_Type;

   FUNCTION Get_Preemptive (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN String;



   FUNCTION Get_Name (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN Scheduler_Type;

   FUNCTION Get_Name (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN Unbounded_String;

   PROCEDURE Set_Quantum (
         My_Scheduler : IN OUT Generic_Scheduler'Class;
         Q            : IN     Natural);

   FUNCTION Get_Quantum (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN Natural;

   FUNCTION Get_Quantum (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN String;

   FUNCTION Get_Quantum (
         My_Scheduler : IN     Generic_Scheduler'Class)
     RETURN Unbounded_String;


   FUNCTION Copy (
         A_Scheduler : IN     Generic_Scheduler)
     RETURN Scheduler_Ptr IS
   ABSTRACT;




   -------------------------------------------------------------
   -------------------------------------------------------------
   -- Exception raised during feasibility test or simulation
   -------------------------------------------------------------
   -------------------------------------------------------------



   -- The processor utilization factor have to bel es than 100 percent
   --
   Processor_Utilization_Exceeded : EXCEPTION;

   -- The required computation can not be done with such scheduler
   --
   Invalid_Scheduler : EXCEPTION;

   -- Raised if during a bound blocking time computation
   -- resources with different protocols are met
   --
   Other_Resources_Protocol_Found : EXCEPTION;

   -- Raised if an invalid resource protocole is found
   --
   Invalid_Protocol : EXCEPTION;





   -------------------------------------------------------------
   -------------------------------------------------------------
   -- Feasiblity tests for a given scheduler
   -- Most of these sub-programs are declared abstract since
   -- they have to be provided by the implemented scheduler
   -------------------------------------------------------------
   -------------------------------------------------------------
   -- Compute the bound on processor utilization factor that the
   -- designed can not overflow if task deadlines must be met
   --
   PROCEDURE Bound_On_Processor_Utilization (
         My_Scheduler   : IN     Generic_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Result         : IN OUT Double;
         Msg            : IN OUT Unbounded_String) IS
   ABSTRACT;


   -- Apply processor utilization factor tests
   --
   PROCEDURE Utilization_Factor_Feasibility_Test (
         My_Scheduler   : IN     Generic_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String) IS
   ABSTRACT;


   -- Compute for each task its response time
   --
   PROCEDURE Compute_Response_Time (
         My_Scheduler   : IN     Generic_Scheduler;
         My_Tasks       : IN OUT Tasks_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Response_Time  :    OUT Response_Time_Table) IS
   ABSTRACT;


   -- Compute bound on blocking time of each task
   --
   PROCEDURE Compute_Blocking_Time (
         My_Scheduler   : IN     Generic_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String;
         Msg            : IN OUT Unbounded_String;
         Blocking_Time  :    OUT Blocking_Time_Table) IS
   ABSTRACT;


   -------------------------------------------------------------
   -------------------------------------------------------------
   --  Procedure and functions to run a scheduling simulation
   -------------------------------------------------------------
   -------------------------------------------------------------


   -- Main entry point for a simulation composed of one processor only
   --
   PROCEDURE Build_Scheduling_Sequence (
         My_Scheduler            : IN OUT Generic_Scheduler'Class;
         My_Tasks                : IN     Tasks_Set;
         My_Resources            : IN     Resources_Set;
         My_Buffers              : IN     Buffers_Set;
         My_Messages             : IN     Messages_Set;
         My_Dependencies         : IN     Tasks_Dependencies_Ptr;
         Processor_Name          : IN     Unbounded_String;
         Result                  :    OUT Scheduling_Sequence_Ptr;
         Msg                     : IN OUT Unbounded_String;
         Event_To_Generate       : IN     Time_Unit_Event_Type_Boolean_Table;
         Given_Last_Time         : IN     Natural                            := 0;
         With_Offsets            : IN     Boolean                            := True;
         With_Precedencies       : IN     Boolean                            := True;
         With_Resources          : IN     Boolean                            := True;
         With_Task_Specific_Seed : IN     Boolean                            := True;
         Global_Seed_Value       : IN     Integer                            := 0;
         Predictable_Global_Seed : IN     Boolean                            := True);



   -- Compute the next time on which the task will be ready to run (task wakeup time)
   -- May be overidden by specific scheduler (ie parametric scheduler)
   --
   PROCEDURE Compute_Activation_Time (
         My_Scheduler : IN     Generic_Scheduler;
         Si           : IN OUT Scheduling_Information;
         Elected      : IN     Tasks_Range;
         Value        : IN OUT Natural);


   -- Store the fact that the "Elected" task ran
   --
   PROCEDURE Run_Task (
         My_Scheduler      : IN OUT Generic_Scheduler'Class;
         Processor_Name    : IN     Unbounded_String;
         Si                : IN OUT Scheduling_Information;
         My_Dependencies   : IN     Tasks_Dependencies_Ptr;
         Elected           : IN     Tasks_Range;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN OUT Natural;
         Last_Time         : IN     Natural;
         With_Offsets      :        Boolean                            := True;
         With_Precedencies :        Boolean                            := True;
         With_Resources    :        Boolean                            := True;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table);


   -- At scheduling simulation starting time, check that the simulation
   -- can be done according to the selected scheduler and the task set
   --
   PROCEDURE Check_Before_Scheduling (
         My_Scheduler   : IN     Generic_Scheduler;
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS
   ABSTRACT;



   -- Initiliaze scheduler data in order to perform the scheduling simulation
   --
   PROCEDURE Specific_Scheduler_Initialization (
         My_Scheduler   : IN OUT Generic_Scheduler;
         Si             : IN OUT Scheduling_Information;
         Processor_Name : IN     Unbounded_String;
         My_Tasks       : IN OUT Tasks_Set;
         My_Resources   : IN OUT Resources_Set;
         My_Buffers     : IN OUT Buffers_Set;
         My_Messages    : IN     Messages_Set;
         Msg            : IN OUT Unbounded_String) IS
   ABSTRACT;


   PROCEDURE Scheduler_Initialization (
         My_Scheduler            : IN OUT Generic_Scheduler'Class;
         Si                      : IN OUT Scheduling_Information;
         Processor_Name          : IN     Unbounded_String;
         My_Tasks                : IN OUT Tasks_Set;
         My_Resources            : IN OUT Resources_Set;
         My_Buffers              : IN OUT Buffers_Set;
         Result                  : IN OUT Scheduling_Sequence_Ptr;
         With_Offsets            :        Boolean;
         With_Precedencies       :        Boolean;
         With_Resources          :        Boolean;
         With_Task_Specific_Seed :        Boolean                            := True;
         Global_Seed_Value       :        Integer                            := 0;
         Predictable_Global_Seed :        Boolean                            := True;
         Given_Last_Time         : IN     Natural;
         Event_To_Generate       : IN     Time_Unit_Event_Type_Boolean_Table);


   -- Choose the task to be run at a given time
   --
   PROCEDURE Do_Election (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Msg               : IN OUT Unbounded_String;
         Current_Time      : IN     Natural;
         Processor_Name    : IN     Unbounded_String;
         My_Dependencies   : IN     Tasks_Dependencies_Ptr;
         With_Offsets      : IN     Boolean;
         With_Precedencies : IN     Boolean;
         With_Resources    : IN     Boolean;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table;
         Elected           : IN OUT Tasks_Range;
         No_Task           : IN OUT Boolean) IS
   ABSTRACT;


   -- Sub-programs used to build the scheduling information
   --
   FUNCTION Build_Tcb (
         My_Scheduler : IN     Generic_Scheduler;
         A_Task       :        Task_Ptr)
     RETURN Tcb_Ptr;


   FUNCTION Build_Resource (
         My_Scheduler : IN     Generic_Scheduler;
         A_Resource   :        Resource_Ptr)
     RETURN Shared_Resource_Ptr;



   -- Perform treatements when we simulation shared resources access :
   -- allocation of the shared resource, releasing the shared resource
   -- and checking that the resource can be taken
   --
   PROCEDURE Allocate_Resource (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table);


   PROCEDURE Release_Resource (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table);


   PROCEDURE Check_Resource (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             : IN     Tcb_Ptr;
         Is_Ready          :    OUT Boolean;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table);



   -- Check that the task can be run according to its wake up time offset
   -- and according to the tasks precendencies
   --
   FUNCTION Check_Offset (
         A_Tcb   : Tcb_Ptr;
         At_Time : Natural)
     RETURN Boolean;

   FUNCTION Check_Precedencies (
         Si           : IN     Scheduling_Information;
         Deps         :        Tasks_Dependencies_Ptr;
         Current_Time : IN     Natural;
         A_Tcb        :        Tcb_Ptr)
     RETURN Boolean;




   -- Perform treatements when we simulation messages exchanges  :
   -- send a message to a task, and receive a message from a task
   --
   PROCEDURE Send_Message (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Deps              :        Tasks_Dependencies_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table);


   PROCEDURE Receive_Message (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Deps              :        Tasks_Dependencies_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table);




   -- Perform treatements when we simulation buffer access  :
   -- write bytes into a buffer of read bytes from a buffer
   --
   PROCEDURE Buffer_Read (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Deps              :        Tasks_Dependencies_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table);

   PROCEDURE Buffer_Write (
         My_Scheduler      : IN OUT Generic_Scheduler;
         Si                : IN OUT Scheduling_Information;
         Result            : IN OUT Scheduling_Sequence_Ptr;
         Current_Time      : IN     Natural;
         A_Tcb             :        Tcb_Ptr;
         Deps              :        Tasks_Dependencies_Ptr;
         Event_To_Generate : IN     Time_Unit_Event_Type_Boolean_Table);






   -------------------------------------------------------------------
   -------------------------------------------------------------------
   -- Sub program to sort any kind of data managed by schedulers
   -------------------------------------------------------------------
   -------------------------------------------------------------------


   Current_Processor_Name : Unbounded_String;

   FUNCTION Select_Cpu (
         Op1 : IN     Task_Ptr)
     RETURN Boolean;
   FUNCTION Select_Cpu (
         Op1 : IN     Resource_Ptr)
     RETURN Boolean;
   FUNCTION Select_Cpu (
         Op1 : IN     Buffer_Ptr)
     RETURN Boolean;



   -------------------------------------------------------------
   -------------------------------------------------------------
   -- I/O operations
   -------------------------------------------------------------
   -------------------------------------------------------------


   -- XML/AADL export of the scheduler properties
   --
   FUNCTION Export_Aadl_Properties (
         My_Scheduler : IN     Generic_Scheduler;
         Number_Of_Ht : IN     Natural)
     RETURN Unbounded_String;

   FUNCTION Export_Xml (
         My_Scheduler : IN     Generic_Scheduler)
     RETURN Unbounded_String;



   -- XML export of the event table
   -- These sub-programs can be overidden by specific scheduler
   -- if events must provide specific information. This information
   -- is carried on by XML attributes
   --


   -- Save scheduling information at simulation time
   --
   FUNCTION Produce_Running_Task_Event (
         My_Scheduler : IN     Generic_Scheduler;
         A_Task       : IN     Tcb_Ptr)
     RETURN Time_Unit_Event_Ptr;


   -- Export XML event table
   --
   FUNCTION Export_Xml_Event_Write_To_Buffer (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;
   FUNCTION Export_Xml_Event_Read_From_Buffer (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;
   FUNCTION Export_Xml_Event_Running_Task (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;
   FUNCTION Export_Xml_Event_Task_Activation (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;
   FUNCTION Export_Xml_Event_Start_Of_Task_Capacity (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;
   FUNCTION Export_Xml_Event_End_Of_Task_Capacity (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;
   FUNCTION Export_Xml_Event_Send_Message (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;
   FUNCTION Export_Xml_Event_Receive_Message (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;
   FUNCTION Export_Xml_Event_Allocate_Resource (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;
   FUNCTION Export_Xml_Event_Release_Resource (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;
   FUNCTION Export_Xml_Event_Wait_For_Resource (
         My_Scheduler : IN     Generic_Scheduler;
         An_Event     : IN     Time_Unit_Event_Ptr)
     RETURN Unbounded_String;


PRIVATE



   TYPE Generic_Scheduler IS ABSTRACT NEW Ada.Finalization.Controlled WITH
      RECORD
         Name          : Scheduler_Type;
         Is_Preemptive : Preemptive_Type;
         Quantum       : Natural         := 0;

         Previously_Elected : Tasks_Range := Tasks_Range'First;
      END RECORD;




END Scheduler ;


