------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Events;
USE Events;

WITH Translate;
USE Translate;

WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Text_Io;
USE Text_Io;

WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;

WITH Buffers;
USE Buffers;
USE Buffers.Buffer_Roles;
WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;


WITH Resources;
USE Resources;
WITH Resource_Set;
USE Resource_Set;
USE Resource_Set.Generic_Resource_Set;

WITH Time_Units;
USE Time_Units;
USE Time_Units.Time_Unit_List;

WITH Natural_Util;
USE Natural_Util;
USE Natural_Util.Natural_List;

WITH Queueing_System;
USE Queueing_System;
USE Queueing_System.A_Resp_Time_Consumer;



PACKAGE BODY Scheduling_Sequences IS



   PROCEDURE Initialize (
         B :    OUT Buffer_Size_Item) IS
   BEGIN
      B.Time:=0;
      B.Size:=0;
   END Initialize;


   PROCEDURE Put (
         B : IN     Buffer_Size_Item) IS
   BEGIN
      Put(B.Time'Img & "/" & B.Size'Img);
   END Put;


   PROCEDURE Initialize (
         Wt :    OUT Double) IS

   BEGIN
      Wt := 0.0;
   END Initialize;

   PROCEDURE Put (
         Wt : IN     Double) IS
   BEGIN
      Put(Wt'Img);
   END Put;




   FUNCTION Compute_Average_Buffer_Size_From_Simulation (
         Buff : IN     Buffer_Size_Table_Ptr)
     RETURN Double IS

      Sum : Double := 0.0;

   BEGIN

      FOR I IN 0..Buff.Nb_Entries-1 LOOP
         Sum:=Sum+Double(Buff.Entries(I).Size);
      END LOOP;

      RETURN Sum/Double(Buff.Nb_Entries);

   END Compute_Average_Buffer_Size_From_Simulation;



   FUNCTION Compute_Maximum_Buffer_Size_From_Simulation (
         Buff : IN     Buffer_Size_Table_Ptr)
     RETURN Natural IS

      Max : Natural := 0;
   BEGIN

      FOR I IN 0..Buff.Nb_Entries-1 LOOP
         IF Buff.Entries(I).Size > Max
               THEN
            Max:=Buff.Entries(I).Size;
         END IF;
      END LOOP;

      RETURN Max;

   END Compute_Maximum_Buffer_Size_From_Simulation;



   PROCEDURE Compute_Buffer_Size_From_Simulation (
         Sched   : IN     Scheduling_Sequence_Ptr;
         My_Buff : IN     Buffer_Ptr;
         Buff    : IN OUT Buffer_Size_Table_Ptr) IS


      Current_Size : Natural          := 0;
      A_Buff_Size  : Buffer_Size_Item;

   BEGIN

      FOR I IN 0 .. Sched.Nb_Entries-1 LOOP

         IF (Sched.Entries(I).Item.Event_Type=Read_From_Buffer)
               THEN
            IF (Sched.Entries(I).Item.Read_Buffer.Name = My_Buff.Name)
                  THEN
               A_Buff_Size.Time:=Sched.Entries(I).Data;
               Current_Size:=Current_Size-Sched.Entries(I).
                  Item.Read_Size;
               A_Buff_Size.Size:=Current_Size;
               Add(Buff.All, A_Buff_Size);
            END IF;
         END IF;

         IF (Sched.Entries(I).Item.Event_Type=Write_To_Buffer)
               THEN
            IF (Sched.Entries(I).Item.Write_Buffer.Name = My_Buff.Name)
                  THEN
               A_Buff_Size.Time:=Sched.Entries(I).Data;
               Current_Size:=Current_Size+Sched.Entries(I).
                  Item.Write_Size;
               A_Buff_Size.Size:=Current_Size;
               Add(Buff.All, A_Buff_Size);
            END IF;
         END IF;

      END LOOP;
   END Compute_Buffer_Size_From_Simulation;




   FUNCTION Compute_Maximum_Waiting_Time_From_Simulation (
         Buff                    : IN     Buffer_Size_Table_Ptr;
         Average_Consumer_Period : IN     Double)
     RETURN Natural IS

      Max : Natural := 0;
   BEGIN

      FOR I IN 0..Buff.Nb_Entries-1 LOOP
         IF Buff.Entries(I).Size > Max
               THEN
            Max:=Buff.Entries(I).Size;
         END IF;
      END LOOP;

      RETURN (Max*Natural(Average_Consumer_Period));

   END Compute_Maximum_Waiting_Time_From_Simulation;


   FUNCTION Compute_Average_Waiting_Time_From_Simulation (
         Buff                    : IN     Buffer_Size_Table_Ptr;
         Average_Consumer_Period : IN     Double)
     RETURN Double IS


      Sum : Double := 0.0;

   BEGIN

      FOR I IN 0..Buff.Nb_Entries-1 LOOP
         Sum:=Sum+Double(Buff.Entries(I).Size);
      END LOOP;

      RETURN ( (Sum/Double(Buff.Nb_Entries))*Average_Consumer_Period );

   END Compute_Average_Waiting_Time_From_Simulation;





   PROCEDURE Cyclic_Response_Time_By_Simulation (
         My_Task : IN     Task_Ptr;
         Sched   : IN     Scheduling_Sequence_Ptr;
         Msg     : IN OUT Unbounded_String;
         Max     : IN OUT Natural;
         Min     : IN OUT Natural;
         Average : IN OUT Double) IS


      Task_Occurence_Number        : Task_Occurence_Range := 0;
      Awoken_Task_Occurence_Number : Task_Occurence_Range := 0;
      Running_Awoken_Task_Index    : Task_Occurence_Range := 0;
      Next_My_Task_Occurence_Awake : Task_Occurence_Range := 0;

      A_Running_Task : Task_Ptr;
      Effective_End  : Natural  := 0;

      Current_Date            : Natural                  := 0;
      Number_Of_Response_Time : Natural;
      My_Task_Occurence_Table : Task_Occurence_Table_Ptr;






   BEGIN

      -- initialize all occurences
      -- dates are taken from scheduling table

      My_Task_Occurence_Table:= NEW Task_Occurence_Table;
      Min:=Natural'Last;
      Max:=Natural'First;
      Average:=0.0;
      Number_Of_Response_Time:=0;
      Msg:=Empty_String;


      FOR J IN 0..Sched.Nb_Entries-1 LOOP

         IF Sched.Entries(J).Item.Event_Type =
               Task_Activation THEN
            IF Sched.Entries(J).Item.Activation_Task.Name =
                  My_Task.Name THEN
               Initialize_Cyclic(My_Task_Occurence_Table.Entries(
                     Task_Occurence_Number),
                  My_Task,
                  Sched.Entries(J).Data ) ;
               Task_Occurence_Number := Task_Occurence_Number + 1;
               My_Task_Occurence_Table.Nb_Entries:=
                  My_Task_Occurence_Table.Nb_Entries+1;
            END IF;
         END IF;
      END LOOP ;


      FOR T IN 0 .. Sched.Nb_Entries-1 LOOP

         -- the current date
         --
         Current_Date := Sched.Entries(T).Data ;

         -- the current task running is the running occurence
         --
         CASE Sched.Entries(T).Item.Event_Type IS

            WHEN Running_Task =>
               A_Running_Task := Sched.Entries(T).Item.
                  Running_Task;


               -- New occurence of My_task
               --
               IF ( Current_Date = My_Task_Occurence_Table.Entries(
                        Next_My_Task_Occurence_Awake).Arrival_Date ) THEN

                  -- one more task is awoken
                  --
                  Awoken_Task_Occurence_Number :=
                     Awoken_Task_Occurence_Number + 1 ;

                  -- activate the occurence
                  --
                  My_Task_Occurence_Table.Entries(
                     Next_My_Task_Occurence_Awake).
                     Is_Active := True ;

                  -- if there is not active occurence, run the new occurence
                  --
                  IF ( Awoken_Task_Occurence_Number = 0 ) THEN
                     Running_Awoken_Task_Index :=
                        Next_My_Task_Occurence_Awake ;
                  END IF ;

                  -- the next occurence to awake
                  --
                  Next_My_Task_Occurence_Awake :=
                     Next_My_Task_Occurence_Awake + 1 ;

               END IF ;

               -- is there any activated occurences
               --
               IF ( Awoken_Task_Occurence_Number > 0 ) THEN

                  -- increment response time for all active occurences
                  --
                  FOR I_Occurence IN 0 ..My_Task_Occurence_Table.
                        Nb_Entries-1  LOOP
                     IF ( My_Task_Occurence_Table.Entries(I_Occurence).
                           Is_Active )
                           THEN
                        My_Task_Occurence_Table.Entries(I_Occurence).
                           Response_Time :=
                           My_Task_Occurence_Table.Entries(I_Occurence).
                           Response_Time + 1 ;
                     END IF ;
                  END LOOP ;

                  IF ( A_Running_Task.Name = My_Task.Name ) THEN

                     My_Task_Occurence_Table.Entries(
                        Running_Awoken_Task_Index).
                        Residual_Capacity :=
                        My_Task_Occurence_Table.Entries(
                        Running_Awoken_Task_Index).
                        Residual_Capacity - 1 ;

                     -- is The Current Occurence finished
                     --
                     IF (My_Task_Occurence_Table.Entries(
                              Running_Awoken_Task_Index).
                           Residual_Capacity = 0) THEN
                        Effective_End := Current_Date + 1 ;

                        -- Check deadline missed
                        --
                        IF ( My_Task_Occurence_Table.Entries(
                                 Running_Awoken_Task_Index).
                              Response_Time
                              > My_Task.Deadline ) THEN
                           Msg := Msg
                              & Lb_Comma & Lb_Check_Deadline(
                              Current_Language)
                              & My_Task_Occurence_Table.Entries(
                              Running_Awoken_Task_Index).
                              Absolute_Deadline'Img
                              & Lb_Completion_Time(Current_Language)
                              & Effective_End'Img
                              & To_Unbounded_String(")");

                        END IF ;

                        -- Update worst/best/average case response time
                        --
                        IF (My_Task_Occurence_Table.Entries(
                                 Running_Awoken_Task_Index).
                              Response_Time
                              < Min) THEN
                           Min :=
                              My_Task_Occurence_Table.Entries(
                              Running_Awoken_Task_Index).Response_Time ;
                        END IF ;
                        IF (My_Task_Occurence_Table.Entries(
                                 Running_Awoken_Task_Index).
                              Response_Time
                              > Max) THEN
                           Max :=
                              My_Task_Occurence_Table.Entries(
                              Running_Awoken_Task_Index).Response_Time ;
                        END IF ;
                        Average:=Average+Double(My_Task_Occurence_Table.
                           Entries(
                              Running_Awoken_Task_Index).Response_Time);
                        Number_Of_Response_Time:=
                           Number_Of_Response_Time+
                           1;


                        -- desactivate the current occurence
                        --
                        My_Task_Occurence_Table.Entries(
                           Running_Awoken_Task_Index).
                           Is_Active := False ;

                        -- give hand to the next awoken occurence
                        --
                        Running_Awoken_Task_Index :=
                           Running_Awoken_Task_Index + 1 ;

                        -- one less Awoken Occurence
                        --
                        Awoken_Task_Occurence_Number :=
                           Awoken_Task_Occurence_Number - 1 ;

                     END IF ;

                  END IF ;

               END IF ;

            WHEN Task_Activation =>
               A_Running_Task := Sched.Entries(T).
                  Item.Activation_Task;

            WHEN OTHERS  =>
               NULL;


         END CASE;


      END LOOP ;

      IF Number_Of_Response_Time>0 THEN
         Average:=Average/Double(Number_Of_Response_Time);
      END IF;
      IF Min=Natural'Last THEN
         Min:=0;
      END IF;


   END Cyclic_Response_Time_By_Simulation;


   PROCEDURE Acyclic_Response_Time_By_Simulation (
         My_Task : IN     Aperiodic_Task_Ptr;
         Sched   : IN     Scheduling_Sequence_Ptr;
         Msg     : IN OUT Unbounded_String;
         Max     : IN OUT Natural;
         Min     : IN OUT Natural;
         Average : IN OUT Double) IS

      My_Task_Occurence : Task_Occurence;
      A_Running_Task    : Task_Ptr;
      Effective_End     : Natural        := 0;

      Current_Date : Natural := 0;

   BEGIN

      Initialize_Acyclic(My_Task_Occurence , My_Task);
      Msg:=To_Unbounded_String ("");


      FOR T IN 0 .. Sched.Nb_Entries-1 LOOP
         -- the current date
         --
         Current_Date := Sched.Entries(T).Data ;

         -- the current task running is the running occurence
         --
         CASE Sched.Entries(T).Item.Event_Type IS
            WHEN Running_Task =>
               A_Running_Task := Sched.Entries(T).
                  Item.Running_Task;

               -- if the occurence of My_task starts
               --
               IF ( Current_Date = My_Task_Occurence.Arrival_Date ) THEN
                  My_Task_Occurence.Is_Active := True ;
               END IF ;

               IF ( A_Running_Task.Name = My_Task.Name ) THEN

                  My_Task_Occurence.Residual_Capacity :=
                     My_Task_Occurence.
                     Residual_Capacity - 1 ;

                  -- is The Current Occurence finished
                  --
                  IF (My_Task_Occurence.Residual_Capacity = 0) THEN

                     My_Task_Occurence.Is_Active := False ;

                     Effective_End := Current_Date + 1 ;

                     -- Check deadline missed
                     --
                     IF ( Effective_End > My_Task_Occurence.
                           Absolute_Deadline)
                           THEN
                        Msg := Msg
                           & Lb_Comma & Lb_Check_Deadline(
                           Current_Language)
                           & My_Task_Occurence.Absolute_Deadline'Img
                           & Lb_Completion_Time(Current_Language)
                           & Effective_End'Img
                           & To_Unbounded_String(")");
                     END IF ;

                  END IF;

               END IF;


               -- increment response time if the occurence is Active
               --
               IF ( My_Task_Occurence.Is_Active ) THEN
                  My_Task_Occurence.Response_Time := My_Task_Occurence.
                     Response_Time + 1 ;
               END IF ;
            -- if the current task running is the occurence of My_Task
            --
            WHEN Task_Activation =>
               A_Running_Task := Sched.Entries(T).
                  Item.Activation_Task;
            WHEN OTHERS =>
               NULL;
         END CASE;

      END LOOP;

      IF Effective_End>0
            THEN
         Average:=Double(Effective_End-My_Task_Occurence.Arrival_Date) ;
         Min:=Effective_End-My_Task_Occurence.Arrival_Date ;
         Max:=Effective_End-My_Task_Occurence.Arrival_Date ;
      ELSE
         Average:=0.0;
         Min:=0;
         Max:=0;
      END IF;

   END Acyclic_Response_Time_By_Simulation;


   PROCEDURE Compute_Response_Time_Distribution (
         My_Task : IN     Task_Ptr;
         Sched   : IN     Scheduling_Sequence_Ptr;
         Result  : IN OUT Density_Table) IS


      Awoken_Task_Occurence_Number : Task_Occurence_Range := 0;
      Task_Occurence_Number        : Task_Occurence_Range := 0;
      Running_Awoken_Task_Index    : Task_Occurence_Range := 0;
      Next_My_Task_Occurence_Awake : Task_Occurence_Range := 0;
      A_Running_Task               : Task_Ptr;
      Current_Date                 : Natural              := 0;
      Found                        : Boolean              := False;
      A_Density                    : Density_Item;
      Number_Of_Response_Time      : Double;

      My_Task_Occurence_Table : Task_Occurence_Table_Ptr;






   BEGIN


      My_Task_Occurence_Table:= NEW Task_Occurence_Table;


      -- initialize all occurences
      IF My_Task.Task_Type=Aperiodic_Type THEN

         Initialize_Acyclic(My_Task_Occurence_Table.Entries(
               Task_Occurence_Number),
            Aperiodic_Task_Ptr(My_Task) ) ;
         Task_Occurence_Number := Task_Occurence_Number + 1;

      ELSE


         -- initialize all occurences
         -- dates are taken from scheduling table
         --
         FOR J IN 0..Sched.Nb_Entries-1 LOOP

            IF Sched.Entries(J).Item.Event_Type =
                  Task_Activation THEN
               IF Sched.Entries(J).Item.Activation_Task.Name =
                     My_Task.Name THEN
                  Initialize_Cyclic(My_Task_Occurence_Table.Entries(
                        Task_Occurence_Number),
                     My_Task,
                     Sched.Entries(J).Data ) ;
                  Task_Occurence_Number := Task_Occurence_Number + 1;
                  My_Task_Occurence_Table.Nb_Entries:=
                     My_Task_Occurence_Table.Nb_Entries+1;
               END IF;
            END IF;
         END LOOP ;

      END IF;





      FOR T IN 0 .. Sched.Nb_Entries-1 LOOP

         IF Sched.Entries(T).Item.Event_Type = Running_Task THEN

            -- the current date
            Current_Date := Sched.Entries(T).Data ;

            -- the current task running is the running occurence
            A_Running_Task := Sched.Entries(T).Item.
               Running_Task ;
            -- New occurence of My_task
            IF ( Current_Date = My_Task_Occurence_Table.Entries(
                     Next_My_Task_Occurence_Awake).Arrival_Date ) THEN

               -- one more task is awoken
               Awoken_Task_Occurence_Number :=
                  Awoken_Task_Occurence_Number + 1 ;

               -- activate the occurence
               My_Task_Occurence_Table.Entries(
                  Next_My_Task_Occurence_Awake).
                  Is_Active := True ;

               -- if there is not active occurence, run the new occurence
               IF ( Awoken_Task_Occurence_Number = 0 ) THEN
                  Running_Awoken_Task_Index :=
                     Next_My_Task_Occurence_Awake ;
               END IF ;

               -- the next occurence to awake
               Next_My_Task_Occurence_Awake :=
                  Next_My_Task_Occurence_Awake + 1 ;

            END IF ;

            -- is there any activated occurences
            IF ( Awoken_Task_Occurence_Number > 0 ) THEN

               -- increment response time for all active occurences
               --
               FOR I_Occurence IN 0 .. My_Task_Occurence_Table.
                     Nb_Entries-
                     1 LOOP
                  IF ( My_Task_Occurence_Table.Entries(I_Occurence).
                        Is_Active ) THEN
                     My_Task_Occurence_Table.Entries(I_Occurence).
                        Response_Time :=
                        My_Task_Occurence_Table.Entries(I_Occurence).
                        Response_Time + 1 ;
                  END IF ;
               END LOOP ;


               IF ( A_Running_Task.Name = My_Task.Name ) THEN

                  My_Task_Occurence_Table.Entries(
                     Running_Awoken_Task_Index).
                     Residual_Capacity :=
                     My_Task_Occurence_Table.Entries(
                     Running_Awoken_Task_Index).
                     Residual_Capacity
                     - 1 ;

                  -- is The Current Occurence finished
                  IF (My_Task_Occurence_Table.Entries(
                           Running_Awoken_Task_Index).
                        Residual_Capacity = 0) THEN


                     -- We have a response time. We look for the
                     --   density table its response time is already
                     -- known. If so, increment the frequency, if
                     -- not, add it to the table
                     --
                     Found:=False;
                     FOR K IN 0..Result.Nb_Entries-1 LOOP
                        IF Result.Entries(K).Response_Time=
                              My_Task_Occurence_Table.Entries(
                              Running_Awoken_Task_Index).Response_Time
                              THEN
                           Result.Entries(K).Probability:=Result.
                              Entries(
                              K).Probability+1.0;
                           Found:=True;
                        END IF;
                     END LOOP;
                     IF NOT Found THEN
                        A_Density.Response_Time:=
                           My_Task_Occurence_Table.Entries(
                           Running_Awoken_Task_Index).Response_Time;
                        A_Density.Probability:=1.0;
                        Add(Result, A_Density);
                     END IF;

                     -- desactivate the current occurence
                     My_Task_Occurence_Table.Entries(
                        Running_Awoken_Task_Index).
                        Is_Active := False ;

                     -- give hand to the next awoken occurence
                     Running_Awoken_Task_Index :=
                        Running_Awoken_Task_Index + 1 ;

                     -- one less Awoken Occurence
                     Awoken_Task_Occurence_Number :=
                        Awoken_Task_Occurence_Number - 1 ;

                  END IF ;

               END IF ;

            END IF ;

         END IF;

      END LOOP ;


      -- Now, compute density : find number of activation done and
      -- divide all entries of the table by this number
      --
      Number_Of_Response_Time:=0.0;
      FOR K IN 0..Result.Nb_Entries-1 LOOP
         Number_Of_Response_Time:=Number_Of_Response_Time+Result.
            Entries(
            K).Probability;
      END LOOP;
      FOR K IN 0..Result.Nb_Entries-1 LOOP
         Result.Entries(K).Probability:=Result.Entries(K).Probability/
            Number_Of_Response_Time;
      END LOOP;


   END Compute_Response_Time_Distribution ;


   PROCEDURE Initialize (
         D : IN OUT Density_Item) IS
   BEGIN
      D.Response_Time:=0;
      D.Probability:=0.0;
   END Initialize;


   PROCEDURE Put (
         D : IN     Density_Item) IS
   BEGIN
      Put(D.Response_Time'Img & " " & D.Probability'Img);
   END Put;



   PROCEDURE Compute_Response_Time_Distribution (
         My_Tasks       : IN     Tasks_Set;
         Sched          : IN     Scheduling_Sequence_Ptr;
         Processor_Name :        Unbounded_String;
         Result         : IN OUT Densities_Table) IS


      A_Task          : Task_Ptr;
      My_Iterator     : Tasks_Iterator;
      Working_Density : Density_Table;

   BEGIN

      Initialize(Result);

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF (Processor_Name=A_Task.Cpu_Name)
               THEN
            Initialize(Working_Density);
            Compute_Response_Time_Distribution(A_Task,
               Sched,
               Working_Density);
            Add(Result, A_Task, Working_Density);

         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;

   END Compute_Response_Time_Distribution;





   PROCEDURE Response_Time_By_Simulation (
         My_Task :        Task_Ptr;
         Sched   : IN     Scheduling_Sequence_Ptr;
         Msg     : IN OUT Unbounded_String;
         Max     : IN OUT Natural;
         Min     : IN OUT Natural;
         Average : IN OUT Double) IS

   BEGIN

      CASE My_Task.Task_Type IS
         WHEN Aperiodic_Type =>
            Acyclic_Response_Time_By_Simulation(
               Aperiodic_Task_Ptr(
                  My_Task), Sched, Msg, Max, Min, Average);
         WHEN OTHERS =>
            Cyclic_Response_Time_By_Simulation(My_Task,
               Sched, Msg, Max, Min, Average);
      END CASE;
   END Response_Time_By_Simulation;






   PROCEDURE All_Response_Time_By_Simulation (
         My_Task : IN     Task_Ptr;
         Sched   : IN     Scheduling_Sequence_Ptr;
         Result  : IN OUT Task_Occurence_Table_Ptr) IS


      Result_Number                : Task_Occurence_Range     := 0;
      Task_Occurence_Number        : Task_Occurence_Range     := 0;
      Awoken_Task_Occurence_Number : Task_Occurence_Range     := 0;
      Running_Awoken_Task_Index    : Task_Occurence_Range     := 0;
      Next_My_Task_Occurence_Awake : Task_Occurence_Range     := 0;
      A_Running_Task               : Task_Ptr;
      Current_Date                 : Natural                  := 0;
      My_Task_Occurence_Table      : Task_Occurence_Table_Ptr;


   BEGIN

      -- initialize all occurences
      -- dates are taken from scheduling table
      --
      My_Task_Occurence_Table:= NEW Task_Occurence_Table;
      FOR J IN 0..Sched.Nb_Entries-1 LOOP

         IF Sched.Entries(J).Item.Event_Type =
               Task_Activation THEN
            IF Sched.Entries(J).Item.Activation_Task.Name =
                  My_Task.Name THEN
               Initialize_Cyclic(My_Task_Occurence_Table.Entries(
                     Task_Occurence_Number),
                  My_Task,
                  Sched.Entries(J).Data ) ;
               Task_Occurence_Number := Task_Occurence_Number + 1;
            END IF;
         END IF;
      END LOOP ;

      FOR T IN 0 .. Sched.Nb_Entries-1 LOOP

         -- the current date
         Current_Date := Sched.Entries(T).Data ;

         -- the current task running is the running occurence
         CASE Sched.Entries(T).Item.Event_Type IS

            WHEN Running_Task =>
               A_Running_Task := Sched.Entries(T).Item.
                  Running_Task;
               --               Put();

               -- New occurence of My_task
               IF ( Current_Date = My_Task_Occurence_Table.Entries(
                        Next_My_Task_Occurence_Awake).Arrival_Date ) THEN

                  -- one more task is awoken
                  Awoken_Task_Occurence_Number :=
                     Awoken_Task_Occurence_Number + 1 ;

                  -- activate the occurence
                  My_Task_Occurence_Table.Entries(
                     Next_My_Task_Occurence_Awake).
                     Is_Active := True ;

                  -- if there is not active occurence, run the new occurence
                  IF ( Awoken_Task_Occurence_Number = 0 ) THEN
                     Running_Awoken_Task_Index :=
                        Next_My_Task_Occurence_Awake ;
                  END IF ;

                  -- the next occurence to awake
                  Next_My_Task_Occurence_Awake :=
                     Next_My_Task_Occurence_Awake + 1 ;

               END IF ;

               -- is there any activated occurences
               IF ( Awoken_Task_Occurence_Number > 0 ) THEN

                  -- increment response time for all active occurences
                  -- for I_Occurence in Running_Awoken_Task_Index ..
                  -- (Running_Awoken_Task_Index +
                  -- Awoken_Task_Occurence_Number - 1) loop
                  FOR I_Occurence IN Task_Occurence_Range  LOOP
                     IF ( My_Task_Occurence_Table.Entries(I_Occurence).
                           Is_Active )
                           THEN
                        My_Task_Occurence_Table.Entries(I_Occurence).
                           Response_Time :=
                           My_Task_Occurence_Table.Entries(I_Occurence).
                           Response_Time + 1 ;
                     END IF ;
                  END LOOP ;

                  IF ( A_Running_Task.Name = My_Task.Name ) THEN

                     My_Task_Occurence_Table.Entries(
                        Running_Awoken_Task_Index).
                        Residual_Capacity :=
                        My_Task_Occurence_Table.Entries(
                        Running_Awoken_Task_Index).
                        Residual_Capacity - 1 ;

                     -- is The Current Occurence finished
                     IF (My_Task_Occurence_Table.Entries(
                              Running_Awoken_Task_Index).
                           Residual_Capacity = 0) THEN

                        -- store the response time
                        Result.Nb_Entries:=Result.Nb_Entries+1;
                        Result.Entries(Result_Number).Response_Time:=
                           My_Task_Occurence_Table.Entries(
                           Running_Awoken_Task_Index).
                           Response_Time;
                        Result_Number:=Result_Number+1;

                        -- desactivate the current occurence
                        My_Task_Occurence_Table.Entries(
                           Running_Awoken_Task_Index).
                           Is_Active := False ;

                        -- give hand to the next awoken occurence
                        Running_Awoken_Task_Index :=
                           Running_Awoken_Task_Index + 1 ;

                        -- one less Awoken Occurence
                        --
                        Awoken_Task_Occurence_Number :=
                           Awoken_Task_Occurence_Number - 1 ;

                     END IF ;

                  END IF ;

               END IF ;

            WHEN Task_Activation =>
               A_Running_Task := Sched.Entries(T).
                  Item.Activation_Task;

            WHEN OTHERS  =>
               NULL;
         END CASE;

      END LOOP ;


      Free(My_Task_Occurence_Table);

   END All_Response_Time_By_Simulation;



   PROCEDURE Blocking_Time_From_Simulation (
         My_Task      : IN     Task_Ptr;
         My_Resources : IN     Resources_Set;
         Sched        : IN     Scheduling_Sequence_Ptr;
         Max          : IN OUT Natural;
         Min          : IN OUT Natural;
         Average      : IN OUT Double) IS


      A_Resource   : Resource_Ptr;
      Resource_Ite : Resources_Iterator;

      Running_Task : Task_Ptr;

      Event_List : Time_Unit_List.List;
      An_Event   : Time_Unit_Event_Ptr;
      Wait_Event : Time_Unit_Event_Ptr;
      Event_Ite  : Time_Unit_List.Iterator;

      Date_List : Natural_List.List;
      A_Date    : Natural_Ptr;
      Wait_Date : Natural_Ptr;
      Date_Ite  : Natural_List.Iterator;

      Nb_Blocking_Time : Natural;

   BEGIN

      Max :=0;
      Min:=Natural'Last;
      Average:=0.0;
      Nb_Blocking_Time:=0;


      FOR T IN 0 .. Sched.Nb_Entries-1 LOOP


         -- Store events on shared resources
         --
         CASE Sched.Entries(T).Item.Event_Type IS
            WHEN Wait_For_Resource =>
               Running_Task := Sched.Entries(T).
                  Item.Wait_Task;
               IF Running_Task.Name=My_Task.Name THEN
                  Add(Event_List, Sched.Entries(T).Item);

                  -- the current date
                  --
                  A_Date := NEW Natural;
                  A_Date.All:=Sched.Entries(T).Data;
                  Add(Date_List, A_Date);
               END IF;

            WHEN Allocate_Resource =>
               Running_Task := Sched.Entries(T).
                  Item.Allocate_Task;
               IF Running_Task.Name=My_Task.Name THEN
                  Add(Event_List, Sched.Entries(T).Item);

                  -- the current date
                  --
                  A_Date := NEW Natural;
                  A_Date.All:=Sched.Entries(T).Data;
                  Add(Date_List, A_Date);
               END IF;
            WHEN OTHERS =>
               NULL;
         END CASE;
      END LOOP;


      -- compute maximum, minimum and average blocking time
      --
      IF NOT Is_Empty(Event_List) THEN

         Reset_Iterator(My_Resources, Resource_Ite);
         LOOP
            Current_Element(My_Resources, A_Resource, Resource_Ite);


            -- for each resource allocation, find the corresponding
            -- wait resource event
            --

            Reset_Tail_Iterator(Event_List, Event_Ite);
            Reset_Tail_Iterator(Date_List, Date_Ite);

            LOOP
               Current_Element(Event_List, An_Event, Event_Ite);

               Current_Element(Date_List, A_Date, Date_Ite);

               -- first, find a wait event for the current resource
               --
               IF (An_Event.Event_Type = Wait_For_Resource)
                     AND (Wait_Event = NULL) THEN
                  IF (An_Event.Wait_Resource.Name = A_Resource.Name)
                        THEN
                     Wait_Event:= An_Event;
                     Wait_Date:=A_Date;
                  END IF;
               END IF;
               IF (An_Event.Event_Type = Allocate_Resource)
                     AND (Wait_Event /= NULL) THEN
                  Max :=Natural'Max(Max, (A_Date.All-Wait_Date.All));
                  Min:=Natural'Min(Min, (A_Date.All-Wait_Date.All));
                  Average:=Average+Double(A_Date.All-Wait_Date.All);
                  Nb_Blocking_Time:=Nb_Blocking_Time+1;
                  Wait_Event:= NULL;
               END IF;

               EXIT WHEN Is_Head_Element(Event_List, Event_Ite);

               Previous_Element(Event_List, Event_Ite);
               Previous_Element(Date_List, Date_Ite);
            END LOOP;


            EXIT WHEN Is_Last_Element(My_Resources, Resource_Ite);

            Next_Element(My_Resources, Resource_Ite);
         END LOOP;
      END IF;



      IF Min=Natural'Last THEN
         Min:=0;
      END IF;
      IF Nb_Blocking_Time/=0 THEN
         Average:=Average/Double(Nb_Blocking_Time);
      END IF;


   END Blocking_Time_From_Simulation;




   PROCEDURE Initialize (
         D : IN OUT Deadlock_Item) IS
   BEGIN
      D.Time:=0;
      D.Task_Name := Empty_String;
      D.Resource_Name := Empty_String;
   END Initialize;


   PROCEDURE Put (
         D : IN     Deadlock_Item) IS
   BEGIN
      Put("Time : " & D.Time'Img & " ; ");
      IF D.Task_Name /= Empty_String
            THEN
         Put(To_String(D.Task_Name & " "));
      END IF;
      Put(" ; ");
      IF D.Resource_Name /= Empty_String
            THEN
         Put(To_String(D.Resource_Name & " "));
      END IF;
      New_Line;
   END Put;

   PROCEDURE Put (
         D : IN     Deadlock_Item_Ptr) IS
   BEGIN
      Put(D.All);
   END Put;

   PROCEDURE Initialize (
         D : IN OUT Priority_Inversion_Item) IS
   BEGIN
      D.Start_Time  :=0;
      D.End_Time  :=0;
      D.Task_Name     :=Empty_String;
      D.Resource_Name :=Empty_String;
   END Initialize;


   PROCEDURE Put (
         D : IN     Priority_Inversion_Item) IS
   BEGIN
      Put(" Start/end time  = " & D.Start_Time'Img & D.End_Time'Img);
      Put(" ; Task/Resource name = " & To_String(D.Task_Name)
         & To_String(D.Resource_Name));
      New_Line;
   END Put;

   PROCEDURE Put (
         D : IN     Priority_Inversion_Item_Ptr) IS
   BEGIN
      Put(D.All);
   END Put;




   FUNCTION Deadlock_From_Simulation (
         My_Tasks       : IN     Tasks_Set;
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String;
         Sched          : IN     Scheduling_Sequence_Ptr)
     RETURN Deadlock_List IS

      Result : Deadlock_List;


      -- Variables required to scan the resource and task
      --
      Item          : Deadlock_Item_Ptr;
      Item_Task     : Task_Ptr;
      Item_Resource : Resource_Ptr;

      -- To store number of Tasks
      --
      Number_Of_Tasks : CONSTANT Tasks_Range := Get_Number_Of_Elements (My_Tasks);

      -- To store number of resource
      --
      Number_Of_Resources : CONSTANT Resources_Range := Get_Number_Of_Elements
        (My_Resources);

      -- Stamp to represent the tree graphic for the allocated resources and the awaited resources
      --
      Stamp_For_Allocate : ARRAY (Tasks_Range'RANGE, Resources_Range'RANGE)
        OF Integer;

      Stamp_For_Wait : ARRAY (Tasks_Range'RANGE, Resources_Range'RANGE) OF
        Integer;

      --Tables of names of Tasks
      --
      Table_Name_Tasks : ARRAY (Tasks_Range'RANGE) OF Unbounded_String;

      -- Table of names of resources	
      --
      Table_Name_Resources : ARRAY (Resources_Range'RANGE) OF Unbounded_String;

      -- Tables to store the indices of the tasks and the resources
      --
      Table_Index_Task     : ARRAY (Natural RANGE 0 .. Framework_Config.Max_Resources
        - 1) OF Tasks_Range;
      Table_Index_Resource : ARRAY (Natural RANGE 0 .. Framework_Config.Max_Resources
        - 1) OF Resources_Range;

      -- Index
      --
      Current_Time   : Natural         := 0;
      Index_Task     : Tasks_Range     := 200;
      Index_Resource : Resources_Range := 30;

      --  booleen to say if there is deadlock
      --
      Ok : Boolean;

      -- procedure being used to find the index of a task
      --
      PROCEDURE Index_Of_Tasks (
            Name  : IN OUT Unbounded_String;
            Index : IN OUT Tasks_Range) IS
      BEGIN

         FOR I IN 0..Number_Of_Tasks-1 LOOP
            IF(Table_Name_Tasks(I)= Name)
                  THEN
               Index := I;
            END IF;
         END LOOP;

      END Index_Of_Tasks;

      -- procedure being used to find the index of a resource
      --

      PROCEDURE Index_Of_Resources (
            Name  : IN OUT Unbounded_String;
            Index : IN OUT Resources_Range) IS
      BEGIN

         FOR I IN 0..Number_Of_Resources-1 LOOP
            IF(Table_Name_Resources(I) = Name)
                  THEN
               Index := I;
            END IF;
         END LOOP;

      END Index_Of_Resources;

      --  procedure being used to find if a resource is held by a task
      --

      PROCEDURE Seek_Of_Task (
            R     : IN OUT Resources_Range;
            Index : IN OUT Tasks_Range) IS
      BEGIN

         FOR I IN 0..Number_Of_Tasks-1 LOOP
            IF(Stamp_For_Allocate(I,R) = 1)
                  THEN
               Index := I;
            END IF;
         END LOOP;

      END Seek_Of_Task;

      --  procedure being used to find if a task awaits a resource
      --

      PROCEDURE Seek_Of_Resource (
            T     : IN OUT Tasks_Range;
            Index : IN OUT Resources_Range) IS
      BEGIN

         FOR I IN 0..Number_Of_Resources-1 LOOP
            IF(Stamp_For_Wait(T,I) = 1)
                  THEN
               Index := I;
            END IF;
         END LOOP;

      END Seek_Of_Resource;



      PROCEDURE Perform_Analysis_Of_Resources IS

         Item2    : Deadlock_Item_Ptr;
         List_Ite : Deadlock_Iterator;
         Found    : Boolean           := False;

      BEGIN
         -- Initiazation of stamps
         --

         FOR I IN 0..Number_Of_Tasks-1 LOOP
            FOR J IN 0..Number_Of_Resources-1 LOOP
               Stamp_For_Allocate(I,J) := 0;
               Stamp_For_Wait(I,J) := 0;
            END LOOP;
         END LOOP;

         FOR I IN 0..Sched.Entries(Sched.Nb_Entries-1).Data LOOP
            FOR J IN 0..Sched.Nb_Entries-1 LOOP

               -- construction of the tree
               --
               IF (Sched.Entries(J).Data = I)
                     THEN
                  IF(Sched.Entries(J).Item.Event_Type =
                        Allocate_Resource)
                        THEN
                     Item_Task := Sched.Entries(J).Item.Allocate_Task;
                     Item_Resource := Sched.Entries(J).
                        Item.Allocate_Resource;
                     Index_Of_Tasks(Item_Task.Name, Index_Task);
                     Index_Of_Resources(Item_Resource.Name, Index_Resource);
                     Stamp_For_Allocate(Index_Task,Index_Resource) := 1;
                     Stamp_For_Wait(Index_Task,Index_Resource) := 0;
                  END IF;

                  IF(Sched.Entries(J).Item.Event_Type =
                        Wait_For_Resource)
                        THEN
                     Item_Task := Sched.Entries(J).Item.Wait_Task;
                     Item_Resource := Sched.Entries(J).Item.Wait_Resource;
                     Index_Of_Tasks(Item_Task.Name,Index_Task);
                     Index_Of_Resources(Item_Resource.Name, Index_Resource);
                     Stamp_For_Wait(Index_Task,Index_Resource) := 1;
                  END IF;

                  IF(Sched.Entries(J).Item.Event_Type =
                        Release_Resource)
                        THEN
                     Item_Task := Sched.Entries(J).Item.Release_Task;
                     Item_Resource := Sched.Entries(J).
                        Item.Release_Resource;
                     Index_Of_Tasks(Item_Task.Name,Index_Task);
                     Index_Of_Resources(Item_Resource.Name,Index_Resource);
                     Stamp_For_Allocate(Index_Task,Index_Resource) := 0;
                  END IF;
               END IF;

            END LOOP;

            --analyse of stamps
            --

            FOR K IN 0..Number_Of_Resources-1 LOOP
               FOR L IN 0..Framework_Config.Max_Resources-1 LOOP
                  Table_Index_Task(L)  := 0;
                  Table_Index_Resource(L)  := 0;
               END LOOP;
               Current_Time := 0;
               Index_Resource := K;
               LOOP
                  Index_Task := 30;
                  Ok := False;
                  Seek_Of_Task(Index_Resource,Index_Task);
                  IF(Index_Task /= 30)
                        THEN
                     Table_Index_Task(Current_Time) := Index_Task;
                     Table_Index_Resource(Current_Time) := Index_Resource;
                     Index_Resource := 30;
                     Seek_Of_Resource(Index_Task,Index_Resource);
                     IF(Index_Resource /= 30)
                           THEN
                        FOR P IN 0..Current_Time LOOP
                           IF( Table_Index_Resource(P) =
                                 Index_Resource)
                                 THEN
                              Ok := True;
                           END IF;
                        END LOOP;
                        IF Ok
                              THEN
                           FOR P IN 0..Current_Time LOOP
                              Item := NEW Deadlock_Item;
                              Item.Time := I;
                              Item.Task_Name := Table_Name_Tasks(
                                 Table_Index_Task(P));
                              Item.Resource_Name := Table_Name_Resources(
                                 Table_Index_Resource(P));
                              Found := False;

                              -- to check if item is already present in the list
                              --

                              IF NOT Is_Empty(Result)
                                    THEN
                                 Reset_Head_Iterator(Result, List_Ite);
                                 LOOP
                                    Current_Element (Result, Item2,
                                       List_Ite);

                                    IF((Item2.Task_Name = Item.Task_Name)
                                          AND (Item2.Resource_Name =
                                             Item.Resource_Name))
                                          THEN
                                       Found:=True;
                                       EXIT;
                                    END IF;

                                    IF  Is_Tail_Element (Result,
                                          List_Ite)
                                          THEN
                                       EXIT;
                                    END IF;
                                    Next_Element (Result, List_Ite);
                                 END LOOP;
                              END IF;

                              -- addition of item in the list
                              --	

                              IF NOT Found
                                    THEN
                                 Add(Result, Item);
                              END IF;

                           END LOOP;

                           EXIT;
                        END IF;
                     ELSE
                        EXIT;
                     END IF;
                  ELSE
                     EXIT;
                  END IF;
                  Current_Time := Current_Time + 1;
               END LOOP;
            END LOOP;

         END LOOP;

      END Perform_Analysis_Of_Resources;

   BEGIN
      -- Initialization of tables of tasks names
      --
      FOR I IN 0..Number_Of_Tasks-1 LOOP

         Get_Element_Number(My_Tasks, Item_Task,I);
         Table_Name_Tasks(I) := Item_Task.Name;

      END LOOP;

      -- Initialization of tables of resources of names
      --
      FOR I IN 0..Number_Of_Resources-1 LOOP

         Get_Element_Number(My_Resources, Item_Resource,I);
         Table_Name_Resources(I) := Item_Resource.Name;
      END LOOP;

      Perform_Analysis_Of_Resources;
      RETURN Result;

   END  Deadlock_From_Simulation;

   FUNCTION Priority_Inversion_From_Simulation (
         My_Tasks       : IN     Tasks_Set;
         My_Resources   : IN     Resources_Set;
         Processor_Name : IN     Unbounded_String;
         Sched          : IN     Scheduling_Sequence_Ptr)
     RETURN Priority_Inversion_List IS


      Result : Priority_Inversion_List;


      -- Variables required to scan the resource set
      --
      Item        : Priority_Inversion_Item_Ptr;
      A_Resource  : Resource_Ptr;
      My_Iterator : Resources_Iterator;

      -- Store the event of the resource allocation
      --
      Allocation : Time_Unit_Event_Ptr;

      -- Store all Wait_For_Resource events
      --
      Wait_Events : ARRAY (Time_Unit_Range) OF Time_Unit_Event_Ptr;

      -- Simulation time of the stored event
      --
      Wait_Event_Time : ARRAY (Time_Unit_Range) OF Natural;

      -- Index on  Wait_Events and Indexes  tables
      --
      Index : Time_Unit_Range := 0;

      -- Index on the scheduling table
      --
      Current_Time : Time_Unit_Range := 0;

      PROCEDURE Perform_Analysis_Of_One_Resource IS

      BEGIN

         Current_Time := 0;

         -- First, find an Allocation Event
         --
         LOOP
            IF (Current_Time >= Sched.Nb_Entries-1)
                  THEN
               RETURN;
            END IF;

            IF (Sched.Entries(Current_Time).Item.Event_Type =
                  Allocate_Resource) THEN
               IF (Sched.Entries(Current_Time).Item.
                     Allocate_Resource.Name
                     = A_Resource.Name)
                     THEN
                  Allocation:=Sched.Entries(Current_Time).
                     Item;
                  EXIT;
               END IF;
            END IF;

            Current_Time:=Current_Time+1;
         END LOOP;

         LOOP

            Current_Time:=Current_Time+1;


            -- Search all Wait_For_Resource events upto the next allocation
            --
            IF (Current_Time >= Sched.
                  Nb_Entries-1) THEN
               RETURN;
            END IF;


            IF (Sched.Entries(Current_Time).Item.
                  Event_Type =
                  Allocate_Resource) THEN
               IF  (Sched.Entries(Current_Time).
                     Item.
                     Allocate_Resource.Name
                     = A_Resource.Name)
                     THEN

                  -- The next Allocate event is found, stop collecting data
                  -- and do analysis
                  --

                  -- Check that between the two "Allocate" events when have middle priority
                  -- tasks and that the priority of the first allocating task is less than
                  -- 2nd one.
                  --
                  IF (Allocation.Allocate_Task.Priority <
                        Sched
                        .Entries(Current_Time).Item.
                        Allocate_Task.Priority) THEN

                     DECLARE

                        -- Start time of the priority inversion
                        --
                        Start_Time : Natural := Sched.Entries (Current_Time).Data;






                     BEGIN
                        -- look for the start time of the priority inversion
                        --
                        FOR Z IN REVERSE 0..Index-1 LOOP
                           IF Wait_Events(Z).
                                 Wait_Task.
                                 Name =
                                 Sched.Entries(Current_Time).
                                 Item.
                                 Allocate_Task.Name
                                 THEN
                              Start_Time:=Wait_Event_Time(
                                 Z);
                           END IF;
                        END LOOP;


                        -- Store the detected priority inversion
                        --
                        IF Start_Time < Sched.
                              Entries(Current_Time).
                              Data

                              THEN
                           Item := NEW Priority_Inversion_Item;

                           Item.Start_Time:=Start_Time;
                           Item.End_Time:=Sched.
                              Entries(Current_Time).
                              Data;
                           Item.Task_Name:=Sched.
                              Entries(Current_Time).
                              Item.Allocate_Task.Name;
                           Item.Resource_Name:=Sched.
                              Entries(Current_Time).
                              Item.Allocate_Resource.Name;

                           Add(Result, Item);
                        END IF;
                     END;

                  END IF;



                  -- Ready for the next priority inversion analysis
                  --
                  Index:=0;
                  Allocation:=Sched.Entries(Current_Time).
                     Item;


               END IF;
            END IF;


            IF (Sched.Entries(Current_Time).Item.Event_Type =
                  Wait_For_Resource)
                  THEN
               IF (Sched.Entries(Current_Time).Item.
                     Wait_Resource.
                     Name
                     = A_Resource.Name)
                     THEN

                  Wait_Events(Index):=Sched.Entries(
                     Current_Time).
                     Item;
                  Wait_Event_Time(Index):=Sched.
                     Entries(Current_Time).
                     Data;
                  Index:=Index+1;
               END IF;
            END IF;
         END LOOP;

      END Perform_Analysis_Of_One_Resource;



   BEGIN

      Reset_Iterator (My_Resources, My_Iterator);

      Initialize(Result);

      LOOP
         Current_Element(My_Resources, A_Resource, My_Iterator);

         IF (Processor_Name = A_Resource.Cpu_Name) THEN
            Perform_Analysis_Of_One_Resource;
         END IF;

         EXIT WHEN Is_Last_Element(My_Resources, My_Iterator);

         Next_Element(My_Resources, My_Iterator);
      END LOOP;

      RETURN Result;

   END Priority_Inversion_From_Simulation;


   FUNCTION Number_Of_Preemption_From_Simulation (
         Sched    : IN     Scheduling_Sequence_Ptr;
         My_Tasks : IN     Tasks_Set)
     RETURN Natural IS

      Nb         : Natural  := 0;
      A_Task     : Task_Ptr;
      Usage_Time : Natural;

   BEGIN

      FOR I IN 0..Sched.Nb_Entries-1 LOOP
         IF (Sched.Entries(I).Item.Event_Type=Running_Task)
               THEN

            FOR K IN REVERSE 0..I-1  LOOP
               IF (Sched.Entries(K).Item.Event_Type=Running_Task)
                     THEN
                  IF ((Sched.Entries(K).
                           Item.Running_Task.Name /=
                           Sched.Entries(I).
                           Item.Running_Task.Name)
                        AND (Sched.Entries(K).Data + 1=
                           Sched.Entries(I).Data))
                        THEN

                     -- We know that the previous unit of time, another
                     -- task was running. Check now that is was a preeemption
                     --

                     -- look for the previously elected task and its state
                     --
                     A_Task:=Search_Task(My_Tasks, Sched.Entries(I).
                        Item.
                        Running_Task.Name);

                     -- Compute the running time of the task until its wake up time

                     --
                     Usage_Time :=0;
                     FOR Z IN REVERSE 0..K-1  LOOP
                        IF (Sched.Entries(Z).Item.Event_Type=
                              Running_Task)
                              THEN
                           IF A_Task.Name=Sched.Entries(Z).Item.
                                 Running_Task.Name
                                 THEN
                              Usage_Time:=Usage_Time+1;
                           END IF;
                        END IF;
                        IF (Sched.Entries(Z).Item.Event_Type=
                              Task_Activation)
                              THEN
                           IF A_Task.Name=Sched.Entries(Z).Item.
                                 Activation_Task.Name
                                 THEN
                              IF (Usage_Time/=A_Task.Capacity)
                                    AND (Usage_Time/=0)
                                    THEN
                                 Nb:=Nb+1;
                              END IF;
                              EXIT;
                           END IF;
                        END IF;
                     END LOOP;
                     EXIT;
                  END IF;
               END IF;
            END LOOP;

         END IF;
      END LOOP;

      RETURN  Nb;
   END  Number_Of_Preemption_From_Simulation;



   FUNCTION Number_Of_Switch_Context_From_Simulation (
         Sched : IN     Scheduling_Sequence_Ptr)
     RETURN Natural IS

      Nb : Natural := 0;

   BEGIN


      FOR I IN 0..Sched.Nb_Entries-1 LOOP
         IF (Sched.Entries(I).Item.Event_Type=Running_Task)
               THEN

            FOR K IN REVERSE 0..I-1 LOOP
               IF (Sched.Entries(K).Item.Event_Type=Running_Task)
                     THEN

                  IF (Sched.Entries(K).
                        Item.Running_Task.Name /=
                        Sched.Entries(I).
                        Item.Running_Task.Name)
                        THEN
                     Nb:=Nb + 1;
                  END IF;


                  EXIT;
               END IF;
            END LOOP;
         END IF;
      END LOOP;

      RETURN Nb;
   END  Number_Of_Switch_Context_From_Simulation;

   PROCEDURE Consumer_Response_Time (
         Sched              : IN     Scheduling_Sequence_Ptr;
         Consumer_Resp_Time : IN OUT Resp_Time_Consumer_Table_Ptr;
         My_Consumer_Task   : IN     Task_Ptr) IS

      My_Consumer_Task_Occurence_Table : Task_Occurence_Table_Ptr;
      Index                            : Resp_Time_Consumer_Range := 0;


   BEGIN

      -- initialize
      --
      My_Consumer_Task_Occurence_Table := NEW Task_Occurence_Table;
      All_Response_Time_By_Simulation (
         My_Consumer_Task,
         Sched,
         My_Consumer_Task_Occurence_Table);

      -- affect response time value
      --
      FOR I_Occurence IN Task_Occurence_Range  LOOP
         IF (My_Consumer_Task_Occurence_Table.Entries(
                  I_Occurence).Response_Time /= 0) THEN

            Consumer_Resp_Time.Entries(Index).Data :=
               Double(My_Consumer_Task_Occurence_Table.Entries(
                  I_Occurence).Response_Time);
            Index := Index + 1;

         END IF;
      END LOOP ;
      Consumer_Resp_Time.Nb_Entries := Index;

      Free(My_Consumer_Task_Occurence_Table);
   END Consumer_Response_Time;

END Scheduling_Sequences;

