------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Text_Io;
USE Text_Io;

WITH Scheduler;
USE Scheduler;

WITH Aadl_Parsers;
USE Aadl_Parsers;
WITH Xml_Parsers;
USE Xml_Parsers;
WITH Input_Sources.File;
USE Input_Sources.File;
WITH Sax.Readers;
USE Sax.Readers;

WITH Task_Offsets;
USE Task_Offsets;

WITH Integer_Util;
USE Integer_Util;

WITH Unbounded_Strings;
USE Unbounded_Strings;

PACKAGE BODY Systems IS



   PROCEDURE Delete_Task (
         A_Sys  : IN OUT System;
         A_Task :        Task_Ptr) IS
   BEGIN
      Delete_Task(A_Sys.Resources, A_Task.Name);
      Delete_Task(A_Sys.Buffers, A_Task.Name);
      Delete_All_Task_Dependencies(A_Sys.Dependencies, A_Task);
      Delete(A_Sys.Tasks, A_Task);
   END Delete_Task;


   PROCEDURE Delete_Buffer (
         A_Sys    : IN OUT System;
         A_Buffer :        Buffer_Ptr) IS
   BEGIN
      Delete_All_Task_Dependencies(A_Sys.Dependencies,
         A_Buffer);
      Delete(A_Sys.Buffers, A_Buffer);
   END Delete_Buffer;


   PROCEDURE Delete_Message (
         A_Sys     : IN OUT System;
         A_Message :        Message_Ptr) IS
   BEGIN
      Delete_All_Task_Dependencies(A_Sys.Dependencies,
         A_Message);
      Delete(A_Sys.Messages, A_Message);
   END Delete_Message;



   PROCEDURE Delete_Processor (
         A_Sys       : IN OUT System;
         A_Processor :        Processor_Ptr) IS
   BEGIN

      Delete_Processor_All_Task_Dependencies(A_Sys.Dependencies,
         A_Processor.Name);
      Delete_Processor(A_Sys.Resources, A_Processor.Name);
      Delete_Processor(A_Sys.Buffers, A_Processor.Name);
      Delete_Processor(A_Sys.Tasks, A_Processor.Name);
      Delete_Processor(A_Sys.Address_Spaces, A_Processor.Name);
      Delete(A_Sys.Processors, A_Processor);
   END Delete_Processor;



   PROCEDURE Delete_Address_Space (
         A_Sys            : IN OUT System;
         An_Address_Space :        Address_Space_Ptr) IS
   BEGIN
      Delete_Address_Space_All_Task_Dependencies(A_Sys.Dependencies,
         An_Address_Space.Name);
      Delete_Address_Space(A_Sys.Resources, An_Address_Space.Name);
      Delete_Address_Space(A_Sys.Buffers, An_Address_Space.Name);
      Delete_Address_Space(A_Sys.Tasks, An_Address_Space.Name);
      Delete(A_Sys.Address_Spaces, An_Address_Space);
   END Delete_Address_Space;



   PROCEDURE Duplicate (
         Src  : IN     System;
         Dest : IN OUT System) IS
   BEGIN

      Duplicate (Src.Tasks, Dest.Tasks);
      Duplicate (Src.Resources, Dest.Resources);
      Duplicate (Src.Messages, Dest.Messages);
      Duplicate (Src.Dependencies, Dest.Dependencies);
      Duplicate (Src.Processors, Dest.Processors);
      Duplicate (Src.Buffers, Dest.Buffers);
      Duplicate (Src.Networks, Dest.Networks);
      Duplicate (Src.Event_Analyzers, Dest.Event_Analyzers);
      Duplicate (Src.Address_Spaces, Dest.Address_Spaces);

   END Duplicate;



   PROCEDURE Put_Xml (
         A_System : IN     System) IS
   BEGIN

      Put(To_String(Export_Xml_Cheddar_Dtd (A_System)));
      Put(To_String(Export_Xml(A_System)));

   END Put_Xml;

   PROCEDURE Put_Aadl (
         A_System : IN     System) IS
   BEGIN
      New_Line;

      -- Export AADL standard properties and Cheddar's properties
      --
      --   Put_Line(To_String(Export_Aadl_Standard_Properties(A_System)));
      --   Put_Line(To_String(Export_Aadl_Project_Properties(A_System)));
      --   Put_Line(To_String(Export_Aadl_Cheddar_Properties(A_System)));


      -- Export the system itself
      --
      Put_Line(To_String(Export_Aadl_Implementations (A_System)));

      New_Line;

   END Put_Aadl;



   PROCEDURE Put (
         A_System : IN     System) IS
   BEGIN

      Put_Line("***** LIST OF PROCESSORS *****");
      Put(A_System.Processors);
      New_Line;
      New_Line;
      New_Line;
      Put_Line("***** LIST OF TASKS *****");
      Put(A_System.Tasks);
      New_Line;
      New_Line;
      New_Line;
      Put_Line("***** LIST OF RESSOURCES *****");
      Put(A_System.Resources);
      New_Line;
      New_Line;
      New_Line;
      Put_Line("***** LIST OF BUFFERS *****");
      Put(A_System.Buffers);
      New_Line;
      New_Line;
      New_Line;
      Put_Line("***** LIST OF MESSAGES *****");
      Put(A_System.Messages);
      New_Line;
      New_Line;
      New_Line;
      Put_Line("***** LIST OF DEPENCENCIES *****");
      Put(A_System.Dependencies);
      New_Line;
      New_Line;
      New_Line;
      Put_Line("***** LIST OF NETWORKS *****");
      Put(A_System.Networks);
      New_Line;
      New_Line;
      New_Line;
      Put_Line("***** LIST OF EVENTS ANALYZERS *****");
      Put(A_System.Event_Analyzers);
      New_Line;
      New_Line;
      New_Line;
      Put_Line("***** LIST OF ADDRESS SPACES *****");
      Put(A_System.Address_Spaces);
      New_Line;
      New_Line;
      New_Line;

   END Put;



   PROCEDURE Put (
         A_System : IN     System_Ptr) IS
   BEGIN
      Put(A_System.All);
   END Put;



   ------------------------------------------------




   PROCEDURE Initialize (
         A_System : IN OUT System) IS
   BEGIN

      Reset(A_System.Networks);
      Reset(A_System.Tasks);
      Reset(A_System.Messages);
      Reset(A_System.Resources);
      IF A_System.Dependencies /= NULL
            THEN
         Free(A_System.Dependencies);
      END IF;
      A_System.Dependencies:= NEW Tasks_Dependencies;
      Reset(A_System.Dependencies);
      Reset(A_System.Buffers);
      Reset(A_System.Processors);
      Reset(A_System.Event_Analyzers);
      Reset(A_System.Address_Spaces);
   END Initialize;


   PROCEDURE Initialize (
         A_System : IN     System_Ptr) IS
   BEGIN
      Initialize(A_System.All);
   END Initialize;




   ------------------------------------------------


   PROCEDURE Read_From_Xml_File (
         A_System  : IN OUT System;
         Dir_List  : IN     Unbounded_String_List;
         File_Name : IN     String) IS
   BEGIN
      Read_From_Xml_File(A_System, Dir_List, To_Unbounded_String(
            File_Name));
   END Read_From_Xml_File;


   PROCEDURE Write_To_Xml_File (
         A_System  : IN     System;
         File_Name : IN     String) IS
   BEGIN
      Write_To_Xml_File(A_System, To_Unbounded_String(File_Name));
   END Write_To_Xml_File;


   PROCEDURE Read_From_Xml_File (
         A_System  : IN OUT System;
         Dir_List  : IN     Unbounded_String_List;
         File_Name : IN     Unbounded_String) IS

      Read       :          File_Input;
      My_Reader  :          Xml_Project_Parser;
      Name_Start :          Natural;
      S          : CONSTANT String             := To_String (File_Name);


   BEGIN

      --  Base file name should be used as the public Id
      --
      Name_Start := S'Last;
      WHILE Name_Start >= S'First  AND THEN S (Name_Start) /= '/' LOOP
         Name_Start := Name_Start - 1;
      END LOOP;
      Set_Public_Id (Read, S (Name_Start + 1 .. S'Last));
      Set_System_Id (Read, S);
      Open (S, Read);

      --  xmlns:* attributes will be reported in Start_Element
      --
      Set_Feature (My_Reader, Namespace_Prefixes_Feature, False);
      Set_Feature (My_Reader, Validation_Feature, True);

      Parse (My_Reader, Read);
      A_System:=Get_Parsed_System(My_Reader);
      Close (Read);


   EXCEPTION
      WHEN Xml_Fatal_Error =>
         Close(Read);
         RAISE Xml_Read_Error;


   END Read_From_Xml_File;



   PROCEDURE Write_To_Xml_File (
         A_System  : IN     System;
         File_Name : IN     Unbounded_String) IS


      Into : File_Type;

   BEGIN

      -- Open file and Write DTD/XML Header
      --
      Create(Into,
         Mode => Out_File,
         Name => To_String(File_Name));

      Put(Into, To_String(Export_Xml_Cheddar_Dtd (A_System)));
      Put(Into, To_String(Export_Xml(A_System)));

      Close(Into);

   END Write_To_Xml_File;


   PROCEDURE Read_From_Aadl_File (
         A_System          : IN OUT System;
         Dir_List          : IN     Unbounded_String_List;
         Project_File_List : IN     Unbounded_String_List) IS

      My_Reader : Aadl_Project_Parser;

   BEGIN

      Parse (My_Reader, Dir_List, Project_File_List);
      A_System:=Get_Parsed_System(My_Reader);

   END Read_From_Aadl_File;


   PROCEDURE Write_To_Aadl_File (
         A_System  : IN     System;
         File_Name : IN     Unbounded_String) IS

      Into : File_Type;

   BEGIN

      -- Open file and Write aadl specification
      --
      Create(Into,
         Mode => Out_File,
         Name => To_String(File_Name));


      New_Line(Into);


      -- Export AADL standard customized properties
      --
      -- Put_Line(Into, To_String(Export_Aadl_Standard_Properties(A_System)));
      -- Put_Line(Into, To_String(Export_Aadl_Project_Properties(A_System)));
      -- Put_Line(Into, To_String(Export_Aadl_Cheddar_Properties(A_System)));



      -- Export the system itself
      --
      Put_Line(Into, To_String(Export_Aadl_Implementations (A_System)));

      New_Line(Into);

      Close(Into);

   END Write_To_Aadl_File;



   PROCEDURE Write_Aadl_Standard_Properties_Set_To_File (
         A_System : IN     System) IS

      Into : File_Type;

   BEGIN

      -- Open file and Write aadl specification
      --
      Create(Into,
         Mode => Out_File,
         Name => "AADL_Properties.aadl");


      New_Line(Into);
      Put_Line(Into, To_String(Export_Aadl_Standard_Properties(A_System)));
      New_Line(Into);

      Close(Into);
   END  Write_Aadl_Standard_Properties_Set_To_File;



   PROCEDURE Write_Cheddar_Property_Sets_To_File (
         A_System : IN     System) IS

      Into : File_Type;

   BEGIN

      -- Open file and Write aadl specification
      --
      Create(Into,
         Mode => Out_File,
         Name => "AADL_Project.aadl");
      New_Line(Into);
      Put_Line(Into, To_String(Export_Aadl_Project_Properties(A_System)));
      New_Line(Into);
      Close(Into);


      Create(Into,
         Mode => Out_File,
         Name => "Cheddar_Properties.aadl");
      New_Line(Into);
      Put_Line(Into, To_String(Export_Aadl_Cheddar_Properties(A_System)));
      New_Line(Into);
      Close(Into);


      Create(Into,
         Mode => Out_File,
         Name => "User_Defined_Cheddar_Properties.aadl");
      New_Line(Into);
      Put_Line(Into, To_String(
            Export_Aadl_User_Defined_Cheddar_Properties(A_System)));
      New_Line(Into);
      Close(Into);



   END Write_Cheddar_Property_Sets_To_File;


   ------------------------------------------------

   FUNCTION Export_Aadl_User_Defined_Cheddar_Properties (
         A_System : IN     System)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;
      Tmp    : Unbounded_String := Empty_String;

   BEGIN

      -- Export user's defined properties
      --
      Tmp:=  Export_Aadl_User_Defined_Properties(
         A_System.Tasks);
      Tmp:= Tmp & Export_Aadl_User_Defined_Properties(
         A_System.Messages);


      Result:=Result & To_Unbounded_String(
         "property set User_Defined_Cheddar_Properties is") &
         Unbounded_Lf & Unbounded_Lf;

      IF Tmp=Empty_String
            THEN
         Result:=Result &
            "Dummy_User_Defined_Cheddar_Properties : aadlboolean"
            & Unbounded_Lf & "applies to (thread, thread group);";
      ELSE
         Result:=Result & Tmp;
      END IF;


      Result:=Result & Unbounded_Lf & To_Unbounded_String(
         "end User_Defined_Cheddar_Properties;") & Unbounded_Lf &
         Unbounded_Lf;

      RETURN Result;

   END Export_Aadl_User_Defined_Cheddar_Properties;






   FUNCTION Export_Aadl_Standard_Properties (
         A_System : IN     System)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;

   BEGIN

      Result:=Result & To_Unbounded_String(
         "property set AADL_Properties is") & Unbounded_Lf;

      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Activate_Deadline : Time applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Activate_Execution_Time : Time_Range applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Activate_Entrypoint : aadlstring  applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Active_Thread_Handling_Protocol : inherit Supported_Active_Thread_Handling_Protocols => value(Default_Active_Thread_Handling_Protocol) applies to (thread, thread group, process, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Active_Thread_Queue_Handling_Protocol :  enumeration (flush, hold) => flush applies to (thread, thread group, process, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Actual_Connection_Binding : inherit  reference(bus, processor, device) applies to (port connections, thread, thread group, process, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Actual_Latency : Time applies to (flow);") & Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Actual_Memory_Binding : inherit  reference(memory) applies to (thread, thread group, process, system, data port, event data port, subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Actual_Processor_Binding : inherit  reference(processor) applies to (thread, thread group, process, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Actual_Subprogram_Call :  reference(server subprogram) applies to (subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Actual_Subprogram_Call_Binding :  reference(bus, processor, memory) applies to (subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Actual_Thoughput : Data_Volume applies to (flow);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Aggregate_Data_Port : aadlboolean  => false applies to (port group);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Access_Protocol : list of  enumeration (Memory_Access, Device_Access) applies to (bus);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Connection_Binding : inherit list of  reference(bus, processor, device) applies to (port connections, thread, thread group, process, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Connection_Binding_Class : inherit list of  classifier(processor, bus, device) applies to (port connections, thread, thread group, process, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Connection_Protocol : list of  enumeration (Data_Connection, Event_Connection) applies to (bus);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Dispatch_Protocol : list of Supported_Dispatch_Protocols applies to (processor);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Memory_Binding : inherit list of  reference(memory, system, processor) applies to (thread, thread group, process, system, device, data port, event data port, subprogram, processor);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Memory_Binding_Class : inherit list of  classifier(memory, system, processor) applies to (thread, thread group, process, system, device, data port, event data port, subprogram, processor);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Message_Size : Size_Range applies to (bus);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Period : list of Time_Range applies to (processor, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Processor_Binding : inherit list of  reference(processor, system) applies to (thread, thread group, process, system, device);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Processor_Binding_Class : inherit list of  classifier(processor, system) applies to (thread, thread group, process, system, device);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Subprogram_Call : list of  reference(server subprogram) applies to (subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Allowed_Subprogram_Call_Binding : inherit list of  reference(bus, processor, device) applies to (subprogram, thread, thread group, process, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Assign_Time : Time applies to (processor, bus);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Assign_Byte_Time : Time applies to (processor, bus);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Assign_Fixed_Time : Time applies to (processor, bus);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Available_Memory_Binding : inherit list of  reference(memory, system) applies to (system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Available_Processor_Binding : inherit list of  reference(processor, system) applies to (system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Base_Address : access aadlinteger 0 .. value(Max_Base_Address) applies to (data);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Client_Subprogram_Execution_Time : Time applies to (subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Clock_Jitter : Time applies to (processor, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Clock_Period : Time applies to (processor, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Clock_Period_Range : Time_Range applies to (processor, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Compute_Execution_Time : Time_Range applies to (thread, subprogram, event port, event data port);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Compute_Deadline : Time applies to (thread, subprogram, event port, event data port);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Compute_Entrypoint : aadlstring  applies to (thread, subprogram, event port, event data port);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Concurrency_Control_Protocol : Supported_Concurrency_Control_Protocols => NoneSpecified applies to (data);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Connection_Protocol : Supported_Connection_Protocols applies to (connections);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Deactivate_Deadline : Time applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Deactivate_Execution_Time : Time_Range applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Deactivate_Entrypoint : aadlstring  applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Deadline : inherit Time => value(Period) applies to (thread, thread group, process, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Dequeue_Protocol :  enumeration (OneItem, AllItems) => OneItem applies to (event port, event data port);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Device_Dispatch_Protocol : Supported_Dispatch_Protocols => Aperiodic applies to (thread, device);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Device_Register_Address : aadlinteger  applies to (port, port group);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Dispatch_Protocol : Supported_Dispatch_Protocols applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Expected_Latency : Time applies to (flow);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Expected_Throughput : Data_Volume applies to (flow);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Finalize_Deadline : Time applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Finalize_Execution_Time : Time_Range applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Finalize_Entrypoint : aadlstring  applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Hardware_Description_Source_Text : inherit list of aadlstring  applies to (memory, bus, device, processor, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Hardware_Source_Language : Supported_Hardware_Source_Languages applies to (memory, bus, device, processor, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Initialize_Deadline : Time applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Initialize_Execution_Time : Time_Range applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Initialize_Entrypoint : aadlstring  applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Latency : Time applies to (flow, connections);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Load_Deadline : Time applies to (process);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Load_Time : Time_Range applies to (process, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Memory_Protocol :  enumeration (Read_Only, Write_Only, Read_Write) => Read_Write applies to (memory);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Not_Collocated : list of  reference(data, thread, process, system, connections) applies to (data, thread, process, system, connections);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Overflow_Handling_Protocol :  enumeration (Dropoldest, Dropnewest, Error) => Dropoldest applies to (event port, event data port, subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Period : inherit Time applies to (thread, thread group, process, system, device);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Process_Swap_Execution_Time : Time_Range applies to (processor);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Propagation_Delay : Time_Range applies to (bus);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Provided_Access : access  enumeration (Read_Only, Write_Only, Read_Write, By_Method) => Read_Write applies to (data);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Queue_Processing_Protocol : Supported_Queue_Processing_Protocols => FIFO applies to (event port, event data port, subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Queue_Size : aadlinteger 0 .. value(Max_Queue_Size) => 0 applies to (event port, event data port, subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Read_Time : list of Time_Range applies to (memory);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Recover_Deadline : Time applies to (thread, server subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Recover_Execution_Time : Time_Range applies to (thread, server subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Recover_Entrypoint : aadlstring  applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Required_Access : access  enumeration (Read_Only, Write_Only, Read_Write, By_Method) => Read_Write applies to (data, bus);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Required_Connection : aadlboolean  => true applies to (port);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Runtime_Protection : inherit aadlboolean  => true applies to (process);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Scheduling_Protocol : list of Supported_Scheduling_Protocols applies to (processor, process);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Server_Subprogram_Call_Binding : inherit list of  reference(thread, processor) applies to (subprogram, thread, thread group, process, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Source_Code_Size : Size applies to (data, thread, process, subprogram, processor, device);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Source_Data_Size : Size applies to (data, subprogram, thread, process, processor, device);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Source_Heap_Size : Size applies to (thread);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Source_Language : Supported_Source_Languages applies to (subprogram, data, thread, process, bus, device, processor);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Source_Name : aadlstring  applies to (data, port, subprogram, parameter);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Source_Stack_Size : Size applies to (thread, subprogram, processor, device);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Source_Text : inherit list of aadlstring  applies to (data, thread, thread group, process, system, memory, bus, device, processor, port, subprogram, parameter, port group);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Startup_Deadline : inherit Time applies to (processor, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Subprogram_Execution_Time : Time_Range applies to (subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Supported_Source_Language : list of Supported_Source_Languages applies to (processor, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Synchronized_Component : inherit aadlboolean  applies to (mode);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Thread_Limit : aadlinteger 0 .. value(Max_Thread_Limit) => value(Max_Thread_Limit) applies to (processor);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Thread_Swap_Execution_Time : Time_Range applies to (processor, system);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Throughput : Data_Volume applies to (flow, connections);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Transmission_Time : list of Time_Range applies to (bus);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Type_Source_Name : aadlstring  applies to (data, port, subprogram);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Urgency : aadlinteger 0 .. value(Max_Urgency) applies to (port);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Word_Count : aadlinteger 0 .. value(Max_Word_Count) applies to (memory);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Word_Size : Size => 8 Bits applies to (memory);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Word_Space : aadlinteger 1 .. value(Max_Word_Space) => 1 applies to (memory);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Write_Time : list of Time_Range applies to (memory);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Data_Volume : type aadlinteger  units (bitsps, Bps => bitsps * 8, Kbps => Bps * 1000, Mbps => Kbps * 1000, Gbps => Mbps * 1000);") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Size : type aadlinteger 0 B .. value(Max_Memory_Size) units Size_Units;") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Size_Range : type range of Size;") & Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Time : type aadlinteger 0 Ns .. value(Max_Time) units Time_Units;") &
         Unbounded_Lf;
      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Time_Range : type range of Time;") & Unbounded_Lf;

      Result:=Result & To_Unbounded_String(
         "end AADL_Properties;") & Unbounded_Lf;

      RETURN Result;

   END Export_Aadl_Standard_Properties;



   FUNCTION Export_Aadl_Project_Properties (
         A_System : IN     System)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;

   BEGIN
      Result:=Result & To_Unbounded_String(
         "property set AADL_Project is") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Default_Active_Thread_Handling_Protocol : constant Supported_Active_Thread_Handling_Protocols => abort;") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Max_Aadlinteger : constant aadlinteger => 2#1#e32;") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Max_Memory_Size : constant aadlinteger Size_Units => 2#1#e32;") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Max_Queue_Size : constant aadlinteger => 512;") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Max_Base_Address : constant aadlinteger => 512;") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Max_Urgency : constant aadlinteger => 12;" ) &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Max_Time : constant aadlinteger Time_Units => 1000 Hr;") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Max_Thread_Limit : constant aadlinteger => 32;") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Max_Word_Count : constant aadlinteger => 64000;") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Max_Word_Space : constant aadlinteger => 64;") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Supported_Active_Thread_Handling_Protocols : type enumeration (abort, complete_one_flush_queue, complete_one_transfer_queue, complete_one_preserve_queue, complete_all);")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Supported_Connection_Protocols : type enumeration (UDP, TCP);") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Supported_Dispatch_Protocols : type enumeration (Periodic, Background, Poisson_Process, Parametric);")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Supported_Concurrency_Control_Protocols : type enumeration (No_Protocol, Priority_Ceiling_Protocol, Priority_Inheritance_Protocol);")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Supported_Hardware_Source_Languages : type enumeration (VHDL);")&
         Unbounded_Lf;

      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Supported_Scheduling_Protocols : type enumeration (");
      FOR I IN Scheduler_Type'First..Scheduler_Type'Last LOOP
         Result:= Result & To_Unbounded_String(Scheduler_Type'Image(I));
         IF I /= Scheduler_Type'Last
               THEN
            Result:=Result & To_Unbounded_String(", ");
         ELSE
            Result := Result & To_Unbounded_String(");") & Unbounded_Lf;
         END IF;
      END LOOP;

      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Supported_Queue_Processing_Protocols : type enumeration (FIFO);")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Supported_Source_Languages : type enumeration (Ada95, C);") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Time_Units : type units (ps, Ns => ps * 1000, Us => Ns * 1000, Ms => Us * 1000, Sec => Ms * 1000, Min => Sec * 60, Hr => Min * 60);") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(Ascii.Ht &
         "Size_Units : type units (Bits, B => Bits * 8, KB => B * 1000, MB => KB * 1000, GB => MB * 1000);") &
         Unbounded_Lf;
      Result:= Result & To_Unbounded_String("end AADL_Project;") &
         Unbounded_Lf;

      RETURN Result;

   END Export_Aadl_Project_Properties;


   FUNCTION Export_Aadl_Cheddar_Properties (
         A_System : IN     System)
     RETURN Unbounded_String IS

      Result : Unbounded_String;

   BEGIN

      -- Export cheddar's specific properties
      --
      Result:=Result & Unbounded_Lf & To_Unbounded_String(
         "property set Cheddar_Properties is") & Unbounded_Lf;



      -- Thread and thread group properties
      --
      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Dispatch_Seed_is_Predictable : aadlboolean") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (thread, thread group);") &
         Unbounded_Lf;

      Result:=Result & Ascii.Ht & To_Unbounded_String(
         "Source_Text : inherit list of aadlstring  applies to (data, thread, thread group, process, system, memory, bus, device, processor, port, subprogram, parameter, port group);") &
         Unbounded_Lf;

      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Dispatch_Seed_Value : aadlinteger") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (thread, thread group);") &
         Unbounded_Lf;

      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Dispatch_Absolute_Time : aadlinteger") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (thread, thread group);") &
         Unbounded_Lf;

      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Criticality : aadlinteger") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (thread, thread group);") &
         Unbounded_Lf;

      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Bound_On_Data_Blocking_Time : aadlinteger") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (thread, thread group);") &
         Unbounded_Lf;

      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Dispatch_Jitter : aadlinteger") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (thread, thread group);") &
         Unbounded_Lf;

      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Fixed_Priority : aadlinteger 0..255") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (thread, thread group);") &
         Unbounded_Lf;

      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "POSIX_Scheduling_Policy : enumeration (SCHED_FIFO, SCHED_RR, SCHED_OTHERS)") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (thread, thread group);") &
         Unbounded_Lf;
      FOR I IN Offsets_Range'RANGE LOOP
         Result:=Result & To_Unbounded_String( Ascii.Ht &
            "Dispatch_Offset_Value" &
            To_String(
               Integer_Util.Format(Integer(I))) & " : aadlinteger") &
            Unbounded_Lf;
         Result:=Result & To_Unbounded_String(
            Ascii.Ht & Ascii.Ht &
            "applies to (thread, thread group);") &
            Unbounded_Lf;
         Result:=Result & To_Unbounded_String( Ascii.Ht &
            "Dispatch_Offset_Time" & To_String(
               Integer_Util.Format(Integer(I)))  & " : aadlinteger") &
            Unbounded_Lf;
         Result:=Result & To_Unbounded_String(
            Ascii.Ht & Ascii.Ht &
            "applies to (thread, thread group);") &
            Unbounded_Lf;
      END LOOP;


      -- System properties
      --
      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Task_Precedencies : list of aadlstring") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (system);") &
         Unbounded_Lf;


      -- Processor properties
      --
      Result:=Result & Unbounded_Lf & To_Unbounded_String( Ascii.Ht &
         "Scheduler_Quantum : aadlinteger") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (processor);") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Preemptive_Scheduler : aadlboolean") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (processor);") &
         Unbounded_Lf & Unbounded_Lf;



      -- Data properties
      --
      Result:=Result & Unbounded_Lf & To_Unbounded_String( Ascii.Ht &
         "Data_Concurrency_State : aadlinteger") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (data);") & Unbounded_Lf &
         Unbounded_Lf;


      -- Process properties
      --
      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Critical_Section : list of aadlstring") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (process);") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Source_Global_Heap_Size : aadlinteger") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (process);") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Source_Global_Stack_Size : aadlinteger") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (process);") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Source_Global_Text_Size : aadlinteger") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (process);") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String( Ascii.Ht &
         "Source_Global_Data_Size : aadlinteger") & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         Ascii.Ht & Ascii.Ht & "applies to (process);") &
         Unbounded_Lf & Unbounded_Lf;



      Result:=Result & To_Unbounded_String(
         "end Cheddar_Properties;") &
         Unbounded_Lf & Unbounded_Lf;


      RETURN Result;

   END Export_Aadl_Cheddar_Properties;




   FUNCTION Export_Aadl_Implementations (
         A_System : IN     System)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;
      Tmp    : Unbounded_String := Empty_String;

   BEGIN

      -- Export all sub-components of the system
      --
      Result:=Result & Export_Aadl_Implementations(
         A_System.Resources);
      Result:=Result & Export_Aadl_Implementations(A_System.Tasks,
         A_System.Resources);
      Result:=Result & Export_Aadl_Implementations(
         A_System.Address_Spaces,
         A_System.
         Tasks,  A_System.Resources);
      Result:=Result & Export_Aadl_Implementations(A_System.Buffers);
      Result:=Result & Export_Aadl_Implementations(
         A_System.Messages);
      Result:=Result & Export_Aadl_Implementations(
         A_System.Networks);
      Result:=Result & Export_Aadl_Implementations(
         A_System.Processors);


      -- Export the system composition
      --
      Result:=Result & Unbounded_Lf & To_Unbounded_String(
         "system Cheddar") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String( "end Cheddar;") &
         Unbounded_Lf;
      Result:=Result & Unbounded_Lf;

      Result:=Result & To_Unbounded_String(
         "system implementation Cheddar.Impl" ) & Unbounded_Lf;
      Result:=Result & To_Unbounded_String(  Ascii.Ht & "subcomponents") &
         Unbounded_Lf;

      -- List all processors, processes and networks
      --
      Result:=Result & Export_Aadl_Declarations(A_System.Networks, 2);
      Result:=Result & Export_Aadl_Declarations(A_System.Processors,
         2);
      Result:=Result & Export_Aadl_Declarations(
         A_System.Address_Spaces,
         2);


      -- Properties related to the global system : processor and bus connections, event
      -- analyzers, task precedencies
      --
      Tmp:=Export_Aadl_Properties(A_System.Address_Spaces, 2);
      Tmp:= Tmp & Export_Aadl_Properties(A_System.Event_Analyzers, 2);
      Tmp:= Tmp & Export_Aadl_Properties(A_System.Dependencies, 2);
      IF Tmp/=Empty_String
            THEN
         Result:=Result & To_Unbounded_String(  Ascii.Ht & "properties") &
            Unbounded_Lf & Tmp;
      END IF;

      Result:=Result & To_Unbounded_String( "end Cheddar.Impl;") &
         Unbounded_Lf & Unbounded_Lf;

      RETURN Result;

   END Export_Aadl_Implementations;


   FUNCTION Export_Aadl_Declarations (
         A_System     : IN     System;
         Number_Of_Ht : IN     Natural)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;

   BEGIN
      RETURN Result;
   END Export_Aadl_Declarations;



   FUNCTION Export_Xml_Cheddar_Dtd (
         A_System : IN     System)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;
   BEGIN

      Result:=Result & To_Unbounded_String(
         "<?xml version=""1.0"" standalone=""yes""?>  ") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<?xml-stylesheet type=""text/xsl"" href=""cheddar_project.xsl""?>")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String( "<!DOCTYPE cheddar [  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT name (#PCDATA) > ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT scheduler (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT parameter (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT network_link (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT offset (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT period (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT capacity (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT deadline (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT blocking_time (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT policy (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT priority (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT criticality (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT cpu_name (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT address_space_name (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT jitter (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT seed (#PCDATA) > ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT predictable_seed (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT activation_rule (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT state (#PCDATA) > ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT protocol (#PCDATA) > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT size (#PCDATA) > ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT qs (#PCDATA) > ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT time (#PCDATA) > ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT buffer_user (#PCDATA) >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT resource_user (#PCDATA) >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT event_Analyzer (#PCDATA) >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT type (#PCDATA) >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT text_memory_size (#PCDATA) >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT heap_memory_size (#PCDATA) >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT stack_memory_size (#PCDATA) >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT data_memory_size (#PCDATA) >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT cheddar (processors, (address_spaces)?, (tasks)?, ((event_analyzers)?|(networks)?|(buffers)?|(resources)?|(messages)?),(dependencies)?  )  >   ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT processors (processor)+ >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT processor (name|scheduler|network_link) >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT address_spaces (address_space)+ >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT address_space (name|text_memory_size|data_memory_size|stack_memory_size|heap_memory_size) >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT networks (network)+ >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT network (name|type) >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT tasks (task)+>  ") &
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT task (name|cpu_name|address_space_name|capacity|start_time|(stack_memory_size)?|(text_memory_size)?|(period)?|");
      Result:=Result & To_Unbounded_String(
         "(deadline)?|(parameters)?|(offsets)?|(jitter)?|(policy)?|(criticality)?|(priority)?|(predictable_seed)?|(blocking_time)?|(seed)?")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "|(activation_rule)?) >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT offsets (offset)+ >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT parameters (parameter)+ >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT messages (message)+ >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT message (name|size|(period)?|(deadline)?|(jitter)?) >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT buffers (buffer)+ >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT buffer (cpu_name|address_space_name|qs|name|size|(buffer_used_by)?) >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT buffer_used_by (buffer_user)+ >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT resources (resource)+ >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT resource (cpu_name|address_space_name|name|protocol|(state)?|(resource_used_by)?) >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT resource_used_by (resource_user)+ >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT dependencies (dependency)+ >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT event_analyzers (event_analyzer)+ >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ELEMENT dependency (#PCDATA)  >  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(  "<!ATTLIST scheduler ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        quantum CDATA  ""0""  ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        is_preemptive CDATA  ""PREEMPTIVE""  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        parametric_file_name CDATA """" >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "<!ATTLIST event_analyzer  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        parametric_file_name CDATA """" >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(  "<!ATTLIST task  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        task_type CDATA  ""APERIODIC_TYPE""  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        x CDATA ""0""  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        y CDATA ""0"" > ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String( "<!ATTLIST buffer  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(  "        x CDATA ""0"" ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        y CDATA ""0"" > ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(  "<!ATTLIST message  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(  "        x CDATA ""0"" ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        y CDATA ""0"" > ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(  "<!ATTLIST buffer_user ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        buffer_role CDATA  ""producer"" >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(  "<!ATTLIST dependency ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        from_type CDATA ""task"" ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        to_type CDATA ""task"" > ")& Unbounded_Lf;
      Result:=Result & To_Unbounded_String(  "<!ATTLIST parameter " )&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(
         "        parameter_type CDATA  ""integer"" >  ")&
         Unbounded_Lf;
      Result:=Result & To_Unbounded_String(  "] >  ")& Unbounded_Lf;
      Result:=Result & Unbounded_Lf& Unbounded_Lf& Unbounded_Lf;


      RETURN Result;

   END Export_Xml_Cheddar_Dtd;


   FUNCTION Export_Xml (
         A_System : IN     System)
     RETURN Unbounded_String IS

      Result : Unbounded_String := Empty_String;

   BEGIN


      Result:= Result & To_Unbounded_String("<cheddar>") &
         Unbounded_Lf;


      -- Call each export part of the system
      --
      Result:= Result & Export_Xml(A_System.Processors);
      Result:= Result & Export_Xml(A_System.Address_Spaces);
      Result:= Result & Export_Xml(A_System.Tasks);
      Result:= Result & Export_Xml(A_System.Networks);
      Result:= Result & Export_Xml(A_System.Buffers);
      Result:= Result & Export_Xml(A_System.Messages);
      Result:= Result & Export_Xml(A_System.Resources);
      Result:= Result & Export_Xml(A_System.Dependencies);
      Result:= Result & Export_Xml(A_System.Event_Analyzers);


      Result:= Result & To_Unbounded_String("</cheddar>") &
         Unbounded_Lf &
         Unbounded_Lf;
      Result:= Result & Unbounded_Lf & Unbounded_Lf;


      RETURN Result;

   END Export_Xml;




END Systems;


