------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Aadl_Config;
USE Aadl_Config;

WITH Buffers;
USE Buffers;
USE Buffers.Buffer_Roles;

WITH Messages;
USE Messages;

WITH Tasks;
USE Tasks;

WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Text_Io;
USE Text_Io;



PACKAGE BODY Task_Dependencies IS


   -- Add a task in the Dependent task list
   -- only if it now already added
   --
   PROCEDURE Add_Dependent_Tasks (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         My_Task         : IN     Task_Ptr) IS

      Ite1   : Tasks_Iterator;
      A_Task : Task_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.Dependent_Tasks) THEN
         Reset_Iterator(My_Dependencies.Dependent_Tasks, Ite1);
         LOOP
            Current_Element(My_Dependencies.Dependent_Tasks, A_Task, Ite1);

            IF A_Task.Name=My_Task.Name
                  THEN
               RETURN;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.Dependent_Tasks,
               Ite1);
            Next_Element(My_Dependencies.Dependent_Tasks, Ite1);
         END LOOP;
      END IF;

      Add(My_Dependencies.Dependent_Tasks,  My_Task);

   END Add_Dependent_Tasks;



   -------------------------------------------------------------------------



   PROCEDURE Free (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr) IS
   BEGIN
      Free(My_Dependencies.Dependent_Tasks);
      Free(Half_Dep_Set.Set(My_Dependencies.All));
   END Free;



   FUNCTION Export_Xml (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Unbounded_String IS


      My_Iterator : Tasks_Dependencies_Iterator;
      A_Half_Dep  : Dependency_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator);


         Result:= Result & To_Unbounded_String( "<dependencies>") &
            Unbounded_Lf;
         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator);

            Result:= Result & To_Unbounded_String(  Ascii.Ht &
               "<dependency");

            IF A_Half_Dep.Depend_Type=
                  Buffer_Dependency THEN
               IF A_Half_Dep.Buffer_Orientation =
                     From_Task_To_Object
                     THEN
                  Result:= Result & To_Unbounded_String(
                     " to_type=""buffer""> " &
                     To_String(A_Half_Dep.Buffer_Dependent_Task.
                        Name) & " " &
                     To_String(A_Half_Dep.
                        Buffer_Dependency_Object.Name) &
                     " ");
               ELSE
                  Result:= Result & To_Unbounded_String(
                     " from_type=""buffer""> " &
                     To_String(A_Half_Dep.
                        Buffer_Dependency_Object.Name) & " " &
                     To_String(A_Half_Dep.Buffer_Dependent_Task.
                        Name) & " ");
               END IF;
            ELSE
               IF A_Half_Dep.Depend_Type=
                     Communication_Dependency THEN
                  IF A_Half_Dep.Communication_Orientation =
                        From_Task_To_Object
                        THEN
                     Result:= Result & To_Unbounded_String(
                        " to_type=""message""> " &
                        To_String(
                           A_Half_Dep.Communication_Dependent_Task.
                           Name) & " " &
                        To_String(A_Half_Dep.
                           Communication_Dependency_Object.Name) &
                        " ");
                  ELSE
                     Result:= Result & To_Unbounded_String(
                        " from_type=""message""> " &
                        To_String(A_Half_Dep.
                           Communication_Dependency_Object.Name) & " " &

                        To_String(
                           A_Half_Dep.Communication_Dependent_Task.
                           Name)  &  " ");
                  END IF;
               ELSE
                  Result:= Result & To_Unbounded_String(  "> " & To_String(
                        A_Half_Dep.
                        Precedency_From_Task.Name) & " " & To_String(
                        A_Half_Dep.Precedency_To_Task.Name) &
                     " ");
               END IF;
            END IF;

            Result:= Result & To_Unbounded_String("</dependency>") &
               Unbounded_Lf;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator);

            Next_Element(My_Dependencies.All, My_Iterator);

         END LOOP;
         Result:= Result & To_Unbounded_String("</dependencies>") &
            Unbounded_Lf;

      END IF;

      RETURN Result;

   END Export_Xml;




   FUNCTION Export_Aadl_Properties (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         Number_Of_Ht    : IN     Natural)
     RETURN Unbounded_String IS

      My_Iterator     : Tasks_Dependencies_Iterator;
      A_Half_Dep      : Dependency_Ptr;
      Head_Is_Written : Boolean                     := False;

      Result : Unbounded_String := Empty_String;

   BEGIN


      IF Aadl_Export_Precedencies_To_Properties
            THEN

         IF NOT Is_Empty(My_Dependencies.All) THEN

            Reset_Iterator(My_Dependencies.All, My_Iterator);
            LOOP
               Current_Element(My_Dependencies.All, A_Half_Dep,
                  My_Iterator);


               IF A_Half_Dep.Depend_Type=Precedence_Dependency
                     THEN

                  IF NOT Head_Is_Written
                        THEN
                     FOR I IN 1..Number_Of_Ht LOOP
                        Result:= Result & Ascii.Ht;
                     END LOOP;
                     Result:= Result & To_Unbounded_String(
                        "Cheddar_Properties::Task_Precedencies => (") &
                        Unbounded_Lf;
                     Head_Is_Written:=True;
                  ELSE
                     Result:=Result & "," & Unbounded_Lf;
                  END IF;

                  FOR I IN 1..Number_Of_Ht+1 LOOP
                     Result:= Result & Ascii.Ht;
                  END LOOP;
                  Result:= Result &  """" & "instancied_" &
                     A_Half_Dep.Precedency_From_Task.Address_Space_Name &
                     ".instancied_" & A_Half_Dep.Precedency_From_Task.Name &
                     """" & ", " & """" &
                     "instancied_" &
                     A_Half_Dep.Precedency_From_Task.Address_Space_Name &
                     ".instancied_" & A_Half_Dep.Precedency_To_Task.Name &
                     """";
               END IF;

               EXIT WHEN Is_Last_Element(My_Dependencies.All,
                  My_Iterator);


               Next_Element(My_Dependencies.All, My_Iterator);

            END LOOP;

            IF Head_Is_Written
                  THEN
               Result:= Result & To_Unbounded_String(");") &
                  Unbounded_Lf;
            END IF;
         END IF;

      END IF;


      RETURN Result;


   END Export_Aadl_Properties;





   PROCEDURE Duplicate (
         Src  : IN     Tasks_Dependencies_Ptr;
         Dest : IN OUT Tasks_Dependencies_Ptr) IS

   BEGIN
      Duplicate(Half_Dep_Set.Set(Src.All), Half_Dep_Set.Set(Dest.All));
      Duplicate(Src.Dependent_Tasks, Dest.Dependent_Tasks);
   END Duplicate;



   PROCEDURE Reset (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         Free_Object     : IN     Boolean                := False) IS

   BEGIN
      Reset(Half_Dep_Set.Set(My_Dependencies.All), Free_Object);
      Reset(My_Dependencies.Dependent_Tasks, Free_Object);
   END Reset;


   PROCEDURE Put (
         My_Dependencies : IN     Tasks_Dependencies_Ptr) IS

   BEGIN
      Put("Task list/index : ");
      New_Line;

      Put("Dependencies : ");
      New_Line;
      Put(My_Dependencies.All);
   END Put;




   ------------------------------------------------------------------------




   ---------------------------------------
   --  Services below allow to add
   --  dependencies.
   ---------------------------------------




   -- Add ALL dependencies taken from the definition
   -- of a given buffer
   --
   PROCEDURE Add_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         My_Tasks        : IN     Tasks_Set;
         A_Buffer        : IN     Buffer_Ptr) IS


      Task1 : Task_Ptr;

   BEGIN

      Delete_All_Task_Dependencies(My_Dependencies,
         A_Buffer);

      FOR I IN 0..A_Buffer.Roles.Nb_Entries-1 LOOP
         Task1:=Search_Task(My_Tasks, A_Buffer.Roles.Entries(I).Item);
         IF A_Buffer.Roles.Entries(I).Data.The_Role=Producer
               THEN
            Add_One_Task_Dependency(My_Dependencies,
               Task1,
               A_Buffer,
               From_Task_To_Object);
         ELSE
            Add_One_Task_Dependency(My_Dependencies,
               Task1,
               A_Buffer,
               From_Object_To_Task);
         END IF;
      END LOOP;

   END Add_All_Task_Dependencies;


   PROCEDURE Add_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Task_Ptr;
         A_Dep           : IN     Message_Ptr;
         A_Type          : IN     Orientation_Dependency_Type) IS

      A_Dep_Ptr : Dependency_Ptr;
   BEGIN
      A_Dep_Ptr := NEW Dependency(Communication_Dependency);
      A_Dep_Ptr.Communication_Dependent_Task:=A_Task;
      A_Dep_Ptr.Communication_Orientation:=A_Type;
      A_Dep_Ptr.Communication_Dependency_Object:=A_Dep;

      Add(My_Dependencies.All, A_Dep_Ptr);

      Add_Dependent_Tasks(My_Dependencies, A_Task);
   END  Add_One_Task_Dependency;



   PROCEDURE Add_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Task_Ptr;
         A_Dep           : IN     Buffer_Ptr;
         A_Type          : IN     Orientation_Dependency_Type) IS

      A_Dep_Ptr : Dependency_Ptr;
   BEGIN
      A_Dep_Ptr := NEW Dependency(Buffer_Dependency);
      A_Dep_Ptr.Buffer_Dependent_Task:=A_Task;
      A_Dep_Ptr.Buffer_Orientation:=A_Type;
      A_Dep_Ptr.Buffer_Dependency_Object:=A_Dep;

      Add(My_Dependencies.All, A_Dep_Ptr);

      Add_Dependent_Tasks(My_Dependencies, A_Task);
   END  Add_One_Task_Dependency;


   PROCEDURE Add_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         From_Task       : IN     Task_Ptr;
         To_Task         : IN     Task_Ptr) IS

      A_Dep_Ptr : Dependency_Ptr;
   BEGIN
      A_Dep_Ptr := NEW Dependency(Precedence_Dependency);
      A_Dep_Ptr.Precedency_From_Task:=From_Task;
      A_Dep_Ptr.Precedency_To_Task:=To_Task;

      Add(My_Dependencies.All, A_Dep_Ptr);

      Add_Dependent_Tasks(My_Dependencies, From_Task);
      Add_Dependent_Tasks(My_Dependencies, To_Task);
   END  Add_One_Task_Dependency;



   ------------------------------------------------------------------


   FUNCTION Has_Buffer_To_Read (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Task_Ptr)
     RETURN Boolean IS

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Buffer_Dependency THEN
               IF (A_Half_Dep.Buffer_Dependent_Task.Name = My_Task.
                     Name)
                     AND (A_Half_Dep.Buffer_Orientation=
                     From_Object_To_Task)
                     THEN
                  RETURN True;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

      RETURN False;

   END Has_Buffer_To_Read;


   FUNCTION Has_Buffer_To_Write (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Task_Ptr)
     RETURN Boolean IS

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Buffer_Dependency THEN
               IF (A_Half_Dep.Buffer_Dependent_Task .Name = My_Task.
                     Name)
                     AND (A_Half_Dep.Buffer_Orientation=
                     From_Task_To_Object)
                     THEN
                  RETURN True;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

      RETURN False;

   END Has_Buffer_To_Write;



   FUNCTION Get_A_Successor (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Task_Ptr)
     RETURN Task_Ptr IS


      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_From_Task.Name = My_Task.Name)
                     THEN
                  RETURN Copy(A_Half_Dep.Precedency_To_Task);
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All, My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

      RAISE Dependency_Not_Found;
      RETURN NULL;

   END Get_A_Successor;



   FUNCTION Get_A_Predecessor (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Task_Ptr)
     RETURN Task_Ptr IS

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_To_Task.Name = My_Task.Name)
                     THEN
                  RETURN Copy(A_Half_Dep.Precedency_From_Task);
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All, My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

      RAISE Dependency_Not_Found;
      RETURN NULL;

   END Get_A_Predecessor;




   FUNCTION Has_Predecessor (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Task_Ptr)
     RETURN Boolean IS
      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_To_Task.Name = My_Task.
                     Name)
                     THEN
                  RETURN True;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

      RETURN False;
   END Has_Predecessor;




   FUNCTION Has_Successor (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Task_Ptr)
     RETURN Boolean IS


      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_From_Task.Name = My_Task.
                     Name)
                     THEN
                  RETURN True;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

      RETURN False;
   END Has_Successor;



   FUNCTION Get_Successors_List (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Task_Ptr)
     RETURN Tasks_Set IS

      Result   : Tasks_Set;
      New_Task : Task_Ptr;

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_From_Task.Name =
                     My_Task.
                     Name)
                     THEN
                  New_Task:=Copy(A_Half_Dep.
                     Precedency_To_Task);
                  Add(Result, New_Task);
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

      RETURN Result;
   END Get_Successors_List;



   FUNCTION Get_Predecessors_List (
         My_Dependencies : IN     Tasks_Dependencies_Ptr;
         My_Task         : IN     Task_Ptr)
     RETURN Tasks_Set IS

      Result   : Tasks_Set;
      New_Task : Task_Ptr;

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_To_Task.Name =
                     My_Task.
                     Name)
                     THEN
                  New_Task:=Copy(A_Half_Dep.
                     Precedency_From_Task);
                  Add(Result, New_Task);
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

      RETURN Result;
   END Get_Predecessors_List;





   FUNCTION Get_A_Root_Task (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Task_Ptr IS

      Is_Root : Boolean;

      My_Iterator2 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;
      My_Iterator1 : Tasks_Iterator;
      A_Task       : Task_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.Dependent_Tasks) THEN

         Reset_Iterator(My_Dependencies.Dependent_Tasks, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.
               Dependent_Tasks,
               A_Task,
               My_Iterator1);

            Reset_Iterator(My_Dependencies.All, My_Iterator2);

            Is_Root:=True;
            LOOP
               Current_Element(My_Dependencies.All,
                  A_Half_Dep,
                  My_Iterator2);


               IF A_Half_Dep.Depend_Type=
                     Buffer_Dependency THEN
                  IF (A_Half_Dep.Buffer_Orientation =
                        From_Object_To_Task)
                        AND (A_Task.Name = A_Half_Dep.
                        Buffer_Dependent_Task.Name)
                        THEN
                     Is_Root:=False;
                  END IF;
               END IF;

               IF A_Half_Dep.Depend_Type=
                     Communication_Dependency THEN
                  IF (A_Half_Dep.
                        Communication_Orientation =
                        From_Object_To_Task)
                        AND (A_Task.Name = A_Half_Dep.
                        Communication_Dependent_Task.
                        Name)
                        THEN
                     Is_Root:=False;
                  END IF;
               END IF;

               IF A_Half_Dep.Depend_Type=
                     Precedence_Dependency THEN
                  IF (A_Half_Dep.Precedency_To_Task.
                        Name =
                        A_Task.
                        Name)
                        THEN
                     Is_Root:=False;
                  END IF;
               END IF;

               EXIT WHEN Is_Last_Element(
                  My_Dependencies.All,
                  My_Iterator2);

               Next_Element(My_Dependencies.All,
                  My_Iterator2);
            END LOOP;


            IF Is_Root THEN
               RETURN A_Task;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.
               Dependent_Tasks,
               My_Iterator1);

            Next_Element(My_Dependencies.
               Dependent_Tasks,
               My_Iterator1);

         END LOOP;
      END IF;


      RAISE Dependency_Not_Found;
      RETURN NULL;
   END Get_A_Root_Task;



   FUNCTION Get_A_Leaf_Task (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Task_Ptr IS

      Is_Leaf : Boolean;

      My_Iterator2 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;
      My_Iterator1 : Tasks_Iterator;
      A_Task       : Task_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.Dependent_Tasks) THEN

         Reset_Iterator(My_Dependencies.Dependent_Tasks, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.
               Dependent_Tasks,
               A_Task,
               My_Iterator1);

            Reset_Iterator(My_Dependencies.All, My_Iterator2);

            Is_Leaf:=True;
            LOOP
               Current_Element(My_Dependencies.All,
                  A_Half_Dep,
                  My_Iterator2);


               IF A_Half_Dep.Depend_Type=
                     Buffer_Dependency THEN
                  IF (A_Half_Dep.Buffer_Orientation =
                        From_Task_To_Object)
                        AND (A_Task.Name = A_Half_Dep.
                        Buffer_Dependent_Task.Name)
                        THEN
                     Is_Leaf:=False;
                  END IF;
               END IF;

               IF A_Half_Dep.Depend_Type=
                     Communication_Dependency THEN
                  IF (A_Half_Dep.
                        Communication_Orientation =
                        From_Task_To_Object)
                        AND (A_Task.Name = A_Half_Dep.
                        Communication_Dependent_Task.
                        Name)
                        THEN
                     Is_Leaf:=False;
                  END IF;
               END IF;

               IF A_Half_Dep.Depend_Type=
                     Precedence_Dependency THEN
                  IF (A_Half_Dep.Precedency_From_Task.
                        Name =
                        A_Task.
                        Name)
                        THEN
                     Is_Leaf:=False;
                  END IF;
               END IF;

               EXIT WHEN Is_Last_Element(
                  My_Dependencies.All,
                  My_Iterator2);

               Next_Element(My_Dependencies.All,
                  My_Iterator2);
            END LOOP;


            IF Is_Leaf THEN
               RETURN A_Task;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.
               Dependent_Tasks,
               My_Iterator1);

            Next_Element(My_Dependencies.
               Dependent_Tasks,
               My_Iterator1);

         END LOOP;
      END IF;


      RAISE Dependency_Not_Found;
      RETURN NULL;
   END Get_A_Leaf_Task;




   FUNCTION Get_Leaf_Tasks (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Tasks_Set IS

      Result   : Tasks_Set;
      Is_Leaf  : Boolean;
      New_Task : Task_Ptr;

      My_Iterator2 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;
      My_Iterator1 : Tasks_Iterator;
      A_Task       : Task_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.Dependent_Tasks) THEN

         Reset_Iterator(My_Dependencies.Dependent_Tasks, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.
               Dependent_Tasks,
               A_Task,
               My_Iterator1);


            Reset_Iterator(My_Dependencies.All, My_Iterator2);
            Is_Leaf:=True;
            LOOP
               Current_Element(My_Dependencies.All,
                  A_Half_Dep,
                  My_Iterator2);


               IF A_Half_Dep.Depend_Type=
                     Buffer_Dependency THEN
                  IF (A_Half_Dep.Buffer_Orientation =
                        From_Task_To_Object)
                        AND (A_Task.Name = A_Half_Dep.
                        Buffer_Dependent_Task.Name)
                        THEN
                     Is_Leaf:=False;
                  END IF;
               END IF;

               IF A_Half_Dep.Depend_Type=
                     Communication_Dependency THEN
                  IF (A_Half_Dep.
                        Communication_Orientation =
                        From_Task_To_Object)
                        AND (A_Task.Name = A_Half_Dep.
                        Communication_Dependent_Task.
                        Name)
                        THEN
                     Is_Leaf:=False;
                  END IF;
               END IF;

               IF A_Half_Dep.Depend_Type=
                     Precedence_Dependency THEN
                  IF (A_Half_Dep.Precedency_From_Task.
                        Name =
                        A_Task.
                        Name)
                        THEN
                     Is_Leaf:=False;
                  END IF;
               END IF;

               EXIT WHEN Is_Last_Element(
                  My_Dependencies.All,
                  My_Iterator2);

               Next_Element(My_Dependencies.All,
                  My_Iterator2);
            END LOOP;


            IF Is_Leaf THEN
               New_Task:=Copy(A_Task);
               Add(Result, New_Task);
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.
               Dependent_Tasks,
               My_Iterator1);

            Next_Element(My_Dependencies.
               Dependent_Tasks,
               My_Iterator1);

         END LOOP;
      END IF;
      RETURN Result;
   END Get_Leaf_Tasks;





   FUNCTION Get_Root_Tasks (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Tasks_Set IS

      Result   : Tasks_Set;
      Is_Root  : Boolean;
      New_Task : Task_Ptr;

      My_Iterator2 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;
      My_Iterator1 : Tasks_Iterator;
      A_Task       : Task_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.Dependent_Tasks) THEN

         Reset_Iterator(My_Dependencies.Dependent_Tasks, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.
               Dependent_Tasks,
               A_Task,
               My_Iterator1);

            Reset_Iterator(My_Dependencies.All, My_Iterator2);

            Is_Root:=True;
            LOOP
               Current_Element(My_Dependencies.All,
                  A_Half_Dep,
                  My_Iterator2);


               IF A_Half_Dep.Depend_Type=
                     Buffer_Dependency THEN
                  IF (A_Half_Dep.Buffer_Orientation =
                        From_Object_To_Task)
                        AND (A_Task.Name = A_Half_Dep.
                        Buffer_Dependent_Task.Name)
                        THEN
                     Is_Root:=False;
                  END IF;
               END IF;

               IF A_Half_Dep.Depend_Type=
                     Communication_Dependency THEN
                  IF (A_Half_Dep.
                        Communication_Orientation =
                        From_Object_To_Task)
                        AND (A_Task.Name = A_Half_Dep.
                        Communication_Dependent_Task.
                        Name)
                        THEN
                     Is_Root:=False;
                  END IF;
               END IF;

               IF A_Half_Dep.Depend_Type=
                     Precedence_Dependency THEN
                  IF (A_Half_Dep.Precedency_To_Task.
                        Name =
                        A_Task.
                        Name)
                        THEN
                     Is_Root:=False;
                  END IF;
               END IF;

               EXIT WHEN Is_Last_Element(
                  My_Dependencies.All,
                  My_Iterator2);

               Next_Element(My_Dependencies.All,
                  My_Iterator2);
            END LOOP;


            IF Is_Root THEN
               New_Task:=Copy(A_Task);
               Add(Result, New_Task);
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.
               Dependent_Tasks,
               My_Iterator1);

            Next_Element(My_Dependencies.
               Dependent_Tasks,
               My_Iterator1);

         END LOOP;
      END IF;
      RETURN Result;

   END Get_Root_Tasks;




   -------------------------------------------------------------------


   -- Delete ALL dependencies  of an
   -- address space (parameter "An_Address_Space")
   --
   PROCEDURE Delete_Address_Space_All_Task_Dependencies (
         My_Dependencies  : IN OUT Tasks_Dependencies_Ptr;
         An_Address_Space : IN     Unbounded_String) IS

      My_Iterator1  : Tasks_Dependencies_Iterator;
      A_Half_Dep    : Dependency_Ptr;
      To_Be_Deleted : Tasks_Dependencies_Ptr      :=
      NEW Tasks_Dependencies;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_From_Task.Address_Space_Name =
                     An_Address_Space)
                     OR (
                     A_Half_Dep.Precedency_To_Task.Address_Space_Name =
                     An_Address_Space)
                     THEN
                  Add(To_Be_Deleted.All, A_Half_Dep);
               END IF;
            END IF;

            IF A_Half_Dep.Depend_Type=
                  Buffer_Dependency THEN
               IF (A_Half_Dep.Buffer_Orientation =
                     From_Object_To_Task)
                     AND (An_Address_Space = A_Half_Dep.
                     Buffer_Dependent_Task.Address_Space_Name)
                     THEN
                  Add(To_Be_Deleted.All, A_Half_Dep);
               END IF;
            END IF;

            IF A_Half_Dep.Depend_Type=
                  Communication_Dependency THEN
               IF (A_Half_Dep.
                     Communication_Orientation =
                     From_Object_To_Task)
                     AND (An_Address_Space = A_Half_Dep.
                     Communication_Dependent_Task.
                     Address_Space_Name)
                     THEN
                  Add(To_Be_Deleted.All, A_Half_Dep);
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;


      IF NOT Is_Empty(To_Be_Deleted.All)
            THEN
         Delete(My_Dependencies.All, To_Be_Deleted.All);
         Free(Half_Dep_Set.Set_Ptr(To_Be_Deleted), False);
      END IF;

   END Delete_Address_Space_All_Task_Dependencies;



   -- Delete ALL dependencies  of a
   -- processor (parameter "A_Processor")
   --
   PROCEDURE Delete_Processor_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Processor     : IN     Unbounded_String) IS

      My_Iterator1  : Tasks_Dependencies_Iterator;
      A_Half_Dep    : Dependency_Ptr;
      To_Be_Deleted : Tasks_Dependencies_Ptr      :=
      NEW Tasks_Dependencies;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_From_Task.Cpu_Name =
                     A_Processor)
                     OR (A_Half_Dep.Precedency_To_Task.Cpu_Name =
                     A_Processor)
                     THEN
                  Add(To_Be_Deleted.All, A_Half_Dep);
               END IF;
            END IF;

            IF A_Half_Dep.Depend_Type=
                  Buffer_Dependency THEN
               IF (A_Half_Dep.Buffer_Orientation =
                     From_Object_To_Task)
                     AND (A_Processor = A_Half_Dep.
                     Buffer_Dependent_Task.Cpu_Name)
                     THEN
                  Add(To_Be_Deleted.All, A_Half_Dep);
               END IF;
            END IF;

            IF A_Half_Dep.Depend_Type=
                  Communication_Dependency THEN
               IF (A_Half_Dep.
                     Communication_Orientation =
                     From_Object_To_Task)
                     AND (A_Processor = A_Half_Dep.
                     Communication_Dependent_Task.
                     Cpu_Name)
                     THEN
                  Add(To_Be_Deleted.All, A_Half_Dep);
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;


      IF NOT Is_Empty(To_Be_Deleted.All)
            THEN
         Delete(My_Dependencies.All, To_Be_Deleted.All);
         Free(Half_Dep_Set.Set_Ptr(To_Be_Deleted), False);
      END IF;

   END Delete_Processor_All_Task_Dependencies;



   -- Delete ALL dependencies :
   --  1) of a task (if My_Dep_Name is a Task_ptr)
   --  2) of a message  (if My_Dep_Name is a Message_ptr)
   --  3) of a buffer  (if My_Dep_Name is a Buffer_ptr)
   --
   PROCEDURE Delete_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Task_Ptr) IS

      My_Iterator1  : Tasks_Dependencies_Iterator;
      A_Half_Dep    : Dependency_Ptr;
      To_Be_Deleted : Tasks_Dependencies_Ptr      :=
      NEW Tasks_Dependencies;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Buffer_Dependency THEN
               IF (A_Half_Dep.Buffer_Dependent_Task.Name = A_Task.
                     Name)
                     THEN
                  Add(To_Be_Deleted.All, A_Half_Dep);
               END IF;
            END IF;
            IF A_Half_Dep.Depend_Type=
                  Communication_Dependency THEN
               IF (A_Half_Dep.Communication_Dependent_Task.Name =
                     A_Task.
                     Name)
                     THEN
                  Add(To_Be_Deleted.All, A_Half_Dep);
               END IF;
            END IF;
            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_From_Task.Name = A_Task.
                     Name)
                     OR (A_Half_Dep.Precedency_To_Task.Name = A_Task.
                     Name)
                     THEN
                  Add(To_Be_Deleted.All, A_Half_Dep);
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;


      IF NOT Is_Empty(To_Be_Deleted.All)
            THEN
         Delete(My_Dependencies.All, To_Be_Deleted.All);
         Free(Half_Dep_Set.Set_Ptr(To_Be_Deleted), False);
      END IF;

   END Delete_All_Task_Dependencies;



   PROCEDURE Delete_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Buffer        : IN     Buffer_Ptr) IS

      My_Iterator1  : Tasks_Dependencies_Iterator;
      A_Half_Dep    : Dependency_Ptr;
      To_Be_Deleted : Tasks_Dependencies_Ptr      :=
      NEW Tasks_Dependencies;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Buffer_Dependency THEN
               IF (A_Half_Dep.Buffer_Dependency_Object.Name = A_Buffer.
                     Name)
                     THEN
                  Add(To_Be_Deleted.All, A_Half_Dep);
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;



      IF NOT Is_Empty(To_Be_Deleted.All)
            THEN
         Delete(My_Dependencies.All, To_Be_Deleted.All);
         Free(Half_Dep_Set.Set_Ptr(To_Be_Deleted), False);
      END IF;

   END Delete_All_Task_Dependencies;



   PROCEDURE Delete_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Message       : IN     Message_Ptr) IS

      My_Iterator1  : Tasks_Dependencies_Iterator;
      A_Half_Dep    : Dependency_Ptr;
      To_Be_Deleted : Tasks_Dependencies_Ptr      :=
      NEW Tasks_Dependencies;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Communication_Dependency THEN
               IF (A_Half_Dep.Communication_Dependency_Object.Name =
                     A_Message.
                     Name)
                     THEN
                  Add(To_Be_Deleted.All, A_Half_Dep);
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;


      IF NOT Is_Empty(To_Be_Deleted.All)
            THEN
         Delete(My_Dependencies.All, To_Be_Deleted.All);
         Free(Half_Dep_Set.Set_Ptr(To_Be_Deleted), False);
      END IF;

   END Delete_All_Task_Dependencies;



   -- Remove ONE dependency between two tasks
   -- (remove a precedency relationship, not with buffer of message)
   --
   PROCEDURE Delete_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         From_Task       : IN     Task_Ptr;
         To_Task         : IN     Task_Ptr) IS

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_From_Task.Name = From_Task.
                     Name)
                     AND (A_Half_Dep.Precedency_To_Task.Name = To_Task.
                     Name)
                     THEN
                  Delete(My_Dependencies.All, A_Half_Dep);
                  RETURN;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

      RAISE Dependency_Not_Found;
   END Delete_One_Task_Dependency;


   -- Remove ONE dependency between a task and
   --   a message
   --
   PROCEDURE Delete_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Task_Ptr;
         A_Buffer        : IN     Buffer_Ptr;
         A_Type          : IN     Orientation_Dependency_Type) IS

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Buffer_Dependency THEN
               IF (A_Half_Dep.Buffer_Dependent_Task.Name = A_Task.
                     Name)
                     AND (A_Half_Dep.Buffer_Dependency_Object.Name =
                     A_Buffer.
                     Name)
                     AND (A_Half_Dep.Buffer_Orientation  = A_Type)
                     THEN
                  Delete(My_Dependencies.All, A_Half_Dep);
                  RETURN;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

      RAISE Dependency_Not_Found;
   END Delete_One_Task_Dependency;


   -- Remove ONE dependency between a task and
   --   a message
   --
   PROCEDURE Delete_One_Task_Dependency (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Task_Ptr;
         A_Message       : IN     Message_Ptr;
         A_Type          : IN     Orientation_Dependency_Type) IS

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Communication_Dependency THEN
               IF (A_Half_Dep.Communication_Dependent_Task.Name =
                     A_Task.
                     Name)
                     AND (
                     A_Half_Dep.Communication_Dependency_Object.Name =
                     A_Message.
                     Name)
                     AND (A_Half_Dep.Communication_Orientation  =
                     A_Type)
                     THEN
                  Delete(My_Dependencies.All, A_Half_Dep);
                  RETURN;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

      RAISE Dependency_Not_Found;
   END Delete_One_Task_Dependency;




   -------------------------------------------------------------------


   PROCEDURE Update_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Task          : IN     Task_Ptr) IS

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Buffer_Dependency THEN
               IF (A_Half_Dep.Buffer_Dependent_Task.Name = A_Task.
                     Name)
                     THEN
                  A_Half_Dep.Buffer_Dependent_Task:=A_Task;
               END IF;
            END IF;
            IF A_Half_Dep.Depend_Type=
                  Communication_Dependency THEN
               IF (A_Half_Dep.Communication_Dependent_Task.Name =
                     A_Task.
                     Name)
                     THEN
                  A_Half_Dep.Communication_Dependent_Task:=A_Task;
               END IF;
            END IF;
            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_From_Task.Name = A_Task.
                     Name)
                     THEN
                  A_Half_Dep.Precedency_From_Task:=A_Task;
               END IF;
            END IF;
            IF A_Half_Dep.Depend_Type=
                  Precedence_Dependency THEN
               IF (A_Half_Dep.Precedency_To_Task.Name = A_Task.
                     Name)
                     THEN
                  A_Half_Dep.Precedency_To_Task:=A_Task;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

   END Update_All_Task_Dependencies;


   PROCEDURE Update_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Buffer        : IN     Buffer_Ptr) IS

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Buffer_Dependency THEN
               IF (A_Half_Dep.Buffer_Dependency_Object.Name = A_Buffer.
                     Name)
                     THEN
                  A_Half_Dep.Buffer_Dependency_Object:=A_Buffer;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;

   END Update_All_Task_Dependencies;



   PROCEDURE Update_All_Task_Dependencies (
         My_Dependencies : IN OUT Tasks_Dependencies_Ptr;
         A_Message       : IN     Message_Ptr) IS

      My_Iterator1 : Tasks_Dependencies_Iterator;
      A_Half_Dep   : Dependency_Ptr;

   BEGIN

      IF NOT Is_Empty(My_Dependencies.All) THEN

         Reset_Iterator(My_Dependencies.All, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.All, A_Half_Dep,
               My_Iterator1);

            IF A_Half_Dep.Depend_Type=
                  Communication_Dependency THEN
               IF (A_Half_Dep.Communication_Dependency_Object.Name =
                     A_Message.
                     Name)
                     THEN
                  A_Half_Dep.Communication_Dependency_Object:=
                     A_Message;
               END IF;
            END IF;

            EXIT WHEN Is_Last_Element(My_Dependencies.All,
               My_Iterator1);

            Next_Element(My_Dependencies.All, My_Iterator1);
         END LOOP;
      END IF;


   END Update_All_Task_Dependencies;




   -------------------------------------------------------------------


   -- Check is a task dependencies is acyclic or not
   --
   FUNCTION Is_Cyclic (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN NOT Is_Acyclic(My_Dependencies);
   END Is_Cyclic;




   FUNCTION Is_Acyclic (
         My_Dependencies : IN     Tasks_Dependencies_Ptr)
     RETURN Boolean IS

      TYPE Tasks_Table_Range IS NEW Integer RANGE 0 .. Framework_Config.Max_Tasks_Dependencies;
      TYPE Tasks_Table IS ARRAY (Tasks_Table_Range'RANGE) OF Task_Ptr;


      My_Iterator1 : Tasks_Iterator;
      A_Task       : Task_Ptr;


      -- Stay true until we find a cycle
      --
      Result : Boolean := True;


      --  Check backward arrows (cycle detection)
      --
      FUNCTION Has_Backward_Arrow (
            Current,
            Next_Of_Current  : Task_Ptr;
            Number_Of_Arrows : Tasks_Table_Range;
            From,
            To               : Tasks_Table)
        RETURN Boolean IS
      BEGIN
         FOR K IN 0..Number_Of_Arrows-1 LOOP
            IF (Current.Name=From(K).Name)
                  AND (Next_Of_Current.Name=To(K).Name)
                  THEN
               RETURN True;
            END IF;
         END LOOP;
         RETURN False;
      END Has_Backward_Arrow;


      -- Each task of the set has to be inspected
      --
      PROCEDURE Scan_A_Task (
            Current          :        Task_Ptr;
            Number_Of_Arrows : IN     Tasks_Table_Range;
            From,
            To               : IN     Tasks_Table) IS


         Next            : Tasks_Set;
         Ite             : Tasks_Iterator;
         Next_Of_Current : Task_Ptr;

         Tmp_To : Tasks_Table := To;

         Tmp_From   :          Tasks_Table       := From;
         Tmp_Number : CONSTANT Tasks_Table_Range := Number_Of_Arrows + 1;

      BEGIN

         IF Has_Successor(My_Dependencies, Current)
               THEN

            Reset(Next, False);
            Next:=Get_Successors_List(My_Dependencies,
               Current);

            -- Run for each task a cyclicity test
            --
            Reset_Iterator(Next, Ite);
            FOR J IN 0 .. Get_Number_Of_Elements(Next)-
                  1 LOOP
               Current_Element(Next, Next_Of_Current,
                  Ite);

               -- Memorize that we had check this arrow
               --
               IF Has_Backward_Arrow(Current,
                     Next_Of_Current, Number_Of_Arrows, From, To)
                     THEN
                  Result:=False;
               ELSE
                  Tmp_From(Tmp_Number-1):=Current;
                  Tmp_To(Tmp_Number-1):=Next_Of_Current;
                  Scan_A_Task(Next_Of_Current, Tmp_Number, Tmp_From,
                     Tmp_To);
               END IF;
               Next_Element(Next, Ite);
            END LOOP;
         END IF;

      END Scan_A_Task;


      -- Used to store backward arrows
      --
      From,
      To   : Tasks_Table;
      M    : Tasks_Table_Range := 0;

   BEGIN

      -- Run for Each task A Cyclicity Test
      --
      IF NOT Is_Empty(My_Dependencies.Dependent_Tasks) THEN

         Reset_Iterator(My_Dependencies.Dependent_Tasks, My_Iterator1);

         LOOP
            Current_Element(My_Dependencies.
               Dependent_Tasks,
               A_Task,
               My_Iterator1);

            IF Result
                  THEN
               M:=0;
               Scan_A_Task(A_Task, M, From, To);
            END IF;


            EXIT WHEN Is_Last_Element(My_Dependencies.
               Dependent_Tasks,
               My_Iterator1);

            Next_Element(My_Dependencies.
               Dependent_Tasks,
               My_Iterator1);

         END LOOP;
      END IF;

      RETURN Result;

   END Is_Acyclic;




END Task_Dependencies;

