------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------





WITH Ada.Exceptions;
USE Ada.Exceptions;


WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Tasks;
USE Tasks;
WITH Resources;
USE Resources;
USE Resources.Task_Lists;

WITH Time_Units;
USE  Time_Units;
USE  Time_Units.A_Time_Unit;

WITH Parameters;
USE Parameters;

WITH Natural_Util;
USE Natural_Util;
WITH Double_Util;
USE Double_Util;
WITH Integer_Util;
USE Integer_Util;


WITH Ada.Numerics.Elementary_Functions;
USE Ada.Numerics.Elementary_Functions;

WITH Objects;
USE Objects;

WITH Text_Io;
USE Text_Io;

WITH Translate;
USE Translate;


WITH Big_Number;



PACKAGE BODY Task_Set IS


   PROCEDURE Check_Task_Critical_Section (
         My_Tasks          : IN     Tasks_Set;
         Resource_Name     : IN     Unbounded_String;
         Resource_Cpu_Name : IN     Unbounded_String;
         Task_Name         : IN     Unbounded_String;
         Share_Begin       : IN     Integer;
         Share_End         : IN     Integer) IS

      A_Task : Task_Ptr;

   BEGIN

      A_Task:=Search_Task(My_Tasks, Task_Name);


      --
      -- Check if begin and end time are consistent
      --
      IF (Share_Begin<=0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(
               Resource_Name & " ; " & Task_Name & " ; " &
               Lb_Task_Begin(Current_Language) & Lb_Must_Be(
                  Current_Language)
               & To_Unbounded_String(" >0 ")
               ));
      END IF;

      IF (Share_End<=0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(
               Resource_Name & " ; " & Task_Name & " ; " &
               Lb_Task_End(Current_Language) & Lb_Must_Be(
                  Current_Language)
               & To_Unbounded_String(" >0 ")
               ));
      END IF;

      IF Share_Begin > Share_End
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(
               Resource_Name & " ; " & Task_Name & " ; " &
               Lb_Task_Begin(Current_Language) & Lb_Must_Be(
                  Current_Language)
               & To_Unbounded_String(" <= ")
               & Lb_Task_End(Current_Language) ));
      END IF;



      -- Check that the critical section is consistent    according to
      -- the task definition
      --
      IF (Natural(Share_End)>A_Task.Capacity)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Resource_Name & " ; " & Task_Name & " ; " &
               Lb_Task_End(Current_Language) &
               Lb_Must_Be(Current_Language)
               & To_Unbounded_String(" <= ")
               & Lb_Capacity(Current_Language)));
      END IF;

      IF (Natural(Share_Begin)>A_Task.Capacity)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Resource_Name & " ; " & Task_Name & " ; " &
               Lb_Task_Begin(Current_Language) &
               Lb_Must_Be(Current_Language)
               & To_Unbounded_String(" <= ")
               & Lb_Capacity(Current_Language)));
      END IF;


      -- Check that the task is on the same processor that the
      -- resource
      --
      IF A_Task.Cpu_Name/=Resource_Cpu_Name
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(
               Resource_Name & " ; " & Task_Name & " ; " &
               Lb_Have_To_Be_On_The_Same_Processor(Current_Language)
               ));
      END IF;

   END  Check_Task_Critical_Section;


   PROCEDURE Check_Buffer_Roles (
         My_Tasks    : IN     Tasks_Set;
         Buffer_Name : IN     Unbounded_String;
         Task_Name   : IN     Unbounded_String;
         Time        : IN     Integer;
         Size        : IN     Integer) IS

      A_Task : Task_Ptr;

   BEGIN

      IF (Size<=0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(
               Buffer_Name & " ; " & Task_Name & " ; " &
               Lb_Size(Current_Language) & Lb_Must_Be(
                  Current_Language)
               & To_Unbounded_String(" >0 ")
               ));
      END IF;


      IF (Time<=0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(
               Buffer_Name & " ; " & Task_Name & " ; " &
               Lb_Time(Current_Language) & Lb_Must_Be(
                  Current_Language)
               & To_Unbounded_String(" >0 ")
               ));
      END IF;

      -- Check that capacity is <= time
      --
      A_Task:=Search_Task(My_Tasks, Task_Name);
      IF A_Task.Capacity < Time
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Buffer_Name & " ; " & Task_Name & " ; " &
               Lb_Time (Current_Language) & Lb_Must_Be(
                  Current_Language) & To_Unbounded_String(" <= ") &
               Lb_Capacity(
                  Current_Language)));
      END IF;


   END  Check_Buffer_Roles;




   PROCEDURE Add_Task (
         My_Tasks             : IN OUT Tasks_Set;
         Name                 : IN     Unbounded_String;
         Cpu_Name             : IN     Unbounded_String;
         Address_Space_Name   : IN     Unbounded_String;
         Task_Type            : IN     Tasks_Type;
         Start_Time           : IN     Integer;
         Capacity             : IN     Integer;
         Period               : IN     Integer;
         Deadline             : IN     Integer;
         Jitter               : IN     Integer;
         Blocking_Time        : IN     Integer;
         Priority             : IN     Integer;
         Criticality          : IN     Integer;
         Policy               : IN     Policies;
         Offset               :        Offsets_Table;
         Stack_Memory_Size    : IN     Integer                       := 0;
         Text_Memory_Size     : IN     Integer                       := 0;
         Param                : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Parametric_Rule_Name : IN     Unbounded_String              := Empty_String;
         Seed_Value           : IN     Integer                       := 0;
         Predictable          : IN     Boolean                       := True) IS


      Dummy : Task_Ptr;

   BEGIN
      Add_Task(My_Tasks, Dummy, Name, Cpu_Name, Address_Space_Name,
         Task_Type, Start_Time,
         Capacity, Period, Deadline, Jitter, Blocking_Time,
         Priority, Criticality, Policy, Offset, Stack_Memory_Size,
         Text_Memory_Size,
         Param, Parametric_Rule_Name,
         Seed_Value, Predictable);
   END Add_Task;



   PROCEDURE Check_Task (
         My_Tasks             : IN     Tasks_Set;
         Name                 : IN     Unbounded_String;
         Cpu_Name             : IN     Unbounded_String;
         Address_Space_Name   : IN     Unbounded_String;
         Task_Type            : IN     Tasks_Type;
         Start_Time           : IN     Integer;
         Capacity             : IN     Integer;
         Period               : IN     Integer;
         Deadline             : IN     Integer;
         Jitter               : IN     Integer;
         Blocking_Time        : IN     Integer;
         Priority             : IN     Integer;
         Criticality          : IN     Integer;
         Policy               : IN     Policies;
         Offset               :        Offsets_Table;
         Stack_Memory_Size    : IN     Integer                       := 0;
         Text_Memory_Size     : IN     Integer                       := 0;
         Param                : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Parametric_Rule_Name : IN     Unbounded_String              := Empty_String;
         Seed_Value           : IN     Integer                       := 0;
         Predictable          : IN     Boolean                       := True) IS



   BEGIN



      IF Name=""
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;

      IF NOT Is_A_Valid_Identifier(Name)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Task_Name(Current_Language) &
               Lb_Colon & Lb_Invalid_Identifier(Current_Language)));
      END IF;


      IF NOT Is_A_Valid_Identifier(Parametric_Rule_Name)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Activation_Rule(Current_Language) &
               Lb_Colon & Lb_Invalid_Identifier(Current_Language)));
      END IF;


      IF (Cpu_Name = "")
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Processor_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;

      IF (Address_Space_Name = "")
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Address_Space_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      END IF;


      CASE Task_Type IS
         WHEN Periodic_Type =>
            IF Period<=0 THEN
               Raise_Exception(Invalid_Parameter'Identity,
                  To_String(Lb_Task(Current_Language) & " " & Name &
                     " : " & Lb_Periodic(Current_Language) &
                     Lb_Require_Period(Current_Language)));
            END IF;

         WHEN Aperiodic_Type =>
            IF Period/=0 THEN
               Raise_Exception(Invalid_Parameter'Identity,
                  To_String(Lb_Task(Current_Language) & " " & Name &
                     " : " & Lb_Aperiodic(Current_Language) &
                     Lb_No_Period(Current_Language)));
            END IF;

         WHEN Poisson_Type =>
            IF Period<=0 THEN
               Raise_Exception(Invalid_Parameter'Identity,
                  To_String(Lb_Task(Current_Language) & " " & Name &
                     " : " & Lb_Poisson_Process_Task(Current_Language) &
                     Lb_Require_Period(Current_Language)));
            END IF;

         WHEN Sporadic_Type =>
            IF Period<=0 THEN
               Raise_Exception(Invalid_Parameter'Identity,
                  To_String(Lb_Task(Current_Language) & " " & Name &
                     " : " & Lb_Sporadic_Task(Current_Language) &
                     Lb_Require_Period(Current_Language)));
            END IF;

         WHEN Parametric_Type =>
            IF Parametric_Rule_Name = Empty_String THEN
               Raise_Exception(Invalid_Parameter'Identity,
                  To_String(Lb_Task(Current_Language) & " " & Name &
                     " : " & Lb_Parametric_Task(Current_Language) &
                     Lb_Require_Activation_Rule(Current_Language)));
            END IF;

      END CASE;


      IF (Capacity<=0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Capacity(Current_Language) & Lb_Must_Be(
                  Current_Language) & ">0"));
      END IF;

      IF (Jitter<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Jitter(Current_Language) & Lb_Must_Be(
                  Current_Language) & "=>0"));
      END IF;

      IF (Criticality<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Criticality(Current_Language) & Lb_Must_Be(
                  Current_Language) & "=>0"));
      END IF;


      IF (Period<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Period(Current_Language) & Lb_Must_Be(
                  Current_Language) & "=>0"));
      END IF;

      IF (Deadline<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Deadline(Current_Language) & Lb_Must_Be(
                  Current_Language) & ">=0"));
      END IF;

      IF (Deadline<Jitter)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Deadline(Current_Language) & Lb_Must_Be(
                  Current_Language) & Lb_Superior(Current_Language) &
               Lb_Jitter(Current_Language)));
      END IF;

      IF (Start_Time<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Start_Time(Current_Language) & Lb_Must_Be(
                  Current_Language) & ">=0"));
      END IF;

      IF (Blocking_Time<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Blocking_Time(Current_Language) & Lb_Must_Be(
                  Current_Language) & ">=0"));
      END IF;

      IF (Text_Memory_Size<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Text_Memory_Size(Current_Language) &
               Lb_Must_Be(
                  Current_Language) & ">=0"));
      END IF;

      IF (Stack_Memory_Size<0)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Stack_Memory_Size(Current_Language) &
               Lb_Must_Be(Current_Language) & ">=0"));
      END IF;

      IF (Priority<Integer(Priority_Range'First))
            OR (Priority>Integer(Priority_Range'Last))
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Invalid_Priority(Current_Language)));
      END IF;

      IF (Priority /= 0) AND
            (Policy=Sched_Others)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Policy_Control(Current_Language)));
      END IF;

      IF (Priority = 0) AND
            (Policy/=Sched_Others)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Policy_Control(Current_Language)));
      END IF;

      IF (To_String(Parametric_Rule_Name) /= ""
            AND Task_Type /= Parametric_Type)
            THEN
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Task(Current_Language) & " " & Name &
               " : " & Lb_Rule_For_Parametric_Only(Current_Language)));
      END IF;



      -- Offset integrity checks
      --

      FOR I IN 0..Offset.Nb_Entries-1 LOOP

         IF Offset.Entries(I).Value<0
               THEN
            Raise_Exception(Invalid_Parameter'Identity,
               To_String(Lb_Task(Current_Language) & " " & Name &
                  " : " & To_Unbounded_String("Offset/val.") &
                  Lb_Must_Be(Current_Language)
                  & To_Unbounded_String(" >= 0")));
         END IF;

         IF Offset.Entries(I).Activation<0
               THEN
            Raise_Exception(Invalid_Parameter'Identity,
               To_String(Lb_Task(Current_Language) & " " & Name &
                  " : " & To_Unbounded_String("Offset/activation") &
                  Lb_Must_Be(Current_Language)
                  & To_Unbounded_String(" >= 0")));

         END IF;

      END LOOP;


   END Check_Task;


   PROCEDURE Add_Task (
         My_Tasks             : IN OUT Tasks_Set;
         A_Task               : IN OUT Task_Ptr;
         Name                 : IN     Unbounded_String;
         Cpu_Name             : IN     Unbounded_String;
         Address_Space_Name   : IN     Unbounded_String;
         Task_Type            : IN     Tasks_Type;
         Start_Time           : IN     Integer;
         Capacity             : IN     Integer;
         Period               : IN     Integer;
         Deadline             : IN     Integer;
         Jitter               : IN     Integer;
         Blocking_Time        : IN     Integer;
         Priority             : IN     Integer;
         Criticality          : IN     Integer;
         Policy               : IN     Policies;
         Offset               :        Offsets_Table;
         Stack_Memory_Size    : IN     Integer                       := 0;
         Text_Memory_Size     : IN     Integer                       := 0;
         Param                : IN     User_Defined_Parameters_Table := No_User_Defined_Parameter;
         Parametric_Rule_Name : IN     Unbounded_String              := Empty_String;
         Seed_Value           : IN     Integer                       := 0;
         Predictable          : IN     Boolean                       := True) IS



      New_Periodic_Task   : Periodic_Task_Ptr;
      New_Aperiodic_Task  : Aperiodic_Task_Ptr;
      New_Poisson_Task    : Poisson_Task_Ptr;
      New_Parametric_Task : Parametric_Task_Ptr;
      New_Sporadic_Task   : Sporadic_Task_Ptr;

      My_Iterator : Iterator;


   BEGIN


      Check_Task(My_Tasks, Name, Cpu_Name, Address_Space_Name,
         Task_Type, Start_Time,
         Capacity, Period, Deadline, Jitter, Blocking_Time, Priority,
         Criticality, Policy, Offset, Stack_Memory_Size,
         Text_Memory_Size, Param, Parametric_Rule_Name, Seed_Value,
         Predictable);



      IF(Get_Number_Of_Elements(My_Tasks) > 0)
            THEN

         Reset_Iterator(My_Tasks, My_Iterator);

         LOOP
            Current_Element(My_Tasks, A_Task, My_Iterator);
            IF (Name = A_Task.Name) THEN
               Raise_Exception(Invalid_Parameter'Identity,
                  To_String(Lb_Task(Current_Language) & " " & Name &
                     " : " & Lb_Task_Name(Current_Language) &
                     Lb_Already_Defined(Current_Language)));
            END IF;
            EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);
            Next_Element(My_Tasks, My_Iterator);
         END LOOP;
      END IF;


      CASE Task_Type IS
         WHEN Periodic_Type   =>
            New_Periodic_Task:=NEW Periodic_Task;
            New_Periodic_Task.Jitter:=Natural(Jitter);
            New_Periodic_Task.Period:=Natural(Period);
            A_Task:=Task_Ptr(New_Periodic_Task);
         WHEN Aperiodic_Type =>
            New_Aperiodic_Task:=NEW Aperiodic_Task;
            A_Task:=Task_Ptr(New_Aperiodic_Task);
         WHEN Poisson_Type    =>
            New_Poisson_Task:=NEW Poisson_Task;
            New_Poisson_Task.Jitter:=Natural(Jitter);
            New_Poisson_Task.Period:=Natural(Period);
            New_Poisson_Task.Seed:=Seed_Value;
            New_Poisson_Task.Predictable:=Predictable;
            A_Task:=Task_Ptr(New_Poisson_Task);
         WHEN Sporadic_Type    =>
            New_Sporadic_Task:=NEW Sporadic_Task;
            New_Sporadic_Task.Jitter:=Natural(Jitter);
            New_Sporadic_Task.Period:=Natural(Period);
            New_Sporadic_Task.Seed:=Seed_Value;
            New_Sporadic_Task.Predictable:=Predictable;
            A_Task:=Task_Ptr(New_Sporadic_Task);
         WHEN Parametric_Type =>
            New_Parametric_Task:=NEW Parametric_Task;
            New_Parametric_Task.Jitter:=Natural(Jitter);
            New_Parametric_Task.Period:=Natural(Period);
            New_Parametric_Task.Seed:=Seed_Value;
            New_Parametric_Task.Predictable:=Predictable;
            New_Parametric_Task.Activation_Rule:=
               Parametric_Rule_Name;
            A_Task:=Task_Ptr(New_Parametric_Task);
      END CASE;

      A_Task.Offset:=Offset;
      A_Task.Deadline:=Natural(Deadline);
      A_Task.Name:=Name;
      A_Task.Capacity:=Natural(Capacity);
      A_Task.Start_Time:=Natural(Start_Time);
      A_Task.Cpu_Name:=Cpu_Name;
      A_Task.Address_Space_Name:=Address_Space_Name;
      A_Task.Priority:=Priority_Range(Priority);
      A_Task.Criticality:=Natural(Criticality);
      A_Task.Policy:=Policy;
      A_Task.Task_Type:=Task_Type;
      A_Task.Blocking_Time:=Natural(Blocking_Time);
      A_Task.Stack_Memory_Size:=Natural(Stack_Memory_Size);
      A_Task.Text_Memory_Size:=Natural(Text_Memory_Size);
      A_Task.Param:=Param;


      Add(My_Tasks,A_Task);



   EXCEPTION
      WHEN Full_Set =>
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Can_Not_Define_More_Tasks(
                  Current_Language)));

   END Add_Task;




   FUNCTION Task_Is_Present (
         My_Tasks : IN     Tasks_Set;
         Name     : IN     Unbounded_String)
     RETURN Boolean IS

      My_Iterator : Iterator;
      A_Task      : Task_Ptr;

      Found : Boolean := False;

   BEGIN

      IF  Is_Empty(My_Tasks) THEN
         RETURN False;
      ELSE
         Reset_Iterator(My_Tasks, My_Iterator);

         LOOP
            Current_Element(My_Tasks,A_Task,My_Iterator);

            IF(A_Task.Name = Name)
                  THEN
               Found:=True;
            END IF;

            EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

            Next_Element(My_Tasks, My_Iterator);

         END LOOP;

         RETURN Found;
      END IF;

   END Task_Is_Present;



   FUNCTION Search_Task (
         My_Tasks : IN     Tasks_Set;
         Name     : IN     Unbounded_String)
     RETURN Task_Ptr IS

      My_Iterator : Iterator;
      A_Task      : Task_Ptr;
      Result      : Task_Ptr;

      Found : Boolean := False;

   BEGIN

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF(A_Task.Name = Name)
               THEN
            Found:=True;
            Result:=A_Task;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;

      IF NOT Found
            THEN
         Raise_Exception(Task_Not_Found'Identity,
            To_String(Lb_Task_Name(Current_Language) & "=" & Name));
      END IF;

      RETURN Result;
   END Search_Task;



   FUNCTION Increasing_Si (
         Op1 : IN     Task_Ptr;
         Op2 : IN     Task_Ptr)
     RETURN Boolean IS
      Si1,
      Si2 : Float;
   BEGIN
      Si1 := Log(Float(Periodic_Task_Ptr(Op1).Period), 2.0)-
         Float'Floor(Log(Float(Periodic_Task_Ptr(Op1).Period), 2.0));
      Si2 := Float(Log(Float(Periodic_Task_Ptr(Op2).Period), 2.0))-
         Float'Floor(Float(Log(Float(Periodic_Task_Ptr(Op2).Period),
               2.0)));
      RETURN (Si1 <= Si2);
   END Increasing_Si;

   FUNCTION Increasing_Utilization (
         Op1 : IN     Task_Ptr;
         Op2 : IN     Task_Ptr)
     RETURN Boolean IS
      U1,
      U2 : Float;
   BEGIN
      U1 := Float(Op1.Capacity) / Float(Periodic_Task_Ptr(Op1).
         Period);
      U2 := Float(Op2.Capacity) / Float(Periodic_Task_Ptr(Op2).
         Period);
      RETURN (U1 <= U2);
   END Increasing_Utilization;


   FUNCTION Increasing_Priority (
         Op1 : IN     Task_Ptr;
         Op2 : IN     Task_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Op1.Priority<= Op2.Priority);
   END Increasing_Priority;

   FUNCTION Decreasing_Priority (
         Op1 : IN     Task_Ptr;
         Op2 : IN     Task_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Op1.Priority>= Op2.Priority);
   END Decreasing_Priority;


   FUNCTION Increasing_Period (
         Op1 : IN     Task_Ptr;
         Op2 : IN     Task_Ptr)
     RETURN Boolean IS
   BEGIN
      CASE Op1.Task_Type IS
         WHEN Periodic_Type =>
            RETURN (Periodic_Task_Ptr(Op1).Period <=
               Periodic_Task_Ptr(Op2).Period);
         WHEN Poisson_Type =>
            RETURN (Poisson_Task_Ptr(Op1).Period <=
               Poisson_Task_Ptr(Op2).Period);
         WHEN OTHERS =>
            RETURN True;
      END CASE;

   END Increasing_Period;


   FUNCTION Decreasing_Period (
         Op1 : IN     Task_Ptr;
         Op2 : IN     Task_Ptr)
     RETURN Boolean IS
   BEGIN
      CASE Op1.Task_Type IS
         WHEN Periodic_Type =>
            RETURN (Periodic_Task_Ptr(Op1).Period >=
               Periodic_Task_Ptr(
                  Op2).
               Period);
         WHEN Poisson_Type =>
            RETURN (Poisson_Task_Ptr(Op1).Period >= Poisson_Task_Ptr(
                  Op2).
               Period);
         WHEN OTHERS =>
            RETURN True;
      END CASE;

   END Decreasing_Period;


   FUNCTION Increasing_Deadline (
         Op1 : IN     Task_Ptr;
         Op2 : IN     Task_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Op1.Deadline <= Op2.Deadline);
   END Increasing_Deadline;


   FUNCTION Decreasing_Deadline (
         Op1 : IN     Task_Ptr;
         Op2 : IN     Task_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Op1.Deadline >= Op2.Deadline);
   END Decreasing_Deadline;

   FUNCTION Increasing_Name (
         Op1 : IN     Task_Ptr;
         Op2 : IN     Task_Ptr)
     RETURN Boolean IS
   BEGIN
      RETURN (Op1.Name<=Op2.Name);
   END Increasing_Name;



   FUNCTION Get (
         My_Tasks   : IN     Tasks_Set;
         Task_Name  : IN     Unbounded_String;
         Param_Name : IN     Task_Parameters)
     RETURN Unbounded_String IS

      A_Task      : Task_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF (Param_Name /= Cpu_Name)
            AND (Param_Name /= Address_Space_Name)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF (A_Task.Name = Task_Name)
               THEN
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;

      IF (Param_Name = Cpu_Name)
            THEN
         RETURN A_Task.Cpu_Name;
      ELSE
         RETURN A_Task.Address_Space_Name;
      END IF;

   END Get;


   FUNCTION Get (
         My_Tasks   : IN     Tasks_Set;
         Task_Name  : IN     Unbounded_String;
         Param_Name : IN     Task_Parameters)
     RETURN Boolean IS

      A_Task      : Task_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF (Param_Name /= Predictable)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF (A_Task.Name = Task_Name)
               THEN
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;

      IF (A_Task.Task_Type /= Poisson_Type)
            AND (A_Task.Task_Type /= Parametric_Type)

            THEN
         RAISE Invalid_Parameter;
      END IF;

      RETURN Poisson_Task_Ptr(A_Task).Predictable;

   END Get;


   FUNCTION Get (
         My_Tasks   : IN     Tasks_Set;
         Task_Name  : IN     Unbounded_String;
         Param_Name : IN     Task_Parameters)
     RETURN Priority_Range IS

      A_Task      : Task_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF (Param_Name /= Priority)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF (A_Task.Name = Task_Name)
               THEN
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;

      RETURN A_Task.Priority;

   END Get;


   FUNCTION Get (
         My_Tasks   : IN     Tasks_Set;
         Task_Name  : IN     Unbounded_String;
         Param_Name : IN     Task_Parameters)
     RETURN Natural IS

      A_Task      : Task_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF ( (Param_Name /= Start_Time) AND
            (Param_Name /= Blocking_Time) AND
            (Param_Name /= Capacity) AND
            (Param_Name /= Period) AND
            (Param_Name /= Deadline) AND
            (Param_Name /= Jitter) )
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF (A_Task.Name = Task_Name)
               THEN
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;

      IF (Param_Name = Start_Time)
            THEN
         RETURN A_Task.Start_Time;
      ELSE
         IF (Param_Name = Deadline)
               THEN
            RETURN A_Task.Deadline;
         ELSE
            IF (Param_Name = Period)
                  THEN
               RETURN Periodic_Task_Ptr(A_Task).Period;
            ELSE
               IF (Param_Name = Jitter)
                     THEN
                  RETURN Periodic_Task_Ptr(A_Task).Jitter;
               ELSE
                  IF (Param_Name = Capacity)
                        THEN
                     RETURN A_Task.Capacity;
                  ELSE
                     RETURN A_Task.Blocking_Time;
                  END IF;
               END IF;
            END IF;
         END IF;
      END IF;

   END Get;


   PROCEDURE Set (
         My_Tasks    : IN OUT Tasks_Set;
         Task_Name   : IN     Unbounded_String;
         Param_Name  : IN     Task_Parameters;
         Param_Value : IN     Unbounded_String) IS

      A_Task      : Task_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF (Param_Name /= Cpu_Name)
            AND (Param_Name /= Address_Space_Name)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF (A_Task.Name = Task_Name)
               THEN
            IF (Param_Name = Cpu_Name)
                  THEN
               A_Task.Cpu_Name:=Param_Value;
            ELSE
               A_Task.Address_Space_Name:=Param_Value;
            END IF;
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;

   END Set;


   PROCEDURE Set (
         My_Tasks    : IN OUT Tasks_Set;
         Task_Name   : IN     Unbounded_String;
         Param_Name  : IN     Task_Parameters;
         Param_Value : IN     Priority_Range) IS

      A_Task      : Task_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF (Param_Name /= Priority)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF (A_Task.Name = Task_Name)
               THEN
            A_Task.Priority:=Param_Value;
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;
   END Set;


   PROCEDURE Set (
         My_Tasks    : IN OUT Tasks_Set;
         Task_Name   : IN     Unbounded_String;
         Param_Name  : IN     Task_Parameters;
         Param_Value : IN     Boolean) IS

      A_Task      : Task_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF (Param_Name /= Predictable)
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF (A_Task.Name = Task_Name)
               THEN
            IF (A_Task.Task_Type /= Poisson_Type)
                  AND (A_Task.Task_Type /= Parametric_Type)
                  THEN
               RAISE Invalid_Parameter;
            END IF;
            Poisson_Task_Ptr(A_Task).Predictable:=Param_Value;
            EXIT;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;
   END Set;


   PROCEDURE Set (
         My_Tasks    : IN OUT Tasks_Set;
         Task_Name   : IN     Unbounded_String;
         Param_Name  : IN     Task_Parameters;
         Param_Value : IN     Natural) IS
      A_Task      : Task_Ptr;
      My_Iterator : Iterator;

   BEGIN

      IF ( (Param_Name /= Start_Time) AND
            (Param_Name /= Blocking_Time) AND
            (Param_Name /= Capacity) AND
            (Param_Name /= Period) AND
            (Param_Name /= Deadline) AND
            (Param_Name /= Jitter) )
            THEN
         RAISE Invalid_Parameter;
      END IF;

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF (A_Task.Name = Task_Name)
               THEN
            IF (Param_Name = Start_Time)
                  THEN
               A_Task.Start_Time:=Param_Value;
               EXIT;
            END IF;
            IF (Param_Name = Deadline)
                  THEN
               A_Task.Deadline:=Param_Value;
            END IF;
            IF (Param_Name = Period AND A_Task.Task_Type =
                  Periodic_Type)
                  THEN
               Periodic_Task_Ptr(A_Task).Period:=Param_Value;
               EXIT;
            END IF;
            IF (Param_Name = Jitter AND A_Task.Task_Type =
                  Periodic_Type)
                  THEN
               Periodic_Task_Ptr(A_Task).Jitter:=Param_Value;
               EXIT;
            END IF;
            IF (Param_Name = Period AND A_Task.Task_Type =
                  Poisson_Type)
                  THEN
               Poisson_Task_Ptr(A_Task).Period:=Param_Value;
               EXIT;
            END IF;
            IF (Param_Name = Jitter AND A_Task.Task_Type =
                  Poisson_Type)
                  THEN
               Poisson_Task_Ptr(A_Task).Jitter:=Param_Value;
               EXIT;
            END IF;
            IF (Param_Name = Capacity)
                  THEN
               A_Task.Capacity:=Param_Value;
               EXIT;
            END IF;
            IF (Param_Name = Blocking_Time)
                  THEN
               A_Task.Blocking_Time:=Param_Value;
               EXIT;
            END IF;
         END IF;
         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;
   END Set;


   FUNCTION Get_Number_Of_Task_From_Processor (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Tasks_Range IS

      Number      : Tasks_Range := 0;
      A_Task      : Task_Ptr;
      My_Iterator : Iterator;

   BEGIN


      IF Is_Empty(My_Tasks)
            THEN
         RETURN 0;
      END IF;


      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks, A_Task, My_Iterator);

         IF (A_Task.Cpu_Name = Processor_Name)
               THEN
            Number:=Number+1;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;

      RETURN Number;
   END Get_Number_Of_Task_From_Processor;


   PROCEDURE Periodic_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

      Iterator1 : Tasks_Iterator;
      Task1     : Task_Ptr;

   BEGIN

      Reset_Iterator(My_Tasks, Iterator1);

      LOOP
         Current_Element(My_Tasks, Task1, Iterator1);


         IF (Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String )
               THEN
            IF  (Task1.Task_Type /= Periodic_Type)
                  THEN
               RAISE Task_Must_Be_Periodic;
            END IF;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, Iterator1);
         Next_Element(My_Tasks, Iterator1);

      END LOOP;

   END Periodic_Control;

   PROCEDURE Periodic_Or_Sporadic_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

      Iterator1 : Tasks_Iterator;
      Task1     : Task_Ptr;

   BEGIN

      Reset_Iterator(My_Tasks, Iterator1);

      LOOP
         Current_Element(My_Tasks, Task1, Iterator1);


         IF (Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String )
               THEN
            IF  (Task1.Task_Type /= Periodic_Type) AND  (Task1.Task_Type /=
                  Sporadic_Type)
                  THEN
               RAISE Task_Must_Be_Periodic;
            END IF;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, Iterator1);
         Next_Element(My_Tasks, Iterator1);

      END LOOP;

   END Periodic_Or_Sporadic_Control;


   PROCEDURE Periodic_Or_Poisson_Process_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS


      Iterator1 : Tasks_Iterator;
      Task1     : Task_Ptr;

   BEGIN

      Reset_Iterator(My_Tasks, Iterator1);

      LOOP
         Current_Element(My_Tasks, Task1, Iterator1);


         IF (Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String )
               THEN
            IF  (Task1.Task_Type /= Periodic_Type)
                  AND (Task1.Task_Type /= Poisson_Type)

                  THEN
               RAISE Task_Must_Be_Periodic_Or_Poisson_Process;
            END IF;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, Iterator1);
         Next_Element(My_Tasks, Iterator1);

      END LOOP;

   END Periodic_Or_Poisson_Process_Control;


   PROCEDURE Priority_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS


      Iterator1 : Tasks_Iterator;
      Task1     : Task_Ptr;
      Iterator2 : Tasks_Iterator;
      Task2     : Task_Ptr;

   BEGIN
      -- No duplicate priority
      --

      Reset_Iterator(My_Tasks, Iterator1);


      LOOP
         Current_Element(My_Tasks, Task1, Iterator1);


         IF (Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String )
               THEN
            Reset_Iterator(My_Tasks, Iterator2);

            LOOP
               Current_Element(My_Tasks, Task2, Iterator2);


               IF (Task2.Cpu_Name = Task1.Cpu_Name)
                     AND (Task1.Name /= Task2.Name)
                     AND (Task1.Priority = Task2.Priority)
                     THEN
                  RAISE Priority_Error;
               END IF;

               EXIT WHEN Is_Last_Element(My_Tasks, Iterator2);
               Next_Element(My_Tasks, Iterator2);

            END LOOP;


         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, Iterator1);
         Next_Element(My_Tasks, Iterator1);

      END LOOP;
   END Priority_Control;


   -- Check start time
   --
   PROCEDURE Start_Time_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

      Iterator1 : Tasks_Iterator;
      Task1     : Task_Ptr;

   BEGIN

      Reset_Iterator(My_Tasks, Iterator1);

      LOOP
         Current_Element(My_Tasks, Task1, Iterator1);


         IF ( (Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String )  )
               AND (Task1.Start_Time /= 0)
               THEN
            RAISE Start_Time_Error;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, Iterator1);
         Next_Element(My_Tasks, Iterator1);

      END LOOP;
   END Start_Time_Control;


   -- Check offset
   --
   PROCEDURE Offset_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

      Iterator1 : Tasks_Iterator;
      Task1     : Task_Ptr;

   BEGIN

      Reset_Iterator(My_Tasks, Iterator1);

      LOOP
         Current_Element(My_Tasks, Task1, Iterator1);


         IF (Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String )
               THEN
            FOR I IN 0..Task1.Offset.Nb_Entries-1 LOOP
               IF (Task1.Offset.Entries(I).Activation /= 0)
                     THEN
                  RAISE Offset_Error;
               END IF;
            END LOOP;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, Iterator1);
         Next_Element(My_Tasks, Iterator1);

      END LOOP;
   END Offset_Control;


   PROCEDURE Have_Deadlines_Equal_Than_Periods_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

      A_Task      : Task_Ptr;
      My_Iterator : Iterator;

   BEGIN

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF ( NOT ((A_Task.Task_Type = Periodic_Type) OR
                  (A_Task.Task_Type = Poisson_Type)) AND
               (A_Task.Cpu_Name = Processor_Name) )
               THEN
            RAISE Task_Must_Be_Periodic;
         END IF;

         IF   (A_Task.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String )
               THEN
            IF A_Task.Task_Type = Periodic_Type THEN
               IF (A_Task.Deadline /=
                     Periodic_Task_Ptr(A_Task).Period)
                     THEN
                  RAISE Task_Must_Have_Period_Equal_To_Deadline;
               END IF;
            END IF;

            IF A_Task.Task_Type = Poisson_Type THEN  -- check this
               IF (A_Task.Deadline /=
                     Poisson_Task_Ptr(A_Task).Period)
                     THEN
                  RAISE Task_Must_Have_Period_Equal_To_Deadline;
               END IF;
            END IF;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;

   END Have_Deadlines_Equal_Than_Periods_Control ;



   PROCEDURE Deadline_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

      Iterator1 : Tasks_Iterator;
      Task1     : Task_Ptr;

   BEGIN

      Reset_Iterator(My_Tasks, Iterator1);

      LOOP
         Current_Element(My_Tasks, Task1, Iterator1);


         IF (Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String )
               THEN
            IF Task1.Deadline = 0
                  THEN
               RAISE Deadline_Error;
            END IF;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, Iterator1);
         Next_Element(My_Tasks, Iterator1);

      END LOOP;

   END Deadline_Control;


   PROCEDURE Have_Same_Period_Control (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String) IS

      Iterator1    : Tasks_Iterator;
      Task1        : Task_Ptr;
      Period_Value : Natural        := 0;
      First        : Boolean        := True;

   BEGIN

      Reset_Iterator(My_Tasks, Iterator1);

      LOOP
         Current_Element(My_Tasks, Task1, Iterator1);

         IF  NOT ((Task1.Task_Type = Periodic_Type) OR
               (Task1.Task_Type = Aperiodic_Type))
               THEN
            RAISE Task_Model_Error;
         END IF;

         IF (Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String )
               THEN
            IF First THEN
               First:=False;
               IF Task1.Task_Type = Periodic_Type
                     THEN
                  Period_Value:=Periodic_Task_Ptr(Task1).Period;
               END IF;
            ELSE
               IF Task1.Task_Type= Aperiodic_Type AND
                     Period_Value /= 0
                     THEN
                  RAISE Task_Must_Have_The_Same_Period_Value;
               END IF;
               IF Task1.Task_Type= Periodic_Type AND
                     Period_Value /= Periodic_Task_Ptr(Task1).Period
                     THEN
                  RAISE Task_Must_Have_The_Same_Period_Value;
               END IF;
            END IF;
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, Iterator1);
         Next_Element(My_Tasks, Iterator1);

      END LOOP;

   END Have_Same_Period_Control;


   FUNCTION Deadline_Inferior_To_Period (
         My_Tasks : IN     Tasks_Set)
     RETURN Boolean IS

      My_Iterator : Iterator;
      A_Task      : Task_Ptr;
      Result      : Boolean  := True;

   BEGIN
      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF A_Task.Task_Type = Periodic_Type THEN
            IF(Periodic_Task_Ptr(A_Task).Period <  A_Task.Deadline)
                  THEN
               Result:=False;
            END IF;
         END IF;

         IF A_Task.Task_Type = Poisson_Type THEN
            IF(Poisson_Task_Ptr(A_Task).Period <  A_Task.Deadline)
                  THEN
               Result:=False;
            END IF;
         END IF;


         -- Next task
         --
         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);
         Next_Element(My_Tasks, My_Iterator);
      END LOOP;

      RETURN Result;

   END Deadline_Inferior_To_Period;


   FUNCTION Is_Harmonic (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Boolean IS

      My_Iterator1,
      My_Iterator2 : Iterator;
      A_Task1,
      A_Task2      : Task_Ptr;
      Result       : Boolean  := True;
      Tmp_Mod      : Natural;

   BEGIN
      -- First loop
      --
      Reset_Iterator(My_Tasks, My_Iterator1);

      LOOP
         Current_Element(My_Tasks,A_Task1,My_Iterator1);

         IF (A_Task1.Cpu_Name = Processor_Name)
               OR (Processor_Name = Empty_String )
               THEN

            -- Second loop
            --
            Reset_Iterator(My_Tasks, My_Iterator2);

            LOOP
               Current_Element(My_Tasks,A_Task2,My_Iterator2);

               IF (A_Task2.Cpu_Name = Processor_Name)
                     OR (Processor_Name = Empty_String )
                     THEN

                  IF A_Task1.Task_Type = Periodic_Type THEN
                     IF (Periodic_Task_Ptr(A_Task1).Period >
                           Periodic_Task_Ptr(
                              A_Task2).Period)
                           THEN
                        Tmp_Mod := Periodic_Task_Ptr(A_Task1).Period mod
                           Periodic_Task_Ptr(A_Task2).Period;
                     ELSE
                        Tmp_Mod := Periodic_Task_Ptr(A_Task2).Period mod
                           Periodic_Task_Ptr(A_Task1).Period;
                     END IF;

                     IF ( A_Task1.Name /= A_Task2.Name AND Tmp_Mod /= 0 )
                           THEN
                        Result:=False;
                     END IF;
                  END IF;


                  IF A_Task1.Task_Type = Poisson_Type THEN
                     IF (Poisson_Task_Ptr(A_Task1).Period >
                           Poisson_Task_Ptr(
                              A_Task2).Period)
                           THEN
                        Tmp_Mod := Poisson_Task_Ptr(A_Task1).Period mod
                           Poisson_Task_Ptr(A_Task2).Period;
                     ELSE
                        Tmp_Mod := Poisson_Task_Ptr(A_Task2).Period mod
                           Poisson_Task_Ptr(A_Task1).Period;
                     END IF;

                     IF ( A_Task1.Name /= A_Task2.Name AND Tmp_Mod /= 0 )
                           THEN
                        Result:=False;
                     END IF;
                  END IF;

               END IF;

               -- next task
               --
               EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator2);
               Next_Element(My_Tasks, My_Iterator2);

            END LOOP;


         END IF;

         -- next task
         --
         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator1);
         Next_Element(My_Tasks, My_Iterator1);
      END LOOP;

      RETURN Result;
   END Is_Harmonic;


   PROCEDURE Delete_Address_Space (
         My_Tasks : IN OUT Tasks_Set;
         A_Addr   : IN     Unbounded_String) IS

      Tmp         : Tasks_Set;
      A_Task      : Task_Ptr;
      My_Iterator : Tasks_Iterator;

   BEGIN


      IF(Get_Number_Of_Elements(My_Tasks) > 0)
            THEN

         Reset_Iterator(My_Tasks, My_Iterator);
         LOOP
            Current_Element(My_Tasks, A_Task, My_Iterator);

            IF(A_Task.Address_Space_Name = A_Addr )
                  THEN
               Add(Tmp, A_Task);
            END IF;

            EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);
            Next_Element(My_Tasks, My_Iterator);
         END LOOP;

         IF NOT Is_Empty(Tmp) THEN
            Reset_Iterator(Tmp, My_Iterator);
            LOOP
               Current_Element(Tmp, A_Task, My_Iterator);

               Delete(My_Tasks,A_Task);

               EXIT WHEN Is_Last_Element(Tmp, My_Iterator);
               Next_Element(Tmp, My_Iterator);
            END LOOP;

            Reset(Tmp,False);
         END IF;

      END IF;


   END Delete_Address_Space;


   FUNCTION Scheduling_Period (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Double IS

      Period     : Double  := 1.0;
      Start_Time : Integer := 0;

      A_Task      : Task_Ptr;
      My_Iterator : Tasks_Iterator;

      PRAGMA Warnings (Off);

      TYPE My_Index IS mod 2**16;
      FOR My_Index'Size USE 16;

      PACKAGE My_Big IS NEW Big_Number(My_Index, My_Index(
            Framework_Config.
            Max_Big_Number_Size));
      USE My_Big;
      USE My_Big.Conversion;
      USE My_Big.Generic_Conversion;
      USE My_Big.String_Conversion;
      USE My_Big.Unsigned_Number;

      Res,
      Current : Big_Unsigned;

      FUNCTION Set IS
      NEW Int_Number2big_Unsigned(Integer);

      PRAGMA Warnings (On);


   BEGIN


      -- Check if Tasks Are Periodic
      --
      Periodic_Control(My_Tasks, Processor_Name);


      Res:=Set(1);

      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks, A_Task, My_Iterator);

         IF (A_Task.Cpu_Name = Processor_Name AND
               A_Task.Task_Type = Periodic_Type)
               THEN
            Start_Time:=Natural'Max(Start_Time, A_Task.Start_Time);
            Current:=Set(Periodic_Task_Ptr(A_Task).Period);
            Res:=Lcm(Res,Current);
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);
      END LOOP;


      IF Start_Time/=0
            THEN
         Current:=Set(2);
         Res:=Res*Current;
         Current:=Set(Start_Time);
         Res:=Res+Current;
      END IF;

      Period:=Double(Big_Unsigned2long_Long_Float(Res));


      RETURN Period;

   END Scheduling_Period;



   FUNCTION Scheduling_Period (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Natural IS

      Period     : Natural := 1;
      Start_Time : Natural := 0;

      A_Task      : Task_Ptr;
      My_Iterator : Tasks_Iterator;

   BEGIN


      -- Check if tasks are periodic
      --
      Periodic_Control(My_Tasks, Processor_Name);


      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks, A_Task, My_Iterator);

         IF (A_Task.Cpu_Name = Processor_Name AND
               A_Task.Task_Type = Periodic_Type)
               THEN
            Start_Time:=Natural'Max(Start_Time, A_Task.Start_Time);
            Period:= Natural_Util.Lcm(Period,
               Periodic_Task_Ptr(A_Task).Period);
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);
      END LOOP;


      IF Start_Time/=0
            THEN
         Period:=(Period*2)+Start_Time;
      END IF;

      RETURN Period;


   END Scheduling_Period;


   FUNCTION Processor_Utilization_Over_Deadline (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Double IS

      Utilization : Double := 0.0;

      My_Iterator : Tasks_Iterator;
      A_Task      : Task_Ptr;

   BEGIN

      -- Check if tasks are periodic
      --
      Periodic_Control(My_Tasks, Processor_Name);


      -- Check if deadline a greater than 0
      --
      Deadline_Control(My_Tasks, Processor_Name);




      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF (A_Task.Cpu_Name = Processor_Name)
               THEN
            Utilization:=Utilization+(Double(A_Task.Capacity) /
               Double(A_Task.Deadline));
         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;


      RETURN Utilization;

   END Processor_Utilization_Over_Deadline;


   FUNCTION Processor_Utilization_Over_Period (
         My_Tasks       : IN     Tasks_Set;
         Processor_Name : IN     Unbounded_String)
     RETURN Double IS

      Utilization : Double := 0.0;

      My_Iterator : Tasks_Iterator;
      A_Task      : Task_Ptr;

   BEGIN

      -- Check if tasks are periodics
      --
      Periodic_Control(My_Tasks, Processor_Name);



      Reset_Iterator(My_Tasks, My_Iterator);

      LOOP
         Current_Element(My_Tasks,A_Task,My_Iterator);

         IF (A_Task.Cpu_Name = Processor_Name)
               THEN
            IF A_Task.Task_Type /= Aperiodic_Type THEN
               Utilization:=Utilization+(Double(A_Task.Capacity) /
                  Double(Periodic_Task_Ptr(A_Task).Period));
            END IF;

         END IF;

         EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

         Next_Element(My_Tasks, My_Iterator);

      END LOOP;

      RETURN Utilization;

   END Processor_Utilization_Over_Period;




   PROCEDURE Delete_Processor (
         My_Tasks    : IN OUT Tasks_Set;
         A_Processor : IN     Unbounded_String) IS

      Tmp         : Tasks_Set;
      A_Task      : Task_Ptr;
      My_Iterator : Tasks_Iterator;

   BEGIN

      IF(Get_Number_Of_Elements(My_Tasks) > 0)
            THEN

         Reset_Iterator(My_Tasks, My_Iterator);
         LOOP
            Current_Element(My_Tasks, A_Task,
               My_Iterator);

            IF(A_Task.Cpu_Name = A_Processor)
                  THEN
               Add(Tmp, A_Task);
            END IF;

            EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);
            Next_Element(My_Tasks, My_Iterator);
         END LOOP;

         IF NOT Is_Empty(Tmp) THEN
            Reset_Iterator(Tmp, My_Iterator);
            LOOP
               Current_Element(Tmp, A_Task, My_Iterator);


               Delete(My_Tasks,A_Task);

               EXIT WHEN Is_Last_Element(Tmp, My_Iterator);
               Next_Element(Tmp, My_Iterator);
            END LOOP;

            Reset(Tmp,False);
         END IF;
      END IF;

   END Delete_Processor;




   FUNCTION Export_Xml (
         My_Tasks : IN     Tasks_Set)
     RETURN Unbounded_String IS

      My_Iterator : Tasks_Iterator;
      A_Task      : Task_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN


      IF NOT Is_Empty(My_Tasks) THEN

         Result:= Result & To_Unbounded_String("<tasks>") &
            Unbounded_Lf;

         Reset_Iterator(My_Tasks, My_Iterator);

         LOOP
            Current_Element(My_Tasks,A_Task,My_Iterator);

            Result:= Result & To_Unbounded_String(Ascii.Ht & "<task");
            IF A_Task.X/=0
                  THEN
               Result:= Result & To_Unbounded_String(" x=""" &
                  A_Task.X'
                  Img & """ ");
            END IF;
            IF A_Task.Y/=0
                  THEN
               Result:= Result & To_Unbounded_String("y=""" &
                  A_Task.Y'
                  Img & """ ");
            END IF;

            CASE A_Task.Task_Type IS
               WHEN Periodic_Type =>
                  Result:= Result & To_Unbounded_String(
                     " task_type=""PERIODIC_TYPE"" ");
               WHEN Poisson_Type  =>
                  Result:= Result & To_Unbounded_String(
                     " task_type=""POISSON_TYPE"" ");
               WHEN Sporadic_Type  =>
                  Result:= Result & To_Unbounded_String(
                     " task_type=""SPORADIC_TYPE"" ");
               WHEN Parametric_Type  =>
                  Result:= Result & To_Unbounded_String(
                     " task_type=""PARAMETRIC_TYPE"" ");
               WHEN OTHERS =>
                  NULL;
            END CASE;
            Result:= Result & To_Unbounded_String(">") &
               Unbounded_Lf;


            IF A_Task.Param.Nb_Entries>0 THEN
               Result:= Result & To_Unbounded_String(Ascii.Ht &
                  Ascii.Ht &
                  "<parameters>") & Unbounded_Lf;
               FOR I IN 0..A_Task.Param.Nb_Entries-1 LOOP

                  Result:= Result & To_Unbounded_String(Ascii.Ht &
                     Ascii.Ht & Ascii.Ht &
                     "<parameter");
                  IF A_Task.Param.Entries(I).Ptype =
                        Boolean_Parameter THEN
                     Result:= Result & To_Unbounded_String(
                        " parameter_type=""boolean""> ");
                     Result:= Result & To_Unbounded_String(" " &
                        A_Task.Param.Entries(I).
                        Boolean_Value'
                        Img & " " );
                  ELSE
                     IF A_Task.Param.Entries(I).Ptype =
                           Integer_Parameter THEN
                        Result:= Result & To_Unbounded_String(
                           " parameter_type=""integer""> ");
                        Result:= Result & To_Unbounded_String(" " &
                           A_Task.Param.Entries(I).
                           Integer_Value'Img & " " );
                     ELSE
                        IF A_Task.Param.Entries(I).Ptype =
                              Double_Parameter THEN
                           Result:= Result & To_Unbounded_String(
                              " parameter_type=""double""> ");
                           Result:= Result & To_Unbounded_String(
                              " " &
                              To_String(Format(A_Task.
                                    Param.
                                    Entries(I).
                                    Double_Value)) & " " );
                        ELSE
                           Result:= Result & To_Unbounded_String(
                              " parameter_type=""string""> ");
                           Result:= Result & To_Unbounded_String(
                              " " &
                              To_String(A_Task.Param.
                                 Entries(
                                    I).
                                 String_Value) & " " );
                        END IF;
                     END IF;
                  END IF;
                  Result:= Result & To_Unbounded_String(To_String(
                        " " &
                        A_Task.Param. Entries(I).
                        Name) &
                     " ");
                  Result:= Result & To_Unbounded_String(
                     "</parameter>") &
                     Unbounded_Lf;
               END LOOP;
               Result:= Result & To_Unbounded_String(Ascii.Ht &
                  Ascii.Ht &
                  "</parameters>") & Unbounded_Lf;
            END IF;


            IF A_Task.Offset.Nb_Entries>0 THEN
               Result:= Result & To_Unbounded_String(Ascii.Ht &
                  Ascii.Ht &
                  "<offsets>") & Unbounded_Lf;
               FOR I IN 0..A_Task.Offset.Nb_Entries-1 LOOP
                  Result:= Result & To_Unbounded_String(Ascii.Ht &
                     Ascii.Ht & Ascii.Ht & "<offset>");
                  Result:= Result & To_Unbounded_String(" " &
                     A_Task.Offset.Entries(I).Activation'
                     Img &
                     " ");
                  Result:= Result & To_Unbounded_String(" " &
                     A_Task.Offset.Entries(I).Value'Img &
                     " ");
                  Result:= Result & To_Unbounded_String( "</offset>") &
                     Unbounded_Lf;
               END LOOP;
               Result:= Result & To_Unbounded_String(Ascii.Ht &
                  Ascii.Ht &
                  "</offsets>") & Unbounded_Lf;
            END IF;


            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<cpu_name>" &
               To_String(
                  A_Task.Cpu_Name) & "</cpu_name>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<address_space_name>" &
               To_String(
                  A_Task.Address_Space_Name) &
               "</address_space_name>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<name>" & To_String(
                  A_Task.Name) & "</name>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<capacity>" & A_Task.
               Capacity'Img & "</capacity>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<start_time>" & A_Task.
               Start_Time'Img & "</start_time>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<policy>" & A_Task.
               Policy'Img & "</policy>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<deadline>" & A_Task.
               Deadline'Img & "</deadline>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<criticality>" & A_Task.
               Criticality'Img & "</criticality>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<blocking_time>" &
               A_Task.
               Blocking_Time'Img & "</blocking_time>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<priority>" & A_Task.
               Priority'Img & "</priority>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<text_memory_size>" &
               A_Task.
               Text_Memory_Size'
               Img & "</text_memory_size>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "<stack_memory_size>" &
               A_Task.
               Stack_Memory_Size'
               Img & "</stack_memory_size>") & Unbounded_Lf;


            IF (A_Task.Task_Type /= Aperiodic_Type)  THEN
               Result:= Result & To_Unbounded_String(Ascii.Ht &
                  Ascii.Ht &
                  "<period>" &
                  Periodic_Task_Ptr(A_Task).Period'Img & "</period>") &
                  Unbounded_Lf;
               Result:= Result & To_Unbounded_String(Ascii.Ht &
                  Ascii.Ht &
                  "<jitter>" &
                  Periodic_Task_Ptr(A_Task).Jitter'Img & "</jitter>") &
                  Unbounded_Lf;

            END IF;


            IF (A_Task.Task_Type = Poisson_Type)  OR
                  (A_Task.Task_Type = Sporadic_Type)  OR
                  (A_Task.Task_Type =  Parametric_Type)
                  THEN
               Result:= Result & To_Unbounded_String(Ascii.Ht &
                  Ascii.Ht &
                  "<seed>" &
                  Poisson_Task_Ptr(A_Task).Seed'Img & "</seed>") &
                  Unbounded_Lf;
               Result:= Result & To_Unbounded_String(Ascii.Ht &
                  Ascii.Ht &
                  "<predictable_seed>" &
                  Poisson_Task_Ptr(A_Task).Predictable'Img &
                  "</predictable_seed>") & Unbounded_Lf;
            END IF;


            IF   (A_Task.Task_Type = Parametric_Type)
                  THEN
               Result:= Result & To_Unbounded_String(Ascii.Ht &
                  Ascii.Ht &
                  "<activation_rule>"  & To_String(
                     Parametric_Task_Ptr(A_Task).Activation_Rule) &
                  "</activation_rule>") & Unbounded_Lf;
            END IF;



            Result:= Result & To_Unbounded_String(Ascii.Ht &
               "</task>") &
               Unbounded_Lf;

            EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

            Next_Element(My_Tasks, My_Iterator);

         END LOOP;
         Result:= Result & To_Unbounded_String("</tasks>") &
            Unbounded_Lf;
      END IF;

      RETURN Result;

   END Export_Xml;




   FUNCTION Export_Aadl_Implementations (
         My_Tasks     : IN     Tasks_Set;
         My_Resources : IN     Resources_Set)
     RETURN Unbounded_String IS

      My_Iterator          : Tasks_Iterator;
      A_Task               : Task_Ptr;
      My_Resource_Iterator : Resources_Iterator;
      A_Resource           : Resource_Ptr;

      Print_Feature_Header : Boolean;

      Result : Unbounded_String := Empty_String;

   BEGIN


      IF NOT Is_Empty(My_Tasks) THEN


         Reset_Iterator(My_Tasks, My_Iterator);

         LOOP
            Current_Element(My_Tasks,A_Task,My_Iterator);

            Result:= Result & To_Unbounded_String("thread " &
               To_String(
                  A_Task.Name)) & Unbounded_Lf;
            Print_Feature_Header:=True;

            -- Look for shared resource access
            --
            IF NOT Is_Empty(My_Resources) THEN

               Reset_Iterator(My_Resources, My_Resource_Iterator);

               LOOP
                  Current_Element(My_Resources, A_Resource,
                     My_Resource_Iterator);

                  IF A_Resource.Task_Tab.Nb_Entries/=0 THEN
                     FOR I IN 0..A_Resource.Task_Tab.Nb_Entries-1 LOOP
                        IF A_Resource.Task_Tab.Entries(I).Item =
                              A_Task.
                              Name THEN
                           IF Print_Feature_Header THEN
                              Result:= Result & To_Unbounded_String(
                                 Ascii.Ht & "features") &
                                 Unbounded_Lf;
                              Print_Feature_Header:=False;
                           END IF;
                           Result:= Result & To_Unbounded_String(
                              Ascii.Ht &
                              Ascii.Ht & To_String(A_Resource.Name) &
                              "_features" &
                              " : requires data access "
                              & To_String(A_Resource.Name) &
                              ".Impl;") &
                              Unbounded_Lf;
                           EXIT;
                        END IF;
                     END LOOP;
                  END IF;

                  EXIT WHEN Is_Last_Element(My_Resources,
                     My_Resource_Iterator);

                  Next_Element(My_Resources, My_Resource_Iterator);

               END LOOP;
            END IF;

            Result:= Result & To_Unbounded_String( "end " &
               To_String(
                  A_Task.Name) & ";" ) & Unbounded_Lf & Unbounded_Lf;

            Result:= Result & To_Unbounded_String(
               "thread implementation " & To_String(A_Task.
                  Name) & ".Impl") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               "properties") &
               Unbounded_Lf;


            CASE A_Task.Task_Type IS
               WHEN Aperiodic_Type =>
                  Result:= Result & To_Unbounded_String(Ascii.Ht &
                     Ascii.Ht &
                     "Dispatch_Protocol => Background;") &
                     Unbounded_Lf;
               WHEN Periodic_Type =>
                  Result:= Result & To_Unbounded_String(Ascii.Ht &
                     Ascii.Ht &
                     "Dispatch_Protocol => Periodic;") &
                     Unbounded_Lf;
               WHEN Poisson_Type  =>
                  Result:= Result & To_Unbounded_String(Ascii.Ht &
                     Ascii.Ht &
                     "Dispatch_Protocol => Poisson_Process;") &
                     Unbounded_Lf;
               WHEN Sporadic_Type  =>
                  Result:= Result & To_Unbounded_String(Ascii.Ht &
                     Ascii.Ht &
                     "Dispatch_Protocol => Sporadic;") &
                     Unbounded_Lf;
               WHEN Parametric_Type  =>
                  Result:= Result & To_Unbounded_String(Ascii.Ht &
                     Ascii.Ht &
                     "Dispatch_Protocol => Parametric;") &
                     Unbounded_Lf;
                  Result:= Result & To_Unbounded_String(Ascii.Ht &
                     Ascii.Ht &
                     "Cheddar_Properties::Source_Text => " & """" &
                     To_String(
                        Parametric_Task_Ptr(A_Task).Activation_Rule) &
                     """" & ";") & Unbounded_Lf;
            END CASE;


            IF A_Task.Param.Nb_Entries>0 THEN

               FOR I IN 0..A_Task.Param.Nb_Entries-1 LOOP

                  Result:= Result & To_Unbounded_String( Ascii.Ht &
                     Ascii.Ht &
                     "User_Defined_Cheddar_Properties::" &
                     To_String(A_Task.
                        Param.Entries(I).Name) & " => ");
                  IF A_Task.Param.Entries(I).Ptype =
                        Boolean_Parameter THEN
                     Result:= Result & To_Unbounded_String(
                        A_Task.Param.Entries(I).Boolean_Value'Img);
                  ELSE
                     IF A_Task.Param.Entries(I).Ptype =
                           Integer_Parameter THEN
                        Result:= Result & To_Unbounded_String(
                           A_Task.Param.Entries(I).Integer_Value'
                           Img);
                     ELSE
                        IF A_Task.Param.Entries(I).Ptype =
                              Double_Parameter THEN
                           Result:= Result & To_Unbounded_String(
                              To_String(Format(A_Task.Param.
                                    Entries(I).Double_Value))  );
                        ELSE
                           Result:= Result & To_Unbounded_String("""" &
                              To_String(A_Task.Param.Entries(I).
                                 String_Value) & """");
                        END IF;
                     END IF;
                  END IF;
                  Result:= Result & To_Unbounded_String( ";") &
                     Unbounded_Lf;
               END LOOP;
            END IF;



            IF A_Task.Offset.Nb_Entries>0 THEN
               FOR I IN 0..A_Task.Offset.Nb_Entries-1 LOOP
                  Result:= Result & To_Unbounded_String( Ascii.Ht &
                     Ascii.Ht &
                     "Cheddar_Properties::Dispatch_Offset_Value" &
                     To_String(Integer_Util.
                        Format(Integer(I))) & " => " &
                     A_Task.Offset.
                     Entries(I).Value'Img &  ";");
                  Result:= Result & To_Unbounded_String( Ascii.Ht &
                     Ascii.Ht &
                     "Cheddar_Properties::Dispatch_Offset_Time" &
                     To_String(Integer_Util.
                        Format(Integer(I))) & " => " &
                     A_Task.Offset.
                     Entries(I).Activation'Img &  ";") &
                     Unbounded_Lf;
               END LOOP;
            END IF;

            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "Source_Code_Size => " &
               A_Task.
               Text_Memory_Size'
               Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "Source_Stack_Size => " &
               A_Task.
               Stack_Memory_Size'
               Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "Compute_Execution_Time => " & A_Task.Capacity'Img &
               " ms .. " & A_Task.Capacity'Img & " ms;") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "Cheddar_Properties::Dispatch_Absolute_Time => " &
               A_Task.
               Start_Time'Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "Cheddar_Properties::POSIX_Scheduling_Policy => " &
               A_Task.
               Policy'Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "Cheddar_Properties::Fixed_Priority => " & A_Task.
               Priority'Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "Cheddar_Properties::Bound_On_Data_Blocking_Time => " &
               A_Task.
               Blocking_Time'Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "Deadline  => " &
               A_Task.
               Deadline'Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "Cheddar_Properties::Criticality  => " &
               A_Task.
               Criticality'Img & ";") & Unbounded_Lf;

            IF (A_Task.Task_Type /= Aperiodic_Type)  THEN
               Result:= Result & To_Unbounded_String( Ascii.Ht &
                  Ascii.Ht &
                  "Period => " &
                  Periodic_Task_Ptr(A_Task).Period'Img & ";") &
                  Unbounded_Lf;
               Result:= Result & To_Unbounded_String( Ascii.Ht &
                  Ascii.Ht &
                  "Cheddar_Properties::Dispatch_Jitter => " &
                  Periodic_Task_Ptr(A_Task).Jitter'Img & ";") &
                  Unbounded_Lf;
            END IF;


            IF (A_Task.Task_Type = Poisson_Type)  OR
                  (A_Task.Task_Type = Sporadic_Type)  OR
                  (A_Task.Task_Type =  Parametric_Type)
                  THEN
               Result:= Result & To_Unbounded_String( Ascii.Ht &
                  Ascii.Ht &
                  "Cheddar_Properties::Dispatch_Seed_Value => " &
                  Poisson_Task_Ptr(A_Task).
                  Seed'Img & ";") & Unbounded_Lf;
               Result:= Result & To_Unbounded_String( Ascii.Ht &
                  Ascii.Ht &
                  "Cheddar_Properties::Dispatch_Seed_is_Predictable => " &
                  Poisson_Task_Ptr(
                     A_Task).Predictable'Img & ";") & Unbounded_Lf;
            END IF;


            Result:= Result & To_Unbounded_String( "end " &
               To_String(
                  A_Task.Name) & ".Impl;" ) & Unbounded_Lf &
               Unbounded_Lf;


            EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

            Next_Element(My_Tasks, My_Iterator);

         END LOOP;



      END IF;

      RETURN Result;

   END Export_Aadl_Implementations;





   FUNCTION Export_Aadl_User_Defined_Properties (
         My_Tasks : IN     Tasks_Set)
     RETURN Unbounded_String IS

      My_Iterator : Tasks_Iterator;
      A_Task      : Task_Ptr;

      Find            : Boolean                   := False;
      Properties_List : Unbounded_String_List;
      List_Ite        : Unbounded_String_Iterator;
      Str             : Unbounded_String_Ptr;
      USE Unbounded_Strings.Unbounded_String_List_Package;

      Result : Unbounded_String := Empty_String;

   BEGIN


      IF NOT Is_Empty(My_Tasks) THEN


         Reset_Iterator(My_Tasks, My_Iterator);

         LOOP
            Current_Element(My_Tasks,A_Task,My_Iterator);

            IF A_Task.Param.Nb_Entries>0 THEN
               FOR I IN 0..A_Task.Param.Nb_Entries-1 LOOP

                  -- Check that the property does not exist
                  --
                  Find:=False;
                  IF NOT Is_Empty(Properties_List) THEN

                     Reset_Head_Iterator(Properties_List, List_Ite);
                     LOOP
                        Current_Element (Properties_List, Str,
                           List_Ite);

                        IF Str.All=A_Task.Param.Entries(I).Name
                              THEN
                           Find:=True;
                           EXIT;
                        END IF;
                        IF  Is_Tail_Element (Properties_List,
                              List_Ite) THEN
                           EXIT;
                        END IF;
                        Next_Element (Properties_List, List_Ite);
                     END LOOP;

                  END IF;


                  IF NOT Find THEN
                     Str:= NEW Unbounded_String;
                     Str.All:=A_Task.Param.Entries(I).Name;
                     Add(Properties_List, Str);

                     Result:= Result & To_Unbounded_String( Ascii.Ht &
                        To_String(A_Task.Param.Entries(I).Name) &
                        " : ");
                     IF A_Task.Param.Entries(I).Ptype =
                           Boolean_Parameter THEN
                        Result:= Result & To_Unbounded_String(
                           "aadlboolean") & Unbounded_Lf;
                     ELSE
                        IF A_Task.Param.Entries(I).Ptype =
                              Integer_Parameter THEN
                           Result:= Result & To_Unbounded_String(
                              "aadlinteger") & Unbounded_Lf;
                        ELSE
                           IF A_Task.Param.Entries(I).Ptype =
                                 Double_Parameter THEN
                              Result:= Result & To_Unbounded_String(
                                 "aadlreal") & Unbounded_Lf;
                           ELSE
                              Result:= Result & To_Unbounded_String(
                                 "aadlstring") & Unbounded_Lf;
                           END IF;
                        END IF;
                     END IF;

                     Result:= Result & To_Unbounded_String(Ascii.Ht &
                        Ascii.Ht &
                        "applies to (thread, thread group);") &
                        Unbounded_Lf;
                  END IF;

               END LOOP;

            END IF;


            EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

            Next_Element(My_Tasks, My_Iterator);

         END LOOP;

      END IF;

      RETURN Result;

   END Export_Aadl_User_Defined_Properties;




   FUNCTION Export_Aadl_Declarations (
         My_Tasks           : IN     Tasks_Set;
         Address_Space_Name : IN     Unbounded_String;
         Number_Of_Ht       : IN     Natural)
     RETURN Unbounded_String IS


      My_Iterator : Tasks_Iterator;
      A_Task      : Task_Ptr;

      Result : Unbounded_String := Empty_String;

   BEGIN


      IF NOT Is_Empty(My_Tasks) THEN


         Reset_Iterator(My_Tasks, My_Iterator);

         LOOP
            Current_Element(My_Tasks,A_Task,My_Iterator);

            IF A_Task.Address_Space_Name = Address_Space_Name THEN
               FOR I IN 1..Number_Of_Ht LOOP
                  Result:= Result & Ascii.Ht;
               END LOOP;
               Result:= Result & To_Unbounded_String("instancied_" &
                  To_String(A_Task.Name) &
                  " : thread " & To_String(A_Task.Name) & ".Impl;") &
                  Unbounded_Lf;
            END IF;


            EXIT WHEN Is_Last_Element(My_Tasks, My_Iterator);

            Next_Element(My_Tasks, My_Iterator);

         END LOOP;

      END IF;


      RETURN Result;

   END Export_Aadl_Declarations;




END Task_Set;



