------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Text_Io;
USE Text_Io;



PACKAGE BODY Tasks IS


   --------------------------------------------------

   PROCEDURE Initialize (
         A_Task : IN OUT Generic_Task) IS
   BEGIN
      A_Task.Object_Type:=Task_Object_Type;
      Initialize(A_Task.Offset);
      Initialize(A_Task.Param);
   END Initialize;




   PROCEDURE Put_Name (
         A_Task : IN     Task_Ptr) IS
   BEGIN
      Put(To_String(A_Task.Name));
   END Put_Name;


   PROCEDURE Put (
         A_Task : IN     Task_Ptr) IS
   BEGIN
      Put(A_Task.All);
   END Put;


   PROCEDURE Initialize (
         A_Task : IN OUT Aperiodic_Task) IS
   BEGIN
      Initialize(Generic_Task(A_Task));
      A_Task.Task_Type:=Aperiodic_Type;
   END Initialize;



   PROCEDURE Initialize (
         A_Task : IN OUT Periodic_Task) IS
   BEGIN
      Initialize(Generic_Task(A_Task));
      A_Task.Task_Type:=Periodic_Type;
   END Initialize;



   PROCEDURE Initialize (
         A_Task : IN OUT Poisson_Task) IS
   BEGIN
      Initialize(Generic_Task(A_Task));
      A_Task.Task_Type := Poisson_Type;
   END Initialize;



   PROCEDURE Initialize (
         A_Task : IN OUT Sporadic_Task) IS
   BEGIN
      Initialize(Generic_Task(A_Task));
      A_Task.Task_Type := Sporadic_Type;
   END Initialize;


   PROCEDURE Initialize (
         A_Task : IN OUT Parametric_Task) IS
   BEGIN
      Initialize(Generic_Task(A_Task));
      A_Task.Task_Type:=Parametric_Type;
   END Initialize;



   --------------------------------------------------



   PROCEDURE Put (
         My_Task : IN     Generic_Task) IS
   BEGIN
      Put ("Name : " & To_String(My_Task.Name));
      Put (" ; Processor : " & To_String(My_Task.Cpu_Name));
      Put (" ; Address space : " & To_String(My_Task.Address_Space_Name));
      Put (" ; Type : " & My_Task.Task_Type'Img);
      Put (" ; Capacity : " &
         My_Task.Capacity'Img & " ; Start time : " & My_Task.Start_Time'
         Img & " ; ");
      Put ("Deadline : " & My_Task.Deadline'Img & " ; " );
      Put ("Priority : " & My_Task.Priority'Img & " ; " );
      Put ("Criticality : " & My_Task.Criticality'Img & " ; " );
      Put ("Stack Memory Size : " & My_Task.Stack_Memory_Size'Img & " ; " );
      Put ("Text Memory Size : " & My_Task.Text_Memory_Size'Img & " ; " );
      Put ("Blocking time : " & My_Task.Blocking_Time'Img & " ; " );
      Put ("Scheduling policy : " & My_Task.Policy'Img & " ; " );
      Put(" ; x : " & My_Task.X'Img);
      Put(" ; y : " & My_Task.Y'Img);

      Put(My_Task.Offset);
      Put(My_Task.Param);

   END Put;


   PROCEDURE Put (
         My_Task : IN     Periodic_Task) IS
   BEGIN
      Put(Generic_Task(My_Task));
      Put ("Period : " & My_Task.Period'Img & " ; " );
      Put ("Jitter : " & My_Task.Jitter'Img);
   END Put;

   PROCEDURE Put (
         My_Task : IN     Poisson_Task) IS
   BEGIN
      Put(Periodic_Task(My_Task));
      Put ("Seed : " & My_Task.Seed'Img);
      Put ("Predictable  : " & My_Task.Predictable'Img);

   END Put;

   PROCEDURE Put (
         My_Task : IN     Sporadic_Task) IS
   BEGIN
      Put(Poisson_Task(My_Task));
   END Put;

   PROCEDURE Put (
         My_Task : IN     Parametric_Task) IS
   BEGIN
      Put(Poisson_Task(My_Task));
      Put ("Activation rule : " & To_String(My_Task.Activation_Rule) &
         " ; " );
   END Put;


   -----------------------------------------------------


   FUNCTION Copy (
         My_Task : IN     Task_Ptr)
     RETURN Task_Ptr IS

      New_Periodic_Task   : Periodic_Task_Ptr;
      New_Aperiodic_Task  : Aperiodic_Task_Ptr;
      New_Poisson_Task    : Poisson_Task_Ptr;
      New_Sporadic_Task   : Sporadic_Task_Ptr;
      New_Parametric_Task : Parametric_Task_Ptr;
      Ptr                 : Task_Ptr;

   BEGIN

      CASE My_Task.Task_Type IS
         WHEN Periodic_Type   =>
            New_Periodic_Task       := NEW Periodic_Task;
            New_Periodic_Task.All   := Periodic_Task_Ptr(My_Task).ALL;
            Ptr                     := Task_Ptr(New_Periodic_Task);
         WHEN Aperiodic_Type  =>
            New_Aperiodic_Task      := NEW Aperiodic_Task;
            New_Aperiodic_Task.All  := Aperiodic_Task_Ptr(My_Task).ALL;
            Ptr                     := Task_Ptr(New_Aperiodic_Task);
         WHEN Poisson_Type    =>
            New_Poisson_Task        := NEW Poisson_Task;
            New_Poisson_Task.All    := Poisson_Task_Ptr(My_Task).ALL;
            Ptr                     := Task_Ptr(New_Poisson_Task);
         WHEN Sporadic_Type    =>
            New_Sporadic_Task        := NEW Sporadic_Task;
            New_Sporadic_Task.All    := Sporadic_Task_Ptr(My_Task).ALL;
            Ptr                     := Task_Ptr(New_Sporadic_Task);
         WHEN Parametric_Type =>
            New_Parametric_Task     := NEW Parametric_Task;
            New_Parametric_Task.All := Parametric_Task_Ptr(My_Task).ALL;
            Ptr                     := Task_Ptr(New_Parametric_Task);
      END CASE;

      RETURN Ptr;
   END Copy;


   --------------------------------------------------------------


   PROCEDURE Put (
         My_Task_Occurence : IN     Task_Occurence) IS

   BEGIN

      Put ("Arrival_Date : " & My_Task_Occurence.Arrival_Date'Img & " ; " );
      Put ("Absolute_Deadline : " & My_Task_Occurence.Absolute_Deadline'
         Img);
      Put ("Residual_capacity : " & My_Task_Occurence.Residual_Capacity'
         Img);
      Put ("Response_Time : " & My_Task_Occurence.Response_Time'Img);
   END Put;


   PROCEDURE Initialize_Cyclic (
         My_Task_Occurence : IN OUT Task_Occurence;
         A_Task            : IN     Task_Ptr;
         Arrival_Date      : IN     Natural        := 0) IS
   BEGIN
      My_Task_Occurence.Parent_Task       := A_Task ;
      My_Task_Occurence.Residual_Capacity := A_Task.Capacity ;
      My_Task_Occurence.Arrival_Date      := Arrival_Date;
      My_Task_Occurence.Absolute_Deadline := My_Task_Occurence.
         Arrival_Date + A_Task.Deadline ;

   END Initialize_Cyclic;


   PROCEDURE Initialize_Acyclic (
         My_Task_Occurence : IN OUT Task_Occurence;
         A_Task            : IN     Aperiodic_Task_Ptr) IS
   BEGIN
      My_Task_Occurence.Parent_Task       := Task_Ptr(A_Task) ;
      My_Task_Occurence.Residual_Capacity := A_Task.Capacity ;
      My_Task_Occurence.Arrival_Date      := A_Task.Start_Time ;
      My_Task_Occurence.Absolute_Deadline := My_Task_Occurence.
         Arrival_Date + A_Task.Deadline ;
   END Initialize_Acyclic;



END Tasks;
