------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Task_Offsets;
USE Task_Offsets;
USE Task_Offsets.Offsets_Table_Package;
WITH Parameters;
USE Parameters;
USE Parameters.User_Defined_Parameters_Table_Package;

WITH Convert_Strings;
WITH Convert_Unbounded_Strings;

WITH Objects;
USE Objects;

WITH Framework_Config;
USE  Framework_Config;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;

WITH Unchecked_Deallocation;



PACKAGE Tasks IS

   Compute_Aborded : EXCEPTION;

   TYPE Tasks_Type IS
         (Periodic_Type,
          Aperiodic_Type,
          Sporadic_Type,
          Poisson_Type,
          Parametric_Type);


   PROCEDURE To_Task_Type IS
   NEW Convert_Strings(Tasks_Type,Aperiodic_Type);
   PROCEDURE To_Task_Type IS
   NEW Convert_Unbounded_Strings(Tasks_Type,Aperiodic_Type);


   TYPE Policies IS
         (Sched_Fifo,
          Sched_Rr,
          Sched_Others);


   PROCEDURE To_Policy IS
   NEW Convert_Strings(Policies ,Sched_Fifo);
   PROCEDURE To_Policy IS
   NEW Convert_Unbounded_Strings(Policies ,Sched_Fifo);


   --------------------------------------------------

   TYPE Generic_Task IS ABSTRACT NEW Generic_Object WITH
      RECORD
         -- Type of task activation (periodic, aperiodic, sporadic...)
         --
         Task_Type : Tasks_Type;

         -- Address space and processor hosting the task
         --
         Cpu_Name           : Unbounded_String := To_Unbounded_String ("");
         Address_Space_Name : Unbounded_String := To_Unbounded_String ("");

         Capacity    : Natural        := 0;
         Deadline    : Natural        := 0;
         Start_Time  : Natural        := 0;
         Criticality : Natural        := 0;
         Priority    : Priority_Range := Priority_Range'First;

         Blocking_Time     : Natural                       := 0;
         Policy            : Policies;
         X                 : Natural                       := 0;
         Y                 : Natural                       := 0;
         Offset            : Offsets_Table;
         Text_Memory_Size  : Natural                       := 0;
         Stack_Memory_Size : Natural                       := 0;
         Param             : User_Defined_Parameters_Table;
      END RECORD;

   TYPE Task_Ptr IS ACCESS ALL Generic_Task'Class;



   PROCEDURE Initialize (
         A_Task : IN OUT Generic_Task);


   PROCEDURE Put (
         My_Task : IN     Generic_Task);


   PROCEDURE Put_Name (
         A_Task : IN     Task_Ptr);


   PROCEDURE Put (
         A_Task : IN     Task_Ptr);


   FUNCTION Copy (
         My_Task : IN     Task_Ptr)
     RETURN Task_Ptr;


   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Generic_Task'Class, Task_Ptr);


   --------------------------------------------------

   TYPE Periodic_Task IS NEW Generic_Task WITH
      RECORD
         Period : Natural := 0;
         Jitter : Natural := 0;
      END RECORD;

   TYPE Periodic_Task_Ptr IS ACCESS ALL Periodic_Task'Class;


   PROCEDURE Initialize (
         A_Task : IN OUT Periodic_Task);


   PROCEDURE Put (
         My_Task : IN     Periodic_Task);


   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Periodic_Task'Class, Periodic_Task_Ptr);


   --------------------------------------------------


   TYPE Aperiodic_Task IS NEW Generic_Task WITH NULL RECORD;

   TYPE Aperiodic_Task_Ptr IS ACCESS ALL Aperiodic_Task'Class;


   PROCEDURE Initialize (
         A_Task : IN OUT Aperiodic_Task);

   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Aperiodic_Task'Class, Aperiodic_Task_Ptr);

   --------------------------------------------------

   TYPE Poisson_Task IS NEW Periodic_Task WITH
      RECORD
         Seed        : Integer := 0;
         Predictable : Boolean := True;
      END RECORD;

   TYPE Poisson_Task_Ptr IS ACCESS ALL Poisson_Task'Class;


   PROCEDURE Initialize (
         A_Task : IN OUT Poisson_Task);


   PROCEDURE Put (
         My_Task : IN     Poisson_Task);

   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Poisson_Task'Class, Poisson_Task_Ptr);

   --------------------------------------------------


   TYPE Sporadic_Task IS NEW Poisson_Task WITH NULL RECORD;

   TYPE Sporadic_Task_Ptr IS ACCESS ALL Sporadic_Task'Class;

   PROCEDURE Initialize (
         A_Task : IN OUT Sporadic_Task);

   PROCEDURE Put (
         My_Task : IN     Sporadic_Task);

   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Sporadic_Task'Class, Sporadic_Task_Ptr);


   --------------------------------------------------



   TYPE Parametric_Task IS NEW Poisson_Task WITH
      RECORD
         Activation_Rule : Unbounded_String := To_Unbounded_String ("");
      END RECORD;

   TYPE Parametric_Task_Ptr IS ACCESS ALL Parametric_Task'Class;


   PROCEDURE Initialize (
         A_Task : IN OUT Parametric_Task);


   PROCEDURE Put (
         My_Task : IN     Parametric_Task);


   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Parametric_Task'Class, Parametric_Task_Ptr);

   --------------------------------------------------


   TYPE Task_Occurence IS
      RECORD
         Parent_Task       : Task_Ptr;
         Arrival_Date      : Natural  := 0;
         Residual_Capacity : Natural  := 0;
         Response_Time     : Natural  := 0;
         Absolute_Deadline : Natural  := 0;
         Is_Active         : Boolean  := False;
      END RECORD;


   PROCEDURE Put (
         My_Task_Occurence : IN     Task_Occurence);


   PROCEDURE Initialize_Cyclic (
         My_Task_Occurence : IN OUT Task_Occurence;
         A_Task            : IN     Task_Ptr;
         Arrival_Date      : IN     Natural        := 0);


   PROCEDURE Initialize_Acyclic (
         My_Task_Occurence : IN OUT Task_Occurence;
         A_Task            : IN     Aperiodic_Task_Ptr);


END Tasks;





