------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developped by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Translate;
USE Translate;
WITH Translate.English_Labels;
WITH Translate.Francais_Labels;

WITH Text_Io;
USE Text_Io;

WITH Processors;
USE Processors;
WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Unbounded_Strings;
USE Unbounded_Strings;
USE Unbounded_Strings.Strings_Table_Package;
USE Unbounded_Strings.Unbounded_String_List_Package;

WITH Parameters;
USE Parameters;
USE Parameters.Framework_Parameters_Table_Package;

WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;

WITH Systems;
USE Systems;

WITH Call_Framework;
USE Call_Framework;
USE Call_Framework.Framework_Response_Package;
USE Call_Framework.Framework_Request_Package;

WITH   Call_Scheduling_Framework;
USE   Call_Scheduling_Framework;

WITH Multiprocessor_Services;
USE Multiprocessor_Services;
USE Multiprocessor_Services.Scheduling_Result_Per_Processor;

WITH Config;
USE Config;





PROCEDURE Scheduling_Simulation IS


   PROCEDURE Call_Processor_Analysis (
         A_Sys : IN OUT System;
         Name  : IN     Unbounded_String) IS

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;
      A_Param       : Parameter_Ptr;

   BEGIN
      Initialize(Response_List);
      Initialize(Request_List);


      Initialize(A_Request);
      A_Request.Target:=Name;
      A_Request.Statement:=Scheduling_Simulation_Preemption_Number;
      Add(Request_List, A_Request);

      Initialize(A_Request);
      A_Request.Target:=Name;
      A_Request.Statement:=
         Scheduling_Simulation_Context_Switch_Number;
      Add(Request_List, A_Request);


      Sequential_Framework_Request (
         A_Sys, Request_List, Response_List);


      Put_Line(To_String(Response_List.Entries(0).
            Title));
      FOR J IN 0..Response_List.Nb_Entries-1 LOOP
         Put_Line(To_String(Response_List.Entries(J).
               Text));
      END LOOP;

   END Call_Processor_Analysis;



   PROCEDURE Compute_Scheduling (
         A_Sys : IN OUT System) IS

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;
      A_Param       : Parameter_Ptr;

   BEGIN

      Initialize(Response_List);
      Initialize(Request_List);
      Initialize(A_Request);
      A_Request.Statement:=Scheduling_Simulation_Time_Line;
      A_Param := NEW Parameter(Integer_Parameter);
      A_Param.Name:=To_Unbounded_String("period");
      A_Param.Integer_Value:=10000;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Boolean_Parameter);
      A_Param.Name:=To_Unbounded_String("schedule_with_offsets");
      A_Param.Boolean_Value:=True;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Boolean_Parameter);
      A_Param.Name:=To_Unbounded_String(
         "schedule_with_precedencies");
      A_Param.Boolean_Value:=True;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Boolean_Parameter);
      A_Param.Name:=To_Unbounded_String("schedule_with_resources");
      A_Param.Boolean_Value:=True;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Integer_Parameter);
      A_Param.Name:=To_Unbounded_String("seed_value");
      A_Param.Integer_Value:=0;
      Add(A_Request.Param, A_Param);
      Add(Request_List, A_Request);
      Sequential_Framework_Request (
         A_Sys, Request_List, Response_List);



      Put_Line(To_String(Response_List.Entries(0).
            Title));
      FOR J IN 0..Response_List.Nb_Entries-1 LOOP
         Put_Line(To_String(Response_List.Entries(J).
               Text));
      END LOOP;


   END Compute_Scheduling;




   Sys : System;

   Project_File_List     : Unbounded_String_List;
   A_Processor           : Processor_Ptr;
   My_Processor_Iterator : Processors_Iterator;
   Project_File_Dir_List : Unbounded_String_List;


BEGIN

   -- Initialize labels
   --
   Translate.Initialize;
   IF Current_Language=English THEN
      Translate.English_Labels.Initialize;
   ELSE
      Translate.Francais_Labels.Initialize;
   END IF;



   -- Read project file
   --
   Initialize(Project_File_List);
   Systems.Read_From_Xml_File(Sys, Project_File_Dir_List,
      To_Unbounded_String("uniform.xml"));


   -- Compute the scheduling
   --
   Compute_Scheduling(Sys);

   -- Perform analysis on the computed scheduling for each processor
   --
   Reset_Iterator(Sys.Processors, My_Processor_Iterator);
   LOOP
      Current_Element(Sys.Processors, A_Processor,
         My_Processor_Iterator);

      Call_Processor_Analysis(Sys, A_Processor.Name);

      EXIT WHEN Is_Last_Element(Sys.Processors,
         My_Processor_Iterator);
      Next_Element(Sys.Processors, My_Processor_Iterator);
   END LOOP;


END Scheduling_Simulation;




