
-----------------------------------------------------------------------
--          GtkAda - Ada95 binding for the Gimp Toolkit              --
--                                                                   --
--                     Copyright (C) 1998-1999                       --
--        Emmanuel Briot, Joel Brobecker and Arnaud Charlet          --
--                                                                   --
-- This library is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public               --
-- License as published by the Free Software Foundation; either      --
-- version 2 of the License, or (at your option) any later version.  --
--                                                                   --
-- This library is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of    --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU --
-- General Public License for more details.                          --
--                                                                   --
-- You should have received a copy of the GNU General Public         --
-- License along with this library; if not, write to the             --
-- Free Software Foundation, Inc., 59 Temple Place - Suite 330,      --
-- Boston, MA 02111-1307, USA.                                       --
--                                                                   --
-- As a special exception, if other files instantiate generics from  --
-- this unit, or you link this unit with other files to produce an   --
-- executable, this  unit  does not  by itself cause  the resulting  --
-- executable to be covered by the GNU General Public License. This  --
-- exception does not however invalidate any other reasons why the   --
-- executable file  might be covered by the  GNU Public License.     --
-----------------------------------------------------------------------

WITH Glib;
USE Glib;
WITH Gtk.Enums;
USE Gtk.Enums;
WITH Gtk.Style;
USE Gtk.Style;
WITH Gdk.Drawable;
USE Gdk.Drawable;
WITH Gdk.Event;
USE Gdk.Event;
WITH Gdk.Window;
USE Gdk.Window;
WITH Gdk.Rectangle;
USE Gdk.Rectangle;
WITH Gtk.Drawing_Area;
USE Gtk.Drawing_Area;
WITH Gtk.Handlers;
USE Gtk.Handlers;

PACKAGE BODY Double_Buffer IS

   PACKAGE Event_Cb IS NEW Gtk.Handlers.Return_Callback
      (
      Widget_Type => Gtk_Double_Buffer_Record,
      Return_Type => Boolean);
   PACKAGE Void_Cb IS NEW Gtk.Handlers.Callback
      (Widget_Type => Gtk_Double_Buffer_Record);

   FUNCTION Configure (
         Buffer : ACCESS Gtk_Double_Buffer_Record'Class;
         Event  :        Gdk.Event.Gdk_Event)
     RETURN Boolean;
   --  Callback when the buffer is resized

   FUNCTION Expose (
         Buffer : ACCESS Gtk_Double_Buffer_Record'Class;
         Event  :        Gdk.Event.Gdk_Event)
     RETURN Boolean;
   --  Callback when the window needs to redraw itself on the screen

   FUNCTION Create_Internal_Pixmap (
         Buffer : ACCESS Gtk_Double_Buffer_Record'Class)
     RETURN Gdk.Pixmap.Gdk_Pixmap;
   --  Create one of the internal pixmaps used by the buffer, with the
   --  correct size.

   PROCEDURE Destroy_Cb (
         Buffer : ACCESS Gtk_Double_Buffer_Record'Class);
   --  Called when the widget is destroyed, to free up some memory

   -------------
   -- Gtk_New --
   -------------

   PROCEDURE Gtk_New (
         Buffer :    OUT Gtk_Double_Buffer) IS
   BEGIN
      Buffer := NEW Gtk_Double_Buffer_Record;
      Double_Buffer.Initialize (Buffer);
   END Gtk_New;

   ----------------
   -- Initialize --
   ----------------

   PROCEDURE Initialize (
         Buffer : ACCESS Gtk_Double_Buffer_Record'Class) IS
   BEGIN
      Gtk.Drawing_Area.Initialize (Buffer);

      --  Note: the pixmap is created in the resize routine
      --  and the GC is created in the Realize_Cb routine.

      --  Connect the signals
      Event_Cb.Connect (Buffer, "configure_event",
         Event_Cb.To_Marshaller (Configure'ACCESS));
      Event_Cb.Connect (Buffer, "expose_event",
         Event_Cb.To_Marshaller (Expose'ACCESS));
      Void_Cb.Connect (Buffer, "destroy",
         Void_Cb.To_Marshaller (Destroy_Cb'ACCESS));
   END Initialize;

   ----------------------------
   -- Create_Internal_Pixmap --
   ----------------------------

   FUNCTION Create_Internal_Pixmap (
         Buffer : ACCESS Gtk_Double_Buffer_Record'Class)
     RETURN Gdk.Pixmap.Gdk_Pixmap IS
      Pix : Gdk.Pixmap.Gdk_Pixmap;
   BEGIN
      Gdk.Pixmap.Gdk_New (Pix,
         Get_Window (Buffer),
         Gint (Get_Allocation_Width (Buffer)),
         Gint (Get_Allocation_Height (Buffer)));
      RETURN Pix;
   END Create_Internal_Pixmap;

   ---------------
   -- Configure --
   ---------------

   FUNCTION Configure (
         Buffer : ACCESS Gtk_Double_Buffer_Record'Class;
         Event  :        Gdk.Event.Gdk_Event)
     RETURN Boolean IS
      Old_Pixmap : CONSTANT Gdk.Pixmap.Gdk_Pixmap := Buffer.Pixmap;
      Old_Triple : CONSTANT Gdk.Pixmap.Gdk_Pixmap := Buffer.Triple_Buffer;

      USE TYPE Gdk.Pixmap.Gdk_Pixmap;

   BEGIN
      Buffer.Pixmap := Create_Internal_Pixmap (Buffer);

      IF Buffer.Use_Triple_Buffer THEN
         Buffer.Triple_Buffer := Create_Internal_Pixmap (Buffer);
      END IF;

      IF Old_Pixmap /= NULL THEN

         --  If Back_Store is True, then we want to keep the content
         --  of the old pixmap, and thus we copy it to the new one.

         IF Buffer.Back_Store THEN
            DECLARE
               Old_Width  : Gint;
               Old_Height : Gint;

            BEGIN
               Get_Size (Gdk_Window (Old_Pixmap), Old_Width, Old_Height);

               Gdk.Drawable.Draw_Rectangle
                  (Buffer.Pixmap,
                  Get_Background_Gc (Get_Style (Buffer),
                     State_Normal),
                  Filled => True,
                  X      => 0,
                  Y      => 0,
                  Width  => Gdk.Event.Get_Width (Event),
                  Height => Gdk.Event.Get_Height (Event));

               Gdk.Drawable.Copy_Area (Buffer.Pixmap,
                  Get_Foreground_Gc (Get_Style (Buffer),
                     State_Normal),
                  X        => 0,
                  Y        => 0,
                  Source   => Old_Pixmap,
                  Source_X => 0,
                  Source_Y => 0,
                  Width    => Old_Width,
                  Height   => Old_Height);
            END;
         END IF;

         Gdk.Pixmap.Unref (Old_Pixmap);
      END IF;

      IF Old_Triple /= NULL THEN

         --  If Back_Store is True, then we want to keep the content
         --  of the old pixmap, and thus we copy it to the new one.

         IF Buffer.Back_Store THEN
            DECLARE
               Old_Width  : Gint;
               Old_Height : Gint;

            BEGIN
               Get_Size (Gdk_Window (Old_Triple), Old_Width, Old_Height);

               Gdk.Drawable.Draw_Rectangle
                  (Buffer.Triple_Buffer,
                  Get_Background_Gc (Get_Style (Buffer),
                     State_Normal),
                  Filled => True,
                  X      => 0,
                  Y      => 0,
                  Width  => Gdk.Event.Get_Width (Event),
                  Height => Gdk.Event.Get_Height (Event));

               Gdk.Drawable.Copy_Area (Buffer.Triple_Buffer,
                  Get_Foreground_Gc (Get_Style (Buffer),
                     State_Normal),
                  X        => 0,
                  Y        => 0,
                  Source   => Old_Triple,
                  Source_X => 0,
                  Source_Y => 0,
                  Width    => Old_Width,
                  Height   => Old_Height);
            END;
         END IF;

         Gdk.Pixmap.Unref (Old_Triple);
      END IF;

      RETURN True;
   END Configure;

   ----------------
   -- Destroy_Cb --
   ----------------

   PROCEDURE Destroy_Cb (
         Buffer : ACCESS Gtk_Double_Buffer_Record'Class) IS
      USE TYPE Gdk.Pixmap.Gdk_Pixmap;
   BEGIN
      IF Buffer.Pixmap /= NULL THEN
         Gdk.Pixmap.Unref (Buffer.Pixmap);
      END IF;

      IF Buffer.Triple_Buffer /= NULL THEN
         Gdk.Pixmap.Unref (Buffer.Triple_Buffer);
      END IF;
   END Destroy_Cb;

   ------------
   -- Expose --
   ------------

   FUNCTION Expose (
         Buffer : ACCESS Gtk_Double_Buffer_Record'Class;
         Event  :        Gdk.Event.Gdk_Event)
     RETURN Boolean IS
      Area : CONSTANT Gdk.Rectangle.Gdk_Rectangle := Gdk.Event.Get_Area (Event);








   BEGIN
      IF Buffer.Use_Triple_Buffer THEN
         --  If the event was generated manually, we need to update
         --  the Triple_Buffer with the value of the double_buffer,
         --  unless of course the buffer is frozen.

         IF Gdk.Event.Get_Send_Event (Event) THEN
            IF Buffer.Is_Frozen THEN
               Buffer.Should_Update_On_Screen := True;
            ELSE
               Gdk.Drawable.Copy_Area
                  (Buffer.Triple_Buffer,
                  Get_Foreground_Gc (Get_Style (Buffer),
                     State_Normal),
                  X        => 0,
                  Y        => 0,
                  Source   => Buffer.Pixmap,
                  Source_X => 0,
                  Source_Y => 0,
                  Width    => Gint (Get_Allocation_Width (Buffer)),
                  Height   => Gint (Get_Allocation_Height (Buffer)));
            END IF;
         END IF;

         --  Copy the triple buffer to the screen
         Gdk.Drawable.Copy_Area (Get_Window (Buffer),
            Get_Foreground_Gc (Get_Style (Buffer),
               State_Normal),
            X        => Area.X,
            Y        => Area.Y,
            Source   => Buffer.Triple_Buffer,
            Source_X => Area.X,
            Source_Y => Area.Y,
            Width    => Gint (Area.Width),
            Height   => Gint (Area.Height));

      ELSIF NOT Buffer.Is_Frozen THEN
         --  Copy the double buffer to the screen
         Gdk.Drawable.Copy_Area (Get_Window (Buffer),
            Get_Foreground_Gc (Get_Style (Buffer),
               State_Normal),
            X        => Area.X,
            Y        => Area.Y,
            Source   => Buffer.Pixmap,
            Source_X => Area.X,
            Source_Y => Area.Y,
            Width    => Gint (Area.Width),
            Height   => Gint (Area.Height));

      ELSE
         Buffer.Should_Update_On_Screen := True;
      END IF;

      RETURN False;
   END Expose;

   ----------------
   -- Get_Pixmap --
   ----------------

   FUNCTION Get_Pixmap (
         Buffer : ACCESS Gtk_Double_Buffer_Record)
     RETURN Gdk.Drawable.Gdk_Drawable IS
   BEGIN
      RETURN Gdk.Drawable.Gdk_Drawable (Buffer.Pixmap);
   END Get_Pixmap;

   --------------------
   -- Set_Back_Store --
   --------------------

   PROCEDURE Set_Back_Store (
         Buffer     : ACCESS Gtk_Double_Buffer_Record;
         Back_Store :        Boolean                  := True) IS
   BEGIN
      Buffer.Back_Store := Back_Store;
   END Set_Back_Store;

   ------------
   -- Freeze --
   ------------

   PROCEDURE Freeze (
         Buffer : ACCESS Gtk_Double_Buffer_Record) IS
   BEGIN
      Buffer.Is_Frozen := True;
   END Freeze;

   ----------
   -- Thaw --
   ----------

   PROCEDURE Thaw (
         Buffer : ACCESS Gtk_Double_Buffer_Record) IS
   BEGIN
      Buffer.Is_Frozen := False;

      IF Buffer.Should_Update_On_Screen THEN
         Draw (Buffer);
      END IF;
   END Thaw;

   -----------------------
   -- Set_Triple_Buffer --
   -----------------------

   PROCEDURE Set_Triple_Buffer (
         Buffer            : ACCESS Gtk_Double_Buffer_Record;
         Use_Triple_Buffer :        Boolean                  := True) IS
      USE TYPE Gdk.Gdk_Window;
   BEGIN
      Buffer.Use_Triple_Buffer := Use_Triple_Buffer;

      --  If we do not want triple_buffer but already have one, delete it
      IF NOT Use_Triple_Buffer
            AND THEN Buffer.Triple_Buffer /= NULL
            THEN
         Gdk.Pixmap.Unref (Buffer.Triple_Buffer);
      END IF;

      --  If we do want a triple buffer, create the pixmap

      IF Use_Triple_Buffer THEN
         Buffer.Triple_Buffer := Create_Internal_Pixmap (Buffer);

         --  Copy the current value of the pixmap in it
         Gdk.Drawable.Copy_Area
            (Buffer.Triple_Buffer,
            Get_Foreground_Gc (Get_Style (Buffer),
               State_Normal),
            X        => 0,
            Y        => 0,
            Source   => Buffer.Pixmap,
            Source_X => 0,
            Source_Y => 0,
            Width    => Gint (Get_Allocation_Width (Buffer)),
            Height   => Gint (Get_Allocation_Height (Buffer)));
      END IF;
   END Set_Triple_Buffer;

END Double_Buffer;
