------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Gtk.Drawing_Area;
USE Gtk.Drawing_Area;
WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Gdk.Drawable;
USE Gdk.Drawable;
WITH Gtk.Handlers;
USE Gtk.Handlers;
WITH Gtk.Arguments;
USE Gtk.Arguments;
WITH Gtk.Object;
USE Gtk.Object;
WITH Gdk;
USE Gdk;
WITH Gtkada.Types;
USE Gtkada.Types;
WITH System;

WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;

WITH Framework_Config;
USE Framework_Config;
WITH Translate;
USE Translate;

WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;



PACKAGE BODY Draw_Buffer_Pkg IS

   --  This pointer will keep a pointer to the C 'class record' for
   --  gtk. To avoid allocating memory for each widget, this may be done
   --  only once, and reused
   Class_Record : System.Address := System.Null_Address;




   PACKAGE Draw_Cb IS NEW Gtk.Handlers.Return_Callback (
      Widget_Type => Draw_Buffer_Record,
      Return_Type => Boolean);
   PACKAGE Allocation_Cb IS NEW Handlers.Callback (
      Draw_Buffer_Record);
   PACKAGE Allocation_Marshaller IS NEW
      Allocation_Cb.Marshallers.Generic_Marshaller
      (Gtk_Allocation_Access, To_Allocation);



   PROCEDURE Gtk_New (
         Draw_Buffer :    OUT Draw_Buffer_Access) IS

   BEGIN
      Draw_Buffer:=NEW Draw_Buffer_Record;
      Draw_Buffer_Pkg.Initialize (Draw_Buffer);
   END Gtk_New;


   PROCEDURE Initialize (
         Draw_Buffer : ACCESS Draw_Buffer_Record'Class) IS



   BEGIN

      Draw_Buffer.Current_Width:=Draw_Buffer.First_Width;
      Draw_Buffer.Current_Height:=Draw_Buffer.First_Height;



      --  We need to call the ancestor's Initialize function to create
      --  the underlying C object.
      Double_Buffer.Initialize (Draw_Buffer);

      --  The following call is required to initialize the class record,
      --  and the new signals created for this widget.
      --  Note that for now you can only create basic signals (whose
      --  callbacks do not have any parameters), and that you just have
      --  to put their name in a table.
      --  Note also that we keep Class_Record, so that the memory allocation
      --  is done only once.
      Gtk.Object.Initialize_Class_Record (Draw_Buffer, Null_Array,
         Class_Record);



      Gtk.Drawing_Area.Set_Usize (Gtk_Drawing_Area(Draw_Buffer),
         Draw_Buffer.First_Width, Draw_Buffer.First_Height);


      Load (Draw_Buffer.Font,
         "-adobe-helvetica-medium-o-normal--14-140-75-75-p-78-iso8859-1");

      Draw_Cb.Connect (Draw_Buffer, "configure_event",
         Draw_Cb.To_Marshaller (Configure'ACCESS));
      Allocation_Cb.Connect (Draw_Buffer, "size_allocate",
         Allocation_Marshaller.To_Marshaller (Size_Allocate'ACCESS));


   END Initialize;



   PROCEDURE Color_Allocation (
         Draw_Buffer : ACCESS Draw_Buffer_Record'Class) IS
   BEGIN

      IF NOT Draw_Buffer.Have_Colors THEN

         Gdk.Gc.Gdk_New (Draw_Buffer.White_Gc, Get_Window (
               Draw_Buffer));
         Gdk.Gc.Set_Foreground
            (Draw_Buffer.White_Gc, Gdk.Color.White (Gtk.Widget.
               Get_Default_Colormap));

         Gdk.Gc.Gdk_New (Draw_Buffer.Black_Gc, Get_Window (
               Draw_Buffer));
         Gdk.Gc.Set_Foreground
            (Draw_Buffer.Black_Gc, Gdk.Color.Black (Gtk.Widget.
               Get_Default_Colormap));

         Draw_Buffer.Red_Color := Gdk.Color.Parse ("Red");
         Gdk.Color.Alloc (Gtk.Widget.Get_Default_Colormap, Draw_Buffer.
            Red_Color);


         Draw_Buffer.Have_Colors:=True;

      END IF;
   END Color_Allocation;


   PROCEDURE Clear (
         Draw_Buffer : ACCESS Draw_Buffer_Record'Class) IS
   BEGIN


      Color_Allocation(Draw_Buffer);

      Draw_Buffer.Initialized:=False;


      Draw_Rectangle (Get_Pixmap(Draw_Buffer), Draw_Buffer.
         White_Gc, Filled => True,
         X     => 0,   Y      => 0,
         Width => Draw_Buffer.Current_Width, Height => Draw_Buffer.
         Current_Height);

      Double_Buffer.Draw (Gtk_Double_Buffer(Draw_Buffer));

   END Clear;


   PROCEDURE Draw (
         Draw_Buffer       : ACCESS Draw_Buffer_Record'Class;
         Buff              :        Buffer_Ptr;
         Buff1             :        Buffer_Size_Table;
         Last_Time_Unit1   :        Natural;
         Time_Unit_Number1 :        Time_Unit_Range) IS

      I  : Gint    := 0;
      Ok : Boolean;

   BEGIN

      Draw_Buffer.Last_Time_Unit:=Last_Time_Unit1;
      Draw_Buffer.Time_Unit_Number:=Time_Unit_Number1;
      Draw_Buffer.Buff:= NEW Buffer_Size_Table;
      Draw_Buffer.Buff.All:=Buff1;
      Draw_Buffer.Buff_Ptr:=Buff;


      Ok:=False;
      I:=Gint(Draw_Buffer.Last_Time_Unit+4)*
         Draw_Buffer.Timeunit_Margin+Draw_Buffer.Left_Margin;
      IF I > Draw_Buffer.Current_Width
            THEN
         Draw_Buffer.Current_Width:=I;
         Ok:=True;
      END IF;


      I:=Draw_Buffer.High_Margin+Draw_Buffer.Buffer_Margin*4;
      IF I > Draw_Buffer.Current_Height
            THEN
         Draw_Buffer.Current_Height:=I;
         Ok:=True;
      END IF;

      IF Ok THEN
         Gtk.Drawing_Area.Set_Usize (Gtk_Drawing_Area(Draw_Buffer),
            Draw_Buffer.Current_Width, Draw_Buffer.Current_Height);
      END IF;



      Draw_Buffer.Initialized:=True;

      Redraw(Draw_Buffer);


   END Draw;


   PROCEDURE Redraw (
         Draw_Buffer : ACCESS Draw_Buffer_Record'Class) IS


      Pixel_Buffer_Size : Gint             := 0;
      Pixel_Buffer_Time : Gint             := 0;
      Msg               : Unbounded_String;


   BEGIN

      ---------------------------------
      -- Clear work space
      ---------------------------------


      Draw_Rectangle (Get_Pixmap(Draw_Buffer), Draw_Buffer.
         White_Gc, Filled => True,
         X     => 0,   Y      => 0,
         Width => Draw_Buffer.Current_Width, Height => Draw_Buffer.
         Current_Height);


      ----------------------------------------------------------
      -- Draw buffer utilization
      --    in red color
      ----------------------------------------------------------


      FOR J IN 0..Draw_Buffer.Time_Unit_Number-2  LOOP
         Pixel_Buffer_Time:=Gint(Draw_Buffer.Buff.Entries(
               Buffer_Size_Range(J+1)).
            Time - Draw_Buffer.Buff.Entries(Buffer_Size_Range(J)).Time);
         Pixel_Buffer_Time:=Pixel_Buffer_Time*
            Draw_Buffer.Timeunit_Margin;


         Pixel_Buffer_Size:=Gint(Double(Draw_Buffer.Buff.Entries(
                  Buffer_Size_Range(J)).Size)*100.0/Double(Draw_Buffer.
               Buff_Ptr.Size));
         Pixel_Buffer_Size:=Pixel_Buffer_Size*Draw_Buffer.Buffer_Margin/
            10;

         Draw_Rectangle (Get_Pixmap(Draw_Buffer),
            Draw_Buffer.Black_Gc, True,

            Draw_Buffer.Left_Margin +
            (Gint(Draw_Buffer.Buff.Entries(Buffer_Size_Range(J)).Time)*
               Draw_Buffer.
               Timeunit_Margin),

            Draw_Buffer.High_Margin+20+Gint(10)*
            Draw_Buffer.Buffer_Margin-Pixel_Buffer_Size,
            Pixel_Buffer_Time,
            Pixel_Buffer_Size);
      END LOOP;


      -- Draw the last buffer size unit up to the end of
      -- the x axis
      --
      Pixel_Buffer_Time:=Gint(Draw_Buffer.Last_Time_Unit
         - Draw_Buffer.Buff.Entries(Buffer_Size_Range(Draw_Buffer.
               Time_Unit_Number-1)).Time);
      Pixel_Buffer_Time:=Pixel_Buffer_Time*
         Draw_Buffer.Timeunit_Margin;


      Pixel_Buffer_Size:=Gint(Double(Draw_Buffer.Buff.Entries(
               Buffer_Size_Range(Draw_Buffer.Time_Unit_Number-1)).Size)*
         100.0/Double(Draw_Buffer.
            Buff_Ptr.Size));
      Pixel_Buffer_Size:=Pixel_Buffer_Size*Draw_Buffer.Buffer_Margin/
         10;

      Draw_Rectangle (Get_Pixmap(Draw_Buffer),
         Draw_Buffer.Black_Gc, True,

         Draw_Buffer.Left_Margin +
         (Gint(Draw_Buffer.Buff.Entries(Buffer_Size_Range(Draw_Buffer.
                     Time_Unit_Number-1)).Time)*
            Draw_Buffer.
            Timeunit_Margin),

         Draw_Buffer.High_Margin+20+Gint(10)*
         Draw_Buffer.Buffer_Margin-Pixel_Buffer_Size,
         Pixel_Buffer_Time,
         Pixel_Buffer_Size);

      Gdk.Gc.Set_Foreground
         (Draw_Buffer.Black_Gc, Gdk.Color.Black (Gtk.Widget.
            Get_Default_Colormap));

      ---------------------------------
      -- Draw x and y axis
      ---------------------------------

      Msg:=Lb_Time(Current_Language);
      Draw_Text (Get_Pixmap(Draw_Buffer), Draw_Buffer.
         Font,
         Draw_Buffer.Black_Gc,
         Draw_Buffer.Left_Margin+20+Gint(Draw_Buffer.Last_Time_Unit)*
         Draw_Buffer.Timeunit_Margin,
         Draw_Buffer.High_Margin-35+Gint(12)*
         Draw_Buffer.
         Buffer_Margin,
         To_String(Msg));

      Msg:=Lb_Utilization_Level(Current_Language);
      Draw_Text (Get_Pixmap(Draw_Buffer), Draw_Buffer.
         Font,
         Draw_Buffer.Black_Gc,
         Draw_Buffer.Left_Margin,
         Draw_Buffer.High_Margin,
         To_String(Msg));

      Draw_Line (Get_Pixmap(Draw_Buffer),
         Draw_Buffer.Black_Gc,
         Draw_Buffer.Left_Margin,
         Draw_Buffer.High_Margin+20,
         Draw_Buffer.Left_Margin,
         Draw_Buffer.High_Margin+20+10* Draw_Buffer.Buffer_Margin);

      Draw_Line (Get_Pixmap(Draw_Buffer),
         Draw_Buffer.Black_Gc,
         Draw_Buffer.Left_Margin,
         Draw_Buffer.High_Margin+20+Gint(10)*
         Draw_Buffer.Buffer_Margin,
         Draw_Buffer.Left_Margin+Gint(Draw_Buffer.Last_Time_Unit)*
         Draw_Buffer.Timeunit_Margin,
         Draw_Buffer.High_Margin+20+Gint(10)*
         Draw_Buffer.Buffer_Margin);



      ----------------------------------------------------------
      -- Draw Time unit and buffer size unit
      --    in black color
      ----------------------------------------------------------

      FOR J IN 0 .. Gint(10)  LOOP
         Draw_Rectangle (Get_Pixmap(Draw_Buffer),
            Draw_Buffer.Black_Gc, True,
            Draw_Buffer.Left_Margin-5,
            Draw_Buffer.High_Margin+20+J*Draw_Buffer.
            Buffer_Margin,
            10, 2 );
      END LOOP;

      FOR J IN 0 .. Gint(Draw_Buffer.Last_Time_Unit-1)  LOOP
         Draw_Rectangle (Get_Pixmap(Draw_Buffer),
            Draw_Buffer.Black_Gc, True,
            Draw_Buffer.Left_Margin+20+J*Draw_Buffer.Timeunit_Margin,
            Draw_Buffer.High_Margin-5+Gint(11)*
            Draw_Buffer.
            Buffer_Margin,
            2, 10 );
      END LOOP;

      ---------------------
      ---------------------
      -- Flush data in the window
      ---------------------
      Draw (Draw_Buffer);

   END Redraw;





   ---------------
   -- Configure --
   ---------------

   FUNCTION Configure (
         Draw_Buffer : ACCESS Draw_Buffer_Record'Class;
         Event       :        Gdk.Event.Gdk_Event)
     RETURN Boolean IS


   BEGIN
      IF Draw_Buffer.Initialized
            THEN
         Redraw(Draw_Buffer);
      ELSE
         Color_Allocation(Draw_Buffer);

         -- Check if window size has changed !
         --

         Draw_Buffer.Current_Width:=Gint(Get_Allocation_Width(
               Draw_Buffer));
         Draw_Buffer.Current_Height:=Gint(Get_Allocation_Height(
               Draw_Buffer));

         Draw_Rectangle (Get_Pixmap(Draw_Buffer),
            Draw_Buffer.White_Gc, Filled => True,
            X     => 0,   Y      => 0,
            Width => Draw_Buffer.Current_Width, Height =>
            Draw_Buffer.Current_Height);

         Double_Buffer.Draw (Gtk_Double_Buffer(Draw_Buffer));

      END IF;

      RETURN True;
   END Configure;


   PROCEDURE Size_Allocate (
         Draw_Buffer : ACCESS Draw_Buffer_Record'Class;
         Allocation  :        Gtk_Allocation_Access) IS
   BEGIN
      IF Realized_Is_Set (Draw_Buffer) THEN
         Draw_Buffer.Current_Width:=Gint(Get_Allocation_Width(
               Draw_Buffer));
         Draw_Buffer.Current_Height:=Gint(Get_Allocation_Height(
               Draw_Buffer));
         IF Draw_Buffer.Initialized
               THEN
            Redraw(Draw_Buffer);
         ELSE
            Color_Allocation(Draw_Buffer);

            Draw_Rectangle (Get_Pixmap(Draw_Buffer),
               Draw_Buffer.White_Gc, Filled => True,
               X     => 0,   Y      => 0,
               Width => Draw_Buffer.Current_Width, Height =>
               Draw_Buffer.Current_Height);

            Draw (Draw_Buffer);

         END IF;
      END IF;
      Gtk.Handlers.Emit_Stop_By_Name (Draw_Buffer, "size_allocate");
   END Size_Allocate;


END Draw_Buffer_Pkg;





