------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Gdk.Drawable;
USE Gdk.Drawable;
WITH Gtk.Arguments;
USE Gtk.Arguments;
WITH Gtk.Drawing_Area;
USE Gtk.Drawing_Area;
WITH Gtk.Handlers;
USE Gtk.Handlers;


WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;

WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;

WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;
WITH Tasks;
USE Tasks;

WITH Resource_Set;
USE Resource_Set;
USE Resource_Set.Generic_Resource_Set;
WITH Resources;
USE Resources;

WITH Message_Set;
USE Message_Set;
USE Message_Set.Generic_Message_Set;
WITH Messages;
USE Messages;

WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;
WITH Buffers;
USE Buffers;


WITH Time_Units;
USE Time_Units;
USE Time_Units.A_Time_Unit;

WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Text_Io;
USE Text_Io;

WITH Events;
USE Events;
WITH Framework_Config;
USE Framework_Config;
WITH Translate;
USE Translate;


WITH Select_Time_Line_Pkg;
USE Select_Time_Line_Pkg;
USE Select_Time_Line_Pkg.Generic_Time_Line_Object_Set;
WITH Objects;
USE Objects;
WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;

WITH Editor_Config;
USE Editor_Config;


PACKAGE BODY Draw_Scheduling_Pkg IS

   PACKAGE Draw_Cb IS NEW Gtk.Handlers.Return_Callback
      (
      Widget_Type => Gtk_Double_Buffer_Record,
      Return_Type => Boolean);

   PACKAGE Allocation_Cb IS NEW Handlers.Callback (
      Gtk_Double_Buffer_Record);
   PACKAGE Allocation_Marshaller IS NEW
      Allocation_Cb.Marshallers.Generic_Marshaller
      (Gtk_Allocation_Access, To_Allocation);



   PROCEDURE Color_Allocation (
         Draw_Scheduling : ACCESS Draw_Scheduling_Record) IS
   BEGIN

      IF NOT Draw_Scheduling.Have_Colors THEN

         Gdk.Gc.Gdk_New (Draw_Scheduling.White_Gc, Get_Window (
               Draw_Scheduling.Buffer));
         Gdk.Gc.Set_Foreground
            (Draw_Scheduling.White_Gc, Gdk.Color.White (Gtk.Widget.
               Get_Default_Colormap));

         Gdk.Gc.Gdk_New (Draw_Scheduling.Black_Gc, Get_Window (
               Draw_Scheduling.Buffer));
         Gdk.Gc.Set_Foreground
            (Draw_Scheduling.Black_Gc, Gdk.Color.Black (Gtk.Widget.
               Get_Default_Colormap));

         Draw_Scheduling.Red_Color := Gdk.Color.Parse ("Red");
         Gdk.Color.Alloc (Gtk.Widget.Get_Default_Colormap,
            Draw_Scheduling.
            Red_Color);

         Draw_Scheduling.Blue_Color := Gdk.Color.Parse ("Blue");
         Gdk.Color.Alloc (Gtk.Widget.Get_Default_Colormap,
            Draw_Scheduling.
            Blue_Color);


         Draw_Scheduling.Have_Colors:=True;

      END IF;
   END Color_Allocation;


   FUNCTION Configure (
         Buffer : ACCESS Gtk_Double_Buffer_Record'Class;
         Event  :        Gdk.Event.Gdk_Event)
     RETURN Boolean IS


   BEGIN
      IF Draw_Scheduling.Initialized
            THEN
         Redraw_Time_Line(Draw_Scheduling);
      ELSE
         Color_Allocation(Draw_Scheduling);

         -- Check if window size has changed !
         --

         Draw_Scheduling.Current_Width:=Gint(Get_Allocation_Width(
               Draw_Scheduling.Buffer));
         Draw_Scheduling.Current_Height:=Gint(Get_Allocation_Height(
               Draw_Scheduling.Buffer));

         Draw_Rectangle (Get_Pixmap(Draw_Scheduling.Buffer),
            Draw_Scheduling.White_Gc, Filled => True,
            X     => 0,   Y      => 0,
            Width => Draw_Scheduling.Current_Width, Height =>
            Draw_Scheduling.Current_Height);

         Double_Buffer.Draw (Draw_Scheduling.Buffer);

      END IF;

      RETURN True;
   END Configure;


   --  This function is called once gtk has decided what size and position
   --  the widget will actually have, or everytime the widget is resized.
   --  This would be a good time for instance for resizing the component
   --  sub-widgets.
   --  Note that we have to move and resize the widget ourselves, and that
   --  for such a simple case, we could simply rely on the ancestor's
   --  size_allocation function.
   --
   PROCEDURE Size_Allocate (
         Buffer     : ACCESS Gtk_Double_Buffer_Record'Class;
         Allocation :        Gtk_Allocation_Access) IS
   BEGIN
      IF Realized_Is_Set (Buffer) THEN
         Draw_Scheduling.Current_Width:=Gint(Get_Allocation_Width(
               Draw_Scheduling.Buffer));
         Draw_Scheduling.Current_Height:=Gint(Get_Allocation_Height(
               Draw_Scheduling.Buffer));
         IF Draw_Scheduling.Initialized
               THEN
            Redraw_Time_Line(Draw_Scheduling);
         ELSE
            Color_Allocation(Draw_Scheduling);

            Draw_Rectangle (Get_Pixmap(Draw_Scheduling.Buffer),
               Draw_Scheduling.White_Gc, Filled => True,
               X     => 0,   Y      => 0,
               Width => Draw_Scheduling.Current_Width, Height =>
               Draw_Scheduling.Current_Height);

            Double_Buffer.Draw (Draw_Scheduling.Buffer);

         END IF;
      END IF;
      Gtk.Handlers.Emit_Stop_By_Name (Buffer, "size_allocate");
   END Size_Allocate;



   PROCEDURE Gtk_New (
         Draw_Scheduling :    OUT Draw_Scheduling_Access) IS

   BEGIN
      Draw_Scheduling:=NEW Draw_Scheduling_Record;
      Draw_Scheduling_Pkg.Initialize (Draw_Scheduling);
   END Gtk_New;


   PROCEDURE Initialize (
         Draw_Scheduling : ACCESS Draw_Scheduling_Record'Class) IS

   BEGIN


      Draw_Scheduling.Current_Width:=Draw_Scheduling.First_Width;
      Draw_Scheduling.Current_Height:=Draw_Scheduling.First_Height;

      Gtk_New(Draw_Scheduling.Buffer);

      Gtk.Drawing_Area.Set_Usize (Gtk_Drawing_Area(Draw_Scheduling.Buffer),
         Draw_Scheduling.First_Width, Draw_Scheduling.First_Height);
      Show_All (Draw_Scheduling.Buffer);

      Load (Draw_Scheduling.Font,
         "-adobe-helvetica-medium-o-normal--14-140-75-75-p-78-iso8859-1");


      Draw_Cb.Connect (Draw_Scheduling.Buffer, "configure_event",
         Draw_Cb.To_Marshaller (Configure'ACCESS));
      Allocation_Cb.Connect (Draw_Scheduling.Buffer, "size_allocate",
         Allocation_Marshaller.To_Marshaller (Size_Allocate'ACCESS));

   END Initialize;



   PROCEDURE Clear_Time_Line (
         Draw_Scheduling : ACCESS Draw_Scheduling_Record) IS
   BEGIN


      Color_Allocation(Draw_Scheduling);

      Draw_Scheduling.Initialized:=False;


      Draw_Rectangle (Get_Pixmap(Draw_Scheduling.Buffer), Draw_Scheduling.
         White_Gc, Filled => True,
         X     => 0,   Y      => 0,
         Width => Draw_Scheduling.Current_Width, Height =>
         Draw_Scheduling.
         Current_Height);

      Double_Buffer.Draw (Draw_Scheduling.Buffer);

   END Clear_Time_Line;




   PROCEDURE Draw_Time_Line (
         Draw_Scheduling : ACCESS Draw_Scheduling_Record;
         Task1           :        Tasks_Set;
         Resource1       :        Resources_Set;
         Message1        :        Messages_Set;
         Buffer1         :        Buffers_Set;
         Period1         :        Natural;
         Start_Draw1     :        Natural;
         Sched1          :        Scheduling_Table_Ptr) IS

      I  : Gint    := 0;
      Ok : Boolean;


   BEGIN



      Draw_Scheduling.Tasks:=Task1 ;
      Draw_Scheduling.Resources:=Resource1;
      Draw_Scheduling.Messages:=Message1;
      Draw_Scheduling.Buffers:=Buffer1;
      Draw_Scheduling.Period:=Period1;
      Draw_Scheduling.Start_Draw:=Start_Draw1;
      Draw_Scheduling.Sched:=Sched1;


      -- Set Left_margin
      --
      Draw_Scheduling.Left_Margin:=20;



      -- Compute Widget size
      --
      Ok:=False;
      I:=Gint(Period1-Start_Draw1)*Draw_Scheduling.Timeunit_Margin+
         Draw_Scheduling.
         Left_Margin+Draw_Scheduling.Right_Margin;
      IF I > Draw_Scheduling.Current_Width
            THEN
         Draw_Scheduling.Current_Width:=I;
         Ok:=True;
      END IF;

      IF Is_Empty(Time_Line_To_Display)
            THEN
         I:=Draw_Scheduling.High_Margin+Draw_Scheduling.Interligne*Gint(
            Get_Number_Of_Elements(Task1));
         IF Schedule_With_Resources
               THEN
            I:=I+Draw_Scheduling.Interligne*Gint(
               Get_Number_Of_Elements(Resource1));
         END IF;
         IF Schedule_With_Precedencies
               THEN
            I:=I+Draw_Scheduling.Interligne*Gint(
               Get_Number_Of_Elements(Buffer1));
            I:=I+Draw_Scheduling.Interligne*Gint(
               Get_Number_Of_Elements(Message1));
         END IF;
      ELSE
         I:=Draw_Scheduling.Interligne*Gint(Get_Number_Of_Elements(
               Time_Line_To_Display));
      END IF;
      IF I > Draw_Scheduling.Current_Height
            THEN
         Draw_Scheduling.Current_Height:=I;
         Ok:=True;
      END IF;


      IF Ok THEN
         Gtk.Drawing_Area.Set_Usize (Gtk_Drawing_Area(Draw_Scheduling.
               Buffer),
            Draw_Scheduling.Current_Width, Draw_Scheduling.Current_Height);
      END IF;

      Draw_Scheduling.Initialized:=True;

      Redraw_Time_Line(Draw_Scheduling);


   END Draw_Time_Line;




   PROCEDURE Redraw_Time_Line (
         Draw_Scheduling : ACCESS Draw_Scheduling_Record'Class) IS


      My_Task_Iterator     : Tasks_Iterator;
      My_Resource_Iterator : Resources_Iterator;
      My_Buffer_Iterator   : Buffers_Iterator;
      My_Message_Iterator  : Messages_Iterator;
      A_Task               : Task_Ptr;
      A_Message            : Message_Ptr;
      A_Resource           : Resource_Ptr;
      A_Buffer             : Buffer_Ptr;

      I : Gint := 0;

      Msg,
      Name : Unbounded_String;

      TYPE Object_Range IS NEW Integer RANGE 0 ..
            (Max_Tasks + Max_Resources + Max_Buffers + Max_Messages);
      Task_Name     : ARRAY (Object_Range) OF Unbounded_String;
      Resource_Name : ARRAY (Object_Range) OF Unbounded_String;
      Message_Name  : ARRAY (Object_Range) OF Unbounded_String;
      Buffer_Name   : ARRAY (Object_Range) OF Unbounded_String;

      Task_Id    : Gint;
      Message_Id : Gint;
      Buffer_Id  : Gint;

      Time_U       : Gint;
      Save_Start_U : ARRAY (Object_Range) OF Gint := (OTHERS => 0);
      Save_Start_M : ARRAY (Object_Range) OF Gint := (OTHERS => 0);
      Ok           : Boolean;

   BEGIN


      ---------------------------------
      -- Clear work space
      ---------------------------------


      Draw_Rectangle (Get_Pixmap(Draw_Scheduling.Buffer),
         Draw_Scheduling.
         White_Gc, Filled => True,
         X     => 0,   Y      => 0,
         Width => Draw_Scheduling.Current_Width, Height =>
         Draw_Scheduling.
         Current_Height);


      ---------------------------------
      -- Draw a line per task
      ---------------------------------


      I:=0;
      Reset_Iterator(Draw_Scheduling.Tasks, My_Task_Iterator);

      LOOP

         Current_Element(Draw_Scheduling.Tasks, A_Task, My_Task_Iterator);

         IF To_Display(A_Task.Name, Task_Object_Type)
               THEN

            Msg:=Lb_Task_Name(Current_Language) & "=" & A_Task.Name &
               "        ";

            CASE A_Task.Task_Type IS
               WHEN Aperiodic_Type =>
                  Msg:=Msg & Lb_Capacity(Current_Language) &
                     To_Unbounded_String("=")
                     & A_Task.Capacity'Img
                     & To_Unbounded_String("; ") & Lb_Deadline(
                     Current_Language) & To_Unbounded_String("=")
                     & Aperiodic_Task_Ptr(A_Task).Deadline'Img
                     & To_Unbounded_String("; ") & Lb_Start_Time(
                     Current_Language) & To_Unbounded_String("=")
                     & A_Task.Start_Time'Img
                     & To_Unbounded_String("; ") & Lb_Priority(
                     Current_Language) & To_Unbounded_String("=")
                     & A_Task.Priority'Img
                     & To_Unbounded_String("; Cpu=")
                     & A_Task.Cpu_Name;
               WHEN OTHERS =>
                  Msg:=Msg & Lb_Period(Current_Language) &
                     To_Unbounded_String("=")
                     & Periodic_Task_Ptr(A_Task).Period'Img
                     & To_Unbounded_String("; ") & Lb_Capacity(
                     Current_Language) & To_Unbounded_String("=")
                     & A_Task.Capacity'Img
                     & To_Unbounded_String("; ") & Lb_Deadline(
                     Current_Language) & To_Unbounded_String("=")
                     & Periodic_Task_Ptr(A_Task).Deadline'Img
                     & To_Unbounded_String("; ") & Lb_Start_Time(
                     Current_Language) & To_Unbounded_String("=")
                     & A_Task.Start_Time'Img
                     & To_Unbounded_String("; ") & Lb_Priority(
                     Current_Language) & To_Unbounded_String("=")
                     & A_Task.Priority'Img
                     & To_Unbounded_String("; Cpu=")
                     & A_Task.Cpu_Name;
            END CASE;
            IF A_Task.Task_Type=Parametric_Type
                  THEN
               Msg:=Msg & To_Unbounded_String(";") & Lb_Activation_Rule(
                  Current_Language) & To_Unbounded_String("=") &
                  Parametric_Task_Ptr(A_Task).Activation_Rule;
            END IF;


            Draw_Text (Get_Pixmap(Draw_Scheduling.Buffer),
               Draw_Scheduling.
               Font,
               Draw_Scheduling.Black_Gc,
               Draw_Scheduling.Left_Margin,
               Draw_Scheduling.High_Margin+I*Draw_Scheduling.
               Interligne+20,
               To_String(Msg));


            Draw_Line (Get_Pixmap(Draw_Scheduling.Buffer),
               Draw_Scheduling.Black_Gc,
               Draw_Scheduling.Left_Margin,  Draw_Scheduling.
               High_Margin+I*
               Draw_Scheduling.Interligne,
               Draw_Scheduling.Left_Margin+Gint(Draw_Scheduling.Period-
                  Draw_Scheduling.Start_Draw)*
               Draw_Scheduling.Timeunit_Margin, Draw_Scheduling.
               High_Margin+
               I*Draw_Scheduling.Interligne);

            -- For each line, draw unit time lines
            --
            FOR J IN 0 .. Gint(Draw_Scheduling.Period-Draw_Scheduling.
                  Start_Draw)   LOOP
               Draw_Line (Get_Pixmap(Draw_Scheduling.Buffer),
                  Draw_Scheduling.Black_Gc,
                  Draw_Scheduling.Left_Margin+J*Draw_Scheduling.
                  Timeunit_Margin,  Draw_Scheduling.High_Margin+I*
                  Draw_Scheduling.Interligne-3,
                  Draw_Scheduling.Left_Margin+J*Draw_Scheduling.
                  Timeunit_Margin, Draw_Scheduling.High_Margin+I*
                  Draw_Scheduling.Interligne+3);
            END LOOP;


            -- Register each task in the name table
            --
            Task_Name(Object_Range(I)):=A_Task.Name;

            I:=I+1;
         END IF;

         EXIT WHEN Is_Last_Element(Draw_Scheduling.Tasks,
            My_Task_Iterator);
         Next_Element(Draw_Scheduling.Tasks, My_Task_Iterator);

      END LOOP;


      -------------------------------------
      -- Draw a line for each resource
      -------------------------------------

      IF Schedule_With_Resources
            THEN

         IF NOT Is_Empty(Draw_Scheduling.Resources) THEN

            Reset_Iterator(Draw_Scheduling.Resources, My_Resource_Iterator);

            LOOP

               Current_Element(Draw_Scheduling.Resources,A_Resource,
                  My_Resource_Iterator);

               IF To_Display(A_Resource.Name, Resource_Object_Type)
                     THEN

                  Msg:=Lb_Resource_Name(Current_Language) & "=" &
                     A_Resource.Name & "        ";

                  IF (A_Resource.Protocol = Priority_Ceiling_Protocol)
                        THEN
                     Msg:=Msg & To_Unbounded_String("Protocol = PCP");
                  ELSE
                     IF (A_Resource.Protocol =
                           Priority_Inheritance_Protocol) THEN
                        Msg:=Msg & To_Unbounded_String("Protocol = PIP");
                     ELSE
                        IF (A_Resource.Protocol =
                              Immediate_Priority_Ceiling_Protocol) THEN
                           Msg:=Msg & To_Unbounded_String(
                              "Protocol = IPCP");

                        ELSE
                           Msg:=Msg & To_Unbounded_String(
                              "Protocol = No protocol");
                        END IF;
                     END IF;
                  END IF;

                  Msg:= Msg & To_Unbounded_String("; Cpu=")
                     & A_Task.Cpu_Name;

                  Draw_Text (Get_Pixmap(Draw_Scheduling.Buffer),
                     Draw_Scheduling.
                     Font,
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,
                     Draw_Scheduling.High_Margin+I*Draw_Scheduling.
                     Interligne+
                     20,
                     To_String(Msg));

                  Draw_Line (Get_Pixmap(Draw_Scheduling.Buffer),
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,  Draw_Scheduling.
                     High_Margin+
                     I*
                     Draw_Scheduling.Interligne,
                     Draw_Scheduling.Left_Margin+Gint(Draw_Scheduling.
                        Period-
                        Draw_Scheduling.Start_Draw)*
                     Draw_Scheduling.Timeunit_Margin, Draw_Scheduling.
                     High_Margin+
                     I*Draw_Scheduling.Interligne);

                  -- For each line, draw unit time lines
                  --
                  FOR J IN 0 .. Gint(Draw_Scheduling.Period-
                        Draw_Scheduling.
                        Start_Draw)   LOOP
                     Draw_Line (Get_Pixmap(Draw_Scheduling.Buffer),
                        Draw_Scheduling.Black_Gc,
                        Draw_Scheduling.Left_Margin+J*Draw_Scheduling.
                        Timeunit_Margin,  Draw_Scheduling.High_Margin+I*
                        Draw_Scheduling.Interligne-3,
                        Draw_Scheduling.Left_Margin+J*Draw_Scheduling.
                        Timeunit_Margin, Draw_Scheduling.High_Margin+I*
                        Draw_Scheduling.Interligne+3);
                  END LOOP;


                  -- Register each resource in the name table
                  --
                  Resource_Name(Object_Range(I)):=A_Resource.Name;

                  I := I + 1;
               END IF;


               EXIT WHEN Is_Last_Element(Draw_Scheduling.Resources,
                  My_Resource_Iterator);
               Next_Element(Draw_Scheduling.Resources,
                  My_Resource_Iterator);

            END LOOP;
         END IF;
      END IF;


      -------------------------------------
      -- Draw a line for each message and buffer
      -------------------------------------

      IF Schedule_With_Precedencies
            THEN


         IF NOT Is_Empty(Draw_Scheduling.Messages) THEN

            Reset_Iterator(Draw_Scheduling.Messages, My_Message_Iterator);

            LOOP

               Current_Element(Draw_Scheduling.Messages,A_Message,
                  My_Message_Iterator);

               IF To_Display(A_Message.Name, Message_Object_Type)
                     THEN

                  Msg:=Lb_Message_Name(Current_Language) & "=" &
                     A_Message.Name & "        ";

                  IF A_Message.Message_Type = Periodic_Type
                        THEN
                     Msg:=Msg & To_Unbounded_String("Pe=") &
                        Periodic_Message_Ptr(A_Message).Period'Img;
                  ELSE
                     Msg:=Msg & To_Unbounded_String(" ;Pe= 0");
                  END IF;

                  Msg:=Msg & To_Unbounded_String(" ; Delay=") &
                     A_Message.Response_Time'Img ;
                  Msg:=Msg & To_Unbounded_String(" ; Size = ") &
                     A_Message.Size'Img ;


                  Draw_Text (Get_Pixmap(Draw_Scheduling.Buffer),
                     Draw_Scheduling.
                     Font,
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,
                     Draw_Scheduling.High_Margin+I*Draw_Scheduling.
                     Interligne+
                     20,
                     To_String(Msg));

                  Draw_Line (Get_Pixmap(Draw_Scheduling.Buffer),
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,  Draw_Scheduling.
                     High_Margin+
                     I*
                     Draw_Scheduling.Interligne,
                     Draw_Scheduling.Left_Margin+Gint(Draw_Scheduling.
                        Period-
                        Draw_Scheduling.Start_Draw)*
                     Draw_Scheduling.Timeunit_Margin, Draw_Scheduling.
                     High_Margin+
                     I*Draw_Scheduling.Interligne);

                  -- For each line, draw unit time lines
                  --
                  FOR J IN 0 .. Gint(Draw_Scheduling.Period-
                        Draw_Scheduling.
                        Start_Draw)   LOOP
                     Draw_Line (Get_Pixmap(Draw_Scheduling.Buffer),
                        Draw_Scheduling.Black_Gc,
                        Draw_Scheduling.Left_Margin+J*Draw_Scheduling.
                        Timeunit_Margin,  Draw_Scheduling.High_Margin+I*
                        Draw_Scheduling.Interligne-3,
                        Draw_Scheduling.Left_Margin+J*Draw_Scheduling.
                        Timeunit_Margin, Draw_Scheduling.High_Margin+I*
                        Draw_Scheduling.Interligne+3);
                  END LOOP;


                  -- Register each message in the name table
                  --
                  Message_Name(Object_Range(I)):=A_Message.Name;

                  I := I +1;
               END IF;


               EXIT WHEN Is_Last_Element(Draw_Scheduling.Messages,
                  My_Message_Iterator);
               Next_Element(Draw_Scheduling.Messages,
                  My_Message_Iterator);

            END LOOP;
         END IF;


         IF NOT Is_Empty(Draw_Scheduling.Buffers) THEN

            Reset_Iterator(Draw_Scheduling.Buffers, My_Buffer_Iterator);

            LOOP

               Current_Element(Draw_Scheduling.Buffers,A_Buffer,
                  My_Buffer_Iterator);

               IF To_Display(A_Buffer.Name, Buffer_Object_Type)
                     THEN

                  Msg:=Lb_Buffer_Name(Current_Language) & "=" &
                     A_Buffer.Name & "        ";

                  Msg:=Msg & To_Unbounded_String("Size = ") &
                     A_Buffer.Size'Img ;

                  Msg:=Msg & To_Unbounded_String(" ; Cpu = ") &
                     A_Buffer.Cpu_Name;


                  Draw_Text (Get_Pixmap(Draw_Scheduling.Buffer),
                     Draw_Scheduling.
                     Font,
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,
                     Draw_Scheduling.High_Margin+I*Draw_Scheduling.
                     Interligne+
                     20,
                     To_String(Msg));


                  Draw_Line (Get_Pixmap(Draw_Scheduling.Buffer),
                     Draw_Scheduling.Black_Gc,
                     Draw_Scheduling.Left_Margin,  Draw_Scheduling.
                     High_Margin+
                     I*
                     Draw_Scheduling.Interligne,
                     Draw_Scheduling.Left_Margin+Gint(Draw_Scheduling.
                        Period-
                        Draw_Scheduling.Start_Draw)*
                     Draw_Scheduling.Timeunit_Margin, Draw_Scheduling.
                     High_Margin+
                     I*Draw_Scheduling.Interligne);

                  -- For each line, draw unit time lines
                  --
                  FOR J IN 0 .. Gint(Draw_Scheduling.Period-
                        Draw_Scheduling.
                        Start_Draw)   LOOP
                     Draw_Line (Get_Pixmap(Draw_Scheduling.Buffer),
                        Draw_Scheduling.Black_Gc,
                        Draw_Scheduling.Left_Margin+J*Draw_Scheduling.
                        Timeunit_Margin,  Draw_Scheduling.High_Margin+I*
                        Draw_Scheduling.Interligne-3,
                        Draw_Scheduling.Left_Margin+J*Draw_Scheduling.
                        Timeunit_Margin, Draw_Scheduling.High_Margin+I*
                        Draw_Scheduling.Interligne+3);
                  END LOOP;


                  -- Register each buffer in the name table
                  --
                  Buffer_Name(Object_Range(I)):=A_Buffer.Name;

                  I := I +1;
               END IF;


               EXIT WHEN Is_Last_Element(Draw_Scheduling.Buffers,
                  My_Buffer_Iterator);
               Next_Element(Draw_Scheduling.Buffers, My_Buffer_Iterator);

            END LOOP;
         END IF;

      END IF;





      ------------------------------------------
      -- Draw information from scheduling table
      ------------------------------------------

      FOR U IN 0..Draw_Scheduling.Sched.Nb_Entries-1 LOOP -- processors
         IF  Draw_Scheduling.Sched.Entries(U).Data.Error_Msg=
               Empty_String THEN


            -- Scan scheduling sequence, but only for period-start-draw units of times
            --
            FOR K IN 0..Draw_Scheduling.Sched.Entries(U).
                  Data.Result.Nb_Entries-1 LOOP

               IF (Draw_Scheduling.Sched.Entries(U).Data.Result.
                     Entries(K).Data >=
                     Draw_Scheduling.Start_Draw)
                     AND (Draw_Scheduling.Sched.Entries(U).Data.Result.
                     Entries(K).Data <= Draw_Scheduling.Period)
                     THEN


                  CASE Draw_Scheduling.Sched.Entries(U).Data.Result.
                           Entries(K).Item.Event_Type IS


                     ----------------------------------------
                     -- Draw task running time in black
                     -- (time when the task owns the processor)
                     ----------------------------------------
                     WHEN Running_Task =>

                        Name := Draw_Scheduling.Sched.Entries(U).
                           Data.
                           Result.
                           Entries(K).Item.Running_Task.
                           Name;

                        IF To_Display(Name, Task_Object_Type)
                              THEN
                           Ok:=False;
                           FOR L IN Object_Range LOOP
                              IF (Task_Name(L) = Name) THEN
                                 Ok:=True;
                                 Task_Id:=Gint(L);
                                 Time_U:=
                                    Gint(Draw_Scheduling.Sched.Entries(
                                       U).
                                    Data.Result.Entries(K).
                                    Data-
                                    Draw_Scheduling.Start_Draw);
                              END IF;
                           END LOOP;

                           IF NOT Ok THEN
                              RAISE Task_Not_Found;
                           END IF;

                           Gdk.Gc.Set_Foreground
                              (Draw_Scheduling.Black_Gc, Gdk.Color.
                              Black
                              (Gtk.Widget.Get_Default_Colormap));

                           Draw_Rectangle (Get_Pixmap(Draw_Scheduling.
                                 Buffer),
                              Draw_Scheduling.Black_Gc, True,
                              Draw_Scheduling.Timeunit_Margin*
                              Time_U+Draw_Scheduling.Left_Margin,
                              Task_Id*Draw_Scheduling.Interligne+
                              Draw_Scheduling.High_Margin-3,
                              Draw_Scheduling.Timeunit_Margin, 7);
                        END IF;

                     -----------------------------------------------------
                     -- Draw each task wake up time in red
                     -----------------------------------------------------
                     WHEN Task_Activation =>

                        Name := Draw_Scheduling.Sched.Entries(U).
                           Data.
                           Result.
                           Entries(K).Item.
                           Activation_Task.Name;

                        IF To_Display(Name, Task_Object_Type)
                              THEN
                           Ok:=False;
                           FOR L IN Object_Range LOOP
                              IF (Task_Name(L) = Name) THEN
                                 Ok:=True;
                                 Task_Id:=Gint(L);
                                 Time_U:=
                                    Gint(Draw_Scheduling.Sched.Entries(
                                       U).
                                    Data.Result.Entries(K).
                                    Data-
                                    Draw_Scheduling.Start_Draw);
                              END IF;
                           END LOOP;

                           IF NOT Ok THEN
                              RAISE Task_Not_Found;
                           END IF;

                           Gdk.Gc.Set_Foreground (Draw_Scheduling.
                              Black_Gc,
                              Draw_Scheduling.Red_Color );

                           Draw_Rectangle (Get_Pixmap(Draw_Scheduling.
                                 Buffer),
                              Draw_Scheduling.Black_Gc, True,
                              Draw_Scheduling.Left_Margin+
                              Draw_Scheduling.Timeunit_Margin*Time_U,
                              Draw_Scheduling.High_Margin+Task_Id*
                              Draw_Scheduling.Interligne-8, 3, 16 );

                        END IF;

                     -----------------------------------------------------
                     -- Draw shared resource allocation in blue color
                     -----------------------------------------------------
                     WHEN Allocate_Resource =>

                        Name := Draw_Scheduling.Sched.Entries(U).
                           Data.
                           Result.
                           Entries(K).Item.
                           Allocate_Resource.
                           Name;

                        IF To_Display(Name, Resource_Object_Type)
                              THEN

                           Ok:=False;
                           FOR L IN Object_Range LOOP
                              IF (Resource_Name(L) = Name) THEN
                                 Ok:=True;
                                 Task_Id:=Gint(L);
                                 Time_U:=
                                    Gint(Draw_Scheduling.Sched.Entries(
                                       U).
                                    Data.Result.Entries(K).
                                    Data-
                                    Draw_Scheduling.Start_Draw);
                              END IF;
                           END LOOP;

                           IF NOT Ok THEN
                              RAISE Resource_Not_Found;
                           END IF;

                           Gdk.Gc.Set_Foreground (Draw_Scheduling.
                              Black_Gc,
                              Draw_Scheduling.Blue_Color );

                           Draw_Rectangle (Get_Pixmap(Draw_Scheduling.
                                 Buffer),
                              Draw_Scheduling.Black_Gc, True,
                              Draw_Scheduling.Left_Margin+
                              Draw_Scheduling.Timeunit_Margin*Time_U,
                              Draw_Scheduling.High_Margin+Task_Id*
                              Draw_Scheduling.Interligne-8, 3, 16 );

                           Save_Start_U(Object_Range(Task_Id)) :=
                              Draw_Scheduling.Left_Margin+
                              Draw_Scheduling.Timeunit_Margin*Time_U;

                        END IF;

                     ----------------------------------------
                     -- Draw shared resource releasing in black
                     ----------------------------------------
                     WHEN Release_Resource =>


                        Name := Draw_Scheduling.Sched.Entries(U).
                           Data.
                           Result.
                           Entries(K).Item.
                           Release_Resource.
                           Name;

                        IF To_Display(Name, Resource_Object_Type)
                              THEN
                           Ok:=False;
                           FOR L IN Object_Range LOOP
                              IF (Resource_Name(L) = Name) THEN
                                 Ok:=True;
                                 Task_Id:=Gint(L);
                                 Time_U:=
                                    Gint(Draw_Scheduling.Sched.Entries(
                                       U).
                                    Data.Result.Entries(K).
                                    Data-
                                    Draw_Scheduling.Start_Draw) +
                                    1;
                              END IF;
                           END LOOP;

                           IF NOT Ok THEN
                              RAISE Task_Not_Found;
                           END IF;

                           Gdk.Gc.Set_Foreground
                              (Draw_Scheduling.Black_Gc, Gdk.Color.
                              Black
                              (Gtk.Widget.Get_Default_Colormap));

                           Draw_Rectangle (Get_Pixmap(Draw_Scheduling.
                                 Buffer),
                              Draw_Scheduling.Black_Gc, True,
                              Save_Start_U(Object_Range(Task_Id)),
                              Task_Id*Draw_Scheduling.Interligne+
                              Draw_Scheduling.High_Margin-3,
                              Draw_Scheduling.
                              Left_Margin+
                              Draw_Scheduling.Timeunit_Margin*Time_U-
                              Save_Start_U(Object_Range(Task_Id))
                              , 7);
                        END IF;
                     ----------------------------------------------------
                     -- Event about buffers
                     -- Draw in red buffer write and in blue buffer read
                     ----------------------------------------------------

                     WHEN Write_To_Buffer =>

                        Name := Draw_Scheduling.Sched.Entries(U).
                           Data.
                           Result.
                           Entries(K).Item.Write_Buffer.
                           Name;

                        IF To_Display(Name, Buffer_Object_Type)
                              THEN

                           Ok:=False;
                           FOR L IN Object_Range LOOP
                              IF (Buffer_Name(L) = Name) THEN
                                 Ok:=True;
                                 Buffer_Id:=Gint(L);
                                 Time_U:=
                                    Gint(Draw_Scheduling.Sched.Entries(
                                       U).
                                    Data.Result.Entries(K).
                                    Data);
                              END IF;
                           END LOOP;
                           IF NOT Ok THEN
                              RAISE Buffer_Not_Found;
                           END IF;


                           Gdk.Gc.Set_Foreground (Draw_Scheduling.
                              Black_Gc,
                              Draw_Scheduling.Blue_Color );

                           Draw_Rectangle (Get_Pixmap(Draw_Scheduling.
                                 Buffer),
                              Draw_Scheduling.Black_Gc, True,
                              Draw_Scheduling.Timeunit_Margin*
                              Time_U+Draw_Scheduling.Left_Margin,
                              Buffer_Id*Draw_Scheduling.Interligne+
                              Draw_Scheduling.High_Margin-3,
                              Draw_Scheduling.Timeunit_Margin, 7);


                        END IF;


                     WHEN Read_From_Buffer =>

                        Name := Draw_Scheduling.Sched.Entries(U).
                           Data.
                           Result.
                           Entries(K).Item.
                           Read_Buffer.
                           Name;

                        IF To_Display(Name, Buffer_Object_Type)
                              THEN

                           Ok:=False;
                           FOR L IN Object_Range LOOP
                              IF (Buffer_Name(L) = Name) THEN
                                 Ok:=True;
                                 Buffer_Id:=Gint(L);
                                 Time_U:=
                                    Gint(Draw_Scheduling.Sched.Entries(
                                       U).
                                    Data.Result.Entries(K).
                                    Data-
                                    Draw_Scheduling.Start_Draw);
                              END IF;
                           END LOOP;

                           IF NOT Ok THEN
                              RAISE Buffer_Not_Found;
                           END IF;

                           Gdk.Gc.Set_Foreground (Draw_Scheduling.
                              Black_Gc,
                              Draw_Scheduling.Red_Color );

                           Draw_Rectangle (Get_Pixmap(Draw_Scheduling.
                                 Buffer),
                              Draw_Scheduling.Black_Gc, True,
                              Draw_Scheduling.Timeunit_Margin*
                              Time_U+Draw_Scheduling.Left_Margin,
                              Buffer_Id*Draw_Scheduling.Interligne+
                              Draw_Scheduling.High_Margin-3,
                              Draw_Scheduling.Timeunit_Margin, 7);


                        END IF;

                     ----------------------------------------
                     -- Event about messages
                     -- Draw in red message communication time
                     ----------------------------------------

                     WHEN Send_Message =>

                        Name := Draw_Scheduling.Sched.Entries(U).
                           Data.
                           Result.
                           Entries(K).Item.Send_Message.
                           Name;

                        IF To_Display(Name, Message_Object_Type)
                              THEN

                           Ok:=False;
                           FOR L IN Object_Range LOOP
                              IF (Message_Name(L) = Name) THEN
                                 Ok:=True;
                                 Message_Id:=Gint(L);
                                 Time_U:=
                                    Gint(Draw_Scheduling.Sched.Entries(
                                       U).
                                    Data.Result.Entries(K).
                                    Data);
                              END IF;
                           END LOOP;
                           IF NOT Ok THEN
                              RAISE Message_Not_Found;
                           END IF;

                           -- Save the time the message is sent
                           --
                           Save_Start_M(Object_Range(Message_Id)):=Time_U;
                        END IF;


                     WHEN Receive_Message =>

                        Name := Draw_Scheduling.Sched.Entries(U).
                           Data.
                           Result.
                           Entries(K).Item.
                           Receive_Message.
                           Name;

                        IF To_Display(Name, Message_Object_Type)
                              THEN

                           Ok:=False;
                           FOR L IN Object_Range LOOP
                              IF (Message_Name(L) = Name) THEN
                                 Ok:=True;
                                 Message_Id:=Gint(L);
                                 Time_U:=
                                    Gint(Draw_Scheduling.Sched.Entries(
                                       U).
                                    Data.Result.Entries(K).
                                    Data-
                                    Draw_Scheduling.Start_Draw);
                              END IF;
                           END LOOP;

                           IF NOT Ok THEN
                              RAISE Message_Not_Found;
                           END IF;
                           Gdk.Gc.Set_Foreground (Draw_Scheduling.
                              Black_Gc,
                              Draw_Scheduling.Red_Color );


                           Draw_Rectangle (Get_Pixmap(Draw_Scheduling.
                                 Buffer),
                              Draw_Scheduling.Black_Gc, True,

                              Draw_Scheduling.Left_Margin+
                              Save_Start_M(Object_Range(Message_Id))*
                              Draw_Scheduling.Timeunit_Margin,

                              Message_Id*Draw_Scheduling.Interligne+
                              Draw_Scheduling.High_Margin-3,

                              Draw_Scheduling.
                              Timeunit_Margin*(Time_U-
                                 Save_Start_M(Object_Range(Message_Id))),
                              7);

                        END IF;

                     -------------------------------------------
                     -- Events we have nothing to display
                     -------------------------------------------
                     WHEN OTHERS  =>
                        NULL;



                  END CASE;
               END IF;


            END LOOP;
         END IF;
      END LOOP;

      Gdk.Gc.Set_Foreground
         (Draw_Scheduling.Black_Gc, Gdk.Color.Black (Gtk.Widget.
            Get_Default_Colormap));


      ---------------------
      -- Flush data in the window
      ---------------------
      Draw (Draw_Scheduling.Buffer);

   END Redraw_Time_Line;




END Draw_Scheduling_Pkg;





