------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;


WITH Call_Framework;
USE Call_Framework;
USE Call_Framework.Framework_Request_Package;
USE Call_Framework.Framework_Response_Package;

WITH Parameters;
USE Parameters;
USE Parameters.Framework_Parameters_Table_Package;

WITH Call_Scheduling_Framework;
USE Call_Scheduling_Framework;

WITH Multiprocessor_Services;
USE Multiprocessor_Services;
USE Multiprocessor_Services.Scheduling_Result_Per_Processor;

WITH Time_Units;
USE Time_Units;
WITH Framework_Config;
USE Framework_Config;

WITH Strings;
USE Strings;

WITH Project;
USE Project;

WITH Unbounded_Strings;
USE Unbounded_Strings;

WITH Mmi_Tools;
USE Mmi_Tools;

WITH Translate;
USE Translate;

WITH Text_Tools;
USE Text_Tools;

WITH Text_Io;
USE Text_Io;

WITH Processors;
USE Processors;
WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;

WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;
WITH Resource_Set;
USE Resource_Set;
USE Resource_Set.Generic_Resource_Set;
WITH Message_Set;
USE Message_Set;
USE Message_Set.Generic_Message_Set;
WITH Buffer_Set;
USE Buffer_Set;
USE Buffer_Set.Generic_Buffer_Set;


WITH Select_Time_Line_Pkg;
USE Select_Time_Line_Pkg;

WITH Editor_Config;
USE Editor_Config;
WITH Events;
USE Events;


PACKAGE BODY Schedule_From_To_Pkg.Callbacks IS



   Draw_From,
   Draw_Upto,
   Schedule_To : Unbounded_String;

   Last_Period : CONSTANT Natural := Framework_Config.Max_Scheduling_Period - 1;

   Max_Period : Natural := 0;
   Start_Draw : Integer := 0;
   Stop_Draw  : Integer := 0;

   Ok : Boolean;




   PROCEDURE On_Cancel_Pressed (
         Object : ACCESS Gtk_Button_Record'Class) IS
   BEGIN
      Destroy(Schedule_From_To);
   END On_Cancel_Pressed;




   -- Now, call scheduling analysis for each processor
   --
   PROCEDURE Call_Processor_Analysis (
         Name : Unbounded_String) IS

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;
      A_Param       : Parameter_Ptr;

   BEGIN
      Initialize(Response_List);
      Initialize(Request_List);


      IF Preemption_From_Simulation THEN
         Initialize(A_Request);
         A_Request.Target:=Name;
         A_Request.Statement:=Scheduling_Simulation_Preemption_Number;
         Add(Request_List, A_Request);
      END IF;

      IF Context_Switch_From_Simulation THEN
         Initialize(A_Request);
         A_Request.Target:=Name;
         A_Request.Statement:=
            Scheduling_Simulation_Context_Switch_Number;
         Add(Request_List, A_Request);
      END IF;

      IF  ( Response_Time_From_Simulation(Worst_Case) OR
            Response_Time_From_Simulation(Best_Case) OR
            Response_Time_From_Simulation(Average_Case) OR
            Response_Time_From_Simulation(Plot_Case) ) THEN

         Initialize(A_Request);

         A_Param := NEW Parameter(Boolean_Parameter);
         A_Param.Name:=To_Unbounded_String("worst_case");
         IF Response_Time_From_Simulation(Worst_Case) THEN
            A_Param.Boolean_Value:=True;
         ELSE
            A_Param.Boolean_Value:=False;
         END IF;
         Add(A_Request.Param, A_Param);

         A_Param := NEW Parameter(Boolean_Parameter);
         A_Param.Name:=To_Unbounded_String("best_case");
         IF Response_Time_From_Simulation(Best_Case) THEN
            A_Param.Boolean_Value:=True;
         ELSE
            A_Param.Boolean_Value:=False;
         END IF;
         Add(A_Request.Param, A_Param);

         A_Param := NEW Parameter(Boolean_Parameter);
         A_Param.Name:=To_Unbounded_String("average_case");
         IF Response_Time_From_Simulation(Average_Case) THEN
            A_Param.Boolean_Value:=True;
         ELSE
            A_Param.Boolean_Value:=False;
         END IF;
         Add(A_Request.Param, A_Param);

         A_Request.Target:=Name;


         A_Request.Statement:=Scheduling_Simulation_Response_Time;
         Add(Request_List, A_Request);
      END IF;

      IF  ( Blocking_Time_From_Simulation(Worst_Case) OR
            Blocking_Time_From_Simulation(Best_Case) OR
            Blocking_Time_From_Simulation(Average_Case) OR
            Blocking_Time_From_Simulation(Plot_Case) ) THEN

         Initialize(A_Request);

         A_Param := NEW Parameter(Boolean_Parameter);
         A_Param.Name:=To_Unbounded_String("worst_case");
         IF Blocking_Time_From_Simulation(Worst_Case) THEN
            A_Param.Boolean_Value:=True;
         ELSE
            A_Param.Boolean_Value:=False;
         END IF;
         Add(A_Request.Param, A_Param);

         A_Param := NEW Parameter(Boolean_Parameter);
         A_Param.Name:=To_Unbounded_String("best_case");
         IF Blocking_Time_From_Simulation(Best_Case) THEN
            A_Param.Boolean_Value:=True;
         ELSE
            A_Param.Boolean_Value:=False;
         END IF;
         Add(A_Request.Param, A_Param);

         A_Param := NEW Parameter(Boolean_Parameter);
         A_Param.Name:=To_Unbounded_String("average_case");
         IF Blocking_Time_From_Simulation(Average_Case) THEN
            A_Param.Boolean_Value:=True;
         ELSE
            A_Param.Boolean_Value:=False;
         END IF;
         Add(A_Request.Param, A_Param);

         A_Request.Target:=Name;

         A_Request.Statement:=Scheduling_Simulation_Blocking_Time;
         Add(Request_List, A_Request);
      END IF;


      Sequential_Framework_Request (
         Sys, Request_List, Response_List);


      Write_Title(Schedule_From_To.Text, Response_List.Entries(0).
         Title);
      FOR J IN 0..Response_List.Nb_Entries-1 LOOP
         Write_Text(Schedule_From_To.Text, Response_List.Entries(J).
            Text);
      END LOOP;

   END Call_Processor_Analysis;


   PROCEDURE Call_Uncustomized_Simulation IS

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;
      A_Param       : Parameter_Ptr;

   BEGIN


      Initialize(A_Request);
      A_Request.Statement:=Scheduling_Simulation_Time_Line;
      A_Param := NEW Parameter(Integer_Parameter);
      A_Param.Name:=To_Unbounded_String("period");
      A_Param.Integer_Value:=Max_Period;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Boolean_Parameter);
      A_Param.Name:=To_Unbounded_String("schedule_with_offsets");
      A_Param.Boolean_Value:=Schedule_With_Offsets;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Boolean_Parameter);
      A_Param.Name:=To_Unbounded_String("schedule_with_precedencies");
      A_Param.Boolean_Value:=Schedule_With_Precedencies;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Boolean_Parameter);
      A_Param.Name:=To_Unbounded_String("schedule_with_resources");
      A_Param.Boolean_Value:=Schedule_With_Resources;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Boolean_Parameter);
      A_Param.Name:=To_Unbounded_String("predictable");
      A_Param.Boolean_Value:=Is_Global_Predictable_Seed;
      Add(A_Request.Param, A_Param);

      IF Has_Global_Seed THEN
         A_Param := NEW Parameter(Integer_Parameter);
         A_Param.Name:=To_Unbounded_String("seed_value");
         A_Param.Integer_Value:=Global_Seed_Value;
         Add(A_Request.Param, A_Param);
      END IF;

      FOR I IN Time_Unit_Event_Type'Range LOOP
         A_Param := NEW Parameter(Boolean_Parameter);
         A_Param.Name:=To_Unbounded_String(To_Lower(I'Img));
         IF   Generate_Events(I) THEN
            A_Param.Boolean_Value:=True;
         ELSE
            A_Param.Boolean_Value:=False;
         END IF;
         Add(A_Request.Param, A_Param);
      END LOOP;


      Add(Request_List, A_Request);


      Initialize(A_Request);
      A_Request.Target:=Empty_String;
      A_Request.Statement:=Scheduling_Simulation_Basics;
      Add(Request_List, A_Request);



      Sequential_Framework_Request (
         Sys, Request_List, Response_List);

      Write_Title_And_Text(Schedule_From_To.Text, Response_List);

      Draw_Time_Line(Schedule_From_To.Draw_Scheduling,
         Sys.Tasks,
         Sys.Resources,
         Sys.Messages,
         Sys.Buffers,
         Stop_Draw,
         Start_Draw,
         Call_Scheduling_Framework.Sched);

   END Call_Uncustomized_Simulation;





   PROCEDURE Call_Customized_Simulation IS

      A_Processor           : Processor_Ptr;
      My_Processor_Iterator : Processors_Iterator;

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;
      A_Param       : Parameter_Ptr;

   BEGIN


      Initialize(A_Request);

      A_Request.Statement:=Scheduling_Simulation_Time_Line;
      A_Param := NEW Parameter(Integer_Parameter);
      A_Param.Name:=To_Unbounded_String("period");
      A_Param.Integer_Value:=Max_Period;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Boolean_Parameter);
      A_Param.Name:=To_Unbounded_String("schedule_with_offsets");
      A_Param.Boolean_Value:=Schedule_With_Offsets;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Boolean_Parameter);
      A_Param.Name:=To_Unbounded_String("schedule_with_precedencies");
      A_Param.Boolean_Value:=Schedule_With_Precedencies;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Boolean_Parameter);
      A_Param.Name:=To_Unbounded_String("schedule_with_resources");
      A_Param.Boolean_Value:=Schedule_With_Resources;
      Add(A_Request.Param, A_Param);
      A_Param := NEW Parameter(Boolean_Parameter);
      A_Param.Name:=To_Unbounded_String("predictable");
      A_Param.Boolean_Value:=Is_Global_Predictable_Seed;
      Add(A_Request.Param, A_Param);

      IF Has_Global_Seed THEN
         A_Param := NEW Parameter(Integer_Parameter);
         A_Param.Name:=To_Unbounded_String("seed_value");
         A_Param.Integer_Value:=Global_Seed_Value;
         Add(A_Request.Param, A_Param);
      END IF;


      FOR I IN Time_Unit_Event_Type'Range LOOP
         A_Param := NEW Parameter(Boolean_Parameter);
         A_Param.Name:=To_Unbounded_String(To_Lower(I'Img));
         IF   Generate_Events(I) THEN
            A_Param.Boolean_Value:=True;
         ELSE
            A_Param.Boolean_Value:=False;
         END IF;
         Add(A_Request.Param, A_Param);
      END LOOP;



      Add(Request_List, A_Request);

      Sequential_Framework_Request (
         Sys, Request_List, Response_List);


      -- Put to the screen the resulting event table
      --
      IF Schedule_And_Display_Event_Table THEN
         Put(Call_Scheduling_Framework.Sched);
      END IF;


      Write_Title_And_Text(Schedule_From_To.Text, Response_List);

      Draw_Time_Line(Schedule_From_To.Draw_Scheduling,
         Sys.Tasks,
         Sys.Resources,
         Sys.Messages,
         Sys.Buffers,
         Stop_Draw,
         Start_Draw,
         Call_Scheduling_Framework.Sched);


      -- export event table in the targeted XML file
      --
      IF Schedule_And_Export_Event_Table
            THEN
         Write_To_Xml_File (Call_Scheduling_Framework.Sched, Sys,
            Xml_Event_Table_File_Name);
      END IF;


      IF Schedule_All_Processeurs=False THEN
         Call_Processor_Analysis(Last_Selected_Processor);
      ELSE

         Reset_Iterator(Sys.Processors, My_Processor_Iterator);
         LOOP
            Current_Element(Sys.Processors, A_Processor,
               My_Processor_Iterator);

            Call_Processor_Analysis(A_Processor.Name);

            EXIT WHEN Is_Last_Element(Sys.Processors,
               My_Processor_Iterator);
            Next_Element(Sys.Processors, My_Processor_Iterator);
         END LOOP;

      END IF;


      -- Now run, event analyzers,
      --

      IF Run_Event_Analyzer_From_Simulation THEN

         Initialize(Response_List);
         Initialize(Request_List);

         Initialize(A_Request);
         A_Request.Statement:=Scheduling_Simulation_Run_Event_Handler;
         Add(Request_List, A_Request);

         Sequential_Framework_Request (
            Sys, Request_List, Response_List);

         Write_Title_And_Text(Schedule_From_To.Text, Response_List);

      END IF;
   END Call_Customized_Simulation;



   PROCEDURE Call_Simulation IS
   BEGIN
      -- According to from when the scheduling simulation is demanded
      -- by the user, we build a different request
      --
      IF Schedule_From_To.Customized_Simulation
            THEN
         Call_Customized_Simulation;
      ELSE
         Call_Uncustomized_Simulation;
      END IF;
   END Call_Simulation;




   PROCEDURE On_Ok_Pressed (
         Object : ACCESS Gtk_Button_Record'Class) IS


      -- Number of time line
      -- to display
      --
      Object_Number : Natural := 0;


   BEGIN


      Draw_From:=To_Unbounded_String(Get_Text(Schedule_From_To.
            Draw_From_Entry));
      Draw_Upto:=To_Unbounded_String(Get_Text(Schedule_From_To.
            Draw_Upto_Entry));
      Schedule_To:=To_Unbounded_String(Get_Text(Schedule_From_To.
            Schedule_To_Entry));



      To_Natural(Draw_Upto,Stop_Draw,Ok);
      IF NOT Ok
            THEN
         Show_Message_Box(Lb_Double_Quote & Lb_Draw_To(Current_Language) &
            Lb_Double_Quote &
            Lb_Must_Be_Numeric(Current_Language));
         RETURN;
      END IF;


      To_Integer(Draw_From,Start_Draw,Ok);
      IF NOT Ok
            THEN
         Show_Message_Box(Lb_Double_Quote & Lb_Draw_From(Current_Language) &
            Lb_Double_Quote &
            Lb_Must_Be_Numeric(Current_Language));
         RETURN;
      END IF;


      To_Integer(Schedule_To,Max_Period,Ok);
      IF NOT Ok
            THEN
         Show_Message_Box(Lb_Double_Quote & Lb_Schedule_To(
               Current_Language) & Lb_Double_Quote &
            Lb_Must_Be_Numeric(Current_Language));
         RETURN;
      END IF;


      IF(Max_Period>Last_Period)
            THEN
         Show_Message_Box(Lb_Double_Quote & Lb_Schedule_To(
               Current_Language) & Lb_Double_Quote & Lb_Must_Be(
               Current_Language) & To_Unbounded_String(" <= ")
            & Last_Period'Img);
         RETURN;
      END IF;

      IF(Max_Period<=0)
            THEN
         Show_Message_Box(Lb_Double_Quote & Lb_Schedule_To(
               Current_Language) & Lb_Double_Quote &
            Lb_Must_Be(Current_Language) & To_Unbounded_String(" > 0"));
         RETURN;
      END IF;

      IF(Start_Draw<0)
            THEN
         Show_Message_Box(Lb_Double_Quote & Lb_Draw_From(Current_Language) &
            Lb_Double_Quote & Lb_Must_Be(Current_Language) &
            To_Unbounded_String(" >= 0"));
         RETURN;
      END IF;

      IF(Stop_Draw<0)
            THEN
         Show_Message_Box(Lb_Double_Quote & Lb_Draw_To(Current_Language) &
            Lb_Double_Quote & Lb_Must_Be(Current_Language) &
            To_Unbounded_String(" >= 0"));
         RETURN;
      END IF;



      IF(Start_Draw>Max_Period)
            THEN
         Show_Message_Box(Lb_Double_Quote & Lb_Draw_From(Current_Language) &
            Lb_Must_Be(
               Current_Language) & Lb_Double_Quote & To_Unbounded_String(
               " <= ")
            & Lb_Double_Quote & Lb_Schedule_To(Current_Language) &
            Lb_Double_Quote );
         RETURN;
      END IF;

      IF (Stop_Draw-Start_Draw>1500)
            THEN
         Show_Message_Box(Lb_Draw_Time_Line_Too_Large(Current_Language));
         RETURN;
      END IF;




      -- Check if we need to select which tasks/messages/buffers
      -- to display with time lines
      --
      Object_Number:=Natural(Get_Number_Of_Elements(Sys.Tasks))+
         Natural(Get_Number_Of_Elements(Sys.Messages))+
         Natural(Get_Number_Of_Elements(Sys.Buffers));

      IF Schedule_With_Resources THEN
         Object_Number:=Object_Number+Natural(Get_Number_Of_Elements(
               Sys.
               Resources));
      END IF;


      IF Object_Number > Editor_Config.Max_Time_Line_To_Display
            THEN
         Gtk_New(Select_Time_Line, Sys, Call_Simulation'ACCESS);
         Show_All(Select_Time_Line);
      ELSE
         Call_Simulation;
      END IF;

      -- The window can not be destroyed ... because
      -- the callback used by the Select_Time_Line widget
      -- will call it !!!
      -- So, just hide it, and do not destroy it !!!!
      --
      Hide(Schedule_From_To);


   END  On_Ok_Pressed;

END Schedule_From_To_Pkg.Callbacks;

