------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Framework_Config;
USE Framework_Config;
WITH Translate;
USE Translate;


WITH Tasks;
USE Tasks;
WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;
WITH Processors;
USE Processors;
WITH Processor_Set;
USE Processor_Set;
USE Processor_Set.Generic_Processor_Set;
WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;


WITH Gtk;
USE Gtk;
WITH Gtk.Enums;
USE Gtk.Enums;

WITH Callbacks_Cheddar;
USE Callbacks_Cheddar;
WITH Schedule_From_To_Pkg.Callbacks;
USE Schedule_From_To_Pkg.Callbacks;



PACKAGE BODY Schedule_From_To_Pkg IS



   PROCEDURE Gtk_New (
         Schedule_From_To      :    OUT Schedule_From_To_Access;
         Text                  :        Gtk_Text;
         Draw_Scheduling       :        Draw_Scheduling_Access;
         Sys                   : IN     System;
         Customized_Simulation :        Boolean                 := False) IS

   BEGIN
      Schedule_From_To := NEW Schedule_From_To_Record;
      Schedule_From_To_Pkg.Initialize (Schedule_From_To, Text,
         Draw_Scheduling, Sys, Customized_Simulation);

   END Gtk_New;




   PROCEDURE Initialize (
         Schedule_From_To      : ACCESS Schedule_From_To_Record'Class;
         Text                  :        Gtk_Text;
         Draw_Scheduling       :        Draw_Scheduling_Access;
         Sys                   : IN     System;
         Customized_Simulation :        Boolean                       := False) IS


      A_Task      :          Task_Ptr;
      Iterator1   :          Tasks_Iterator;
      A_Processor :          Processor_Ptr;
      Iterator2   :          Processors_Iterator;
      Max_Period  :          Natural             := 0;
      Period      :          Natural             := 0;
      Last_Period : CONSTANT Natural             := Framework_Config.Max_Scheduling_Period
        - 1;



   BEGIN



      -- First of all : get the largest period of all
      -- the defined tasks
      --
      Reset_Iterator(Sys.Tasks, Iterator1);
      LOOP

         Current_Element(Sys.Tasks, A_Task, Iterator1);
         IF (A_Task.Task_Type = Periodic_Type)
               OR (A_Task.Task_Type = Poisson_Type)
               OR (A_Task.Task_Type = Sporadic_Type)
               THEN
            Max_Period:=Natural'Max(Max_Period,
               Periodic_Task_Ptr(A_Task).Period);
         ELSE
            Max_Period:=Natural'Max(Max_Period,
               A_Task.Deadline);
         END IF;

         EXIT WHEN Is_Last_Element(Sys.Tasks, Iterator1);

         Next_Element(Sys.Tasks, Iterator1);

      END LOOP;




      -- Then, compute the base period (or hyper-period)
      --
      Reset_Iterator(Sys.Processors, Iterator2);

      LOOP


         Current_Element(Sys.Processors, A_Processor, Iterator2);

         IF (Get_Number_Of_Task_From_Processor(Sys.Tasks, A_Processor.
                  Name)/=0)
               THEN

            BEGIN
               Period:=Scheduling_Period(Sys.
                  Tasks, A_Processor.Name);
               IF Period> Max_Period
                     THEN
                  Max_Period:=Period;
               END IF;

            EXCEPTION
               WHEN OTHERS  =>
                  NULL;
            END;
         END IF;

         EXIT WHEN Is_Last_Element(Sys.Processors, Iterator2);

         Next_Element(Sys.Processors, Iterator2);


      END LOOP;


      IF(Max_Period>=Last_Period)
            THEN
         Max_Period:=Last_Period;
      END IF;





      Schedule_From_To.Customized_Simulation:=Customized_Simulation;
      Schedule_From_To.Text:=Text;
      Schedule_From_To.Draw_Scheduling:=Draw_Scheduling;

      Gtk.Window.Initialize (Schedule_From_To, Window_Dialog);
      Set_Title (Schedule_From_To, To_String(Lb_Scheduling_Simulation(
               Current_Language)));
      Set_Policy (Schedule_From_To, False, True, False);
      Set_Position (Schedule_From_To, Win_Pos_Center_Always);
      Set_Modal (Schedule_From_To, True);

      Gtk_New_Vbox (Schedule_From_To.Vbox, False, 0);


      Add (Schedule_From_To, Schedule_From_To.Vbox);



      Gtk_New_Hbox (Schedule_From_To.Hbox, False, 0);
      Set_Border_Width (Schedule_From_To.Hbox, 16);
      Pack_Start (Schedule_From_To.Vbox, Schedule_From_To.Hbox, True,
         True, 0);

      Gtk_New (Schedule_From_To.Schedule_From_To_Label, To_String(
            Lb_Schedule_To(
               Current_Language) & Lb_Colon ));
      Set_Alignment (Schedule_From_To.Schedule_From_To_Label, 0.5,
         0.5);
      Set_Padding (Schedule_From_To.Schedule_From_To_Label, 0, 0);
      Set_Justify (Schedule_From_To.Schedule_From_To_Label,
         Justify_Left);
      Set_Line_Wrap (Schedule_From_To.Schedule_From_To_Label, False);
      Pack_Start (Schedule_From_To.Hbox, Schedule_From_To.
         Schedule_From_To_Label, False, False, 0);

      Gtk_New (Schedule_From_To.Schedule_To_Entry);
      Set_Editable (Schedule_From_To.Schedule_To_Entry, True);
      Set_Max_Length (Schedule_From_To.Schedule_To_Entry, 0);
      Set_Text (Schedule_From_To.Schedule_To_Entry, Max_Period'Img);
      Set_Visibility (Schedule_From_To.Schedule_To_Entry, True);
      Pack_Start (Schedule_From_To.Hbox, Schedule_From_To.
         Schedule_To_Entry,
         True, True, 0);



      Gtk_New_Hbox (Schedule_From_To.Hbox, False, 0);
      Set_Border_Width (Schedule_From_To.Hbox, 16);
      Pack_Start (Schedule_From_To.Vbox, Schedule_From_To.Hbox, True,
         True, 0);

      Gtk_New (Schedule_From_To.Schedule_From_To_Label, To_String(
            Lb_Draw_From(
               Current_Language) & Lb_Colon ));
      Set_Alignment (Schedule_From_To.Schedule_From_To_Label, 0.5,
         0.5);
      Set_Padding (Schedule_From_To.Schedule_From_To_Label, 0, 0);
      Set_Justify (Schedule_From_To.Schedule_From_To_Label,
         Justify_Left);
      Set_Line_Wrap (Schedule_From_To.Schedule_From_To_Label, False);
      Pack_Start (Schedule_From_To.Hbox, Schedule_From_To.
         Schedule_From_To_Label, False, False, 0);

      Gtk_New (Schedule_From_To.Draw_From_Entry);
      Set_Editable (Schedule_From_To.Draw_From_Entry, True);
      Set_Max_Length (Schedule_From_To.Draw_From_Entry, 0);
      Set_Text (Schedule_From_To.Draw_From_Entry, "0");
      Set_Visibility (Schedule_From_To.Draw_From_Entry, True);
      Pack_Start (Schedule_From_To.Hbox, Schedule_From_To.
         Draw_From_Entry,
         True, True, 0);


      Gtk_New_Hbox (Schedule_From_To.Hbox, True, 0);
      Set_Border_Width (Schedule_From_To.Hbox, 16);
      Pack_Start (Schedule_From_To.Vbox, Schedule_From_To.Hbox, True,
         True, 0);

      Gtk_New (Schedule_From_To.Schedule_From_To_Label, To_String(
            Lb_Draw_To(
               Current_Language) & Lb_Colon ));
      Set_Alignment (Schedule_From_To.Schedule_From_To_Label, 0.5,
         0.5);
      Set_Padding (Schedule_From_To.Schedule_From_To_Label, 0, 0);
      Set_Justify (Schedule_From_To.Schedule_From_To_Label,
         Justify_Left);
      Set_Line_Wrap (Schedule_From_To.Schedule_From_To_Label, False);
      Pack_Start (Schedule_From_To.Hbox, Schedule_From_To.
         Schedule_From_To_Label, False, False, 0);


      Gtk_New (Schedule_From_To.Draw_Upto_Entry);
      Set_Editable (Schedule_From_To.Draw_Upto_Entry, True);
      Set_Max_Length (Schedule_From_To.Draw_Upto_Entry, 0);
      Set_Text (Schedule_From_To.Draw_Upto_Entry, Max_Period'Img);
      Set_Visibility (Schedule_From_To.Draw_Upto_Entry, True);
      Pack_Start (Schedule_From_To.Hbox, Schedule_From_To.
         Draw_Upto_Entry,
         True, True, 0);



      Gtk_New_Hseparator (Schedule_From_To.Separator);
      Pack_Start (Schedule_From_To.Vbox, Schedule_From_To.Separator,
         True,
         True, 0);

      Gtk_New_Hbox (Schedule_From_To.Hbox, True, 0);
      Set_Border_Width (Schedule_From_To.Hbox, 9);
      Pack_Start (Schedule_From_To.Vbox, Schedule_From_To.Hbox, True,
         True, 0);

      Gtk_New (Schedule_From_To.Ok_Button, "Ok");
      Set_Usize (Schedule_From_To.Ok_Button, 70, 25);
      Button_Callback.Connect
         (Schedule_From_To.Ok_Button, "clicked",
         Button_Callback.To_Marshaller (On_Ok_Pressed'ACCESS));
      Pack_Start (Schedule_From_To.Hbox, Schedule_From_To.Ok_Button,
         False, False, 0);

      Gtk_New (Schedule_From_To.Cancel_Button, To_String(Lb_Cancel(
               Current_Language)));
      Set_Usize (Schedule_From_To.Cancel_Button, 70, 25);
      Button_Callback.Connect
         (Schedule_From_To.Cancel_Button, "clicked",
         Button_Callback.To_Marshaller (On_Cancel_Pressed'ACCESS));

      Pack_Start (Schedule_From_To.Hbox, Schedule_From_To.
         Cancel_Button,
         False, False, 0);

   END Initialize;





END Schedule_From_To_Pkg;
