------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------




WITH  Gnat.Current_Exception;
USE  Gnat.Current_Exception;

WITH Text_Io;
USE Text_Io;

WITH Translate;
USE Translate;

WITH Framework_Config;
USE Framework_Config;

WITH Unbounded_Strings;
USE Unbounded_Strings;
USE Unbounded_Strings.Strings_Table_Package;



WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;

WITH Gtk.Gentry;
USE Gtk.Gentry;


WITH Resources;
USE Resources;
WITH Resource_Set;
USE Resources.Task_Lists;
USE Resource_Set;
USE Resource_Set.Generic_Resource_Set;


WITH Task_Set;
USE Task_Set;
USE Task_Set.Generic_Task_Set;

WITH Project;
USE Project;



PACKAGE BODY Update_Resources_Set_Pkg.Callbacks.Integrity_Checks IS



   PROCEDURE Check_Integrity_Constraints (
         Update_Resources_Set : IN OUT Update_Resources_Set_Access;
         Error_Message        : IN OUT Unbounded_String;
         Ok                   : IN OUT Boolean) IS


      Resource_State_String    : Unbounded_String;
      Resource_Protocol_String : Unbounded_String;

   BEGIN

      Ok:=True;

      Resource_Name := To_Unbounded_String(Get_Text(
            Update_Resources_Set.
            Name));


      Resource_State_String := To_Unbounded_String(Get_Text(
            Update_Resources_Set.State));

      IF Resource_State_String=""
            THEN
         Resource_State:=0;
      ELSE
         To_Integer(Resource_State_String, Resource_State, Ok);
         IF NOT Ok
               THEN
            Error_Message:=Lb_Resource_State(Current_Language) &
               Lb_Must_Be_Numeric(Current_Language);
            RETURN;
         END IF;
      END IF;


      Resource_Protocol_String := To_Unbounded_String(Gtk.Gentry.Get_Text(
            Get_Entry(Update_Resources_Set.Protocol)));


      IF Resource_Protocol_String="PCP" THEN
         Resource_Protocol:=Priority_Ceiling_Protocol;
      ELSE
         IF Resource_Protocol_String="PIP" THEN
            Resource_Protocol:=Priority_Inheritance_Protocol;
         ELSE
            IF Resource_Protocol_String="IPCP" THEN
               Resource_Protocol:=Immediate_Priority_Ceiling_Protocol;
            ELSE
               Resource_Protocol:=No_Protocol;
            END IF;
         END IF;
      END IF;


      Resource_Processor :=To_Unbounded_String(Get_Text(
            Update_Resources_Set.
            Cpu_Name));
      Resource_Address_Space:=To_Unbounded_String(Get_Text(Get_Entry(
               Update_Resources_Set.Address_Space_Menu)));


      Check_Resource (Sys.Resources,
         Integrity_Checks.Resource_Name,
         Integrity_Checks.Resource_State,
         Integrity_Checks.Resource_Processor,
         Integrity_Checks.Resource_Address_Space,
         Integrity_Checks.Resource_Protocol);


      FOR I IN 0 .. Update_Resources_Set.Task_List_Share.Nb_Entries-1
            LOOP

         Check_Task_Critical_Section (Sys.Tasks,
            Integrity_Checks.Resource_Name,
            Integrity_Checks.Resource_Processor,
            Update_Resources_Set.Task_List_Share.Entries(I).Item,
            Update_Resources_Set.Task_List_Share.Entries(I).
            Data.Task_Begin,
            Update_Resources_Set.Task_List_Share.Entries(I).
            Data.Task_End);

      END LOOP;




   EXCEPTION
      WHEN OTHERS =>
         Ok:=False;
         Error_Message:=To_Unbounded_String(Exception_Message);


   END Check_Integrity_Constraints;





   PROCEDURE Check_Task_Sharing_Integrity_Constraints (
         Update_Resources_Set : IN OUT Update_Resources_Set_Access;
         Error_Message        : IN OUT Unbounded_String;
         Ok                   : IN OUT Boolean) IS

      Share_Begin_String : Unbounded_String;
      Share_End_String   : Unbounded_String;
      Integer_Ok         : Boolean;

   BEGIN

      Ok:=True;


      Resource_Task_Name := To_Unbounded_String(Get_Text(Get_Entry(
               Update_Resources_Set.Resource_Affect_Task)));
      Share_Begin_String := To_Unbounded_String(Get_Text(
            Update_Resources_Set.Resource_Affect_Begin));
      Share_End_String := To_Unbounded_String(Get_Text(
            Update_Resources_Set.
            Resource_Affect_End));


      To_Integer(Share_Begin_String, Resource_Share_Begin, Integer_Ok);
      IF NOT Integer_Ok
            THEN
         Error_Message:=Lb_Task_Begin(Current_Language) &
            Lb_Must_Be_Numeric(Current_Language);
         Ok:=False;
         RETURN;
      END IF;


      To_Integer(Share_End_String, Resource_Share_End, Integer_Ok);
      IF NOT Integer_Ok
            THEN
         Error_Message:=Lb_Task_End(Current_Language) &
            Lb_Must_Be_Numeric(Current_Language);
         Ok:=False;
         RETURN;
      END IF;


   EXCEPTION
      WHEN OTHERS =>
         Ok:=False;
         Error_Message:=To_Unbounded_String(Exception_Message);


   END Check_Task_Sharing_Integrity_Constraints;



END Update_Resources_Set_Pkg.Callbacks.Integrity_Checks;

