------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Text_Io;
USE Text_Io;

WITH Ada.Sequential_Io;



PACKAGE BODY Io_Tools IS


   -- Return the directory separator for Ubnix or windows
   -- depending on the file_name and the already directory
   -- separators it contains
   --
   FUNCTION Directory_Separator (
         File_Name : Unbounded_String)
     RETURN String IS

   BEGIN

      -- Suppress blank
      --
      FOR I IN 1..Length(File_Name) LOOP
         IF Element(File_Name,I) = '/'
               THEN
            RETURN "/";
         END IF;
      END LOOP;
      RETURN "\";
   END Directory_Separator;



   -- Check is a file can be read
   -- Return True if the file can be read
   --
   FUNCTION Can_Be_Read (
         File_Name : Unbounded_String)
     RETURN Boolean IS

      A_File : File_Type;
      Result : Boolean   := True;

   BEGIN

      BEGIN
         Open(A_File,  In_File, To_String(File_Name));
         Close(A_File);

      EXCEPTION
         WHEN OTHERS =>
            Result:=False;
      END;

      RETURN Result;

   END Can_Be_Read;



   PROCEDURE Reraise;
   PRAGMA No_Return (Reraise);
   PRAGMA Import (C, Reraise, "__gnat_reraise");
   --  Reraises the exception referenced by the Current_Excep field of
   --  the TSD (all fields of this exception occurrence are set). Abort
   --  is deferred before the reraise operation.


   PROCEDURE Write_Sequential_File (
         File_Name : IN     Unbounded_String;
         Data      : IN     Unbounded_String) IS


      PACKAGE Set_File IS NEW Ada.Sequential_Io(Character);
      USE Set_File;
      A_File : Set_File.File_Type;

   BEGIN
      Create(A_File,
         Mode => Out_File,
         Name => To_String(File_Name));


      IF NOT Is_Open(A_File)
            THEN
         RAISE Can_Not_Open_File_For_Writing;
      END IF;


      FOR I IN 1..Length(Data) LOOP
         Write(A_File,Element(Data,I));
      END LOOP;

      Close(A_File);

   EXCEPTION
      WHEN OTHERS =>
         Close(A_File);
         Reraise;

   END Write_Sequential_File;


   PROCEDURE Write_Sequential_File (
         File_Name : IN     String;
         Data      : IN     Unbounded_String) IS
   BEGIN
      Write_Sequential_File(To_Unbounded_String(File_Name), Data);
   END Write_Sequential_File;




   FUNCTION Read_Sequential_File (
         File_Name : String)
     RETURN Unbounded_String IS
   BEGIN
      RETURN Read_Sequential_File(To_Unbounded_String(File_Name));
   END Read_Sequential_File;


   FUNCTION Read_Sequential_File (
         File_Name : Unbounded_String)
     RETURN Unbounded_String IS

      Desc   : Character;
      Result : Unbounded_String;

      PACKAGE Set_File IS NEW Ada.Sequential_Io(Character);
      USE Set_File;
      A_File : Set_File.File_Type;

   BEGIN
      Open(A_File,
         Mode => In_File,
         Name => To_String(File_Name));


      IF NOT Is_Open(A_File)
            THEN
         RAISE Can_Not_Open_File_For_Reading;
      END IF;


      WHILE NOT End_Of_File(A_File)
            LOOP
         Read(A_File,Desc);
         Result := Result & Desc;
      END LOOP;

      Close(A_File);

      RETURN Result;

   EXCEPTION
      WHEN OTHERS =>
         Close(A_File);
         Reraise;


   END Read_Sequential_File;




END Io_Tools;


