------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------



WITH Text_Io;
USE Text_Io;





WITH Unchecked_Deallocation;




PACKAGE BODY Lists IS


   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Cell, Cell_Ptr);





   PROCEDURE Initialize (
         L : IN OUT List) IS

      Current  : Cell_Ptr := L.Head;
      Previous : Cell_Ptr;

   BEGIN
      WHILE Current/=NULL LOOP
         Free(Current.Info);
         Previous:=Current;
         Current:=Current.Next;
         Free(Previous);
      END LOOP;

      L.Head:=NULL;
      L.Tail:=NULL;
   END Initialize;


   PROCEDURE Add (
         L : IN OUT List;
         E : IN     Element) IS

      New_Cell : Cell_Ptr;

   BEGIN
      New_Cell:=NEW Cell;
      New_Cell.Info:=E;

      New_Cell.Next:=L.Head;
      L.Head:=New_Cell;

      IF (L.Tail = NULL)
            THEN
         L.Tail := New_Cell;
      ELSE
         New_Cell.Next.Previous:=New_Cell;
      END IF;


   END Add;



   PROCEDURE Put (
         L : IN     List) IS

      Current : Cell_Ptr := L.Head;

   BEGIN
      WHILE Current/=NULL LOOP
         Put(Current.Info);
         Current:=Current.Next;
      END LOOP;
   END Put;




   FUNCTION Get_Number_Of_Elements (
         L : IN     List)
     RETURN Natural IS

      Current : Cell_Ptr := L.Head;
      Size    : Natural  := 0;

   BEGIN
      WHILE Current/=NULL LOOP
         Size:=Size+1;
         Current:=Current.Next;
      END LOOP;

      RETURN Size;

   END Get_Number_Of_Elements;



   PROCEDURE Delete (
         L : IN OUT List;
         E : IN     List) IS

      Current : Cell_Ptr := E.Head;

   BEGIN
      WHILE Current/=NULL LOOP
         Delete(L, Current.Info);
         Current:=Current.Next;
      END LOOP;
   END Delete;


   PROCEDURE Delete (
         L : IN OUT List;
         E : IN     Element) IS

      Current  : Cell_Ptr := L.Head;
      Is_Found : Boolean  := False;

   BEGIN
      WHILE Current/=NULL LOOP
         IF Current.Info=E
               THEN
            Is_Found:=True;

            -- Delete the last item
            --
            IF (Current.Previous=NULL)
                  AND (Current.Next=NULL)
                  THEN
               L.Head:=NULL;
               L.Tail:=NULL;
            ELSE

               -- Delete the Head item
               --
               IF (Current.Previous=NULL)
                     THEN
                  L.Head:=Current.Next;
                  Current.Next.Previous:=NULL;
               ELSE
                  -- delete the Tail item
                  --
                  IF (Current.Next = NULL) THEN
                     L.Tail := Current.Previous;
                     Current.Previous.Next:=NULL;
                     -- Middle item
                     --
                  ELSE
                     Current.Previous.Next:=Current.Next;
                     Current.Next.Previous:=Current.Previous;
                  END IF;
               END IF;
            END IF;

            Free(Current.Info);
            Free(Current);
            EXIT;
         END IF;

         Current:=Current.Next;
      END LOOP;

      IF NOT Is_Found THEN
         RAISE Element_Not_Found;
      END IF;

   END Delete;


   PROCEDURE Delete (
         L : IN OUT List) IS

      Current  : Cell_Ptr := L.Head;
      Previous : Cell_Ptr;

   BEGIN
      WHILE Current/=NULL LOOP
         Free(Current.Info);
         Previous:=Current;
         Current:=Current.Next;
         Free(Previous);
      END LOOP;

   END Delete;


   PROCEDURE Duplicate (
         Src  : IN     List;
         Dest : IN OUT List) IS

      Current : Cell_Ptr := Src.Tail;

   BEGIN

      WHILE Current/=NULL LOOP


         Add(Dest,Current.Info);

         Current:=Current.Previous;
      END LOOP;

   END Duplicate;


   FUNCTION Get_Head (
         L : IN     List)
     RETURN Element IS

   BEGIN
      RETURN(L.Head.Info);
   END Get_Head;


   FUNCTION Get_Tail (
         L : IN     List)
     RETURN Element IS

   BEGIN
      RETURN(L.Tail.Info);
   END Get_Tail;


   FUNCTION Is_Empty (
         L : IN     List)
     RETURN Boolean IS
   BEGIN
      RETURN (L.Head=NULL) OR (L.Tail=NULL);
   END Is_Empty;


   PROCEDURE Previous_Element (
         L           : IN     List;
         My_Iterator : IN OUT Iterator) IS
   BEGIN
      My_Iterator.Current:=My_Iterator.Current.Previous;
   END Previous_Element;


   PROCEDURE Next_Element (
         L           : IN     List;
         My_Iterator : IN OUT Iterator) IS
   BEGIN
      My_Iterator.Current:=My_Iterator.Current.Next;
   END Next_Element;


   FUNCTION Is_Tail_Element (
         L           : IN     List;
         My_Iterator : IN     Iterator)
     RETURN Boolean IS
   BEGIN
      RETURN (My_Iterator.Current = L.Tail) OR (L.Tail=NULL);
   END Is_Tail_Element;


   FUNCTION Is_Head_Element (
         L           : IN     List;
         My_Iterator : IN     Iterator)
     RETURN Boolean IS
   BEGIN
      RETURN (My_Iterator.Current = L.Head) OR (L.Head=NULL);
   END Is_Head_Element;


   PROCEDURE Current_Element (
         L              : IN     List;
         Return_Element :    OUT Element;
         My_Iterator    : IN     Iterator) IS
   BEGIN
      Return_Element:= My_Iterator.Current.Info;
   END Current_Element;


   PROCEDURE Reset_Head_Iterator (
         L           : IN     List;
         My_Iterator : IN OUT Iterator) IS
   BEGIN
      My_Iterator.Current:=L.Head;
   END Reset_Head_Iterator;


   PROCEDURE Reset_Tail_Iterator (
         L           : IN     List;
         My_Iterator : IN OUT Iterator) IS
   BEGIN
      My_Iterator.Current:=L.Tail;
   END Reset_Tail_Iterator;




END Lists;
