------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Text_Io;
USE Text_Io;



WITH Ada.Sequential_Io;


PACKAGE BODY Sets IS




   PROCEDURE Reset (
         My_Set      : IN OUT Set;
         Free_Object : IN     Boolean := False) IS
   BEGIN
      IF(Free_Object = True)
            THEN
         FOR I IN 0 .. My_Set.Number_Of_Elements-1  LOOP
            Free(My_Set.Container(I));
         END LOOP;
      END IF;
      My_Set.Number_Of_Elements:=0;
   END Reset;


   PROCEDURE Initialize (
         My_Set : IN OUT Set) IS
   BEGIN
      My_Set.Container:= NEW Element_Table;
      Reset(My_Set);
   END Initialize;



   PROCEDURE Write_To_File (
         My_Set     : IN     Set'Class;
         File_Name  : IN     Unbounded_String;
         Write_Mode : IN     Io_Tools.File_Mode) IS

      Desc : File_Descriptor;

      PACKAGE Set_File IS NEW Ada.Sequential_Io( File_Descriptor );
      USE Set_File;

      A_Set_File : Set_File.File_Type;
   BEGIN

      IF(Write_Mode = Save_In_New_File)
            THEN
         Create(A_Set_File,
            Mode => Out_File,
            Name => To_String(File_Name));
      ELSE
         Open(A_Set_File,
            Mode => Append_File,
            Name => To_String(File_Name));
      END IF;

      IF NOT Is_Open(A_Set_File)
            THEN
         RAISE Can_Not_Open_File_For_Writing;
      END IF;

      FOR I IN 0 .. (My_Set.Number_Of_Elements - 1) LOOP

         Fill_Descriptor(Desc, My_Set.Container(I));

         Write(A_Set_File,Desc);
      END LOOP;

      Close(A_Set_File);
   END Write_To_File;


   PROCEDURE Read_From_File (
         My_Set    : IN OUT Set'Class;
         File_Name : IN     Unbounded_String) IS


      Desc : File_Descriptor;

      PACKAGE Set_File IS NEW Ada.Sequential_Io(File_Descriptor);
      USE Set_File;

      A_Set_File : Set_File.File_Type;

   BEGIN

      Reset(My_Set);

      Open(A_Set_File,
         Mode => In_File,
         Name => To_String(File_Name));


      IF NOT Is_Open(A_Set_File)
            THEN
         RAISE Can_Not_Open_File_For_Reading;
      END IF;


      WHILE NOT End_Of_File(A_Set_File)
            LOOP

         Read(A_Set_File,Desc);
         Extract_Descriptor(Desc, My_Set.Container(My_Set.
               Number_Of_Elements));
         My_Set.Number_Of_Elements:=My_Set.Number_Of_Elements + 1;

      END LOOP;

      Close(A_Set_File);
   END Read_From_File;





   PROCEDURE Add (
         My_Set    : IN OUT Set;
         A_Element : IN     Element) IS
   BEGIN

      IF (My_Set.Number_Of_Elements = Element_Range(Max_Element))
            THEN
         RAISE Full_Set;
      END IF;

      My_Set.Container(My_Set.Number_Of_Elements):=A_Element;
      My_Set.Number_Of_Elements:=My_Set.Number_Of_Elements + 1;
   END Add;


   PROCEDURE Delete (
         My_Set      : IN OUT Set;
         A_Element   : IN     Element;
         Free_Object : IN     Boolean := False) IS

      Found : Boolean := False;
   BEGIN

      IF (My_Set.Number_Of_Elements = 0)
            THEN
         RAISE Empty_Set;
      END IF;

      FOR I IN  0 .. (My_Set.Number_Of_Elements - 1) LOOP
         IF (A_Element = My_Set.Container(I))
               THEN
            IF (Free_Object = True)
                  THEN
               Free(My_Set.Container(I));
            END IF;
            My_Set.Container(I):=My_Set.Container(My_Set.
               Number_Of_Elements - 1);
            My_Set.Number_Of_Elements:=My_Set.
               Number_Of_Elements - 1;
            Found:=True;
            EXIT;
         END IF;
      END LOOP;
      IF (NOT Found)
            THEN
         RAISE Element_Not_Found;
      END IF;
   END Delete;


   PROCEDURE Next_Element (
         My_Set      : IN     Set;
         My_Iterator : IN OUT Iterator) IS
   BEGIN
      IF (My_Set.Number_Of_Elements = 0)
            THEN
         RAISE Empty_Set;
      END IF;

      My_Iterator.Current:=My_Iterator.Current + 1;

   END Next_Element;


   PROCEDURE Current_Element (
         My_Set         : IN     Set;
         Return_Element :    OUT Element;
         My_Iterator    : IN     Iterator) IS
   BEGIN
      IF (My_Set.Number_Of_Elements = 0)
            THEN
         RAISE Empty_Set;
      END IF;

      Return_Element:=My_Set.Container(My_Iterator.Current);
   END Current_Element;


   PROCEDURE Get_Element_Number (
         My_Set         : IN     Set;
         Return_Element :    OUT Element;
         Position       : IN     Element_Range) IS
   BEGIN
      IF (My_Set.Number_Of_Elements < Position)
            THEN
         RAISE Invalid_Argument;
      END IF;

      Return_Element:=My_Set.Container(Position);

   END Get_Element_Number;



   FUNCTION Get_Number_Of_Elements (
         My_Set : IN     Set)
     RETURN Element_Range IS
   BEGIN
      RETURN (My_Set.Number_Of_Elements);
   END Get_Number_Of_Elements;


   FUNCTION Is_Empty (
         My_Set : IN     Set)
     RETURN Boolean IS
   BEGIN
      RETURN (My_Set.Number_Of_Elements = 0);
   END Is_Empty;


   FUNCTION Is_Last_Element (
         My_Set      : IN     Set;
         My_Iterator : IN     Iterator)
     RETURN Boolean IS
   BEGIN
      RETURN (My_Iterator.Current = (My_Set.Number_Of_Elements - 1));
   END Is_Last_Element;


   FUNCTION Is_First_Element (
         My_Set      : IN     Set;
         My_Iterator : IN     Iterator)
     RETURN Boolean IS
   BEGIN
      RETURN (My_Iterator.Current = 0);
   END Is_First_Element;


   PROCEDURE Reset_Iterator (
         My_Iterator : IN OUT Iterator) IS
   BEGIN
      My_Iterator.Current:=0;
   END Reset_Iterator;


   PROCEDURE Sort (
         My_Set : IN OUT Set;
         Order  :        Order_Function) IS

      Temp : Element;

   BEGIN
      FOR I IN 0 .. My_Set.Number_Of_Elements-1  LOOP
         FOR J IN 0 .. (My_Set.Number_Of_Elements-2) LOOP
            IF NOT Order(My_Set.Container(J), My_Set.Container(J+1))
                  THEN
               Temp:=My_Set.Container(J);
               My_Set.Container(J):=My_Set.Container(J+1);
               My_Set.Container(J+1):=Temp;
            END IF;
         END LOOP;
      END LOOP;
   END Sort;


   PROCEDURE Select_And_Copy (
         Src         : IN     Set;
         Dest        : IN OUT Set;
         Must_Select :        Select_Function) IS

      Tmp : Element;
   BEGIN
      FOR I IN 0 .. Src.Number_Of_Elements-1  LOOP
         IF Must_Select(Src.Container(I))
               THEN
            Tmp:=Copy(Src.Container(I));
            Add(Dest,Tmp);
         END IF;
      END LOOP;
   END Select_And_Copy;



   PROCEDURE Duplicate (
         Src  : IN     Set;
         Dest : IN OUT Set) IS

      Tmp : Element;
   BEGIN
      Dest.Number_Of_Elements:=Src.Number_Of_Elements;
      FOR I IN 0 .. Src.Number_Of_Elements-1  LOOP
         Tmp:=Copy(Src.Container(I));
         Dest.Container(I):=Tmp;
      END LOOP;
   END Duplicate;


   PROCEDURE Put (
         My_Set : IN     Set) IS
   BEGIN
      Put_Line("Number of elements : "
         & My_Set.Number_Of_Elements'Img);
      New_Line;
      FOR I IN 0 .. My_Set.Number_Of_Elements-1  LOOP
         Put(My_Set.Container(I));
         New_Line;
      END LOOP;
   END Put;



   PRAGMA Warnings (Off);

   PROCEDURE Extract_Descriptor (
         Desc      : IN     File_Descriptor;
         A_Element :    OUT Element) IS
   BEGIN
      NULL;
   END Extract_Descriptor;


   PROCEDURE Fill_Descriptor (
         Desc      :    OUT File_Descriptor;
         A_Element : IN     Element) IS
   BEGIN
      NULL;
   END Fill_Descriptor;


   PRAGMA Warnings (On);


END Sets;
