------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Ada.Finalization;


WITH Ada.Strings.Unbounded;
USE  Ada.Strings.Unbounded;

WITH Io_Tools;
USE Io_Tools;


GENERIC

   Max_Element : Natural := 100;

   TYPE Element IS PRIVATE;
   TYPE File_Descriptor IS PRIVATE;

   WITH PROCEDURE Put (
         A_Element : IN     Element);
   WITH PROCEDURE Free (
         A_Element : IN OUT Element);
   WITH FUNCTION Copy (
         E_Lement : IN     Element)
     RETURN Element;

PACKAGE Sets IS

   TYPE Set IS ABSTRACT TAGGED PRIVATE;
   TYPE Iterator IS PRIVATE;

   TYPE Element_Range IS NEW Natural RANGE 0 .. Max_Element;
   TYPE Set_Ptr IS ACCESS ALL Set'Class;


   Invalid_Argument : EXCEPTION;
   Element_Not_Found : EXCEPTION;
   Full_Set : EXCEPTION;
   Empty_Set : EXCEPTION;




   PROCEDURE Fill_Descriptor (
         Desc      :    OUT File_Descriptor;
         A_Element : IN     Element);


   PROCEDURE Extract_Descriptor (
         Desc      : IN     File_Descriptor;
         A_Element :    OUT Element);


   PROCEDURE Write_To_File (
         My_Set     : IN     Set'Class;
         File_Name  : IN     Unbounded_String;
         Write_Mode : IN     Io_Tools.File_Mode);


   PROCEDURE Initialize (
         My_Set : IN OUT Set);


   PROCEDURE Reset (
         My_Set      : IN OUT Set;
         Free_Object : IN     Boolean := False);


   PROCEDURE Put (
         My_Set : IN     Set);


   PROCEDURE Duplicate (
         Src  : IN     Set;
         Dest : IN OUT Set);


   TYPE Select_Function IS ACCESS FUNCTION
   (Op1 : IN Element) RETURN Boolean;

   PROCEDURE Select_And_Copy (
         Src         : IN     Set;
         Dest        : IN OUT Set;
         Must_Select :        Select_Function);


   PROCEDURE Read_From_File (
         My_Set    : IN OUT Set'Class;
         File_Name : IN     Unbounded_String);


   PROCEDURE Add (
         My_Set    : IN OUT Set;
         A_Element : IN     Element);


   PROCEDURE Delete (
         My_Set      : IN OUT Set;
         A_Element   : IN     Element;
         Free_Object : IN     Boolean := False);


   -----------------------------------------------
   -- Theses functions/procedures can be used to
   --
   -----------------------------------------------

   -- Lets go to the next element of the set
   --
   PROCEDURE Next_Element (
         My_Set      : IN     Set;
         My_Iterator : IN OUT Iterator);

   -- Get the element currently pointed by
   -- the iterator
   --
   PROCEDURE Current_Element (
         My_Set         : IN     Set;
         Return_Element :    OUT Element;
         My_Iterator    : IN     Iterator);


   PROCEDURE Get_Element_Number (
         My_Set         : IN     Set;
         Return_Element :    OUT Element;
         Position       : IN     Element_Range);


   PROCEDURE Reset_Iterator (
         My_Iterator : IN OUT Iterator);


   FUNCTION Get_Number_Of_Elements (
         My_Set : IN     Set)
     RETURN Element_Range;


   FUNCTION Is_Empty (
         My_Set : IN     Set)
     RETURN Boolean;


   FUNCTION Is_Last_Element (
         My_Set      : IN     Set;
         My_Iterator : IN     Iterator)
     RETURN Boolean;


   FUNCTION Is_First_Element (
         My_Set      : IN     Set;
         My_Iterator : IN     Iterator)
     RETURN Boolean;



   -- Sort utilitiy. The "order_function" return
   -- if op1 and op2 are not n the right order
   --

   TYPE Order_Function IS ACCESS FUNCTION
   (Op1 : IN Element;  Op2 : IN Element) RETURN Boolean;

   PROCEDURE Sort (
         My_Set : IN OUT Set;
         Order  :        Order_Function);


PRIVATE



   TYPE Element_Table IS ARRAY (Element_Range) OF Element;
   TYPE Element_Table_Ptr IS ACCESS Element_Table;


   TYPE Set IS NEW Ada.Finalization.Controlled WITH
      RECORD
         Number_Of_Elements : Element_Range     := 0;
         Container          : Element_Table_Ptr;
      END RECORD;


   TYPE Iterator IS
      RECORD
         Current : Element_Range := Element_Range'First;
      END RECORD;


   PRAGMA Inline (Current_Element);
   PRAGMA Inline (Next_Element);
   PRAGMA Inline (Is_Last_Element);



END Sets;




