------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

WITH Text_Io;
USE Text_Io;

WITH Unchecked_Deallocation;


PACKAGE BODY Sets IS




   PROCEDURE Reset (
         My_Set      : IN OUT Set;
         Free_Object : IN     Boolean := False) IS
   BEGIN
      IF(Free_Object = True)
            THEN
         Free(My_Set);
      END IF;
      My_Set.Number_Of_Elements:=0;
   END Reset;



   PROCEDURE Free (
         My_Set : IN OUT Set) IS

   BEGIN
      FOR I IN 0 .. My_Set.Number_Of_Elements-1  LOOP
         Free(My_Set.Container(I));
      END LOOP;
   END Free;


   PROCEDURE Free (
         My_Set      : IN OUT Set_Ptr;
         Free_Object : IN     Boolean := True) IS

      PROCEDURE Free_Pointer IS
      NEW Unchecked_Deallocation
         (Set'Class, Set_Ptr);

   BEGIN
      IF Free_Object THEN
         FOR I IN 0 .. My_Set.Number_Of_Elements-1  LOOP
            Free(My_Set.Container(I));
         END LOOP;
      END IF;
      Free_Pointer(My_Set);
   END Free;



   PROCEDURE Initialize (
         My_Set : IN OUT Set) IS
   BEGIN
      My_Set.Container:= NEW Element_Table;
      Reset(My_Set);
   END Initialize;


   PROCEDURE Add (
         My_Set    : IN OUT Set;
         A_Element : IN     Element) IS
   BEGIN

      IF (My_Set.Number_Of_Elements = Element_Range(Max_Element))
            THEN
         RAISE Full_Set;
      END IF;

      My_Set.Container(My_Set.Number_Of_Elements):=A_Element;
      My_Set.Number_Of_Elements:=My_Set.Number_Of_Elements + 1;
   END Add;


   PROCEDURE Delete (
         My_Set      : IN OUT Set;
         Elements    : IN     Set;
         Free_Object : IN     Boolean := False) IS

   BEGIN


      IF (My_Set.Number_Of_Elements = 0)
            THEN
         RAISE Empty_Set;
      END IF;

      FOR I IN 0 .. Elements.Number_Of_Elements-1  LOOP
         Delete(My_Set, Elements.Container(I), Free_Object);
      END LOOP;

   END Delete;



   PROCEDURE Delete (
         My_Set      : IN OUT Set;
         A_Element   : IN     Element;
         Free_Object : IN     Boolean := False) IS

      Found : Boolean := False;
   BEGIN

      IF (My_Set.Number_Of_Elements = 0)
            THEN
         RAISE Empty_Set;
      END IF;

      FOR I IN  0 .. (My_Set.Number_Of_Elements - 1) LOOP
         IF (A_Element = My_Set.Container(I))
               THEN
            IF (Free_Object = True)
                  THEN
               Free(My_Set.Container(I));
            END IF;
            My_Set.Container(I):=My_Set.Container(My_Set.
               Number_Of_Elements - 1);
            My_Set.Number_Of_Elements:=My_Set.
               Number_Of_Elements - 1;
            Found:=True;
            EXIT;
         END IF;
      END LOOP;
      IF (NOT Found)
            THEN
         RAISE Element_Not_Found;
      END IF;
   END Delete;


   PROCEDURE Next_Element (
         My_Set      : IN     Set;
         My_Iterator : IN OUT Iterator) IS
   BEGIN
      IF (My_Set.Number_Of_Elements = 0)
            THEN
         RAISE Empty_Set;
      END IF;

      My_Iterator.Current:=My_Iterator.Current + 1;

   END Next_Element;


   PROCEDURE Current_Element (
         My_Set         : IN     Set;
         Return_Element :    OUT Element;
         My_Iterator    : IN     Iterator) IS
   BEGIN
      IF (My_Set.Number_Of_Elements = 0)
            THEN
         RAISE Empty_Set;
      END IF;

      Return_Element:=My_Set.Container(My_Iterator.Current);
   END Current_Element;


   PROCEDURE Get_Element_Number (
         My_Set         : IN     Set;
         Return_Element :    OUT Element;
         Position       : IN     Element_Range) IS
   BEGIN
      IF (My_Set.Number_Of_Elements < Position)
            THEN
         RAISE Invalid_Argument;
      END IF;

      Return_Element:=My_Set.Container(Position);

   END Get_Element_Number;



   FUNCTION Get_Number_Of_Elements (
         My_Set : IN     Set)
     RETURN Element_Range IS
   BEGIN
      RETURN (My_Set.Number_Of_Elements);
   END Get_Number_Of_Elements;


   FUNCTION Is_Empty (
         My_Set : IN     Set)
     RETURN Boolean IS
   BEGIN
      RETURN (My_Set.Number_Of_Elements = 0);
   END Is_Empty;


   FUNCTION Is_Last_Element (
         My_Set      : IN     Set;
         My_Iterator : IN     Iterator)
     RETURN Boolean IS
   BEGIN
      RETURN (My_Iterator.Current = (My_Set.Number_Of_Elements - 1));
   END Is_Last_Element;


   FUNCTION Is_First_Element (
         My_Iterator : IN     Iterator)
     RETURN Boolean IS
   BEGIN
      RETURN (My_Iterator.Current = 0);
   END Is_First_Element;


   PROCEDURE Reset_Iterator (
         My_Iterator : IN OUT Iterator) IS
   BEGIN
      My_Iterator.Current:=0;
   END Reset_Iterator;


   PROCEDURE Sort (
         My_Set : IN OUT Set;
         Order  :        Order_Function) IS

      Temp : Element;

   BEGIN
      FOR I IN 0 .. My_Set.Number_Of_Elements-1  LOOP
         FOR J IN 0 .. (My_Set.Number_Of_Elements-2) LOOP
            IF NOT Order(My_Set.Container(J), My_Set.Container(J+1))
                  THEN
               Temp:=My_Set.Container(J);
               My_Set.Container(J):=My_Set.Container(J+1);
               My_Set.Container(J+1):=Temp;
            END IF;
         END LOOP;
      END LOOP;
   END Sort;


   PROCEDURE Select_And_Copy (
         Src         : IN     Set;
         Dest        : IN OUT Set;
         Must_Select :        Select_Function) IS

      Tmp : Element;
   BEGIN
      FOR I IN 0 .. Src.Number_Of_Elements-1  LOOP
         IF Must_Select(Src.Container(I))
               THEN
            Tmp:=Copy(Src.Container(I));
            Add(Dest,Tmp);
         END IF;
      END LOOP;
   END Select_And_Copy;



   PROCEDURE Duplicate (
         Src  : IN     Set;
         Dest : IN OUT Set) IS

      Tmp : Element;
   BEGIN
      Dest.Number_Of_Elements:=Src.Number_Of_Elements;
      FOR I IN 0 .. Src.Number_Of_Elements-1  LOOP
         Tmp:=Copy(Src.Container(I));
         Dest.Container(I):=Tmp;
      END LOOP;
   END Duplicate;


   PROCEDURE Put (
         My_Set : IN     Set) IS
   BEGIN
      Put_Line("Number of elements : "
         & My_Set.Number_Of_Elements'Img);
      New_Line;
      FOR I IN 0 .. My_Set.Number_Of_Elements-1  LOOP
         Put(My_Set.Container(I));
         New_Line;
      END LOOP;
   END Put;


END Sets;
