------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------


WITH Ada.Numerics.Aux;
USE Ada.Numerics.Aux;

WITH Ada.Strings.Unbounded;
USE Ada.Strings.Unbounded;
WITH Ada.Strings.Unbounded.Text_Io;
USE Ada.Strings.Unbounded.Text_Io;

WITH Convert_Unbounded_Strings;

WITH Tables;
WITH Lists;
WITH Unchecked_Deallocation;


PACKAGE Unbounded_Strings IS


   -- Some standard strings constants
   --
   Unbounded_Sp : CONSTANT Unbounded_String := To_Unbounded_String (" ");

   Unbounded_Lf : CONSTANT Unbounded_String := To_Unbounded_String ("" &
     Ascii.Lf);

   Unbounded_Ht : CONSTANT Unbounded_String := To_Unbounded_String ("" &
     Ascii.Ht);

   Empty_String : CONSTANT Unbounded_String := To_Unbounded_String ("");




   ----------------------------------------------
   -- String tables
   ----------------------------------------------
   -- Maximum number of strings composing a table of strings
   --
   Max_Strings : CONSTANT Natural := 100;


   PROCEDURE Initialize (
         St :    OUT Unbounded_String);

   PACKAGE Strings_Table_Package IS NEW Tables(Unbounded_String,
      Max_Strings,Put);
   USE Strings_Table_Package;

   SUBTYPE Strings_Range IS Strings_Table_Package.Table_Range;
   SUBTYPE Strings_Table IS Strings_Table_Package.Table;


   ------------------------------------------------------
   -- Convert unbounded_strings to standard numeric types
   ------------------------------------------------------

   PROCEDURE To_Natural IS
   NEW Convert_Unbounded_Strings(Natural,0);
   PROCEDURE To_Integer IS
   NEW Convert_Unbounded_Strings(Integer,0);
   PROCEDURE To_Boolean IS
   NEW Convert_Unbounded_Strings(Boolean,True);


   PROCEDURE To_Double (
         From : IN     Unbounded_String;
         To   :    OUT Double;
         Ok   :    OUT Boolean);


   FUNCTION To_Upper (
         S : Unbounded_String)
     RETURN Unbounded_String;

   FUNCTION To_Lower (
         S : Unbounded_String)
     RETURN Unbounded_String;


   FUNCTION To_Upper (
         S : String)
     RETURN Unbounded_String;

   FUNCTION To_Lower (
         S : String)
     RETURN Unbounded_String;




   --------------------------------------------------------------------
   -- Pointers to unbounded_string and sub programms to manipulate it
   --------------------------------------------------------------------

   TYPE Unbounded_String_Ptr IS ACCESS Unbounded_String;


   PROCEDURE Free IS
   NEW Unchecked_Deallocation
      (Unbounded_String, Unbounded_String_Ptr);


   PROCEDURE Put (
         S : Unbounded_String_Ptr);


   ------------------------------------------------------
   -- List of unbounded_string
   ------------------------------------------------------


   PACKAGE Unbounded_String_List_Package IS NEW Lists(
      Unbounded_String_Ptr, Put, Free);
   USE Unbounded_String_List_Package;
   SUBTYPE Unbounded_String_List IS
      Unbounded_String_List_Package.List;
   SUBTYPE Unbounded_String_Iterator IS
      Unbounded_String_List_Package.Iterator;





END Unbounded_Strings;
